/* Generated by Cython 3.0.0 */

/* BEGIN: Cython Metadata
{
    "distutils": {
        "depends": [
            "pyreadstat/conditional_includes.h",
            "src/readstat.h",
            "src/readstat_io_unistd.h"
        ],
        "extra_compile_args": [
            "-Ireadstat",
            "-DHAVE_ZLIB=1"
        ],
        "include_dirs": [
            "pyreadstat",
            "src",
            "src/stata",
            "src/spss",
            "src/sas",
            "."
        ],
        "libraries": [
            "m",
            "z"
        ],
        "name": "pyreadstat._readstat_parser",
        "sources": [
            "pyreadstat/_readstat_parser.pyx",
            "./src/CKHashTable.c",
            "./src/readstat_bits.c",
            "./src/readstat_convert.c",
            "./src/readstat_error.c",
            "./src/readstat_io_unistd.c",
            "./src/readstat_malloc.c",
            "./src/readstat_metadata.c",
            "./src/readstat_parser.c",
            "./src/readstat_value.c",
            "./src/readstat_variable.c",
            "./src/readstat_writer.c",
            "./src/sas/ieee.c",
            "./src/sas/readstat_sas.c",
            "./src/sas/readstat_sas7bcat_read.c",
            "./src/sas/readstat_sas7bcat_write.c",
            "./src/sas/readstat_sas7bdat_read.c",
            "./src/sas/readstat_sas7bdat_write.c",
            "./src/sas/readstat_sas_rle.c",
            "./src/sas/readstat_xport.c",
            "./src/sas/readstat_xport_parse_format.c",
            "./src/sas/readstat_xport_read.c",
            "./src/sas/readstat_xport_write.c",
            "./src/spss/readstat_por.c",
            "./src/spss/readstat_por_parse.c",
            "./src/spss/readstat_por_read.c",
            "./src/spss/readstat_por_write.c",
            "./src/spss/readstat_sav.c",
            "./src/spss/readstat_sav_compress.c",
            "./src/spss/readstat_sav_parse.c",
            "./src/spss/readstat_sav_parse_timestamp.c",
            "./src/spss/readstat_sav_read.c",
            "./src/spss/readstat_sav_write.c",
            "./src/spss/readstat_spss.c",
            "./src/spss/readstat_spss_parse.c",
            "./src/spss/readstat_zsav_compress.c",
            "./src/spss/readstat_zsav_read.c",
            "./src/spss/readstat_zsav_write.c",
            "./src/stata/readstat_dta.c",
            "./src/stata/readstat_dta_parse_timestamp.c",
            "./src/stata/readstat_dta_read.c",
            "./src/stata/readstat_dta_write.c"
        ]
    },
    "module_name": "pyreadstat._readstat_parser"
}
END: Cython Metadata */

#ifndef PY_SSIZE_T_CLEAN
#define PY_SSIZE_T_CLEAN
#endif /* PY_SSIZE_T_CLEAN */
#if defined(CYTHON_LIMITED_API) && 0
  #ifndef Py_LIMITED_API
    #if CYTHON_LIMITED_API+0 > 0x03030000
      #define Py_LIMITED_API CYTHON_LIMITED_API
    #else
      #define Py_LIMITED_API 0x03030000
    #endif
  #endif
#endif

#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02070000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03030000)
    #error Cython requires Python 2.7+ or Python 3.3+.
#else
#define CYTHON_ABI "3_0_0"
#define __PYX_ABI_MODULE_NAME "_cython_" CYTHON_ABI
#define __PYX_TYPE_MODULE_PREFIX __PYX_ABI_MODULE_NAME "."
#define CYTHON_HEX_VERSION 0x030000F0
#define CYTHON_FUTURE_DIVISION 0
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(_WIN32) && !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef HAVE_LONG_LONG
  #define HAVE_LONG_LONG
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#if defined(GRAALVM_PYTHON)
  /* For very preliminary testing purposes. Most variables are set the same as PyPy.
     The existence of this section does not imply that anything works or is even tested */
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 1
  #define CYTHON_COMPILING_IN_NOGIL 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_TYPE_SPECS
  #define CYTHON_USE_TYPE_SPECS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS (PY_MAJOR_VERSION >= 3)
  #endif
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #undef CYTHON_USE_MODULE_STATE
  #define CYTHON_USE_MODULE_STATE 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
#elif defined(PYPY_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 0
  #define CYTHON_COMPILING_IN_NOGIL 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_TYPE_SPECS
  #define CYTHON_USE_TYPE_SPECS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS (PY_MAJOR_VERSION >= 3)
  #endif
  #if PY_VERSION_HEX < 0x03090000
    #undef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #elif !defined(CYTHON_PEP489_MULTI_PHASE_INIT)
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #undef CYTHON_USE_MODULE_STATE
  #define CYTHON_USE_MODULE_STATE 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE (PY_VERSION_HEX >= 0x030400a1 && PYPY_VERSION_NUM >= 0x07030C00)
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
#elif defined(CYTHON_LIMITED_API)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 1
  #define CYTHON_COMPILING_IN_GRAAL 0
  #define CYTHON_COMPILING_IN_NOGIL 0
  #undef CYTHON_CLINE_IN_TRACEBACK
  #define CYTHON_CLINE_IN_TRACEBACK 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_TYPE_SPECS
  #define CYTHON_USE_TYPE_SPECS 1
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #endif
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_MODULE_STATE
  #define CYTHON_USE_MODULE_STATE 1
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 1
  #endif
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
#elif defined(PY_NOGIL)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 0
  #define CYTHON_COMPILING_IN_NOGIL 1
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #ifndef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 1
  #endif
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 0
  #define CYTHON_COMPILING_IN_NOGIL 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #ifndef CYTHON_USE_TYPE_SPECS
    #define CYTHON_USE_TYPE_SPECS 0
  #endif
  #ifndef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #if PY_MAJOR_VERSION < 3
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #ifndef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if PY_VERSION_HEX < 0x030300F0 || PY_VERSION_HEX >= 0x030B00A2
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #ifndef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #ifndef CYTHON_FAST_GIL
    #define CYTHON_FAST_GIL (PY_MAJOR_VERSION < 3 || PY_VERSION_HEX >= 0x03060000 && PY_VERSION_HEX < 0x030C00A6)
  #endif
  #ifndef CYTHON_METH_FASTCALL
    #define CYTHON_METH_FASTCALL (PY_VERSION_HEX >= 0x030700A1)
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL 1
  #endif
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #elif !defined(CYTHON_PEP489_MULTI_PHASE_INIT)
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #ifndef CYTHON_USE_MODULE_STATE
    #define CYTHON_USE_MODULE_STATE 0
  #endif
  #if PY_VERSION_HEX < 0x030400a1
    #undef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 0
  #elif !defined(CYTHON_USE_TP_FINALIZE)
    #define CYTHON_USE_TP_FINALIZE 1
  #endif
  #if PY_VERSION_HEX < 0x030600B1
    #undef CYTHON_USE_DICT_VERSIONS
    #define CYTHON_USE_DICT_VERSIONS 0
  #elif !defined(CYTHON_USE_DICT_VERSIONS)
    #define CYTHON_USE_DICT_VERSIONS  (PY_VERSION_HEX < 0x030C00A5)
  #endif
  #if PY_VERSION_HEX < 0x030700A3
    #undef CYTHON_USE_EXC_INFO_STACK
    #define CYTHON_USE_EXC_INFO_STACK 0
  #elif !defined(CYTHON_USE_EXC_INFO_STACK)
    #define CYTHON_USE_EXC_INFO_STACK 1
  #endif
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 1
  #endif
#endif
#if !defined(CYTHON_FAST_PYCCALL)
#define CYTHON_FAST_PYCCALL  (CYTHON_FAST_PYCALL && PY_VERSION_HEX >= 0x030600B1)
#endif
#if !defined(CYTHON_VECTORCALL)
#define CYTHON_VECTORCALL  (CYTHON_FAST_PYCCALL && PY_VERSION_HEX >= 0x030800B1)
#endif
#define CYTHON_BACKPORT_VECTORCALL (CYTHON_METH_FASTCALL && PY_VERSION_HEX < 0x030800B1)
#if CYTHON_USE_PYLONG_INTERNALS
  #if PY_MAJOR_VERSION < 3
    #include "longintrepr.h"
  #endif
  #undef SHIFT
  #undef BASE
  #undef MASK
  #ifdef SIZEOF_VOID_P
    enum { __pyx_check_sizeof_voidp = 1 / (int)(SIZEOF_VOID_P == sizeof(void*)) };
  #endif
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
  #if defined(__cplusplus)
    /* for clang __has_cpp_attribute(maybe_unused) is true even before C++17
     * but leads to warnings with -pedantic, since it is a C++17 feature */
    #if ((defined(_MSVC_LANG) && _MSVC_LANG >= 201703L) || __cplusplus >= 201703L)
      #if __has_cpp_attribute(maybe_unused)
        #define CYTHON_UNUSED [[maybe_unused]]
      #endif
    #endif
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
  #define CYTHON_MAYBE_UNUSED_VAR(x) CYTHON_UNUSED_VAR(x)
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#ifdef _MSC_VER
    #ifndef _MSC_STDINT_H_
        #if _MSC_VER < 1300
            typedef unsigned char     uint8_t;
            typedef unsigned short    uint16_t;
            typedef unsigned int      uint32_t;
        #else
            typedef unsigned __int8   uint8_t;
            typedef unsigned __int16  uint16_t;
            typedef unsigned __int32  uint32_t;
        #endif
    #endif
    #if _MSC_VER < 1300
        #ifdef _WIN64
            typedef unsigned long long  __pyx_uintptr_t;
        #else
            typedef unsigned int        __pyx_uintptr_t;
        #endif
    #else
        #ifdef _WIN64
            typedef unsigned __int64    __pyx_uintptr_t;
        #else
            typedef unsigned __int32    __pyx_uintptr_t;
        #endif
    #endif
#else
    #include <stdint.h>
    typedef uintptr_t  __pyx_uintptr_t;
#endif
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus)
    /* for clang __has_cpp_attribute(fallthrough) is true even before C++17
     * but leads to warnings with -pedantic, since it is a C++17 feature */
    #if ((defined(_MSVC_LANG) && _MSVC_LANG >= 201703L) || __cplusplus >= 201703L)
      #if __has_cpp_attribute(fallthrough)
        #define CYTHON_FALLTHROUGH [[fallthrough]]
      #endif
    #endif
    #ifndef CYTHON_FALLTHROUGH
      #if __has_cpp_attribute(clang::fallthrough)
        #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
      #elif __has_cpp_attribute(gnu::fallthrough)
        #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
      #endif
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif
#ifdef __cplusplus
  template <typename T>
  struct __PYX_IS_UNSIGNED_IMPL {static const bool value = T(0) < T(-1);};
  #define __PYX_IS_UNSIGNED(type) (__PYX_IS_UNSIGNED_IMPL<type>::value)
#else
  #define __PYX_IS_UNSIGNED(type) (((type)-1) > 0)
#endif
#if CYTHON_COMPILING_IN_PYPY == 1
  #define __PYX_NEED_TP_PRINT_SLOT  (PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x030A0000)
#else
  #define __PYX_NEED_TP_PRINT_SLOT  (PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000)
#endif
#define __PYX_REINTERPRET_FUNCION(func_pointer, other_pointer) ((func_pointer)(void(*)(void))(other_pointer))

#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #elif defined(__GNUC__)
    #define CYTHON_INLINE __inline__
  #elif defined(_MSC_VER)
    #define CYTHON_INLINE __inline
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_INLINE inline
  #else
    #define CYTHON_INLINE
  #endif
#endif

#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_DefaultClassType PyClass_Type
  #define __Pyx_PyCode_New(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
  #define __Pyx_DefaultClassType PyType_Type
#if PY_VERSION_HEX >= 0x030B00A1
    static CYTHON_INLINE PyCodeObject* __Pyx_PyCode_New(int a, int p, int k, int l, int s, int f,
                                                    PyObject *code, PyObject *c, PyObject* n, PyObject *v,
                                                    PyObject *fv, PyObject *cell, PyObject* fn,
                                                    PyObject *name, int fline, PyObject *lnos) {
        PyObject *kwds=NULL, *argcount=NULL, *posonlyargcount=NULL, *kwonlyargcount=NULL;
        PyObject *nlocals=NULL, *stacksize=NULL, *flags=NULL, *replace=NULL, *empty=NULL;
        const char *fn_cstr=NULL;
        const char *name_cstr=NULL;
        PyCodeObject *co=NULL, *result=NULL;
        PyObject *type, *value, *traceback;
        PyErr_Fetch(&type, &value, &traceback);
        if (!(kwds=PyDict_New())) goto end;
        if (!(argcount=PyLong_FromLong(a))) goto end;
        if (PyDict_SetItemString(kwds, "co_argcount", argcount) != 0) goto end;
        if (!(posonlyargcount=PyLong_FromLong(p))) goto end;
        if (PyDict_SetItemString(kwds, "co_posonlyargcount", posonlyargcount) != 0) goto end;
        if (!(kwonlyargcount=PyLong_FromLong(k))) goto end;
        if (PyDict_SetItemString(kwds, "co_kwonlyargcount", kwonlyargcount) != 0) goto end;
        if (!(nlocals=PyLong_FromLong(l))) goto end;
        if (PyDict_SetItemString(kwds, "co_nlocals", nlocals) != 0) goto end;
        if (!(stacksize=PyLong_FromLong(s))) goto end;
        if (PyDict_SetItemString(kwds, "co_stacksize", stacksize) != 0) goto end;
        if (!(flags=PyLong_FromLong(f))) goto end;
        if (PyDict_SetItemString(kwds, "co_flags", flags) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_code", code) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_consts", c) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_names", n) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_varnames", v) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_freevars", fv) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_cellvars", cell) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_linetable", lnos) != 0) goto end;
        if (!(fn_cstr=PyUnicode_AsUTF8AndSize(fn, NULL))) goto end;
        if (!(name_cstr=PyUnicode_AsUTF8AndSize(name, NULL))) goto end;
        if (!(co = PyCode_NewEmpty(fn_cstr, name_cstr, fline))) goto end;
        if (!(replace = PyObject_GetAttrString((PyObject*)co, "replace"))) goto end;
        if (!(empty = PyTuple_New(0))) goto end;
        result = (PyCodeObject*) PyObject_Call(replace, empty, kwds);
    end:
        Py_XDECREF((PyObject*) co);
        Py_XDECREF(kwds);
        Py_XDECREF(argcount);
        Py_XDECREF(posonlyargcount);
        Py_XDECREF(kwonlyargcount);
        Py_XDECREF(nlocals);
        Py_XDECREF(stacksize);
        Py_XDECREF(replace);
        Py_XDECREF(empty);
        if (type) {
            PyErr_Restore(type, value, traceback);
        }
        return result;
    }
#elif PY_VERSION_HEX >= 0x030800B2 && !CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyCode_New(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_NewWithPosOnlyArgs(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#else
  #define __Pyx_PyCode_New(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#endif
#endif
#if PY_VERSION_HEX >= 0x030900A4 || defined(Py_IS_TYPE)
  #define __Pyx_IS_TYPE(ob, type) Py_IS_TYPE(ob, type)
#else
  #define __Pyx_IS_TYPE(ob, type) (((const PyObject*)ob)->ob_type == (type))
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_Is)
  #define __Pyx_Py_Is(x, y)  Py_Is(x, y)
#else
  #define __Pyx_Py_Is(x, y) ((x) == (y))
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_IsNone)
  #define __Pyx_Py_IsNone(ob) Py_IsNone(ob)
#else
  #define __Pyx_Py_IsNone(ob) __Pyx_Py_Is((ob), Py_None)
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_IsTrue)
  #define __Pyx_Py_IsTrue(ob) Py_IsTrue(ob)
#else
  #define __Pyx_Py_IsTrue(ob) __Pyx_Py_Is((ob), Py_True)
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_IsFalse)
  #define __Pyx_Py_IsFalse(ob) Py_IsFalse(ob)
#else
  #define __Pyx_Py_IsFalse(ob) __Pyx_Py_Is((ob), Py_False)
#endif
#define __Pyx_NoneAsNull(obj)  (__Pyx_Py_IsNone(obj) ? NULL : (obj))
#if PY_VERSION_HEX >= 0x030900F0 && !CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyObject_GC_IsFinalized(o) PyObject_GC_IsFinalized(o)
#else
  #define __Pyx_PyObject_GC_IsFinalized(o) _PyGC_FINALIZED(o)
#endif
#ifndef CO_COROUTINE
  #define CO_COROUTINE 0x80
#endif
#ifndef CO_ASYNC_GENERATOR
  #define CO_ASYNC_GENERATOR 0x200
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef Py_TPFLAGS_SEQUENCE
  #define Py_TPFLAGS_SEQUENCE 0
#endif
#ifndef Py_TPFLAGS_MAPPING
  #define Py_TPFLAGS_MAPPING 0
#endif
#ifndef METH_STACKLESS
  #define METH_STACKLESS 0
#endif
#if PY_VERSION_HEX <= 0x030700A3 || !defined(METH_FASTCALL)
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #define __Pyx_PyCFunctionFast _PyCFunctionFast
  #define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
#endif
#if CYTHON_METH_FASTCALL
  #define __Pyx_METH_FASTCALL METH_FASTCALL
  #define __Pyx_PyCFunction_FastCall __Pyx_PyCFunctionFast
  #define __Pyx_PyCFunction_FastCallWithKeywords __Pyx_PyCFunctionFastWithKeywords
#else
  #define __Pyx_METH_FASTCALL METH_VARARGS
  #define __Pyx_PyCFunction_FastCall PyCFunction
  #define __Pyx_PyCFunction_FastCallWithKeywords PyCFunctionWithKeywords
#endif
#if CYTHON_VECTORCALL
  #define __pyx_vectorcallfunc vectorcallfunc
  #define __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET  PY_VECTORCALL_ARGUMENTS_OFFSET
  #define __Pyx_PyVectorcall_NARGS(n)  PyVectorcall_NARGS((size_t)(n))
#elif CYTHON_BACKPORT_VECTORCALL
  typedef PyObject *(*__pyx_vectorcallfunc)(PyObject *callable, PyObject *const *args,
                                            size_t nargsf, PyObject *kwnames);
  #define __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET  ((size_t)1 << (8 * sizeof(size_t) - 1))
  #define __Pyx_PyVectorcall_NARGS(n)  ((Py_ssize_t)(((size_t)(n)) & ~__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET))
#else
  #define __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET  0
  #define __Pyx_PyVectorcall_NARGS(n)  ((Py_ssize_t)(n))
#endif
#if PY_VERSION_HEX < 0x030900B1
  #define __Pyx_PyType_FromModuleAndSpec(m, s, b)  ((void)m, PyType_FromSpecWithBases(s, b))
  typedef PyObject *(*__Pyx_PyCMethod)(PyObject *, PyTypeObject *, PyObject *const *, size_t, PyObject *);
#else
  #define __Pyx_PyType_FromModuleAndSpec(m, s, b)  PyType_FromModuleAndSpec(m, s, b)
  #define __Pyx_PyCMethod  PyCMethod
#endif
#ifndef METH_METHOD
  #define METH_METHOD 0x200
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyThreadState_Current PyThreadState_Get()
#elif !CYTHON_FAST_THREAD_STATE
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x03060000
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#elif PY_VERSION_HEX >= 0x03000000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_Current
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
static CYTHON_INLINE void *__Pyx_PyModule_GetState(PyObject *op)
{
    void *result;
    result = PyModule_GetState(op);
    if (!result)
        Py_FatalError("Couldn't find the module state");
    return result;
}
#endif
#define __Pyx_PyObject_GetSlot(obj, name, func_ctype)  __Pyx_PyType_GetSlot(Py_TYPE(obj), name, func_ctype)
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyType_GetSlot(type, name, func_ctype)  ((func_ctype) PyType_GetSlot((type), Py_##name))
#else
  #define __Pyx_PyType_GetSlot(type, name, func_ctype)  ((type)->name)
#endif
#if PY_VERSION_HEX < 0x030700A2 && !defined(PyThread_tss_create) && !defined(Py_tss_NEEDS_INIT)
#include "pythread.h"
#define Py_tss_NEEDS_INIT 0
typedef int Py_tss_t;
static CYTHON_INLINE int PyThread_tss_create(Py_tss_t *key) {
  *key = PyThread_create_key();
  return 0;
}
static CYTHON_INLINE Py_tss_t * PyThread_tss_alloc(void) {
  Py_tss_t *key = (Py_tss_t *)PyObject_Malloc(sizeof(Py_tss_t));
  *key = Py_tss_NEEDS_INIT;
  return key;
}
static CYTHON_INLINE void PyThread_tss_free(Py_tss_t *key) {
  PyObject_Free(key);
}
static CYTHON_INLINE int PyThread_tss_is_created(Py_tss_t *key) {
  return *key != Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE void PyThread_tss_delete(Py_tss_t *key) {
  PyThread_delete_key(*key);
  *key = Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE int PyThread_tss_set(Py_tss_t *key, void *value) {
  return PyThread_set_key_value(*key, value);
}
static CYTHON_INLINE void * PyThread_tss_get(Py_tss_t *key) {
  return PyThread_get_key_value(*key);
}
#endif
#if PY_MAJOR_VERSION < 3
    #if CYTHON_COMPILING_IN_PYPY
        #if PYPY_VERSION_NUM < 0x07030600
            #if defined(__cplusplus) && __cplusplus >= 201402L
                [[deprecated("`with nogil:` inside a nogil function will not release the GIL in PyPy2 < 7.3.6")]]
            #elif defined(__GNUC__) || defined(__clang__)
                __attribute__ ((__deprecated__("`with nogil:` inside a nogil function will not release the GIL in PyPy2 < 7.3.6")))
            #elif defined(_MSC_VER)
                __declspec(deprecated("`with nogil:` inside a nogil function will not release the GIL in PyPy2 < 7.3.6"))
            #endif
            static CYTHON_INLINE int PyGILState_Check(void) {
                return 0;
            }
        #else  // PYPY_VERSION_NUM < 0x07030600
        #endif  // PYPY_VERSION_NUM < 0x07030600
    #else
        static CYTHON_INLINE int PyGILState_Check(void) {
            PyThreadState * tstate = _PyThreadState_Current;
            return tstate && (tstate == PyGILState_GetThisThreadState());
        }
    #endif
#endif
#if CYTHON_COMPILING_IN_CPYTHON || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#if PY_MAJOR_VERSION >= 3 || CYTHON_FUTURE_DIVISION
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX > 0x030600B4 && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStrWithError(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
static CYTHON_INLINE PyObject * __Pyx_PyDict_GetItemStr(PyObject *dict, PyObject *name) {
    PyObject *res = __Pyx_PyDict_GetItemStrWithError(dict, name);
    if (res == NULL) PyErr_Clear();
    return res;
}
#elif PY_MAJOR_VERSION >= 3 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07020000)
#define __Pyx_PyDict_GetItemStrWithError  PyDict_GetItemWithError
#define __Pyx_PyDict_GetItemStr           PyDict_GetItem
#else
static CYTHON_INLINE PyObject * __Pyx_PyDict_GetItemStrWithError(PyObject *dict, PyObject *name) {
#if CYTHON_COMPILING_IN_PYPY
    return PyDict_GetItem(dict, name);
#else
    PyDictEntry *ep;
    PyDictObject *mp = (PyDictObject*) dict;
    long hash = ((PyStringObject *) name)->ob_shash;
    assert(hash != -1);
    ep = (mp->ma_lookup)(mp, name, hash);
    if (ep == NULL) {
        return NULL;
    }
    return ep->me_value;
#endif
}
#define __Pyx_PyDict_GetItemStr           PyDict_GetItem
#endif
#if CYTHON_USE_TYPE_SLOTS
  #define __Pyx_PyType_GetFlags(tp)   (((PyTypeObject *)tp)->tp_flags)
  #define __Pyx_PyType_HasFeature(type, feature)  ((__Pyx_PyType_GetFlags(type) & (feature)) != 0)
  #define __Pyx_PyObject_GetIterNextFunc(obj)  (Py_TYPE(obj)->tp_iternext)
#else
  #define __Pyx_PyType_GetFlags(tp)   (PyType_GetFlags((PyTypeObject *)tp))
  #define __Pyx_PyType_HasFeature(type, feature)  PyType_HasFeature(type, feature)
  #define __Pyx_PyObject_GetIterNextFunc(obj)  PyIter_Next
#endif
#if CYTHON_USE_TYPE_SPECS && PY_VERSION_HEX >= 0x03080000
#define __Pyx_PyHeapTypeObject_GC_Del(obj)  {\
    PyTypeObject *type = Py_TYPE(obj);\
    assert(__Pyx_PyType_HasFeature(type, Py_TPFLAGS_HEAPTYPE));\
    PyObject_GC_Del(obj);\
    Py_DECREF(type);\
}
#else
#define __Pyx_PyHeapTypeObject_GC_Del(obj)  PyObject_GC_Del(obj)
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define CYTHON_PEP393_ENABLED 1
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GetLength(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_ReadChar(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((void)u, 1114111U)
  #define __Pyx_PyUnicode_KIND(u)         ((void)u, (0))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)u)
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)k, PyUnicode_ReadChar((PyObject*)(d), i))
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GetLength(u))
#elif PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #if PY_VERSION_HEX >= 0x030C0000
    #define __Pyx_PyUnicode_READY(op)       (0)
  #else
    #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                                0 : _PyUnicode_Ready((PyObject *)(op)))
  #endif
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         ((int)PyUnicode_KIND(u))
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, (Py_UCS4) ch)
  #if PY_VERSION_HEX >= 0x030C0000
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_LENGTH(u))
  #else
    #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03090000
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : ((PyCompactUnicodeObject *)(u))->wstr_length))
    #else
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
    #endif
  #endif
#else
  #define CYTHON_PEP393_ENABLED 0
  #define PyUnicode_1BYTE_KIND  1
  #define PyUnicode_2BYTE_KIND  2
  #define PyUnicode_4BYTE_KIND  4
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((sizeof(Py_UNICODE) == 2) ? 65535U : 1114111U)
  #define __Pyx_PyUnicode_KIND(u)         ((int)sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  (((void)(k)), ((Py_UNICODE*)d)[i] = (Py_UNICODE) ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #if !defined(PyUnicode_DecodeUnicodeEscape)
    #define PyUnicode_DecodeUnicodeEscape(s, size, errors)  PyUnicode_Decode(s, size, "unicode_escape", errors)
  #endif
  #if !defined(PyUnicode_Contains) || (PY_MAJOR_VERSION == 2 && PYPY_VERSION_NUM < 0x07030500)
    #undef PyUnicode_Contains
    #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
  #endif
  #if !defined(PyByteArray_Check)
    #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
  #endif
  #if !defined(PyObject_Format)
    #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
  #endif
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None || (PyString_Check(b) && !PyString_CheckExact(b)))) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None || (PyUnicode_Check(b) && !PyUnicode_CheckExact(b)))) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
#ifndef PyObject_Unicode
  #define PyObject_Unicode             PyObject_Str
#endif
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#if CYTHON_COMPILING_IN_CPYTHON
  #define __Pyx_PySequence_ListKeepNew(obj)\
    (likely(PyList_CheckExact(obj) && Py_REFCNT(obj) == 1) ? __Pyx_NewRef(obj) : PySequence_List(obj))
#else
  #define __Pyx_PySequence_ListKeepNew(obj)  PySequence_List(obj)
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        __Pyx_IS_TYPE(obj, &PySet_Type)
#endif
#if PY_VERSION_HEX >= 0x030900A4
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_SET_REFCNT(obj, refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SET_SIZE(obj, size)
#else
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_REFCNT(obj) = (refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SIZE(obj) = (size)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
#else
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define __Pyx_Py3Int_Check(op)       PyLong_Check(op)
  #define __Pyx_Py3Int_CheckExact(op)  PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#else
  #define __Pyx_Py3Int_Check(op)       (PyLong_Check(op) || PyInt_Check(op))
  #define __Pyx_Py3Int_CheckExact(op)  (PyLong_CheckExact(op) || PyInt_CheckExact(op))
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   __Pyx_PyIndex_AsHash_t
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   __Pyx_PyIndex_AsSsize_t
#endif
#if CYTHON_USE_ASYNC_SLOTS
  #if PY_VERSION_HEX >= 0x030500B1
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
    #define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
  #else
    #define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
  #endif
#else
  #define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef __Pyx_PyAsyncMethodsStruct
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
    } __Pyx_PyAsyncMethodsStruct;
#endif

#if defined(_WIN32) || defined(WIN32) || defined(MS_WINDOWS)
  #if !defined(_USE_MATH_DEFINES)
    #define _USE_MATH_DEFINES
  #endif
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif

#define __PYX_MARK_ERR_POS(f_index, lineno) \
    { __pyx_filename = __pyx_f[f_index]; (void)__pyx_filename; __pyx_lineno = lineno; (void)__pyx_lineno; __pyx_clineno = __LINE__; (void)__pyx_clineno; }
#define __PYX_ERR(f_index, lineno, Ln_error) \
    { __PYX_MARK_ERR_POS(f_index, lineno) goto Ln_error; }

#ifdef CYTHON_EXTERN_C
    #undef __PYX_EXTERN_C
    #define __PYX_EXTERN_C CYTHON_EXTERN_C
#elif defined(__PYX_EXTERN_C)
    #ifdef _MSC_VER
    #pragma message ("Please do not define the '__PYX_EXTERN_C' macro externally. Use 'CYTHON_EXTERN_C' instead.")
    #else
    #warning Please do not define the '__PYX_EXTERN_C' macro externally. Use 'CYTHON_EXTERN_C' instead.
    #endif
#else
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__pyreadstat___readstat_parser
#define __PYX_HAVE_API__pyreadstat___readstat_parser
/* Early includes */
#include <stddef.h>
#include <sys/types.h>
#include <stdint.h>
#include "readstat.h"
#include "readstat_io_unistd.h"
#include "conditional_includes.h"
#include <string.h>
#include <stdio.h>
#include "datetime.h"

    /* Backport for Python 2.x */
    #if PY_MAJOR_VERSION < 3
        #ifndef PyDateTime_DELTA_GET_DAYS
            #define PyDateTime_DELTA_GET_DAYS(o) (((PyDateTime_Delta*)o)->days)
        #endif
        #ifndef PyDateTime_DELTA_GET_SECONDS
            #define PyDateTime_DELTA_GET_SECONDS(o) (((PyDateTime_Delta*)o)->seconds)
        #endif
        #ifndef PyDateTime_DELTA_GET_MICROSECONDS
            #define PyDateTime_DELTA_GET_MICROSECONDS(o) (((PyDateTime_Delta*)o)->microseconds)
        #endif
    #endif

    /* Backport for Python < 3.6 */
    #if PY_VERSION_HEX < 0x030600a4
        #ifndef PyDateTime_TIME_GET_FOLD
            #define PyDateTime_TIME_GET_FOLD(o) ((void)(o), 0)
        #endif
        #ifndef PyDateTime_DATE_GET_FOLD
            #define PyDateTime_DATE_GET_FOLD(o) ((void)(o), 0)
        #endif
    #endif

    /* Backport for Python < 3.6 */
    #if PY_VERSION_HEX < 0x030600a4
        #define __Pyx_DateTime_DateTimeWithFold(year, month, day, hour, minute, second, microsecond, tz, fold)             ((void)(fold), PyDateTimeAPI->DateTime_FromDateAndTime(year, month, day, hour, minute, second,                 microsecond, tz, PyDateTimeAPI->DateTimeType))
        #define __Pyx_DateTime_TimeWithFold(hour, minute, second, microsecond, tz, fold)             ((void)(fold), PyDateTimeAPI->Time_FromTime(hour, minute, second, microsecond, tz, PyDateTimeAPI->TimeType))
    #else /* For Python 3.6+ so that we can pass tz */
        #define __Pyx_DateTime_DateTimeWithFold(year, month, day, hour, minute, second, microsecond, tz, fold)             PyDateTimeAPI->DateTime_FromDateAndTimeAndFold(year, month, day, hour, minute, second,                 microsecond, tz, fold, PyDateTimeAPI->DateTimeType)
        #define __Pyx_DateTime_TimeWithFold(hour, minute, second, microsecond, tz, fold)             PyDateTimeAPI->Time_FromTimeAndFold(hour, minute, second, microsecond, tz, fold, PyDateTimeAPI->TimeType)
    #endif

    /* Backport for Python < 3.7 */
    #if PY_VERSION_HEX < 0x030700b1
        #define __Pyx_TimeZone_UTC NULL
        #define __Pyx_TimeZone_FromOffsetAndName(offset, name) ((void)(offset), (void)(name), (PyObject*)NULL)
    #else
        #define __Pyx_TimeZone_UTC PyDateTime_TimeZone_UTC
        #define __Pyx_TimeZone_FromOffsetAndName(offset, name) PyTimeZone_FromOffsetAndName(offset, name)
    #endif

    /* Backport for Python < 3.10 */
    #if PY_VERSION_HEX < 0x030a00a1
        #ifndef PyDateTime_TIME_GET_TZINFO
            #define PyDateTime_TIME_GET_TZINFO(o)                 ((((PyDateTime_Time*)o)->hastzinfo) ? ((PyDateTime_Time*)o)->tzinfo : Py_None)
        #endif
        #ifndef PyDateTime_DATE_GET_TZINFO
            #define PyDateTime_DATE_GET_TZINFO(o)                 ((((PyDateTime_DateTime*)o)->hastzinfo) ? ((PyDateTime_DateTime*)o)->tzinfo : Py_None)
        #endif
    #endif
    
#include <math.h>
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_UTF8 1
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT (PY_MAJOR_VERSION >= 3 && __PYX_DEFAULT_STRING_ENCODING_IS_UTF8)
#define __PYX_DEFAULT_STRING_ENCODING "utf8"
#define __Pyx_PyObject_FromString __Pyx_PyUnicode_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
static CYTHON_INLINE int __Pyx_is_valid_index(Py_ssize_t i, Py_ssize_t limit) {
    return (size_t) i < (size_t) limit;
}
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
#define __Pyx_PyByteArray_FromString(s) PyByteArray_FromStringAndSize((const char*)s, strlen((const char*)s))
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyObject_AsWritableString(s)    ((char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
#if CYTHON_COMPILING_IN_LIMITED_API
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const wchar_t *u)
{
    const wchar_t *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#else
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const Py_UNICODE *u)
{
    const Py_UNICODE *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#endif
#define __Pyx_PyUnicode_FromOrdinal(o)       PyUnicode_FromOrdinal((int)o)
#define __Pyx_PyUnicode_FromUnicode(u)       PyUnicode_FromUnicode(u, __Pyx_Py_UNICODE_strlen(u))
#define __Pyx_PyUnicode_FromUnicodeAndLength PyUnicode_FromUnicode
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject*);
#if CYTHON_ASSUME_SAFE_MACROS
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #if PY_VERSION_HEX >= 0x030C00A7
  #ifndef _PyLong_SIGN_MASK
    #define _PyLong_SIGN_MASK 3
  #endif
  #ifndef _PyLong_NON_SIZE_BITS
    #define _PyLong_NON_SIZE_BITS 3
  #endif
  #define __Pyx_PyLong_Sign(x)  (((PyLongObject*)x)->long_value.lv_tag & _PyLong_SIGN_MASK)
  #define __Pyx_PyLong_IsNeg(x)  ((__Pyx_PyLong_Sign(x) & 2) != 0)
  #define __Pyx_PyLong_IsNonNeg(x)  (!__Pyx_PyLong_IsNeg(x))
  #define __Pyx_PyLong_IsZero(x)  (__Pyx_PyLong_Sign(x) & 1)
  #define __Pyx_PyLong_IsPos(x)  (__Pyx_PyLong_Sign(x) == 0)
  #define __Pyx_PyLong_CompactValueUnsigned(x)  (__Pyx_PyLong_Digits(x)[0])
  #define __Pyx_PyLong_DigitCount(x)  ((Py_ssize_t) (((PyLongObject*)x)->long_value.lv_tag >> _PyLong_NON_SIZE_BITS))
  #define __Pyx_PyLong_SignedDigitCount(x)\
        ((1 - (Py_ssize_t) __Pyx_PyLong_Sign(x)) * __Pyx_PyLong_DigitCount(x))
  #if defined(PyUnstable_Long_IsCompact) && defined(PyUnstable_Long_CompactValue)
    #define __Pyx_PyLong_IsCompact(x)     PyUnstable_Long_IsCompact((PyLongObject*) x)
    #define __Pyx_PyLong_CompactValue(x)  PyUnstable_Long_CompactValue((PyLongObject*) x)
  #else
    #define __Pyx_PyLong_IsCompact(x)     (((PyLongObject*)x)->long_value.lv_tag < (2 << _PyLong_NON_SIZE_BITS))
    #define __Pyx_PyLong_CompactValue(x)  ((1 - (Py_ssize_t) __Pyx_PyLong_Sign(x)) * (Py_ssize_t) __Pyx_PyLong_Digits(x)[0])
  #endif
  typedef Py_ssize_t  __Pyx_compact_pylong;
  typedef size_t  __Pyx_compact_upylong;
  #else  // Py < 3.12
  #define __Pyx_PyLong_IsNeg(x)  (Py_SIZE(x) < 0)
  #define __Pyx_PyLong_IsNonNeg(x)  (Py_SIZE(x) >= 0)
  #define __Pyx_PyLong_IsZero(x)  (Py_SIZE(x) == 0)
  #define __Pyx_PyLong_IsPos(x)  (Py_SIZE(x) > 0)
  #define __Pyx_PyLong_CompactValueUnsigned(x)  ((Py_SIZE(x) == 0) ? 0 : __Pyx_PyLong_Digits(x)[0])
  #define __Pyx_PyLong_DigitCount(x)  __Pyx_sst_abs(Py_SIZE(x))
  #define __Pyx_PyLong_SignedDigitCount(x)  Py_SIZE(x)
  #define __Pyx_PyLong_IsCompact(x)  (Py_SIZE(x) == 0 || Py_SIZE(x) == 1 || Py_SIZE(x) == -1)
  #define __Pyx_PyLong_CompactValue(x)\
        ((Py_SIZE(x) == 0) ? (sdigit) 0 : ((Py_SIZE(x) < 0) ? -(sdigit)__Pyx_PyLong_Digits(x)[0] : (sdigit)__Pyx_PyLong_Digits(x)[0]))
  typedef sdigit  __Pyx_compact_pylong;
  typedef digit  __Pyx_compact_upylong;
  #endif
  #if PY_VERSION_HEX >= 0x030C00A5
  #define __Pyx_PyLong_Digits(x)  (((PyLongObject*)x)->long_value.ob_digit)
  #else
  #define __Pyx_PyLong_Digits(x)  (((PyLongObject*)x)->ob_digit)
  #endif
#endif
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = (char) c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c) + 1);
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }

#if !CYTHON_USE_MODULE_STATE
static PyObject *__pyx_m = NULL;
#endif
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm = __FILE__;
static const char *__pyx_filename;

/* #### Code section: filename_table ### */

static const char *__pyx_f[] = {
  "pyreadstat/_readstat_parser.pyx",
  "<stringsource>",
  "datetime.pxd",
  "type.pxd",
};
/* #### Code section: utility_code_proto_before_types ### */
/* #### Code section: numeric_typedefs ### */
/* #### Code section: complex_type_declarations ### */
/* #### Code section: type_declarations ### */

/*--- Type declarations ---*/
struct __pyx_obj_10pyreadstat_16_readstat_parser_data_container;
struct __pyx_opt_args_7cpython_8datetime_time_new;
struct __pyx_opt_args_7cpython_8datetime_datetime_new;
struct __pyx_opt_args_7cpython_8datetime_timezone_new;
struct __pyx_opt_args_7cpython_8datetime_datetime_from_timestamp;

/* "cpython/datetime.pxd":292
 * # Create time object using DateTime CAPI factory function
 * # Note, there are no range checks for any of the arguments.
 * cdef inline time time_new(int hour, int minute, int second, int microsecond, object tz, int fold=0):             # <<<<<<<<<<<<<<
 *     return __Pyx_DateTime_TimeWithFold(hour, minute, second, microsecond, tz, fold)
 * 
 */
struct __pyx_opt_args_7cpython_8datetime_time_new {
  int __pyx_n;
  int fold;
};

/* "cpython/datetime.pxd":297
 * # Create datetime object using DateTime CAPI factory function.
 * # Note, there are no range checks for any of the arguments.
 * cdef inline datetime datetime_new(int year, int month, int day, int hour, int minute, int second, int microsecond, object tz, int fold=0):             # <<<<<<<<<<<<<<
 *     return __Pyx_DateTime_DateTimeWithFold(year, month, day, hour, minute, second, microsecond, tz, fold)
 * 
 */
struct __pyx_opt_args_7cpython_8datetime_datetime_new {
  int __pyx_n;
  int fold;
};

/* "cpython/datetime.pxd":306
 * 
 * # Create timedelta object using DateTime CAPI factory function.
 * cdef inline object timezone_new(object offset, object name=None):             # <<<<<<<<<<<<<<
 *     if PY_VERSION_HEX < 0x030700b1:
 *         raise RuntimeError('Time zones are not available from the C-API.')
 */
struct __pyx_opt_args_7cpython_8datetime_timezone_new {
  int __pyx_n;
  PyObject *name;
};

/* "cpython/datetime.pxd":312
 * 
 * # Create datetime object using DB API constructor.
 * cdef inline datetime datetime_from_timestamp(timestamp, tz=None):             # <<<<<<<<<<<<<<
 *     return PyDateTimeAPI.DateTime_FromTimestamp(
 *         <PyObject*>PyDateTimeAPI.DateTimeType, (timestamp, tz) if tz is not None else (timestamp,), NULL)
 */
struct __pyx_opt_args_7cpython_8datetime_datetime_from_timestamp {
  int __pyx_n;
  PyObject *tz;
};

/* "pyreadstat/_readstat_parser.pxd":22
 * 
 * # Definitions of enum types
 * ctypedef enum py_file_extension:             # <<<<<<<<<<<<<<
 *     FILE_EXT_SAV
 *     FILE_EXT_SAS7BDAT
 */
enum __pyx_t_10pyreadstat_16_readstat_parser_py_file_extension {
  __pyx_e_10pyreadstat_16_readstat_parser_FILE_EXT_SAV,
  __pyx_e_10pyreadstat_16_readstat_parser_FILE_EXT_SAS7BDAT,
  __pyx_e_10pyreadstat_16_readstat_parser_FILE_EXT_DTA,
  __pyx_e_10pyreadstat_16_readstat_parser_FILE_EXT_XPORT,
  __pyx_e_10pyreadstat_16_readstat_parser_FILE_EXT_POR,
  __pyx_e_10pyreadstat_16_readstat_parser_FILE_EXT_SAS7BCAT
};
typedef enum __pyx_t_10pyreadstat_16_readstat_parser_py_file_extension __pyx_t_10pyreadstat_16_readstat_parser_py_file_extension;

/* "pyreadstat/_readstat_parser.pxd":30
 *     FILE_EXT_SAS7BCAT
 * 
 * ctypedef enum py_file_format:             # <<<<<<<<<<<<<<
 *     FILE_FORMAT_SAS
 *     FILE_FORMAT_SPSS
 */
enum __pyx_t_10pyreadstat_16_readstat_parser_py_file_format {
  __pyx_e_10pyreadstat_16_readstat_parser_FILE_FORMAT_SAS,
  __pyx_e_10pyreadstat_16_readstat_parser_FILE_FORMAT_SPSS,
  __pyx_e_10pyreadstat_16_readstat_parser_FILE_FORMAT_STATA
};
typedef enum __pyx_t_10pyreadstat_16_readstat_parser_py_file_format __pyx_t_10pyreadstat_16_readstat_parser_py_file_format;

/* "pyreadstat/_readstat_parser.pxd":35
 *     FILE_FORMAT_STATA
 * 
 * ctypedef enum py_datetime_format:             # <<<<<<<<<<<<<<
 *     DATE_FORMAT_NOTADATE
 *     DATE_FORMAT_DATE
 */
enum __pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format {
  __pyx_e_10pyreadstat_16_readstat_parser_DATE_FORMAT_NOTADATE,
  __pyx_e_10pyreadstat_16_readstat_parser_DATE_FORMAT_DATE,
  __pyx_e_10pyreadstat_16_readstat_parser_DATE_FORMAT_DATETIME,
  __pyx_e_10pyreadstat_16_readstat_parser_DATE_FORMAT_TIME
};
typedef enum __pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format __pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format;

/* "pyreadstat/_readstat_parser.pxd":41
 *     DATE_FORMAT_TIME
 * 
 * ctypedef enum py_variable_format:             # <<<<<<<<<<<<<<
 *     VAR_FORMAT_STRING
 *     VAR_FORMAT_LONG
 */
enum __pyx_t_10pyreadstat_16_readstat_parser_py_variable_format {
  __pyx_e_10pyreadstat_16_readstat_parser_VAR_FORMAT_STRING,
  __pyx_e_10pyreadstat_16_readstat_parser_VAR_FORMAT_LONG,
  __pyx_e_10pyreadstat_16_readstat_parser_VAR_FORMAT_FLOAT,
  __pyx_e_10pyreadstat_16_readstat_parser_VAR_FORMAT_MISSING
};
typedef enum __pyx_t_10pyreadstat_16_readstat_parser_py_variable_format __pyx_t_10pyreadstat_16_readstat_parser_py_variable_format;

/* "pyreadstat/_readstat_parser.pxd":49
 * # Definitions of extension types
 * 
 * cdef class data_container:             # <<<<<<<<<<<<<<
 *     """
 *     This extension type holds all the data we need to get from our file from
 */
struct __pyx_obj_10pyreadstat_16_readstat_parser_data_container {
  PyObject_HEAD
  int n_obs;
  int n_vars;
  int max_n_obs;
  PyObject *col_data;
  PyObject *col_data_len;
  PyObject *col_names;
  PyObject *col_labels;
  PyObject *col_dtypes;
  PyObject *col_numpy_dtypes;
  PyObject *col_dtypes_isobject;
  PyObject *col_dytpes_isfloat;
  PyObject *col_formats;
  PyObject *col_formats_original;
  PyObject *origin;
  __pyx_t_10pyreadstat_16_readstat_parser_py_file_format file_format;
  int is_unkown_number_rows;
  PyObject *file_label;
  PyObject *file_encoding;
  int metaonly;
  int dates_as_pandas;
  PyObject *label_to_var_name;
  PyObject *labels_raw;
  PyObject *notes;
  PyObject *user_encoding;
  PyObject *table_name;
  int filter_cols;
  PyObject *use_cols;
  int usernan;
  PyObject *missing_ranges;
  PyObject *missing_user_values;
  PyObject *variable_storage_width;
  PyObject *variable_display_width;
  PyObject *variable_alignment;
  PyObject *variable_measure;
  int no_datetime_conversion;
  int ctime;
  int mtime;
};

/* #### Code section: utility_code_proto ### */

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, Py_ssize_t);
    void (*DECREF)(void*, PyObject*, Py_ssize_t);
    void (*GOTREF)(void*, PyObject*, Py_ssize_t);
    void (*GIVEREF)(void*, PyObject*, Py_ssize_t);
    void* (*SetupContext)(const char*, Py_ssize_t, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), (__LINE__), (__FILE__));\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), (__LINE__), (__FILE__));\
          }
  #define __Pyx_RefNannyFinishContextNogil() {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __Pyx_RefNannyFinishContext();\
              PyGILState_Release(__pyx_gilstate_save);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), (__LINE__), (__FILE__))
  #define __Pyx_RefNannyFinishContextNogil() __Pyx_RefNannyFinishContext()
#endif
  #define __Pyx_RefNannyFinishContextNogil() {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __Pyx_RefNannyFinishContext();\
              PyGILState_Release(__pyx_gilstate_save);\
          }
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_XINCREF(r)  do { if((r) == NULL); else {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) == NULL); else {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) == NULL); else {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) == NULL); else {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContextNogil()
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_Py_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; Py_XDECREF(tmp);\
    } while (0)
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyErrExceptionMatches.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* PyThreadStateGet.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#if PY_VERSION_HEX >= 0x030C00A6
#define __Pyx_PyErr_Occurred()  (__pyx_tstate->current_exception != NULL)
#define __Pyx_PyErr_CurrentExceptionType()  (__pyx_tstate->current_exception ? (PyObject*) Py_TYPE(__pyx_tstate->current_exception) : (PyObject*) NULL)
#else
#define __Pyx_PyErr_Occurred()  (__pyx_tstate->curexc_type != NULL)
#define __Pyx_PyErr_CurrentExceptionType()  (__pyx_tstate->curexc_type)
#endif
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  (PyErr_Occurred() != NULL)
#define __Pyx_PyErr_CurrentExceptionType()  PyErr_Occurred()
#endif

/* PyErrFetchRestore.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A6
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* PyObjectGetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* PyObjectGetAttrStrNoError.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name);

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* ExtTypeTest.proto */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type);

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* RaiseException.proto */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* TupleAndListFromArray.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyList_FromArray(PyObject *const *src, Py_ssize_t n);
static CYTHON_INLINE PyObject* __Pyx_PyTuple_FromArray(PyObject *const *src, Py_ssize_t n);
#endif

/* IncludeStringH.proto */
#include <string.h>

/* BytesEquals.proto */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals);

/* UnicodeEquals.proto */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals);

/* fastcall.proto */
#define __Pyx_Arg_VARARGS(args, i) PyTuple_GET_ITEM(args, i)
#define __Pyx_NumKwargs_VARARGS(kwds) PyDict_Size(kwds)
#define __Pyx_KwValues_VARARGS(args, nargs) NULL
#define __Pyx_GetKwValue_VARARGS(kw, kwvalues, s) __Pyx_PyDict_GetItemStrWithError(kw, s)
#define __Pyx_KwargsAsDict_VARARGS(kw, kwvalues) PyDict_Copy(kw)
#if CYTHON_METH_FASTCALL
    #define __Pyx_Arg_FASTCALL(args, i) args[i]
    #define __Pyx_NumKwargs_FASTCALL(kwds) PyTuple_GET_SIZE(kwds)
    #define __Pyx_KwValues_FASTCALL(args, nargs) ((args) + (nargs))
    static CYTHON_INLINE PyObject * __Pyx_GetKwValue_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues, PyObject *s);
    #define __Pyx_KwargsAsDict_FASTCALL(kw, kwvalues) _PyStack_AsDict(kwvalues, kw)
#else
    #define __Pyx_Arg_FASTCALL __Pyx_Arg_VARARGS
    #define __Pyx_NumKwargs_FASTCALL __Pyx_NumKwargs_VARARGS
    #define __Pyx_KwValues_FASTCALL __Pyx_KwValues_VARARGS
    #define __Pyx_GetKwValue_FASTCALL __Pyx_GetKwValue_VARARGS
    #define __Pyx_KwargsAsDict_FASTCALL __Pyx_KwargsAsDict_VARARGS
#endif
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_ArgsSlice_VARARGS(args, start, stop) __Pyx_PyTuple_FromArray(&__Pyx_Arg_VARARGS(args, start), stop - start)
#define __Pyx_ArgsSlice_FASTCALL(args, start, stop) __Pyx_PyTuple_FromArray(&__Pyx_Arg_FASTCALL(args, start), stop - start)
#else
#define __Pyx_ArgsSlice_VARARGS(args, start, stop) PyTuple_GetSlice(args, start, stop)
#define __Pyx_ArgsSlice_FASTCALL(args, start, stop) PyTuple_GetSlice(args, start, stop)
#endif

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* KeywordStringCheck.proto */
static int __Pyx_CheckKeywordStrings(PyObject *kw, const char* function_name, int kw_allowed);

/* RaiseDoubleKeywords.proto */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywords.proto */
static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject *const *kwvalues,
    PyObject **argnames[],
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args,
    const char* function_name);

/* PyObjectSetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
#define __Pyx_PyObject_DelAttrStr(o,n) __Pyx_PyObject_SetAttrStr(o, n, NULL)
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value);
#else
#define __Pyx_PyObject_DelAttrStr(o,n)   PyObject_DelAttr(o,n)
#define __Pyx_PyObject_SetAttrStr(o,n,v) PyObject_SetAttr(o,n,v)
#endif

/* PySequenceContains.proto */
static CYTHON_INLINE int __Pyx_PySequence_ContainsTF(PyObject* item, PyObject* seq, int eq) {
    int result = PySequence_Contains(seq, item);
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

/* ModFloat[double].proto */
static CYTHON_INLINE double __Pyx_mod_double(double, double);

/* PyFunctionFastCall.proto */
#if CYTHON_FAST_PYCALL
#if !CYTHON_VECTORCALL
#define __Pyx_PyFunction_FastCall(func, args, nargs)\
    __Pyx_PyFunction_FastCallDict((func), (args), (nargs), NULL)
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs);
#endif
#define __Pyx_BUILD_ASSERT_EXPR(cond)\
    (sizeof(char [1 - 2*!(cond)]) - 1)
#ifndef Py_MEMBER_SIZE
#define Py_MEMBER_SIZE(type, member) sizeof(((type *)0)->member)
#endif
#if !CYTHON_VECTORCALL
#if PY_VERSION_HEX >= 0x03080000
  #include "frameobject.h"
#if PY_VERSION_HEX >= 0x030b00a6
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
  #define __Pxy_PyFrame_Initialize_Offsets()
  #define __Pyx_PyFrame_GetLocalsplus(frame)  ((frame)->f_localsplus)
#else
  static size_t __pyx_pyframe_localsplus_offset = 0;
  #include "frameobject.h"
  #define __Pxy_PyFrame_Initialize_Offsets()\
    ((void)__Pyx_BUILD_ASSERT_EXPR(sizeof(PyFrameObject) == offsetof(PyFrameObject, f_localsplus) + Py_MEMBER_SIZE(PyFrameObject, f_localsplus)),\
     (void)(__pyx_pyframe_localsplus_offset = ((size_t)PyFrame_Type.tp_basicsize) - Py_MEMBER_SIZE(PyFrameObject, f_localsplus)))
  #define __Pyx_PyFrame_GetLocalsplus(frame)\
    (assert(__pyx_pyframe_localsplus_offset), (PyObject **)(((char *)(frame)) + __pyx_pyframe_localsplus_offset))
#endif
#endif
#endif

/* PyObjectCallMethO.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectFastCall.proto */
#define __Pyx_PyObject_FastCall(func, args, nargs)  __Pyx_PyObject_FastCallDict(func, args, (size_t)(nargs), NULL)
static CYTHON_INLINE PyObject* __Pyx_PyObject_FastCallDict(PyObject *func, PyObject **args, size_t nargs, PyObject *kwargs);

/* GetItemInt.proto */
#define __Pyx_GetItemInt(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Fast(o, (Py_ssize_t)i, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL) :\
               __Pyx_GetItemInt_Generic(o, to_py_func(i))))
#define __Pyx_GetItemInt_List(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_List_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
#define __Pyx_GetItemInt_Tuple(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Tuple_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "tuple index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i,
                                                     int is_list, int wraparound, int boundscheck);

/* PyDictVersioning.proto */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
#define __PYX_DICT_VERSION_INIT  ((PY_UINT64_T) -1)
#define __PYX_GET_DICT_VERSION(dict)  (((PyDictObject*)(dict))->ma_version_tag)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)\
    (version_var) = __PYX_GET_DICT_VERSION(dict);\
    (cache_var) = (value);
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP) {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    if (likely(__PYX_GET_DICT_VERSION(DICT) == __pyx_dict_version)) {\
        (VAR) = __pyx_dict_cached_value;\
    } else {\
        (VAR) = __pyx_dict_cached_value = (LOOKUP);\
        __pyx_dict_version = __PYX_GET_DICT_VERSION(DICT);\
    }\
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj);
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj);
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version);
#else
#define __PYX_GET_DICT_VERSION(dict)  (0)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP)  (VAR) = (LOOKUP);
#endif

/* GetModuleGlobalName.proto */
#if CYTHON_USE_DICT_VERSIONS
#define __Pyx_GetModuleGlobalName(var, name)  do {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    (var) = (likely(__pyx_dict_version == __PYX_GET_DICT_VERSION(__pyx_d))) ?\
        (likely(__pyx_dict_cached_value) ? __Pyx_NewRef(__pyx_dict_cached_value) : __Pyx_GetBuiltinName(name)) :\
        __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
} while(0)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  do {\
    PY_UINT64_T __pyx_dict_version;\
    PyObject *__pyx_dict_cached_value;\
    (var) = __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
} while(0)
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value);
#else
#define __Pyx_GetModuleGlobalName(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name);
#endif

/* RaiseUnboundLocalError.proto */
static CYTHON_INLINE void __Pyx_RaiseUnboundLocalError(const char *varname);

/* UnicodeConcatInPlace.proto */
# if CYTHON_COMPILING_IN_CPYTHON && PY_MAJOR_VERSION >= 3
    #if CYTHON_REFNANNY
        #define __Pyx_PyUnicode_ConcatInPlace(left, right) __Pyx_PyUnicode_ConcatInPlaceImpl(&left, right, __pyx_refnanny)
    #else
        #define __Pyx_PyUnicode_ConcatInPlace(left, right) __Pyx_PyUnicode_ConcatInPlaceImpl(&left, right)
    #endif
    static CYTHON_INLINE PyObject *__Pyx_PyUnicode_ConcatInPlaceImpl(PyObject **p_left, PyObject *right
        #if CYTHON_REFNANNY
        , void* __pyx_refnanny
        #endif
    );
#else
#define __Pyx_PyUnicode_ConcatInPlace __Pyx_PyUnicode_Concat
#endif
#define __Pyx_PyUnicode_ConcatInPlaceSafe(left, right) ((unlikely((left) == Py_None) || unlikely((right) == Py_None)) ?\
    PyNumber_InPlaceAdd(left, right) : __Pyx_PyUnicode_ConcatInPlace(left, right))

/* StrConcatInPlace.proto */
#if PY_MAJOR_VERSION >= 3
    #define __Pyx_PyStr_Concat __Pyx_PyUnicode_Concat
    #define __Pyx_PyStr_ConcatInPlace __Pyx_PyUnicode_ConcatInPlace
#else
    #define __Pyx_PyStr_Concat PyNumber_Add
    #define __Pyx_PyStr_ConcatInPlace PyNumber_InPlaceAdd
#endif
#define __Pyx_PyStr_ConcatSafe(a, b) ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
    PyNumber_Add(a, b) : __Pyx_PyStr_Concat(a, b))
#define __Pyx_PyStr_ConcatInPlaceSafe(a, b) ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
    PyNumber_InPlaceAdd(a, b) : __Pyx_PyStr_ConcatInPlace(a, b))

/* PyObject_Str.proto */
#define __Pyx_PyObject_Str(obj)\
    (likely(PyString_CheckExact(obj)) ? __Pyx_NewRef(obj) : PyObject_Str(obj))

/* RaiseUnexpectedTypeError.proto */
static int __Pyx_RaiseUnexpectedTypeError(const char *expected, PyObject *obj);

/* ListAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_PyList_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len) & likely(len > (L->allocated >> 1))) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        __Pyx_SET_SIZE(list, len + 1);
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_PyList_Append(L,x) PyList_Append(L,x)
#endif

/* DictGetItem.proto */
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key);
#define __Pyx_PyObject_Dict_GetItem(obj, name)\
    (likely(PyDict_CheckExact(obj)) ?\
     __Pyx_PyDict_GetItem(obj, name) : PyObject_GetItem(obj, name))
#else
#define __Pyx_PyDict_GetItem(d, key) PyObject_GetItem(d, key)
#define __Pyx_PyObject_Dict_GetItem(obj, name)  PyObject_GetItem(obj, name)
#endif

/* SetItemInt.proto */
#define __Pyx_SetItemInt(o, i, v, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_SetItemInt_Fast(o, (Py_ssize_t)i, v, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list assignment index out of range"), -1) :\
               __Pyx_SetItemInt_Generic(o, to_py_func(i), v)))
static int __Pyx_SetItemInt_Generic(PyObject *o, PyObject *j, PyObject *v);
static CYTHON_INLINE int __Pyx_SetItemInt_Fast(PyObject *o, Py_ssize_t i, PyObject *v,
                                               int is_list, int wraparound, int boundscheck);

/* SliceObject.proto */
#define __Pyx_PyObject_DelSlice(obj, cstart, cstop, py_start, py_stop, py_slice, has_cstart, has_cstop, wraparound)\
    __Pyx_PyObject_SetSlice(obj, (PyObject*)NULL, cstart, cstop, py_start, py_stop, py_slice, has_cstart, has_cstop, wraparound)
static CYTHON_INLINE int __Pyx_PyObject_SetSlice(
        PyObject* obj, PyObject* value, Py_ssize_t cstart, Py_ssize_t cstop,
        PyObject** py_start, PyObject** py_stop, PyObject** py_slice,
        int has_cstart, int has_cstop, int wraparound);

/* PyObjectCallOneArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* dict_getitem_default.proto */
static PyObject* __Pyx_PyDict_GetItemDefault(PyObject* d, PyObject* key, PyObject* default_value);

/* UnpackUnboundCMethod.proto */
typedef struct {
    PyObject *type;
    PyObject **method_name;
    PyCFunction func;
    PyObject *method;
    int flag;
} __Pyx_CachedCFunction;

/* CallUnboundCMethod1.proto */
static PyObject* __Pyx__CallUnboundCMethod1(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg);
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_CallUnboundCMethod1(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg);
#else
#define __Pyx_CallUnboundCMethod1(cfunc, self, arg)  __Pyx__CallUnboundCMethod1(cfunc, self, arg)
#endif

/* CallUnboundCMethod2.proto */
static PyObject* __Pyx__CallUnboundCMethod2(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg1, PyObject* arg2);
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030600B1
static CYTHON_INLINE PyObject *__Pyx_CallUnboundCMethod2(__Pyx_CachedCFunction *cfunc, PyObject *self, PyObject *arg1, PyObject *arg2);
#else
#define __Pyx_CallUnboundCMethod2(cfunc, self, arg1, arg2)  __Pyx__CallUnboundCMethod2(cfunc, self, arg1, arg2)
#endif

/* PyObjectCall2Args.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2);

/* PyObjectGetMethod.proto */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method);

/* PyObjectCallMethod1.proto */
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg);

/* append.proto */
static CYTHON_INLINE int __Pyx_PyObject_Append(PyObject* L, PyObject* x);

/* StrEquals.proto */
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyString_Equals __Pyx_PyUnicode_Equals
#else
#define __Pyx_PyString_Equals __Pyx_PyBytes_Equals
#endif

/* SliceObject.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetSlice(
        PyObject* obj, Py_ssize_t cstart, Py_ssize_t cstop,
        PyObject** py_start, PyObject** py_stop, PyObject** py_slice,
        int has_cstart, int has_cstop, int wraparound);

/* RaiseTooManyValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected);

/* RaiseNeedMoreValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index);

/* IterFinish.proto */
static CYTHON_INLINE int __Pyx_IterFinish(void);

/* UnpackItemEndCheck.proto */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected);

/* ObjectGetItem.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject *key);
#else
#define __Pyx_PyObject_GetItem(obj, key)  PyObject_GetItem(obj, key)
#endif

/* py_dict_items.proto */
static CYTHON_INLINE PyObject* __Pyx_PyDict_Items(PyObject* d);

/* CallUnboundCMethod0.proto */
static PyObject* __Pyx__CallUnboundCMethod0(__Pyx_CachedCFunction* cfunc, PyObject* self);
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_CallUnboundCMethod0(cfunc, self)\
    (likely((cfunc)->func) ?\
        (likely((cfunc)->flag == METH_NOARGS) ?  (*((cfunc)->func))(self, NULL) :\
         (PY_VERSION_HEX >= 0x030600B1 && likely((cfunc)->flag == METH_FASTCALL) ?\
            (PY_VERSION_HEX >= 0x030700A0 ?\
                (*(__Pyx_PyCFunctionFast)(void*)(PyCFunction)(cfunc)->func)(self, &__pyx_empty_tuple, 0) :\
                (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)(cfunc)->func)(self, &__pyx_empty_tuple, 0, NULL)) :\
          (PY_VERSION_HEX >= 0x030700A0 && (cfunc)->flag == (METH_FASTCALL | METH_KEYWORDS) ?\
            (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)(cfunc)->func)(self, &__pyx_empty_tuple, 0, NULL) :\
            (likely((cfunc)->flag == (METH_VARARGS | METH_KEYWORDS)) ?  ((*(PyCFunctionWithKeywords)(void*)(PyCFunction)(cfunc)->func)(self, __pyx_empty_tuple, NULL)) :\
               ((cfunc)->flag == METH_VARARGS ?  (*((cfunc)->func))(self, __pyx_empty_tuple) :\
               __Pyx__CallUnboundCMethod0(cfunc, self)))))) :\
        __Pyx__CallUnboundCMethod0(cfunc, self))
#else
#define __Pyx_CallUnboundCMethod0(cfunc, self)  __Pyx__CallUnboundCMethod0(cfunc, self)
#endif

/* GetAttr.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *, PyObject *);

/* HasAttr.proto */
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *, PyObject *);

/* GetTopmostException.proto */
#if CYTHON_USE_EXC_INFO_STACK && CYTHON_FAST_THREAD_STATE
static _PyErr_StackItem * __Pyx_PyErr_GetTopmostException(PyThreadState *tstate);
#endif

/* SaveResetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSave(type, value, tb)  __Pyx__ExceptionSave(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#define __Pyx_ExceptionReset(type, value, tb)  __Pyx__ExceptionReset(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
#else
#define __Pyx_ExceptionSave(type, value, tb)   PyErr_GetExcInfo(type, value, tb)
#define __Pyx_ExceptionReset(type, value, tb)  PyErr_SetExcInfo(type, value, tb)
#endif

/* GetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_GetException(type, value, tb)  __Pyx__GetException(__pyx_tstate, type, value, tb)
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* pyfrozenset_new.proto */
static CYTHON_INLINE PyObject* __Pyx_PyFrozenSet_New(PyObject* it);

/* PySetContains.proto */
static CYTHON_INLINE int __Pyx_PySet_ContainsTF(PyObject* key, PyObject* set, int eq);

/* ListExtend.proto */
static CYTHON_INLINE int __Pyx_PyList_Extend(PyObject* L, PyObject* v) {
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject* none = _PyList_Extend((PyListObject*)L, v);
    if (unlikely(!none))
        return -1;
    Py_DECREF(none);
    return 0;
#else
    return PyList_SetSlice(L, PY_SSIZE_T_MAX, PY_SSIZE_T_MAX, v);
#endif
}

/* PyObjectCallNoArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func);

/* IncludeStructmemberH.proto */
#include <structmember.h>

/* FixUpExtensionType.proto */
#if CYTHON_USE_TYPE_SPECS
static int __Pyx_fix_up_extension_type_from_spec(PyType_Spec *spec, PyTypeObject *type);
#endif

/* PyObjectCallMethod0.proto */
static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name);

/* ValidateBasesTuple.proto */
#if CYTHON_COMPILING_IN_CPYTHON || CYTHON_COMPILING_IN_LIMITED_API || CYTHON_USE_TYPE_SPECS
static int __Pyx_validate_bases_tuple(const char *type_name, Py_ssize_t dictoffset, PyObject *bases);
#endif

/* PyType_Ready.proto */
CYTHON_UNUSED static int __Pyx_PyType_Ready(PyTypeObject *t);

/* PyObject_GenericGetAttrNoDict.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttrNoDict PyObject_GenericGetAttr
#endif

/* PyObject_GenericGetAttr.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttr PyObject_GenericGetAttr
#endif

/* SetupReduce.proto */
#if !CYTHON_COMPILING_IN_LIMITED_API
static int __Pyx_setup_reduce(PyObject* type_obj);
#endif

/* TypeImport.proto */
#ifndef __PYX_HAVE_RT_ImportType_proto_3_0_0
#define __PYX_HAVE_RT_ImportType_proto_3_0_0
#if defined (__STDC_VERSION__) && __STDC_VERSION__ >= 201112L
#include <stdalign.h>
#endif
#if (defined (__STDC_VERSION__) && __STDC_VERSION__ >= 201112L) || __cplusplus >= 201103L
#define __PYX_GET_STRUCT_ALIGNMENT_3_0_0(s) alignof(s)
#else
#define __PYX_GET_STRUCT_ALIGNMENT_3_0_0(s) sizeof(void*)
#endif
enum __Pyx_ImportType_CheckSize_3_0_0 {
   __Pyx_ImportType_CheckSize_Error_3_0_0 = 0,
   __Pyx_ImportType_CheckSize_Warn_3_0_0 = 1,
   __Pyx_ImportType_CheckSize_Ignore_3_0_0 = 2
};
static PyTypeObject *__Pyx_ImportType_3_0_0(PyObject* module, const char *module_name, const char *class_name, size_t size, size_t alignment, enum __Pyx_ImportType_CheckSize_3_0_0 check_size);
#endif

/* Import.proto */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level);

/* ImportFrom.proto */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name);

/* ImportDottedModule.proto */
static PyObject *__Pyx_ImportDottedModule(PyObject *name, PyObject *parts_tuple);
#if PY_MAJOR_VERSION >= 3
static PyObject *__Pyx_ImportDottedModule_WalkParts(PyObject *module, PyObject *name, PyObject *parts_tuple);
#endif

/* ImportDottedModuleRelFirst.proto */
static PyObject *__Pyx_ImportDottedModuleRelFirst(PyObject *name, PyObject *parts_tuple);

/* FetchSharedCythonModule.proto */
static PyObject *__Pyx_FetchSharedCythonABIModule(void);

/* FetchCommonType.proto */
#if !CYTHON_USE_TYPE_SPECS
static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type);
#else
static PyTypeObject* __Pyx_FetchCommonTypeFromSpec(PyObject *module, PyType_Spec *spec, PyObject *bases);
#endif

/* PyMethodNew.proto */
#if PY_MAJOR_VERSION >= 3
static PyObject *__Pyx_PyMethod_New(PyObject *func, PyObject *self, PyObject *typ) {
    CYTHON_UNUSED_VAR(typ);
    if (!self)
        return __Pyx_NewRef(func);
    return PyMethod_New(func, self);
}
#else
    #define __Pyx_PyMethod_New PyMethod_New
#endif

/* PyVectorcallFastCallDict.proto */
#if CYTHON_METH_FASTCALL
static CYTHON_INLINE PyObject *__Pyx_PyVectorcall_FastCallDict(PyObject *func, __pyx_vectorcallfunc vc, PyObject *const *args, size_t nargs, PyObject *kw);
#endif

/* CythonFunctionShared.proto */
#define __Pyx_CyFunction_USED
#define __Pyx_CYFUNCTION_STATICMETHOD  0x01
#define __Pyx_CYFUNCTION_CLASSMETHOD   0x02
#define __Pyx_CYFUNCTION_CCLASS        0x04
#define __Pyx_CYFUNCTION_COROUTINE     0x08
#define __Pyx_CyFunction_GetClosure(f)\
    (((__pyx_CyFunctionObject *) (f))->func_closure)
#if PY_VERSION_HEX < 0x030900B1
  #define __Pyx_CyFunction_GetClassObj(f)\
      (((__pyx_CyFunctionObject *) (f))->func_classobj)
#else
  #define __Pyx_CyFunction_GetClassObj(f)\
      ((PyObject*) ((PyCMethodObject *) (f))->mm_class)
#endif
#define __Pyx_CyFunction_SetClassObj(f, classobj)\
    __Pyx__CyFunction_SetClassObj((__pyx_CyFunctionObject *) (f), (classobj))
#define __Pyx_CyFunction_Defaults(type, f)\
    ((type *)(((__pyx_CyFunctionObject *) (f))->defaults))
#define __Pyx_CyFunction_SetDefaultsGetter(f, g)\
    ((__pyx_CyFunctionObject *) (f))->defaults_getter = (g)
typedef struct {
#if PY_VERSION_HEX < 0x030900B1
    PyCFunctionObject func;
#else
    PyCMethodObject func;
#endif
#if CYTHON_BACKPORT_VECTORCALL
    __pyx_vectorcallfunc func_vectorcall;
#endif
#if PY_VERSION_HEX < 0x030500A0
    PyObject *func_weakreflist;
#endif
    PyObject *func_dict;
    PyObject *func_name;
    PyObject *func_qualname;
    PyObject *func_doc;
    PyObject *func_globals;
    PyObject *func_code;
    PyObject *func_closure;
#if PY_VERSION_HEX < 0x030900B1
    PyObject *func_classobj;
#endif
    void *defaults;
    int defaults_pyobjects;
    size_t defaults_size;  // used by FusedFunction for copying defaults
    int flags;
    PyObject *defaults_tuple;
    PyObject *defaults_kwdict;
    PyObject *(*defaults_getter)(PyObject *);
    PyObject *func_annotations;
    PyObject *func_is_coroutine;
} __pyx_CyFunctionObject;
#define __Pyx_CyFunction_Check(obj)  __Pyx_TypeCheck(obj, __pyx_CyFunctionType)
#define __Pyx_IsCyOrPyCFunction(obj)  __Pyx_TypeCheck2(obj, __pyx_CyFunctionType, &PyCFunction_Type)
#define __Pyx_CyFunction_CheckExact(obj)  __Pyx_IS_TYPE(obj, __pyx_CyFunctionType)
static PyObject *__Pyx_CyFunction_Init(__pyx_CyFunctionObject* op, PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *closure,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);
static CYTHON_INLINE void __Pyx__CyFunction_SetClassObj(__pyx_CyFunctionObject* f, PyObject* classobj);
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *m,
                                                         size_t size,
                                                         int pyobjects);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *m,
                                                            PyObject *tuple);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *m,
                                                             PyObject *dict);
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *m,
                                                              PyObject *dict);
static int __pyx_CyFunction_init(PyObject *module);
#if CYTHON_METH_FASTCALL
static PyObject * __Pyx_CyFunction_Vectorcall_NOARGS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
static PyObject * __Pyx_CyFunction_Vectorcall_O(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS_METHOD(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
#if CYTHON_BACKPORT_VECTORCALL
#define __Pyx_CyFunction_func_vectorcall(f) (((__pyx_CyFunctionObject*)f)->func_vectorcall)
#else
#define __Pyx_CyFunction_func_vectorcall(f) (((PyCFunctionObject*)f)->vectorcall)
#endif
#endif

/* CythonFunction.proto */
static PyObject *__Pyx_CyFunction_New(PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *closure,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);

/* SetNameInClass.proto */
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1
#define __Pyx_SetNameInClass(ns, name, value)\
    (likely(PyDict_CheckExact(ns)) ? _PyDict_SetItem_KnownHash(ns, name, value, ((PyASCIIObject *) name)->hash) : PyObject_SetItem(ns, name, value))
#elif CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_SetNameInClass(ns, name, value)\
    (likely(PyDict_CheckExact(ns)) ? PyDict_SetItem(ns, name, value) : PyObject_SetItem(ns, name, value))
#else
#define __Pyx_SetNameInClass(ns, name, value)  PyObject_SetItem(ns, name, value)
#endif

/* CalculateMetaclass.proto */
static PyObject *__Pyx_CalculateMetaclass(PyTypeObject *metaclass, PyObject *bases);

/* PyObjectLookupSpecial.proto */
#if CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
#define __Pyx_PyObject_LookupSpecialNoError(obj, attr_name)  __Pyx__PyObject_LookupSpecial(obj, attr_name, 0)
#define __Pyx_PyObject_LookupSpecial(obj, attr_name)  __Pyx__PyObject_LookupSpecial(obj, attr_name, 1)
static CYTHON_INLINE PyObject* __Pyx__PyObject_LookupSpecial(PyObject* obj, PyObject* attr_name, int with_error);
#else
#define __Pyx_PyObject_LookupSpecialNoError(o,n) __Pyx_PyObject_GetAttrStrNoError(o,n)
#define __Pyx_PyObject_LookupSpecial(o,n) __Pyx_PyObject_GetAttrStr(o,n)
#endif

/* Py3ClassCreate.proto */
static PyObject *__Pyx_Py3MetaclassPrepare(PyObject *metaclass, PyObject *bases, PyObject *name, PyObject *qualname,
                                           PyObject *mkw, PyObject *modname, PyObject *doc);
static PyObject *__Pyx_Py3ClassCreate(PyObject *metaclass, PyObject *name, PyObject *bases, PyObject *dict,
                                      PyObject *mkw, int calculate_metaclass, int allow_py2_metaclass);

/* Py3UpdateBases.proto */
static PyObject* __Pyx_PEP560_update_bases(PyObject *bases);

/* CLineInTraceback.proto */
#ifdef CYTHON_CLINE_IN_TRACEBACK
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#else
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#endif

/* CodeObjectCache.proto */
#if !CYTHON_COMPILING_IN_LIMITED_API
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);
#endif

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* GCCDiagnostics.proto */
#if !defined(__INTEL_COMPILER) && defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 6))
#define __Pyx_HAS_GCC_DIAGNOSTIC
#endif

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_readstat_type_t(readstat_type_t value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value);

/* CIntFromPy.proto */
static CYTHON_INLINE readstat_type_t __Pyx_PyInt_As_readstat_type_t(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE __pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format __Pyx_PyInt_As___pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_char(char value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From___pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format(__pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format value);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* FormatTypeName.proto */
#if CYTHON_COMPILING_IN_LIMITED_API
typedef PyObject *__Pyx_TypeName;
#define __Pyx_FMT_TYPENAME "%U"
static __Pyx_TypeName __Pyx_PyType_GetName(PyTypeObject* tp);
#define __Pyx_DECREF_TypeName(obj) Py_XDECREF(obj)
#else
typedef const char *__Pyx_TypeName;
#define __Pyx_FMT_TYPENAME "%.200s"
#define __Pyx_PyType_GetName(tp) ((tp)->tp_name)
#define __Pyx_DECREF_TypeName(obj)
#endif

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
#define __Pyx_TypeCheck2(obj, type1, type2) __Pyx_IsAnySubtype2(Py_TYPE(obj), (PyTypeObject *)type1, (PyTypeObject *)type2)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_IsAnySubtype2(PyTypeObject *cls, PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_TypeCheck2(obj, type1, type2) (PyObject_TypeCheck(obj, (PyTypeObject *)type1) || PyObject_TypeCheck(obj, (PyTypeObject *)type2))
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
#define __Pyx_PyErr_GivenExceptionMatches2(err, type1, type2) (PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2))
#endif
#define __Pyx_PyErr_ExceptionMatches2(err1, err2)  __Pyx_PyErr_GivenExceptionMatches2(__Pyx_PyErr_CurrentExceptionType(), err1, err2)
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(void);

/* VoidPtrExport.proto */
static int __Pyx_ExportVoidPtr(PyObject *name, void *p, const char *sig);

/* FunctionExport.proto */
static int __Pyx_ExportFunction(const char *name, void (*f)(void), const char *sig);

/* InitStrings.proto */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);

/* #### Code section: module_declarations ### */
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_4date_4year_year(PyDateTime_Date *__pyx_v_self); /* proto*/
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_4date_5month_month(PyDateTime_Date *__pyx_v_self); /* proto*/
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_4date_3day_day(PyDateTime_Date *__pyx_v_self); /* proto*/
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_4time_4hour_hour(PyDateTime_Time *__pyx_v_self); /* proto*/
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_4time_6minute_minute(PyDateTime_Time *__pyx_v_self); /* proto*/
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_4time_6second_second(PyDateTime_Time *__pyx_v_self); /* proto*/
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_4time_11microsecond_microsecond(PyDateTime_Time *__pyx_v_self); /* proto*/
static CYTHON_INLINE PyObject *__pyx_f_7cpython_8datetime_4time_6tzinfo_tzinfo(PyDateTime_Time *__pyx_v_self); /* proto*/
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_4time_4fold_fold(PyDateTime_Time *__pyx_v_self); /* proto*/
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_8datetime_4year_year(PyDateTime_DateTime *__pyx_v_self); /* proto*/
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_8datetime_5month_month(PyDateTime_DateTime *__pyx_v_self); /* proto*/
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_8datetime_3day_day(PyDateTime_DateTime *__pyx_v_self); /* proto*/
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_8datetime_4hour_hour(PyDateTime_DateTime *__pyx_v_self); /* proto*/
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_8datetime_6minute_minute(PyDateTime_DateTime *__pyx_v_self); /* proto*/
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_8datetime_6second_second(PyDateTime_DateTime *__pyx_v_self); /* proto*/
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_8datetime_11microsecond_microsecond(PyDateTime_DateTime *__pyx_v_self); /* proto*/
static CYTHON_INLINE PyObject *__pyx_f_7cpython_8datetime_8datetime_6tzinfo_tzinfo(PyDateTime_DateTime *__pyx_v_self); /* proto*/
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_8datetime_4fold_fold(PyDateTime_DateTime *__pyx_v_self); /* proto*/
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_9timedelta_3day_day(PyDateTime_Delta *__pyx_v_self); /* proto*/
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_9timedelta_6second_second(PyDateTime_Delta *__pyx_v_self); /* proto*/
static CYTHON_INLINE int __pyx_f_7cpython_8datetime_9timedelta_11microsecond_microsecond(PyDateTime_Delta *__pyx_v_self); /* proto*/

/* Module declarations from "libc.stddef" */

/* Module declarations from "pyreadstat.readstat_api" */

/* Module declarations from "libc.string" */

/* Module declarations from "libc.stdio" */

/* Module declarations from "__builtin__" */

/* Module declarations from "cpython.type" */

/* Module declarations from "cpython" */

/* Module declarations from "cpython.object" */

/* Module declarations from "cpython.version" */

/* Module declarations from "datetime" */

/* Module declarations from "cpython.datetime" */
static CYTHON_INLINE void __pyx_f_7cpython_8datetime_import_datetime(void); /*proto*/
static CYTHON_INLINE PyDateTime_DateTime *__pyx_f_7cpython_8datetime_datetime_new(int, int, int, int, int, int, int, PyObject *, struct __pyx_opt_args_7cpython_8datetime_datetime_new *__pyx_optional_args); /*proto*/
static CYTHON_INLINE PyDateTime_Delta *__pyx_f_7cpython_8datetime_timedelta_new(int, int, int); /*proto*/

/* Module declarations from "cpython.exc" */

/* Module declarations from "libc.math" */

/* Module declarations from "pyreadstat._readstat_parser" */
static PyObject *__pyx_v_10pyreadstat_16_readstat_parser_readstat_to_numpy_types = 0;
static PyObject *__pyx_v_10pyreadstat_16_readstat_parser_sas_date_formats = 0;
static PyObject *__pyx_v_10pyreadstat_16_readstat_parser_sas_datetime_formats = 0;
static PyObject *__pyx_v_10pyreadstat_16_readstat_parser_sas_time_formats = 0;
static PyObject *__pyx_v_10pyreadstat_16_readstat_parser_sas_all_formats = 0;
static PyObject *__pyx_v_10pyreadstat_16_readstat_parser_sas_origin = 0;
static PyObject *__pyx_v_10pyreadstat_16_readstat_parser_spss_datetime_formats = 0;
static PyObject *__pyx_v_10pyreadstat_16_readstat_parser_spss_date_formats = 0;
static PyObject *__pyx_v_10pyreadstat_16_readstat_parser_spss_time_formats = 0;
static PyObject *__pyx_v_10pyreadstat_16_readstat_parser_spss_all_formats = 0;
static PyObject *__pyx_v_10pyreadstat_16_readstat_parser_spss_origin = 0;
static PyObject *__pyx_v_10pyreadstat_16_readstat_parser_stata_datetime_formats = 0;
static PyObject *__pyx_v_10pyreadstat_16_readstat_parser_stata_date_formats = 0;
static PyObject *__pyx_v_10pyreadstat_16_readstat_parser_stata_time_formats = 0;
static PyObject *__pyx_v_10pyreadstat_16_readstat_parser_stata_all_formats = 0;
static PyObject *__pyx_v_10pyreadstat_16_readstat_parser_stata_origin = 0;
static __pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format __pyx_f_10pyreadstat_16_readstat_parser_transform_variable_format(PyObject *, __pyx_t_10pyreadstat_16_readstat_parser_py_file_format); /*proto*/
static PyObject *__pyx_f_10pyreadstat_16_readstat_parser_transform_datetime(__pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format, double, __pyx_t_10pyreadstat_16_readstat_parser_py_file_format, PyObject *, int); /*proto*/
static int __pyx_f_10pyreadstat_16_readstat_parser_handle_metadata(readstat_metadata_t *, void *); /*proto*/
static int __pyx_f_10pyreadstat_16_readstat_parser_handle_variable(int, readstat_variable_t *, char *, void *); /*proto*/
static int __pyx_f_10pyreadstat_16_readstat_parser_handle_value(int, readstat_variable_t *, readstat_value_t, void *); /*proto*/
static int __pyx_f_10pyreadstat_16_readstat_parser_handle_value_label(char *, readstat_value_t, char *, void *); /*proto*/
static int __pyx_f_10pyreadstat_16_readstat_parser_handle_note(int, char *, void *); /*proto*/
static void __pyx_f_10pyreadstat_16_readstat_parser_check_exit_status(readstat_error_t); /*proto*/
static void __pyx_f_10pyreadstat_16_readstat_parser_run_readstat_parser(char *, struct __pyx_obj_10pyreadstat_16_readstat_parser_data_container *, __pyx_t_10pyreadstat_16_readstat_parser_py_file_extension, long, long); /*proto*/
static PyObject *__pyx_f_10pyreadstat_16_readstat_parser_data_container_to_dict(struct __pyx_obj_10pyreadstat_16_readstat_parser_data_container *); /*proto*/
static PyObject *__pyx_f_10pyreadstat_16_readstat_parser_dict_to_pandas_dataframe(PyObject *, struct __pyx_obj_10pyreadstat_16_readstat_parser_data_container *); /*proto*/
static PyObject *__pyx_f_10pyreadstat_16_readstat_parser_data_container_extract_metadata(struct __pyx_obj_10pyreadstat_16_readstat_parser_data_container *); /*proto*/
static PyObject *__pyx_f_10pyreadstat_16_readstat_parser_convert_readstat_to_python_value(readstat_value_t, int, struct __pyx_obj_10pyreadstat_16_readstat_parser_data_container *); /*proto*/
static int __pyx_f_10pyreadstat_16_readstat_parser_handle_open(char const *, void *); /*proto*/
/* #### Code section: typeinfo ### */
/* #### Code section: before_global_var ### */
#define __Pyx_MODULE_NAME "pyreadstat._readstat_parser"
extern int __pyx_module_is_main_pyreadstat___readstat_parser;
int __pyx_module_is_main_pyreadstat___readstat_parser = 0;

/* Implementation of "pyreadstat._readstat_parser" */
/* #### Code section: global_var ### */
static PyObject *__pyx_builtin_object;
static PyObject *__pyx_builtin_TypeError;
static PyObject *__pyx_builtin_range;
static PyObject *__pyx_builtin_chr;
static PyObject *__pyx_builtin_enumerate;
static PyObject *__pyx_builtin_zip;
static PyObject *__pyx_builtin_UnicodeError;
static PyObject *__pyx_builtin_RuntimeError;
/* #### Code section: string_decls ### */
static const char __pyx_k_d[] = "%d";
static const char __pyx_k__2[] = "";
static const char __pyx_k__8[] = ".";
static const char __pyx_k__9[] = "*";
static const char __pyx_k_gc[] = "gc";
static const char __pyx_k_hi[] = "hi";
static const char __pyx_k_lo[] = "lo";
static const char __pyx_k_np[] = "np";
static const char __pyx_k_nt[] = "nt";
static const char __pyx_k_os[] = "os";
static const char __pyx_k_pd[] = "pd";
static const char __pyx_k_tC[] = "%tC";
static const char __pyx_k_tc[] = "%tc";
static const char __pyx_k_td[] = "%td";
static const char __pyx_k_TOD[] = "TOD";
static const char __pyx_k__15[] = "?";
static const char __pyx_k_chr[] = "chr";
static const char __pyx_k_doc[] = "__doc__";
static const char __pyx_k_get[] = "get";
static const char __pyx_k_nan[] = "nan";
static const char __pyx_k_sys[] = "sys";
static const char __pyx_k_zip[] = "zip";
static const char __pyx_k_DATE[] = "DATE";
static const char __pyx_k_HHMM[] = "HHMM";
static const char __pyx_k_NULL[] = "NULL";
static const char __pyx_k_TIME[] = "TIME";
static const char __pyx_k_copy[] = "copy";
static const char __pyx_k_date[] = "date";
static const char __pyx_k_dict[] = "dict";
static const char __pyx_k_fill[] = "fill";
static const char __pyx_k_init[] = "__init__";
static const char __pyx_k_int8[] = "int8";
static const char __pyx_k_left[] = "left";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_name[] = "name";
static const char __pyx_k_path[] = "path";
static const char __pyx_k_self[] = "self";
static const char __pyx_k_spec[] = "__spec__";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_time[] = "time";
static const char __pyx_k_warn[] = "warn";
static const char __pyx_k_ADATE[] = "ADATE";
static const char __pyx_k_DATE8[] = "DATE8";
static const char __pyx_k_DATE9[] = "DATE9";
static const char __pyx_k_DTIME[] = "DTIME";
static const char __pyx_k_EDATE[] = "EDATE";
static const char __pyx_k_JDATE[] = "JDATE";
static const char __pyx_k_M8_ns[] = "<M8[ns]";
static const char __pyx_k_SDATE[] = "SDATE";
static const char __pyx_k_TIME5[] = "TIME5";
static const char __pyx_k_TIME8[] = "TIME8";
static const char __pyx_k_dtype[] = "dtype";
static const char __pyx_k_empty[] = "empty";
static const char __pyx_k_float[] = "float";
static const char __pyx_k_int16[] = "int16";
static const char __pyx_k_int32[] = "int32";
static const char __pyx_k_int64[] = "int64";
static const char __pyx_k_items[] = "items";
static const char __pyx_k_notes[] = "notes";
static const char __pyx_k_numpy[] = "numpy";
static const char __pyx_k_range[] = "range";
static const char __pyx_k_right[] = "right";
static const char __pyx_k_scale[] = "scale";
static const char __pyx_k_super[] = "super";
static const char __pyx_k_utf_8[] = "utf-8";
static const char __pyx_k_ADATE8[] = "ADATE8";
static const char __pyx_k_DATE11[] = "DATE11";
static const char __pyx_k_DATE12[] = "DATE12";
static const char __pyx_k_DDMMYY[] = "DDMMYY";
static const char __pyx_k_DTDATE[] = "DTDATE";
static const char __pyx_k_EDATE8[] = "EDATE8";
static const char __pyx_k_JDATE5[] = "JDATE5";
static const char __pyx_k_JDATE7[] = "JDATE7";
static const char __pyx_k_MMDDYY[] = "MMDDYY";
static const char __pyx_k_SDATE8[] = "SDATE8";
static const char __pyx_k_TIME20[] = "TIME20";
static const char __pyx_k_YYMMDD[] = "YYMMDD";
static const char __pyx_k_append[] = "append";
static const char __pyx_k_astype[] = "astype";
static const char __pyx_k_center[] = "center";
static const char __pyx_k_dict_2[] = "__dict__";
static const char __pyx_k_double[] = "double";
static const char __pyx_k_dtypes[] = "dtypes";
static const char __pyx_k_enable[] = "enable";
static const char __pyx_k_encode[] = "encode";
static const char __pyx_k_format[] = "format";
static const char __pyx_k_import[] = "__import__";
static const char __pyx_k_isfile[] = "isfile";
static const char __pyx_k_module[] = "__module__";
static const char __pyx_k_name_2[] = "__name__";
static const char __pyx_k_object[] = "object";
static const char __pyx_k_pandas[] = "pandas";
static const char __pyx_k_reduce[] = "__reduce__";
static const char __pyx_k_string[] = "string";
static const char __pyx_k_tolist[] = "tolist";
static const char __pyx_k_ADATE10[] = "ADATE10";
static const char __pyx_k_B8601DA[] = "B8601DA";
static const char __pyx_k_B8601DN[] = "B8601DN";
static const char __pyx_k_B8601DT[] = "B8601DT";
static const char __pyx_k_B8601TM[] = "B8601TM";
static const char __pyx_k_DDMMYYB[] = "DDMMYYB";
static const char __pyx_k_DDMMYYC[] = "DDMMYYC";
static const char __pyx_k_DDMMYYD[] = "DDMMYYD";
static const char __pyx_k_DDMMYYP[] = "DDMMYYP";
static const char __pyx_k_DDMMYYS[] = "DDMMYYS";
static const char __pyx_k_E8601DA[] = "E8601DA";
static const char __pyx_k_E8601DT[] = "E8601DT";
static const char __pyx_k_E8601TM[] = "E8601TM";
static const char __pyx_k_EDATE10[] = "EDATE10";
static const char __pyx_k_MDYAMPM[] = "MDYAMPM";
static const char __pyx_k_MMDDYYB[] = "MMDDYYB";
static const char __pyx_k_MMDDYYC[] = "MMDDYYC";
static const char __pyx_k_MMDDYYD[] = "MMDDYYD";
static const char __pyx_k_MMDDYYP[] = "MMDDYYP";
static const char __pyx_k_MMDDYYS[] = "MMDDYYS";
static const char __pyx_k_SDATE10[] = "SDATE10";
static const char __pyx_k_YYMMDDB[] = "YYMMDDB";
static const char __pyx_k_YYMMDDD[] = "YYMMDDD";
static const char __pyx_k_YYMMDDN[] = "YYMMDDN";
static const char __pyx_k_YYMMDDP[] = "YYMMDDP";
static const char __pyx_k_YYMMDDS[] = "YYMMDDS";
static const char __pyx_k_columns[] = "columns";
static const char __pyx_k_disable[] = "disable";
static const char __pyx_k_float64[] = "float64";
static const char __pyx_k_nominal[] = "nominal";
static const char __pyx_k_ordinal[] = "ordinal";
static const char __pyx_k_prepare[] = "__prepare__";
static const char __pyx_k_tcHH_MM[] = "%tcHH:MM";
static const char __pyx_k_tdD_m_Y[] = "%tdD_m_Y";
static const char __pyx_k_unknown[] = "unknown";
static const char __pyx_k_DATEAMPM[] = "DATEAMPM";
static const char __pyx_k_DATETIME[] = "DATETIME";
static const char __pyx_k_DDMMYYN6[] = "DDMMYYN6";
static const char __pyx_k_DDMMYYN8[] = "DDMMYYN8";
static const char __pyx_k_IS8601DA[] = "IS8601DA";
static const char __pyx_k_IS8601DT[] = "IS8601DT";
static const char __pyx_k_IS8601TM[] = "IS8601TM";
static const char __pyx_k_MMDDYYN6[] = "MMDDYYN6";
static const char __pyx_k_MMDDYYN8[] = "MMDDYYN8";
static const char __pyx_k_TIME11_2[] = "TIME11.2";
static const char __pyx_k_TIME20_3[] = "TIME20.3";
static const char __pyx_k_TIMEAMPM[] = "TIMEAMPM";
static const char __pyx_k_WEEKDATE[] = "WEEKDATE";
static const char __pyx_k_WEEKDATX[] = "WEEKDATX";
static const char __pyx_k_YMDHMS16[] = "YMDHMS16";
static const char __pyx_k_YMDHMS19[] = "YMDHMS19";
static const char __pyx_k_YMDHMS20[] = "YMDHMS20";
static const char __pyx_k_YYMMDD10[] = "YYMMDD10";
static const char __pyx_k_datetime[] = "datetime";
static const char __pyx_k_fsdecode[] = "fsdecode";
static const char __pyx_k_fsencode[] = "fsencode";
static const char __pyx_k_getstate[] = "__getstate__";
static const char __pyx_k_pyx_capi[] = "__pyx_capi__";
static const char __pyx_k_qualname[] = "__qualname__";
static const char __pyx_k_set_name[] = "__set_name__";
static const char __pyx_k_setstate[] = "__setstate__";
static const char __pyx_k_warnings[] = "warnings";
static const char __pyx_k_DATETIME8[] = "DATETIME8";
static const char __pyx_k_DDMMYYB10[] = "DDMMYYB10";
static const char __pyx_k_DDMMYYC10[] = "DDMMYYC10";
static const char __pyx_k_DDMMYYD10[] = "DDMMYYD10";
static const char __pyx_k_DDMMYYP10[] = "DDMMYYP10";
static const char __pyx_k_DDMMYYS10[] = "DDMMYYS10";
static const char __pyx_k_DataFrame[] = "DataFrame";
static const char __pyx_k_MMDDYYB10[] = "MMDDYYB10";
static const char __pyx_k_MMDDYYC10[] = "MMDDYYC10";
static const char __pyx_k_MMDDYYD10[] = "MMDDYYD10";
static const char __pyx_k_MMDDYYP10[] = "MMDDYYP10";
static const char __pyx_k_MMDDYYS10[] = "MMDDYYS10";
static const char __pyx_k_TypeError[] = "TypeError";
static const char __pyx_k_enumerate[] = "enumerate";
static const char __pyx_k_from_dict[] = "from_dict";
static const char __pyx_k_isenabled[] = "isenabled";
static const char __pyx_k_metaclass[] = "__metaclass__";
static const char __pyx_k_pyx_state[] = "__pyx_state";
static const char __pyx_k_reduce_ex[] = "__reduce_ex__";
static const char __pyx_k_DATETIME17[] = "DATETIME17";
static const char __pyx_k_DATETIME18[] = "DATETIME18";
static const char __pyx_k_DATETIME19[] = "DATETIME19";
static const char __pyx_k_DATETIME20[] = "DATETIME20";
static const char __pyx_k_DATETIME21[] = "DATETIME21";
static const char __pyx_k_DATETIME22[] = "DATETIME22";
static const char __pyx_k_YMDHMS19_2[] = "YMDHMS19.2";
static const char __pyx_k_duplicated[] = "_duplicated";
static const char __pyx_k_expanduser[] = "expanduser";
static const char __pyx_k_file_label[] = "file_label";
static const char __pyx_k_sas_origin[] = "sas_origin";
static const char __pyx_k_table_name[] = "table_name";
static const char __pyx_k_tcHH_MM_SS[] = "%tcHH:MM:SS";
static const char __pyx_k_OrderedDict[] = "OrderedDict";
static const char __pyx_k_collections[] = "collections";
static const char __pyx_k_mro_entries[] = "__mro_entries__";
static const char __pyx_k_number_rows[] = "number_rows";
static const char __pyx_k_spss_origin[] = "spss_origin";
static const char __pyx_k_to_datetime[] = "to_datetime";
static const char __pyx_k_DATETIME23_2[] = "DATETIME23.2";
static const char __pyx_k_RuntimeError[] = "RuntimeError";
static const char __pyx_k_UnicodeError[] = "UnicodeError";
static const char __pyx_k_column_names[] = "column_names";
static const char __pyx_k_initializing[] = "_initializing";
static const char __pyx_k_is_coroutine[] = "_is_coroutine";
static const char __pyx_k_stata_origin[] = "stata_origin";
static const char __pyx_k_stringsource[] = "<stringsource>";
static const char __pyx_k_tdCCYY_NN_DD[] = "%tdCCYY-NN-DD";
static const char __pyx_k_undetermined[] = "undetermined";
static const char __pyx_k_value_labels[] = "value_labels";
static const char __pyx_k_ReadstatError[] = "ReadstatError";
static const char __pyx_k_class_getitem[] = "__class_getitem__";
static const char __pyx_k_column_labels[] = "column_labels";
static const char __pyx_k_creation_time[] = "creation_time";
static const char __pyx_k_file_encoding[] = "file_encoding";
static const char __pyx_k_fromtimestamp[] = "fromtimestamp";
static const char __pyx_k_init_subclass[] = "__init_subclass__";
static const char __pyx_k_output_format[] = "output_format";
static const char __pyx_k_reduce_cython[] = "__reduce_cython__";
static const char __pyx_k_data_container[] = "data_container";
static const char __pyx_k_missing_ranges[] = "missing_ranges";
static const char __pyx_k_number_columns[] = "number_columns";
static const char __pyx_k_PyreadstatError[] = "PyreadstatError";
static const char __pyx_k_allowed_formats[] = "allowed_formats";
static const char __pyx_k_sas_all_formats[] = "sas_all_formats";
static const char __pyx_k_setstate_cython[] = "__setstate_cython__";
static const char __pyx_k_surrogateescape[] = "surrogateescape";
static const char __pyx_k_Unkown_data_type[] = "Unkown data type";
static const char __pyx_k_sas_date_formats[] = "sas_date_formats";
static const char __pyx_k_sas_time_formats[] = "sas_time_formats";
static const char __pyx_k_spss_all_formats[] = "spss_all_formats";
static const char __pyx_k_variable_measure[] = "variable_measure";
static const char __pyx_k_modification_time[] = "modification_time";
static const char __pyx_k_spss_date_formats[] = "spss_date_formats";
static const char __pyx_k_spss_time_formats[] = "spss_time_formats";
static const char __pyx_k_stata_all_formats[] = "stata_all_formats";
static const char __pyx_k_variable_to_label[] = "variable_to_label";
static const char __pyx_k_asyncio_coroutines[] = "asyncio.coroutines";
static const char __pyx_k_cline_in_traceback[] = "cline_in_traceback";
static const char __pyx_k_metadata_container[] = "metadata_container";
static const char __pyx_k_stata_date_formats[] = "stata_date_formats";
static const char __pyx_k_stata_time_formats[] = "stata_time_formats";
static const char __pyx_k_variable_alignment[] = "variable_alignment";
static const char __pyx_k_Unknown_file_format[] = "Unknown file format";
static const char __pyx_k_missing_user_values[] = "missing_user_values";
static const char __pyx_k_sas_datetime_formats[] = "sas_datetime_formats";
static const char __pyx_k_File_0_does_not_exist[] = "File {0} does not exist!";
static const char __pyx_k_getfilesystemencoding[] = "getfilesystemencoding";
static const char __pyx_k_spss_datetime_formats[] = "spss_datetime_formats";
static const char __pyx_k_variable_value_labels[] = "variable_value_labels";
static const char __pyx_k_column_names_to_labels[] = "column_names_to_labels";
static const char __pyx_k_stata_datetime_formats[] = "stata_datetime_formats";
static const char __pyx_k_variable_display_width[] = "variable_display_width";
static const char __pyx_k_variable_storage_width[] = "variable_storage_width";
static const char __pyx_k_original_variable_types[] = "original_variable_types";
static const char __pyx_k_readstat_to_numpy_types[] = "readstat_to_numpy_types";
static const char __pyx_k_readstat_variable_types[] = "readstat_variable_types";
static const char __pyx_k_metadata_container___init[] = "metadata_container.__init__";
static const char __pyx_k_pyreadstat__readstat_parser[] = "pyreadstat._readstat_parser";
static const char __pyx_k_data_container___reduce_cython[] = "data_container.__reduce_cython__";
static const char __pyx_k_Just_defining_a_custom_exceptio[] = "\n    Just defining a custom exception to raise when readstat gives an error return code.\n    ";
static const char __pyx_k_This_class_holds_metadata_we_wa[] = "\n    This class holds metadata we want to give back to python\n    ";
static const char __pyx_k_pyreadstat__readstat_parser_pyx[] = "pyreadstat/_readstat_parser.pyx";
static const char __pyx_k_Failed_convert_C_to_python_value[] = "Failed convert C to python value";
static const char __pyx_k_Failed_to_read_number_of_variabl[] = "Failed to read number of variables";
static const char __pyx_k_STRING_type_with_value_s_with_da[] = "STRING type with value '%s' with date type in column '%s'";
static const char __pyx_k_Time_zones_are_not_available_fro[] = "Time zones are not available from the C-API.";
static const char __pyx_k_column_0_is_duplicated_renamed_t[] = "column '{0}' is duplicated, renamed to '{1}'";
static const char __pyx_k_data_container___setstate_cython[] = "data_container.__setstate_cython__";
static const char __pyx_k_file_path_could_not_be_encoded_w[] = "file path could not be encoded with %s which is set as your system encoding, trying to encode it as utf-8. Please set your system encoding correctly.";
static const char __pyx_k_no_default___reduce___due_to_non[] = "no default __reduce__ due to non-trivial __cinit__";
static const char __pyx_k_output_format_must_be_one_of_all[] = "output format must be one of {allowed_formats}, '{output_format}' was given";
static const char __pyx_k_path_must_be_either_str_or_bytes[] = "path must be either str or bytes";
static const char __pyx_k_path_must_be_str_bytes_or_unicod[] = "path must be str, bytes or unicode";
static const char __pyx_k_Just_defining_a_custom_exceptio_2[] = "\n    Just defining a custom exception to raise when pyreadstat raises an exception.\n    ";
/* #### Code section: decls ### */
static int __pyx_pf_10pyreadstat_16_readstat_parser_14data_container___cinit__(struct __pyx_obj_10pyreadstat_16_readstat_parser_data_container *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10pyreadstat_16_readstat_parser_14data_container_2__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_10pyreadstat_16_readstat_parser_data_container *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10pyreadstat_16_readstat_parser_14data_container_4__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_10pyreadstat_16_readstat_parser_data_container *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_10pyreadstat_16_readstat_parser_18metadata_container___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_tp_new_10pyreadstat_16_readstat_parser_data_container(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static __Pyx_CachedCFunction __pyx_umethod_PyDict_Type_get = {0, 0, 0, 0, 0};
static __Pyx_CachedCFunction __pyx_umethod_PyDict_Type_items = {0, 0, 0, 0, 0};
/* #### Code section: late_includes ### */
/* #### Code section: module_state ### */
typedef struct {
  PyObject *__pyx_d;
  PyObject *__pyx_b;
  PyObject *__pyx_cython_runtime;
  PyObject *__pyx_empty_tuple;
  PyObject *__pyx_empty_bytes;
  PyObject *__pyx_empty_unicode;
  #ifdef __Pyx_CyFunction_USED
  PyTypeObject *__pyx_CyFunctionType;
  #endif
  #ifdef __Pyx_FusedFunction_USED
  PyTypeObject *__pyx_FusedFunctionType;
  #endif
  #ifdef __Pyx_Generator_USED
  PyTypeObject *__pyx_GeneratorType;
  #endif
  #ifdef __Pyx_IterableCoroutine_USED
  PyTypeObject *__pyx_IterableCoroutineType;
  #endif
  #ifdef __Pyx_Coroutine_USED
  PyTypeObject *__pyx_CoroutineAwaitType;
  #endif
  #ifdef __Pyx_Coroutine_USED
  PyTypeObject *__pyx_CoroutineType;
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  PyTypeObject *__pyx_ptype_7cpython_4type_type;
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  PyTypeObject *__pyx_ptype_7cpython_8datetime_date;
  PyTypeObject *__pyx_ptype_7cpython_8datetime_time;
  PyTypeObject *__pyx_ptype_7cpython_8datetime_datetime;
  PyTypeObject *__pyx_ptype_7cpython_8datetime_timedelta;
  PyTypeObject *__pyx_ptype_7cpython_8datetime_tzinfo;
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  PyObject *__pyx_type_10pyreadstat_16_readstat_parser_data_container;
  #endif
  PyTypeObject *__pyx_ptype_10pyreadstat_16_readstat_parser_data_container;
  PyObject *__pyx_n_s_ADATE;
  PyObject *__pyx_n_s_ADATE10;
  PyObject *__pyx_n_s_ADATE8;
  PyObject *__pyx_n_s_B8601DA;
  PyObject *__pyx_n_s_B8601DN;
  PyObject *__pyx_n_s_B8601DT;
  PyObject *__pyx_n_s_B8601TM;
  PyObject *__pyx_n_s_DATE;
  PyObject *__pyx_n_s_DATE11;
  PyObject *__pyx_n_s_DATE12;
  PyObject *__pyx_n_s_DATE8;
  PyObject *__pyx_n_s_DATE9;
  PyObject *__pyx_n_s_DATEAMPM;
  PyObject *__pyx_n_s_DATETIME;
  PyObject *__pyx_n_s_DATETIME17;
  PyObject *__pyx_n_s_DATETIME18;
  PyObject *__pyx_n_s_DATETIME19;
  PyObject *__pyx_n_s_DATETIME20;
  PyObject *__pyx_n_s_DATETIME21;
  PyObject *__pyx_n_s_DATETIME22;
  PyObject *__pyx_kp_s_DATETIME23_2;
  PyObject *__pyx_n_s_DATETIME8;
  PyObject *__pyx_n_s_DDMMYY;
  PyObject *__pyx_n_s_DDMMYYB;
  PyObject *__pyx_n_s_DDMMYYB10;
  PyObject *__pyx_n_s_DDMMYYC;
  PyObject *__pyx_n_s_DDMMYYC10;
  PyObject *__pyx_n_s_DDMMYYD;
  PyObject *__pyx_n_s_DDMMYYD10;
  PyObject *__pyx_n_s_DDMMYYN6;
  PyObject *__pyx_n_s_DDMMYYN8;
  PyObject *__pyx_n_s_DDMMYYP;
  PyObject *__pyx_n_s_DDMMYYP10;
  PyObject *__pyx_n_s_DDMMYYS;
  PyObject *__pyx_n_s_DDMMYYS10;
  PyObject *__pyx_n_s_DTDATE;
  PyObject *__pyx_n_s_DTIME;
  PyObject *__pyx_n_s_DataFrame;
  PyObject *__pyx_n_s_E8601DA;
  PyObject *__pyx_n_s_E8601DT;
  PyObject *__pyx_n_s_E8601TM;
  PyObject *__pyx_n_s_EDATE;
  PyObject *__pyx_n_s_EDATE10;
  PyObject *__pyx_n_s_EDATE8;
  PyObject *__pyx_kp_s_Failed_convert_C_to_python_value;
  PyObject *__pyx_kp_s_Failed_to_read_number_of_variabl;
  PyObject *__pyx_kp_s_File_0_does_not_exist;
  PyObject *__pyx_n_s_HHMM;
  PyObject *__pyx_n_s_IS8601DA;
  PyObject *__pyx_n_s_IS8601DT;
  PyObject *__pyx_n_s_IS8601TM;
  PyObject *__pyx_n_s_JDATE;
  PyObject *__pyx_n_s_JDATE5;
  PyObject *__pyx_n_s_JDATE7;
  PyObject *__pyx_kp_s_Just_defining_a_custom_exceptio;
  PyObject *__pyx_kp_s_Just_defining_a_custom_exceptio_2;
  PyObject *__pyx_kp_s_M8_ns;
  PyObject *__pyx_n_s_MDYAMPM;
  PyObject *__pyx_n_s_MMDDYY;
  PyObject *__pyx_n_s_MMDDYYB;
  PyObject *__pyx_n_s_MMDDYYB10;
  PyObject *__pyx_n_s_MMDDYYC;
  PyObject *__pyx_n_s_MMDDYYC10;
  PyObject *__pyx_n_s_MMDDYYD;
  PyObject *__pyx_n_s_MMDDYYD10;
  PyObject *__pyx_n_s_MMDDYYN6;
  PyObject *__pyx_n_s_MMDDYYN8;
  PyObject *__pyx_n_s_MMDDYYP;
  PyObject *__pyx_n_s_MMDDYYP10;
  PyObject *__pyx_n_s_MMDDYYS;
  PyObject *__pyx_n_s_MMDDYYS10;
  PyObject *__pyx_n_s_NULL;
  PyObject *__pyx_n_s_OrderedDict;
  PyObject *__pyx_n_s_PyreadstatError;
  PyObject *__pyx_n_s_ReadstatError;
  PyObject *__pyx_n_s_RuntimeError;
  PyObject *__pyx_n_s_SDATE;
  PyObject *__pyx_n_s_SDATE10;
  PyObject *__pyx_n_s_SDATE8;
  PyObject *__pyx_kp_s_STRING_type_with_value_s_with_da;
  PyObject *__pyx_n_s_TIME;
  PyObject *__pyx_kp_s_TIME11_2;
  PyObject *__pyx_n_s_TIME20;
  PyObject *__pyx_kp_s_TIME20_3;
  PyObject *__pyx_n_s_TIME5;
  PyObject *__pyx_n_s_TIME8;
  PyObject *__pyx_n_s_TIMEAMPM;
  PyObject *__pyx_n_s_TOD;
  PyObject *__pyx_kp_s_This_class_holds_metadata_we_wa;
  PyObject *__pyx_kp_s_Time_zones_are_not_available_fro;
  PyObject *__pyx_n_s_TypeError;
  PyObject *__pyx_n_s_UnicodeError;
  PyObject *__pyx_kp_s_Unknown_file_format;
  PyObject *__pyx_kp_s_Unkown_data_type;
  PyObject *__pyx_n_s_WEEKDATE;
  PyObject *__pyx_n_s_WEEKDATX;
  PyObject *__pyx_n_s_YMDHMS16;
  PyObject *__pyx_n_s_YMDHMS19;
  PyObject *__pyx_kp_s_YMDHMS19_2;
  PyObject *__pyx_n_s_YMDHMS20;
  PyObject *__pyx_n_s_YYMMDD;
  PyObject *__pyx_n_s_YYMMDD10;
  PyObject *__pyx_n_s_YYMMDDB;
  PyObject *__pyx_n_s_YYMMDDD;
  PyObject *__pyx_n_s_YYMMDDN;
  PyObject *__pyx_n_s_YYMMDDP;
  PyObject *__pyx_n_s_YYMMDDS;
  PyObject *__pyx_n_s__15;
  PyObject *__pyx_kp_s__2;
  PyObject *__pyx_kp_u__8;
  PyObject *__pyx_n_s__9;
  PyObject *__pyx_n_s_allowed_formats;
  PyObject *__pyx_n_s_append;
  PyObject *__pyx_n_s_astype;
  PyObject *__pyx_n_s_asyncio_coroutines;
  PyObject *__pyx_n_s_center;
  PyObject *__pyx_n_s_chr;
  PyObject *__pyx_n_s_class_getitem;
  PyObject *__pyx_n_s_cline_in_traceback;
  PyObject *__pyx_n_s_collections;
  PyObject *__pyx_kp_s_column_0_is_duplicated_renamed_t;
  PyObject *__pyx_n_s_column_labels;
  PyObject *__pyx_n_s_column_names;
  PyObject *__pyx_n_s_column_names_to_labels;
  PyObject *__pyx_n_s_columns;
  PyObject *__pyx_n_s_copy;
  PyObject *__pyx_n_s_creation_time;
  PyObject *__pyx_kp_s_d;
  PyObject *__pyx_n_s_data_container;
  PyObject *__pyx_n_s_data_container___reduce_cython;
  PyObject *__pyx_n_s_data_container___setstate_cython;
  PyObject *__pyx_n_s_date;
  PyObject *__pyx_n_s_datetime;
  PyObject *__pyx_n_s_dict;
  PyObject *__pyx_n_s_dict_2;
  PyObject *__pyx_kp_u_disable;
  PyObject *__pyx_n_s_doc;
  PyObject *__pyx_n_s_double;
  PyObject *__pyx_n_s_dtype;
  PyObject *__pyx_n_s_dtypes;
  PyObject *__pyx_n_s_duplicated;
  PyObject *__pyx_n_s_empty;
  PyObject *__pyx_kp_u_enable;
  PyObject *__pyx_n_s_encode;
  PyObject *__pyx_n_s_enumerate;
  PyObject *__pyx_n_s_expanduser;
  PyObject *__pyx_n_s_file_encoding;
  PyObject *__pyx_n_s_file_label;
  PyObject *__pyx_kp_s_file_path_could_not_be_encoded_w;
  PyObject *__pyx_n_s_fill;
  PyObject *__pyx_n_s_float;
  PyObject *__pyx_n_s_float64;
  PyObject *__pyx_n_s_format;
  PyObject *__pyx_n_s_from_dict;
  PyObject *__pyx_n_s_fromtimestamp;
  PyObject *__pyx_n_s_fsdecode;
  PyObject *__pyx_n_s_fsencode;
  PyObject *__pyx_kp_u_gc;
  PyObject *__pyx_n_s_get;
  PyObject *__pyx_n_s_getfilesystemencoding;
  PyObject *__pyx_n_s_getstate;
  PyObject *__pyx_n_s_hi;
  PyObject *__pyx_n_s_import;
  PyObject *__pyx_n_s_init;
  PyObject *__pyx_n_s_init_subclass;
  PyObject *__pyx_n_s_initializing;
  PyObject *__pyx_n_s_int16;
  PyObject *__pyx_n_s_int32;
  PyObject *__pyx_n_s_int64;
  PyObject *__pyx_n_s_int8;
  PyObject *__pyx_n_s_is_coroutine;
  PyObject *__pyx_kp_u_isenabled;
  PyObject *__pyx_n_s_isfile;
  PyObject *__pyx_n_s_items;
  PyObject *__pyx_n_s_left;
  PyObject *__pyx_n_s_lo;
  PyObject *__pyx_n_s_main;
  PyObject *__pyx_n_s_metaclass;
  PyObject *__pyx_n_s_metadata_container;
  PyObject *__pyx_n_s_metadata_container___init;
  PyObject *__pyx_n_s_missing_ranges;
  PyObject *__pyx_n_s_missing_user_values;
  PyObject *__pyx_n_s_modification_time;
  PyObject *__pyx_n_s_module;
  PyObject *__pyx_n_s_mro_entries;
  PyObject *__pyx_n_s_name;
  PyObject *__pyx_n_s_name_2;
  PyObject *__pyx_n_s_nan;
  PyObject *__pyx_kp_s_no_default___reduce___due_to_non;
  PyObject *__pyx_n_s_nominal;
  PyObject *__pyx_n_s_notes;
  PyObject *__pyx_n_s_np;
  PyObject *__pyx_n_s_nt;
  PyObject *__pyx_n_s_number_columns;
  PyObject *__pyx_n_s_number_rows;
  PyObject *__pyx_n_s_numpy;
  PyObject *__pyx_n_s_object;
  PyObject *__pyx_n_s_ordinal;
  PyObject *__pyx_n_s_original_variable_types;
  PyObject *__pyx_n_s_os;
  PyObject *__pyx_n_s_output_format;
  PyObject *__pyx_kp_s_output_format_must_be_one_of_all;
  PyObject *__pyx_n_s_pandas;
  PyObject *__pyx_n_s_path;
  PyObject *__pyx_kp_s_path_must_be_either_str_or_bytes;
  PyObject *__pyx_kp_s_path_must_be_str_bytes_or_unicod;
  PyObject *__pyx_n_s_pd;
  PyObject *__pyx_n_s_prepare;
  PyObject *__pyx_n_s_pyreadstat__readstat_parser;
  PyObject *__pyx_kp_s_pyreadstat__readstat_parser_pyx;
  PyObject *__pyx_n_s_pyx_capi;
  PyObject *__pyx_n_s_pyx_state;
  PyObject *__pyx_n_s_qualname;
  PyObject *__pyx_n_s_range;
  PyObject *__pyx_n_s_readstat_to_numpy_types;
  PyObject *__pyx_n_s_readstat_variable_types;
  PyObject *__pyx_n_s_reduce;
  PyObject *__pyx_n_s_reduce_cython;
  PyObject *__pyx_n_s_reduce_ex;
  PyObject *__pyx_n_s_right;
  PyObject *__pyx_n_s_sas_all_formats;
  PyObject *__pyx_n_s_sas_date_formats;
  PyObject *__pyx_n_s_sas_datetime_formats;
  PyObject *__pyx_n_s_sas_origin;
  PyObject *__pyx_n_s_sas_time_formats;
  PyObject *__pyx_n_s_scale;
  PyObject *__pyx_n_s_self;
  PyObject *__pyx_n_s_set_name;
  PyObject *__pyx_n_s_setstate;
  PyObject *__pyx_n_s_setstate_cython;
  PyObject *__pyx_n_s_spec;
  PyObject *__pyx_n_s_spss_all_formats;
  PyObject *__pyx_n_s_spss_date_formats;
  PyObject *__pyx_n_s_spss_datetime_formats;
  PyObject *__pyx_n_s_spss_origin;
  PyObject *__pyx_n_s_spss_time_formats;
  PyObject *__pyx_n_s_stata_all_formats;
  PyObject *__pyx_n_s_stata_date_formats;
  PyObject *__pyx_n_s_stata_datetime_formats;
  PyObject *__pyx_n_s_stata_origin;
  PyObject *__pyx_n_s_stata_time_formats;
  PyObject *__pyx_n_s_string;
  PyObject *__pyx_kp_s_stringsource;
  PyObject *__pyx_n_s_super;
  PyObject *__pyx_n_s_surrogateescape;
  PyObject *__pyx_n_s_sys;
  PyObject *__pyx_kp_s_tC;
  PyObject *__pyx_n_s_table_name;
  PyObject *__pyx_kp_s_tc;
  PyObject *__pyx_kp_s_tcHH_MM;
  PyObject *__pyx_kp_s_tcHH_MM_SS;
  PyObject *__pyx_kp_s_td;
  PyObject *__pyx_kp_s_tdCCYY_NN_DD;
  PyObject *__pyx_kp_s_tdD_m_Y;
  PyObject *__pyx_n_s_test;
  PyObject *__pyx_n_s_time;
  PyObject *__pyx_n_s_to_datetime;
  PyObject *__pyx_n_s_tolist;
  PyObject *__pyx_n_s_undetermined;
  PyObject *__pyx_n_s_unknown;
  PyObject *__pyx_kp_s_utf_8;
  PyObject *__pyx_n_s_value_labels;
  PyObject *__pyx_n_s_variable_alignment;
  PyObject *__pyx_n_s_variable_display_width;
  PyObject *__pyx_n_s_variable_measure;
  PyObject *__pyx_n_s_variable_storage_width;
  PyObject *__pyx_n_s_variable_to_label;
  PyObject *__pyx_n_s_variable_value_labels;
  PyObject *__pyx_n_s_warn;
  PyObject *__pyx_n_s_warnings;
  PyObject *__pyx_n_s_zip;
  PyObject *__pyx_int_0;
  PyObject *__pyx_int_1;
  PyObject *__pyx_int_100000;
  PyObject *__pyx_tuple_;
  PyObject *__pyx_tuple__3;
  PyObject *__pyx_tuple__4;
  PyObject *__pyx_tuple__5;
  PyObject *__pyx_tuple__6;
  PyObject *__pyx_tuple__7;
  PyObject *__pyx_tuple__10;
  PyObject *__pyx_tuple__12;
  PyObject *__pyx_codeobj__11;
  PyObject *__pyx_codeobj__13;
  PyObject *__pyx_codeobj__14;
} __pyx_mstate;

#if CYTHON_USE_MODULE_STATE
#ifdef __cplusplus
namespace {
  extern struct PyModuleDef __pyx_moduledef;
} /* anonymous namespace */
#else
static struct PyModuleDef __pyx_moduledef;
#endif

#define __pyx_mstate(o) ((__pyx_mstate *)__Pyx_PyModule_GetState(o))

#define __pyx_mstate_global (__pyx_mstate(PyState_FindModule(&__pyx_moduledef)))

#define __pyx_m (PyState_FindModule(&__pyx_moduledef))
#else
static __pyx_mstate __pyx_mstate_global_static =
#ifdef __cplusplus
    {};
#else
    {0};
#endif
static __pyx_mstate *__pyx_mstate_global = &__pyx_mstate_global_static;
#endif
/* #### Code section: module_state_clear ### */
#if CYTHON_USE_MODULE_STATE
static int __pyx_m_clear(PyObject *m) {
  __pyx_mstate *clear_module_state = __pyx_mstate(m);
  if (!clear_module_state) return 0;
  Py_CLEAR(clear_module_state->__pyx_d);
  Py_CLEAR(clear_module_state->__pyx_b);
  Py_CLEAR(clear_module_state->__pyx_cython_runtime);
  Py_CLEAR(clear_module_state->__pyx_empty_tuple);
  Py_CLEAR(clear_module_state->__pyx_empty_bytes);
  Py_CLEAR(clear_module_state->__pyx_empty_unicode);
  #ifdef __Pyx_CyFunction_USED
  Py_CLEAR(clear_module_state->__pyx_CyFunctionType);
  #endif
  #ifdef __Pyx_FusedFunction_USED
  Py_CLEAR(clear_module_state->__pyx_FusedFunctionType);
  #endif
  Py_CLEAR(clear_module_state->__pyx_ptype_7cpython_4type_type);
  Py_CLEAR(clear_module_state->__pyx_ptype_7cpython_8datetime_date);
  Py_CLEAR(clear_module_state->__pyx_ptype_7cpython_8datetime_time);
  Py_CLEAR(clear_module_state->__pyx_ptype_7cpython_8datetime_datetime);
  Py_CLEAR(clear_module_state->__pyx_ptype_7cpython_8datetime_timedelta);
  Py_CLEAR(clear_module_state->__pyx_ptype_7cpython_8datetime_tzinfo);
  Py_CLEAR(clear_module_state->__pyx_ptype_10pyreadstat_16_readstat_parser_data_container);
  Py_CLEAR(clear_module_state->__pyx_type_10pyreadstat_16_readstat_parser_data_container);
  Py_CLEAR(clear_module_state->__pyx_n_s_ADATE);
  Py_CLEAR(clear_module_state->__pyx_n_s_ADATE10);
  Py_CLEAR(clear_module_state->__pyx_n_s_ADATE8);
  Py_CLEAR(clear_module_state->__pyx_n_s_B8601DA);
  Py_CLEAR(clear_module_state->__pyx_n_s_B8601DN);
  Py_CLEAR(clear_module_state->__pyx_n_s_B8601DT);
  Py_CLEAR(clear_module_state->__pyx_n_s_B8601TM);
  Py_CLEAR(clear_module_state->__pyx_n_s_DATE);
  Py_CLEAR(clear_module_state->__pyx_n_s_DATE11);
  Py_CLEAR(clear_module_state->__pyx_n_s_DATE12);
  Py_CLEAR(clear_module_state->__pyx_n_s_DATE8);
  Py_CLEAR(clear_module_state->__pyx_n_s_DATE9);
  Py_CLEAR(clear_module_state->__pyx_n_s_DATEAMPM);
  Py_CLEAR(clear_module_state->__pyx_n_s_DATETIME);
  Py_CLEAR(clear_module_state->__pyx_n_s_DATETIME17);
  Py_CLEAR(clear_module_state->__pyx_n_s_DATETIME18);
  Py_CLEAR(clear_module_state->__pyx_n_s_DATETIME19);
  Py_CLEAR(clear_module_state->__pyx_n_s_DATETIME20);
  Py_CLEAR(clear_module_state->__pyx_n_s_DATETIME21);
  Py_CLEAR(clear_module_state->__pyx_n_s_DATETIME22);
  Py_CLEAR(clear_module_state->__pyx_kp_s_DATETIME23_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_DATETIME8);
  Py_CLEAR(clear_module_state->__pyx_n_s_DDMMYY);
  Py_CLEAR(clear_module_state->__pyx_n_s_DDMMYYB);
  Py_CLEAR(clear_module_state->__pyx_n_s_DDMMYYB10);
  Py_CLEAR(clear_module_state->__pyx_n_s_DDMMYYC);
  Py_CLEAR(clear_module_state->__pyx_n_s_DDMMYYC10);
  Py_CLEAR(clear_module_state->__pyx_n_s_DDMMYYD);
  Py_CLEAR(clear_module_state->__pyx_n_s_DDMMYYD10);
  Py_CLEAR(clear_module_state->__pyx_n_s_DDMMYYN6);
  Py_CLEAR(clear_module_state->__pyx_n_s_DDMMYYN8);
  Py_CLEAR(clear_module_state->__pyx_n_s_DDMMYYP);
  Py_CLEAR(clear_module_state->__pyx_n_s_DDMMYYP10);
  Py_CLEAR(clear_module_state->__pyx_n_s_DDMMYYS);
  Py_CLEAR(clear_module_state->__pyx_n_s_DDMMYYS10);
  Py_CLEAR(clear_module_state->__pyx_n_s_DTDATE);
  Py_CLEAR(clear_module_state->__pyx_n_s_DTIME);
  Py_CLEAR(clear_module_state->__pyx_n_s_DataFrame);
  Py_CLEAR(clear_module_state->__pyx_n_s_E8601DA);
  Py_CLEAR(clear_module_state->__pyx_n_s_E8601DT);
  Py_CLEAR(clear_module_state->__pyx_n_s_E8601TM);
  Py_CLEAR(clear_module_state->__pyx_n_s_EDATE);
  Py_CLEAR(clear_module_state->__pyx_n_s_EDATE10);
  Py_CLEAR(clear_module_state->__pyx_n_s_EDATE8);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Failed_convert_C_to_python_value);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Failed_to_read_number_of_variabl);
  Py_CLEAR(clear_module_state->__pyx_kp_s_File_0_does_not_exist);
  Py_CLEAR(clear_module_state->__pyx_n_s_HHMM);
  Py_CLEAR(clear_module_state->__pyx_n_s_IS8601DA);
  Py_CLEAR(clear_module_state->__pyx_n_s_IS8601DT);
  Py_CLEAR(clear_module_state->__pyx_n_s_IS8601TM);
  Py_CLEAR(clear_module_state->__pyx_n_s_JDATE);
  Py_CLEAR(clear_module_state->__pyx_n_s_JDATE5);
  Py_CLEAR(clear_module_state->__pyx_n_s_JDATE7);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Just_defining_a_custom_exceptio);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Just_defining_a_custom_exceptio_2);
  Py_CLEAR(clear_module_state->__pyx_kp_s_M8_ns);
  Py_CLEAR(clear_module_state->__pyx_n_s_MDYAMPM);
  Py_CLEAR(clear_module_state->__pyx_n_s_MMDDYY);
  Py_CLEAR(clear_module_state->__pyx_n_s_MMDDYYB);
  Py_CLEAR(clear_module_state->__pyx_n_s_MMDDYYB10);
  Py_CLEAR(clear_module_state->__pyx_n_s_MMDDYYC);
  Py_CLEAR(clear_module_state->__pyx_n_s_MMDDYYC10);
  Py_CLEAR(clear_module_state->__pyx_n_s_MMDDYYD);
  Py_CLEAR(clear_module_state->__pyx_n_s_MMDDYYD10);
  Py_CLEAR(clear_module_state->__pyx_n_s_MMDDYYN6);
  Py_CLEAR(clear_module_state->__pyx_n_s_MMDDYYN8);
  Py_CLEAR(clear_module_state->__pyx_n_s_MMDDYYP);
  Py_CLEAR(clear_module_state->__pyx_n_s_MMDDYYP10);
  Py_CLEAR(clear_module_state->__pyx_n_s_MMDDYYS);
  Py_CLEAR(clear_module_state->__pyx_n_s_MMDDYYS10);
  Py_CLEAR(clear_module_state->__pyx_n_s_NULL);
  Py_CLEAR(clear_module_state->__pyx_n_s_OrderedDict);
  Py_CLEAR(clear_module_state->__pyx_n_s_PyreadstatError);
  Py_CLEAR(clear_module_state->__pyx_n_s_ReadstatError);
  Py_CLEAR(clear_module_state->__pyx_n_s_RuntimeError);
  Py_CLEAR(clear_module_state->__pyx_n_s_SDATE);
  Py_CLEAR(clear_module_state->__pyx_n_s_SDATE10);
  Py_CLEAR(clear_module_state->__pyx_n_s_SDATE8);
  Py_CLEAR(clear_module_state->__pyx_kp_s_STRING_type_with_value_s_with_da);
  Py_CLEAR(clear_module_state->__pyx_n_s_TIME);
  Py_CLEAR(clear_module_state->__pyx_kp_s_TIME11_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_TIME20);
  Py_CLEAR(clear_module_state->__pyx_kp_s_TIME20_3);
  Py_CLEAR(clear_module_state->__pyx_n_s_TIME5);
  Py_CLEAR(clear_module_state->__pyx_n_s_TIME8);
  Py_CLEAR(clear_module_state->__pyx_n_s_TIMEAMPM);
  Py_CLEAR(clear_module_state->__pyx_n_s_TOD);
  Py_CLEAR(clear_module_state->__pyx_kp_s_This_class_holds_metadata_we_wa);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Time_zones_are_not_available_fro);
  Py_CLEAR(clear_module_state->__pyx_n_s_TypeError);
  Py_CLEAR(clear_module_state->__pyx_n_s_UnicodeError);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Unknown_file_format);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Unkown_data_type);
  Py_CLEAR(clear_module_state->__pyx_n_s_WEEKDATE);
  Py_CLEAR(clear_module_state->__pyx_n_s_WEEKDATX);
  Py_CLEAR(clear_module_state->__pyx_n_s_YMDHMS16);
  Py_CLEAR(clear_module_state->__pyx_n_s_YMDHMS19);
  Py_CLEAR(clear_module_state->__pyx_kp_s_YMDHMS19_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_YMDHMS20);
  Py_CLEAR(clear_module_state->__pyx_n_s_YYMMDD);
  Py_CLEAR(clear_module_state->__pyx_n_s_YYMMDD10);
  Py_CLEAR(clear_module_state->__pyx_n_s_YYMMDDB);
  Py_CLEAR(clear_module_state->__pyx_n_s_YYMMDDD);
  Py_CLEAR(clear_module_state->__pyx_n_s_YYMMDDN);
  Py_CLEAR(clear_module_state->__pyx_n_s_YYMMDDP);
  Py_CLEAR(clear_module_state->__pyx_n_s_YYMMDDS);
  Py_CLEAR(clear_module_state->__pyx_n_s__15);
  Py_CLEAR(clear_module_state->__pyx_kp_s__2);
  Py_CLEAR(clear_module_state->__pyx_kp_u__8);
  Py_CLEAR(clear_module_state->__pyx_n_s__9);
  Py_CLEAR(clear_module_state->__pyx_n_s_allowed_formats);
  Py_CLEAR(clear_module_state->__pyx_n_s_append);
  Py_CLEAR(clear_module_state->__pyx_n_s_astype);
  Py_CLEAR(clear_module_state->__pyx_n_s_asyncio_coroutines);
  Py_CLEAR(clear_module_state->__pyx_n_s_center);
  Py_CLEAR(clear_module_state->__pyx_n_s_chr);
  Py_CLEAR(clear_module_state->__pyx_n_s_class_getitem);
  Py_CLEAR(clear_module_state->__pyx_n_s_cline_in_traceback);
  Py_CLEAR(clear_module_state->__pyx_n_s_collections);
  Py_CLEAR(clear_module_state->__pyx_kp_s_column_0_is_duplicated_renamed_t);
  Py_CLEAR(clear_module_state->__pyx_n_s_column_labels);
  Py_CLEAR(clear_module_state->__pyx_n_s_column_names);
  Py_CLEAR(clear_module_state->__pyx_n_s_column_names_to_labels);
  Py_CLEAR(clear_module_state->__pyx_n_s_columns);
  Py_CLEAR(clear_module_state->__pyx_n_s_copy);
  Py_CLEAR(clear_module_state->__pyx_n_s_creation_time);
  Py_CLEAR(clear_module_state->__pyx_kp_s_d);
  Py_CLEAR(clear_module_state->__pyx_n_s_data_container);
  Py_CLEAR(clear_module_state->__pyx_n_s_data_container___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_data_container___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_date);
  Py_CLEAR(clear_module_state->__pyx_n_s_datetime);
  Py_CLEAR(clear_module_state->__pyx_n_s_dict);
  Py_CLEAR(clear_module_state->__pyx_n_s_dict_2);
  Py_CLEAR(clear_module_state->__pyx_kp_u_disable);
  Py_CLEAR(clear_module_state->__pyx_n_s_doc);
  Py_CLEAR(clear_module_state->__pyx_n_s_double);
  Py_CLEAR(clear_module_state->__pyx_n_s_dtype);
  Py_CLEAR(clear_module_state->__pyx_n_s_dtypes);
  Py_CLEAR(clear_module_state->__pyx_n_s_duplicated);
  Py_CLEAR(clear_module_state->__pyx_n_s_empty);
  Py_CLEAR(clear_module_state->__pyx_kp_u_enable);
  Py_CLEAR(clear_module_state->__pyx_n_s_encode);
  Py_CLEAR(clear_module_state->__pyx_n_s_enumerate);
  Py_CLEAR(clear_module_state->__pyx_n_s_expanduser);
  Py_CLEAR(clear_module_state->__pyx_n_s_file_encoding);
  Py_CLEAR(clear_module_state->__pyx_n_s_file_label);
  Py_CLEAR(clear_module_state->__pyx_kp_s_file_path_could_not_be_encoded_w);
  Py_CLEAR(clear_module_state->__pyx_n_s_fill);
  Py_CLEAR(clear_module_state->__pyx_n_s_float);
  Py_CLEAR(clear_module_state->__pyx_n_s_float64);
  Py_CLEAR(clear_module_state->__pyx_n_s_format);
  Py_CLEAR(clear_module_state->__pyx_n_s_from_dict);
  Py_CLEAR(clear_module_state->__pyx_n_s_fromtimestamp);
  Py_CLEAR(clear_module_state->__pyx_n_s_fsdecode);
  Py_CLEAR(clear_module_state->__pyx_n_s_fsencode);
  Py_CLEAR(clear_module_state->__pyx_kp_u_gc);
  Py_CLEAR(clear_module_state->__pyx_n_s_get);
  Py_CLEAR(clear_module_state->__pyx_n_s_getfilesystemencoding);
  Py_CLEAR(clear_module_state->__pyx_n_s_getstate);
  Py_CLEAR(clear_module_state->__pyx_n_s_hi);
  Py_CLEAR(clear_module_state->__pyx_n_s_import);
  Py_CLEAR(clear_module_state->__pyx_n_s_init);
  Py_CLEAR(clear_module_state->__pyx_n_s_init_subclass);
  Py_CLEAR(clear_module_state->__pyx_n_s_initializing);
  Py_CLEAR(clear_module_state->__pyx_n_s_int16);
  Py_CLEAR(clear_module_state->__pyx_n_s_int32);
  Py_CLEAR(clear_module_state->__pyx_n_s_int64);
  Py_CLEAR(clear_module_state->__pyx_n_s_int8);
  Py_CLEAR(clear_module_state->__pyx_n_s_is_coroutine);
  Py_CLEAR(clear_module_state->__pyx_kp_u_isenabled);
  Py_CLEAR(clear_module_state->__pyx_n_s_isfile);
  Py_CLEAR(clear_module_state->__pyx_n_s_items);
  Py_CLEAR(clear_module_state->__pyx_n_s_left);
  Py_CLEAR(clear_module_state->__pyx_n_s_lo);
  Py_CLEAR(clear_module_state->__pyx_n_s_main);
  Py_CLEAR(clear_module_state->__pyx_n_s_metaclass);
  Py_CLEAR(clear_module_state->__pyx_n_s_metadata_container);
  Py_CLEAR(clear_module_state->__pyx_n_s_metadata_container___init);
  Py_CLEAR(clear_module_state->__pyx_n_s_missing_ranges);
  Py_CLEAR(clear_module_state->__pyx_n_s_missing_user_values);
  Py_CLEAR(clear_module_state->__pyx_n_s_modification_time);
  Py_CLEAR(clear_module_state->__pyx_n_s_module);
  Py_CLEAR(clear_module_state->__pyx_n_s_mro_entries);
  Py_CLEAR(clear_module_state->__pyx_n_s_name);
  Py_CLEAR(clear_module_state->__pyx_n_s_name_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_nan);
  Py_CLEAR(clear_module_state->__pyx_kp_s_no_default___reduce___due_to_non);
  Py_CLEAR(clear_module_state->__pyx_n_s_nominal);
  Py_CLEAR(clear_module_state->__pyx_n_s_notes);
  Py_CLEAR(clear_module_state->__pyx_n_s_np);
  Py_CLEAR(clear_module_state->__pyx_n_s_nt);
  Py_CLEAR(clear_module_state->__pyx_n_s_number_columns);
  Py_CLEAR(clear_module_state->__pyx_n_s_number_rows);
  Py_CLEAR(clear_module_state->__pyx_n_s_numpy);
  Py_CLEAR(clear_module_state->__pyx_n_s_object);
  Py_CLEAR(clear_module_state->__pyx_n_s_ordinal);
  Py_CLEAR(clear_module_state->__pyx_n_s_original_variable_types);
  Py_CLEAR(clear_module_state->__pyx_n_s_os);
  Py_CLEAR(clear_module_state->__pyx_n_s_output_format);
  Py_CLEAR(clear_module_state->__pyx_kp_s_output_format_must_be_one_of_all);
  Py_CLEAR(clear_module_state->__pyx_n_s_pandas);
  Py_CLEAR(clear_module_state->__pyx_n_s_path);
  Py_CLEAR(clear_module_state->__pyx_kp_s_path_must_be_either_str_or_bytes);
  Py_CLEAR(clear_module_state->__pyx_kp_s_path_must_be_str_bytes_or_unicod);
  Py_CLEAR(clear_module_state->__pyx_n_s_pd);
  Py_CLEAR(clear_module_state->__pyx_n_s_prepare);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyreadstat__readstat_parser);
  Py_CLEAR(clear_module_state->__pyx_kp_s_pyreadstat__readstat_parser_pyx);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_capi);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_state);
  Py_CLEAR(clear_module_state->__pyx_n_s_qualname);
  Py_CLEAR(clear_module_state->__pyx_n_s_range);
  Py_CLEAR(clear_module_state->__pyx_n_s_readstat_to_numpy_types);
  Py_CLEAR(clear_module_state->__pyx_n_s_readstat_variable_types);
  Py_CLEAR(clear_module_state->__pyx_n_s_reduce);
  Py_CLEAR(clear_module_state->__pyx_n_s_reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_reduce_ex);
  Py_CLEAR(clear_module_state->__pyx_n_s_right);
  Py_CLEAR(clear_module_state->__pyx_n_s_sas_all_formats);
  Py_CLEAR(clear_module_state->__pyx_n_s_sas_date_formats);
  Py_CLEAR(clear_module_state->__pyx_n_s_sas_datetime_formats);
  Py_CLEAR(clear_module_state->__pyx_n_s_sas_origin);
  Py_CLEAR(clear_module_state->__pyx_n_s_sas_time_formats);
  Py_CLEAR(clear_module_state->__pyx_n_s_scale);
  Py_CLEAR(clear_module_state->__pyx_n_s_self);
  Py_CLEAR(clear_module_state->__pyx_n_s_set_name);
  Py_CLEAR(clear_module_state->__pyx_n_s_setstate);
  Py_CLEAR(clear_module_state->__pyx_n_s_setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_spec);
  Py_CLEAR(clear_module_state->__pyx_n_s_spss_all_formats);
  Py_CLEAR(clear_module_state->__pyx_n_s_spss_date_formats);
  Py_CLEAR(clear_module_state->__pyx_n_s_spss_datetime_formats);
  Py_CLEAR(clear_module_state->__pyx_n_s_spss_origin);
  Py_CLEAR(clear_module_state->__pyx_n_s_spss_time_formats);
  Py_CLEAR(clear_module_state->__pyx_n_s_stata_all_formats);
  Py_CLEAR(clear_module_state->__pyx_n_s_stata_date_formats);
  Py_CLEAR(clear_module_state->__pyx_n_s_stata_datetime_formats);
  Py_CLEAR(clear_module_state->__pyx_n_s_stata_origin);
  Py_CLEAR(clear_module_state->__pyx_n_s_stata_time_formats);
  Py_CLEAR(clear_module_state->__pyx_n_s_string);
  Py_CLEAR(clear_module_state->__pyx_kp_s_stringsource);
  Py_CLEAR(clear_module_state->__pyx_n_s_super);
  Py_CLEAR(clear_module_state->__pyx_n_s_surrogateescape);
  Py_CLEAR(clear_module_state->__pyx_n_s_sys);
  Py_CLEAR(clear_module_state->__pyx_kp_s_tC);
  Py_CLEAR(clear_module_state->__pyx_n_s_table_name);
  Py_CLEAR(clear_module_state->__pyx_kp_s_tc);
  Py_CLEAR(clear_module_state->__pyx_kp_s_tcHH_MM);
  Py_CLEAR(clear_module_state->__pyx_kp_s_tcHH_MM_SS);
  Py_CLEAR(clear_module_state->__pyx_kp_s_td);
  Py_CLEAR(clear_module_state->__pyx_kp_s_tdCCYY_NN_DD);
  Py_CLEAR(clear_module_state->__pyx_kp_s_tdD_m_Y);
  Py_CLEAR(clear_module_state->__pyx_n_s_test);
  Py_CLEAR(clear_module_state->__pyx_n_s_time);
  Py_CLEAR(clear_module_state->__pyx_n_s_to_datetime);
  Py_CLEAR(clear_module_state->__pyx_n_s_tolist);
  Py_CLEAR(clear_module_state->__pyx_n_s_undetermined);
  Py_CLEAR(clear_module_state->__pyx_n_s_unknown);
  Py_CLEAR(clear_module_state->__pyx_kp_s_utf_8);
  Py_CLEAR(clear_module_state->__pyx_n_s_value_labels);
  Py_CLEAR(clear_module_state->__pyx_n_s_variable_alignment);
  Py_CLEAR(clear_module_state->__pyx_n_s_variable_display_width);
  Py_CLEAR(clear_module_state->__pyx_n_s_variable_measure);
  Py_CLEAR(clear_module_state->__pyx_n_s_variable_storage_width);
  Py_CLEAR(clear_module_state->__pyx_n_s_variable_to_label);
  Py_CLEAR(clear_module_state->__pyx_n_s_variable_value_labels);
  Py_CLEAR(clear_module_state->__pyx_n_s_warn);
  Py_CLEAR(clear_module_state->__pyx_n_s_warnings);
  Py_CLEAR(clear_module_state->__pyx_n_s_zip);
  Py_CLEAR(clear_module_state->__pyx_int_0);
  Py_CLEAR(clear_module_state->__pyx_int_1);
  Py_CLEAR(clear_module_state->__pyx_int_100000);
  Py_CLEAR(clear_module_state->__pyx_tuple_);
  Py_CLEAR(clear_module_state->__pyx_tuple__3);
  Py_CLEAR(clear_module_state->__pyx_tuple__4);
  Py_CLEAR(clear_module_state->__pyx_tuple__5);
  Py_CLEAR(clear_module_state->__pyx_tuple__6);
  Py_CLEAR(clear_module_state->__pyx_tuple__7);
  Py_CLEAR(clear_module_state->__pyx_tuple__10);
  Py_CLEAR(clear_module_state->__pyx_tuple__12);
  Py_CLEAR(clear_module_state->__pyx_codeobj__11);
  Py_CLEAR(clear_module_state->__pyx_codeobj__13);
  Py_CLEAR(clear_module_state->__pyx_codeobj__14);
  return 0;
}
#endif
/* #### Code section: module_state_traverse ### */
#if CYTHON_USE_MODULE_STATE
static int __pyx_m_traverse(PyObject *m, visitproc visit, void *arg) {
  __pyx_mstate *traverse_module_state = __pyx_mstate(m);
  if (!traverse_module_state) return 0;
  Py_VISIT(traverse_module_state->__pyx_d);
  Py_VISIT(traverse_module_state->__pyx_b);
  Py_VISIT(traverse_module_state->__pyx_cython_runtime);
  Py_VISIT(traverse_module_state->__pyx_empty_tuple);
  Py_VISIT(traverse_module_state->__pyx_empty_bytes);
  Py_VISIT(traverse_module_state->__pyx_empty_unicode);
  #ifdef __Pyx_CyFunction_USED
  Py_VISIT(traverse_module_state->__pyx_CyFunctionType);
  #endif
  #ifdef __Pyx_FusedFunction_USED
  Py_VISIT(traverse_module_state->__pyx_FusedFunctionType);
  #endif
  Py_VISIT(traverse_module_state->__pyx_ptype_7cpython_4type_type);
  Py_VISIT(traverse_module_state->__pyx_ptype_7cpython_8datetime_date);
  Py_VISIT(traverse_module_state->__pyx_ptype_7cpython_8datetime_time);
  Py_VISIT(traverse_module_state->__pyx_ptype_7cpython_8datetime_datetime);
  Py_VISIT(traverse_module_state->__pyx_ptype_7cpython_8datetime_timedelta);
  Py_VISIT(traverse_module_state->__pyx_ptype_7cpython_8datetime_tzinfo);
  Py_VISIT(traverse_module_state->__pyx_ptype_10pyreadstat_16_readstat_parser_data_container);
  Py_VISIT(traverse_module_state->__pyx_type_10pyreadstat_16_readstat_parser_data_container);
  Py_VISIT(traverse_module_state->__pyx_n_s_ADATE);
  Py_VISIT(traverse_module_state->__pyx_n_s_ADATE10);
  Py_VISIT(traverse_module_state->__pyx_n_s_ADATE8);
  Py_VISIT(traverse_module_state->__pyx_n_s_B8601DA);
  Py_VISIT(traverse_module_state->__pyx_n_s_B8601DN);
  Py_VISIT(traverse_module_state->__pyx_n_s_B8601DT);
  Py_VISIT(traverse_module_state->__pyx_n_s_B8601TM);
  Py_VISIT(traverse_module_state->__pyx_n_s_DATE);
  Py_VISIT(traverse_module_state->__pyx_n_s_DATE11);
  Py_VISIT(traverse_module_state->__pyx_n_s_DATE12);
  Py_VISIT(traverse_module_state->__pyx_n_s_DATE8);
  Py_VISIT(traverse_module_state->__pyx_n_s_DATE9);
  Py_VISIT(traverse_module_state->__pyx_n_s_DATEAMPM);
  Py_VISIT(traverse_module_state->__pyx_n_s_DATETIME);
  Py_VISIT(traverse_module_state->__pyx_n_s_DATETIME17);
  Py_VISIT(traverse_module_state->__pyx_n_s_DATETIME18);
  Py_VISIT(traverse_module_state->__pyx_n_s_DATETIME19);
  Py_VISIT(traverse_module_state->__pyx_n_s_DATETIME20);
  Py_VISIT(traverse_module_state->__pyx_n_s_DATETIME21);
  Py_VISIT(traverse_module_state->__pyx_n_s_DATETIME22);
  Py_VISIT(traverse_module_state->__pyx_kp_s_DATETIME23_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_DATETIME8);
  Py_VISIT(traverse_module_state->__pyx_n_s_DDMMYY);
  Py_VISIT(traverse_module_state->__pyx_n_s_DDMMYYB);
  Py_VISIT(traverse_module_state->__pyx_n_s_DDMMYYB10);
  Py_VISIT(traverse_module_state->__pyx_n_s_DDMMYYC);
  Py_VISIT(traverse_module_state->__pyx_n_s_DDMMYYC10);
  Py_VISIT(traverse_module_state->__pyx_n_s_DDMMYYD);
  Py_VISIT(traverse_module_state->__pyx_n_s_DDMMYYD10);
  Py_VISIT(traverse_module_state->__pyx_n_s_DDMMYYN6);
  Py_VISIT(traverse_module_state->__pyx_n_s_DDMMYYN8);
  Py_VISIT(traverse_module_state->__pyx_n_s_DDMMYYP);
  Py_VISIT(traverse_module_state->__pyx_n_s_DDMMYYP10);
  Py_VISIT(traverse_module_state->__pyx_n_s_DDMMYYS);
  Py_VISIT(traverse_module_state->__pyx_n_s_DDMMYYS10);
  Py_VISIT(traverse_module_state->__pyx_n_s_DTDATE);
  Py_VISIT(traverse_module_state->__pyx_n_s_DTIME);
  Py_VISIT(traverse_module_state->__pyx_n_s_DataFrame);
  Py_VISIT(traverse_module_state->__pyx_n_s_E8601DA);
  Py_VISIT(traverse_module_state->__pyx_n_s_E8601DT);
  Py_VISIT(traverse_module_state->__pyx_n_s_E8601TM);
  Py_VISIT(traverse_module_state->__pyx_n_s_EDATE);
  Py_VISIT(traverse_module_state->__pyx_n_s_EDATE10);
  Py_VISIT(traverse_module_state->__pyx_n_s_EDATE8);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Failed_convert_C_to_python_value);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Failed_to_read_number_of_variabl);
  Py_VISIT(traverse_module_state->__pyx_kp_s_File_0_does_not_exist);
  Py_VISIT(traverse_module_state->__pyx_n_s_HHMM);
  Py_VISIT(traverse_module_state->__pyx_n_s_IS8601DA);
  Py_VISIT(traverse_module_state->__pyx_n_s_IS8601DT);
  Py_VISIT(traverse_module_state->__pyx_n_s_IS8601TM);
  Py_VISIT(traverse_module_state->__pyx_n_s_JDATE);
  Py_VISIT(traverse_module_state->__pyx_n_s_JDATE5);
  Py_VISIT(traverse_module_state->__pyx_n_s_JDATE7);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Just_defining_a_custom_exceptio);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Just_defining_a_custom_exceptio_2);
  Py_VISIT(traverse_module_state->__pyx_kp_s_M8_ns);
  Py_VISIT(traverse_module_state->__pyx_n_s_MDYAMPM);
  Py_VISIT(traverse_module_state->__pyx_n_s_MMDDYY);
  Py_VISIT(traverse_module_state->__pyx_n_s_MMDDYYB);
  Py_VISIT(traverse_module_state->__pyx_n_s_MMDDYYB10);
  Py_VISIT(traverse_module_state->__pyx_n_s_MMDDYYC);
  Py_VISIT(traverse_module_state->__pyx_n_s_MMDDYYC10);
  Py_VISIT(traverse_module_state->__pyx_n_s_MMDDYYD);
  Py_VISIT(traverse_module_state->__pyx_n_s_MMDDYYD10);
  Py_VISIT(traverse_module_state->__pyx_n_s_MMDDYYN6);
  Py_VISIT(traverse_module_state->__pyx_n_s_MMDDYYN8);
  Py_VISIT(traverse_module_state->__pyx_n_s_MMDDYYP);
  Py_VISIT(traverse_module_state->__pyx_n_s_MMDDYYP10);
  Py_VISIT(traverse_module_state->__pyx_n_s_MMDDYYS);
  Py_VISIT(traverse_module_state->__pyx_n_s_MMDDYYS10);
  Py_VISIT(traverse_module_state->__pyx_n_s_NULL);
  Py_VISIT(traverse_module_state->__pyx_n_s_OrderedDict);
  Py_VISIT(traverse_module_state->__pyx_n_s_PyreadstatError);
  Py_VISIT(traverse_module_state->__pyx_n_s_ReadstatError);
  Py_VISIT(traverse_module_state->__pyx_n_s_RuntimeError);
  Py_VISIT(traverse_module_state->__pyx_n_s_SDATE);
  Py_VISIT(traverse_module_state->__pyx_n_s_SDATE10);
  Py_VISIT(traverse_module_state->__pyx_n_s_SDATE8);
  Py_VISIT(traverse_module_state->__pyx_kp_s_STRING_type_with_value_s_with_da);
  Py_VISIT(traverse_module_state->__pyx_n_s_TIME);
  Py_VISIT(traverse_module_state->__pyx_kp_s_TIME11_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_TIME20);
  Py_VISIT(traverse_module_state->__pyx_kp_s_TIME20_3);
  Py_VISIT(traverse_module_state->__pyx_n_s_TIME5);
  Py_VISIT(traverse_module_state->__pyx_n_s_TIME8);
  Py_VISIT(traverse_module_state->__pyx_n_s_TIMEAMPM);
  Py_VISIT(traverse_module_state->__pyx_n_s_TOD);
  Py_VISIT(traverse_module_state->__pyx_kp_s_This_class_holds_metadata_we_wa);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Time_zones_are_not_available_fro);
  Py_VISIT(traverse_module_state->__pyx_n_s_TypeError);
  Py_VISIT(traverse_module_state->__pyx_n_s_UnicodeError);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Unknown_file_format);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Unkown_data_type);
  Py_VISIT(traverse_module_state->__pyx_n_s_WEEKDATE);
  Py_VISIT(traverse_module_state->__pyx_n_s_WEEKDATX);
  Py_VISIT(traverse_module_state->__pyx_n_s_YMDHMS16);
  Py_VISIT(traverse_module_state->__pyx_n_s_YMDHMS19);
  Py_VISIT(traverse_module_state->__pyx_kp_s_YMDHMS19_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_YMDHMS20);
  Py_VISIT(traverse_module_state->__pyx_n_s_YYMMDD);
  Py_VISIT(traverse_module_state->__pyx_n_s_YYMMDD10);
  Py_VISIT(traverse_module_state->__pyx_n_s_YYMMDDB);
  Py_VISIT(traverse_module_state->__pyx_n_s_YYMMDDD);
  Py_VISIT(traverse_module_state->__pyx_n_s_YYMMDDN);
  Py_VISIT(traverse_module_state->__pyx_n_s_YYMMDDP);
  Py_VISIT(traverse_module_state->__pyx_n_s_YYMMDDS);
  Py_VISIT(traverse_module_state->__pyx_n_s__15);
  Py_VISIT(traverse_module_state->__pyx_kp_s__2);
  Py_VISIT(traverse_module_state->__pyx_kp_u__8);
  Py_VISIT(traverse_module_state->__pyx_n_s__9);
  Py_VISIT(traverse_module_state->__pyx_n_s_allowed_formats);
  Py_VISIT(traverse_module_state->__pyx_n_s_append);
  Py_VISIT(traverse_module_state->__pyx_n_s_astype);
  Py_VISIT(traverse_module_state->__pyx_n_s_asyncio_coroutines);
  Py_VISIT(traverse_module_state->__pyx_n_s_center);
  Py_VISIT(traverse_module_state->__pyx_n_s_chr);
  Py_VISIT(traverse_module_state->__pyx_n_s_class_getitem);
  Py_VISIT(traverse_module_state->__pyx_n_s_cline_in_traceback);
  Py_VISIT(traverse_module_state->__pyx_n_s_collections);
  Py_VISIT(traverse_module_state->__pyx_kp_s_column_0_is_duplicated_renamed_t);
  Py_VISIT(traverse_module_state->__pyx_n_s_column_labels);
  Py_VISIT(traverse_module_state->__pyx_n_s_column_names);
  Py_VISIT(traverse_module_state->__pyx_n_s_column_names_to_labels);
  Py_VISIT(traverse_module_state->__pyx_n_s_columns);
  Py_VISIT(traverse_module_state->__pyx_n_s_copy);
  Py_VISIT(traverse_module_state->__pyx_n_s_creation_time);
  Py_VISIT(traverse_module_state->__pyx_kp_s_d);
  Py_VISIT(traverse_module_state->__pyx_n_s_data_container);
  Py_VISIT(traverse_module_state->__pyx_n_s_data_container___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_data_container___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_date);
  Py_VISIT(traverse_module_state->__pyx_n_s_datetime);
  Py_VISIT(traverse_module_state->__pyx_n_s_dict);
  Py_VISIT(traverse_module_state->__pyx_n_s_dict_2);
  Py_VISIT(traverse_module_state->__pyx_kp_u_disable);
  Py_VISIT(traverse_module_state->__pyx_n_s_doc);
  Py_VISIT(traverse_module_state->__pyx_n_s_double);
  Py_VISIT(traverse_module_state->__pyx_n_s_dtype);
  Py_VISIT(traverse_module_state->__pyx_n_s_dtypes);
  Py_VISIT(traverse_module_state->__pyx_n_s_duplicated);
  Py_VISIT(traverse_module_state->__pyx_n_s_empty);
  Py_VISIT(traverse_module_state->__pyx_kp_u_enable);
  Py_VISIT(traverse_module_state->__pyx_n_s_encode);
  Py_VISIT(traverse_module_state->__pyx_n_s_enumerate);
  Py_VISIT(traverse_module_state->__pyx_n_s_expanduser);
  Py_VISIT(traverse_module_state->__pyx_n_s_file_encoding);
  Py_VISIT(traverse_module_state->__pyx_n_s_file_label);
  Py_VISIT(traverse_module_state->__pyx_kp_s_file_path_could_not_be_encoded_w);
  Py_VISIT(traverse_module_state->__pyx_n_s_fill);
  Py_VISIT(traverse_module_state->__pyx_n_s_float);
  Py_VISIT(traverse_module_state->__pyx_n_s_float64);
  Py_VISIT(traverse_module_state->__pyx_n_s_format);
  Py_VISIT(traverse_module_state->__pyx_n_s_from_dict);
  Py_VISIT(traverse_module_state->__pyx_n_s_fromtimestamp);
  Py_VISIT(traverse_module_state->__pyx_n_s_fsdecode);
  Py_VISIT(traverse_module_state->__pyx_n_s_fsencode);
  Py_VISIT(traverse_module_state->__pyx_kp_u_gc);
  Py_VISIT(traverse_module_state->__pyx_n_s_get);
  Py_VISIT(traverse_module_state->__pyx_n_s_getfilesystemencoding);
  Py_VISIT(traverse_module_state->__pyx_n_s_getstate);
  Py_VISIT(traverse_module_state->__pyx_n_s_hi);
  Py_VISIT(traverse_module_state->__pyx_n_s_import);
  Py_VISIT(traverse_module_state->__pyx_n_s_init);
  Py_VISIT(traverse_module_state->__pyx_n_s_init_subclass);
  Py_VISIT(traverse_module_state->__pyx_n_s_initializing);
  Py_VISIT(traverse_module_state->__pyx_n_s_int16);
  Py_VISIT(traverse_module_state->__pyx_n_s_int32);
  Py_VISIT(traverse_module_state->__pyx_n_s_int64);
  Py_VISIT(traverse_module_state->__pyx_n_s_int8);
  Py_VISIT(traverse_module_state->__pyx_n_s_is_coroutine);
  Py_VISIT(traverse_module_state->__pyx_kp_u_isenabled);
  Py_VISIT(traverse_module_state->__pyx_n_s_isfile);
  Py_VISIT(traverse_module_state->__pyx_n_s_items);
  Py_VISIT(traverse_module_state->__pyx_n_s_left);
  Py_VISIT(traverse_module_state->__pyx_n_s_lo);
  Py_VISIT(traverse_module_state->__pyx_n_s_main);
  Py_VISIT(traverse_module_state->__pyx_n_s_metaclass);
  Py_VISIT(traverse_module_state->__pyx_n_s_metadata_container);
  Py_VISIT(traverse_module_state->__pyx_n_s_metadata_container___init);
  Py_VISIT(traverse_module_state->__pyx_n_s_missing_ranges);
  Py_VISIT(traverse_module_state->__pyx_n_s_missing_user_values);
  Py_VISIT(traverse_module_state->__pyx_n_s_modification_time);
  Py_VISIT(traverse_module_state->__pyx_n_s_module);
  Py_VISIT(traverse_module_state->__pyx_n_s_mro_entries);
  Py_VISIT(traverse_module_state->__pyx_n_s_name);
  Py_VISIT(traverse_module_state->__pyx_n_s_name_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_nan);
  Py_VISIT(traverse_module_state->__pyx_kp_s_no_default___reduce___due_to_non);
  Py_VISIT(traverse_module_state->__pyx_n_s_nominal);
  Py_VISIT(traverse_module_state->__pyx_n_s_notes);
  Py_VISIT(traverse_module_state->__pyx_n_s_np);
  Py_VISIT(traverse_module_state->__pyx_n_s_nt);
  Py_VISIT(traverse_module_state->__pyx_n_s_number_columns);
  Py_VISIT(traverse_module_state->__pyx_n_s_number_rows);
  Py_VISIT(traverse_module_state->__pyx_n_s_numpy);
  Py_VISIT(traverse_module_state->__pyx_n_s_object);
  Py_VISIT(traverse_module_state->__pyx_n_s_ordinal);
  Py_VISIT(traverse_module_state->__pyx_n_s_original_variable_types);
  Py_VISIT(traverse_module_state->__pyx_n_s_os);
  Py_VISIT(traverse_module_state->__pyx_n_s_output_format);
  Py_VISIT(traverse_module_state->__pyx_kp_s_output_format_must_be_one_of_all);
  Py_VISIT(traverse_module_state->__pyx_n_s_pandas);
  Py_VISIT(traverse_module_state->__pyx_n_s_path);
  Py_VISIT(traverse_module_state->__pyx_kp_s_path_must_be_either_str_or_bytes);
  Py_VISIT(traverse_module_state->__pyx_kp_s_path_must_be_str_bytes_or_unicod);
  Py_VISIT(traverse_module_state->__pyx_n_s_pd);
  Py_VISIT(traverse_module_state->__pyx_n_s_prepare);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyreadstat__readstat_parser);
  Py_VISIT(traverse_module_state->__pyx_kp_s_pyreadstat__readstat_parser_pyx);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_capi);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_state);
  Py_VISIT(traverse_module_state->__pyx_n_s_qualname);
  Py_VISIT(traverse_module_state->__pyx_n_s_range);
  Py_VISIT(traverse_module_state->__pyx_n_s_readstat_to_numpy_types);
  Py_VISIT(traverse_module_state->__pyx_n_s_readstat_variable_types);
  Py_VISIT(traverse_module_state->__pyx_n_s_reduce);
  Py_VISIT(traverse_module_state->__pyx_n_s_reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_reduce_ex);
  Py_VISIT(traverse_module_state->__pyx_n_s_right);
  Py_VISIT(traverse_module_state->__pyx_n_s_sas_all_formats);
  Py_VISIT(traverse_module_state->__pyx_n_s_sas_date_formats);
  Py_VISIT(traverse_module_state->__pyx_n_s_sas_datetime_formats);
  Py_VISIT(traverse_module_state->__pyx_n_s_sas_origin);
  Py_VISIT(traverse_module_state->__pyx_n_s_sas_time_formats);
  Py_VISIT(traverse_module_state->__pyx_n_s_scale);
  Py_VISIT(traverse_module_state->__pyx_n_s_self);
  Py_VISIT(traverse_module_state->__pyx_n_s_set_name);
  Py_VISIT(traverse_module_state->__pyx_n_s_setstate);
  Py_VISIT(traverse_module_state->__pyx_n_s_setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_spec);
  Py_VISIT(traverse_module_state->__pyx_n_s_spss_all_formats);
  Py_VISIT(traverse_module_state->__pyx_n_s_spss_date_formats);
  Py_VISIT(traverse_module_state->__pyx_n_s_spss_datetime_formats);
  Py_VISIT(traverse_module_state->__pyx_n_s_spss_origin);
  Py_VISIT(traverse_module_state->__pyx_n_s_spss_time_formats);
  Py_VISIT(traverse_module_state->__pyx_n_s_stata_all_formats);
  Py_VISIT(traverse_module_state->__pyx_n_s_stata_date_formats);
  Py_VISIT(traverse_module_state->__pyx_n_s_stata_datetime_formats);
  Py_VISIT(traverse_module_state->__pyx_n_s_stata_origin);
  Py_VISIT(traverse_module_state->__pyx_n_s_stata_time_formats);
  Py_VISIT(traverse_module_state->__pyx_n_s_string);
  Py_VISIT(traverse_module_state->__pyx_kp_s_stringsource);
  Py_VISIT(traverse_module_state->__pyx_n_s_super);
  Py_VISIT(traverse_module_state->__pyx_n_s_surrogateescape);
  Py_VISIT(traverse_module_state->__pyx_n_s_sys);
  Py_VISIT(traverse_module_state->__pyx_kp_s_tC);
  Py_VISIT(traverse_module_state->__pyx_n_s_table_name);
  Py_VISIT(traverse_module_state->__pyx_kp_s_tc);
  Py_VISIT(traverse_module_state->__pyx_kp_s_tcHH_MM);
  Py_VISIT(traverse_module_state->__pyx_kp_s_tcHH_MM_SS);
  Py_VISIT(traverse_module_state->__pyx_kp_s_td);
  Py_VISIT(traverse_module_state->__pyx_kp_s_tdCCYY_NN_DD);
  Py_VISIT(traverse_module_state->__pyx_kp_s_tdD_m_Y);
  Py_VISIT(traverse_module_state->__pyx_n_s_test);
  Py_VISIT(traverse_module_state->__pyx_n_s_time);
  Py_VISIT(traverse_module_state->__pyx_n_s_to_datetime);
  Py_VISIT(traverse_module_state->__pyx_n_s_tolist);
  Py_VISIT(traverse_module_state->__pyx_n_s_undetermined);
  Py_VISIT(traverse_module_state->__pyx_n_s_unknown);
  Py_VISIT(traverse_module_state->__pyx_kp_s_utf_8);
  Py_VISIT(traverse_module_state->__pyx_n_s_value_labels);
  Py_VISIT(traverse_module_state->__pyx_n_s_variable_alignment);
  Py_VISIT(traverse_module_state->__pyx_n_s_variable_display_width);
  Py_VISIT(traverse_module_state->__pyx_n_s_variable_measure);
  Py_VISIT(traverse_module_state->__pyx_n_s_variable_storage_width);
  Py_VISIT(traverse_module_state->__pyx_n_s_variable_to_label);
  Py_VISIT(traverse_module_state->__pyx_n_s_variable_value_labels);
  Py_VISIT(traverse_module_state->__pyx_n_s_warn);
  Py_VISIT(traverse_module_state->__pyx_n_s_warnings);
  Py_VISIT(traverse_module_state->__pyx_n_s_zip);
  Py_VISIT(traverse_module_state->__pyx_int_0);
  Py_VISIT(traverse_module_state->__pyx_int_1);
  Py_VISIT(traverse_module_state->__pyx_int_100000);
  Py_VISIT(traverse_module_state->__pyx_tuple_);
  Py_VISIT(traverse_module_state->__pyx_tuple__3);
  Py_VISIT(traverse_module_state->__pyx_tuple__4);
  Py_VISIT(traverse_module_state->__pyx_tuple__5);
  Py_VISIT(traverse_module_state->__pyx_tuple__6);
  Py_VISIT(traverse_module_state->__pyx_tuple__7);
  Py_VISIT(traverse_module_state->__pyx_tuple__10);
  Py_VISIT(traverse_module_state->__pyx_tuple__12);
  Py_VISIT(traverse_module_state->__pyx_codeobj__11);
  Py_VISIT(traverse_module_state->__pyx_codeobj__13);
  Py_VISIT(traverse_module_state->__pyx_codeobj__14);
  return 0;
}
#endif
/* #### Code section: module_state_defines ### */
#define __pyx_d __pyx_mstate_global->__pyx_d
#define __pyx_b __pyx_mstate_global->__pyx_b
#define __pyx_cython_runtime __pyx_mstate_global->__pyx_cython_runtime
#define __pyx_empty_tuple __pyx_mstate_global->__pyx_empty_tuple
#define __pyx_empty_bytes __pyx_mstate_global->__pyx_empty_bytes
#define __pyx_empty_unicode __pyx_mstate_global->__pyx_empty_unicode
#ifdef __Pyx_CyFunction_USED
#define __pyx_CyFunctionType __pyx_mstate_global->__pyx_CyFunctionType
#endif
#ifdef __Pyx_FusedFunction_USED
#define __pyx_FusedFunctionType __pyx_mstate_global->__pyx_FusedFunctionType
#endif
#ifdef __Pyx_Generator_USED
#define __pyx_GeneratorType __pyx_mstate_global->__pyx_GeneratorType
#endif
#ifdef __Pyx_IterableCoroutine_USED
#define __pyx_IterableCoroutineType __pyx_mstate_global->__pyx_IterableCoroutineType
#endif
#ifdef __Pyx_Coroutine_USED
#define __pyx_CoroutineAwaitType __pyx_mstate_global->__pyx_CoroutineAwaitType
#endif
#ifdef __Pyx_Coroutine_USED
#define __pyx_CoroutineType __pyx_mstate_global->__pyx_CoroutineType
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#define __pyx_ptype_7cpython_4type_type __pyx_mstate_global->__pyx_ptype_7cpython_4type_type
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#define __pyx_ptype_7cpython_8datetime_date __pyx_mstate_global->__pyx_ptype_7cpython_8datetime_date
#define __pyx_ptype_7cpython_8datetime_time __pyx_mstate_global->__pyx_ptype_7cpython_8datetime_time
#define __pyx_ptype_7cpython_8datetime_datetime __pyx_mstate_global->__pyx_ptype_7cpython_8datetime_datetime
#define __pyx_ptype_7cpython_8datetime_timedelta __pyx_mstate_global->__pyx_ptype_7cpython_8datetime_timedelta
#define __pyx_ptype_7cpython_8datetime_tzinfo __pyx_mstate_global->__pyx_ptype_7cpython_8datetime_tzinfo
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#define __pyx_type_10pyreadstat_16_readstat_parser_data_container __pyx_mstate_global->__pyx_type_10pyreadstat_16_readstat_parser_data_container
#endif
#define __pyx_ptype_10pyreadstat_16_readstat_parser_data_container __pyx_mstate_global->__pyx_ptype_10pyreadstat_16_readstat_parser_data_container
#define __pyx_n_s_ADATE __pyx_mstate_global->__pyx_n_s_ADATE
#define __pyx_n_s_ADATE10 __pyx_mstate_global->__pyx_n_s_ADATE10
#define __pyx_n_s_ADATE8 __pyx_mstate_global->__pyx_n_s_ADATE8
#define __pyx_n_s_B8601DA __pyx_mstate_global->__pyx_n_s_B8601DA
#define __pyx_n_s_B8601DN __pyx_mstate_global->__pyx_n_s_B8601DN
#define __pyx_n_s_B8601DT __pyx_mstate_global->__pyx_n_s_B8601DT
#define __pyx_n_s_B8601TM __pyx_mstate_global->__pyx_n_s_B8601TM
#define __pyx_n_s_DATE __pyx_mstate_global->__pyx_n_s_DATE
#define __pyx_n_s_DATE11 __pyx_mstate_global->__pyx_n_s_DATE11
#define __pyx_n_s_DATE12 __pyx_mstate_global->__pyx_n_s_DATE12
#define __pyx_n_s_DATE8 __pyx_mstate_global->__pyx_n_s_DATE8
#define __pyx_n_s_DATE9 __pyx_mstate_global->__pyx_n_s_DATE9
#define __pyx_n_s_DATEAMPM __pyx_mstate_global->__pyx_n_s_DATEAMPM
#define __pyx_n_s_DATETIME __pyx_mstate_global->__pyx_n_s_DATETIME
#define __pyx_n_s_DATETIME17 __pyx_mstate_global->__pyx_n_s_DATETIME17
#define __pyx_n_s_DATETIME18 __pyx_mstate_global->__pyx_n_s_DATETIME18
#define __pyx_n_s_DATETIME19 __pyx_mstate_global->__pyx_n_s_DATETIME19
#define __pyx_n_s_DATETIME20 __pyx_mstate_global->__pyx_n_s_DATETIME20
#define __pyx_n_s_DATETIME21 __pyx_mstate_global->__pyx_n_s_DATETIME21
#define __pyx_n_s_DATETIME22 __pyx_mstate_global->__pyx_n_s_DATETIME22
#define __pyx_kp_s_DATETIME23_2 __pyx_mstate_global->__pyx_kp_s_DATETIME23_2
#define __pyx_n_s_DATETIME8 __pyx_mstate_global->__pyx_n_s_DATETIME8
#define __pyx_n_s_DDMMYY __pyx_mstate_global->__pyx_n_s_DDMMYY
#define __pyx_n_s_DDMMYYB __pyx_mstate_global->__pyx_n_s_DDMMYYB
#define __pyx_n_s_DDMMYYB10 __pyx_mstate_global->__pyx_n_s_DDMMYYB10
#define __pyx_n_s_DDMMYYC __pyx_mstate_global->__pyx_n_s_DDMMYYC
#define __pyx_n_s_DDMMYYC10 __pyx_mstate_global->__pyx_n_s_DDMMYYC10
#define __pyx_n_s_DDMMYYD __pyx_mstate_global->__pyx_n_s_DDMMYYD
#define __pyx_n_s_DDMMYYD10 __pyx_mstate_global->__pyx_n_s_DDMMYYD10
#define __pyx_n_s_DDMMYYN6 __pyx_mstate_global->__pyx_n_s_DDMMYYN6
#define __pyx_n_s_DDMMYYN8 __pyx_mstate_global->__pyx_n_s_DDMMYYN8
#define __pyx_n_s_DDMMYYP __pyx_mstate_global->__pyx_n_s_DDMMYYP
#define __pyx_n_s_DDMMYYP10 __pyx_mstate_global->__pyx_n_s_DDMMYYP10
#define __pyx_n_s_DDMMYYS __pyx_mstate_global->__pyx_n_s_DDMMYYS
#define __pyx_n_s_DDMMYYS10 __pyx_mstate_global->__pyx_n_s_DDMMYYS10
#define __pyx_n_s_DTDATE __pyx_mstate_global->__pyx_n_s_DTDATE
#define __pyx_n_s_DTIME __pyx_mstate_global->__pyx_n_s_DTIME
#define __pyx_n_s_DataFrame __pyx_mstate_global->__pyx_n_s_DataFrame
#define __pyx_n_s_E8601DA __pyx_mstate_global->__pyx_n_s_E8601DA
#define __pyx_n_s_E8601DT __pyx_mstate_global->__pyx_n_s_E8601DT
#define __pyx_n_s_E8601TM __pyx_mstate_global->__pyx_n_s_E8601TM
#define __pyx_n_s_EDATE __pyx_mstate_global->__pyx_n_s_EDATE
#define __pyx_n_s_EDATE10 __pyx_mstate_global->__pyx_n_s_EDATE10
#define __pyx_n_s_EDATE8 __pyx_mstate_global->__pyx_n_s_EDATE8
#define __pyx_kp_s_Failed_convert_C_to_python_value __pyx_mstate_global->__pyx_kp_s_Failed_convert_C_to_python_value
#define __pyx_kp_s_Failed_to_read_number_of_variabl __pyx_mstate_global->__pyx_kp_s_Failed_to_read_number_of_variabl
#define __pyx_kp_s_File_0_does_not_exist __pyx_mstate_global->__pyx_kp_s_File_0_does_not_exist
#define __pyx_n_s_HHMM __pyx_mstate_global->__pyx_n_s_HHMM
#define __pyx_n_s_IS8601DA __pyx_mstate_global->__pyx_n_s_IS8601DA
#define __pyx_n_s_IS8601DT __pyx_mstate_global->__pyx_n_s_IS8601DT
#define __pyx_n_s_IS8601TM __pyx_mstate_global->__pyx_n_s_IS8601TM
#define __pyx_n_s_JDATE __pyx_mstate_global->__pyx_n_s_JDATE
#define __pyx_n_s_JDATE5 __pyx_mstate_global->__pyx_n_s_JDATE5
#define __pyx_n_s_JDATE7 __pyx_mstate_global->__pyx_n_s_JDATE7
#define __pyx_kp_s_Just_defining_a_custom_exceptio __pyx_mstate_global->__pyx_kp_s_Just_defining_a_custom_exceptio
#define __pyx_kp_s_Just_defining_a_custom_exceptio_2 __pyx_mstate_global->__pyx_kp_s_Just_defining_a_custom_exceptio_2
#define __pyx_kp_s_M8_ns __pyx_mstate_global->__pyx_kp_s_M8_ns
#define __pyx_n_s_MDYAMPM __pyx_mstate_global->__pyx_n_s_MDYAMPM
#define __pyx_n_s_MMDDYY __pyx_mstate_global->__pyx_n_s_MMDDYY
#define __pyx_n_s_MMDDYYB __pyx_mstate_global->__pyx_n_s_MMDDYYB
#define __pyx_n_s_MMDDYYB10 __pyx_mstate_global->__pyx_n_s_MMDDYYB10
#define __pyx_n_s_MMDDYYC __pyx_mstate_global->__pyx_n_s_MMDDYYC
#define __pyx_n_s_MMDDYYC10 __pyx_mstate_global->__pyx_n_s_MMDDYYC10
#define __pyx_n_s_MMDDYYD __pyx_mstate_global->__pyx_n_s_MMDDYYD
#define __pyx_n_s_MMDDYYD10 __pyx_mstate_global->__pyx_n_s_MMDDYYD10
#define __pyx_n_s_MMDDYYN6 __pyx_mstate_global->__pyx_n_s_MMDDYYN6
#define __pyx_n_s_MMDDYYN8 __pyx_mstate_global->__pyx_n_s_MMDDYYN8
#define __pyx_n_s_MMDDYYP __pyx_mstate_global->__pyx_n_s_MMDDYYP
#define __pyx_n_s_MMDDYYP10 __pyx_mstate_global->__pyx_n_s_MMDDYYP10
#define __pyx_n_s_MMDDYYS __pyx_mstate_global->__pyx_n_s_MMDDYYS
#define __pyx_n_s_MMDDYYS10 __pyx_mstate_global->__pyx_n_s_MMDDYYS10
#define __pyx_n_s_NULL __pyx_mstate_global->__pyx_n_s_NULL
#define __pyx_n_s_OrderedDict __pyx_mstate_global->__pyx_n_s_OrderedDict
#define __pyx_n_s_PyreadstatError __pyx_mstate_global->__pyx_n_s_PyreadstatError
#define __pyx_n_s_ReadstatError __pyx_mstate_global->__pyx_n_s_ReadstatError
#define __pyx_n_s_RuntimeError __pyx_mstate_global->__pyx_n_s_RuntimeError
#define __pyx_n_s_SDATE __pyx_mstate_global->__pyx_n_s_SDATE
#define __pyx_n_s_SDATE10 __pyx_mstate_global->__pyx_n_s_SDATE10
#define __pyx_n_s_SDATE8 __pyx_mstate_global->__pyx_n_s_SDATE8
#define __pyx_kp_s_STRING_type_with_value_s_with_da __pyx_mstate_global->__pyx_kp_s_STRING_type_with_value_s_with_da
#define __pyx_n_s_TIME __pyx_mstate_global->__pyx_n_s_TIME
#define __pyx_kp_s_TIME11_2 __pyx_mstate_global->__pyx_kp_s_TIME11_2
#define __pyx_n_s_TIME20 __pyx_mstate_global->__pyx_n_s_TIME20
#define __pyx_kp_s_TIME20_3 __pyx_mstate_global->__pyx_kp_s_TIME20_3
#define __pyx_n_s_TIME5 __pyx_mstate_global->__pyx_n_s_TIME5
#define __pyx_n_s_TIME8 __pyx_mstate_global->__pyx_n_s_TIME8
#define __pyx_n_s_TIMEAMPM __pyx_mstate_global->__pyx_n_s_TIMEAMPM
#define __pyx_n_s_TOD __pyx_mstate_global->__pyx_n_s_TOD
#define __pyx_kp_s_This_class_holds_metadata_we_wa __pyx_mstate_global->__pyx_kp_s_This_class_holds_metadata_we_wa
#define __pyx_kp_s_Time_zones_are_not_available_fro __pyx_mstate_global->__pyx_kp_s_Time_zones_are_not_available_fro
#define __pyx_n_s_TypeError __pyx_mstate_global->__pyx_n_s_TypeError
#define __pyx_n_s_UnicodeError __pyx_mstate_global->__pyx_n_s_UnicodeError
#define __pyx_kp_s_Unknown_file_format __pyx_mstate_global->__pyx_kp_s_Unknown_file_format
#define __pyx_kp_s_Unkown_data_type __pyx_mstate_global->__pyx_kp_s_Unkown_data_type
#define __pyx_n_s_WEEKDATE __pyx_mstate_global->__pyx_n_s_WEEKDATE
#define __pyx_n_s_WEEKDATX __pyx_mstate_global->__pyx_n_s_WEEKDATX
#define __pyx_n_s_YMDHMS16 __pyx_mstate_global->__pyx_n_s_YMDHMS16
#define __pyx_n_s_YMDHMS19 __pyx_mstate_global->__pyx_n_s_YMDHMS19
#define __pyx_kp_s_YMDHMS19_2 __pyx_mstate_global->__pyx_kp_s_YMDHMS19_2
#define __pyx_n_s_YMDHMS20 __pyx_mstate_global->__pyx_n_s_YMDHMS20
#define __pyx_n_s_YYMMDD __pyx_mstate_global->__pyx_n_s_YYMMDD
#define __pyx_n_s_YYMMDD10 __pyx_mstate_global->__pyx_n_s_YYMMDD10
#define __pyx_n_s_YYMMDDB __pyx_mstate_global->__pyx_n_s_YYMMDDB
#define __pyx_n_s_YYMMDDD __pyx_mstate_global->__pyx_n_s_YYMMDDD
#define __pyx_n_s_YYMMDDN __pyx_mstate_global->__pyx_n_s_YYMMDDN
#define __pyx_n_s_YYMMDDP __pyx_mstate_global->__pyx_n_s_YYMMDDP
#define __pyx_n_s_YYMMDDS __pyx_mstate_global->__pyx_n_s_YYMMDDS
#define __pyx_n_s__15 __pyx_mstate_global->__pyx_n_s__15
#define __pyx_kp_s__2 __pyx_mstate_global->__pyx_kp_s__2
#define __pyx_kp_u__8 __pyx_mstate_global->__pyx_kp_u__8
#define __pyx_n_s__9 __pyx_mstate_global->__pyx_n_s__9
#define __pyx_n_s_allowed_formats __pyx_mstate_global->__pyx_n_s_allowed_formats
#define __pyx_n_s_append __pyx_mstate_global->__pyx_n_s_append
#define __pyx_n_s_astype __pyx_mstate_global->__pyx_n_s_astype
#define __pyx_n_s_asyncio_coroutines __pyx_mstate_global->__pyx_n_s_asyncio_coroutines
#define __pyx_n_s_center __pyx_mstate_global->__pyx_n_s_center
#define __pyx_n_s_chr __pyx_mstate_global->__pyx_n_s_chr
#define __pyx_n_s_class_getitem __pyx_mstate_global->__pyx_n_s_class_getitem
#define __pyx_n_s_cline_in_traceback __pyx_mstate_global->__pyx_n_s_cline_in_traceback
#define __pyx_n_s_collections __pyx_mstate_global->__pyx_n_s_collections
#define __pyx_kp_s_column_0_is_duplicated_renamed_t __pyx_mstate_global->__pyx_kp_s_column_0_is_duplicated_renamed_t
#define __pyx_n_s_column_labels __pyx_mstate_global->__pyx_n_s_column_labels
#define __pyx_n_s_column_names __pyx_mstate_global->__pyx_n_s_column_names
#define __pyx_n_s_column_names_to_labels __pyx_mstate_global->__pyx_n_s_column_names_to_labels
#define __pyx_n_s_columns __pyx_mstate_global->__pyx_n_s_columns
#define __pyx_n_s_copy __pyx_mstate_global->__pyx_n_s_copy
#define __pyx_n_s_creation_time __pyx_mstate_global->__pyx_n_s_creation_time
#define __pyx_kp_s_d __pyx_mstate_global->__pyx_kp_s_d
#define __pyx_n_s_data_container __pyx_mstate_global->__pyx_n_s_data_container
#define __pyx_n_s_data_container___reduce_cython __pyx_mstate_global->__pyx_n_s_data_container___reduce_cython
#define __pyx_n_s_data_container___setstate_cython __pyx_mstate_global->__pyx_n_s_data_container___setstate_cython
#define __pyx_n_s_date __pyx_mstate_global->__pyx_n_s_date
#define __pyx_n_s_datetime __pyx_mstate_global->__pyx_n_s_datetime
#define __pyx_n_s_dict __pyx_mstate_global->__pyx_n_s_dict
#define __pyx_n_s_dict_2 __pyx_mstate_global->__pyx_n_s_dict_2
#define __pyx_kp_u_disable __pyx_mstate_global->__pyx_kp_u_disable
#define __pyx_n_s_doc __pyx_mstate_global->__pyx_n_s_doc
#define __pyx_n_s_double __pyx_mstate_global->__pyx_n_s_double
#define __pyx_n_s_dtype __pyx_mstate_global->__pyx_n_s_dtype
#define __pyx_n_s_dtypes __pyx_mstate_global->__pyx_n_s_dtypes
#define __pyx_n_s_duplicated __pyx_mstate_global->__pyx_n_s_duplicated
#define __pyx_n_s_empty __pyx_mstate_global->__pyx_n_s_empty
#define __pyx_kp_u_enable __pyx_mstate_global->__pyx_kp_u_enable
#define __pyx_n_s_encode __pyx_mstate_global->__pyx_n_s_encode
#define __pyx_n_s_enumerate __pyx_mstate_global->__pyx_n_s_enumerate
#define __pyx_n_s_expanduser __pyx_mstate_global->__pyx_n_s_expanduser
#define __pyx_n_s_file_encoding __pyx_mstate_global->__pyx_n_s_file_encoding
#define __pyx_n_s_file_label __pyx_mstate_global->__pyx_n_s_file_label
#define __pyx_kp_s_file_path_could_not_be_encoded_w __pyx_mstate_global->__pyx_kp_s_file_path_could_not_be_encoded_w
#define __pyx_n_s_fill __pyx_mstate_global->__pyx_n_s_fill
#define __pyx_n_s_float __pyx_mstate_global->__pyx_n_s_float
#define __pyx_n_s_float64 __pyx_mstate_global->__pyx_n_s_float64
#define __pyx_n_s_format __pyx_mstate_global->__pyx_n_s_format
#define __pyx_n_s_from_dict __pyx_mstate_global->__pyx_n_s_from_dict
#define __pyx_n_s_fromtimestamp __pyx_mstate_global->__pyx_n_s_fromtimestamp
#define __pyx_n_s_fsdecode __pyx_mstate_global->__pyx_n_s_fsdecode
#define __pyx_n_s_fsencode __pyx_mstate_global->__pyx_n_s_fsencode
#define __pyx_kp_u_gc __pyx_mstate_global->__pyx_kp_u_gc
#define __pyx_n_s_get __pyx_mstate_global->__pyx_n_s_get
#define __pyx_n_s_getfilesystemencoding __pyx_mstate_global->__pyx_n_s_getfilesystemencoding
#define __pyx_n_s_getstate __pyx_mstate_global->__pyx_n_s_getstate
#define __pyx_n_s_hi __pyx_mstate_global->__pyx_n_s_hi
#define __pyx_n_s_import __pyx_mstate_global->__pyx_n_s_import
#define __pyx_n_s_init __pyx_mstate_global->__pyx_n_s_init
#define __pyx_n_s_init_subclass __pyx_mstate_global->__pyx_n_s_init_subclass
#define __pyx_n_s_initializing __pyx_mstate_global->__pyx_n_s_initializing
#define __pyx_n_s_int16 __pyx_mstate_global->__pyx_n_s_int16
#define __pyx_n_s_int32 __pyx_mstate_global->__pyx_n_s_int32
#define __pyx_n_s_int64 __pyx_mstate_global->__pyx_n_s_int64
#define __pyx_n_s_int8 __pyx_mstate_global->__pyx_n_s_int8
#define __pyx_n_s_is_coroutine __pyx_mstate_global->__pyx_n_s_is_coroutine
#define __pyx_kp_u_isenabled __pyx_mstate_global->__pyx_kp_u_isenabled
#define __pyx_n_s_isfile __pyx_mstate_global->__pyx_n_s_isfile
#define __pyx_n_s_items __pyx_mstate_global->__pyx_n_s_items
#define __pyx_n_s_left __pyx_mstate_global->__pyx_n_s_left
#define __pyx_n_s_lo __pyx_mstate_global->__pyx_n_s_lo
#define __pyx_n_s_main __pyx_mstate_global->__pyx_n_s_main
#define __pyx_n_s_metaclass __pyx_mstate_global->__pyx_n_s_metaclass
#define __pyx_n_s_metadata_container __pyx_mstate_global->__pyx_n_s_metadata_container
#define __pyx_n_s_metadata_container___init __pyx_mstate_global->__pyx_n_s_metadata_container___init
#define __pyx_n_s_missing_ranges __pyx_mstate_global->__pyx_n_s_missing_ranges
#define __pyx_n_s_missing_user_values __pyx_mstate_global->__pyx_n_s_missing_user_values
#define __pyx_n_s_modification_time __pyx_mstate_global->__pyx_n_s_modification_time
#define __pyx_n_s_module __pyx_mstate_global->__pyx_n_s_module
#define __pyx_n_s_mro_entries __pyx_mstate_global->__pyx_n_s_mro_entries
#define __pyx_n_s_name __pyx_mstate_global->__pyx_n_s_name
#define __pyx_n_s_name_2 __pyx_mstate_global->__pyx_n_s_name_2
#define __pyx_n_s_nan __pyx_mstate_global->__pyx_n_s_nan
#define __pyx_kp_s_no_default___reduce___due_to_non __pyx_mstate_global->__pyx_kp_s_no_default___reduce___due_to_non
#define __pyx_n_s_nominal __pyx_mstate_global->__pyx_n_s_nominal
#define __pyx_n_s_notes __pyx_mstate_global->__pyx_n_s_notes
#define __pyx_n_s_np __pyx_mstate_global->__pyx_n_s_np
#define __pyx_n_s_nt __pyx_mstate_global->__pyx_n_s_nt
#define __pyx_n_s_number_columns __pyx_mstate_global->__pyx_n_s_number_columns
#define __pyx_n_s_number_rows __pyx_mstate_global->__pyx_n_s_number_rows
#define __pyx_n_s_numpy __pyx_mstate_global->__pyx_n_s_numpy
#define __pyx_n_s_object __pyx_mstate_global->__pyx_n_s_object
#define __pyx_n_s_ordinal __pyx_mstate_global->__pyx_n_s_ordinal
#define __pyx_n_s_original_variable_types __pyx_mstate_global->__pyx_n_s_original_variable_types
#define __pyx_n_s_os __pyx_mstate_global->__pyx_n_s_os
#define __pyx_n_s_output_format __pyx_mstate_global->__pyx_n_s_output_format
#define __pyx_kp_s_output_format_must_be_one_of_all __pyx_mstate_global->__pyx_kp_s_output_format_must_be_one_of_all
#define __pyx_n_s_pandas __pyx_mstate_global->__pyx_n_s_pandas
#define __pyx_n_s_path __pyx_mstate_global->__pyx_n_s_path
#define __pyx_kp_s_path_must_be_either_str_or_bytes __pyx_mstate_global->__pyx_kp_s_path_must_be_either_str_or_bytes
#define __pyx_kp_s_path_must_be_str_bytes_or_unicod __pyx_mstate_global->__pyx_kp_s_path_must_be_str_bytes_or_unicod
#define __pyx_n_s_pd __pyx_mstate_global->__pyx_n_s_pd
#define __pyx_n_s_prepare __pyx_mstate_global->__pyx_n_s_prepare
#define __pyx_n_s_pyreadstat__readstat_parser __pyx_mstate_global->__pyx_n_s_pyreadstat__readstat_parser
#define __pyx_kp_s_pyreadstat__readstat_parser_pyx __pyx_mstate_global->__pyx_kp_s_pyreadstat__readstat_parser_pyx
#define __pyx_n_s_pyx_capi __pyx_mstate_global->__pyx_n_s_pyx_capi
#define __pyx_n_s_pyx_state __pyx_mstate_global->__pyx_n_s_pyx_state
#define __pyx_n_s_qualname __pyx_mstate_global->__pyx_n_s_qualname
#define __pyx_n_s_range __pyx_mstate_global->__pyx_n_s_range
#define __pyx_n_s_readstat_to_numpy_types __pyx_mstate_global->__pyx_n_s_readstat_to_numpy_types
#define __pyx_n_s_readstat_variable_types __pyx_mstate_global->__pyx_n_s_readstat_variable_types
#define __pyx_n_s_reduce __pyx_mstate_global->__pyx_n_s_reduce
#define __pyx_n_s_reduce_cython __pyx_mstate_global->__pyx_n_s_reduce_cython
#define __pyx_n_s_reduce_ex __pyx_mstate_global->__pyx_n_s_reduce_ex
#define __pyx_n_s_right __pyx_mstate_global->__pyx_n_s_right
#define __pyx_n_s_sas_all_formats __pyx_mstate_global->__pyx_n_s_sas_all_formats
#define __pyx_n_s_sas_date_formats __pyx_mstate_global->__pyx_n_s_sas_date_formats
#define __pyx_n_s_sas_datetime_formats __pyx_mstate_global->__pyx_n_s_sas_datetime_formats
#define __pyx_n_s_sas_origin __pyx_mstate_global->__pyx_n_s_sas_origin
#define __pyx_n_s_sas_time_formats __pyx_mstate_global->__pyx_n_s_sas_time_formats
#define __pyx_n_s_scale __pyx_mstate_global->__pyx_n_s_scale
#define __pyx_n_s_self __pyx_mstate_global->__pyx_n_s_self
#define __pyx_n_s_set_name __pyx_mstate_global->__pyx_n_s_set_name
#define __pyx_n_s_setstate __pyx_mstate_global->__pyx_n_s_setstate
#define __pyx_n_s_setstate_cython __pyx_mstate_global->__pyx_n_s_setstate_cython
#define __pyx_n_s_spec __pyx_mstate_global->__pyx_n_s_spec
#define __pyx_n_s_spss_all_formats __pyx_mstate_global->__pyx_n_s_spss_all_formats
#define __pyx_n_s_spss_date_formats __pyx_mstate_global->__pyx_n_s_spss_date_formats
#define __pyx_n_s_spss_datetime_formats __pyx_mstate_global->__pyx_n_s_spss_datetime_formats
#define __pyx_n_s_spss_origin __pyx_mstate_global->__pyx_n_s_spss_origin
#define __pyx_n_s_spss_time_formats __pyx_mstate_global->__pyx_n_s_spss_time_formats
#define __pyx_n_s_stata_all_formats __pyx_mstate_global->__pyx_n_s_stata_all_formats
#define __pyx_n_s_stata_date_formats __pyx_mstate_global->__pyx_n_s_stata_date_formats
#define __pyx_n_s_stata_datetime_formats __pyx_mstate_global->__pyx_n_s_stata_datetime_formats
#define __pyx_n_s_stata_origin __pyx_mstate_global->__pyx_n_s_stata_origin
#define __pyx_n_s_stata_time_formats __pyx_mstate_global->__pyx_n_s_stata_time_formats
#define __pyx_n_s_string __pyx_mstate_global->__pyx_n_s_string
#define __pyx_kp_s_stringsource __pyx_mstate_global->__pyx_kp_s_stringsource
#define __pyx_n_s_super __pyx_mstate_global->__pyx_n_s_super
#define __pyx_n_s_surrogateescape __pyx_mstate_global->__pyx_n_s_surrogateescape
#define __pyx_n_s_sys __pyx_mstate_global->__pyx_n_s_sys
#define __pyx_kp_s_tC __pyx_mstate_global->__pyx_kp_s_tC
#define __pyx_n_s_table_name __pyx_mstate_global->__pyx_n_s_table_name
#define __pyx_kp_s_tc __pyx_mstate_global->__pyx_kp_s_tc
#define __pyx_kp_s_tcHH_MM __pyx_mstate_global->__pyx_kp_s_tcHH_MM
#define __pyx_kp_s_tcHH_MM_SS __pyx_mstate_global->__pyx_kp_s_tcHH_MM_SS
#define __pyx_kp_s_td __pyx_mstate_global->__pyx_kp_s_td
#define __pyx_kp_s_tdCCYY_NN_DD __pyx_mstate_global->__pyx_kp_s_tdCCYY_NN_DD
#define __pyx_kp_s_tdD_m_Y __pyx_mstate_global->__pyx_kp_s_tdD_m_Y
#define __pyx_n_s_test __pyx_mstate_global->__pyx_n_s_test
#define __pyx_n_s_time __pyx_mstate_global->__pyx_n_s_time
#define __pyx_n_s_to_datetime __pyx_mstate_global->__pyx_n_s_to_datetime
#define __pyx_n_s_tolist __pyx_mstate_global->__pyx_n_s_tolist
#define __pyx_n_s_undetermined __pyx_mstate_global->__pyx_n_s_undetermined
#define __pyx_n_s_unknown __pyx_mstate_global->__pyx_n_s_unknown
#define __pyx_kp_s_utf_8 __pyx_mstate_global->__pyx_kp_s_utf_8
#define __pyx_n_s_value_labels __pyx_mstate_global->__pyx_n_s_value_labels
#define __pyx_n_s_variable_alignment __pyx_mstate_global->__pyx_n_s_variable_alignment
#define __pyx_n_s_variable_display_width __pyx_mstate_global->__pyx_n_s_variable_display_width
#define __pyx_n_s_variable_measure __pyx_mstate_global->__pyx_n_s_variable_measure
#define __pyx_n_s_variable_storage_width __pyx_mstate_global->__pyx_n_s_variable_storage_width
#define __pyx_n_s_variable_to_label __pyx_mstate_global->__pyx_n_s_variable_to_label
#define __pyx_n_s_variable_value_labels __pyx_mstate_global->__pyx_n_s_variable_value_labels
#define __pyx_n_s_warn __pyx_mstate_global->__pyx_n_s_warn
#define __pyx_n_s_warnings __pyx_mstate_global->__pyx_n_s_warnings
#define __pyx_n_s_zip __pyx_mstate_global->__pyx_n_s_zip
#define __pyx_int_0 __pyx_mstate_global->__pyx_int_0
#define __pyx_int_1 __pyx_mstate_global->__pyx_int_1
#define __pyx_int_100000 __pyx_mstate_global->__pyx_int_100000
#define __pyx_tuple_ __pyx_mstate_global->__pyx_tuple_
#define __pyx_tuple__3 __pyx_mstate_global->__pyx_tuple__3
#define __pyx_tuple__4 __pyx_mstate_global->__pyx_tuple__4
#define __pyx_tuple__5 __pyx_mstate_global->__pyx_tuple__5
#define __pyx_tuple__6 __pyx_mstate_global->__pyx_tuple__6
#define __pyx_tuple__7 __pyx_mstate_global->__pyx_tuple__7
#define __pyx_tuple__10 __pyx_mstate_global->__pyx_tuple__10
#define __pyx_tuple__12 __pyx_mstate_global->__pyx_tuple__12
#define __pyx_codeobj__11 __pyx_mstate_global->__pyx_codeobj__11
#define __pyx_codeobj__13 __pyx_mstate_global->__pyx_codeobj__13
#define __pyx_codeobj__14 __pyx_mstate_global->__pyx_codeobj__14
/* #### Code section: module_code ### */

/* "cpython/datetime.pxd":72
 *     ctypedef extern class datetime.date[object PyDateTime_Date]:
 *         @property
 *         cdef inline int year(self):             # <<<<<<<<<<<<<<
 *             return PyDateTime_GET_YEAR(self)
 * 
 */

static CYTHON_INLINE int __pyx_f_7cpython_8datetime_4date_4year_year(PyDateTime_Date *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("year", 0);

  /* "cpython/datetime.pxd":73
 *         @property
 *         cdef inline int year(self):
 *             return PyDateTime_GET_YEAR(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
 */
  __pyx_r = PyDateTime_GET_YEAR(((PyObject *)__pyx_v_self));
  goto __pyx_L0;

  /* "cpython/datetime.pxd":72
 *     ctypedef extern class datetime.date[object PyDateTime_Date]:
 *         @property
 *         cdef inline int year(self):             # <<<<<<<<<<<<<<
 *             return PyDateTime_GET_YEAR(self)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cpython/datetime.pxd":76
 * 
 *         @property
 *         cdef inline int month(self):             # <<<<<<<<<<<<<<
 *             return PyDateTime_GET_MONTH(self)
 * 
 */

static CYTHON_INLINE int __pyx_f_7cpython_8datetime_4date_5month_month(PyDateTime_Date *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("month", 0);

  /* "cpython/datetime.pxd":77
 *         @property
 *         cdef inline int month(self):
 *             return PyDateTime_GET_MONTH(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
 */
  __pyx_r = PyDateTime_GET_MONTH(((PyObject *)__pyx_v_self));
  goto __pyx_L0;

  /* "cpython/datetime.pxd":76
 * 
 *         @property
 *         cdef inline int month(self):             # <<<<<<<<<<<<<<
 *             return PyDateTime_GET_MONTH(self)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cpython/datetime.pxd":80
 * 
 *         @property
 *         cdef inline int day(self):             # <<<<<<<<<<<<<<
 *             return PyDateTime_GET_DAY(self)
 * 
 */

static CYTHON_INLINE int __pyx_f_7cpython_8datetime_4date_3day_day(PyDateTime_Date *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("day", 0);

  /* "cpython/datetime.pxd":81
 *         @property
 *         cdef inline int day(self):
 *             return PyDateTime_GET_DAY(self)             # <<<<<<<<<<<<<<
 * 
 *     ctypedef extern class datetime.time[object PyDateTime_Time]:
 */
  __pyx_r = PyDateTime_GET_DAY(((PyObject *)__pyx_v_self));
  goto __pyx_L0;

  /* "cpython/datetime.pxd":80
 * 
 *         @property
 *         cdef inline int day(self):             # <<<<<<<<<<<<<<
 *             return PyDateTime_GET_DAY(self)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cpython/datetime.pxd":85
 *     ctypedef extern class datetime.time[object PyDateTime_Time]:
 *         @property
 *         cdef inline int hour(self):             # <<<<<<<<<<<<<<
 *             return PyDateTime_TIME_GET_HOUR(self)
 * 
 */

static CYTHON_INLINE int __pyx_f_7cpython_8datetime_4time_4hour_hour(PyDateTime_Time *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("hour", 0);

  /* "cpython/datetime.pxd":86
 *         @property
 *         cdef inline int hour(self):
 *             return PyDateTime_TIME_GET_HOUR(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
 */
  __pyx_r = PyDateTime_TIME_GET_HOUR(((PyObject *)__pyx_v_self));
  goto __pyx_L0;

  /* "cpython/datetime.pxd":85
 *     ctypedef extern class datetime.time[object PyDateTime_Time]:
 *         @property
 *         cdef inline int hour(self):             # <<<<<<<<<<<<<<
 *             return PyDateTime_TIME_GET_HOUR(self)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cpython/datetime.pxd":89
 * 
 *         @property
 *         cdef inline int minute(self):             # <<<<<<<<<<<<<<
 *             return PyDateTime_TIME_GET_MINUTE(self)
 * 
 */

static CYTHON_INLINE int __pyx_f_7cpython_8datetime_4time_6minute_minute(PyDateTime_Time *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("minute", 0);

  /* "cpython/datetime.pxd":90
 *         @property
 *         cdef inline int minute(self):
 *             return PyDateTime_TIME_GET_MINUTE(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
 */
  __pyx_r = PyDateTime_TIME_GET_MINUTE(((PyObject *)__pyx_v_self));
  goto __pyx_L0;

  /* "cpython/datetime.pxd":89
 * 
 *         @property
 *         cdef inline int minute(self):             # <<<<<<<<<<<<<<
 *             return PyDateTime_TIME_GET_MINUTE(self)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cpython/datetime.pxd":93
 * 
 *         @property
 *         cdef inline int second(self):             # <<<<<<<<<<<<<<
 *             return PyDateTime_TIME_GET_SECOND(self)
 * 
 */

static CYTHON_INLINE int __pyx_f_7cpython_8datetime_4time_6second_second(PyDateTime_Time *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("second", 0);

  /* "cpython/datetime.pxd":94
 *         @property
 *         cdef inline int second(self):
 *             return PyDateTime_TIME_GET_SECOND(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
 */
  __pyx_r = PyDateTime_TIME_GET_SECOND(((PyObject *)__pyx_v_self));
  goto __pyx_L0;

  /* "cpython/datetime.pxd":93
 * 
 *         @property
 *         cdef inline int second(self):             # <<<<<<<<<<<<<<
 *             return PyDateTime_TIME_GET_SECOND(self)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cpython/datetime.pxd":97
 * 
 *         @property
 *         cdef inline int microsecond(self):             # <<<<<<<<<<<<<<
 *             return PyDateTime_TIME_GET_MICROSECOND(self)
 * 
 */

static CYTHON_INLINE int __pyx_f_7cpython_8datetime_4time_11microsecond_microsecond(PyDateTime_Time *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("microsecond", 0);

  /* "cpython/datetime.pxd":98
 *         @property
 *         cdef inline int microsecond(self):
 *             return PyDateTime_TIME_GET_MICROSECOND(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
 */
  __pyx_r = PyDateTime_TIME_GET_MICROSECOND(((PyObject *)__pyx_v_self));
  goto __pyx_L0;

  /* "cpython/datetime.pxd":97
 * 
 *         @property
 *         cdef inline int microsecond(self):             # <<<<<<<<<<<<<<
 *             return PyDateTime_TIME_GET_MICROSECOND(self)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cpython/datetime.pxd":101
 * 
 *         @property
 *         cdef inline object tzinfo(self):             # <<<<<<<<<<<<<<
 *             return <object>PyDateTime_TIME_GET_TZINFO(self)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_7cpython_8datetime_4time_6tzinfo_tzinfo(PyDateTime_Time *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1;
  __Pyx_RefNannySetupContext("tzinfo", 0);

  /* "cpython/datetime.pxd":102
 *         @property
 *         cdef inline object tzinfo(self):
 *             return <object>PyDateTime_TIME_GET_TZINFO(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyDateTime_TIME_GET_TZINFO(((PyObject *)__pyx_v_self));
  __Pyx_INCREF(((PyObject *)__pyx_t_1));
  __pyx_r = ((PyObject *)__pyx_t_1);
  goto __pyx_L0;

  /* "cpython/datetime.pxd":101
 * 
 *         @property
 *         cdef inline object tzinfo(self):             # <<<<<<<<<<<<<<
 *             return <object>PyDateTime_TIME_GET_TZINFO(self)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cpython/datetime.pxd":105
 * 
 *         @property
 *         cdef inline int fold(self):             # <<<<<<<<<<<<<<
 *             # For Python < 3.6 this returns 0 no matter what
 *             return PyDateTime_TIME_GET_FOLD(self)
 */

static CYTHON_INLINE int __pyx_f_7cpython_8datetime_4time_4fold_fold(PyDateTime_Time *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("fold", 0);

  /* "cpython/datetime.pxd":107
 *         cdef inline int fold(self):
 *             # For Python < 3.6 this returns 0 no matter what
 *             return PyDateTime_TIME_GET_FOLD(self)             # <<<<<<<<<<<<<<
 * 
 *     ctypedef extern class datetime.datetime[object PyDateTime_DateTime]:
 */
  __pyx_r = PyDateTime_TIME_GET_FOLD(((PyObject *)__pyx_v_self));
  goto __pyx_L0;

  /* "cpython/datetime.pxd":105
 * 
 *         @property
 *         cdef inline int fold(self):             # <<<<<<<<<<<<<<
 *             # For Python < 3.6 this returns 0 no matter what
 *             return PyDateTime_TIME_GET_FOLD(self)
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cpython/datetime.pxd":111
 *     ctypedef extern class datetime.datetime[object PyDateTime_DateTime]:
 *         @property
 *         cdef inline int year(self):             # <<<<<<<<<<<<<<
 *             return PyDateTime_GET_YEAR(self)
 * 
 */

static CYTHON_INLINE int __pyx_f_7cpython_8datetime_8datetime_4year_year(PyDateTime_DateTime *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("year", 0);

  /* "cpython/datetime.pxd":112
 *         @property
 *         cdef inline int year(self):
 *             return PyDateTime_GET_YEAR(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
 */
  __pyx_r = PyDateTime_GET_YEAR(((PyObject *)__pyx_v_self));
  goto __pyx_L0;

  /* "cpython/datetime.pxd":111
 *     ctypedef extern class datetime.datetime[object PyDateTime_DateTime]:
 *         @property
 *         cdef inline int year(self):             # <<<<<<<<<<<<<<
 *             return PyDateTime_GET_YEAR(self)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cpython/datetime.pxd":115
 * 
 *         @property
 *         cdef inline int month(self):             # <<<<<<<<<<<<<<
 *             return PyDateTime_GET_MONTH(self)
 * 
 */

static CYTHON_INLINE int __pyx_f_7cpython_8datetime_8datetime_5month_month(PyDateTime_DateTime *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("month", 0);

  /* "cpython/datetime.pxd":116
 *         @property
 *         cdef inline int month(self):
 *             return PyDateTime_GET_MONTH(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
 */
  __pyx_r = PyDateTime_GET_MONTH(((PyObject *)__pyx_v_self));
  goto __pyx_L0;

  /* "cpython/datetime.pxd":115
 * 
 *         @property
 *         cdef inline int month(self):             # <<<<<<<<<<<<<<
 *             return PyDateTime_GET_MONTH(self)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cpython/datetime.pxd":119
 * 
 *         @property
 *         cdef inline int day(self):             # <<<<<<<<<<<<<<
 *             return PyDateTime_GET_DAY(self)
 * 
 */

static CYTHON_INLINE int __pyx_f_7cpython_8datetime_8datetime_3day_day(PyDateTime_DateTime *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("day", 0);

  /* "cpython/datetime.pxd":120
 *         @property
 *         cdef inline int day(self):
 *             return PyDateTime_GET_DAY(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
 */
  __pyx_r = PyDateTime_GET_DAY(((PyObject *)__pyx_v_self));
  goto __pyx_L0;

  /* "cpython/datetime.pxd":119
 * 
 *         @property
 *         cdef inline int day(self):             # <<<<<<<<<<<<<<
 *             return PyDateTime_GET_DAY(self)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cpython/datetime.pxd":123
 * 
 *         @property
 *         cdef inline int hour(self):             # <<<<<<<<<<<<<<
 *             return PyDateTime_DATE_GET_HOUR(self)
 * 
 */

static CYTHON_INLINE int __pyx_f_7cpython_8datetime_8datetime_4hour_hour(PyDateTime_DateTime *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("hour", 0);

  /* "cpython/datetime.pxd":124
 *         @property
 *         cdef inline int hour(self):
 *             return PyDateTime_DATE_GET_HOUR(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
 */
  __pyx_r = PyDateTime_DATE_GET_HOUR(((PyObject *)__pyx_v_self));
  goto __pyx_L0;

  /* "cpython/datetime.pxd":123
 * 
 *         @property
 *         cdef inline int hour(self):             # <<<<<<<<<<<<<<
 *             return PyDateTime_DATE_GET_HOUR(self)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cpython/datetime.pxd":127
 * 
 *         @property
 *         cdef inline int minute(self):             # <<<<<<<<<<<<<<
 *             return PyDateTime_DATE_GET_MINUTE(self)
 * 
 */

static CYTHON_INLINE int __pyx_f_7cpython_8datetime_8datetime_6minute_minute(PyDateTime_DateTime *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("minute", 0);

  /* "cpython/datetime.pxd":128
 *         @property
 *         cdef inline int minute(self):
 *             return PyDateTime_DATE_GET_MINUTE(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
 */
  __pyx_r = PyDateTime_DATE_GET_MINUTE(((PyObject *)__pyx_v_self));
  goto __pyx_L0;

  /* "cpython/datetime.pxd":127
 * 
 *         @property
 *         cdef inline int minute(self):             # <<<<<<<<<<<<<<
 *             return PyDateTime_DATE_GET_MINUTE(self)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cpython/datetime.pxd":131
 * 
 *         @property
 *         cdef inline int second(self):             # <<<<<<<<<<<<<<
 *             return PyDateTime_DATE_GET_SECOND(self)
 * 
 */

static CYTHON_INLINE int __pyx_f_7cpython_8datetime_8datetime_6second_second(PyDateTime_DateTime *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("second", 0);

  /* "cpython/datetime.pxd":132
 *         @property
 *         cdef inline int second(self):
 *             return PyDateTime_DATE_GET_SECOND(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
 */
  __pyx_r = PyDateTime_DATE_GET_SECOND(((PyObject *)__pyx_v_self));
  goto __pyx_L0;

  /* "cpython/datetime.pxd":131
 * 
 *         @property
 *         cdef inline int second(self):             # <<<<<<<<<<<<<<
 *             return PyDateTime_DATE_GET_SECOND(self)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cpython/datetime.pxd":135
 * 
 *         @property
 *         cdef inline int microsecond(self):             # <<<<<<<<<<<<<<
 *             return PyDateTime_DATE_GET_MICROSECOND(self)
 * 
 */

static CYTHON_INLINE int __pyx_f_7cpython_8datetime_8datetime_11microsecond_microsecond(PyDateTime_DateTime *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("microsecond", 0);

  /* "cpython/datetime.pxd":136
 *         @property
 *         cdef inline int microsecond(self):
 *             return PyDateTime_DATE_GET_MICROSECOND(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
 */
  __pyx_r = PyDateTime_DATE_GET_MICROSECOND(((PyObject *)__pyx_v_self));
  goto __pyx_L0;

  /* "cpython/datetime.pxd":135
 * 
 *         @property
 *         cdef inline int microsecond(self):             # <<<<<<<<<<<<<<
 *             return PyDateTime_DATE_GET_MICROSECOND(self)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cpython/datetime.pxd":139
 * 
 *         @property
 *         cdef inline object tzinfo(self):             # <<<<<<<<<<<<<<
 *             return <object>PyDateTime_DATE_GET_TZINFO(self)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_7cpython_8datetime_8datetime_6tzinfo_tzinfo(PyDateTime_DateTime *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1;
  __Pyx_RefNannySetupContext("tzinfo", 0);

  /* "cpython/datetime.pxd":140
 *         @property
 *         cdef inline object tzinfo(self):
 *             return <object>PyDateTime_DATE_GET_TZINFO(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyDateTime_DATE_GET_TZINFO(((PyObject *)__pyx_v_self));
  __Pyx_INCREF(((PyObject *)__pyx_t_1));
  __pyx_r = ((PyObject *)__pyx_t_1);
  goto __pyx_L0;

  /* "cpython/datetime.pxd":139
 * 
 *         @property
 *         cdef inline object tzinfo(self):             # <<<<<<<<<<<<<<
 *             return <object>PyDateTime_DATE_GET_TZINFO(self)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cpython/datetime.pxd":143
 * 
 *         @property
 *         cdef inline int fold(self):             # <<<<<<<<<<<<<<
 *             # For Python < 3.6 this returns 0 no matter what
 *             return PyDateTime_DATE_GET_FOLD(self)
 */

static CYTHON_INLINE int __pyx_f_7cpython_8datetime_8datetime_4fold_fold(PyDateTime_DateTime *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("fold", 0);

  /* "cpython/datetime.pxd":145
 *         cdef inline int fold(self):
 *             # For Python < 3.6 this returns 0 no matter what
 *             return PyDateTime_DATE_GET_FOLD(self)             # <<<<<<<<<<<<<<
 * 
 *     ctypedef extern class datetime.timedelta[object PyDateTime_Delta]:
 */
  __pyx_r = PyDateTime_DATE_GET_FOLD(((PyObject *)__pyx_v_self));
  goto __pyx_L0;

  /* "cpython/datetime.pxd":143
 * 
 *         @property
 *         cdef inline int fold(self):             # <<<<<<<<<<<<<<
 *             # For Python < 3.6 this returns 0 no matter what
 *             return PyDateTime_DATE_GET_FOLD(self)
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cpython/datetime.pxd":149
 *     ctypedef extern class datetime.timedelta[object PyDateTime_Delta]:
 *         @property
 *         cdef inline int day(self):             # <<<<<<<<<<<<<<
 *             return PyDateTime_DELTA_GET_DAYS(self)
 * 
 */

static CYTHON_INLINE int __pyx_f_7cpython_8datetime_9timedelta_3day_day(PyDateTime_Delta *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("day", 0);

  /* "cpython/datetime.pxd":150
 *         @property
 *         cdef inline int day(self):
 *             return PyDateTime_DELTA_GET_DAYS(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
 */
  __pyx_r = PyDateTime_DELTA_GET_DAYS(((PyObject *)__pyx_v_self));
  goto __pyx_L0;

  /* "cpython/datetime.pxd":149
 *     ctypedef extern class datetime.timedelta[object PyDateTime_Delta]:
 *         @property
 *         cdef inline int day(self):             # <<<<<<<<<<<<<<
 *             return PyDateTime_DELTA_GET_DAYS(self)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cpython/datetime.pxd":153
 * 
 *         @property
 *         cdef inline int second(self):             # <<<<<<<<<<<<<<
 *             return PyDateTime_DELTA_GET_SECONDS(self)
 * 
 */

static CYTHON_INLINE int __pyx_f_7cpython_8datetime_9timedelta_6second_second(PyDateTime_Delta *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("second", 0);

  /* "cpython/datetime.pxd":154
 *         @property
 *         cdef inline int second(self):
 *             return PyDateTime_DELTA_GET_SECONDS(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
 */
  __pyx_r = PyDateTime_DELTA_GET_SECONDS(((PyObject *)__pyx_v_self));
  goto __pyx_L0;

  /* "cpython/datetime.pxd":153
 * 
 *         @property
 *         cdef inline int second(self):             # <<<<<<<<<<<<<<
 *             return PyDateTime_DELTA_GET_SECONDS(self)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cpython/datetime.pxd":157
 * 
 *         @property
 *         cdef inline int microsecond(self):             # <<<<<<<<<<<<<<
 *             return PyDateTime_DELTA_GET_MICROSECONDS(self)
 * 
 */

static CYTHON_INLINE int __pyx_f_7cpython_8datetime_9timedelta_11microsecond_microsecond(PyDateTime_Delta *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("microsecond", 0);

  /* "cpython/datetime.pxd":158
 *         @property
 *         cdef inline int microsecond(self):
 *             return PyDateTime_DELTA_GET_MICROSECONDS(self)             # <<<<<<<<<<<<<<
 * 
 *     ctypedef extern class datetime.tzinfo[object PyDateTime_TZInfo]:
 */
  __pyx_r = PyDateTime_DELTA_GET_MICROSECONDS(((PyObject *)__pyx_v_self));
  goto __pyx_L0;

  /* "cpython/datetime.pxd":157
 * 
 *         @property
 *         cdef inline int microsecond(self):             # <<<<<<<<<<<<<<
 *             return PyDateTime_DELTA_GET_MICROSECONDS(self)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cpython/datetime.pxd":282
 * # Datetime C API initialization function.
 * # You have to call it before any usage of DateTime CAPI functions.
 * cdef inline void import_datetime():             # <<<<<<<<<<<<<<
 *     PyDateTime_IMPORT
 * 
 */

static CYTHON_INLINE void __pyx_f_7cpython_8datetime_import_datetime(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("import_datetime", 0);

  /* "cpython/datetime.pxd":283
 * # You have to call it before any usage of DateTime CAPI functions.
 * cdef inline void import_datetime():
 *     PyDateTime_IMPORT             # <<<<<<<<<<<<<<
 * 
 * # Create date object using DateTime CAPI factory function.
 */
  (void)(PyDateTime_IMPORT);

  /* "cpython/datetime.pxd":282
 * # Datetime C API initialization function.
 * # You have to call it before any usage of DateTime CAPI functions.
 * cdef inline void import_datetime():             # <<<<<<<<<<<<<<
 *     PyDateTime_IMPORT
 * 
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "cpython/datetime.pxd":287
 * # Create date object using DateTime CAPI factory function.
 * # Note, there are no range checks for any of the arguments.
 * cdef inline date date_new(int year, int month, int day):             # <<<<<<<<<<<<<<
 *     return PyDateTimeAPI.Date_FromDate(year, month, day, PyDateTimeAPI.DateType)
 * 
 */

static CYTHON_INLINE PyDateTime_Date *__pyx_f_7cpython_8datetime_date_new(int __pyx_v_year, int __pyx_v_month, int __pyx_v_day) {
  PyDateTime_Date *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("date_new", 0);

  /* "cpython/datetime.pxd":288
 * # Note, there are no range checks for any of the arguments.
 * cdef inline date date_new(int year, int month, int day):
 *     return PyDateTimeAPI.Date_FromDate(year, month, day, PyDateTimeAPI.DateType)             # <<<<<<<<<<<<<<
 * 
 * # Create time object using DateTime CAPI factory function
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __pyx_t_1 = ((PyObject *)PyDateTimeAPI->Date_FromDate(__pyx_v_year, __pyx_v_month, __pyx_v_day, PyDateTimeAPI->DateType)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 288, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = ((PyDateTime_Date *)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "cpython/datetime.pxd":287
 * # Create date object using DateTime CAPI factory function.
 * # Note, there are no range checks for any of the arguments.
 * cdef inline date date_new(int year, int month, int day):             # <<<<<<<<<<<<<<
 *     return PyDateTimeAPI.Date_FromDate(year, month, day, PyDateTimeAPI.DateType)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cpython.datetime.date_new", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cpython/datetime.pxd":292
 * # Create time object using DateTime CAPI factory function
 * # Note, there are no range checks for any of the arguments.
 * cdef inline time time_new(int hour, int minute, int second, int microsecond, object tz, int fold=0):             # <<<<<<<<<<<<<<
 *     return __Pyx_DateTime_TimeWithFold(hour, minute, second, microsecond, tz, fold)
 * 
 */

static CYTHON_INLINE PyDateTime_Time *__pyx_f_7cpython_8datetime_time_new(int __pyx_v_hour, int __pyx_v_minute, int __pyx_v_second, int __pyx_v_microsecond, PyObject *__pyx_v_tz, struct __pyx_opt_args_7cpython_8datetime_time_new *__pyx_optional_args) {
  int __pyx_v_fold = ((int)0);
  PyDateTime_Time *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("time_new", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_fold = __pyx_optional_args->fold;
    }
  }

  /* "cpython/datetime.pxd":293
 * # Note, there are no range checks for any of the arguments.
 * cdef inline time time_new(int hour, int minute, int second, int microsecond, object tz, int fold=0):
 *     return __Pyx_DateTime_TimeWithFold(hour, minute, second, microsecond, tz, fold)             # <<<<<<<<<<<<<<
 * 
 * # Create datetime object using DateTime CAPI factory function.
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __pyx_t_1 = ((PyObject *)__Pyx_DateTime_TimeWithFold(__pyx_v_hour, __pyx_v_minute, __pyx_v_second, __pyx_v_microsecond, __pyx_v_tz, __pyx_v_fold)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 293, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_7cpython_8datetime_time))))) __PYX_ERR(2, 293, __pyx_L1_error)
  __pyx_r = ((PyDateTime_Time *)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "cpython/datetime.pxd":292
 * # Create time object using DateTime CAPI factory function
 * # Note, there are no range checks for any of the arguments.
 * cdef inline time time_new(int hour, int minute, int second, int microsecond, object tz, int fold=0):             # <<<<<<<<<<<<<<
 *     return __Pyx_DateTime_TimeWithFold(hour, minute, second, microsecond, tz, fold)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cpython.datetime.time_new", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cpython/datetime.pxd":297
 * # Create datetime object using DateTime CAPI factory function.
 * # Note, there are no range checks for any of the arguments.
 * cdef inline datetime datetime_new(int year, int month, int day, int hour, int minute, int second, int microsecond, object tz, int fold=0):             # <<<<<<<<<<<<<<
 *     return __Pyx_DateTime_DateTimeWithFold(year, month, day, hour, minute, second, microsecond, tz, fold)
 * 
 */

static CYTHON_INLINE PyDateTime_DateTime *__pyx_f_7cpython_8datetime_datetime_new(int __pyx_v_year, int __pyx_v_month, int __pyx_v_day, int __pyx_v_hour, int __pyx_v_minute, int __pyx_v_second, int __pyx_v_microsecond, PyObject *__pyx_v_tz, struct __pyx_opt_args_7cpython_8datetime_datetime_new *__pyx_optional_args) {
  int __pyx_v_fold = ((int)0);
  PyDateTime_DateTime *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("datetime_new", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_fold = __pyx_optional_args->fold;
    }
  }

  /* "cpython/datetime.pxd":298
 * # Note, there are no range checks for any of the arguments.
 * cdef inline datetime datetime_new(int year, int month, int day, int hour, int minute, int second, int microsecond, object tz, int fold=0):
 *     return __Pyx_DateTime_DateTimeWithFold(year, month, day, hour, minute, second, microsecond, tz, fold)             # <<<<<<<<<<<<<<
 * 
 * # Create timedelta object using DateTime CAPI factory function.
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __pyx_t_1 = ((PyObject *)__Pyx_DateTime_DateTimeWithFold(__pyx_v_year, __pyx_v_month, __pyx_v_day, __pyx_v_hour, __pyx_v_minute, __pyx_v_second, __pyx_v_microsecond, __pyx_v_tz, __pyx_v_fold)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 298, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = ((PyDateTime_DateTime *)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "cpython/datetime.pxd":297
 * # Create datetime object using DateTime CAPI factory function.
 * # Note, there are no range checks for any of the arguments.
 * cdef inline datetime datetime_new(int year, int month, int day, int hour, int minute, int second, int microsecond, object tz, int fold=0):             # <<<<<<<<<<<<<<
 *     return __Pyx_DateTime_DateTimeWithFold(year, month, day, hour, minute, second, microsecond, tz, fold)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cpython.datetime.datetime_new", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cpython/datetime.pxd":302
 * # Create timedelta object using DateTime CAPI factory function.
 * # Note, there are no range checks for any of the arguments.
 * cdef inline timedelta timedelta_new(int days, int seconds, int useconds):             # <<<<<<<<<<<<<<
 *     return PyDateTimeAPI.Delta_FromDelta(days, seconds, useconds, 1, PyDateTimeAPI.DeltaType)
 * 
 */

static CYTHON_INLINE PyDateTime_Delta *__pyx_f_7cpython_8datetime_timedelta_new(int __pyx_v_days, int __pyx_v_seconds, int __pyx_v_useconds) {
  PyDateTime_Delta *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("timedelta_new", 0);

  /* "cpython/datetime.pxd":303
 * # Note, there are no range checks for any of the arguments.
 * cdef inline timedelta timedelta_new(int days, int seconds, int useconds):
 *     return PyDateTimeAPI.Delta_FromDelta(days, seconds, useconds, 1, PyDateTimeAPI.DeltaType)             # <<<<<<<<<<<<<<
 * 
 * # Create timedelta object using DateTime CAPI factory function.
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __pyx_t_1 = ((PyObject *)PyDateTimeAPI->Delta_FromDelta(__pyx_v_days, __pyx_v_seconds, __pyx_v_useconds, 1, PyDateTimeAPI->DeltaType)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 303, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = ((PyDateTime_Delta *)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "cpython/datetime.pxd":302
 * # Create timedelta object using DateTime CAPI factory function.
 * # Note, there are no range checks for any of the arguments.
 * cdef inline timedelta timedelta_new(int days, int seconds, int useconds):             # <<<<<<<<<<<<<<
 *     return PyDateTimeAPI.Delta_FromDelta(days, seconds, useconds, 1, PyDateTimeAPI.DeltaType)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cpython.datetime.timedelta_new", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cpython/datetime.pxd":306
 * 
 * # Create timedelta object using DateTime CAPI factory function.
 * cdef inline object timezone_new(object offset, object name=None):             # <<<<<<<<<<<<<<
 *     if PY_VERSION_HEX < 0x030700b1:
 *         raise RuntimeError('Time zones are not available from the C-API.')
 */

static CYTHON_INLINE PyObject *__pyx_f_7cpython_8datetime_timezone_new(PyObject *__pyx_v_offset, struct __pyx_opt_args_7cpython_8datetime_timezone_new *__pyx_optional_args) {
  PyObject *__pyx_v_name = ((PyObject *)Py_None);
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("timezone_new", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_name = __pyx_optional_args->name;
    }
  }

  /* "cpython/datetime.pxd":307
 * # Create timedelta object using DateTime CAPI factory function.
 * cdef inline object timezone_new(object offset, object name=None):
 *     if PY_VERSION_HEX < 0x030700b1:             # <<<<<<<<<<<<<<
 *         raise RuntimeError('Time zones are not available from the C-API.')
 *     return __Pyx_TimeZone_FromOffsetAndName(offset, <PyObject*>name if name is not None else NULL)
 */
  __pyx_t_1 = (PY_VERSION_HEX < 0x030700b1);
  if (unlikely(__pyx_t_1)) {

    /* "cpython/datetime.pxd":308
 * cdef inline object timezone_new(object offset, object name=None):
 *     if PY_VERSION_HEX < 0x030700b1:
 *         raise RuntimeError('Time zones are not available from the C-API.')             # <<<<<<<<<<<<<<
 *     return __Pyx_TimeZone_FromOffsetAndName(offset, <PyObject*>name if name is not None else NULL)
 * 
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple_, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 308, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(2, 308, __pyx_L1_error)

    /* "cpython/datetime.pxd":307
 * # Create timedelta object using DateTime CAPI factory function.
 * cdef inline object timezone_new(object offset, object name=None):
 *     if PY_VERSION_HEX < 0x030700b1:             # <<<<<<<<<<<<<<
 *         raise RuntimeError('Time zones are not available from the C-API.')
 *     return __Pyx_TimeZone_FromOffsetAndName(offset, <PyObject*>name if name is not None else NULL)
 */
  }

  /* "cpython/datetime.pxd":309
 *     if PY_VERSION_HEX < 0x030700b1:
 *         raise RuntimeError('Time zones are not available from the C-API.')
 *     return __Pyx_TimeZone_FromOffsetAndName(offset, <PyObject*>name if name is not None else NULL)             # <<<<<<<<<<<<<<
 * 
 * # Create datetime object using DB API constructor.
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = (__pyx_v_name != Py_None);
  if (__pyx_t_1) {
    __pyx_t_3 = ((PyObject *)__pyx_v_name);
  } else {
    __pyx_t_3 = NULL;
  }
  __pyx_t_2 = __Pyx_TimeZone_FromOffsetAndName(__pyx_v_offset, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 309, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "cpython/datetime.pxd":306
 * 
 * # Create timedelta object using DateTime CAPI factory function.
 * cdef inline object timezone_new(object offset, object name=None):             # <<<<<<<<<<<<<<
 *     if PY_VERSION_HEX < 0x030700b1:
 *         raise RuntimeError('Time zones are not available from the C-API.')
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cpython.datetime.timezone_new", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cpython/datetime.pxd":312
 * 
 * # Create datetime object using DB API constructor.
 * cdef inline datetime datetime_from_timestamp(timestamp, tz=None):             # <<<<<<<<<<<<<<
 *     return PyDateTimeAPI.DateTime_FromTimestamp(
 *         <PyObject*>PyDateTimeAPI.DateTimeType, (timestamp, tz) if tz is not None else (timestamp,), NULL)
 */

static CYTHON_INLINE PyDateTime_DateTime *__pyx_f_7cpython_8datetime_datetime_from_timestamp(PyObject *__pyx_v_timestamp, struct __pyx_opt_args_7cpython_8datetime_datetime_from_timestamp *__pyx_optional_args) {
  PyObject *__pyx_v_tz = ((PyObject *)Py_None);
  PyDateTime_DateTime *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("datetime_from_timestamp", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_tz = __pyx_optional_args->tz;
    }
  }

  /* "cpython/datetime.pxd":313
 * # Create datetime object using DB API constructor.
 * cdef inline datetime datetime_from_timestamp(timestamp, tz=None):
 *     return PyDateTimeAPI.DateTime_FromTimestamp(             # <<<<<<<<<<<<<<
 *         <PyObject*>PyDateTimeAPI.DateTimeType, (timestamp, tz) if tz is not None else (timestamp,), NULL)
 * 
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);

  /* "cpython/datetime.pxd":314
 * cdef inline datetime datetime_from_timestamp(timestamp, tz=None):
 *     return PyDateTimeAPI.DateTime_FromTimestamp(
 *         <PyObject*>PyDateTimeAPI.DateTimeType, (timestamp, tz) if tz is not None else (timestamp,), NULL)             # <<<<<<<<<<<<<<
 * 
 * # Create date object using DB API constructor.
 */
  __pyx_t_2 = (__pyx_v_tz != Py_None);
  if (__pyx_t_2) {
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 314, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_v_timestamp);
    __Pyx_GIVEREF(__pyx_v_timestamp);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_timestamp);
    __Pyx_INCREF(__pyx_v_tz);
    __Pyx_GIVEREF(__pyx_v_tz);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_tz);
    __pyx_t_1 = __pyx_t_3;
    __pyx_t_3 = 0;
  } else {
    __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 314, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_v_timestamp);
    __Pyx_GIVEREF(__pyx_v_timestamp);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_timestamp);
    __pyx_t_1 = __pyx_t_3;
    __pyx_t_3 = 0;
  }

  /* "cpython/datetime.pxd":313
 * # Create datetime object using DB API constructor.
 * cdef inline datetime datetime_from_timestamp(timestamp, tz=None):
 *     return PyDateTimeAPI.DateTime_FromTimestamp(             # <<<<<<<<<<<<<<
 *         <PyObject*>PyDateTimeAPI.DateTimeType, (timestamp, tz) if tz is not None else (timestamp,), NULL)
 * 
 */
  __pyx_t_3 = ((PyObject *)PyDateTimeAPI->DateTime_FromTimestamp(((PyObject *)PyDateTimeAPI->DateTimeType), __pyx_t_1, NULL)); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 313, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = ((PyDateTime_DateTime *)__pyx_t_3);
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "cpython/datetime.pxd":312
 * 
 * # Create datetime object using DB API constructor.
 * cdef inline datetime datetime_from_timestamp(timestamp, tz=None):             # <<<<<<<<<<<<<<
 *     return PyDateTimeAPI.DateTime_FromTimestamp(
 *         <PyObject*>PyDateTimeAPI.DateTimeType, (timestamp, tz) if tz is not None else (timestamp,), NULL)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("cpython.datetime.datetime_from_timestamp", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cpython/datetime.pxd":317
 * 
 * # Create date object using DB API constructor.
 * cdef inline date date_from_timestamp(timestamp):             # <<<<<<<<<<<<<<
 *     return PyDateTimeAPI.Date_FromTimestamp(<PyObject*>PyDateTimeAPI.DateType, (timestamp,))
 * 
 */

static CYTHON_INLINE PyDateTime_Date *__pyx_f_7cpython_8datetime_date_from_timestamp(PyObject *__pyx_v_timestamp) {
  PyDateTime_Date *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("date_from_timestamp", 0);

  /* "cpython/datetime.pxd":318
 * # Create date object using DB API constructor.
 * cdef inline date date_from_timestamp(timestamp):
 *     return PyDateTimeAPI.Date_FromTimestamp(<PyObject*>PyDateTimeAPI.DateType, (timestamp,))             # <<<<<<<<<<<<<<
 * 
 * # More recognizable getters for date/time/datetime/timedelta.
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 318, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_timestamp);
  __Pyx_GIVEREF(__pyx_v_timestamp);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_timestamp);
  __pyx_t_2 = ((PyObject *)PyDateTimeAPI->Date_FromTimestamp(((PyObject *)PyDateTimeAPI->DateType), __pyx_t_1)); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 318, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = ((PyDateTime_Date *)__pyx_t_2);
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "cpython/datetime.pxd":317
 * 
 * # Create date object using DB API constructor.
 * cdef inline date date_from_timestamp(timestamp):             # <<<<<<<<<<<<<<
 *     return PyDateTimeAPI.Date_FromTimestamp(<PyObject*>PyDateTimeAPI.DateType, (timestamp,))
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cpython.datetime.date_from_timestamp", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cpython/datetime.pxd":326
 * 
 * # Get UTC singleton
 * cdef inline object get_utc():             # <<<<<<<<<<<<<<
 *     if PY_VERSION_HEX < 0x030700b1:
 *         raise RuntimeError('Time zones are not available from the C-API.')
 */

static CYTHON_INLINE PyObject *__pyx_f_7cpython_8datetime_get_utc(void) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_utc", 0);

  /* "cpython/datetime.pxd":327
 * # Get UTC singleton
 * cdef inline object get_utc():
 *     if PY_VERSION_HEX < 0x030700b1:             # <<<<<<<<<<<<<<
 *         raise RuntimeError('Time zones are not available from the C-API.')
 *     return <object>__Pyx_TimeZone_UTC
 */
  __pyx_t_1 = (PY_VERSION_HEX < 0x030700b1);
  if (unlikely(__pyx_t_1)) {

    /* "cpython/datetime.pxd":328
 * cdef inline object get_utc():
 *     if PY_VERSION_HEX < 0x030700b1:
 *         raise RuntimeError('Time zones are not available from the C-API.')             # <<<<<<<<<<<<<<
 *     return <object>__Pyx_TimeZone_UTC
 * 
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple_, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 328, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(2, 328, __pyx_L1_error)

    /* "cpython/datetime.pxd":327
 * # Get UTC singleton
 * cdef inline object get_utc():
 *     if PY_VERSION_HEX < 0x030700b1:             # <<<<<<<<<<<<<<
 *         raise RuntimeError('Time zones are not available from the C-API.')
 *     return <object>__Pyx_TimeZone_UTC
 */
  }

  /* "cpython/datetime.pxd":329
 *     if PY_VERSION_HEX < 0x030700b1:
 *         raise RuntimeError('Time zones are not available from the C-API.')
 *     return <object>__Pyx_TimeZone_UTC             # <<<<<<<<<<<<<<
 * 
 * # Get tzinfo of time
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__Pyx_TimeZone_UTC));
  __pyx_r = ((PyObject *)__Pyx_TimeZone_UTC);
  goto __pyx_L0;

  /* "cpython/datetime.pxd":326
 * 
 * # Get UTC singleton
 * cdef inline object get_utc():             # <<<<<<<<<<<<<<
 *     if PY_VERSION_HEX < 0x030700b1:
 *         raise RuntimeError('Time zones are not available from the C-API.')
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cpython.datetime.get_utc", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cpython/datetime.pxd":332
 * 
 * # Get tzinfo of time
 * cdef inline object time_tzinfo(object o):             # <<<<<<<<<<<<<<
 *     return <object>PyDateTime_TIME_GET_TZINFO(o)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_7cpython_8datetime_time_tzinfo(PyObject *__pyx_v_o) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1;
  __Pyx_RefNannySetupContext("time_tzinfo", 0);

  /* "cpython/datetime.pxd":333
 * # Get tzinfo of time
 * cdef inline object time_tzinfo(object o):
 *     return <object>PyDateTime_TIME_GET_TZINFO(o)             # <<<<<<<<<<<<<<
 * 
 * # Get tzinfo of datetime
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyDateTime_TIME_GET_TZINFO(__pyx_v_o);
  __Pyx_INCREF(((PyObject *)__pyx_t_1));
  __pyx_r = ((PyObject *)__pyx_t_1);
  goto __pyx_L0;

  /* "cpython/datetime.pxd":332
 * 
 * # Get tzinfo of time
 * cdef inline object time_tzinfo(object o):             # <<<<<<<<<<<<<<
 *     return <object>PyDateTime_TIME_GET_TZINFO(o)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cpython/datetime.pxd":336
 * 
 * # Get tzinfo of datetime
 * cdef inline object datetime_tzinfo(object o):             # <<<<<<<<<<<<<<
 *     return <object>PyDateTime_DATE_GET_TZINFO(o)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_7cpython_8datetime_datetime_tzinfo(PyObject *__pyx_v_o) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1;
  __Pyx_RefNannySetupContext("datetime_tzinfo", 0);

  /* "cpython/datetime.pxd":337
 * # Get tzinfo of datetime
 * cdef inline object datetime_tzinfo(object o):
 *     return <object>PyDateTime_DATE_GET_TZINFO(o)             # <<<<<<<<<<<<<<
 * 
 * # Get year of date
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyDateTime_DATE_GET_TZINFO(__pyx_v_o);
  __Pyx_INCREF(((PyObject *)__pyx_t_1));
  __pyx_r = ((PyObject *)__pyx_t_1);
  goto __pyx_L0;

  /* "cpython/datetime.pxd":336
 * 
 * # Get tzinfo of datetime
 * cdef inline object datetime_tzinfo(object o):             # <<<<<<<<<<<<<<
 *     return <object>PyDateTime_DATE_GET_TZINFO(o)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cpython/datetime.pxd":340
 * 
 * # Get year of date
 * cdef inline int date_year(object o):             # <<<<<<<<<<<<<<
 *     return PyDateTime_GET_YEAR(o)
 * 
 */

static CYTHON_INLINE int __pyx_f_7cpython_8datetime_date_year(PyObject *__pyx_v_o) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("date_year", 0);

  /* "cpython/datetime.pxd":341
 * # Get year of date
 * cdef inline int date_year(object o):
 *     return PyDateTime_GET_YEAR(o)             # <<<<<<<<<<<<<<
 * 
 * # Get month of date
 */
  __pyx_r = PyDateTime_GET_YEAR(__pyx_v_o);
  goto __pyx_L0;

  /* "cpython/datetime.pxd":340
 * 
 * # Get year of date
 * cdef inline int date_year(object o):             # <<<<<<<<<<<<<<
 *     return PyDateTime_GET_YEAR(o)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cpython/datetime.pxd":344
 * 
 * # Get month of date
 * cdef inline int date_month(object o):             # <<<<<<<<<<<<<<
 *     return PyDateTime_GET_MONTH(o)
 * 
 */

static CYTHON_INLINE int __pyx_f_7cpython_8datetime_date_month(PyObject *__pyx_v_o) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("date_month", 0);

  /* "cpython/datetime.pxd":345
 * # Get month of date
 * cdef inline int date_month(object o):
 *     return PyDateTime_GET_MONTH(o)             # <<<<<<<<<<<<<<
 * 
 * # Get day of date
 */
  __pyx_r = PyDateTime_GET_MONTH(__pyx_v_o);
  goto __pyx_L0;

  /* "cpython/datetime.pxd":344
 * 
 * # Get month of date
 * cdef inline int date_month(object o):             # <<<<<<<<<<<<<<
 *     return PyDateTime_GET_MONTH(o)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cpython/datetime.pxd":348
 * 
 * # Get day of date
 * cdef inline int date_day(object o):             # <<<<<<<<<<<<<<
 *     return PyDateTime_GET_DAY(o)
 * 
 */

static CYTHON_INLINE int __pyx_f_7cpython_8datetime_date_day(PyObject *__pyx_v_o) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("date_day", 0);

  /* "cpython/datetime.pxd":349
 * # Get day of date
 * cdef inline int date_day(object o):
 *     return PyDateTime_GET_DAY(o)             # <<<<<<<<<<<<<<
 * 
 * # Get year of datetime
 */
  __pyx_r = PyDateTime_GET_DAY(__pyx_v_o);
  goto __pyx_L0;

  /* "cpython/datetime.pxd":348
 * 
 * # Get day of date
 * cdef inline int date_day(object o):             # <<<<<<<<<<<<<<
 *     return PyDateTime_GET_DAY(o)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cpython/datetime.pxd":352
 * 
 * # Get year of datetime
 * cdef inline int datetime_year(object o):             # <<<<<<<<<<<<<<
 *     return PyDateTime_GET_YEAR(o)
 * 
 */

static CYTHON_INLINE int __pyx_f_7cpython_8datetime_datetime_year(PyObject *__pyx_v_o) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("datetime_year", 0);

  /* "cpython/datetime.pxd":353
 * # Get year of datetime
 * cdef inline int datetime_year(object o):
 *     return PyDateTime_GET_YEAR(o)             # <<<<<<<<<<<<<<
 * 
 * # Get month of datetime
 */
  __pyx_r = PyDateTime_GET_YEAR(__pyx_v_o);
  goto __pyx_L0;

  /* "cpython/datetime.pxd":352
 * 
 * # Get year of datetime
 * cdef inline int datetime_year(object o):             # <<<<<<<<<<<<<<
 *     return PyDateTime_GET_YEAR(o)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cpython/datetime.pxd":356
 * 
 * # Get month of datetime
 * cdef inline int datetime_month(object o):             # <<<<<<<<<<<<<<
 *     return PyDateTime_GET_MONTH(o)
 * 
 */

static CYTHON_INLINE int __pyx_f_7cpython_8datetime_datetime_month(PyObject *__pyx_v_o) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("datetime_month", 0);

  /* "cpython/datetime.pxd":357
 * # Get month of datetime
 * cdef inline int datetime_month(object o):
 *     return PyDateTime_GET_MONTH(o)             # <<<<<<<<<<<<<<
 * 
 * # Get day of datetime
 */
  __pyx_r = PyDateTime_GET_MONTH(__pyx_v_o);
  goto __pyx_L0;

  /* "cpython/datetime.pxd":356
 * 
 * # Get month of datetime
 * cdef inline int datetime_month(object o):             # <<<<<<<<<<<<<<
 *     return PyDateTime_GET_MONTH(o)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cpython/datetime.pxd":360
 * 
 * # Get day of datetime
 * cdef inline int datetime_day(object o):             # <<<<<<<<<<<<<<
 *     return PyDateTime_GET_DAY(o)
 * 
 */

static CYTHON_INLINE int __pyx_f_7cpython_8datetime_datetime_day(PyObject *__pyx_v_o) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("datetime_day", 0);

  /* "cpython/datetime.pxd":361
 * # Get day of datetime
 * cdef inline int datetime_day(object o):
 *     return PyDateTime_GET_DAY(o)             # <<<<<<<<<<<<<<
 * 
 * # Get hour of time
 */
  __pyx_r = PyDateTime_GET_DAY(__pyx_v_o);
  goto __pyx_L0;

  /* "cpython/datetime.pxd":360
 * 
 * # Get day of datetime
 * cdef inline int datetime_day(object o):             # <<<<<<<<<<<<<<
 *     return PyDateTime_GET_DAY(o)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cpython/datetime.pxd":364
 * 
 * # Get hour of time
 * cdef inline int time_hour(object o):             # <<<<<<<<<<<<<<
 *     return PyDateTime_TIME_GET_HOUR(o)
 * 
 */

static CYTHON_INLINE int __pyx_f_7cpython_8datetime_time_hour(PyObject *__pyx_v_o) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("time_hour", 0);

  /* "cpython/datetime.pxd":365
 * # Get hour of time
 * cdef inline int time_hour(object o):
 *     return PyDateTime_TIME_GET_HOUR(o)             # <<<<<<<<<<<<<<
 * 
 * # Get minute of time
 */
  __pyx_r = PyDateTime_TIME_GET_HOUR(__pyx_v_o);
  goto __pyx_L0;

  /* "cpython/datetime.pxd":364
 * 
 * # Get hour of time
 * cdef inline int time_hour(object o):             # <<<<<<<<<<<<<<
 *     return PyDateTime_TIME_GET_HOUR(o)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cpython/datetime.pxd":368
 * 
 * # Get minute of time
 * cdef inline int time_minute(object o):             # <<<<<<<<<<<<<<
 *     return PyDateTime_TIME_GET_MINUTE(o)
 * 
 */

static CYTHON_INLINE int __pyx_f_7cpython_8datetime_time_minute(PyObject *__pyx_v_o) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("time_minute", 0);

  /* "cpython/datetime.pxd":369
 * # Get minute of time
 * cdef inline int time_minute(object o):
 *     return PyDateTime_TIME_GET_MINUTE(o)             # <<<<<<<<<<<<<<
 * 
 * # Get second of time
 */
  __pyx_r = PyDateTime_TIME_GET_MINUTE(__pyx_v_o);
  goto __pyx_L0;

  /* "cpython/datetime.pxd":368
 * 
 * # Get minute of time
 * cdef inline int time_minute(object o):             # <<<<<<<<<<<<<<
 *     return PyDateTime_TIME_GET_MINUTE(o)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cpython/datetime.pxd":372
 * 
 * # Get second of time
 * cdef inline int time_second(object o):             # <<<<<<<<<<<<<<
 *     return PyDateTime_TIME_GET_SECOND(o)
 * 
 */

static CYTHON_INLINE int __pyx_f_7cpython_8datetime_time_second(PyObject *__pyx_v_o) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("time_second", 0);

  /* "cpython/datetime.pxd":373
 * # Get second of time
 * cdef inline int time_second(object o):
 *     return PyDateTime_TIME_GET_SECOND(o)             # <<<<<<<<<<<<<<
 * 
 * # Get microsecond of time
 */
  __pyx_r = PyDateTime_TIME_GET_SECOND(__pyx_v_o);
  goto __pyx_L0;

  /* "cpython/datetime.pxd":372
 * 
 * # Get second of time
 * cdef inline int time_second(object o):             # <<<<<<<<<<<<<<
 *     return PyDateTime_TIME_GET_SECOND(o)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cpython/datetime.pxd":376
 * 
 * # Get microsecond of time
 * cdef inline int time_microsecond(object o):             # <<<<<<<<<<<<<<
 *     return PyDateTime_TIME_GET_MICROSECOND(o)
 * 
 */

static CYTHON_INLINE int __pyx_f_7cpython_8datetime_time_microsecond(PyObject *__pyx_v_o) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("time_microsecond", 0);

  /* "cpython/datetime.pxd":377
 * # Get microsecond of time
 * cdef inline int time_microsecond(object o):
 *     return PyDateTime_TIME_GET_MICROSECOND(o)             # <<<<<<<<<<<<<<
 * 
 * # Get fold of time
 */
  __pyx_r = PyDateTime_TIME_GET_MICROSECOND(__pyx_v_o);
  goto __pyx_L0;

  /* "cpython/datetime.pxd":376
 * 
 * # Get microsecond of time
 * cdef inline int time_microsecond(object o):             # <<<<<<<<<<<<<<
 *     return PyDateTime_TIME_GET_MICROSECOND(o)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cpython/datetime.pxd":380
 * 
 * # Get fold of time
 * cdef inline int time_fold(object o):             # <<<<<<<<<<<<<<
 *     # For Python < 3.6 this returns 0 no matter what
 *     return PyDateTime_TIME_GET_FOLD(o)
 */

static CYTHON_INLINE int __pyx_f_7cpython_8datetime_time_fold(PyObject *__pyx_v_o) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("time_fold", 0);

  /* "cpython/datetime.pxd":382
 * cdef inline int time_fold(object o):
 *     # For Python < 3.6 this returns 0 no matter what
 *     return PyDateTime_TIME_GET_FOLD(o)             # <<<<<<<<<<<<<<
 * 
 * # Get hour of datetime
 */
  __pyx_r = PyDateTime_TIME_GET_FOLD(__pyx_v_o);
  goto __pyx_L0;

  /* "cpython/datetime.pxd":380
 * 
 * # Get fold of time
 * cdef inline int time_fold(object o):             # <<<<<<<<<<<<<<
 *     # For Python < 3.6 this returns 0 no matter what
 *     return PyDateTime_TIME_GET_FOLD(o)
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cpython/datetime.pxd":385
 * 
 * # Get hour of datetime
 * cdef inline int datetime_hour(object o):             # <<<<<<<<<<<<<<
 *     return PyDateTime_DATE_GET_HOUR(o)
 * 
 */

static CYTHON_INLINE int __pyx_f_7cpython_8datetime_datetime_hour(PyObject *__pyx_v_o) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("datetime_hour", 0);

  /* "cpython/datetime.pxd":386
 * # Get hour of datetime
 * cdef inline int datetime_hour(object o):
 *     return PyDateTime_DATE_GET_HOUR(o)             # <<<<<<<<<<<<<<
 * 
 * # Get minute of datetime
 */
  __pyx_r = PyDateTime_DATE_GET_HOUR(__pyx_v_o);
  goto __pyx_L0;

  /* "cpython/datetime.pxd":385
 * 
 * # Get hour of datetime
 * cdef inline int datetime_hour(object o):             # <<<<<<<<<<<<<<
 *     return PyDateTime_DATE_GET_HOUR(o)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cpython/datetime.pxd":389
 * 
 * # Get minute of datetime
 * cdef inline int datetime_minute(object o):             # <<<<<<<<<<<<<<
 *     return PyDateTime_DATE_GET_MINUTE(o)
 * 
 */

static CYTHON_INLINE int __pyx_f_7cpython_8datetime_datetime_minute(PyObject *__pyx_v_o) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("datetime_minute", 0);

  /* "cpython/datetime.pxd":390
 * # Get minute of datetime
 * cdef inline int datetime_minute(object o):
 *     return PyDateTime_DATE_GET_MINUTE(o)             # <<<<<<<<<<<<<<
 * 
 * # Get second of datetime
 */
  __pyx_r = PyDateTime_DATE_GET_MINUTE(__pyx_v_o);
  goto __pyx_L0;

  /* "cpython/datetime.pxd":389
 * 
 * # Get minute of datetime
 * cdef inline int datetime_minute(object o):             # <<<<<<<<<<<<<<
 *     return PyDateTime_DATE_GET_MINUTE(o)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cpython/datetime.pxd":393
 * 
 * # Get second of datetime
 * cdef inline int datetime_second(object o):             # <<<<<<<<<<<<<<
 *     return PyDateTime_DATE_GET_SECOND(o)
 * 
 */

static CYTHON_INLINE int __pyx_f_7cpython_8datetime_datetime_second(PyObject *__pyx_v_o) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("datetime_second", 0);

  /* "cpython/datetime.pxd":394
 * # Get second of datetime
 * cdef inline int datetime_second(object o):
 *     return PyDateTime_DATE_GET_SECOND(o)             # <<<<<<<<<<<<<<
 * 
 * # Get microsecond of datetime
 */
  __pyx_r = PyDateTime_DATE_GET_SECOND(__pyx_v_o);
  goto __pyx_L0;

  /* "cpython/datetime.pxd":393
 * 
 * # Get second of datetime
 * cdef inline int datetime_second(object o):             # <<<<<<<<<<<<<<
 *     return PyDateTime_DATE_GET_SECOND(o)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cpython/datetime.pxd":397
 * 
 * # Get microsecond of datetime
 * cdef inline int datetime_microsecond(object o):             # <<<<<<<<<<<<<<
 *     return PyDateTime_DATE_GET_MICROSECOND(o)
 * 
 */

static CYTHON_INLINE int __pyx_f_7cpython_8datetime_datetime_microsecond(PyObject *__pyx_v_o) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("datetime_microsecond", 0);

  /* "cpython/datetime.pxd":398
 * # Get microsecond of datetime
 * cdef inline int datetime_microsecond(object o):
 *     return PyDateTime_DATE_GET_MICROSECOND(o)             # <<<<<<<<<<<<<<
 * 
 * # Get fold of datetime
 */
  __pyx_r = PyDateTime_DATE_GET_MICROSECOND(__pyx_v_o);
  goto __pyx_L0;

  /* "cpython/datetime.pxd":397
 * 
 * # Get microsecond of datetime
 * cdef inline int datetime_microsecond(object o):             # <<<<<<<<<<<<<<
 *     return PyDateTime_DATE_GET_MICROSECOND(o)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cpython/datetime.pxd":401
 * 
 * # Get fold of datetime
 * cdef inline int datetime_fold(object o):             # <<<<<<<<<<<<<<
 *     # For Python < 3.6 this returns 0 no matter what
 *     return PyDateTime_DATE_GET_FOLD(o)
 */

static CYTHON_INLINE int __pyx_f_7cpython_8datetime_datetime_fold(PyObject *__pyx_v_o) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("datetime_fold", 0);

  /* "cpython/datetime.pxd":403
 * cdef inline int datetime_fold(object o):
 *     # For Python < 3.6 this returns 0 no matter what
 *     return PyDateTime_DATE_GET_FOLD(o)             # <<<<<<<<<<<<<<
 * 
 * # Get days of timedelta
 */
  __pyx_r = PyDateTime_DATE_GET_FOLD(__pyx_v_o);
  goto __pyx_L0;

  /* "cpython/datetime.pxd":401
 * 
 * # Get fold of datetime
 * cdef inline int datetime_fold(object o):             # <<<<<<<<<<<<<<
 *     # For Python < 3.6 this returns 0 no matter what
 *     return PyDateTime_DATE_GET_FOLD(o)
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cpython/datetime.pxd":406
 * 
 * # Get days of timedelta
 * cdef inline int timedelta_days(object o):             # <<<<<<<<<<<<<<
 *     return (<PyDateTime_Delta*>o).days
 * 
 */

static CYTHON_INLINE int __pyx_f_7cpython_8datetime_timedelta_days(PyObject *__pyx_v_o) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("timedelta_days", 0);

  /* "cpython/datetime.pxd":407
 * # Get days of timedelta
 * cdef inline int timedelta_days(object o):
 *     return (<PyDateTime_Delta*>o).days             # <<<<<<<<<<<<<<
 * 
 * # Get seconds of timedelta
 */
  __pyx_r = ((PyDateTime_Delta *)__pyx_v_o)->days;
  goto __pyx_L0;

  /* "cpython/datetime.pxd":406
 * 
 * # Get days of timedelta
 * cdef inline int timedelta_days(object o):             # <<<<<<<<<<<<<<
 *     return (<PyDateTime_Delta*>o).days
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cpython/datetime.pxd":410
 * 
 * # Get seconds of timedelta
 * cdef inline int timedelta_seconds(object o):             # <<<<<<<<<<<<<<
 *     return (<PyDateTime_Delta*>o).seconds
 * 
 */

static CYTHON_INLINE int __pyx_f_7cpython_8datetime_timedelta_seconds(PyObject *__pyx_v_o) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("timedelta_seconds", 0);

  /* "cpython/datetime.pxd":411
 * # Get seconds of timedelta
 * cdef inline int timedelta_seconds(object o):
 *     return (<PyDateTime_Delta*>o).seconds             # <<<<<<<<<<<<<<
 * 
 * # Get microseconds of timedelta
 */
  __pyx_r = ((PyDateTime_Delta *)__pyx_v_o)->seconds;
  goto __pyx_L0;

  /* "cpython/datetime.pxd":410
 * 
 * # Get seconds of timedelta
 * cdef inline int timedelta_seconds(object o):             # <<<<<<<<<<<<<<
 *     return (<PyDateTime_Delta*>o).seconds
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cpython/datetime.pxd":414
 * 
 * # Get microseconds of timedelta
 * cdef inline int timedelta_microseconds(object o):             # <<<<<<<<<<<<<<
 *     return (<PyDateTime_Delta*>o).microseconds
 * 
 */

static CYTHON_INLINE int __pyx_f_7cpython_8datetime_timedelta_microseconds(PyObject *__pyx_v_o) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("timedelta_microseconds", 0);

  /* "cpython/datetime.pxd":415
 * # Get microseconds of timedelta
 * cdef inline int timedelta_microseconds(object o):
 *     return (<PyDateTime_Delta*>o).microseconds             # <<<<<<<<<<<<<<
 * 
 * cdef inline double total_seconds(timedelta obj):
 */
  __pyx_r = ((PyDateTime_Delta *)__pyx_v_o)->microseconds;
  goto __pyx_L0;

  /* "cpython/datetime.pxd":414
 * 
 * # Get microseconds of timedelta
 * cdef inline int timedelta_microseconds(object o):             # <<<<<<<<<<<<<<
 *     return (<PyDateTime_Delta*>o).microseconds
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cpython/datetime.pxd":417
 *     return (<PyDateTime_Delta*>o).microseconds
 * 
 * cdef inline double total_seconds(timedelta obj):             # <<<<<<<<<<<<<<
 *     # Mirrors the "timedelta.total_seconds()" method.
 *     # Note that this implementation is not guaranteed to give *exactly* the same
 */

static CYTHON_INLINE double __pyx_f_7cpython_8datetime_total_seconds(PyDateTime_Delta *__pyx_v_obj) {
  double __pyx_v_days;
  double __pyx_v_seconds;
  double __pyx_v_micros;
  double __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("total_seconds", 0);

  /* "cpython/datetime.pxd":423
 *     cdef:
 *         double days, seconds, micros
 *     days = <double>PyDateTime_DELTA_GET_DAYS(obj)             # <<<<<<<<<<<<<<
 *     seconds = <double>PyDateTime_DELTA_GET_SECONDS(obj)
 *     micros = <double>PyDateTime_DELTA_GET_MICROSECONDS(obj)
 */
  __pyx_v_days = ((double)PyDateTime_DELTA_GET_DAYS(((PyObject *)__pyx_v_obj)));

  /* "cpython/datetime.pxd":424
 *         double days, seconds, micros
 *     days = <double>PyDateTime_DELTA_GET_DAYS(obj)
 *     seconds = <double>PyDateTime_DELTA_GET_SECONDS(obj)             # <<<<<<<<<<<<<<
 *     micros = <double>PyDateTime_DELTA_GET_MICROSECONDS(obj)
 *     return days * 24 * 3600 + seconds + micros / 1_000_000
 */
  __pyx_v_seconds = ((double)PyDateTime_DELTA_GET_SECONDS(((PyObject *)__pyx_v_obj)));

  /* "cpython/datetime.pxd":425
 *     days = <double>PyDateTime_DELTA_GET_DAYS(obj)
 *     seconds = <double>PyDateTime_DELTA_GET_SECONDS(obj)
 *     micros = <double>PyDateTime_DELTA_GET_MICROSECONDS(obj)             # <<<<<<<<<<<<<<
 *     return days * 24 * 3600 + seconds + micros / 1_000_000
 */
  __pyx_v_micros = ((double)PyDateTime_DELTA_GET_MICROSECONDS(((PyObject *)__pyx_v_obj)));

  /* "cpython/datetime.pxd":426
 *     seconds = <double>PyDateTime_DELTA_GET_SECONDS(obj)
 *     micros = <double>PyDateTime_DELTA_GET_MICROSECONDS(obj)
 *     return days * 24 * 3600 + seconds + micros / 1_000_000             # <<<<<<<<<<<<<<
 */
  __pyx_r = ((((__pyx_v_days * 24.0) * 3600.0) + __pyx_v_seconds) + (__pyx_v_micros / 1000000.0));
  goto __pyx_L0;

  /* "cpython/datetime.pxd":417
 *     return (<PyDateTime_Delta*>o).microseconds
 * 
 * cdef inline double total_seconds(timedelta obj):             # <<<<<<<<<<<<<<
 *     # Mirrors the "timedelta.total_seconds()" method.
 *     # Note that this implementation is not guaranteed to give *exactly* the same
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyreadstat/_readstat_parser.pyx":81
 *     """
 * 
 *     def __cinit__(self):             # <<<<<<<<<<<<<<
 *         self.n_obs = 0
 *         self.n_vars = 0
 */

/* Python wrapper */
static int __pyx_pw_10pyreadstat_16_readstat_parser_14data_container_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_10pyreadstat_16_readstat_parser_14data_container_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 0, 0, __pyx_nargs); return -1;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_VARARGS(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__cinit__", 0))) return -1;
  __pyx_r = __pyx_pf_10pyreadstat_16_readstat_parser_14data_container___cinit__(((struct __pyx_obj_10pyreadstat_16_readstat_parser_data_container *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10pyreadstat_16_readstat_parser_14data_container___cinit__(struct __pyx_obj_10pyreadstat_16_readstat_parser_data_container *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "pyreadstat/_readstat_parser.pyx":82
 * 
 *     def __cinit__(self):
 *         self.n_obs = 0             # <<<<<<<<<<<<<<
 *         self.n_vars = 0
 *         self.max_n_obs = 0
 */
  __pyx_v_self->n_obs = 0;

  /* "pyreadstat/_readstat_parser.pyx":83
 *     def __cinit__(self):
 *         self.n_obs = 0
 *         self.n_vars = 0             # <<<<<<<<<<<<<<
 *         self.max_n_obs = 0
 *         self.col_data = list()
 */
  __pyx_v_self->n_vars = 0;

  /* "pyreadstat/_readstat_parser.pyx":84
 *         self.n_obs = 0
 *         self.n_vars = 0
 *         self.max_n_obs = 0             # <<<<<<<<<<<<<<
 *         self.col_data = list()
 *         self.col_data_len = list()
 */
  __pyx_v_self->max_n_obs = 0;

  /* "pyreadstat/_readstat_parser.pyx":85
 *         self.n_vars = 0
 *         self.max_n_obs = 0
 *         self.col_data = list()             # <<<<<<<<<<<<<<
 *         self.col_data_len = list()
 *         self.col_names = list()
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 85, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->col_data);
  __Pyx_DECREF(__pyx_v_self->col_data);
  __pyx_v_self->col_data = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyreadstat/_readstat_parser.pyx":86
 *         self.max_n_obs = 0
 *         self.col_data = list()
 *         self.col_data_len = list()             # <<<<<<<<<<<<<<
 *         self.col_names = list()
 *         self.col_labels = list()
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 86, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->col_data_len);
  __Pyx_DECREF(__pyx_v_self->col_data_len);
  __pyx_v_self->col_data_len = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyreadstat/_readstat_parser.pyx":87
 *         self.col_data = list()
 *         self.col_data_len = list()
 *         self.col_names = list()             # <<<<<<<<<<<<<<
 *         self.col_labels = list()
 *         self.col_dtypes = list()
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 87, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->col_names);
  __Pyx_DECREF(__pyx_v_self->col_names);
  __pyx_v_self->col_names = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyreadstat/_readstat_parser.pyx":88
 *         self.col_data_len = list()
 *         self.col_names = list()
 *         self.col_labels = list()             # <<<<<<<<<<<<<<
 *         self.col_dtypes = list()
 *         self.col_numpy_dtypes = list()
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 88, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->col_labels);
  __Pyx_DECREF(__pyx_v_self->col_labels);
  __pyx_v_self->col_labels = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyreadstat/_readstat_parser.pyx":89
 *         self.col_names = list()
 *         self.col_labels = list()
 *         self.col_dtypes = list()             # <<<<<<<<<<<<<<
 *         self.col_numpy_dtypes = list()
 *         self.col_dtypes_isobject = list()
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->col_dtypes);
  __Pyx_DECREF(__pyx_v_self->col_dtypes);
  __pyx_v_self->col_dtypes = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyreadstat/_readstat_parser.pyx":90
 *         self.col_labels = list()
 *         self.col_dtypes = list()
 *         self.col_numpy_dtypes = list()             # <<<<<<<<<<<<<<
 *         self.col_dtypes_isobject = list()
 *         self.col_dytpes_isfloat = list()
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 90, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->col_numpy_dtypes);
  __Pyx_DECREF(__pyx_v_self->col_numpy_dtypes);
  __pyx_v_self->col_numpy_dtypes = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyreadstat/_readstat_parser.pyx":91
 *         self.col_dtypes = list()
 *         self.col_numpy_dtypes = list()
 *         self.col_dtypes_isobject = list()             # <<<<<<<<<<<<<<
 *         self.col_dytpes_isfloat = list()
 *         self.col_formats = list()
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 91, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->col_dtypes_isobject);
  __Pyx_DECREF(__pyx_v_self->col_dtypes_isobject);
  __pyx_v_self->col_dtypes_isobject = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyreadstat/_readstat_parser.pyx":92
 *         self.col_numpy_dtypes = list()
 *         self.col_dtypes_isobject = list()
 *         self.col_dytpes_isfloat = list()             # <<<<<<<<<<<<<<
 *         self.col_formats = list()
 *         self.col_formats_original = list()
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 92, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->col_dytpes_isfloat);
  __Pyx_DECREF(__pyx_v_self->col_dytpes_isfloat);
  __pyx_v_self->col_dytpes_isfloat = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyreadstat/_readstat_parser.pyx":93
 *         self.col_dtypes_isobject = list()
 *         self.col_dytpes_isfloat = list()
 *         self.col_formats = list()             # <<<<<<<<<<<<<<
 *         self.col_formats_original = list()
 *         self.origin = None
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 93, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->col_formats);
  __Pyx_DECREF(__pyx_v_self->col_formats);
  __pyx_v_self->col_formats = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyreadstat/_readstat_parser.pyx":94
 *         self.col_dytpes_isfloat = list()
 *         self.col_formats = list()
 *         self.col_formats_original = list()             # <<<<<<<<<<<<<<
 *         self.origin = None
 *         self.is_unkown_number_rows = 0
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 94, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->col_formats_original);
  __Pyx_DECREF(__pyx_v_self->col_formats_original);
  __pyx_v_self->col_formats_original = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyreadstat/_readstat_parser.pyx":95
 *         self.col_formats = list()
 *         self.col_formats_original = list()
 *         self.origin = None             # <<<<<<<<<<<<<<
 *         self.is_unkown_number_rows = 0
 *         self.file_encoding = None
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->origin);
  __Pyx_DECREF(__pyx_v_self->origin);
  __pyx_v_self->origin = Py_None;

  /* "pyreadstat/_readstat_parser.pyx":96
 *         self.col_formats_original = list()
 *         self.origin = None
 *         self.is_unkown_number_rows = 0             # <<<<<<<<<<<<<<
 *         self.file_encoding = None
 *         self.file_label = None
 */
  __pyx_v_self->is_unkown_number_rows = 0;

  /* "pyreadstat/_readstat_parser.pyx":97
 *         self.origin = None
 *         self.is_unkown_number_rows = 0
 *         self.file_encoding = None             # <<<<<<<<<<<<<<
 *         self.file_label = None
 *         self.metaonly = 0
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->file_encoding);
  __Pyx_DECREF(__pyx_v_self->file_encoding);
  __pyx_v_self->file_encoding = ((PyObject*)Py_None);

  /* "pyreadstat/_readstat_parser.pyx":98
 *         self.is_unkown_number_rows = 0
 *         self.file_encoding = None
 *         self.file_label = None             # <<<<<<<<<<<<<<
 *         self.metaonly = 0
 *         self.dates_as_pandas = 0
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->file_label);
  __Pyx_DECREF(__pyx_v_self->file_label);
  __pyx_v_self->file_label = ((PyObject*)Py_None);

  /* "pyreadstat/_readstat_parser.pyx":99
 *         self.file_encoding = None
 *         self.file_label = None
 *         self.metaonly = 0             # <<<<<<<<<<<<<<
 *         self.dates_as_pandas = 0
 *         self.label_to_var_name = dict()
 */
  __pyx_v_self->metaonly = 0;

  /* "pyreadstat/_readstat_parser.pyx":100
 *         self.file_label = None
 *         self.metaonly = 0
 *         self.dates_as_pandas = 0             # <<<<<<<<<<<<<<
 *         self.label_to_var_name = dict()
 *         self.labels_raw = dict()
 */
  __pyx_v_self->dates_as_pandas = 0;

  /* "pyreadstat/_readstat_parser.pyx":101
 *         self.metaonly = 0
 *         self.dates_as_pandas = 0
 *         self.label_to_var_name = dict()             # <<<<<<<<<<<<<<
 *         self.labels_raw = dict()
 *         self.notes = list()
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 101, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->label_to_var_name);
  __Pyx_DECREF(__pyx_v_self->label_to_var_name);
  __pyx_v_self->label_to_var_name = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pyreadstat/_readstat_parser.pyx":102
 *         self.dates_as_pandas = 0
 *         self.label_to_var_name = dict()
 *         self.labels_raw = dict()             # <<<<<<<<<<<<<<
 *         self.notes = list()
 *         self.user_encoding = None
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 102, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->labels_raw);
  __Pyx_DECREF(__pyx_v_self->labels_raw);
  __pyx_v_self->labels_raw = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pyreadstat/_readstat_parser.pyx":103
 *         self.label_to_var_name = dict()
 *         self.labels_raw = dict()
 *         self.notes = list()             # <<<<<<<<<<<<<<
 *         self.user_encoding = None
 *         self.table_name = None
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 103, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->notes);
  __Pyx_DECREF(__pyx_v_self->notes);
  __pyx_v_self->notes = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pyreadstat/_readstat_parser.pyx":104
 *         self.labels_raw = dict()
 *         self.notes = list()
 *         self.user_encoding = None             # <<<<<<<<<<<<<<
 *         self.table_name = None
 *         self.filter_cols = 0
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->user_encoding);
  __Pyx_DECREF(__pyx_v_self->user_encoding);
  __pyx_v_self->user_encoding = ((PyObject*)Py_None);

  /* "pyreadstat/_readstat_parser.pyx":105
 *         self.notes = list()
 *         self.user_encoding = None
 *         self.table_name = None             # <<<<<<<<<<<<<<
 *         self.filter_cols = 0
 *         self.use_cols = list()
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->table_name);
  __Pyx_DECREF(__pyx_v_self->table_name);
  __pyx_v_self->table_name = ((PyObject*)Py_None);

  /* "pyreadstat/_readstat_parser.pyx":106
 *         self.user_encoding = None
 *         self.table_name = None
 *         self.filter_cols = 0             # <<<<<<<<<<<<<<
 *         self.use_cols = list()
 *         self.usernan = 0
 */
  __pyx_v_self->filter_cols = 0;

  /* "pyreadstat/_readstat_parser.pyx":107
 *         self.table_name = None
 *         self.filter_cols = 0
 *         self.use_cols = list()             # <<<<<<<<<<<<<<
 *         self.usernan = 0
 *         self.missing_ranges = dict()
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 107, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->use_cols);
  __Pyx_DECREF(__pyx_v_self->use_cols);
  __pyx_v_self->use_cols = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyreadstat/_readstat_parser.pyx":108
 *         self.filter_cols = 0
 *         self.use_cols = list()
 *         self.usernan = 0             # <<<<<<<<<<<<<<
 *         self.missing_ranges = dict()
 *         self.missing_user_values = dict()
 */
  __pyx_v_self->usernan = 0;

  /* "pyreadstat/_readstat_parser.pyx":109
 *         self.use_cols = list()
 *         self.usernan = 0
 *         self.missing_ranges = dict()             # <<<<<<<<<<<<<<
 *         self.missing_user_values = dict()
 *         self.variable_storage_width = dict()
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 109, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->missing_ranges);
  __Pyx_DECREF(__pyx_v_self->missing_ranges);
  __pyx_v_self->missing_ranges = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyreadstat/_readstat_parser.pyx":110
 *         self.usernan = 0
 *         self.missing_ranges = dict()
 *         self.missing_user_values = dict()             # <<<<<<<<<<<<<<
 *         self.variable_storage_width = dict()
 *         self.variable_display_width = dict()
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 110, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->missing_user_values);
  __Pyx_DECREF(__pyx_v_self->missing_user_values);
  __pyx_v_self->missing_user_values = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyreadstat/_readstat_parser.pyx":111
 *         self.missing_ranges = dict()
 *         self.missing_user_values = dict()
 *         self.variable_storage_width = dict()             # <<<<<<<<<<<<<<
 *         self.variable_display_width = dict()
 *         self.variable_alignment = dict()
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 111, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->variable_storage_width);
  __Pyx_DECREF(__pyx_v_self->variable_storage_width);
  __pyx_v_self->variable_storage_width = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyreadstat/_readstat_parser.pyx":112
 *         self.missing_user_values = dict()
 *         self.variable_storage_width = dict()
 *         self.variable_display_width = dict()             # <<<<<<<<<<<<<<
 *         self.variable_alignment = dict()
 *         self.variable_measure = dict()
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 112, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->variable_display_width);
  __Pyx_DECREF(__pyx_v_self->variable_display_width);
  __pyx_v_self->variable_display_width = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pyreadstat/_readstat_parser.pyx":113
 *         self.variable_storage_width = dict()
 *         self.variable_display_width = dict()
 *         self.variable_alignment = dict()             # <<<<<<<<<<<<<<
 *         self.variable_measure = dict()
 *         self.no_datetime_conversion = 0
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 113, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->variable_alignment);
  __Pyx_DECREF(__pyx_v_self->variable_alignment);
  __pyx_v_self->variable_alignment = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pyreadstat/_readstat_parser.pyx":114
 *         self.variable_display_width = dict()
 *         self.variable_alignment = dict()
 *         self.variable_measure = dict()             # <<<<<<<<<<<<<<
 *         self.no_datetime_conversion = 0
 *         self.ctime = 0
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 114, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->variable_measure);
  __Pyx_DECREF(__pyx_v_self->variable_measure);
  __pyx_v_self->variable_measure = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pyreadstat/_readstat_parser.pyx":115
 *         self.variable_alignment = dict()
 *         self.variable_measure = dict()
 *         self.no_datetime_conversion = 0             # <<<<<<<<<<<<<<
 *         self.ctime = 0
 *         self.mtime = 0
 */
  __pyx_v_self->no_datetime_conversion = 0;

  /* "pyreadstat/_readstat_parser.pyx":116
 *         self.variable_measure = dict()
 *         self.no_datetime_conversion = 0
 *         self.ctime = 0             # <<<<<<<<<<<<<<
 *         self.mtime = 0
 * 
 */
  __pyx_v_self->ctime = 0;

  /* "pyreadstat/_readstat_parser.pyx":117
 *         self.no_datetime_conversion = 0
 *         self.ctime = 0
 *         self.mtime = 0             # <<<<<<<<<<<<<<
 * 
 * class metadata_container:
 */
  __pyx_v_self->mtime = 0;

  /* "pyreadstat/_readstat_parser.pyx":81
 *     """
 * 
 *     def __cinit__(self):             # <<<<<<<<<<<<<<
 *         self.n_obs = 0
 *         self.n_vars = 0
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyreadstat._readstat_parser.data_container.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_10pyreadstat_16_readstat_parser_14data_container_3__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_10pyreadstat_16_readstat_parser_14data_container_3__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10pyreadstat_16_readstat_parser_14data_container_3__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_10pyreadstat_16_readstat_parser_14data_container_3__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_10pyreadstat_16_readstat_parser_14data_container_2__reduce_cython__(((struct __pyx_obj_10pyreadstat_16_readstat_parser_data_container *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10pyreadstat_16_readstat_parser_14data_container_2__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_10pyreadstat_16_readstat_parser_data_container *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyreadstat._readstat_parser.data_container.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

/* Python wrapper */
static PyObject *__pyx_pw_10pyreadstat_16_readstat_parser_14data_container_5__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_10pyreadstat_16_readstat_parser_14data_container_5__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10pyreadstat_16_readstat_parser_14data_container_5__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_10pyreadstat_16_readstat_parser_14data_container_5__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(1, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 3, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyreadstat._readstat_parser.data_container.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10pyreadstat_16_readstat_parser_14data_container_4__setstate_cython__(((struct __pyx_obj_10pyreadstat_16_readstat_parser_data_container *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10pyreadstat_16_readstat_parser_14data_container_4__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_10pyreadstat_16_readstat_parser_data_container *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyreadstat._readstat_parser.data_container.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyreadstat/_readstat_parser.pyx":123
 *     This class holds metadata we want to give back to python
 *     """
 *     def __init__(self):             # <<<<<<<<<<<<<<
 *         self.column_names = list()
 *         self.column_labels = list()
 */

/* Python wrapper */
static PyObject *__pyx_pw_10pyreadstat_16_readstat_parser_18metadata_container_1__init__(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_10pyreadstat_16_readstat_parser_18metadata_container_1__init__ = {"__init__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10pyreadstat_16_readstat_parser_18metadata_container_1__init__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_10pyreadstat_16_readstat_parser_18metadata_container_1__init__(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_self = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,0};
    PyObject* values[1] = {0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_self)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 123, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(0, 123, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_self = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 123, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyreadstat._readstat_parser.metadata_container.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10pyreadstat_16_readstat_parser_18metadata_container___init__(__pyx_self, __pyx_v_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10pyreadstat_16_readstat_parser_18metadata_container___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pyreadstat/_readstat_parser.pyx":124
 *     """
 *     def __init__(self):
 *         self.column_names = list()             # <<<<<<<<<<<<<<
 *         self.column_labels = list()
 *         self.column_names_to_labels = dict()
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 124, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_column_names, __pyx_t_1) < 0) __PYX_ERR(0, 124, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyreadstat/_readstat_parser.pyx":125
 *     def __init__(self):
 *         self.column_names = list()
 *         self.column_labels = list()             # <<<<<<<<<<<<<<
 *         self.column_names_to_labels = dict()
 *         self.file_encoding = None
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 125, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_column_labels, __pyx_t_1) < 0) __PYX_ERR(0, 125, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyreadstat/_readstat_parser.pyx":126
 *         self.column_names = list()
 *         self.column_labels = list()
 *         self.column_names_to_labels = dict()             # <<<<<<<<<<<<<<
 *         self.file_encoding = None
 *         self.number_columns = None
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 126, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_column_names_to_labels, __pyx_t_1) < 0) __PYX_ERR(0, 126, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyreadstat/_readstat_parser.pyx":127
 *         self.column_labels = list()
 *         self.column_names_to_labels = dict()
 *         self.file_encoding = None             # <<<<<<<<<<<<<<
 *         self.number_columns = None
 *         self.number_rows = None
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_file_encoding, Py_None) < 0) __PYX_ERR(0, 127, __pyx_L1_error)

  /* "pyreadstat/_readstat_parser.pyx":128
 *         self.column_names_to_labels = dict()
 *         self.file_encoding = None
 *         self.number_columns = None             # <<<<<<<<<<<<<<
 *         self.number_rows = None
 *         self.variable_value_labels = dict()
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_number_columns, Py_None) < 0) __PYX_ERR(0, 128, __pyx_L1_error)

  /* "pyreadstat/_readstat_parser.pyx":129
 *         self.file_encoding = None
 *         self.number_columns = None
 *         self.number_rows = None             # <<<<<<<<<<<<<<
 *         self.variable_value_labels = dict()
 *         self.value_labels = dict()
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_number_rows, Py_None) < 0) __PYX_ERR(0, 129, __pyx_L1_error)

  /* "pyreadstat/_readstat_parser.pyx":130
 *         self.number_columns = None
 *         self.number_rows = None
 *         self.variable_value_labels = dict()             # <<<<<<<<<<<<<<
 *         self.value_labels = dict()
 *         self.variable_to_label = dict()
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 130, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_variable_value_labels, __pyx_t_1) < 0) __PYX_ERR(0, 130, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyreadstat/_readstat_parser.pyx":131
 *         self.number_rows = None
 *         self.variable_value_labels = dict()
 *         self.value_labels = dict()             # <<<<<<<<<<<<<<
 *         self.variable_to_label = dict()
 *         self.notes = list()
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 131, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_value_labels, __pyx_t_1) < 0) __PYX_ERR(0, 131, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyreadstat/_readstat_parser.pyx":132
 *         self.variable_value_labels = dict()
 *         self.value_labels = dict()
 *         self.variable_to_label = dict()             # <<<<<<<<<<<<<<
 *         self.notes = list()
 *         self.original_variable_types = dict()
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 132, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_variable_to_label, __pyx_t_1) < 0) __PYX_ERR(0, 132, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyreadstat/_readstat_parser.pyx":133
 *         self.value_labels = dict()
 *         self.variable_to_label = dict()
 *         self.notes = list()             # <<<<<<<<<<<<<<
 *         self.original_variable_types = dict()
 *         self.readstat_variable_types = dict()
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 133, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_notes, __pyx_t_1) < 0) __PYX_ERR(0, 133, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyreadstat/_readstat_parser.pyx":134
 *         self.variable_to_label = dict()
 *         self.notes = list()
 *         self.original_variable_types = dict()             # <<<<<<<<<<<<<<
 *         self.readstat_variable_types = dict()
 *         self.table_name = None
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 134, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_original_variable_types, __pyx_t_1) < 0) __PYX_ERR(0, 134, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyreadstat/_readstat_parser.pyx":135
 *         self.notes = list()
 *         self.original_variable_types = dict()
 *         self.readstat_variable_types = dict()             # <<<<<<<<<<<<<<
 *         self.table_name = None
 *         self.missing_ranges = dict()
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 135, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_readstat_variable_types, __pyx_t_1) < 0) __PYX_ERR(0, 135, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyreadstat/_readstat_parser.pyx":136
 *         self.original_variable_types = dict()
 *         self.readstat_variable_types = dict()
 *         self.table_name = None             # <<<<<<<<<<<<<<
 *         self.missing_ranges = dict()
 *         self.missing_user_values = dict()
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_table_name, Py_None) < 0) __PYX_ERR(0, 136, __pyx_L1_error)

  /* "pyreadstat/_readstat_parser.pyx":137
 *         self.readstat_variable_types = dict()
 *         self.table_name = None
 *         self.missing_ranges = dict()             # <<<<<<<<<<<<<<
 *         self.missing_user_values = dict()
 *         self.variable_storage_width = dict()
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 137, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_missing_ranges, __pyx_t_1) < 0) __PYX_ERR(0, 137, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyreadstat/_readstat_parser.pyx":138
 *         self.table_name = None
 *         self.missing_ranges = dict()
 *         self.missing_user_values = dict()             # <<<<<<<<<<<<<<
 *         self.variable_storage_width = dict()
 *         self.variable_display_width = dict()
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 138, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_missing_user_values, __pyx_t_1) < 0) __PYX_ERR(0, 138, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyreadstat/_readstat_parser.pyx":139
 *         self.missing_ranges = dict()
 *         self.missing_user_values = dict()
 *         self.variable_storage_width = dict()             # <<<<<<<<<<<<<<
 *         self.variable_display_width = dict()
 *         self.variable_alignment = dict()
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 139, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_variable_storage_width, __pyx_t_1) < 0) __PYX_ERR(0, 139, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyreadstat/_readstat_parser.pyx":140
 *         self.missing_user_values = dict()
 *         self.variable_storage_width = dict()
 *         self.variable_display_width = dict()             # <<<<<<<<<<<<<<
 *         self.variable_alignment = dict()
 *         self.variable_measure = dict()
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 140, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_variable_display_width, __pyx_t_1) < 0) __PYX_ERR(0, 140, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyreadstat/_readstat_parser.pyx":141
 *         self.variable_storage_width = dict()
 *         self.variable_display_width = dict()
 *         self.variable_alignment = dict()             # <<<<<<<<<<<<<<
 *         self.variable_measure = dict()
 *         self.creation_time = None
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 141, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_variable_alignment, __pyx_t_1) < 0) __PYX_ERR(0, 141, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyreadstat/_readstat_parser.pyx":142
 *         self.variable_display_width = dict()
 *         self.variable_alignment = dict()
 *         self.variable_measure = dict()             # <<<<<<<<<<<<<<
 *         self.creation_time = None
 *         self.modification_time = None
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 142, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_variable_measure, __pyx_t_1) < 0) __PYX_ERR(0, 142, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyreadstat/_readstat_parser.pyx":143
 *         self.variable_alignment = dict()
 *         self.variable_measure = dict()
 *         self.creation_time = None             # <<<<<<<<<<<<<<
 *         self.modification_time = None
 * 
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_creation_time, Py_None) < 0) __PYX_ERR(0, 143, __pyx_L1_error)

  /* "pyreadstat/_readstat_parser.pyx":144
 *         self.variable_measure = dict()
 *         self.creation_time = None
 *         self.modification_time = None             # <<<<<<<<<<<<<<
 * 
 * 
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_modification_time, Py_None) < 0) __PYX_ERR(0, 144, __pyx_L1_error)

  /* "pyreadstat/_readstat_parser.pyx":123
 *     This class holds metadata we want to give back to python
 *     """
 *     def __init__(self):             # <<<<<<<<<<<<<<
 *         self.column_names = list()
 *         self.column_labels = list()
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyreadstat._readstat_parser.metadata_container.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyreadstat/_readstat_parser.pyx":160
 * 
 * 
 * cdef py_datetime_format transform_variable_format(str var_format, py_file_format file_format):             # <<<<<<<<<<<<<<
 *     """
 *     Transforms a readstat var_format to a date, datetime or time format label
 */

static __pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format __pyx_f_10pyreadstat_16_readstat_parser_transform_variable_format(PyObject *__pyx_v_var_format, __pyx_t_10pyreadstat_16_readstat_parser_py_file_format __pyx_v_file_format) {
  __pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("transform_variable_format", 0);

  /* "pyreadstat/_readstat_parser.pyx":164
 *     Transforms a readstat var_format to a date, datetime or time format label
 *     """
 *     if file_format == FILE_FORMAT_SAS:             # <<<<<<<<<<<<<<
 *         if var_format in sas_all_formats:
 *             if var_format in sas_date_formats:
 */
  switch (__pyx_v_file_format) {
    case __pyx_e_10pyreadstat_16_readstat_parser_FILE_FORMAT_SAS:

    /* "pyreadstat/_readstat_parser.pyx":165
 *     """
 *     if file_format == FILE_FORMAT_SAS:
 *         if var_format in sas_all_formats:             # <<<<<<<<<<<<<<
 *             if var_format in sas_date_formats:
 *                 return DATE_FORMAT_DATE
 */
    __pyx_t_1 = (__Pyx_PySequence_ContainsTF(__pyx_v_var_format, __pyx_v_10pyreadstat_16_readstat_parser_sas_all_formats, Py_EQ)); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 165, __pyx_L1_error)
    if (__pyx_t_1) {

      /* "pyreadstat/_readstat_parser.pyx":166
 *     if file_format == FILE_FORMAT_SAS:
 *         if var_format in sas_all_formats:
 *             if var_format in sas_date_formats:             # <<<<<<<<<<<<<<
 *                 return DATE_FORMAT_DATE
 *             elif var_format in sas_datetime_formats:
 */
      __pyx_t_1 = (__Pyx_PySequence_ContainsTF(__pyx_v_var_format, __pyx_v_10pyreadstat_16_readstat_parser_sas_date_formats, Py_EQ)); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 166, __pyx_L1_error)
      if (__pyx_t_1) {

        /* "pyreadstat/_readstat_parser.pyx":167
 *         if var_format in sas_all_formats:
 *             if var_format in sas_date_formats:
 *                 return DATE_FORMAT_DATE             # <<<<<<<<<<<<<<
 *             elif var_format in sas_datetime_formats:
 *                 return DATE_FORMAT_DATETIME
 */
        __pyx_r = __pyx_e_10pyreadstat_16_readstat_parser_DATE_FORMAT_DATE;
        goto __pyx_L0;

        /* "pyreadstat/_readstat_parser.pyx":166
 *     if file_format == FILE_FORMAT_SAS:
 *         if var_format in sas_all_formats:
 *             if var_format in sas_date_formats:             # <<<<<<<<<<<<<<
 *                 return DATE_FORMAT_DATE
 *             elif var_format in sas_datetime_formats:
 */
      }

      /* "pyreadstat/_readstat_parser.pyx":168
 *             if var_format in sas_date_formats:
 *                 return DATE_FORMAT_DATE
 *             elif var_format in sas_datetime_formats:             # <<<<<<<<<<<<<<
 *                 return DATE_FORMAT_DATETIME
 *             elif var_format in sas_time_formats:
 */
      __pyx_t_1 = (__Pyx_PySequence_ContainsTF(__pyx_v_var_format, __pyx_v_10pyreadstat_16_readstat_parser_sas_datetime_formats, Py_EQ)); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 168, __pyx_L1_error)
      if (__pyx_t_1) {

        /* "pyreadstat/_readstat_parser.pyx":169
 *                 return DATE_FORMAT_DATE
 *             elif var_format in sas_datetime_formats:
 *                 return DATE_FORMAT_DATETIME             # <<<<<<<<<<<<<<
 *             elif var_format in sas_time_formats:
 *                 return DATE_FORMAT_TIME
 */
        __pyx_r = __pyx_e_10pyreadstat_16_readstat_parser_DATE_FORMAT_DATETIME;
        goto __pyx_L0;

        /* "pyreadstat/_readstat_parser.pyx":168
 *             if var_format in sas_date_formats:
 *                 return DATE_FORMAT_DATE
 *             elif var_format in sas_datetime_formats:             # <<<<<<<<<<<<<<
 *                 return DATE_FORMAT_DATETIME
 *             elif var_format in sas_time_formats:
 */
      }

      /* "pyreadstat/_readstat_parser.pyx":170
 *             elif var_format in sas_datetime_formats:
 *                 return DATE_FORMAT_DATETIME
 *             elif var_format in sas_time_formats:             # <<<<<<<<<<<<<<
 *                 return DATE_FORMAT_TIME
 *         else:
 */
      __pyx_t_1 = (__Pyx_PySequence_ContainsTF(__pyx_v_var_format, __pyx_v_10pyreadstat_16_readstat_parser_sas_time_formats, Py_EQ)); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 170, __pyx_L1_error)
      if (__pyx_t_1) {

        /* "pyreadstat/_readstat_parser.pyx":171
 *                 return DATE_FORMAT_DATETIME
 *             elif var_format in sas_time_formats:
 *                 return DATE_FORMAT_TIME             # <<<<<<<<<<<<<<
 *         else:
 *             return DATE_FORMAT_NOTADATE
 */
        __pyx_r = __pyx_e_10pyreadstat_16_readstat_parser_DATE_FORMAT_TIME;
        goto __pyx_L0;

        /* "pyreadstat/_readstat_parser.pyx":170
 *             elif var_format in sas_datetime_formats:
 *                 return DATE_FORMAT_DATETIME
 *             elif var_format in sas_time_formats:             # <<<<<<<<<<<<<<
 *                 return DATE_FORMAT_TIME
 *         else:
 */
      }

      /* "pyreadstat/_readstat_parser.pyx":165
 *     """
 *     if file_format == FILE_FORMAT_SAS:
 *         if var_format in sas_all_formats:             # <<<<<<<<<<<<<<
 *             if var_format in sas_date_formats:
 *                 return DATE_FORMAT_DATE
 */
      goto __pyx_L3;
    }

    /* "pyreadstat/_readstat_parser.pyx":173
 *                 return DATE_FORMAT_TIME
 *         else:
 *             return DATE_FORMAT_NOTADATE             # <<<<<<<<<<<<<<
 * 
 *     elif file_format == FILE_FORMAT_SPSS:
 */
    /*else*/ {
      __pyx_r = __pyx_e_10pyreadstat_16_readstat_parser_DATE_FORMAT_NOTADATE;
      goto __pyx_L0;
    }
    __pyx_L3:;

    /* "pyreadstat/_readstat_parser.pyx":164
 *     Transforms a readstat var_format to a date, datetime or time format label
 *     """
 *     if file_format == FILE_FORMAT_SAS:             # <<<<<<<<<<<<<<
 *         if var_format in sas_all_formats:
 *             if var_format in sas_date_formats:
 */
    break;
    case __pyx_e_10pyreadstat_16_readstat_parser_FILE_FORMAT_SPSS:

    /* "pyreadstat/_readstat_parser.pyx":176
 * 
 *     elif file_format == FILE_FORMAT_SPSS:
 *         if var_format in spss_all_formats:             # <<<<<<<<<<<<<<
 *             if var_format in spss_date_formats:
 *                 return DATE_FORMAT_DATE
 */
    __pyx_t_1 = (__Pyx_PySequence_ContainsTF(__pyx_v_var_format, __pyx_v_10pyreadstat_16_readstat_parser_spss_all_formats, Py_EQ)); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 176, __pyx_L1_error)
    if (__pyx_t_1) {

      /* "pyreadstat/_readstat_parser.pyx":177
 *     elif file_format == FILE_FORMAT_SPSS:
 *         if var_format in spss_all_formats:
 *             if var_format in spss_date_formats:             # <<<<<<<<<<<<<<
 *                 return DATE_FORMAT_DATE
 *             elif var_format in spss_datetime_formats:
 */
      __pyx_t_1 = (__Pyx_PySequence_ContainsTF(__pyx_v_var_format, __pyx_v_10pyreadstat_16_readstat_parser_spss_date_formats, Py_EQ)); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 177, __pyx_L1_error)
      if (__pyx_t_1) {

        /* "pyreadstat/_readstat_parser.pyx":178
 *         if var_format in spss_all_formats:
 *             if var_format in spss_date_formats:
 *                 return DATE_FORMAT_DATE             # <<<<<<<<<<<<<<
 *             elif var_format in spss_datetime_formats:
 *                 return DATE_FORMAT_DATETIME
 */
        __pyx_r = __pyx_e_10pyreadstat_16_readstat_parser_DATE_FORMAT_DATE;
        goto __pyx_L0;

        /* "pyreadstat/_readstat_parser.pyx":177
 *     elif file_format == FILE_FORMAT_SPSS:
 *         if var_format in spss_all_formats:
 *             if var_format in spss_date_formats:             # <<<<<<<<<<<<<<
 *                 return DATE_FORMAT_DATE
 *             elif var_format in spss_datetime_formats:
 */
      }

      /* "pyreadstat/_readstat_parser.pyx":179
 *             if var_format in spss_date_formats:
 *                 return DATE_FORMAT_DATE
 *             elif var_format in spss_datetime_formats:             # <<<<<<<<<<<<<<
 *                 return DATE_FORMAT_DATETIME
 *             elif var_format in spss_time_formats:
 */
      __pyx_t_1 = (__Pyx_PySequence_ContainsTF(__pyx_v_var_format, __pyx_v_10pyreadstat_16_readstat_parser_spss_datetime_formats, Py_EQ)); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 179, __pyx_L1_error)
      if (__pyx_t_1) {

        /* "pyreadstat/_readstat_parser.pyx":180
 *                 return DATE_FORMAT_DATE
 *             elif var_format in spss_datetime_formats:
 *                 return DATE_FORMAT_DATETIME             # <<<<<<<<<<<<<<
 *             elif var_format in spss_time_formats:
 *                 return DATE_FORMAT_TIME
 */
        __pyx_r = __pyx_e_10pyreadstat_16_readstat_parser_DATE_FORMAT_DATETIME;
        goto __pyx_L0;

        /* "pyreadstat/_readstat_parser.pyx":179
 *             if var_format in spss_date_formats:
 *                 return DATE_FORMAT_DATE
 *             elif var_format in spss_datetime_formats:             # <<<<<<<<<<<<<<
 *                 return DATE_FORMAT_DATETIME
 *             elif var_format in spss_time_formats:
 */
      }

      /* "pyreadstat/_readstat_parser.pyx":181
 *             elif var_format in spss_datetime_formats:
 *                 return DATE_FORMAT_DATETIME
 *             elif var_format in spss_time_formats:             # <<<<<<<<<<<<<<
 *                 return DATE_FORMAT_TIME
 *         else:
 */
      __pyx_t_1 = (__Pyx_PySequence_ContainsTF(__pyx_v_var_format, __pyx_v_10pyreadstat_16_readstat_parser_spss_time_formats, Py_EQ)); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 181, __pyx_L1_error)
      if (__pyx_t_1) {

        /* "pyreadstat/_readstat_parser.pyx":182
 *                 return DATE_FORMAT_DATETIME
 *             elif var_format in spss_time_formats:
 *                 return DATE_FORMAT_TIME             # <<<<<<<<<<<<<<
 *         else:
 *             return DATE_FORMAT_NOTADATE
 */
        __pyx_r = __pyx_e_10pyreadstat_16_readstat_parser_DATE_FORMAT_TIME;
        goto __pyx_L0;

        /* "pyreadstat/_readstat_parser.pyx":181
 *             elif var_format in spss_datetime_formats:
 *                 return DATE_FORMAT_DATETIME
 *             elif var_format in spss_time_formats:             # <<<<<<<<<<<<<<
 *                 return DATE_FORMAT_TIME
 *         else:
 */
      }

      /* "pyreadstat/_readstat_parser.pyx":176
 * 
 *     elif file_format == FILE_FORMAT_SPSS:
 *         if var_format in spss_all_formats:             # <<<<<<<<<<<<<<
 *             if var_format in spss_date_formats:
 *                 return DATE_FORMAT_DATE
 */
      goto __pyx_L5;
    }

    /* "pyreadstat/_readstat_parser.pyx":184
 *                 return DATE_FORMAT_TIME
 *         else:
 *             return DATE_FORMAT_NOTADATE             # <<<<<<<<<<<<<<
 * 
 *     elif file_format == FILE_FORMAT_STATA:
 */
    /*else*/ {
      __pyx_r = __pyx_e_10pyreadstat_16_readstat_parser_DATE_FORMAT_NOTADATE;
      goto __pyx_L0;
    }
    __pyx_L5:;

    /* "pyreadstat/_readstat_parser.pyx":175
 *             return DATE_FORMAT_NOTADATE
 * 
 *     elif file_format == FILE_FORMAT_SPSS:             # <<<<<<<<<<<<<<
 *         if var_format in spss_all_formats:
 *             if var_format in spss_date_formats:
 */
    break;
    case __pyx_e_10pyreadstat_16_readstat_parser_FILE_FORMAT_STATA:

    /* "pyreadstat/_readstat_parser.pyx":187
 * 
 *     elif file_format == FILE_FORMAT_STATA:
 *         if var_format in stata_all_formats:             # <<<<<<<<<<<<<<
 *             if var_format in stata_date_formats:
 *                 return DATE_FORMAT_DATE
 */
    __pyx_t_1 = (__Pyx_PySequence_ContainsTF(__pyx_v_var_format, __pyx_v_10pyreadstat_16_readstat_parser_stata_all_formats, Py_EQ)); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 187, __pyx_L1_error)
    if (__pyx_t_1) {

      /* "pyreadstat/_readstat_parser.pyx":188
 *     elif file_format == FILE_FORMAT_STATA:
 *         if var_format in stata_all_formats:
 *             if var_format in stata_date_formats:             # <<<<<<<<<<<<<<
 *                 return DATE_FORMAT_DATE
 *             elif var_format in stata_datetime_formats:
 */
      __pyx_t_1 = (__Pyx_PySequence_ContainsTF(__pyx_v_var_format, __pyx_v_10pyreadstat_16_readstat_parser_stata_date_formats, Py_EQ)); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 188, __pyx_L1_error)
      if (__pyx_t_1) {

        /* "pyreadstat/_readstat_parser.pyx":189
 *         if var_format in stata_all_formats:
 *             if var_format in stata_date_formats:
 *                 return DATE_FORMAT_DATE             # <<<<<<<<<<<<<<
 *             elif var_format in stata_datetime_formats:
 *                 return DATE_FORMAT_DATETIME
 */
        __pyx_r = __pyx_e_10pyreadstat_16_readstat_parser_DATE_FORMAT_DATE;
        goto __pyx_L0;

        /* "pyreadstat/_readstat_parser.pyx":188
 *     elif file_format == FILE_FORMAT_STATA:
 *         if var_format in stata_all_formats:
 *             if var_format in stata_date_formats:             # <<<<<<<<<<<<<<
 *                 return DATE_FORMAT_DATE
 *             elif var_format in stata_datetime_formats:
 */
      }

      /* "pyreadstat/_readstat_parser.pyx":190
 *             if var_format in stata_date_formats:
 *                 return DATE_FORMAT_DATE
 *             elif var_format in stata_datetime_formats:             # <<<<<<<<<<<<<<
 *                 return DATE_FORMAT_DATETIME
 *             elif var_format in stata_time_formats:
 */
      __pyx_t_1 = (__Pyx_PySequence_ContainsTF(__pyx_v_var_format, __pyx_v_10pyreadstat_16_readstat_parser_stata_datetime_formats, Py_EQ)); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 190, __pyx_L1_error)
      if (__pyx_t_1) {

        /* "pyreadstat/_readstat_parser.pyx":191
 *                 return DATE_FORMAT_DATE
 *             elif var_format in stata_datetime_formats:
 *                 return DATE_FORMAT_DATETIME             # <<<<<<<<<<<<<<
 *             elif var_format in stata_time_formats:
 *                 return DATE_FORMAT_TIME
 */
        __pyx_r = __pyx_e_10pyreadstat_16_readstat_parser_DATE_FORMAT_DATETIME;
        goto __pyx_L0;

        /* "pyreadstat/_readstat_parser.pyx":190
 *             if var_format in stata_date_formats:
 *                 return DATE_FORMAT_DATE
 *             elif var_format in stata_datetime_formats:             # <<<<<<<<<<<<<<
 *                 return DATE_FORMAT_DATETIME
 *             elif var_format in stata_time_formats:
 */
      }

      /* "pyreadstat/_readstat_parser.pyx":192
 *             elif var_format in stata_datetime_formats:
 *                 return DATE_FORMAT_DATETIME
 *             elif var_format in stata_time_formats:             # <<<<<<<<<<<<<<
 *                 return DATE_FORMAT_TIME
 *         else:
 */
      __pyx_t_1 = (__Pyx_PySequence_ContainsTF(__pyx_v_var_format, __pyx_v_10pyreadstat_16_readstat_parser_stata_time_formats, Py_EQ)); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 192, __pyx_L1_error)
      if (__pyx_t_1) {

        /* "pyreadstat/_readstat_parser.pyx":193
 *                 return DATE_FORMAT_DATETIME
 *             elif var_format in stata_time_formats:
 *                 return DATE_FORMAT_TIME             # <<<<<<<<<<<<<<
 *         else:
 *             return DATE_FORMAT_NOTADATE
 */
        __pyx_r = __pyx_e_10pyreadstat_16_readstat_parser_DATE_FORMAT_TIME;
        goto __pyx_L0;

        /* "pyreadstat/_readstat_parser.pyx":192
 *             elif var_format in stata_datetime_formats:
 *                 return DATE_FORMAT_DATETIME
 *             elif var_format in stata_time_formats:             # <<<<<<<<<<<<<<
 *                 return DATE_FORMAT_TIME
 *         else:
 */
      }

      /* "pyreadstat/_readstat_parser.pyx":187
 * 
 *     elif file_format == FILE_FORMAT_STATA:
 *         if var_format in stata_all_formats:             # <<<<<<<<<<<<<<
 *             if var_format in stata_date_formats:
 *                 return DATE_FORMAT_DATE
 */
      goto __pyx_L7;
    }

    /* "pyreadstat/_readstat_parser.pyx":195
 *                 return DATE_FORMAT_TIME
 *         else:
 *             return DATE_FORMAT_NOTADATE             # <<<<<<<<<<<<<<
 * 
 * cdef object transform_datetime(py_datetime_format var_format, double tstamp, py_file_format file_format, object origin, bint dates_as_pandas):
 */
    /*else*/ {
      __pyx_r = __pyx_e_10pyreadstat_16_readstat_parser_DATE_FORMAT_NOTADATE;
      goto __pyx_L0;
    }
    __pyx_L7:;

    /* "pyreadstat/_readstat_parser.pyx":186
 *             return DATE_FORMAT_NOTADATE
 * 
 *     elif file_format == FILE_FORMAT_STATA:             # <<<<<<<<<<<<<<
 *         if var_format in stata_all_formats:
 *             if var_format in stata_date_formats:
 */
    break;
    default: break;
  }

  /* "pyreadstat/_readstat_parser.pyx":160
 * 
 * 
 * cdef py_datetime_format transform_variable_format(str var_format, py_file_format file_format):             # <<<<<<<<<<<<<<
 *     """
 *     Transforms a readstat var_format to a date, datetime or time format label
 */

  /* function exit code */
  __pyx_r = (__pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format) 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyreadstat._readstat_parser.transform_variable_format", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = (__pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format) 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyreadstat/_readstat_parser.pyx":197
 *             return DATE_FORMAT_NOTADATE
 * 
 * cdef object transform_datetime(py_datetime_format var_format, double tstamp, py_file_format file_format, object origin, bint dates_as_pandas):             # <<<<<<<<<<<<<<
 *     """
 *     Transforms a tstamp integer value to a date, time or datetime pyton object.
 */

static PyObject *__pyx_f_10pyreadstat_16_readstat_parser_transform_datetime(__pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format __pyx_v_var_format, double __pyx_v_tstamp, __pyx_t_10pyreadstat_16_readstat_parser_py_file_format __pyx_v_file_format, PyObject *__pyx_v_origin, int __pyx_v_dates_as_pandas) {
  PyObject *__pyx_v_tdelta = 0;
  int __pyx_v_days;
  int __pyx_v_secs;
  double __pyx_v_msecs;
  int __pyx_v_usecs;
  PyObject *__pyx_v_mydat = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("transform_datetime", 0);

  /* "pyreadstat/_readstat_parser.pyx":210
 *     cdef object mydat
 * 
 *     if var_format == DATE_FORMAT_DATE:             # <<<<<<<<<<<<<<
 *         if file_format == FILE_FORMAT_SPSS:
 *             # tstamp is in seconds
 */
  switch (__pyx_v_var_format) {
    case __pyx_e_10pyreadstat_16_readstat_parser_DATE_FORMAT_DATE:

    /* "pyreadstat/_readstat_parser.pyx":211
 * 
 *     if var_format == DATE_FORMAT_DATE:
 *         if file_format == FILE_FORMAT_SPSS:             # <<<<<<<<<<<<<<
 *             # tstamp is in seconds
 *             days = <int> (floor(tstamp / 86400))
 */
    __pyx_t_1 = (__pyx_v_file_format == __pyx_e_10pyreadstat_16_readstat_parser_FILE_FORMAT_SPSS);
    if (__pyx_t_1) {

      /* "pyreadstat/_readstat_parser.pyx":213
 *         if file_format == FILE_FORMAT_SPSS:
 *             # tstamp is in seconds
 *             days = <int> (floor(tstamp / 86400))             # <<<<<<<<<<<<<<
 *             secs = <int> (tstamp % 86400)
 *             tdelta = timedelta_new(days, secs, 0)
 */
      __pyx_v_days = ((int)floor((__pyx_v_tstamp / 86400.0)));

      /* "pyreadstat/_readstat_parser.pyx":214
 *             # tstamp is in seconds
 *             days = <int> (floor(tstamp / 86400))
 *             secs = <int> (tstamp % 86400)             # <<<<<<<<<<<<<<
 *             tdelta = timedelta_new(days, secs, 0)
 *             #tdelta = timedelta(seconds=tstamp)
 */
      __pyx_v_secs = ((int)__Pyx_mod_double(__pyx_v_tstamp, 86400.0));

      /* "pyreadstat/_readstat_parser.pyx":215
 *             days = <int> (floor(tstamp / 86400))
 *             secs = <int> (tstamp % 86400)
 *             tdelta = timedelta_new(days, secs, 0)             # <<<<<<<<<<<<<<
 *             #tdelta = timedelta(seconds=tstamp)
 *         else:
 */
      __pyx_t_2 = ((PyObject *)__pyx_f_7cpython_8datetime_timedelta_new(__pyx_v_days, __pyx_v_secs, 0)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 215, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_v_tdelta = __pyx_t_2;
      __pyx_t_2 = 0;

      /* "pyreadstat/_readstat_parser.pyx":211
 * 
 *     if var_format == DATE_FORMAT_DATE:
 *         if file_format == FILE_FORMAT_SPSS:             # <<<<<<<<<<<<<<
 *             # tstamp is in seconds
 *             days = <int> (floor(tstamp / 86400))
 */
      goto __pyx_L3;
    }

    /* "pyreadstat/_readstat_parser.pyx":219
 *         else:
 *             # tstamp is in days
 *             days = <int> tstamp             # <<<<<<<<<<<<<<
 *             tdelta = timedelta_new(days, 0, 0)
 *             #tdelta = timedelta(days=tstamp)
 */
    /*else*/ {
      __pyx_v_days = ((int)__pyx_v_tstamp);

      /* "pyreadstat/_readstat_parser.pyx":220
 *             # tstamp is in days
 *             days = <int> tstamp
 *             tdelta = timedelta_new(days, 0, 0)             # <<<<<<<<<<<<<<
 *             #tdelta = timedelta(days=tstamp)
 *         mydat = origin + tdelta
 */
      __pyx_t_2 = ((PyObject *)__pyx_f_7cpython_8datetime_timedelta_new(__pyx_v_days, 0, 0)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 220, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_v_tdelta = __pyx_t_2;
      __pyx_t_2 = 0;
    }
    __pyx_L3:;

    /* "pyreadstat/_readstat_parser.pyx":222
 *             tdelta = timedelta_new(days, 0, 0)
 *             #tdelta = timedelta(days=tstamp)
 *         mydat = origin + tdelta             # <<<<<<<<<<<<<<
 *         if dates_as_pandas:
 *             return mydat
 */
    __pyx_t_2 = PyNumber_Add(__pyx_v_origin, __pyx_v_tdelta); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 222, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_mydat = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "pyreadstat/_readstat_parser.pyx":223
 *             #tdelta = timedelta(days=tstamp)
 *         mydat = origin + tdelta
 *         if dates_as_pandas:             # <<<<<<<<<<<<<<
 *             return mydat
 *         else:
 */
    if (__pyx_v_dates_as_pandas) {

      /* "pyreadstat/_readstat_parser.pyx":224
 *         mydat = origin + tdelta
 *         if dates_as_pandas:
 *             return mydat             # <<<<<<<<<<<<<<
 *         else:
 *             return mydat.date()
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_v_mydat);
      __pyx_r = __pyx_v_mydat;
      goto __pyx_L0;

      /* "pyreadstat/_readstat_parser.pyx":223
 *             #tdelta = timedelta(days=tstamp)
 *         mydat = origin + tdelta
 *         if dates_as_pandas:             # <<<<<<<<<<<<<<
 *             return mydat
 *         else:
 */
    }

    /* "pyreadstat/_readstat_parser.pyx":226
 *             return mydat
 *         else:
 *             return mydat.date()             # <<<<<<<<<<<<<<
 *     elif var_format == DATE_FORMAT_DATETIME:
 *         if file_format == FILE_FORMAT_STATA:
 */
    /*else*/ {
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_mydat, __pyx_n_s_date); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 226, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = NULL;
      __pyx_t_5 = 0;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
          __pyx_t_5 = 1;
        }
      }
      {
        PyObject *__pyx_callargs[1] = {__pyx_t_4, };
        __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 226, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      }
      __pyx_r = __pyx_t_2;
      __pyx_t_2 = 0;
      goto __pyx_L0;
    }

    /* "pyreadstat/_readstat_parser.pyx":210
 *     cdef object mydat
 * 
 *     if var_format == DATE_FORMAT_DATE:             # <<<<<<<<<<<<<<
 *         if file_format == FILE_FORMAT_SPSS:
 *             # tstamp is in seconds
 */
    break;
    case __pyx_e_10pyreadstat_16_readstat_parser_DATE_FORMAT_DATETIME:

    /* "pyreadstat/_readstat_parser.pyx":228
 *             return mydat.date()
 *     elif var_format == DATE_FORMAT_DATETIME:
 *         if file_format == FILE_FORMAT_STATA:             # <<<<<<<<<<<<<<
 *             # tstamp is in millisecons
 *             days = <int> (floor(tstamp / 86400000))
 */
    __pyx_t_1 = (__pyx_v_file_format == __pyx_e_10pyreadstat_16_readstat_parser_FILE_FORMAT_STATA);
    if (__pyx_t_1) {

      /* "pyreadstat/_readstat_parser.pyx":230
 *         if file_format == FILE_FORMAT_STATA:
 *             # tstamp is in millisecons
 *             days = <int> (floor(tstamp / 86400000))             # <<<<<<<<<<<<<<
 *             msecs = tstamp % 86400000
 *             secs = <int> (msecs/1000)
 */
      __pyx_v_days = ((int)floor((__pyx_v_tstamp / 86400000.0)));

      /* "pyreadstat/_readstat_parser.pyx":231
 *             # tstamp is in millisecons
 *             days = <int> (floor(tstamp / 86400000))
 *             msecs = tstamp % 86400000             # <<<<<<<<<<<<<<
 *             secs = <int> (msecs/1000)
 *             usecs = <int> ((msecs % 1000) * 1000 )
 */
      __pyx_v_msecs = __Pyx_mod_double(__pyx_v_tstamp, 86400000.0);

      /* "pyreadstat/_readstat_parser.pyx":232
 *             days = <int> (floor(tstamp / 86400000))
 *             msecs = tstamp % 86400000
 *             secs = <int> (msecs/1000)             # <<<<<<<<<<<<<<
 *             usecs = <int> ((msecs % 1000) * 1000 )
 *             tdelta = timedelta_new(days, secs, usecs)
 */
      __pyx_v_secs = ((int)(__pyx_v_msecs / 1000.0));

      /* "pyreadstat/_readstat_parser.pyx":233
 *             msecs = tstamp % 86400000
 *             secs = <int> (msecs/1000)
 *             usecs = <int> ((msecs % 1000) * 1000 )             # <<<<<<<<<<<<<<
 *             tdelta = timedelta_new(days, secs, usecs)
 *             #tdelta = timedelta(milliseconds=tstamp)
 */
      __pyx_v_usecs = ((int)(__Pyx_mod_double(__pyx_v_msecs, 1000.0) * 1000.0));

      /* "pyreadstat/_readstat_parser.pyx":234
 *             secs = <int> (msecs/1000)
 *             usecs = <int> ((msecs % 1000) * 1000 )
 *             tdelta = timedelta_new(days, secs, usecs)             # <<<<<<<<<<<<<<
 *             #tdelta = timedelta(milliseconds=tstamp)
 *         else:
 */
      __pyx_t_2 = ((PyObject *)__pyx_f_7cpython_8datetime_timedelta_new(__pyx_v_days, __pyx_v_secs, __pyx_v_usecs)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 234, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_v_tdelta = __pyx_t_2;
      __pyx_t_2 = 0;

      /* "pyreadstat/_readstat_parser.pyx":228
 *             return mydat.date()
 *     elif var_format == DATE_FORMAT_DATETIME:
 *         if file_format == FILE_FORMAT_STATA:             # <<<<<<<<<<<<<<
 *             # tstamp is in millisecons
 *             days = <int> (floor(tstamp / 86400000))
 */
      goto __pyx_L5;
    }

    /* "pyreadstat/_readstat_parser.pyx":238
 *         else:
 *             # tstamp in seconds
 *             days = <int> (floor(tstamp / 86400))             # <<<<<<<<<<<<<<
 *             secs = <int> (tstamp % 86400)
 *             tdelta = timedelta_new(days, secs, 0)
 */
    /*else*/ {
      __pyx_v_days = ((int)floor((__pyx_v_tstamp / 86400.0)));

      /* "pyreadstat/_readstat_parser.pyx":239
 *             # tstamp in seconds
 *             days = <int> (floor(tstamp / 86400))
 *             secs = <int> (tstamp % 86400)             # <<<<<<<<<<<<<<
 *             tdelta = timedelta_new(days, secs, 0)
 *             #tdelta = timedelta(seconds=tstamp)
 */
      __pyx_v_secs = ((int)__Pyx_mod_double(__pyx_v_tstamp, 86400.0));

      /* "pyreadstat/_readstat_parser.pyx":240
 *             days = <int> (floor(tstamp / 86400))
 *             secs = <int> (tstamp % 86400)
 *             tdelta = timedelta_new(days, secs, 0)             # <<<<<<<<<<<<<<
 *             #tdelta = timedelta(seconds=tstamp)
 *         mydat = origin + tdelta
 */
      __pyx_t_2 = ((PyObject *)__pyx_f_7cpython_8datetime_timedelta_new(__pyx_v_days, __pyx_v_secs, 0)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 240, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_v_tdelta = __pyx_t_2;
      __pyx_t_2 = 0;
    }
    __pyx_L5:;

    /* "pyreadstat/_readstat_parser.pyx":242
 *             tdelta = timedelta_new(days, secs, 0)
 *             #tdelta = timedelta(seconds=tstamp)
 *         mydat = origin + tdelta             # <<<<<<<<<<<<<<
 *         return mydat
 *     elif var_format == DATE_FORMAT_TIME:
 */
    __pyx_t_2 = PyNumber_Add(__pyx_v_origin, __pyx_v_tdelta); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 242, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_mydat = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "pyreadstat/_readstat_parser.pyx":243
 *             #tdelta = timedelta(seconds=tstamp)
 *         mydat = origin + tdelta
 *         return mydat             # <<<<<<<<<<<<<<
 *     elif var_format == DATE_FORMAT_TIME:
 *         if file_format == FILE_FORMAT_STATA:
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_mydat);
    __pyx_r = __pyx_v_mydat;
    goto __pyx_L0;

    /* "pyreadstat/_readstat_parser.pyx":227
 *         else:
 *             return mydat.date()
 *     elif var_format == DATE_FORMAT_DATETIME:             # <<<<<<<<<<<<<<
 *         if file_format == FILE_FORMAT_STATA:
 *             # tstamp is in millisecons
 */
    break;
    case __pyx_e_10pyreadstat_16_readstat_parser_DATE_FORMAT_TIME:

    /* "pyreadstat/_readstat_parser.pyx":245
 *         return mydat
 *     elif var_format == DATE_FORMAT_TIME:
 *         if file_format == FILE_FORMAT_STATA:             # <<<<<<<<<<<<<<
 *             # tstamp is in millisecons
 *             days = <int> (floor(tstamp / 86400000))
 */
    __pyx_t_1 = (__pyx_v_file_format == __pyx_e_10pyreadstat_16_readstat_parser_FILE_FORMAT_STATA);
    if (__pyx_t_1) {

      /* "pyreadstat/_readstat_parser.pyx":247
 *         if file_format == FILE_FORMAT_STATA:
 *             # tstamp is in millisecons
 *             days = <int> (floor(tstamp / 86400000))             # <<<<<<<<<<<<<<
 *             msecs = tstamp % 86400000
 *             secs = <int> (msecs/1000)
 */
      __pyx_v_days = ((int)floor((__pyx_v_tstamp / 86400000.0)));

      /* "pyreadstat/_readstat_parser.pyx":248
 *             # tstamp is in millisecons
 *             days = <int> (floor(tstamp / 86400000))
 *             msecs = tstamp % 86400000             # <<<<<<<<<<<<<<
 *             secs = <int> (msecs/1000)
 *             usecs = <int> ((msecs % 1000) * 1000 )
 */
      __pyx_v_msecs = __Pyx_mod_double(__pyx_v_tstamp, 86400000.0);

      /* "pyreadstat/_readstat_parser.pyx":249
 *             days = <int> (floor(tstamp / 86400000))
 *             msecs = tstamp % 86400000
 *             secs = <int> (msecs/1000)             # <<<<<<<<<<<<<<
 *             usecs = <int> ((msecs % 1000) * 1000 )
 *             tdelta = timedelta_new(days, secs, usecs)
 */
      __pyx_v_secs = ((int)(__pyx_v_msecs / 1000.0));

      /* "pyreadstat/_readstat_parser.pyx":250
 *             msecs = tstamp % 86400000
 *             secs = <int> (msecs/1000)
 *             usecs = <int> ((msecs % 1000) * 1000 )             # <<<<<<<<<<<<<<
 *             tdelta = timedelta_new(days, secs, usecs)
 *             #tdelta = timedelta(milliseconds=tstamp)
 */
      __pyx_v_usecs = ((int)(__Pyx_mod_double(__pyx_v_msecs, 1000.0) * 1000.0));

      /* "pyreadstat/_readstat_parser.pyx":251
 *             secs = <int> (msecs/1000)
 *             usecs = <int> ((msecs % 1000) * 1000 )
 *             tdelta = timedelta_new(days, secs, usecs)             # <<<<<<<<<<<<<<
 *             #tdelta = timedelta(milliseconds=tstamp)
 *         else:
 */
      __pyx_t_2 = ((PyObject *)__pyx_f_7cpython_8datetime_timedelta_new(__pyx_v_days, __pyx_v_secs, __pyx_v_usecs)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 251, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_v_tdelta = __pyx_t_2;
      __pyx_t_2 = 0;

      /* "pyreadstat/_readstat_parser.pyx":245
 *         return mydat
 *     elif var_format == DATE_FORMAT_TIME:
 *         if file_format == FILE_FORMAT_STATA:             # <<<<<<<<<<<<<<
 *             # tstamp is in millisecons
 *             days = <int> (floor(tstamp / 86400000))
 */
      goto __pyx_L6;
    }

    /* "pyreadstat/_readstat_parser.pyx":255
 *         else:
 *             # tstamp in seconds
 *             days = <int> (floor(tstamp / 86400))             # <<<<<<<<<<<<<<
 *             secs = <int> (tstamp % 86400)
 *             tdelta = timedelta_new(days, secs, 0)
 */
    /*else*/ {
      __pyx_v_days = ((int)floor((__pyx_v_tstamp / 86400.0)));

      /* "pyreadstat/_readstat_parser.pyx":256
 *             # tstamp in seconds
 *             days = <int> (floor(tstamp / 86400))
 *             secs = <int> (tstamp % 86400)             # <<<<<<<<<<<<<<
 *             tdelta = timedelta_new(days, secs, 0)
 *             #tdelta = timedelta(seconds=tstamp)
 */
      __pyx_v_secs = ((int)__Pyx_mod_double(__pyx_v_tstamp, 86400.0));

      /* "pyreadstat/_readstat_parser.pyx":257
 *             days = <int> (floor(tstamp / 86400))
 *             secs = <int> (tstamp % 86400)
 *             tdelta = timedelta_new(days, secs, 0)             # <<<<<<<<<<<<<<
 *             #tdelta = timedelta(seconds=tstamp)
 *         mydat = origin + tdelta
 */
      __pyx_t_2 = ((PyObject *)__pyx_f_7cpython_8datetime_timedelta_new(__pyx_v_days, __pyx_v_secs, 0)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 257, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_v_tdelta = __pyx_t_2;
      __pyx_t_2 = 0;
    }
    __pyx_L6:;

    /* "pyreadstat/_readstat_parser.pyx":259
 *             tdelta = timedelta_new(days, secs, 0)
 *             #tdelta = timedelta(seconds=tstamp)
 *         mydat = origin + tdelta             # <<<<<<<<<<<<<<
 *         return mydat.time()
 * 
 */
    __pyx_t_2 = PyNumber_Add(__pyx_v_origin, __pyx_v_tdelta); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 259, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_mydat = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "pyreadstat/_readstat_parser.pyx":260
 *             #tdelta = timedelta(seconds=tstamp)
 *         mydat = origin + tdelta
 *         return mydat.time()             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_mydat, __pyx_n_s_time); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 260, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = NULL;
    __pyx_t_5 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_5 = 1;
      }
    }
    {
      PyObject *__pyx_callargs[1] = {__pyx_t_4, };
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 260, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "pyreadstat/_readstat_parser.pyx":244
 *         mydat = origin + tdelta
 *         return mydat
 *     elif var_format == DATE_FORMAT_TIME:             # <<<<<<<<<<<<<<
 *         if file_format == FILE_FORMAT_STATA:
 *             # tstamp is in millisecons
 */
    break;
    default: break;
  }

  /* "pyreadstat/_readstat_parser.pyx":197
 *             return DATE_FORMAT_NOTADATE
 * 
 * cdef object transform_datetime(py_datetime_format var_format, double tstamp, py_file_format file_format, object origin, bint dates_as_pandas):             # <<<<<<<<<<<<<<
 *     """
 *     Transforms a tstamp integer value to a date, time or datetime pyton object.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pyreadstat._readstat_parser.transform_datetime", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_tdelta);
  __Pyx_XDECREF(__pyx_v_mydat);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyreadstat/_readstat_parser.pyx":263
 * 
 * 
 * cdef object convert_readstat_to_python_value(readstat_value_t value, int index, data_container dc):             # <<<<<<<<<<<<<<
 *     """
 *     Converts a readstat value to a python value.
 */

static PyObject *__pyx_f_10pyreadstat_16_readstat_parser_convert_readstat_to_python_value(readstat_value_t __pyx_v_value, int __pyx_v_index, struct __pyx_obj_10pyreadstat_16_readstat_parser_data_container *__pyx_v_dc) {
  readstat_type_t __pyx_v_var_type;
  __pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format __pyx_v_var_format;
  __pyx_t_10pyreadstat_16_readstat_parser_py_variable_format __pyx_v_pyformat;
  PyObject *__pyx_v_origin = 0;
  int __pyx_v_dates_as_pandas;
  __pyx_t_10pyreadstat_16_readstat_parser_py_file_format __pyx_v_file_format;
  PyObject *__pyx_v_result = 0;
  char *__pyx_v_c_str_value;
  PyObject *__pyx_v_py_str_value = 0;
  int8_t __pyx_v_c_int8_value;
  int16_t __pyx_v_c_int16_value;
  int32_t __pyx_v_c_int32_value;
  float __pyx_v_c_float_value;
  double __pyx_v_c_double_value;
  long __pyx_v_py_long_value;
  double __pyx_v_py_float_value;
  double __pyx_v_tstamp;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  readstat_type_t __pyx_t_2;
  __pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format __pyx_t_3;
  int __pyx_t_4;
  __pyx_t_10pyreadstat_16_readstat_parser_py_file_format __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("convert_readstat_to_python_value", 0);

  /* "pyreadstat/_readstat_parser.pyx":290
 *     cdef double tstamp
 * 
 *     var_type = dc.col_dtypes[index]             # <<<<<<<<<<<<<<
 *     var_format = dc.col_formats[index]
 *     origin = dc.origin
 */
  if (unlikely(__pyx_v_dc->col_dtypes == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 290, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_List(__pyx_v_dc->col_dtypes, __pyx_v_index, int, 1, __Pyx_PyInt_From_int, 1, 1, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 290, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = ((readstat_type_t)__Pyx_PyInt_As_readstat_type_t(__pyx_t_1)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 290, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_var_type = __pyx_t_2;

  /* "pyreadstat/_readstat_parser.pyx":291
 * 
 *     var_type = dc.col_dtypes[index]
 *     var_format = dc.col_formats[index]             # <<<<<<<<<<<<<<
 *     origin = dc.origin
 *     dates_as_pandas = dc.dates_as_pandas
 */
  if (unlikely(__pyx_v_dc->col_formats == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 291, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_List(__pyx_v_dc->col_formats, __pyx_v_index, int, 1, __Pyx_PyInt_From_int, 1, 1, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 291, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = ((__pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format)__Pyx_PyInt_As___pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format(__pyx_t_1)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 291, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_var_format = __pyx_t_3;

  /* "pyreadstat/_readstat_parser.pyx":292
 *     var_type = dc.col_dtypes[index]
 *     var_format = dc.col_formats[index]
 *     origin = dc.origin             # <<<<<<<<<<<<<<
 *     dates_as_pandas = dc.dates_as_pandas
 *     file_format = dc.file_format
 */
  __pyx_t_1 = __pyx_v_dc->origin;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_origin = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pyreadstat/_readstat_parser.pyx":293
 *     var_format = dc.col_formats[index]
 *     origin = dc.origin
 *     dates_as_pandas = dc.dates_as_pandas             # <<<<<<<<<<<<<<
 *     file_format = dc.file_format
 * 
 */
  __pyx_t_4 = __pyx_v_dc->dates_as_pandas;
  __pyx_v_dates_as_pandas = __pyx_t_4;

  /* "pyreadstat/_readstat_parser.pyx":294
 *     origin = dc.origin
 *     dates_as_pandas = dc.dates_as_pandas
 *     file_format = dc.file_format             # <<<<<<<<<<<<<<
 * 
 *     # transform to values cython can deal with
 */
  __pyx_t_5 = __pyx_v_dc->file_format;
  __pyx_v_file_format = __pyx_t_5;

  /* "pyreadstat/_readstat_parser.pyx":297
 * 
 *     # transform to values cython can deal with
 *     if var_type == READSTAT_TYPE_STRING or var_type == READSTAT_TYPE_STRING_REF:             # <<<<<<<<<<<<<<
 *         c_str_value = readstat_string_value(value)
 *         if c_str_value != NULL:
 */
  switch (__pyx_v_var_type) {
    case READSTAT_TYPE_STRING:
    case READSTAT_TYPE_STRING_REF:

    /* "pyreadstat/_readstat_parser.pyx":298
 *     # transform to values cython can deal with
 *     if var_type == READSTAT_TYPE_STRING or var_type == READSTAT_TYPE_STRING_REF:
 *         c_str_value = readstat_string_value(value)             # <<<<<<<<<<<<<<
 *         if c_str_value != NULL:
 *             py_str_value = <str> c_str_value
 */
    __pyx_v_c_str_value = readstat_string_value(__pyx_v_value);

    /* "pyreadstat/_readstat_parser.pyx":299
 *     if var_type == READSTAT_TYPE_STRING or var_type == READSTAT_TYPE_STRING_REF:
 *         c_str_value = readstat_string_value(value)
 *         if c_str_value != NULL:             # <<<<<<<<<<<<<<
 *             py_str_value = <str> c_str_value
 *         else:
 */
    __pyx_t_4 = (__pyx_v_c_str_value != NULL);
    if (__pyx_t_4) {

      /* "pyreadstat/_readstat_parser.pyx":300
 *         c_str_value = readstat_string_value(value)
 *         if c_str_value != NULL:
 *             py_str_value = <str> c_str_value             # <<<<<<<<<<<<<<
 *         else:
 *             py_str_value = ''
 */
      __pyx_t_1 = __Pyx_PyStr_FromString(__pyx_v_c_str_value); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 300, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_6 = __pyx_t_1;
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_v_py_str_value = ((PyObject*)__pyx_t_6);
      __pyx_t_6 = 0;

      /* "pyreadstat/_readstat_parser.pyx":299
 *     if var_type == READSTAT_TYPE_STRING or var_type == READSTAT_TYPE_STRING_REF:
 *         c_str_value = readstat_string_value(value)
 *         if c_str_value != NULL:             # <<<<<<<<<<<<<<
 *             py_str_value = <str> c_str_value
 *         else:
 */
      goto __pyx_L3;
    }

    /* "pyreadstat/_readstat_parser.pyx":302
 *             py_str_value = <str> c_str_value
 *         else:
 *             py_str_value = ''             # <<<<<<<<<<<<<<
 *         pyformat = VAR_FORMAT_STRING
 *     elif var_type == READSTAT_TYPE_INT8:
 */
    /*else*/ {
      __Pyx_INCREF(__pyx_kp_s__2);
      __pyx_v_py_str_value = __pyx_kp_s__2;
    }
    __pyx_L3:;

    /* "pyreadstat/_readstat_parser.pyx":303
 *         else:
 *             py_str_value = ''
 *         pyformat = VAR_FORMAT_STRING             # <<<<<<<<<<<<<<
 *     elif var_type == READSTAT_TYPE_INT8:
 *         c_int8_value = readstat_int8_value(value)
 */
    __pyx_v_pyformat = __pyx_e_10pyreadstat_16_readstat_parser_VAR_FORMAT_STRING;

    /* "pyreadstat/_readstat_parser.pyx":297
 * 
 *     # transform to values cython can deal with
 *     if var_type == READSTAT_TYPE_STRING or var_type == READSTAT_TYPE_STRING_REF:             # <<<<<<<<<<<<<<
 *         c_str_value = readstat_string_value(value)
 *         if c_str_value != NULL:
 */
    break;
    case READSTAT_TYPE_INT8:

    /* "pyreadstat/_readstat_parser.pyx":305
 *         pyformat = VAR_FORMAT_STRING
 *     elif var_type == READSTAT_TYPE_INT8:
 *         c_int8_value = readstat_int8_value(value)             # <<<<<<<<<<<<<<
 *         py_long_value = <long> c_int8_value
 *         pyformat = VAR_FORMAT_LONG
 */
    __pyx_v_c_int8_value = readstat_int8_value(__pyx_v_value);

    /* "pyreadstat/_readstat_parser.pyx":306
 *     elif var_type == READSTAT_TYPE_INT8:
 *         c_int8_value = readstat_int8_value(value)
 *         py_long_value = <long> c_int8_value             # <<<<<<<<<<<<<<
 *         pyformat = VAR_FORMAT_LONG
 *     elif var_type == READSTAT_TYPE_INT16:
 */
    __pyx_v_py_long_value = ((long)__pyx_v_c_int8_value);

    /* "pyreadstat/_readstat_parser.pyx":307
 *         c_int8_value = readstat_int8_value(value)
 *         py_long_value = <long> c_int8_value
 *         pyformat = VAR_FORMAT_LONG             # <<<<<<<<<<<<<<
 *     elif var_type == READSTAT_TYPE_INT16:
 *         c_int16_value = readstat_int16_value(value)
 */
    __pyx_v_pyformat = __pyx_e_10pyreadstat_16_readstat_parser_VAR_FORMAT_LONG;

    /* "pyreadstat/_readstat_parser.pyx":304
 *             py_str_value = ''
 *         pyformat = VAR_FORMAT_STRING
 *     elif var_type == READSTAT_TYPE_INT8:             # <<<<<<<<<<<<<<
 *         c_int8_value = readstat_int8_value(value)
 *         py_long_value = <long> c_int8_value
 */
    break;
    case READSTAT_TYPE_INT16:

    /* "pyreadstat/_readstat_parser.pyx":309
 *         pyformat = VAR_FORMAT_LONG
 *     elif var_type == READSTAT_TYPE_INT16:
 *         c_int16_value = readstat_int16_value(value)             # <<<<<<<<<<<<<<
 *         py_long_value = <long> c_int16_value
 *         pyformat = VAR_FORMAT_LONG
 */
    __pyx_v_c_int16_value = readstat_int16_value(__pyx_v_value);

    /* "pyreadstat/_readstat_parser.pyx":310
 *     elif var_type == READSTAT_TYPE_INT16:
 *         c_int16_value = readstat_int16_value(value)
 *         py_long_value = <long> c_int16_value             # <<<<<<<<<<<<<<
 *         pyformat = VAR_FORMAT_LONG
 *     elif var_type == READSTAT_TYPE_INT32:
 */
    __pyx_v_py_long_value = ((long)__pyx_v_c_int16_value);

    /* "pyreadstat/_readstat_parser.pyx":311
 *         c_int16_value = readstat_int16_value(value)
 *         py_long_value = <long> c_int16_value
 *         pyformat = VAR_FORMAT_LONG             # <<<<<<<<<<<<<<
 *     elif var_type == READSTAT_TYPE_INT32:
 *         c_int32_value = readstat_int32_value(value)
 */
    __pyx_v_pyformat = __pyx_e_10pyreadstat_16_readstat_parser_VAR_FORMAT_LONG;

    /* "pyreadstat/_readstat_parser.pyx":308
 *         py_long_value = <long> c_int8_value
 *         pyformat = VAR_FORMAT_LONG
 *     elif var_type == READSTAT_TYPE_INT16:             # <<<<<<<<<<<<<<
 *         c_int16_value = readstat_int16_value(value)
 *         py_long_value = <long> c_int16_value
 */
    break;
    case READSTAT_TYPE_INT32:

    /* "pyreadstat/_readstat_parser.pyx":313
 *         pyformat = VAR_FORMAT_LONG
 *     elif var_type == READSTAT_TYPE_INT32:
 *         c_int32_value = readstat_int32_value(value)             # <<<<<<<<<<<<<<
 *         py_long_value = <long> c_int32_value
 *         pyformat = VAR_FORMAT_LONG
 */
    __pyx_v_c_int32_value = readstat_int32_value(__pyx_v_value);

    /* "pyreadstat/_readstat_parser.pyx":314
 *     elif var_type == READSTAT_TYPE_INT32:
 *         c_int32_value = readstat_int32_value(value)
 *         py_long_value = <long> c_int32_value             # <<<<<<<<<<<<<<
 *         pyformat = VAR_FORMAT_LONG
 *     elif var_type == READSTAT_TYPE_FLOAT:
 */
    __pyx_v_py_long_value = ((long)__pyx_v_c_int32_value);

    /* "pyreadstat/_readstat_parser.pyx":315
 *         c_int32_value = readstat_int32_value(value)
 *         py_long_value = <long> c_int32_value
 *         pyformat = VAR_FORMAT_LONG             # <<<<<<<<<<<<<<
 *     elif var_type == READSTAT_TYPE_FLOAT:
 *         c_float_value = readstat_float_value(value)
 */
    __pyx_v_pyformat = __pyx_e_10pyreadstat_16_readstat_parser_VAR_FORMAT_LONG;

    /* "pyreadstat/_readstat_parser.pyx":312
 *         py_long_value = <long> c_int16_value
 *         pyformat = VAR_FORMAT_LONG
 *     elif var_type == READSTAT_TYPE_INT32:             # <<<<<<<<<<<<<<
 *         c_int32_value = readstat_int32_value(value)
 *         py_long_value = <long> c_int32_value
 */
    break;
    case READSTAT_TYPE_FLOAT:

    /* "pyreadstat/_readstat_parser.pyx":317
 *         pyformat = VAR_FORMAT_LONG
 *     elif var_type == READSTAT_TYPE_FLOAT:
 *         c_float_value = readstat_float_value(value)             # <<<<<<<<<<<<<<
 *         py_float_value = <double> c_float_value
 *         pyformat = VAR_FORMAT_FLOAT
 */
    __pyx_v_c_float_value = readstat_float_value(__pyx_v_value);

    /* "pyreadstat/_readstat_parser.pyx":318
 *     elif var_type == READSTAT_TYPE_FLOAT:
 *         c_float_value = readstat_float_value(value)
 *         py_float_value = <double> c_float_value             # <<<<<<<<<<<<<<
 *         pyformat = VAR_FORMAT_FLOAT
 *     elif var_type == READSTAT_TYPE_DOUBLE:
 */
    __pyx_v_py_float_value = ((double)__pyx_v_c_float_value);

    /* "pyreadstat/_readstat_parser.pyx":319
 *         c_float_value = readstat_float_value(value)
 *         py_float_value = <double> c_float_value
 *         pyformat = VAR_FORMAT_FLOAT             # <<<<<<<<<<<<<<
 *     elif var_type == READSTAT_TYPE_DOUBLE:
 *         c_double_value = readstat_double_value(value);
 */
    __pyx_v_pyformat = __pyx_e_10pyreadstat_16_readstat_parser_VAR_FORMAT_FLOAT;

    /* "pyreadstat/_readstat_parser.pyx":316
 *         py_long_value = <long> c_int32_value
 *         pyformat = VAR_FORMAT_LONG
 *     elif var_type == READSTAT_TYPE_FLOAT:             # <<<<<<<<<<<<<<
 *         c_float_value = readstat_float_value(value)
 *         py_float_value = <double> c_float_value
 */
    break;
    case READSTAT_TYPE_DOUBLE:

    /* "pyreadstat/_readstat_parser.pyx":321
 *         pyformat = VAR_FORMAT_FLOAT
 *     elif var_type == READSTAT_TYPE_DOUBLE:
 *         c_double_value = readstat_double_value(value);             # <<<<<<<<<<<<<<
 *         py_float_value = <double> c_double_value
 *         pyformat = VAR_FORMAT_FLOAT
 */
    __pyx_v_c_double_value = readstat_double_value(__pyx_v_value);

    /* "pyreadstat/_readstat_parser.pyx":322
 *     elif var_type == READSTAT_TYPE_DOUBLE:
 *         c_double_value = readstat_double_value(value);
 *         py_float_value = <double> c_double_value             # <<<<<<<<<<<<<<
 *         pyformat = VAR_FORMAT_FLOAT
 *     else:
 */
    __pyx_v_py_float_value = ((double)__pyx_v_c_double_value);

    /* "pyreadstat/_readstat_parser.pyx":323
 *         c_double_value = readstat_double_value(value);
 *         py_float_value = <double> c_double_value
 *         pyformat = VAR_FORMAT_FLOAT             # <<<<<<<<<<<<<<
 *     else:
 *         raise PyreadstatError("Unkown data type")
 */
    __pyx_v_pyformat = __pyx_e_10pyreadstat_16_readstat_parser_VAR_FORMAT_FLOAT;

    /* "pyreadstat/_readstat_parser.pyx":320
 *         py_float_value = <double> c_float_value
 *         pyformat = VAR_FORMAT_FLOAT
 *     elif var_type == READSTAT_TYPE_DOUBLE:             # <<<<<<<<<<<<<<
 *         c_double_value = readstat_double_value(value);
 *         py_float_value = <double> c_double_value
 */
    break;
    default:

    /* "pyreadstat/_readstat_parser.pyx":325
 *         pyformat = VAR_FORMAT_FLOAT
 *     else:
 *         raise PyreadstatError("Unkown data type")             # <<<<<<<<<<<<<<
 * 
 *     # final transformation and storage
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_PyreadstatError); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 325, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
        __pyx_t_8 = 1;
      }
    }
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_kp_s_Unkown_data_type};
      __pyx_t_6 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 325, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(0, 325, __pyx_L1_error)
    break;
  }

  /* "pyreadstat/_readstat_parser.pyx":329
 *     # final transformation and storage
 * 
 *     if pyformat == VAR_FORMAT_STRING:             # <<<<<<<<<<<<<<
 *         if var_format == DATE_FORMAT_NOTADATE or dc.no_datetime_conversion:
 *             result = py_str_value
 */
  switch (__pyx_v_pyformat) {
    case __pyx_e_10pyreadstat_16_readstat_parser_VAR_FORMAT_STRING:

    /* "pyreadstat/_readstat_parser.pyx":330
 * 
 *     if pyformat == VAR_FORMAT_STRING:
 *         if var_format == DATE_FORMAT_NOTADATE or dc.no_datetime_conversion:             # <<<<<<<<<<<<<<
 *             result = py_str_value
 *         else:
 */
    __pyx_t_9 = (__pyx_v_var_format == __pyx_e_10pyreadstat_16_readstat_parser_DATE_FORMAT_NOTADATE);
    if (!__pyx_t_9) {
    } else {
      __pyx_t_4 = __pyx_t_9;
      goto __pyx_L5_bool_binop_done;
    }
    __pyx_t_4 = __pyx_v_dc->no_datetime_conversion;
    __pyx_L5_bool_binop_done:;
    if (likely(__pyx_t_4)) {

      /* "pyreadstat/_readstat_parser.pyx":331
 *     if pyformat == VAR_FORMAT_STRING:
 *         if var_format == DATE_FORMAT_NOTADATE or dc.no_datetime_conversion:
 *             result = py_str_value             # <<<<<<<<<<<<<<
 *         else:
 *             #str_byte_val = py_str_value.encode("UTF-8")
 */
      if (unlikely(!__pyx_v_py_str_value)) { __Pyx_RaiseUnboundLocalError("py_str_value"); __PYX_ERR(0, 331, __pyx_L1_error) }
      __Pyx_INCREF(__pyx_v_py_str_value);
      __pyx_v_result = __pyx_v_py_str_value;

      /* "pyreadstat/_readstat_parser.pyx":330
 * 
 *     if pyformat == VAR_FORMAT_STRING:
 *         if var_format == DATE_FORMAT_NOTADATE or dc.no_datetime_conversion:             # <<<<<<<<<<<<<<
 *             result = py_str_value
 *         else:
 */
      goto __pyx_L4;
    }

    /* "pyreadstat/_readstat_parser.pyx":334
 *         else:
 *             #str_byte_val = py_str_value.encode("UTF-8")
 *             raise PyreadstatError("STRING type with value '%s' with date type in column '%s'" % (py_str_value, dc.col_names[index]))             # <<<<<<<<<<<<<<
 *     elif pyformat == VAR_FORMAT_LONG:
 *         if var_format == DATE_FORMAT_NOTADATE or dc.no_datetime_conversion:
 */
    /*else*/ {
      __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_PyreadstatError); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 334, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (unlikely(!__pyx_v_py_str_value)) { __Pyx_RaiseUnboundLocalError("py_str_value"); __PYX_ERR(0, 334, __pyx_L1_error) }
      if (unlikely(__pyx_v_dc->col_names == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 334, __pyx_L1_error)
      }
      __pyx_t_7 = __Pyx_GetItemInt_List(__pyx_v_dc->col_names, __pyx_v_index, int, 1, __Pyx_PyInt_From_int, 1, 1, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 334, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_10 = PyTuple_New(2); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 334, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_INCREF(__pyx_v_py_str_value);
      __Pyx_GIVEREF(__pyx_v_py_str_value);
      PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_v_py_str_value);
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_t_7);
      __pyx_t_7 = 0;
      __pyx_t_7 = __Pyx_PyString_Format(__pyx_kp_s_STRING_type_with_value_s_with_da, __pyx_t_10); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 334, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __pyx_t_10 = NULL;
      __pyx_t_8 = 0;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
        __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_1);
        if (likely(__pyx_t_10)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_10);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_1, function);
          __pyx_t_8 = 1;
        }
      }
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_10, __pyx_t_7};
        __pyx_t_6 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 334, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      }
      __Pyx_Raise(__pyx_t_6, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __PYX_ERR(0, 334, __pyx_L1_error)
    }
    __pyx_L4:;

    /* "pyreadstat/_readstat_parser.pyx":329
 *     # final transformation and storage
 * 
 *     if pyformat == VAR_FORMAT_STRING:             # <<<<<<<<<<<<<<
 *         if var_format == DATE_FORMAT_NOTADATE or dc.no_datetime_conversion:
 *             result = py_str_value
 */
    break;
    case __pyx_e_10pyreadstat_16_readstat_parser_VAR_FORMAT_LONG:

    /* "pyreadstat/_readstat_parser.pyx":336
 *             raise PyreadstatError("STRING type with value '%s' with date type in column '%s'" % (py_str_value, dc.col_names[index]))
 *     elif pyformat == VAR_FORMAT_LONG:
 *         if var_format == DATE_FORMAT_NOTADATE or dc.no_datetime_conversion:             # <<<<<<<<<<<<<<
 *             result = py_long_value
 *         else:
 */
    __pyx_t_9 = (__pyx_v_var_format == __pyx_e_10pyreadstat_16_readstat_parser_DATE_FORMAT_NOTADATE);
    if (!__pyx_t_9) {
    } else {
      __pyx_t_4 = __pyx_t_9;
      goto __pyx_L8_bool_binop_done;
    }
    __pyx_t_4 = __pyx_v_dc->no_datetime_conversion;
    __pyx_L8_bool_binop_done:;
    if (__pyx_t_4) {

      /* "pyreadstat/_readstat_parser.pyx":337
 *     elif pyformat == VAR_FORMAT_LONG:
 *         if var_format == DATE_FORMAT_NOTADATE or dc.no_datetime_conversion:
 *             result = py_long_value             # <<<<<<<<<<<<<<
 *         else:
 *             tstamp = <double> py_long_value
 */
      __pyx_t_6 = __Pyx_PyInt_From_long(__pyx_v_py_long_value); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 337, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_v_result = __pyx_t_6;
      __pyx_t_6 = 0;

      /* "pyreadstat/_readstat_parser.pyx":336
 *             raise PyreadstatError("STRING type with value '%s' with date type in column '%s'" % (py_str_value, dc.col_names[index]))
 *     elif pyformat == VAR_FORMAT_LONG:
 *         if var_format == DATE_FORMAT_NOTADATE or dc.no_datetime_conversion:             # <<<<<<<<<<<<<<
 *             result = py_long_value
 *         else:
 */
      goto __pyx_L7;
    }

    /* "pyreadstat/_readstat_parser.pyx":339
 *             result = py_long_value
 *         else:
 *             tstamp = <double> py_long_value             # <<<<<<<<<<<<<<
 *             result = transform_datetime(var_format, tstamp, file_format, origin, dates_as_pandas)
 *     elif pyformat == VAR_FORMAT_FLOAT:
 */
    /*else*/ {
      __pyx_v_tstamp = ((double)__pyx_v_py_long_value);

      /* "pyreadstat/_readstat_parser.pyx":340
 *         else:
 *             tstamp = <double> py_long_value
 *             result = transform_datetime(var_format, tstamp, file_format, origin, dates_as_pandas)             # <<<<<<<<<<<<<<
 *     elif pyformat == VAR_FORMAT_FLOAT:
 *         if var_format == DATE_FORMAT_NOTADATE or dc.no_datetime_conversion:
 */
      __pyx_t_6 = __pyx_f_10pyreadstat_16_readstat_parser_transform_datetime(__pyx_v_var_format, __pyx_v_tstamp, __pyx_v_file_format, __pyx_v_origin, __pyx_v_dates_as_pandas); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 340, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_v_result = __pyx_t_6;
      __pyx_t_6 = 0;
    }
    __pyx_L7:;

    /* "pyreadstat/_readstat_parser.pyx":335
 *             #str_byte_val = py_str_value.encode("UTF-8")
 *             raise PyreadstatError("STRING type with value '%s' with date type in column '%s'" % (py_str_value, dc.col_names[index]))
 *     elif pyformat == VAR_FORMAT_LONG:             # <<<<<<<<<<<<<<
 *         if var_format == DATE_FORMAT_NOTADATE or dc.no_datetime_conversion:
 *             result = py_long_value
 */
    break;
    case __pyx_e_10pyreadstat_16_readstat_parser_VAR_FORMAT_FLOAT:

    /* "pyreadstat/_readstat_parser.pyx":342
 *             result = transform_datetime(var_format, tstamp, file_format, origin, dates_as_pandas)
 *     elif pyformat == VAR_FORMAT_FLOAT:
 *         if var_format == DATE_FORMAT_NOTADATE or dc.no_datetime_conversion:             # <<<<<<<<<<<<<<
 *             result = py_float_value
 *         else:
 */
    __pyx_t_9 = (__pyx_v_var_format == __pyx_e_10pyreadstat_16_readstat_parser_DATE_FORMAT_NOTADATE);
    if (!__pyx_t_9) {
    } else {
      __pyx_t_4 = __pyx_t_9;
      goto __pyx_L11_bool_binop_done;
    }
    __pyx_t_4 = __pyx_v_dc->no_datetime_conversion;
    __pyx_L11_bool_binop_done:;
    if (__pyx_t_4) {

      /* "pyreadstat/_readstat_parser.pyx":343
 *     elif pyformat == VAR_FORMAT_FLOAT:
 *         if var_format == DATE_FORMAT_NOTADATE or dc.no_datetime_conversion:
 *             result = py_float_value             # <<<<<<<<<<<<<<
 *         else:
 *             #tstamp = <int> py_float_value
 */
      __pyx_t_6 = PyFloat_FromDouble(__pyx_v_py_float_value); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 343, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_v_result = __pyx_t_6;
      __pyx_t_6 = 0;

      /* "pyreadstat/_readstat_parser.pyx":342
 *             result = transform_datetime(var_format, tstamp, file_format, origin, dates_as_pandas)
 *     elif pyformat == VAR_FORMAT_FLOAT:
 *         if var_format == DATE_FORMAT_NOTADATE or dc.no_datetime_conversion:             # <<<<<<<<<<<<<<
 *             result = py_float_value
 *         else:
 */
      goto __pyx_L10;
    }

    /* "pyreadstat/_readstat_parser.pyx":346
 *         else:
 *             #tstamp = <int> py_float_value
 *             tstamp = py_float_value             # <<<<<<<<<<<<<<
 *             result = transform_datetime(var_format, tstamp, file_format, origin, dates_as_pandas)
 *     #elif pyformat == VAR_FORMAT_MISSING:
 */
    /*else*/ {
      __pyx_v_tstamp = __pyx_v_py_float_value;

      /* "pyreadstat/_readstat_parser.pyx":347
 *             #tstamp = <int> py_float_value
 *             tstamp = py_float_value
 *             result = transform_datetime(var_format, tstamp, file_format, origin, dates_as_pandas)             # <<<<<<<<<<<<<<
 *     #elif pyformat == VAR_FORMAT_MISSING:
 *     #    pass
 */
      __pyx_t_6 = __pyx_f_10pyreadstat_16_readstat_parser_transform_datetime(__pyx_v_var_format, __pyx_v_tstamp, __pyx_v_file_format, __pyx_v_origin, __pyx_v_dates_as_pandas); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 347, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_v_result = __pyx_t_6;
      __pyx_t_6 = 0;
    }
    __pyx_L10:;

    /* "pyreadstat/_readstat_parser.pyx":341
 *             tstamp = <double> py_long_value
 *             result = transform_datetime(var_format, tstamp, file_format, origin, dates_as_pandas)
 *     elif pyformat == VAR_FORMAT_FLOAT:             # <<<<<<<<<<<<<<
 *         if var_format == DATE_FORMAT_NOTADATE or dc.no_datetime_conversion:
 *             result = py_float_value
 */
    break;
    default:

    /* "pyreadstat/_readstat_parser.pyx":351
 *     #    pass
 *     else:
 *         raise PyreadstatError("Failed convert C to python value")             # <<<<<<<<<<<<<<
 * 
 *     return result
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_PyreadstatError); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 351, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
        __pyx_t_8 = 1;
      }
    }
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_kp_s_Failed_convert_C_to_python_value};
      __pyx_t_6 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 351, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(0, 351, __pyx_L1_error)
    break;
  }

  /* "pyreadstat/_readstat_parser.pyx":353
 *         raise PyreadstatError("Failed convert C to python value")
 * 
 *     return result             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_result);
  __pyx_r = __pyx_v_result;
  goto __pyx_L0;

  /* "pyreadstat/_readstat_parser.pyx":263
 * 
 * 
 * cdef object convert_readstat_to_python_value(readstat_value_t value, int index, data_container dc):             # <<<<<<<<<<<<<<
 *     """
 *     Converts a readstat value to a python value.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("pyreadstat._readstat_parser.convert_readstat_to_python_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_origin);
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XDECREF(__pyx_v_py_str_value);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyreadstat/_readstat_parser.pyx":356
 * 
 * 
 * cdef int handle_metadata(readstat_metadata_t *metadata, void *ctx) except READSTAT_HANDLER_ABORT:             # <<<<<<<<<<<<<<
 *     """
 *     This function sets the number of observations(rows), number of variables
 */

static int __pyx_f_10pyreadstat_16_readstat_parser_handle_metadata(readstat_metadata_t *__pyx_v_metadata, void *__pyx_v_ctx) {
  int __pyx_v_var_count;
  int __pyx_v_obs_count;
  struct __pyx_obj_10pyreadstat_16_readstat_parser_data_container *__pyx_v_dc = 0;
  char *__pyx_v_flabel_orig;
  char *__pyx_v_fencoding_orig;
  PyObject *__pyx_v_flabel = 0;
  PyObject *__pyx_v_fencoding = 0;
  CYTHON_UNUSED int __pyx_v_metaonly;
  char *__pyx_v_table;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("handle_metadata", 0);

  /* "pyreadstat/_readstat_parser.pyx":365
 * 
 *     cdef int var_count, obs_count
 *     cdef  data_container dc = <data_container> ctx             # <<<<<<<<<<<<<<
 *     #cdef object row
 *     cdef char * flabel_orig
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_ctx);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_dc = ((struct __pyx_obj_10pyreadstat_16_readstat_parser_data_container *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyreadstat/_readstat_parser.pyx":375
 *     cdef int mtime
 * 
 *     metaonly = dc.metaonly             # <<<<<<<<<<<<<<
 * 
 *     var_count = readstat_get_var_count(metadata)
 */
  __pyx_t_2 = __pyx_v_dc->metaonly;
  __pyx_v_metaonly = __pyx_t_2;

  /* "pyreadstat/_readstat_parser.pyx":377
 *     metaonly = dc.metaonly
 * 
 *     var_count = readstat_get_var_count(metadata)             # <<<<<<<<<<<<<<
 *     if var_count<0:
 *         raise PyreadstatError("Failed to read number of variables")
 */
  __pyx_v_var_count = readstat_get_var_count(__pyx_v_metadata);

  /* "pyreadstat/_readstat_parser.pyx":378
 * 
 *     var_count = readstat_get_var_count(metadata)
 *     if var_count<0:             # <<<<<<<<<<<<<<
 *         raise PyreadstatError("Failed to read number of variables")
 *     obs_count = readstat_get_row_count(metadata)
 */
  __pyx_t_2 = (__pyx_v_var_count < 0);
  if (unlikely(__pyx_t_2)) {

    /* "pyreadstat/_readstat_parser.pyx":379
 *     var_count = readstat_get_var_count(metadata)
 *     if var_count<0:
 *         raise PyreadstatError("Failed to read number of variables")             # <<<<<<<<<<<<<<
 *     obs_count = readstat_get_row_count(metadata)
 *     if obs_count <0:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_PyreadstatError); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 379, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = NULL;
    __pyx_t_5 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_5 = 1;
      }
    }
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_kp_s_Failed_to_read_number_of_variabl};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 379, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 379, __pyx_L1_error)

    /* "pyreadstat/_readstat_parser.pyx":378
 * 
 *     var_count = readstat_get_var_count(metadata)
 *     if var_count<0:             # <<<<<<<<<<<<<<
 *         raise PyreadstatError("Failed to read number of variables")
 *     obs_count = readstat_get_row_count(metadata)
 */
  }

  /* "pyreadstat/_readstat_parser.pyx":380
 *     if var_count<0:
 *         raise PyreadstatError("Failed to read number of variables")
 *     obs_count = readstat_get_row_count(metadata)             # <<<<<<<<<<<<<<
 *     if obs_count <0:
 *         # if <0 it means the number of rows is not known, allocate 100 000
 */
  __pyx_v_obs_count = readstat_get_row_count(__pyx_v_metadata);

  /* "pyreadstat/_readstat_parser.pyx":381
 *         raise PyreadstatError("Failed to read number of variables")
 *     obs_count = readstat_get_row_count(metadata)
 *     if obs_count <0:             # <<<<<<<<<<<<<<
 *         # if <0 it means the number of rows is not known, allocate 100 000
 *         obs_count = 100000
 */
  __pyx_t_2 = (__pyx_v_obs_count < 0);
  if (__pyx_t_2) {

    /* "pyreadstat/_readstat_parser.pyx":383
 *     if obs_count <0:
 *         # if <0 it means the number of rows is not known, allocate 100 000
 *         obs_count = 100000             # <<<<<<<<<<<<<<
 *         dc.is_unkown_number_rows = 1
 * 
 */
    __pyx_v_obs_count = 0x186A0;

    /* "pyreadstat/_readstat_parser.pyx":384
 *         # if <0 it means the number of rows is not known, allocate 100 000
 *         obs_count = 100000
 *         dc.is_unkown_number_rows = 1             # <<<<<<<<<<<<<<
 * 
 *     dc.n_obs = obs_count
 */
    __pyx_v_dc->is_unkown_number_rows = 1;

    /* "pyreadstat/_readstat_parser.pyx":381
 *         raise PyreadstatError("Failed to read number of variables")
 *     obs_count = readstat_get_row_count(metadata)
 *     if obs_count <0:             # <<<<<<<<<<<<<<
 *         # if <0 it means the number of rows is not known, allocate 100 000
 *         obs_count = 100000
 */
  }

  /* "pyreadstat/_readstat_parser.pyx":386
 *         dc.is_unkown_number_rows = 1
 * 
 *     dc.n_obs = obs_count             # <<<<<<<<<<<<<<
 *     dc.n_vars = var_count
 * 
 */
  __pyx_v_dc->n_obs = __pyx_v_obs_count;

  /* "pyreadstat/_readstat_parser.pyx":387
 * 
 *     dc.n_obs = obs_count
 *     dc.n_vars = var_count             # <<<<<<<<<<<<<<
 * 
 *     dc.col_data_len = [obs_count] * var_count
 */
  __pyx_v_dc->n_vars = __pyx_v_var_count;

  /* "pyreadstat/_readstat_parser.pyx":389
 *     dc.n_vars = var_count
 * 
 *     dc.col_data_len = [obs_count] * var_count             # <<<<<<<<<<<<<<
 *     dc.col_numpy_dtypes = [None] * var_count
 *     dc.col_dytpes_isfloat = [0] * var_count
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_obs_count); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 389, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyList_New(1 * ((__pyx_v_var_count<0) ? 0:__pyx_v_var_count)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 389, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  { Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < __pyx_v_var_count; __pyx_temp++) {
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_GIVEREF(__pyx_t_1);
      PyList_SET_ITEM(__pyx_t_3, __pyx_temp, __pyx_t_1);
    }
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_GIVEREF(__pyx_t_3);
  __Pyx_GOTREF(__pyx_v_dc->col_data_len);
  __Pyx_DECREF(__pyx_v_dc->col_data_len);
  __pyx_v_dc->col_data_len = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "pyreadstat/_readstat_parser.pyx":390
 * 
 *     dc.col_data_len = [obs_count] * var_count
 *     dc.col_numpy_dtypes = [None] * var_count             # <<<<<<<<<<<<<<
 *     dc.col_dytpes_isfloat = [0] * var_count
 *     dc.col_dtypes_isobject = [0] * var_count
 */
  __pyx_t_3 = PyList_New(1 * ((__pyx_v_var_count<0) ? 0:__pyx_v_var_count)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 390, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  { Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < __pyx_v_var_count; __pyx_temp++) {
      __Pyx_INCREF(Py_None);
      __Pyx_GIVEREF(Py_None);
      PyList_SET_ITEM(__pyx_t_3, __pyx_temp, Py_None);
    }
  }
  __Pyx_GIVEREF(__pyx_t_3);
  __Pyx_GOTREF(__pyx_v_dc->col_numpy_dtypes);
  __Pyx_DECREF(__pyx_v_dc->col_numpy_dtypes);
  __pyx_v_dc->col_numpy_dtypes = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "pyreadstat/_readstat_parser.pyx":391
 *     dc.col_data_len = [obs_count] * var_count
 *     dc.col_numpy_dtypes = [None] * var_count
 *     dc.col_dytpes_isfloat = [0] * var_count             # <<<<<<<<<<<<<<
 *     dc.col_dtypes_isobject = [0] * var_count
 * 
 */
  __pyx_t_3 = PyList_New(1 * ((__pyx_v_var_count<0) ? 0:__pyx_v_var_count)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 391, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  { Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < __pyx_v_var_count; __pyx_temp++) {
      __Pyx_INCREF(__pyx_int_0);
      __Pyx_GIVEREF(__pyx_int_0);
      PyList_SET_ITEM(__pyx_t_3, __pyx_temp, __pyx_int_0);
    }
  }
  __Pyx_GIVEREF(__pyx_t_3);
  __Pyx_GOTREF(__pyx_v_dc->col_dytpes_isfloat);
  __Pyx_DECREF(__pyx_v_dc->col_dytpes_isfloat);
  __pyx_v_dc->col_dytpes_isfloat = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "pyreadstat/_readstat_parser.pyx":392
 *     dc.col_numpy_dtypes = [None] * var_count
 *     dc.col_dytpes_isfloat = [0] * var_count
 *     dc.col_dtypes_isobject = [0] * var_count             # <<<<<<<<<<<<<<
 * 
 *     # read other metadata
 */
  __pyx_t_3 = PyList_New(1 * ((__pyx_v_var_count<0) ? 0:__pyx_v_var_count)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 392, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  { Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < __pyx_v_var_count; __pyx_temp++) {
      __Pyx_INCREF(__pyx_int_0);
      __Pyx_GIVEREF(__pyx_int_0);
      PyList_SET_ITEM(__pyx_t_3, __pyx_temp, __pyx_int_0);
    }
  }
  __Pyx_GIVEREF(__pyx_t_3);
  __Pyx_GOTREF(__pyx_v_dc->col_dtypes_isobject);
  __Pyx_DECREF(__pyx_v_dc->col_dtypes_isobject);
  __pyx_v_dc->col_dtypes_isobject = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "pyreadstat/_readstat_parser.pyx":395
 * 
 *     # read other metadata
 *     flabel_orig = readstat_get_file_label(metadata);             # <<<<<<<<<<<<<<
 *     fencoding_orig = readstat_get_file_encoding(metadata)
 *     if flabel_orig != NULL and flabel_orig[0]:
 */
  __pyx_v_flabel_orig = readstat_get_file_label(__pyx_v_metadata);

  /* "pyreadstat/_readstat_parser.pyx":396
 *     # read other metadata
 *     flabel_orig = readstat_get_file_label(metadata);
 *     fencoding_orig = readstat_get_file_encoding(metadata)             # <<<<<<<<<<<<<<
 *     if flabel_orig != NULL and flabel_orig[0]:
 *         flabel = <str> flabel_orig
 */
  __pyx_v_fencoding_orig = readstat_get_file_encoding(__pyx_v_metadata);

  /* "pyreadstat/_readstat_parser.pyx":397
 *     flabel_orig = readstat_get_file_label(metadata);
 *     fencoding_orig = readstat_get_file_encoding(metadata)
 *     if flabel_orig != NULL and flabel_orig[0]:             # <<<<<<<<<<<<<<
 *         flabel = <str> flabel_orig
 *     else:
 */
  __pyx_t_6 = (__pyx_v_flabel_orig != NULL);
  if (__pyx_t_6) {
  } else {
    __pyx_t_2 = __pyx_t_6;
    goto __pyx_L6_bool_binop_done;
  }
  __pyx_t_6 = ((__pyx_v_flabel_orig[0]) != 0);
  __pyx_t_2 = __pyx_t_6;
  __pyx_L6_bool_binop_done:;
  if (__pyx_t_2) {

    /* "pyreadstat/_readstat_parser.pyx":398
 *     fencoding_orig = readstat_get_file_encoding(metadata)
 *     if flabel_orig != NULL and flabel_orig[0]:
 *         flabel = <str> flabel_orig             # <<<<<<<<<<<<<<
 *     else:
 *         flabel = None
 */
    __pyx_t_3 = __Pyx_PyStr_FromString(__pyx_v_flabel_orig); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 398, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __pyx_t_3;
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_flabel = ((PyObject*)__pyx_t_1);
    __pyx_t_1 = 0;

    /* "pyreadstat/_readstat_parser.pyx":397
 *     flabel_orig = readstat_get_file_label(metadata);
 *     fencoding_orig = readstat_get_file_encoding(metadata)
 *     if flabel_orig != NULL and flabel_orig[0]:             # <<<<<<<<<<<<<<
 *         flabel = <str> flabel_orig
 *     else:
 */
    goto __pyx_L5;
  }

  /* "pyreadstat/_readstat_parser.pyx":400
 *         flabel = <str> flabel_orig
 *     else:
 *         flabel = None             # <<<<<<<<<<<<<<
 *     if fencoding_orig != NULL and fencoding_orig[0]:
 *         fencoding = <str> fencoding_orig
 */
  /*else*/ {
    __Pyx_INCREF(Py_None);
    __pyx_v_flabel = ((PyObject*)Py_None);
  }
  __pyx_L5:;

  /* "pyreadstat/_readstat_parser.pyx":401
 *     else:
 *         flabel = None
 *     if fencoding_orig != NULL and fencoding_orig[0]:             # <<<<<<<<<<<<<<
 *         fencoding = <str> fencoding_orig
 *     else:
 */
  __pyx_t_6 = (__pyx_v_fencoding_orig != NULL);
  if (__pyx_t_6) {
  } else {
    __pyx_t_2 = __pyx_t_6;
    goto __pyx_L9_bool_binop_done;
  }
  __pyx_t_6 = ((__pyx_v_fencoding_orig[0]) != 0);
  __pyx_t_2 = __pyx_t_6;
  __pyx_L9_bool_binop_done:;
  if (__pyx_t_2) {

    /* "pyreadstat/_readstat_parser.pyx":402
 *         flabel = None
 *     if fencoding_orig != NULL and fencoding_orig[0]:
 *         fencoding = <str> fencoding_orig             # <<<<<<<<<<<<<<
 *     else:
 *         fencoding = None
 */
    __pyx_t_1 = __Pyx_PyStr_FromString(__pyx_v_fencoding_orig); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 402, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __pyx_t_1;
    __Pyx_INCREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_fencoding = ((PyObject*)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "pyreadstat/_readstat_parser.pyx":401
 *     else:
 *         flabel = None
 *     if fencoding_orig != NULL and fencoding_orig[0]:             # <<<<<<<<<<<<<<
 *         fencoding = <str> fencoding_orig
 *     else:
 */
    goto __pyx_L8;
  }

  /* "pyreadstat/_readstat_parser.pyx":404
 *         fencoding = <str> fencoding_orig
 *     else:
 *         fencoding = None             # <<<<<<<<<<<<<<
 * 
 *     dc.file_encoding = fencoding
 */
  /*else*/ {
    __Pyx_INCREF(Py_None);
    __pyx_v_fencoding = ((PyObject*)Py_None);
  }
  __pyx_L8:;

  /* "pyreadstat/_readstat_parser.pyx":406
 *         fencoding = None
 * 
 *     dc.file_encoding = fencoding             # <<<<<<<<<<<<<<
 *     dc.file_label = flabel
 * 
 */
  __Pyx_INCREF(__pyx_v_fencoding);
  __Pyx_GIVEREF(__pyx_v_fencoding);
  __Pyx_GOTREF(__pyx_v_dc->file_encoding);
  __Pyx_DECREF(__pyx_v_dc->file_encoding);
  __pyx_v_dc->file_encoding = __pyx_v_fencoding;

  /* "pyreadstat/_readstat_parser.pyx":407
 * 
 *     dc.file_encoding = fencoding
 *     dc.file_label = flabel             # <<<<<<<<<<<<<<
 * 
 *     table = readstat_get_table_name(metadata)
 */
  __Pyx_INCREF(__pyx_v_flabel);
  __Pyx_GIVEREF(__pyx_v_flabel);
  __Pyx_GOTREF(__pyx_v_dc->file_label);
  __Pyx_DECREF(__pyx_v_dc->file_label);
  __pyx_v_dc->file_label = __pyx_v_flabel;

  /* "pyreadstat/_readstat_parser.pyx":409
 *     dc.file_label = flabel
 * 
 *     table = readstat_get_table_name(metadata)             # <<<<<<<<<<<<<<
 *     if table != NULL and table[0]:
 *         dc.table_name = <str> table
 */
  __pyx_v_table = readstat_get_table_name(__pyx_v_metadata);

  /* "pyreadstat/_readstat_parser.pyx":410
 * 
 *     table = readstat_get_table_name(metadata)
 *     if table != NULL and table[0]:             # <<<<<<<<<<<<<<
 *         dc.table_name = <str> table
 * 
 */
  __pyx_t_6 = (__pyx_v_table != NULL);
  if (__pyx_t_6) {
  } else {
    __pyx_t_2 = __pyx_t_6;
    goto __pyx_L12_bool_binop_done;
  }
  __pyx_t_6 = ((__pyx_v_table[0]) != 0);
  __pyx_t_2 = __pyx_t_6;
  __pyx_L12_bool_binop_done:;
  if (__pyx_t_2) {

    /* "pyreadstat/_readstat_parser.pyx":411
 *     table = readstat_get_table_name(metadata)
 *     if table != NULL and table[0]:
 *         dc.table_name = <str> table             # <<<<<<<<<<<<<<
 * 
 *     dc.ctime = readstat_get_creation_time(metadata)
 */
    __pyx_t_3 = __Pyx_PyStr_FromString(__pyx_v_table); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 411, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __pyx_t_3;
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GIVEREF(__pyx_t_1);
    __Pyx_GOTREF(__pyx_v_dc->table_name);
    __Pyx_DECREF(__pyx_v_dc->table_name);
    __pyx_v_dc->table_name = ((PyObject*)__pyx_t_1);
    __pyx_t_1 = 0;

    /* "pyreadstat/_readstat_parser.pyx":410
 * 
 *     table = readstat_get_table_name(metadata)
 *     if table != NULL and table[0]:             # <<<<<<<<<<<<<<
 *         dc.table_name = <str> table
 * 
 */
  }

  /* "pyreadstat/_readstat_parser.pyx":413
 *         dc.table_name = <str> table
 * 
 *     dc.ctime = readstat_get_creation_time(metadata)             # <<<<<<<<<<<<<<
 *     dc.mtime = readstat_get_modified_time(metadata)
 * 
 */
  __pyx_v_dc->ctime = readstat_get_creation_time(__pyx_v_metadata);

  /* "pyreadstat/_readstat_parser.pyx":414
 * 
 *     dc.ctime = readstat_get_creation_time(metadata)
 *     dc.mtime = readstat_get_modified_time(metadata)             # <<<<<<<<<<<<<<
 * 
 *     return READSTAT_HANDLER_OK
 */
  __pyx_v_dc->mtime = readstat_get_modified_time(__pyx_v_metadata);

  /* "pyreadstat/_readstat_parser.pyx":416
 *     dc.mtime = readstat_get_modified_time(metadata)
 * 
 *     return READSTAT_HANDLER_OK             # <<<<<<<<<<<<<<
 * 
 * cdef int handle_variable(int index, readstat_variable_t *variable,
 */
  __pyx_r = READSTAT_HANDLER_OK;
  goto __pyx_L0;

  /* "pyreadstat/_readstat_parser.pyx":356
 * 
 * 
 * cdef int handle_metadata(readstat_metadata_t *metadata, void *ctx) except READSTAT_HANDLER_ABORT:             # <<<<<<<<<<<<<<
 *     """
 *     This function sets the number of observations(rows), number of variables
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pyreadstat._readstat_parser.handle_metadata", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = READSTAT_HANDLER_ABORT;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_dc);
  __Pyx_XDECREF(__pyx_v_flabel);
  __Pyx_XDECREF(__pyx_v_fencoding);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyreadstat/_readstat_parser.pyx":418
 *     return READSTAT_HANDLER_OK
 * 
 * cdef int handle_variable(int index, readstat_variable_t *variable,             # <<<<<<<<<<<<<<
 *                          char *val_labels, void *ctx) except READSTAT_HANDLER_ABORT:
 *     """
 */

static int __pyx_f_10pyreadstat_16_readstat_parser_handle_variable(int __pyx_v_index, readstat_variable_t *__pyx_v_variable, char *__pyx_v_val_labels, void *__pyx_v_ctx) {
  char *__pyx_v_var_name;
  char *__pyx_v_var_label;
  char *__pyx_v_var_format;
  PyObject *__pyx_v_col_name = 0;
  PyObject *__pyx_v_col_label = 0;
  PyObject *__pyx_v_label_name = 0;
  PyObject *__pyx_v_col_format_original = 0;
  __pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format __pyx_v_col_format_final;
  readstat_type_t __pyx_v_var_type;
  __pyx_t_10pyreadstat_16_readstat_parser_py_file_format __pyx_v_file_format;
  int __pyx_v_n_ranges;
  readstat_value_t __pyx_v_loval;
  readstat_value_t __pyx_v_hival;
  PyObject *__pyx_v_pyloval = 0;
  PyObject *__pyx_v_pyhival = 0;
  PyObject *__pyx_v_missing_ranges = 0;
  PyObject *__pyx_v_curnptype = 0;
  PyObject *__pyx_v_row = 0;
  int __pyx_v_metaonly;
  int __pyx_v_obs_count;
  int __pyx_v_iscurnptypefloat;
  int __pyx_v_iscurnptypeobject;
  PyObject *__pyx_v_newcolname = 0;
  int __pyx_v_dupcolcnt;
  struct __pyx_obj_10pyreadstat_16_readstat_parser_data_container *__pyx_v_dc = 0;
  PyObject *__pyx_v_msg = NULL;
  long __pyx_v_i;
  size_t __pyx_v_storage_width;
  readstat_alignment_t __pyx_v_align;
  PyObject *__pyx_v_pyalign = 0;
  readstat_measure_t __pyx_v_measure;
  PyObject *__pyx_v_pymeasure = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_t_7;
  __pyx_t_10pyreadstat_16_readstat_parser_py_file_format __pyx_t_8;
  __pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  long __pyx_t_12;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("handle_variable", 0);

  /* "pyreadstat/_readstat_parser.pyx":447
 *     cdef int dupcolcnt
 * 
 *     cdef  data_container dc = <data_container> ctx             # <<<<<<<<<<<<<<
 * 
 *     # get variable name, label, format and type and put into our data container
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_ctx);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_dc = ((struct __pyx_obj_10pyreadstat_16_readstat_parser_data_container *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyreadstat/_readstat_parser.pyx":450
 * 
 *     # get variable name, label, format and type and put into our data container
 *     var_name = readstat_variable_get_name(variable)             # <<<<<<<<<<<<<<
 *     if var_name == NULL:
 *         col_name = None
 */
  __pyx_v_var_name = readstat_variable_get_name(__pyx_v_variable);

  /* "pyreadstat/_readstat_parser.pyx":451
 *     # get variable name, label, format and type and put into our data container
 *     var_name = readstat_variable_get_name(variable)
 *     if var_name == NULL:             # <<<<<<<<<<<<<<
 *         col_name = None
 *     else:
 */
  __pyx_t_2 = (__pyx_v_var_name == NULL);
  if (__pyx_t_2) {

    /* "pyreadstat/_readstat_parser.pyx":452
 *     var_name = readstat_variable_get_name(variable)
 *     if var_name == NULL:
 *         col_name = None             # <<<<<<<<<<<<<<
 *     else:
 *         col_name = <str>var_name
 */
    __Pyx_INCREF(Py_None);
    __pyx_v_col_name = ((PyObject*)Py_None);

    /* "pyreadstat/_readstat_parser.pyx":451
 *     # get variable name, label, format and type and put into our data container
 *     var_name = readstat_variable_get_name(variable)
 *     if var_name == NULL:             # <<<<<<<<<<<<<<
 *         col_name = None
 *     else:
 */
    goto __pyx_L3;
  }

  /* "pyreadstat/_readstat_parser.pyx":454
 *         col_name = None
 *     else:
 *         col_name = <str>var_name             # <<<<<<<<<<<<<<
 * 
 *     # if the user introduced a list of columns to include, continue only if the column is in the list
 */
  /*else*/ {
    __pyx_t_1 = __Pyx_PyStr_FromString(__pyx_v_var_name); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 454, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __pyx_t_1;
    __Pyx_INCREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_col_name = ((PyObject*)__pyx_t_3);
    __pyx_t_3 = 0;
  }
  __pyx_L3:;

  /* "pyreadstat/_readstat_parser.pyx":457
 * 
 *     # if the user introduced a list of columns to include, continue only if the column is in the list
 *     if dc.filter_cols and not (col_name in dc.use_cols):             # <<<<<<<<<<<<<<
 *         dc.n_vars -= 1
 *         return READSTAT_HANDLER_SKIP_VARIABLE
 */
  if (__pyx_v_dc->filter_cols) {
  } else {
    __pyx_t_2 = __pyx_v_dc->filter_cols;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_4 = (__Pyx_PySequence_ContainsTF(__pyx_v_col_name, __pyx_v_dc->use_cols, Py_NE)); if (unlikely((__pyx_t_4 < 0))) __PYX_ERR(0, 457, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_4;
  __pyx_L5_bool_binop_done:;
  if (__pyx_t_2) {

    /* "pyreadstat/_readstat_parser.pyx":458
 *     # if the user introduced a list of columns to include, continue only if the column is in the list
 *     if dc.filter_cols and not (col_name in dc.use_cols):
 *         dc.n_vars -= 1             # <<<<<<<<<<<<<<
 *         return READSTAT_HANDLER_SKIP_VARIABLE
 * 
 */
    __pyx_v_dc->n_vars = (__pyx_v_dc->n_vars - 1);

    /* "pyreadstat/_readstat_parser.pyx":459
 *     if dc.filter_cols and not (col_name in dc.use_cols):
 *         dc.n_vars -= 1
 *         return READSTAT_HANDLER_SKIP_VARIABLE             # <<<<<<<<<<<<<<
 * 
 *     index = readstat_variable_get_index_after_skipping(variable)
 */
    __pyx_r = READSTAT_HANDLER_SKIP_VARIABLE;
    goto __pyx_L0;

    /* "pyreadstat/_readstat_parser.pyx":457
 * 
 *     # if the user introduced a list of columns to include, continue only if the column is in the list
 *     if dc.filter_cols and not (col_name in dc.use_cols):             # <<<<<<<<<<<<<<
 *         dc.n_vars -= 1
 *         return READSTAT_HANDLER_SKIP_VARIABLE
 */
  }

  /* "pyreadstat/_readstat_parser.pyx":461
 *         return READSTAT_HANDLER_SKIP_VARIABLE
 * 
 *     index = readstat_variable_get_index_after_skipping(variable)             # <<<<<<<<<<<<<<
 * 
 *     if col_name in dc.col_names:
 */
  __pyx_v_index = readstat_variable_get_index_after_skipping(__pyx_v_variable);

  /* "pyreadstat/_readstat_parser.pyx":463
 *     index = readstat_variable_get_index_after_skipping(variable)
 * 
 *     if col_name in dc.col_names:             # <<<<<<<<<<<<<<
 *         dupcolcnt = 1
 *         while True:
 */
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_v_col_name, __pyx_v_dc->col_names, Py_EQ)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 463, __pyx_L1_error)
  if (__pyx_t_2) {

    /* "pyreadstat/_readstat_parser.pyx":464
 * 
 *     if col_name in dc.col_names:
 *         dupcolcnt = 1             # <<<<<<<<<<<<<<
 *         while True:
 *             newcolname = col_name + "_duplicated" + str(dupcolcnt)
 */
    __pyx_v_dupcolcnt = 1;

    /* "pyreadstat/_readstat_parser.pyx":465
 *     if col_name in dc.col_names:
 *         dupcolcnt = 1
 *         while True:             # <<<<<<<<<<<<<<
 *             newcolname = col_name + "_duplicated" + str(dupcolcnt)
 *             if newcolname in col_name:
 */
    while (1) {

      /* "pyreadstat/_readstat_parser.pyx":466
 *         dupcolcnt = 1
 *         while True:
 *             newcolname = col_name + "_duplicated" + str(dupcolcnt)             # <<<<<<<<<<<<<<
 *             if newcolname in col_name:
 *                 dupcolcnt += 1
 */
      __pyx_t_3 = __Pyx_PyStr_ConcatSafe(__pyx_v_col_name, __pyx_n_s_duplicated); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 466, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_dupcolcnt); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 466, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_5 = __Pyx_PyObject_Str(__pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 466, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = PyNumber_Add(__pyx_t_3, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 466, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (!(likely(PyString_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("str", __pyx_t_1))) __PYX_ERR(0, 466, __pyx_L1_error)
      __Pyx_XDECREF_SET(__pyx_v_newcolname, ((PyObject*)__pyx_t_1));
      __pyx_t_1 = 0;

      /* "pyreadstat/_readstat_parser.pyx":467
 *         while True:
 *             newcolname = col_name + "_duplicated" + str(dupcolcnt)
 *             if newcolname in col_name:             # <<<<<<<<<<<<<<
 *                 dupcolcnt += 1
 *                 continue
 */
      __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_v_newcolname, __pyx_v_col_name, Py_EQ)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 467, __pyx_L1_error)
      if (__pyx_t_2) {

        /* "pyreadstat/_readstat_parser.pyx":468
 *             newcolname = col_name + "_duplicated" + str(dupcolcnt)
 *             if newcolname in col_name:
 *                 dupcolcnt += 1             # <<<<<<<<<<<<<<
 *                 continue
 *             else:
 */
        __pyx_v_dupcolcnt = (__pyx_v_dupcolcnt + 1);

        /* "pyreadstat/_readstat_parser.pyx":469
 *             if newcolname in col_name:
 *                 dupcolcnt += 1
 *                 continue             # <<<<<<<<<<<<<<
 *             else:
 *                 msg = "column '{0}' is duplicated, renamed to '{1}'".format(col_name, newcolname)
 */
        goto __pyx_L8_continue;

        /* "pyreadstat/_readstat_parser.pyx":467
 *         while True:
 *             newcolname = col_name + "_duplicated" + str(dupcolcnt)
 *             if newcolname in col_name:             # <<<<<<<<<<<<<<
 *                 dupcolcnt += 1
 *                 continue
 */
      }

      /* "pyreadstat/_readstat_parser.pyx":471
 *                 continue
 *             else:
 *                 msg = "column '{0}' is duplicated, renamed to '{1}'".format(col_name, newcolname)             # <<<<<<<<<<<<<<
 *                 warnings.warn(msg)
 *                 col_name = newcolname
 */
      /*else*/ {
        __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_kp_s_column_0_is_duplicated_renamed_t, __pyx_n_s_format); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 471, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_3 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_5);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_5, function);
            __pyx_t_6 = 1;
          }
        }
        {
          PyObject *__pyx_callargs[3] = {__pyx_t_3, __pyx_v_col_name, __pyx_v_newcolname};
          __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_6, 2+__pyx_t_6);
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 471, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __pyx_v_msg = __pyx_t_1;
        __pyx_t_1 = 0;

        /* "pyreadstat/_readstat_parser.pyx":472
 *             else:
 *                 msg = "column '{0}' is duplicated, renamed to '{1}'".format(col_name, newcolname)
 *                 warnings.warn(msg)             # <<<<<<<<<<<<<<
 *                 col_name = newcolname
 *                 break
 */
        __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_warnings); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 472, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_warn); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 472, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __pyx_t_5 = NULL;
        __pyx_t_6 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_6 = 1;
          }
        }
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_v_msg};
          __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 472, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pyreadstat/_readstat_parser.pyx":473
 *                 msg = "column '{0}' is duplicated, renamed to '{1}'".format(col_name, newcolname)
 *                 warnings.warn(msg)
 *                 col_name = newcolname             # <<<<<<<<<<<<<<
 *                 break
 * 
 */
        __Pyx_INCREF(__pyx_v_newcolname);
        __Pyx_DECREF_SET(__pyx_v_col_name, __pyx_v_newcolname);

        /* "pyreadstat/_readstat_parser.pyx":474
 *                 warnings.warn(msg)
 *                 col_name = newcolname
 *                 break             # <<<<<<<<<<<<<<
 * 
 *     dc.col_names.append(col_name)
 */
        goto __pyx_L9_break;
      }
      __pyx_L8_continue:;
    }
    __pyx_L9_break:;

    /* "pyreadstat/_readstat_parser.pyx":463
 *     index = readstat_variable_get_index_after_skipping(variable)
 * 
 *     if col_name in dc.col_names:             # <<<<<<<<<<<<<<
 *         dupcolcnt = 1
 *         while True:
 */
  }

  /* "pyreadstat/_readstat_parser.pyx":476
 *                 break
 * 
 *     dc.col_names.append(col_name)             # <<<<<<<<<<<<<<
 * 
 *     # the name of the value label for the variable
 */
  if (unlikely(__pyx_v_dc->col_names == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "append");
    __PYX_ERR(0, 476, __pyx_L1_error)
  }
  __pyx_t_7 = __Pyx_PyList_Append(__pyx_v_dc->col_names, __pyx_v_col_name); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(0, 476, __pyx_L1_error)

  /* "pyreadstat/_readstat_parser.pyx":479
 * 
 *     # the name of the value label for the variable
 *     if val_labels != NULL:             # <<<<<<<<<<<<<<
 *         label_name = <str> val_labels
 *         if label_name:
 */
  __pyx_t_2 = (__pyx_v_val_labels != NULL);
  if (__pyx_t_2) {

    /* "pyreadstat/_readstat_parser.pyx":480
 *     # the name of the value label for the variable
 *     if val_labels != NULL:
 *         label_name = <str> val_labels             # <<<<<<<<<<<<<<
 *         if label_name:
 *             dc.label_to_var_name[col_name] = label_name
 */
    __pyx_t_1 = __Pyx_PyStr_FromString(__pyx_v_val_labels); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 480, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __pyx_t_1;
    __Pyx_INCREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_label_name = ((PyObject*)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "pyreadstat/_readstat_parser.pyx":481
 *     if val_labels != NULL:
 *         label_name = <str> val_labels
 *         if label_name:             # <<<<<<<<<<<<<<
 *             dc.label_to_var_name[col_name] = label_name
 * 
 */
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_label_name); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 481, __pyx_L1_error)
    if (__pyx_t_2) {

      /* "pyreadstat/_readstat_parser.pyx":482
 *         label_name = <str> val_labels
 *         if label_name:
 *             dc.label_to_var_name[col_name] = label_name             # <<<<<<<<<<<<<<
 * 
 *     var_label = readstat_variable_get_label(variable)
 */
      if (unlikely((PyObject_SetItem(__pyx_v_dc->label_to_var_name, __pyx_v_col_name, __pyx_v_label_name) < 0))) __PYX_ERR(0, 482, __pyx_L1_error)

      /* "pyreadstat/_readstat_parser.pyx":481
 *     if val_labels != NULL:
 *         label_name = <str> val_labels
 *         if label_name:             # <<<<<<<<<<<<<<
 *             dc.label_to_var_name[col_name] = label_name
 * 
 */
    }

    /* "pyreadstat/_readstat_parser.pyx":479
 * 
 *     # the name of the value label for the variable
 *     if val_labels != NULL:             # <<<<<<<<<<<<<<
 *         label_name = <str> val_labels
 *         if label_name:
 */
  }

  /* "pyreadstat/_readstat_parser.pyx":484
 *             dc.label_to_var_name[col_name] = label_name
 * 
 *     var_label = readstat_variable_get_label(variable)             # <<<<<<<<<<<<<<
 *     if var_label == NULL:
 *         col_label = None
 */
  __pyx_v_var_label = readstat_variable_get_label(__pyx_v_variable);

  /* "pyreadstat/_readstat_parser.pyx":485
 * 
 *     var_label = readstat_variable_get_label(variable)
 *     if var_label == NULL:             # <<<<<<<<<<<<<<
 *         col_label = None
 *     else:
 */
  __pyx_t_2 = (__pyx_v_var_label == NULL);
  if (__pyx_t_2) {

    /* "pyreadstat/_readstat_parser.pyx":486
 *     var_label = readstat_variable_get_label(variable)
 *     if var_label == NULL:
 *         col_label = None             # <<<<<<<<<<<<<<
 *     else:
 *         col_label = <str>var_label
 */
    __Pyx_INCREF(Py_None);
    __pyx_v_col_label = ((PyObject*)Py_None);

    /* "pyreadstat/_readstat_parser.pyx":485
 * 
 *     var_label = readstat_variable_get_label(variable)
 *     if var_label == NULL:             # <<<<<<<<<<<<<<
 *         col_label = None
 *     else:
 */
    goto __pyx_L13;
  }

  /* "pyreadstat/_readstat_parser.pyx":488
 *         col_label = None
 *     else:
 *         col_label = <str>var_label             # <<<<<<<<<<<<<<
 *     dc.col_labels.append(col_label)
 * 
 */
  /*else*/ {
    __pyx_t_3 = __Pyx_PyStr_FromString(__pyx_v_var_label); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 488, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __pyx_t_3;
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_col_label = ((PyObject*)__pyx_t_1);
    __pyx_t_1 = 0;
  }
  __pyx_L13:;

  /* "pyreadstat/_readstat_parser.pyx":489
 *     else:
 *         col_label = <str>var_label
 *     dc.col_labels.append(col_label)             # <<<<<<<<<<<<<<
 * 
 *     # format, we have to transform it in something more usable
 */
  if (unlikely(__pyx_v_dc->col_labels == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "append");
    __PYX_ERR(0, 489, __pyx_L1_error)
  }
  __pyx_t_7 = __Pyx_PyList_Append(__pyx_v_dc->col_labels, __pyx_v_col_label); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(0, 489, __pyx_L1_error)

  /* "pyreadstat/_readstat_parser.pyx":492
 * 
 *     # format, we have to transform it in something more usable
 *     var_format = readstat_variable_get_format(variable)             # <<<<<<<<<<<<<<
 *     if var_format == NULL:
 *         col_format_original = "NULL"
 */
  __pyx_v_var_format = readstat_variable_get_format(__pyx_v_variable);

  /* "pyreadstat/_readstat_parser.pyx":493
 *     # format, we have to transform it in something more usable
 *     var_format = readstat_variable_get_format(variable)
 *     if var_format == NULL:             # <<<<<<<<<<<<<<
 *         col_format_original = "NULL"
 *     else:
 */
  __pyx_t_2 = (__pyx_v_var_format == NULL);
  if (__pyx_t_2) {

    /* "pyreadstat/_readstat_parser.pyx":494
 *     var_format = readstat_variable_get_format(variable)
 *     if var_format == NULL:
 *         col_format_original = "NULL"             # <<<<<<<<<<<<<<
 *     else:
 *         col_format_original = <str>var_format
 */
    __Pyx_INCREF(__pyx_n_s_NULL);
    __pyx_v_col_format_original = __pyx_n_s_NULL;

    /* "pyreadstat/_readstat_parser.pyx":493
 *     # format, we have to transform it in something more usable
 *     var_format = readstat_variable_get_format(variable)
 *     if var_format == NULL:             # <<<<<<<<<<<<<<
 *         col_format_original = "NULL"
 *     else:
 */
    goto __pyx_L14;
  }

  /* "pyreadstat/_readstat_parser.pyx":496
 *         col_format_original = "NULL"
 *     else:
 *         col_format_original = <str>var_format             # <<<<<<<<<<<<<<
 *     file_format = dc.file_format
 *     dc.col_formats_original.append(col_format_original)
 */
  /*else*/ {
    __pyx_t_1 = __Pyx_PyStr_FromString(__pyx_v_var_format); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 496, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __pyx_t_1;
    __Pyx_INCREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_col_format_original = ((PyObject*)__pyx_t_3);
    __pyx_t_3 = 0;
  }
  __pyx_L14:;

  /* "pyreadstat/_readstat_parser.pyx":497
 *     else:
 *         col_format_original = <str>var_format
 *     file_format = dc.file_format             # <<<<<<<<<<<<<<
 *     dc.col_formats_original.append(col_format_original)
 *     col_format_final = transform_variable_format(col_format_original, file_format)
 */
  __pyx_t_8 = __pyx_v_dc->file_format;
  __pyx_v_file_format = __pyx_t_8;

  /* "pyreadstat/_readstat_parser.pyx":498
 *         col_format_original = <str>var_format
 *     file_format = dc.file_format
 *     dc.col_formats_original.append(col_format_original)             # <<<<<<<<<<<<<<
 *     col_format_final = transform_variable_format(col_format_original, file_format)
 *     dc.col_formats.append(col_format_final)
 */
  if (unlikely(__pyx_v_dc->col_formats_original == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "append");
    __PYX_ERR(0, 498, __pyx_L1_error)
  }
  __pyx_t_7 = __Pyx_PyList_Append(__pyx_v_dc->col_formats_original, __pyx_v_col_format_original); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(0, 498, __pyx_L1_error)

  /* "pyreadstat/_readstat_parser.pyx":499
 *     file_format = dc.file_format
 *     dc.col_formats_original.append(col_format_original)
 *     col_format_final = transform_variable_format(col_format_original, file_format)             # <<<<<<<<<<<<<<
 *     dc.col_formats.append(col_format_final)
 *     # readstat type
 */
  __pyx_t_9 = __pyx_f_10pyreadstat_16_readstat_parser_transform_variable_format(__pyx_v_col_format_original, __pyx_v_file_format); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 499, __pyx_L1_error)
  __pyx_v_col_format_final = __pyx_t_9;

  /* "pyreadstat/_readstat_parser.pyx":500
 *     dc.col_formats_original.append(col_format_original)
 *     col_format_final = transform_variable_format(col_format_original, file_format)
 *     dc.col_formats.append(col_format_final)             # <<<<<<<<<<<<<<
 *     # readstat type
 *     var_type = readstat_variable_get_type(variable)
 */
  if (unlikely(__pyx_v_dc->col_formats == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "append");
    __PYX_ERR(0, 500, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyInt_From___pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format(__pyx_v_col_format_final); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 500, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = __Pyx_PyList_Append(__pyx_v_dc->col_formats, __pyx_t_3); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(0, 500, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "pyreadstat/_readstat_parser.pyx":502
 *     dc.col_formats.append(col_format_final)
 *     # readstat type
 *     var_type = readstat_variable_get_type(variable)             # <<<<<<<<<<<<<<
 *     dc.col_dtypes.append(var_type)
 *     # equivalent numpy type
 */
  __pyx_v_var_type = readstat_variable_get_type(__pyx_v_variable);

  /* "pyreadstat/_readstat_parser.pyx":503
 *     # readstat type
 *     var_type = readstat_variable_get_type(variable)
 *     dc.col_dtypes.append(var_type)             # <<<<<<<<<<<<<<
 *     # equivalent numpy type
 *     # if it's a date then we need object
 */
  if (unlikely(__pyx_v_dc->col_dtypes == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "append");
    __PYX_ERR(0, 503, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyInt_From_readstat_type_t(__pyx_v_var_type); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 503, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = __Pyx_PyList_Append(__pyx_v_dc->col_dtypes, __pyx_t_3); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(0, 503, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "pyreadstat/_readstat_parser.pyx":506
 *     # equivalent numpy type
 *     # if it's a date then we need object
 *     if col_format_final != DATE_FORMAT_NOTADATE and dc.no_datetime_conversion == 0:             # <<<<<<<<<<<<<<
 *         curnptype = object
 *     else:
 */
  __pyx_t_4 = (__pyx_v_col_format_final != __pyx_e_10pyreadstat_16_readstat_parser_DATE_FORMAT_NOTADATE);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L16_bool_binop_done;
  }
  __pyx_t_4 = (__pyx_v_dc->no_datetime_conversion == 0);
  __pyx_t_2 = __pyx_t_4;
  __pyx_L16_bool_binop_done:;
  if (__pyx_t_2) {

    /* "pyreadstat/_readstat_parser.pyx":507
 *     # if it's a date then we need object
 *     if col_format_final != DATE_FORMAT_NOTADATE and dc.no_datetime_conversion == 0:
 *         curnptype = object             # <<<<<<<<<<<<<<
 *     else:
 *         curnptype = readstat_to_numpy_types[var_type]
 */
    __Pyx_INCREF(__pyx_builtin_object);
    __pyx_v_curnptype = __pyx_builtin_object;

    /* "pyreadstat/_readstat_parser.pyx":506
 *     # equivalent numpy type
 *     # if it's a date then we need object
 *     if col_format_final != DATE_FORMAT_NOTADATE and dc.no_datetime_conversion == 0:             # <<<<<<<<<<<<<<
 *         curnptype = object
 *     else:
 */
    goto __pyx_L15;
  }

  /* "pyreadstat/_readstat_parser.pyx":509
 *         curnptype = object
 *     else:
 *         curnptype = readstat_to_numpy_types[var_type]             # <<<<<<<<<<<<<<
 *     iscurnptypefloat = 0
 *     iscurnptypeobject = 0
 */
  /*else*/ {
    if (unlikely(__pyx_v_10pyreadstat_16_readstat_parser_readstat_to_numpy_types == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 509, __pyx_L1_error)
    }
    __pyx_t_3 = __Pyx_PyInt_From_readstat_type_t(__pyx_v_var_type); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 509, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyDict_GetItem(__pyx_v_10pyreadstat_16_readstat_parser_readstat_to_numpy_types, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 509, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_curnptype = __pyx_t_1;
    __pyx_t_1 = 0;
  }
  __pyx_L15:;

  /* "pyreadstat/_readstat_parser.pyx":510
 *     else:
 *         curnptype = readstat_to_numpy_types[var_type]
 *     iscurnptypefloat = 0             # <<<<<<<<<<<<<<
 *     iscurnptypeobject = 0
 *     # book keeping numpy types
 */
  __pyx_v_iscurnptypefloat = 0;

  /* "pyreadstat/_readstat_parser.pyx":511
 *         curnptype = readstat_to_numpy_types[var_type]
 *     iscurnptypefloat = 0
 *     iscurnptypeobject = 0             # <<<<<<<<<<<<<<
 *     # book keeping numpy types
 *     dc.col_numpy_dtypes[index] = curnptype
 */
  __pyx_v_iscurnptypeobject = 0;

  /* "pyreadstat/_readstat_parser.pyx":513
 *     iscurnptypeobject = 0
 *     # book keeping numpy types
 *     dc.col_numpy_dtypes[index] = curnptype             # <<<<<<<<<<<<<<
 *     if curnptype == object:
 *         iscurnptypeobject = 1
 */
  if (unlikely(__pyx_v_dc->col_numpy_dtypes == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 513, __pyx_L1_error)
  }
  if (unlikely((__Pyx_SetItemInt(__pyx_v_dc->col_numpy_dtypes, __pyx_v_index, __pyx_v_curnptype, int, 1, __Pyx_PyInt_From_int, 1, 1, 1) < 0))) __PYX_ERR(0, 513, __pyx_L1_error)

  /* "pyreadstat/_readstat_parser.pyx":514
 *     # book keeping numpy types
 *     dc.col_numpy_dtypes[index] = curnptype
 *     if curnptype == object:             # <<<<<<<<<<<<<<
 *         iscurnptypeobject = 1
 *     if curnptype == np.float64:
 */
  __pyx_t_1 = PyObject_RichCompare(__pyx_v_curnptype, __pyx_builtin_object, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 514, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 514, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "pyreadstat/_readstat_parser.pyx":515
 *     dc.col_numpy_dtypes[index] = curnptype
 *     if curnptype == object:
 *         iscurnptypeobject = 1             # <<<<<<<<<<<<<<
 *     if curnptype == np.float64:
 *         iscurnptypefloat = 1
 */
    __pyx_v_iscurnptypeobject = 1;

    /* "pyreadstat/_readstat_parser.pyx":514
 *     # book keeping numpy types
 *     dc.col_numpy_dtypes[index] = curnptype
 *     if curnptype == object:             # <<<<<<<<<<<<<<
 *         iscurnptypeobject = 1
 *     if curnptype == np.float64:
 */
  }

  /* "pyreadstat/_readstat_parser.pyx":516
 *     if curnptype == object:
 *         iscurnptypeobject = 1
 *     if curnptype == np.float64:             # <<<<<<<<<<<<<<
 *         iscurnptypefloat = 1
 *     dc.col_dtypes_isobject[index] = iscurnptypeobject
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 516, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_float64); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 516, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyObject_RichCompare(__pyx_v_curnptype, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 516, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 516, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "pyreadstat/_readstat_parser.pyx":517
 *         iscurnptypeobject = 1
 *     if curnptype == np.float64:
 *         iscurnptypefloat = 1             # <<<<<<<<<<<<<<
 *     dc.col_dtypes_isobject[index] = iscurnptypeobject
 *     dc.col_dytpes_isfloat[index] = iscurnptypefloat
 */
    __pyx_v_iscurnptypefloat = 1;

    /* "pyreadstat/_readstat_parser.pyx":516
 *     if curnptype == object:
 *         iscurnptypeobject = 1
 *     if curnptype == np.float64:             # <<<<<<<<<<<<<<
 *         iscurnptypefloat = 1
 *     dc.col_dtypes_isobject[index] = iscurnptypeobject
 */
  }

  /* "pyreadstat/_readstat_parser.pyx":518
 *     if curnptype == np.float64:
 *         iscurnptypefloat = 1
 *     dc.col_dtypes_isobject[index] = iscurnptypeobject             # <<<<<<<<<<<<<<
 *     dc.col_dytpes_isfloat[index] = iscurnptypefloat
 *     metaonly = dc.metaonly
 */
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_iscurnptypeobject); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 518, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (unlikely(__pyx_v_dc->col_dtypes_isobject == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 518, __pyx_L1_error)
  }
  if (unlikely((__Pyx_SetItemInt(__pyx_v_dc->col_dtypes_isobject, __pyx_v_index, __pyx_t_1, int, 1, __Pyx_PyInt_From_int, 1, 1, 1) < 0))) __PYX_ERR(0, 518, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyreadstat/_readstat_parser.pyx":519
 *         iscurnptypefloat = 1
 *     dc.col_dtypes_isobject[index] = iscurnptypeobject
 *     dc.col_dytpes_isfloat[index] = iscurnptypefloat             # <<<<<<<<<<<<<<
 *     metaonly = dc.metaonly
 *     # pre-allocate data
 */
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_iscurnptypefloat); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 519, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (unlikely(__pyx_v_dc->col_dytpes_isfloat == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 519, __pyx_L1_error)
  }
  if (unlikely((__Pyx_SetItemInt(__pyx_v_dc->col_dytpes_isfloat, __pyx_v_index, __pyx_t_1, int, 1, __Pyx_PyInt_From_int, 1, 1, 1) < 0))) __PYX_ERR(0, 519, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyreadstat/_readstat_parser.pyx":520
 *     dc.col_dtypes_isobject[index] = iscurnptypeobject
 *     dc.col_dytpes_isfloat[index] = iscurnptypefloat
 *     metaonly = dc.metaonly             # <<<<<<<<<<<<<<
 *     # pre-allocate data
 *     if metaonly:
 */
  __pyx_t_2 = __pyx_v_dc->metaonly;
  __pyx_v_metaonly = __pyx_t_2;

  /* "pyreadstat/_readstat_parser.pyx":522
 *     metaonly = dc.metaonly
 *     # pre-allocate data
 *     if metaonly:             # <<<<<<<<<<<<<<
 *         row = np.empty(1, dtype=curnptype)
 *     else:
 */
  if (__pyx_v_metaonly) {

    /* "pyreadstat/_readstat_parser.pyx":523
 *     # pre-allocate data
 *     if metaonly:
 *         row = np.empty(1, dtype=curnptype)             # <<<<<<<<<<<<<<
 *     else:
 *         obs_count = dc.n_obs
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 523, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_empty); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 523, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 523, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_dtype, __pyx_v_curnptype) < 0) __PYX_ERR(0, 523, __pyx_L1_error)
    __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__3, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 523, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_row = __pyx_t_5;
    __pyx_t_5 = 0;

    /* "pyreadstat/_readstat_parser.pyx":522
 *     metaonly = dc.metaonly
 *     # pre-allocate data
 *     if metaonly:             # <<<<<<<<<<<<<<
 *         row = np.empty(1, dtype=curnptype)
 *     else:
 */
    goto __pyx_L20;
  }

  /* "pyreadstat/_readstat_parser.pyx":525
 *         row = np.empty(1, dtype=curnptype)
 *     else:
 *         obs_count = dc.n_obs             # <<<<<<<<<<<<<<
 *         row = np.empty(obs_count, dtype=curnptype)
 *         if iscurnptypeobject or iscurnptypefloat:
 */
  /*else*/ {
    __pyx_t_6 = __pyx_v_dc->n_obs;
    __pyx_v_obs_count = __pyx_t_6;

    /* "pyreadstat/_readstat_parser.pyx":526
 *     else:
 *         obs_count = dc.n_obs
 *         row = np.empty(obs_count, dtype=curnptype)             # <<<<<<<<<<<<<<
 *         if iscurnptypeobject or iscurnptypefloat:
 *             row.fill(np.nan)
 */
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_np); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 526, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_empty); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 526, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_obs_count); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 526, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 526, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5);
    __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 526, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_dtype, __pyx_v_curnptype) < 0) __PYX_ERR(0, 526, __pyx_L1_error)
    __pyx_t_10 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_3, __pyx_t_5); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 526, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_v_row = __pyx_t_10;
    __pyx_t_10 = 0;

    /* "pyreadstat/_readstat_parser.pyx":527
 *         obs_count = dc.n_obs
 *         row = np.empty(obs_count, dtype=curnptype)
 *         if iscurnptypeobject or iscurnptypefloat:             # <<<<<<<<<<<<<<
 *             row.fill(np.nan)
 *     dc.col_data.append(row)
 */
    if (!__pyx_v_iscurnptypeobject) {
    } else {
      __pyx_t_2 = __pyx_v_iscurnptypeobject;
      goto __pyx_L22_bool_binop_done;
    }
    __pyx_t_2 = __pyx_v_iscurnptypefloat;
    __pyx_L22_bool_binop_done:;
    if (__pyx_t_2) {

      /* "pyreadstat/_readstat_parser.pyx":528
 *         row = np.empty(obs_count, dtype=curnptype)
 *         if iscurnptypeobject or iscurnptypefloat:
 *             row.fill(np.nan)             # <<<<<<<<<<<<<<
 *     dc.col_data.append(row)
 * 
 */
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_row, __pyx_n_s_fill); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 528, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 528, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_nan); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 528, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = NULL;
      __pyx_t_6 = 0;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_3)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
          __pyx_t_6 = 1;
        }
      }
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_t_1};
        __pyx_t_10 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 528, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

      /* "pyreadstat/_readstat_parser.pyx":527
 *         obs_count = dc.n_obs
 *         row = np.empty(obs_count, dtype=curnptype)
 *         if iscurnptypeobject or iscurnptypefloat:             # <<<<<<<<<<<<<<
 *             row.fill(np.nan)
 *     dc.col_data.append(row)
 */
    }
  }
  __pyx_L20:;

  /* "pyreadstat/_readstat_parser.pyx":529
 *         if iscurnptypeobject or iscurnptypefloat:
 *             row.fill(np.nan)
 *     dc.col_data.append(row)             # <<<<<<<<<<<<<<
 * 
 *     # missing values
 */
  if (unlikely(__pyx_v_dc->col_data == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "append");
    __PYX_ERR(0, 529, __pyx_L1_error)
  }
  __pyx_t_7 = __Pyx_PyList_Append(__pyx_v_dc->col_data, __pyx_v_row); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(0, 529, __pyx_L1_error)

  /* "pyreadstat/_readstat_parser.pyx":532
 * 
 *     # missing values
 *     if dc.usernan:             # <<<<<<<<<<<<<<
 *         n_ranges = readstat_variable_get_missing_ranges_count(variable)
 *         if n_ranges>0:
 */
  if (__pyx_v_dc->usernan) {

    /* "pyreadstat/_readstat_parser.pyx":533
 *     # missing values
 *     if dc.usernan:
 *         n_ranges = readstat_variable_get_missing_ranges_count(variable)             # <<<<<<<<<<<<<<
 *         if n_ranges>0:
 *             missing_ranges = list()
 */
    __pyx_v_n_ranges = readstat_variable_get_missing_ranges_count(__pyx_v_variable);

    /* "pyreadstat/_readstat_parser.pyx":534
 *     if dc.usernan:
 *         n_ranges = readstat_variable_get_missing_ranges_count(variable)
 *         if n_ranges>0:             # <<<<<<<<<<<<<<
 *             missing_ranges = list()
 *             for i in range(0, n_ranges):
 */
    __pyx_t_2 = (__pyx_v_n_ranges > 0);
    if (__pyx_t_2) {

      /* "pyreadstat/_readstat_parser.pyx":535
 *         n_ranges = readstat_variable_get_missing_ranges_count(variable)
 *         if n_ranges>0:
 *             missing_ranges = list()             # <<<<<<<<<<<<<<
 *             for i in range(0, n_ranges):
 *                 loval = readstat_variable_get_missing_range_lo(variable, i)
 */
      __pyx_t_10 = PyList_New(0); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 535, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_v_missing_ranges = ((PyObject*)__pyx_t_10);
      __pyx_t_10 = 0;

      /* "pyreadstat/_readstat_parser.pyx":536
 *         if n_ranges>0:
 *             missing_ranges = list()
 *             for i in range(0, n_ranges):             # <<<<<<<<<<<<<<
 *                 loval = readstat_variable_get_missing_range_lo(variable, i)
 *                 pyloval = convert_readstat_to_python_value(loval, index, dc)
 */
      __pyx_t_6 = __pyx_v_n_ranges;
      __pyx_t_11 = __pyx_t_6;
      for (__pyx_t_12 = 0; __pyx_t_12 < __pyx_t_11; __pyx_t_12+=1) {
        __pyx_v_i = __pyx_t_12;

        /* "pyreadstat/_readstat_parser.pyx":537
 *             missing_ranges = list()
 *             for i in range(0, n_ranges):
 *                 loval = readstat_variable_get_missing_range_lo(variable, i)             # <<<<<<<<<<<<<<
 *                 pyloval = convert_readstat_to_python_value(loval, index, dc)
 *                 hival = readstat_variable_get_missing_range_hi(variable, i)
 */
        __pyx_v_loval = readstat_variable_get_missing_range_lo(__pyx_v_variable, __pyx_v_i);

        /* "pyreadstat/_readstat_parser.pyx":538
 *             for i in range(0, n_ranges):
 *                 loval = readstat_variable_get_missing_range_lo(variable, i)
 *                 pyloval = convert_readstat_to_python_value(loval, index, dc)             # <<<<<<<<<<<<<<
 *                 hival = readstat_variable_get_missing_range_hi(variable, i)
 *                 pyhival = convert_readstat_to_python_value(hival, index, dc)
 */
        __pyx_t_10 = __pyx_f_10pyreadstat_16_readstat_parser_convert_readstat_to_python_value(__pyx_v_loval, __pyx_v_index, __pyx_v_dc); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 538, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        __Pyx_XDECREF_SET(__pyx_v_pyloval, __pyx_t_10);
        __pyx_t_10 = 0;

        /* "pyreadstat/_readstat_parser.pyx":539
 *                 loval = readstat_variable_get_missing_range_lo(variable, i)
 *                 pyloval = convert_readstat_to_python_value(loval, index, dc)
 *                 hival = readstat_variable_get_missing_range_hi(variable, i)             # <<<<<<<<<<<<<<
 *                 pyhival = convert_readstat_to_python_value(hival, index, dc)
 *                 missing_ranges.append({'lo':pyloval, 'hi':pyhival})
 */
        __pyx_v_hival = readstat_variable_get_missing_range_hi(__pyx_v_variable, __pyx_v_i);

        /* "pyreadstat/_readstat_parser.pyx":540
 *                 pyloval = convert_readstat_to_python_value(loval, index, dc)
 *                 hival = readstat_variable_get_missing_range_hi(variable, i)
 *                 pyhival = convert_readstat_to_python_value(hival, index, dc)             # <<<<<<<<<<<<<<
 *                 missing_ranges.append({'lo':pyloval, 'hi':pyhival})
 *             dc.missing_ranges[col_name] = missing_ranges
 */
        __pyx_t_10 = __pyx_f_10pyreadstat_16_readstat_parser_convert_readstat_to_python_value(__pyx_v_hival, __pyx_v_index, __pyx_v_dc); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 540, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        __Pyx_XDECREF_SET(__pyx_v_pyhival, __pyx_t_10);
        __pyx_t_10 = 0;

        /* "pyreadstat/_readstat_parser.pyx":541
 *                 hival = readstat_variable_get_missing_range_hi(variable, i)
 *                 pyhival = convert_readstat_to_python_value(hival, index, dc)
 *                 missing_ranges.append({'lo':pyloval, 'hi':pyhival})             # <<<<<<<<<<<<<<
 *             dc.missing_ranges[col_name] = missing_ranges
 * 
 */
        __pyx_t_10 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 541, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        if (PyDict_SetItem(__pyx_t_10, __pyx_n_s_lo, __pyx_v_pyloval) < 0) __PYX_ERR(0, 541, __pyx_L1_error)
        if (PyDict_SetItem(__pyx_t_10, __pyx_n_s_hi, __pyx_v_pyhival) < 0) __PYX_ERR(0, 541, __pyx_L1_error)
        __pyx_t_7 = __Pyx_PyList_Append(__pyx_v_missing_ranges, __pyx_t_10); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(0, 541, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      }

      /* "pyreadstat/_readstat_parser.pyx":542
 *                 pyhival = convert_readstat_to_python_value(hival, index, dc)
 *                 missing_ranges.append({'lo':pyloval, 'hi':pyhival})
 *             dc.missing_ranges[col_name] = missing_ranges             # <<<<<<<<<<<<<<
 * 
 *     cdef size_t storage_width
 */
      if (unlikely(__pyx_v_dc->missing_ranges == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 542, __pyx_L1_error)
      }
      if (unlikely((PyDict_SetItem(__pyx_v_dc->missing_ranges, __pyx_v_col_name, __pyx_v_missing_ranges) < 0))) __PYX_ERR(0, 542, __pyx_L1_error)

      /* "pyreadstat/_readstat_parser.pyx":534
 *     if dc.usernan:
 *         n_ranges = readstat_variable_get_missing_ranges_count(variable)
 *         if n_ranges>0:             # <<<<<<<<<<<<<<
 *             missing_ranges = list()
 *             for i in range(0, n_ranges):
 */
    }

    /* "pyreadstat/_readstat_parser.pyx":532
 * 
 *     # missing values
 *     if dc.usernan:             # <<<<<<<<<<<<<<
 *         n_ranges = readstat_variable_get_missing_ranges_count(variable)
 *         if n_ranges>0:
 */
  }

  /* "pyreadstat/_readstat_parser.pyx":545
 * 
 *     cdef size_t storage_width
 *     storage_width = readstat_variable_get_storage_width(variable)             # <<<<<<<<<<<<<<
 *     dc.variable_storage_width[col_name] = <int>storage_width
 * 
 */
  __pyx_v_storage_width = readstat_variable_get_storage_width(__pyx_v_variable);

  /* "pyreadstat/_readstat_parser.pyx":546
 *     cdef size_t storage_width
 *     storage_width = readstat_variable_get_storage_width(variable)
 *     dc.variable_storage_width[col_name] = <int>storage_width             # <<<<<<<<<<<<<<
 * 
 *     dc.variable_display_width[col_name] = readstat_variable_get_display_width(variable)
 */
  __pyx_t_10 = __Pyx_PyInt_From_int(((int)__pyx_v_storage_width)); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 546, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  if (unlikely(__pyx_v_dc->variable_storage_width == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 546, __pyx_L1_error)
  }
  if (unlikely((PyDict_SetItem(__pyx_v_dc->variable_storage_width, __pyx_v_col_name, __pyx_t_10) < 0))) __PYX_ERR(0, 546, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

  /* "pyreadstat/_readstat_parser.pyx":548
 *     dc.variable_storage_width[col_name] = <int>storage_width
 * 
 *     dc.variable_display_width[col_name] = readstat_variable_get_display_width(variable)             # <<<<<<<<<<<<<<
 * 
 *     cdef readstat_alignment_t align
 */
  __pyx_t_10 = __Pyx_PyInt_From_int(readstat_variable_get_display_width(__pyx_v_variable)); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 548, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  if (unlikely((PyObject_SetItem(__pyx_v_dc->variable_display_width, __pyx_v_col_name, __pyx_t_10) < 0))) __PYX_ERR(0, 548, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

  /* "pyreadstat/_readstat_parser.pyx":552
 *     cdef readstat_alignment_t align
 *     cdef str pyalign
 *     align = readstat_variable_get_alignment(variable)             # <<<<<<<<<<<<<<
 *     if align == READSTAT_ALIGNMENT_UNKNOWN:
 *         pyalign = "unknown"
 */
  __pyx_v_align = readstat_variable_get_alignment(__pyx_v_variable);

  /* "pyreadstat/_readstat_parser.pyx":553
 *     cdef str pyalign
 *     align = readstat_variable_get_alignment(variable)
 *     if align == READSTAT_ALIGNMENT_UNKNOWN:             # <<<<<<<<<<<<<<
 *         pyalign = "unknown"
 *     elif align == READSTAT_ALIGNMENT_LEFT:
 */
  switch (__pyx_v_align) {
    case READSTAT_ALIGNMENT_UNKNOWN:

    /* "pyreadstat/_readstat_parser.pyx":554
 *     align = readstat_variable_get_alignment(variable)
 *     if align == READSTAT_ALIGNMENT_UNKNOWN:
 *         pyalign = "unknown"             # <<<<<<<<<<<<<<
 *     elif align == READSTAT_ALIGNMENT_LEFT:
 *         pyalign = "left"
 */
    __Pyx_INCREF(__pyx_n_s_unknown);
    __pyx_v_pyalign = __pyx_n_s_unknown;

    /* "pyreadstat/_readstat_parser.pyx":553
 *     cdef str pyalign
 *     align = readstat_variable_get_alignment(variable)
 *     if align == READSTAT_ALIGNMENT_UNKNOWN:             # <<<<<<<<<<<<<<
 *         pyalign = "unknown"
 *     elif align == READSTAT_ALIGNMENT_LEFT:
 */
    break;
    case READSTAT_ALIGNMENT_LEFT:

    /* "pyreadstat/_readstat_parser.pyx":556
 *         pyalign = "unknown"
 *     elif align == READSTAT_ALIGNMENT_LEFT:
 *         pyalign = "left"             # <<<<<<<<<<<<<<
 *     elif align == READSTAT_ALIGNMENT_CENTER:
 *         pyalign = "center"
 */
    __Pyx_INCREF(__pyx_n_s_left);
    __pyx_v_pyalign = __pyx_n_s_left;

    /* "pyreadstat/_readstat_parser.pyx":555
 *     if align == READSTAT_ALIGNMENT_UNKNOWN:
 *         pyalign = "unknown"
 *     elif align == READSTAT_ALIGNMENT_LEFT:             # <<<<<<<<<<<<<<
 *         pyalign = "left"
 *     elif align == READSTAT_ALIGNMENT_CENTER:
 */
    break;
    case READSTAT_ALIGNMENT_CENTER:

    /* "pyreadstat/_readstat_parser.pyx":558
 *         pyalign = "left"
 *     elif align == READSTAT_ALIGNMENT_CENTER:
 *         pyalign = "center"             # <<<<<<<<<<<<<<
 *     elif align == READSTAT_ALIGNMENT_RIGHT:
 *         pyalign = "right"
 */
    __Pyx_INCREF(__pyx_n_s_center);
    __pyx_v_pyalign = __pyx_n_s_center;

    /* "pyreadstat/_readstat_parser.pyx":557
 *     elif align == READSTAT_ALIGNMENT_LEFT:
 *         pyalign = "left"
 *     elif align == READSTAT_ALIGNMENT_CENTER:             # <<<<<<<<<<<<<<
 *         pyalign = "center"
 *     elif align == READSTAT_ALIGNMENT_RIGHT:
 */
    break;
    case READSTAT_ALIGNMENT_RIGHT:

    /* "pyreadstat/_readstat_parser.pyx":560
 *         pyalign = "center"
 *     elif align == READSTAT_ALIGNMENT_RIGHT:
 *         pyalign = "right"             # <<<<<<<<<<<<<<
 *     else:
 *         pyalign = "undetermined"
 */
    __Pyx_INCREF(__pyx_n_s_right);
    __pyx_v_pyalign = __pyx_n_s_right;

    /* "pyreadstat/_readstat_parser.pyx":559
 *     elif align == READSTAT_ALIGNMENT_CENTER:
 *         pyalign = "center"
 *     elif align == READSTAT_ALIGNMENT_RIGHT:             # <<<<<<<<<<<<<<
 *         pyalign = "right"
 *     else:
 */
    break;
    default:

    /* "pyreadstat/_readstat_parser.pyx":562
 *         pyalign = "right"
 *     else:
 *         pyalign = "undetermined"             # <<<<<<<<<<<<<<
 * 
 *     dc.variable_alignment[col_name] = pyalign
 */
    __Pyx_INCREF(__pyx_n_s_undetermined);
    __pyx_v_pyalign = __pyx_n_s_undetermined;
    break;
  }

  /* "pyreadstat/_readstat_parser.pyx":564
 *         pyalign = "undetermined"
 * 
 *     dc.variable_alignment[col_name] = pyalign             # <<<<<<<<<<<<<<
 * 
 *     cdef readstat_measure_t measure
 */
  if (unlikely((PyObject_SetItem(__pyx_v_dc->variable_alignment, __pyx_v_col_name, __pyx_v_pyalign) < 0))) __PYX_ERR(0, 564, __pyx_L1_error)

  /* "pyreadstat/_readstat_parser.pyx":568
 *     cdef readstat_measure_t measure
 *     cdef str pymeasure
 *     measure = readstat_variable_get_measure(variable)             # <<<<<<<<<<<<<<
 *     if measure == READSTAT_MEASURE_UNKNOWN:
 *         pymeasure = "unknown"
 */
  __pyx_v_measure = readstat_variable_get_measure(__pyx_v_variable);

  /* "pyreadstat/_readstat_parser.pyx":569
 *     cdef str pymeasure
 *     measure = readstat_variable_get_measure(variable)
 *     if measure == READSTAT_MEASURE_UNKNOWN:             # <<<<<<<<<<<<<<
 *         pymeasure = "unknown"
 *     elif measure == READSTAT_MEASURE_NOMINAL:
 */
  switch (__pyx_v_measure) {
    case READSTAT_MEASURE_UNKNOWN:

    /* "pyreadstat/_readstat_parser.pyx":570
 *     measure = readstat_variable_get_measure(variable)
 *     if measure == READSTAT_MEASURE_UNKNOWN:
 *         pymeasure = "unknown"             # <<<<<<<<<<<<<<
 *     elif measure == READSTAT_MEASURE_NOMINAL:
 *         pymeasure = "nominal"
 */
    __Pyx_INCREF(__pyx_n_s_unknown);
    __pyx_v_pymeasure = __pyx_n_s_unknown;

    /* "pyreadstat/_readstat_parser.pyx":569
 *     cdef str pymeasure
 *     measure = readstat_variable_get_measure(variable)
 *     if measure == READSTAT_MEASURE_UNKNOWN:             # <<<<<<<<<<<<<<
 *         pymeasure = "unknown"
 *     elif measure == READSTAT_MEASURE_NOMINAL:
 */
    break;
    case READSTAT_MEASURE_NOMINAL:

    /* "pyreadstat/_readstat_parser.pyx":572
 *         pymeasure = "unknown"
 *     elif measure == READSTAT_MEASURE_NOMINAL:
 *         pymeasure = "nominal"             # <<<<<<<<<<<<<<
 *     elif measure == READSTAT_MEASURE_ORDINAL:
 *         pymeasure = "ordinal"
 */
    __Pyx_INCREF(__pyx_n_s_nominal);
    __pyx_v_pymeasure = __pyx_n_s_nominal;

    /* "pyreadstat/_readstat_parser.pyx":571
 *     if measure == READSTAT_MEASURE_UNKNOWN:
 *         pymeasure = "unknown"
 *     elif measure == READSTAT_MEASURE_NOMINAL:             # <<<<<<<<<<<<<<
 *         pymeasure = "nominal"
 *     elif measure == READSTAT_MEASURE_ORDINAL:
 */
    break;
    case READSTAT_MEASURE_ORDINAL:

    /* "pyreadstat/_readstat_parser.pyx":574
 *         pymeasure = "nominal"
 *     elif measure == READSTAT_MEASURE_ORDINAL:
 *         pymeasure = "ordinal"             # <<<<<<<<<<<<<<
 *     elif measure == READSTAT_MEASURE_SCALE:
 *         pymeasure = "scale"
 */
    __Pyx_INCREF(__pyx_n_s_ordinal);
    __pyx_v_pymeasure = __pyx_n_s_ordinal;

    /* "pyreadstat/_readstat_parser.pyx":573
 *     elif measure == READSTAT_MEASURE_NOMINAL:
 *         pymeasure = "nominal"
 *     elif measure == READSTAT_MEASURE_ORDINAL:             # <<<<<<<<<<<<<<
 *         pymeasure = "ordinal"
 *     elif measure == READSTAT_MEASURE_SCALE:
 */
    break;
    case READSTAT_MEASURE_SCALE:

    /* "pyreadstat/_readstat_parser.pyx":576
 *         pymeasure = "ordinal"
 *     elif measure == READSTAT_MEASURE_SCALE:
 *         pymeasure = "scale"             # <<<<<<<<<<<<<<
 *     else:
 *         pymeasure = "undetermined"
 */
    __Pyx_INCREF(__pyx_n_s_scale);
    __pyx_v_pymeasure = __pyx_n_s_scale;

    /* "pyreadstat/_readstat_parser.pyx":575
 *     elif measure == READSTAT_MEASURE_ORDINAL:
 *         pymeasure = "ordinal"
 *     elif measure == READSTAT_MEASURE_SCALE:             # <<<<<<<<<<<<<<
 *         pymeasure = "scale"
 *     else:
 */
    break;
    default:

    /* "pyreadstat/_readstat_parser.pyx":578
 *         pymeasure = "scale"
 *     else:
 *         pymeasure = "undetermined"             # <<<<<<<<<<<<<<
 * 
 *     dc.variable_measure[col_name] = pymeasure
 */
    __Pyx_INCREF(__pyx_n_s_undetermined);
    __pyx_v_pymeasure = __pyx_n_s_undetermined;
    break;
  }

  /* "pyreadstat/_readstat_parser.pyx":580
 *         pymeasure = "undetermined"
 * 
 *     dc.variable_measure[col_name] = pymeasure             # <<<<<<<<<<<<<<
 * 
 *     return READSTAT_HANDLER_OK
 */
  if (unlikely((PyObject_SetItem(__pyx_v_dc->variable_measure, __pyx_v_col_name, __pyx_v_pymeasure) < 0))) __PYX_ERR(0, 580, __pyx_L1_error)

  /* "pyreadstat/_readstat_parser.pyx":582
 *     dc.variable_measure[col_name] = pymeasure
 * 
 *     return READSTAT_HANDLER_OK             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = READSTAT_HANDLER_OK;
  goto __pyx_L0;

  /* "pyreadstat/_readstat_parser.pyx":418
 *     return READSTAT_HANDLER_OK
 * 
 * cdef int handle_variable(int index, readstat_variable_t *variable,             # <<<<<<<<<<<<<<
 *                          char *val_labels, void *ctx) except READSTAT_HANDLER_ABORT:
 *     """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("pyreadstat._readstat_parser.handle_variable", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = READSTAT_HANDLER_ABORT;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_col_name);
  __Pyx_XDECREF(__pyx_v_col_label);
  __Pyx_XDECREF(__pyx_v_label_name);
  __Pyx_XDECREF(__pyx_v_col_format_original);
  __Pyx_XDECREF(__pyx_v_pyloval);
  __Pyx_XDECREF(__pyx_v_pyhival);
  __Pyx_XDECREF(__pyx_v_missing_ranges);
  __Pyx_XDECREF(__pyx_v_curnptype);
  __Pyx_XDECREF(__pyx_v_row);
  __Pyx_XDECREF(__pyx_v_newcolname);
  __Pyx_XDECREF((PyObject *)__pyx_v_dc);
  __Pyx_XDECREF(__pyx_v_msg);
  __Pyx_XDECREF(__pyx_v_pyalign);
  __Pyx_XDECREF(__pyx_v_pymeasure);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyreadstat/_readstat_parser.pyx":585
 * 
 * 
 * cdef int handle_value(int obs_index, readstat_variable_t * variable, readstat_value_t value, void *ctx) except READSTAT_HANDLER_ABORT:             # <<<<<<<<<<<<<<
 *     """
 *     This function transforms every value to python types, and to datetime if
 */

static int __pyx_f_10pyreadstat_16_readstat_parser_handle_value(int __pyx_v_obs_index, readstat_variable_t *__pyx_v_variable, readstat_value_t __pyx_v_value, void *__pyx_v_ctx) {
  struct __pyx_obj_10pyreadstat_16_readstat_parser_data_container *__pyx_v_dc = 0;
  int __pyx_v_index;
  int __pyx_v_max_n_obs;
  int __pyx_v_is_unkown_number_rows;
  int __pyx_v_var_max_rows;
  PyObject *__pyx_v_buf_list = 0;
  int __pyx_v_iscurnptypeobject;
  int __pyx_v_iscurnptypefloat;
  int __pyx_v_missing_tag;
  PyObject *__pyx_v_pyvalue = 0;
  PyObject *__pyx_v_curset = 0;
  PyObject *__pyx_v_curnptype = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("handle_value", 0);

  /* "pyreadstat/_readstat_parser.pyx":607
 * 
 *     # extract variables we need from data container
 *     dc = <data_container> ctx             # <<<<<<<<<<<<<<
 *     index = readstat_variable_get_index_after_skipping(variable)
 *     max_n_obs = dc.max_n_obs
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_ctx);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_dc = ((struct __pyx_obj_10pyreadstat_16_readstat_parser_data_container *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyreadstat/_readstat_parser.pyx":608
 *     # extract variables we need from data container
 *     dc = <data_container> ctx
 *     index = readstat_variable_get_index_after_skipping(variable)             # <<<<<<<<<<<<<<
 *     max_n_obs = dc.max_n_obs
 *     is_unkown_number_rows = dc.is_unkown_number_rows
 */
  __pyx_v_index = readstat_variable_get_index_after_skipping(__pyx_v_variable);

  /* "pyreadstat/_readstat_parser.pyx":609
 *     dc = <data_container> ctx
 *     index = readstat_variable_get_index_after_skipping(variable)
 *     max_n_obs = dc.max_n_obs             # <<<<<<<<<<<<<<
 *     is_unkown_number_rows = dc.is_unkown_number_rows
 * 
 */
  __pyx_t_2 = __pyx_v_dc->max_n_obs;
  __pyx_v_max_n_obs = __pyx_t_2;

  /* "pyreadstat/_readstat_parser.pyx":610
 *     index = readstat_variable_get_index_after_skipping(variable)
 *     max_n_obs = dc.max_n_obs
 *     is_unkown_number_rows = dc.is_unkown_number_rows             # <<<<<<<<<<<<<<
 * 
 *     # check that we still have enough room in our pre-allocated lists
 */
  __pyx_t_3 = __pyx_v_dc->is_unkown_number_rows;
  __pyx_v_is_unkown_number_rows = __pyx_t_3;

  /* "pyreadstat/_readstat_parser.pyx":614
 *     # check that we still have enough room in our pre-allocated lists
 *     # if not, add more room
 *     iscurnptypeobject = dc.col_dtypes_isobject[index]             # <<<<<<<<<<<<<<
 *     iscurnptypefloat = dc.col_dytpes_isfloat[index]
 *     if is_unkown_number_rows:
 */
  if (unlikely(__pyx_v_dc->col_dtypes_isobject == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 614, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_List(__pyx_v_dc->col_dtypes_isobject, __pyx_v_index, int, 1, __Pyx_PyInt_From_int, 1, 1, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 614, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_3 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 614, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_iscurnptypeobject = __pyx_t_3;

  /* "pyreadstat/_readstat_parser.pyx":615
 *     # if not, add more room
 *     iscurnptypeobject = dc.col_dtypes_isobject[index]
 *     iscurnptypefloat = dc.col_dytpes_isfloat[index]             # <<<<<<<<<<<<<<
 *     if is_unkown_number_rows:
 *         if max_n_obs <= obs_index:
 */
  if (unlikely(__pyx_v_dc->col_dytpes_isfloat == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 615, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_List(__pyx_v_dc->col_dytpes_isfloat, __pyx_v_index, int, 1, __Pyx_PyInt_From_int, 1, 1, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 615, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_3 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 615, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_iscurnptypefloat = __pyx_t_3;

  /* "pyreadstat/_readstat_parser.pyx":616
 *     iscurnptypeobject = dc.col_dtypes_isobject[index]
 *     iscurnptypefloat = dc.col_dytpes_isfloat[index]
 *     if is_unkown_number_rows:             # <<<<<<<<<<<<<<
 *         if max_n_obs <= obs_index:
 *             dc.max_n_obs = obs_index + 1
 */
  if (__pyx_v_is_unkown_number_rows) {

    /* "pyreadstat/_readstat_parser.pyx":617
 *     iscurnptypefloat = dc.col_dytpes_isfloat[index]
 *     if is_unkown_number_rows:
 *         if max_n_obs <= obs_index:             # <<<<<<<<<<<<<<
 *             dc.max_n_obs = obs_index + 1
 *         var_max_rows = dc.col_data_len[index]
 */
    __pyx_t_3 = (__pyx_v_max_n_obs <= __pyx_v_obs_index);
    if (__pyx_t_3) {

      /* "pyreadstat/_readstat_parser.pyx":618
 *     if is_unkown_number_rows:
 *         if max_n_obs <= obs_index:
 *             dc.max_n_obs = obs_index + 1             # <<<<<<<<<<<<<<
 *         var_max_rows = dc.col_data_len[index]
 *         if var_max_rows <= obs_index:
 */
      __pyx_v_dc->max_n_obs = (__pyx_v_obs_index + 1);

      /* "pyreadstat/_readstat_parser.pyx":617
 *     iscurnptypefloat = dc.col_dytpes_isfloat[index]
 *     if is_unkown_number_rows:
 *         if max_n_obs <= obs_index:             # <<<<<<<<<<<<<<
 *             dc.max_n_obs = obs_index + 1
 *         var_max_rows = dc.col_data_len[index]
 */
    }

    /* "pyreadstat/_readstat_parser.pyx":619
 *         if max_n_obs <= obs_index:
 *             dc.max_n_obs = obs_index + 1
 *         var_max_rows = dc.col_data_len[index]             # <<<<<<<<<<<<<<
 *         if var_max_rows <= obs_index:
 *             curnptype = dc.col_numpy_dtypes[index]
 */
    if (unlikely(__pyx_v_dc->col_data_len == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 619, __pyx_L1_error)
    }
    __pyx_t_1 = __Pyx_GetItemInt_List(__pyx_v_dc->col_data_len, __pyx_v_index, int, 1, __Pyx_PyInt_From_int, 1, 1, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 619, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 619, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_var_max_rows = __pyx_t_2;

    /* "pyreadstat/_readstat_parser.pyx":620
 *             dc.max_n_obs = obs_index + 1
 *         var_max_rows = dc.col_data_len[index]
 *         if var_max_rows <= obs_index:             # <<<<<<<<<<<<<<
 *             curnptype = dc.col_numpy_dtypes[index]
 *             buf_list = np.empty(100000, dtype=curnptype)
 */
    __pyx_t_3 = (__pyx_v_var_max_rows <= __pyx_v_obs_index);
    if (__pyx_t_3) {

      /* "pyreadstat/_readstat_parser.pyx":621
 *         var_max_rows = dc.col_data_len[index]
 *         if var_max_rows <= obs_index:
 *             curnptype = dc.col_numpy_dtypes[index]             # <<<<<<<<<<<<<<
 *             buf_list = np.empty(100000, dtype=curnptype)
 *             if iscurnptypeobject or iscurnptypefloat:
 */
      if (unlikely(__pyx_v_dc->col_numpy_dtypes == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 621, __pyx_L1_error)
      }
      __pyx_t_1 = __Pyx_GetItemInt_List(__pyx_v_dc->col_numpy_dtypes, __pyx_v_index, int, 1, __Pyx_PyInt_From_int, 1, 1, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 621, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_v_curnptype = __pyx_t_1;
      __pyx_t_1 = 0;

      /* "pyreadstat/_readstat_parser.pyx":622
 *         if var_max_rows <= obs_index:
 *             curnptype = dc.col_numpy_dtypes[index]
 *             buf_list = np.empty(100000, dtype=curnptype)             # <<<<<<<<<<<<<<
 *             if iscurnptypeobject or iscurnptypefloat:
 *                 buf_list.fill(np.nan)
 */
      __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 622, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_empty); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 622, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 622, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_dtype, __pyx_v_curnptype) < 0) __PYX_ERR(0, 622, __pyx_L1_error)
      __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_tuple__4, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 622, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_v_buf_list = __pyx_t_5;
      __pyx_t_5 = 0;

      /* "pyreadstat/_readstat_parser.pyx":623
 *             curnptype = dc.col_numpy_dtypes[index]
 *             buf_list = np.empty(100000, dtype=curnptype)
 *             if iscurnptypeobject or iscurnptypefloat:             # <<<<<<<<<<<<<<
 *                 buf_list.fill(np.nan)
 *             dc.col_data[index] = np.append(dc.col_data[index], buf_list)
 */
      if (!__pyx_v_iscurnptypeobject) {
      } else {
        __pyx_t_3 = __pyx_v_iscurnptypeobject;
        goto __pyx_L7_bool_binop_done;
      }
      __pyx_t_3 = __pyx_v_iscurnptypefloat;
      __pyx_L7_bool_binop_done:;
      if (__pyx_t_3) {

        /* "pyreadstat/_readstat_parser.pyx":624
 *             buf_list = np.empty(100000, dtype=curnptype)
 *             if iscurnptypeobject or iscurnptypefloat:
 *                 buf_list.fill(np.nan)             # <<<<<<<<<<<<<<
 *             dc.col_data[index] = np.append(dc.col_data[index], buf_list)
 *             var_max_rows += 100000
 */
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buf_list, __pyx_n_s_fill); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 624, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 624, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_nan); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 624, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = NULL;
        __pyx_t_2 = 0;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_1);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_1, function);
            __pyx_t_2 = 1;
          }
        }
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_6};
          __pyx_t_5 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_2, 1+__pyx_t_2);
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 624, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        }
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

        /* "pyreadstat/_readstat_parser.pyx":623
 *             curnptype = dc.col_numpy_dtypes[index]
 *             buf_list = np.empty(100000, dtype=curnptype)
 *             if iscurnptypeobject or iscurnptypefloat:             # <<<<<<<<<<<<<<
 *                 buf_list.fill(np.nan)
 *             dc.col_data[index] = np.append(dc.col_data[index], buf_list)
 */
      }

      /* "pyreadstat/_readstat_parser.pyx":625
 *             if iscurnptypeobject or iscurnptypefloat:
 *                 buf_list.fill(np.nan)
 *             dc.col_data[index] = np.append(dc.col_data[index], buf_list)             # <<<<<<<<<<<<<<
 *             var_max_rows += 100000
 *             dc.col_data_len[index] = var_max_rows
 */
      __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 625, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_append); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 625, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (unlikely(__pyx_v_dc->col_data == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 625, __pyx_L1_error)
      }
      __pyx_t_1 = __Pyx_GetItemInt_List(__pyx_v_dc->col_data, __pyx_v_index, int, 1, __Pyx_PyInt_From_int, 1, 1, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 625, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_4 = NULL;
      __pyx_t_2 = 0;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_6);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_6, function);
          __pyx_t_2 = 1;
        }
      }
      {
        PyObject *__pyx_callargs[3] = {__pyx_t_4, __pyx_t_1, __pyx_v_buf_list};
        __pyx_t_5 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_2, 2+__pyx_t_2);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 625, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
      if (unlikely(__pyx_v_dc->col_data == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 625, __pyx_L1_error)
      }
      if (unlikely((__Pyx_SetItemInt(__pyx_v_dc->col_data, __pyx_v_index, __pyx_t_5, int, 1, __Pyx_PyInt_From_int, 1, 1, 1) < 0))) __PYX_ERR(0, 625, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "pyreadstat/_readstat_parser.pyx":626
 *                 buf_list.fill(np.nan)
 *             dc.col_data[index] = np.append(dc.col_data[index], buf_list)
 *             var_max_rows += 100000             # <<<<<<<<<<<<<<
 *             dc.col_data_len[index] = var_max_rows
 * 
 */
      __pyx_v_var_max_rows = (__pyx_v_var_max_rows + 0x186A0);

      /* "pyreadstat/_readstat_parser.pyx":627
 *             dc.col_data[index] = np.append(dc.col_data[index], buf_list)
 *             var_max_rows += 100000
 *             dc.col_data_len[index] = var_max_rows             # <<<<<<<<<<<<<<
 * 
 *     # transform to python value types
 */
      __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_var_max_rows); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 627, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (unlikely(__pyx_v_dc->col_data_len == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 627, __pyx_L1_error)
      }
      if (unlikely((__Pyx_SetItemInt(__pyx_v_dc->col_data_len, __pyx_v_index, __pyx_t_5, int, 1, __Pyx_PyInt_From_int, 1, 1, 1) < 0))) __PYX_ERR(0, 627, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "pyreadstat/_readstat_parser.pyx":620
 *             dc.max_n_obs = obs_index + 1
 *         var_max_rows = dc.col_data_len[index]
 *         if var_max_rows <= obs_index:             # <<<<<<<<<<<<<<
 *             curnptype = dc.col_numpy_dtypes[index]
 *             buf_list = np.empty(100000, dtype=curnptype)
 */
    }

    /* "pyreadstat/_readstat_parser.pyx":616
 *     iscurnptypeobject = dc.col_dtypes_isobject[index]
 *     iscurnptypefloat = dc.col_dytpes_isfloat[index]
 *     if is_unkown_number_rows:             # <<<<<<<<<<<<<<
 *         if max_n_obs <= obs_index:
 *             dc.max_n_obs = obs_index + 1
 */
  }

  /* "pyreadstat/_readstat_parser.pyx":630
 * 
 *     # transform to python value types
 *     if readstat_value_is_missing(value, variable):             # <<<<<<<<<<<<<<
 *         # The user does not want to retrieve missing values
 *         if not dc.usernan or readstat_value_is_system_missing(value):
 */
  __pyx_t_3 = (readstat_value_is_missing(__pyx_v_value, __pyx_v_variable) != 0);
  if (__pyx_t_3) {

    /* "pyreadstat/_readstat_parser.pyx":632
 *     if readstat_value_is_missing(value, variable):
 *         # The user does not want to retrieve missing values
 *         if not dc.usernan or readstat_value_is_system_missing(value):             # <<<<<<<<<<<<<<
 *             if iscurnptypefloat == 1 or iscurnptypeobject == 1:
 *                 # already allocated
 */
    __pyx_t_7 = (!__pyx_v_dc->usernan);
    if (!__pyx_t_7) {
    } else {
      __pyx_t_3 = __pyx_t_7;
      goto __pyx_L11_bool_binop_done;
    }
    __pyx_t_7 = (readstat_value_is_system_missing(__pyx_v_value) != 0);
    __pyx_t_3 = __pyx_t_7;
    __pyx_L11_bool_binop_done:;
    if (__pyx_t_3) {

      /* "pyreadstat/_readstat_parser.pyx":633
 *         # The user does not want to retrieve missing values
 *         if not dc.usernan or readstat_value_is_system_missing(value):
 *             if iscurnptypefloat == 1 or iscurnptypeobject == 1:             # <<<<<<<<<<<<<<
 *                 # already allocated
 *                 pass
 */
      __pyx_t_7 = (__pyx_v_iscurnptypefloat == 1);
      if (!__pyx_t_7) {
      } else {
        __pyx_t_3 = __pyx_t_7;
        goto __pyx_L14_bool_binop_done;
      }
      __pyx_t_7 = (__pyx_v_iscurnptypeobject == 1);
      __pyx_t_3 = __pyx_t_7;
      __pyx_L14_bool_binop_done:;
      if (__pyx_t_3) {
        goto __pyx_L13;
      }

      /* "pyreadstat/_readstat_parser.pyx":639
 *             # for any type except float, the numpy type will be object as now we have nans
 *             else:
 *                 dc.col_numpy_dtypes[index] = object             # <<<<<<<<<<<<<<
 *                 dc.col_dtypes_isobject[index] = 1
 *                 iscurnptypeobject = 1
 */
      /*else*/ {
        if (unlikely(__pyx_v_dc->col_numpy_dtypes == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
          __PYX_ERR(0, 639, __pyx_L1_error)
        }
        if (unlikely((__Pyx_SetItemInt(__pyx_v_dc->col_numpy_dtypes, __pyx_v_index, __pyx_builtin_object, int, 1, __Pyx_PyInt_From_int, 1, 1, 1) < 0))) __PYX_ERR(0, 639, __pyx_L1_error)

        /* "pyreadstat/_readstat_parser.pyx":640
 *             else:
 *                 dc.col_numpy_dtypes[index] = object
 *                 dc.col_dtypes_isobject[index] = 1             # <<<<<<<<<<<<<<
 *                 iscurnptypeobject = 1
 *                 dc.col_data[index] = dc.col_data[index].astype(object, copy=False)
 */
        if (unlikely(__pyx_v_dc->col_dtypes_isobject == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
          __PYX_ERR(0, 640, __pyx_L1_error)
        }
        if (unlikely((__Pyx_SetItemInt(__pyx_v_dc->col_dtypes_isobject, __pyx_v_index, __pyx_int_1, int, 1, __Pyx_PyInt_From_int, 1, 1, 1) < 0))) __PYX_ERR(0, 640, __pyx_L1_error)

        /* "pyreadstat/_readstat_parser.pyx":641
 *                 dc.col_numpy_dtypes[index] = object
 *                 dc.col_dtypes_isobject[index] = 1
 *                 iscurnptypeobject = 1             # <<<<<<<<<<<<<<
 *                 dc.col_data[index] = dc.col_data[index].astype(object, copy=False)
 *                 dc.col_data[index][obs_index:] = np.nan
 */
        __pyx_v_iscurnptypeobject = 1;

        /* "pyreadstat/_readstat_parser.pyx":642
 *                 dc.col_dtypes_isobject[index] = 1
 *                 iscurnptypeobject = 1
 *                 dc.col_data[index] = dc.col_data[index].astype(object, copy=False)             # <<<<<<<<<<<<<<
 *                 dc.col_data[index][obs_index:] = np.nan
 *                 #dc.col_data[index][obs_index] = NAN
 */
        if (unlikely(__pyx_v_dc->col_data == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
          __PYX_ERR(0, 642, __pyx_L1_error)
        }
        __pyx_t_5 = __Pyx_GetItemInt_List(__pyx_v_dc->col_data, __pyx_v_index, int, 1, __Pyx_PyInt_From_int, 1, 1, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 642, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_astype); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 642, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 642, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_copy, Py_False) < 0) __PYX_ERR(0, 642, __pyx_L1_error)
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_tuple__5, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 642, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        if (unlikely(__pyx_v_dc->col_data == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
          __PYX_ERR(0, 642, __pyx_L1_error)
        }
        if (unlikely((__Pyx_SetItemInt(__pyx_v_dc->col_data, __pyx_v_index, __pyx_t_1, int, 1, __Pyx_PyInt_From_int, 1, 1, 1) < 0))) __PYX_ERR(0, 642, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pyreadstat/_readstat_parser.pyx":643
 *                 iscurnptypeobject = 1
 *                 dc.col_data[index] = dc.col_data[index].astype(object, copy=False)
 *                 dc.col_data[index][obs_index:] = np.nan             # <<<<<<<<<<<<<<
 *                 #dc.col_data[index][obs_index] = NAN
 *         elif readstat_value_is_defined_missing(value, variable):
 */
        __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 643, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_nan); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 643, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (unlikely(__pyx_v_dc->col_data == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
          __PYX_ERR(0, 643, __pyx_L1_error)
        }
        __pyx_t_1 = __Pyx_GetItemInt_List(__pyx_v_dc->col_data, __pyx_v_index, int, 1, __Pyx_PyInt_From_int, 1, 1, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 643, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        if (__Pyx_PyObject_SetSlice(__pyx_t_1, __pyx_t_5, __pyx_v_obs_index, 0, NULL, NULL, NULL, 1, 0, 1) < 0) __PYX_ERR(0, 643, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
      __pyx_L13:;

      /* "pyreadstat/_readstat_parser.pyx":632
 *     if readstat_value_is_missing(value, variable):
 *         # The user does not want to retrieve missing values
 *         if not dc.usernan or readstat_value_is_system_missing(value):             # <<<<<<<<<<<<<<
 *             if iscurnptypefloat == 1 or iscurnptypeobject == 1:
 *                 # already allocated
 */
      goto __pyx_L10;
    }

    /* "pyreadstat/_readstat_parser.pyx":645
 *                 dc.col_data[index][obs_index:] = np.nan
 *                 #dc.col_data[index][obs_index] = NAN
 *         elif readstat_value_is_defined_missing(value, variable):             # <<<<<<<<<<<<<<
 *             # SPSS missing values
 *             pyvalue = convert_readstat_to_python_value(value, index, dc)
 */
    __pyx_t_3 = (readstat_value_is_defined_missing(__pyx_v_value, __pyx_v_variable) != 0);
    if (__pyx_t_3) {

      /* "pyreadstat/_readstat_parser.pyx":647
 *         elif readstat_value_is_defined_missing(value, variable):
 *             # SPSS missing values
 *             pyvalue = convert_readstat_to_python_value(value, index, dc)             # <<<<<<<<<<<<<<
 *             dc.col_data[index][obs_index] = pyvalue
 *         elif readstat_value_is_tagged_missing(value):
 */
      __pyx_t_5 = __pyx_f_10pyreadstat_16_readstat_parser_convert_readstat_to_python_value(__pyx_v_value, __pyx_v_index, __pyx_v_dc); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 647, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_v_pyvalue = __pyx_t_5;
      __pyx_t_5 = 0;

      /* "pyreadstat/_readstat_parser.pyx":648
 *             # SPSS missing values
 *             pyvalue = convert_readstat_to_python_value(value, index, dc)
 *             dc.col_data[index][obs_index] = pyvalue             # <<<<<<<<<<<<<<
 *         elif readstat_value_is_tagged_missing(value):
 *             iscurnptypeobject = dc.col_dtypes_isobject[index]
 */
      if (unlikely(__pyx_v_dc->col_data == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 648, __pyx_L1_error)
      }
      __pyx_t_5 = __Pyx_GetItemInt_List(__pyx_v_dc->col_data, __pyx_v_index, int, 1, __Pyx_PyInt_From_int, 1, 1, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 648, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (unlikely((__Pyx_SetItemInt(__pyx_t_5, __pyx_v_obs_index, __pyx_v_pyvalue, int, 1, __Pyx_PyInt_From_int, 0, 1, 1) < 0))) __PYX_ERR(0, 648, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "pyreadstat/_readstat_parser.pyx":645
 *                 dc.col_data[index][obs_index:] = np.nan
 *                 #dc.col_data[index][obs_index] = NAN
 *         elif readstat_value_is_defined_missing(value, variable):             # <<<<<<<<<<<<<<
 *             # SPSS missing values
 *             pyvalue = convert_readstat_to_python_value(value, index, dc)
 */
      goto __pyx_L10;
    }

    /* "pyreadstat/_readstat_parser.pyx":649
 *             pyvalue = convert_readstat_to_python_value(value, index, dc)
 *             dc.col_data[index][obs_index] = pyvalue
 *         elif readstat_value_is_tagged_missing(value):             # <<<<<<<<<<<<<<
 *             iscurnptypeobject = dc.col_dtypes_isobject[index]
 *             # SAS and Stata missing values
 */
    __pyx_t_3 = (readstat_value_is_tagged_missing(__pyx_v_value) != 0);
    if (__pyx_t_3) {

      /* "pyreadstat/_readstat_parser.pyx":650
 *             dc.col_data[index][obs_index] = pyvalue
 *         elif readstat_value_is_tagged_missing(value):
 *             iscurnptypeobject = dc.col_dtypes_isobject[index]             # <<<<<<<<<<<<<<
 *             # SAS and Stata missing values
 *             missing_tag = <int> readstat_value_tag(value)
 */
      if (unlikely(__pyx_v_dc->col_dtypes_isobject == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 650, __pyx_L1_error)
      }
      __pyx_t_5 = __Pyx_GetItemInt_List(__pyx_v_dc->col_dtypes_isobject, __pyx_v_index, int, 1, __Pyx_PyInt_From_int, 1, 1, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 650, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely((__pyx_t_3 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 650, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_v_iscurnptypeobject = __pyx_t_3;

      /* "pyreadstat/_readstat_parser.pyx":652
 *             iscurnptypeobject = dc.col_dtypes_isobject[index]
 *             # SAS and Stata missing values
 *             missing_tag = <int> readstat_value_tag(value)             # <<<<<<<<<<<<<<
 *             # In SAS missing values are A to Z or _ in stata a to z
 *             # if (missing_tag >=65 and missing_tag <= 90) or missing_tag == 95 or (missing_tag >=61 and missing_tag <= 122):
 */
      __pyx_v_missing_tag = ((int)readstat_value_tag(__pyx_v_value));

      /* "pyreadstat/_readstat_parser.pyx":655
 *             # In SAS missing values are A to Z or _ in stata a to z
 *             # if (missing_tag >=65 and missing_tag <= 90) or missing_tag == 95 or (missing_tag >=61 and missing_tag <= 122):
 *             if iscurnptypeobject == 1:             # <<<<<<<<<<<<<<
 *                 dc.col_data[index][obs_index] =  chr(missing_tag)
 *             else:
 */
      __pyx_t_3 = (__pyx_v_iscurnptypeobject == 1);
      if (__pyx_t_3) {

        /* "pyreadstat/_readstat_parser.pyx":656
 *             # if (missing_tag >=65 and missing_tag <= 90) or missing_tag == 95 or (missing_tag >=61 and missing_tag <= 122):
 *             if iscurnptypeobject == 1:
 *                 dc.col_data[index][obs_index] =  chr(missing_tag)             # <<<<<<<<<<<<<<
 *             else:
 *                 dc.col_numpy_dtypes[index] = object
 */
        __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_missing_tag); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 656, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_chr, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 656, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        if (unlikely(__pyx_v_dc->col_data == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
          __PYX_ERR(0, 656, __pyx_L1_error)
        }
        __pyx_t_5 = __Pyx_GetItemInt_List(__pyx_v_dc->col_data, __pyx_v_index, int, 1, __Pyx_PyInt_From_int, 1, 1, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 656, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        if (unlikely((__Pyx_SetItemInt(__pyx_t_5, __pyx_v_obs_index, __pyx_t_1, int, 1, __Pyx_PyInt_From_int, 0, 1, 1) < 0))) __PYX_ERR(0, 656, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "pyreadstat/_readstat_parser.pyx":655
 *             # In SAS missing values are A to Z or _ in stata a to z
 *             # if (missing_tag >=65 and missing_tag <= 90) or missing_tag == 95 or (missing_tag >=61 and missing_tag <= 122):
 *             if iscurnptypeobject == 1:             # <<<<<<<<<<<<<<
 *                 dc.col_data[index][obs_index] =  chr(missing_tag)
 *             else:
 */
        goto __pyx_L16;
      }

      /* "pyreadstat/_readstat_parser.pyx":658
 *                 dc.col_data[index][obs_index] =  chr(missing_tag)
 *             else:
 *                 dc.col_numpy_dtypes[index] = object             # <<<<<<<<<<<<<<
 *                 dc.col_dtypes_isobject[index] = 1
 *                 dc.col_dytpes_isfloat[index] = 0
 */
      /*else*/ {
        if (unlikely(__pyx_v_dc->col_numpy_dtypes == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
          __PYX_ERR(0, 658, __pyx_L1_error)
        }
        if (unlikely((__Pyx_SetItemInt(__pyx_v_dc->col_numpy_dtypes, __pyx_v_index, __pyx_builtin_object, int, 1, __Pyx_PyInt_From_int, 1, 1, 1) < 0))) __PYX_ERR(0, 658, __pyx_L1_error)

        /* "pyreadstat/_readstat_parser.pyx":659
 *             else:
 *                 dc.col_numpy_dtypes[index] = object
 *                 dc.col_dtypes_isobject[index] = 1             # <<<<<<<<<<<<<<
 *                 dc.col_dytpes_isfloat[index] = 0
 *                 iscurnptypeobject = 1
 */
        if (unlikely(__pyx_v_dc->col_dtypes_isobject == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
          __PYX_ERR(0, 659, __pyx_L1_error)
        }
        if (unlikely((__Pyx_SetItemInt(__pyx_v_dc->col_dtypes_isobject, __pyx_v_index, __pyx_int_1, int, 1, __Pyx_PyInt_From_int, 1, 1, 1) < 0))) __PYX_ERR(0, 659, __pyx_L1_error)

        /* "pyreadstat/_readstat_parser.pyx":660
 *                 dc.col_numpy_dtypes[index] = object
 *                 dc.col_dtypes_isobject[index] = 1
 *                 dc.col_dytpes_isfloat[index] = 0             # <<<<<<<<<<<<<<
 *                 iscurnptypeobject = 1
 *                 dc.col_data[index] = dc.col_data[index].astype(object, copy=False)
 */
        if (unlikely(__pyx_v_dc->col_dytpes_isfloat == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
          __PYX_ERR(0, 660, __pyx_L1_error)
        }
        if (unlikely((__Pyx_SetItemInt(__pyx_v_dc->col_dytpes_isfloat, __pyx_v_index, __pyx_int_0, int, 1, __Pyx_PyInt_From_int, 1, 1, 1) < 0))) __PYX_ERR(0, 660, __pyx_L1_error)

        /* "pyreadstat/_readstat_parser.pyx":661
 *                 dc.col_dtypes_isobject[index] = 1
 *                 dc.col_dytpes_isfloat[index] = 0
 *                 iscurnptypeobject = 1             # <<<<<<<<<<<<<<
 *                 dc.col_data[index] = dc.col_data[index].astype(object, copy=False)
 *                 dc.col_data[index][obs_index] =  chr(missing_tag)
 */
        __pyx_v_iscurnptypeobject = 1;

        /* "pyreadstat/_readstat_parser.pyx":662
 *                 dc.col_dytpes_isfloat[index] = 0
 *                 iscurnptypeobject = 1
 *                 dc.col_data[index] = dc.col_data[index].astype(object, copy=False)             # <<<<<<<<<<<<<<
 *                 dc.col_data[index][obs_index] =  chr(missing_tag)
 *             curset = dc.missing_user_values.get(index)
 */
        if (unlikely(__pyx_v_dc->col_data == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
          __PYX_ERR(0, 662, __pyx_L1_error)
        }
        __pyx_t_1 = __Pyx_GetItemInt_List(__pyx_v_dc->col_data, __pyx_v_index, int, 1, __Pyx_PyInt_From_int, 1, 1, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 662, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_astype); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 662, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 662, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_copy, Py_False) < 0) __PYX_ERR(0, 662, __pyx_L1_error)
        __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_tuple__6, __pyx_t_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 662, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (unlikely(__pyx_v_dc->col_data == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
          __PYX_ERR(0, 662, __pyx_L1_error)
        }
        if (unlikely((__Pyx_SetItemInt(__pyx_v_dc->col_data, __pyx_v_index, __pyx_t_6, int, 1, __Pyx_PyInt_From_int, 1, 1, 1) < 0))) __PYX_ERR(0, 662, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

        /* "pyreadstat/_readstat_parser.pyx":663
 *                 iscurnptypeobject = 1
 *                 dc.col_data[index] = dc.col_data[index].astype(object, copy=False)
 *                 dc.col_data[index][obs_index] =  chr(missing_tag)             # <<<<<<<<<<<<<<
 *             curset = dc.missing_user_values.get(index)
 *             if curset is None:
 */
        __pyx_t_6 = __Pyx_PyInt_From_int(__pyx_v_missing_tag); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 663, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_chr, __pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 663, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        if (unlikely(__pyx_v_dc->col_data == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
          __PYX_ERR(0, 663, __pyx_L1_error)
        }
        __pyx_t_6 = __Pyx_GetItemInt_List(__pyx_v_dc->col_data, __pyx_v_index, int, 1, __Pyx_PyInt_From_int, 1, 1, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 663, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        if (unlikely((__Pyx_SetItemInt(__pyx_t_6, __pyx_v_obs_index, __pyx_t_1, int, 1, __Pyx_PyInt_From_int, 0, 1, 1) < 0))) __PYX_ERR(0, 663, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      }
      __pyx_L16:;

      /* "pyreadstat/_readstat_parser.pyx":664
 *                 dc.col_data[index] = dc.col_data[index].astype(object, copy=False)
 *                 dc.col_data[index][obs_index] =  chr(missing_tag)
 *             curset = dc.missing_user_values.get(index)             # <<<<<<<<<<<<<<
 *             if curset is None:
 *                 curset = set()
 */
      if (unlikely(__pyx_v_dc->missing_user_values == Py_None)) {
        PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "get");
        __PYX_ERR(0, 664, __pyx_L1_error)
      }
      __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_index); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 664, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_6 = __Pyx_PyDict_GetItemDefault(__pyx_v_dc->missing_user_values, __pyx_t_1, Py_None); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 664, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (!(likely(PySet_CheckExact(__pyx_t_6))||((__pyx_t_6) == Py_None) || __Pyx_RaiseUnexpectedTypeError("set", __pyx_t_6))) __PYX_ERR(0, 664, __pyx_L1_error)
      __pyx_v_curset = ((PyObject*)__pyx_t_6);
      __pyx_t_6 = 0;

      /* "pyreadstat/_readstat_parser.pyx":665
 *                 dc.col_data[index][obs_index] =  chr(missing_tag)
 *             curset = dc.missing_user_values.get(index)
 *             if curset is None:             # <<<<<<<<<<<<<<
 *                 curset = set()
 *             curset.add(chr(missing_tag))
 */
      __pyx_t_3 = (__pyx_v_curset == ((PyObject*)Py_None));
      if (__pyx_t_3) {

        /* "pyreadstat/_readstat_parser.pyx":666
 *             curset = dc.missing_user_values.get(index)
 *             if curset is None:
 *                 curset = set()             # <<<<<<<<<<<<<<
 *             curset.add(chr(missing_tag))
 *             dc.missing_user_values[index] = curset
 */
        __pyx_t_6 = PySet_New(0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 666, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF_SET(__pyx_v_curset, ((PyObject*)__pyx_t_6));
        __pyx_t_6 = 0;

        /* "pyreadstat/_readstat_parser.pyx":665
 *                 dc.col_data[index][obs_index] =  chr(missing_tag)
 *             curset = dc.missing_user_values.get(index)
 *             if curset is None:             # <<<<<<<<<<<<<<
 *                 curset = set()
 *             curset.add(chr(missing_tag))
 */
      }

      /* "pyreadstat/_readstat_parser.pyx":667
 *             if curset is None:
 *                 curset = set()
 *             curset.add(chr(missing_tag))             # <<<<<<<<<<<<<<
 *             dc.missing_user_values[index] = curset
 *     else:
 */
      if (unlikely(__pyx_v_curset == Py_None)) {
        PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "add");
        __PYX_ERR(0, 667, __pyx_L1_error)
      }
      __pyx_t_6 = __Pyx_PyInt_From_int(__pyx_v_missing_tag); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 667, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_chr, __pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 667, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_8 = PySet_Add(__pyx_v_curset, __pyx_t_1); if (unlikely(__pyx_t_8 == ((int)-1))) __PYX_ERR(0, 667, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "pyreadstat/_readstat_parser.pyx":668
 *                 curset = set()
 *             curset.add(chr(missing_tag))
 *             dc.missing_user_values[index] = curset             # <<<<<<<<<<<<<<
 *     else:
 *         pyvalue = convert_readstat_to_python_value(value, index, dc)
 */
      if (unlikely(__pyx_v_dc->missing_user_values == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 668, __pyx_L1_error)
      }
      __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_index); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 668, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (unlikely((PyDict_SetItem(__pyx_v_dc->missing_user_values, __pyx_t_1, __pyx_v_curset) < 0))) __PYX_ERR(0, 668, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "pyreadstat/_readstat_parser.pyx":649
 *             pyvalue = convert_readstat_to_python_value(value, index, dc)
 *             dc.col_data[index][obs_index] = pyvalue
 *         elif readstat_value_is_tagged_missing(value):             # <<<<<<<<<<<<<<
 *             iscurnptypeobject = dc.col_dtypes_isobject[index]
 *             # SAS and Stata missing values
 */
    }
    __pyx_L10:;

    /* "pyreadstat/_readstat_parser.pyx":630
 * 
 *     # transform to python value types
 *     if readstat_value_is_missing(value, variable):             # <<<<<<<<<<<<<<
 *         # The user does not want to retrieve missing values
 *         if not dc.usernan or readstat_value_is_system_missing(value):
 */
    goto __pyx_L9;
  }

  /* "pyreadstat/_readstat_parser.pyx":670
 *             dc.missing_user_values[index] = curset
 *     else:
 *         pyvalue = convert_readstat_to_python_value(value, index, dc)             # <<<<<<<<<<<<<<
 *         dc.col_data[index][obs_index] = pyvalue
 * 
 */
  /*else*/ {
    __pyx_t_1 = __pyx_f_10pyreadstat_16_readstat_parser_convert_readstat_to_python_value(__pyx_v_value, __pyx_v_index, __pyx_v_dc); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 670, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_pyvalue = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "pyreadstat/_readstat_parser.pyx":671
 *     else:
 *         pyvalue = convert_readstat_to_python_value(value, index, dc)
 *         dc.col_data[index][obs_index] = pyvalue             # <<<<<<<<<<<<<<
 * 
 *     return READSTAT_HANDLER_OK
 */
    if (unlikely(__pyx_v_dc->col_data == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 671, __pyx_L1_error)
    }
    __pyx_t_1 = __Pyx_GetItemInt_List(__pyx_v_dc->col_data, __pyx_v_index, int, 1, __Pyx_PyInt_From_int, 1, 1, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 671, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (unlikely((__Pyx_SetItemInt(__pyx_t_1, __pyx_v_obs_index, __pyx_v_pyvalue, int, 1, __Pyx_PyInt_From_int, 0, 1, 1) < 0))) __PYX_ERR(0, 671, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }
  __pyx_L9:;

  /* "pyreadstat/_readstat_parser.pyx":673
 *         dc.col_data[index][obs_index] = pyvalue
 * 
 *     return READSTAT_HANDLER_OK             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = READSTAT_HANDLER_OK;
  goto __pyx_L0;

  /* "pyreadstat/_readstat_parser.pyx":585
 * 
 * 
 * cdef int handle_value(int obs_index, readstat_variable_t * variable, readstat_value_t value, void *ctx) except READSTAT_HANDLER_ABORT:             # <<<<<<<<<<<<<<
 *     """
 *     This function transforms every value to python types, and to datetime if
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pyreadstat._readstat_parser.handle_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = READSTAT_HANDLER_ABORT;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_dc);
  __Pyx_XDECREF(__pyx_v_buf_list);
  __Pyx_XDECREF(__pyx_v_pyvalue);
  __Pyx_XDECREF(__pyx_v_curset);
  __Pyx_XDECREF(__pyx_v_curnptype);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyreadstat/_readstat_parser.pyx":676
 * 
 * 
 * cdef int handle_value_label(char *val_labels, readstat_value_t value, char *label, void *ctx) except READSTAT_HANDLER_ABORT:             # <<<<<<<<<<<<<<
 *     """
 *     Reads the label for the value that belongs to the label set val_labels. In Handle variable we need to do a map
 */

static int __pyx_f_10pyreadstat_16_readstat_parser_handle_value_label(char *__pyx_v_val_labels, readstat_value_t __pyx_v_value, char *__pyx_v_label, void *__pyx_v_ctx) {
  struct __pyx_obj_10pyreadstat_16_readstat_parser_data_container *__pyx_v_dc = 0;
  char *__pyx_v_c_str_value;
  PyObject *__pyx_v_py_str_value = 0;
  int8_t __pyx_v_c_int8_value;
  int16_t __pyx_v_c_int16_value;
  int32_t __pyx_v_c_int32_value;
  float __pyx_v_c_float_value;
  double __pyx_v_c_double_value;
  long __pyx_v_py_long_value;
  double __pyx_v_py_float_value;
  __pyx_t_10pyreadstat_16_readstat_parser_py_variable_format __pyx_v_pyformat;
  PyObject *__pyx_v_labels_raw = 0;
  PyObject *__pyx_v_var_label = 0;
  PyObject *__pyx_v_value_label_name = 0;
  readstat_type_t __pyx_v_value_type;
  PyObject *__pyx_v_cur_dict = NULL;
  int __pyx_v_missing_tag;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("handle_value_label", 0);

  /* "pyreadstat/_readstat_parser.pyx":682
 *     """
 * 
 *     cdef  data_container dc = <data_container> ctx             # <<<<<<<<<<<<<<
 * 
 *     cdef char * c_str_value
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_ctx);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_dc = ((struct __pyx_obj_10pyreadstat_16_readstat_parser_data_container *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyreadstat/_readstat_parser.pyx":700
 *     cdef str value_label_name
 * 
 *     var_label = <str> val_labels             # <<<<<<<<<<<<<<
 *     value_label_name = <str> label
 * 
 */
  __pyx_t_1 = __Pyx_PyStr_FromString(__pyx_v_val_labels); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 700, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __pyx_t_1;
  __Pyx_INCREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_var_label = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "pyreadstat/_readstat_parser.pyx":701
 * 
 *     var_label = <str> val_labels
 *     value_label_name = <str> label             # <<<<<<<<<<<<<<
 * 
 *     cdef readstat_type_t value_type
 */
  __pyx_t_2 = __Pyx_PyStr_FromString(__pyx_v_label); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 701, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __pyx_t_2;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_value_label_name = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyreadstat/_readstat_parser.pyx":704
 * 
 *     cdef readstat_type_t value_type
 *     value_type = readstat_value_type(value)             # <<<<<<<<<<<<<<
 * 
 *     labels_raw = dc.labels_raw
 */
  __pyx_v_value_type = readstat_value_type(__pyx_v_value);

  /* "pyreadstat/_readstat_parser.pyx":706
 *     value_type = readstat_value_type(value)
 * 
 *     labels_raw = dc.labels_raw             # <<<<<<<<<<<<<<
 *     cur_dict = labels_raw.get(var_label)
 *     if not cur_dict:
 */
  __pyx_t_1 = __pyx_v_dc->labels_raw;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_labels_raw = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pyreadstat/_readstat_parser.pyx":707
 * 
 *     labels_raw = dc.labels_raw
 *     cur_dict = labels_raw.get(var_label)             # <<<<<<<<<<<<<<
 *     if not cur_dict:
 *         cur_dict = dict()
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_labels_raw, __pyx_n_s_get); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 707, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_var_label};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 707, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_v_cur_dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pyreadstat/_readstat_parser.pyx":708
 *     labels_raw = dc.labels_raw
 *     cur_dict = labels_raw.get(var_label)
 *     if not cur_dict:             # <<<<<<<<<<<<<<
 *         cur_dict = dict()
 * 
 */
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_v_cur_dict); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 708, __pyx_L1_error)
  __pyx_t_6 = (!__pyx_t_5);
  if (__pyx_t_6) {

    /* "pyreadstat/_readstat_parser.pyx":709
 *     cur_dict = labels_raw.get(var_label)
 *     if not cur_dict:
 *         cur_dict = dict()             # <<<<<<<<<<<<<<
 * 
 *     if readstat_value_is_tagged_missing(value):
 */
    __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 709, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF_SET(__pyx_v_cur_dict, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "pyreadstat/_readstat_parser.pyx":708
 *     labels_raw = dc.labels_raw
 *     cur_dict = labels_raw.get(var_label)
 *     if not cur_dict:             # <<<<<<<<<<<<<<
 *         cur_dict = dict()
 * 
 */
  }

  /* "pyreadstat/_readstat_parser.pyx":711
 *         cur_dict = dict()
 * 
 *     if readstat_value_is_tagged_missing(value):             # <<<<<<<<<<<<<<
 *         # SAS and Stata missing values
 *         missing_tag = <int> readstat_value_tag(value)
 */
  __pyx_t_6 = (readstat_value_is_tagged_missing(__pyx_v_value) != 0);
  if (__pyx_t_6) {

    /* "pyreadstat/_readstat_parser.pyx":713
 *     if readstat_value_is_tagged_missing(value):
 *         # SAS and Stata missing values
 *         missing_tag = <int> readstat_value_tag(value)             # <<<<<<<<<<<<<<
 *         # In SAS missing values are A to Z or _ in stata a to z
 *         cur_dict[chr(missing_tag)] = value_label_name
 */
    __pyx_v_missing_tag = ((int)readstat_value_tag(__pyx_v_value));

    /* "pyreadstat/_readstat_parser.pyx":715
 *         missing_tag = <int> readstat_value_tag(value)
 *         # In SAS missing values are A to Z or _ in stata a to z
 *         cur_dict[chr(missing_tag)] = value_label_name             # <<<<<<<<<<<<<<
 *     else:
 * 
 */
    __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_missing_tag); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 715, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_chr, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 715, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely((PyObject_SetItem(__pyx_v_cur_dict, __pyx_t_2, __pyx_v_value_label_name) < 0))) __PYX_ERR(0, 715, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "pyreadstat/_readstat_parser.pyx":711
 *         cur_dict = dict()
 * 
 *     if readstat_value_is_tagged_missing(value):             # <<<<<<<<<<<<<<
 *         # SAS and Stata missing values
 *         missing_tag = <int> readstat_value_tag(value)
 */
    goto __pyx_L4;
  }

  /* "pyreadstat/_readstat_parser.pyx":718
 *     else:
 * 
 *         if value_type == READSTAT_TYPE_STRING or value_type == READSTAT_TYPE_STRING_REF:             # <<<<<<<<<<<<<<
 *             c_str_value = readstat_string_value(value)
 *             py_str_value = <str> c_str_value
 */
  /*else*/ {

    /* "pyreadstat/_readstat_parser.pyx":738
 *             py_float_value = <double> c_float_value
 *             pyformat = VAR_FORMAT_FLOAT
 *         elif value_type == READSTAT_TYPE_DOUBLE:             # <<<<<<<<<<<<<<
 *             c_double_value = readstat_double_value(value);
 *             py_float_value = <double> c_double_value
 */
    switch (__pyx_v_value_type) {
      case READSTAT_TYPE_STRING:

      /* "pyreadstat/_readstat_parser.pyx":718
 *     else:
 * 
 *         if value_type == READSTAT_TYPE_STRING or value_type == READSTAT_TYPE_STRING_REF:             # <<<<<<<<<<<<<<
 *             c_str_value = readstat_string_value(value)
 *             py_str_value = <str> c_str_value
 */
      case READSTAT_TYPE_STRING_REF:

      /* "pyreadstat/_readstat_parser.pyx":719
 * 
 *         if value_type == READSTAT_TYPE_STRING or value_type == READSTAT_TYPE_STRING_REF:
 *             c_str_value = readstat_string_value(value)             # <<<<<<<<<<<<<<
 *             py_str_value = <str> c_str_value
 *             pyformat = VAR_FORMAT_STRING
 */
      __pyx_v_c_str_value = readstat_string_value(__pyx_v_value);

      /* "pyreadstat/_readstat_parser.pyx":720
 *         if value_type == READSTAT_TYPE_STRING or value_type == READSTAT_TYPE_STRING_REF:
 *             c_str_value = readstat_string_value(value)
 *             py_str_value = <str> c_str_value             # <<<<<<<<<<<<<<
 *             pyformat = VAR_FORMAT_STRING
 *         elif value_type == READSTAT_TYPE_INT8:
 */
      __pyx_t_2 = __Pyx_PyStr_FromString(__pyx_v_c_str_value); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 720, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_1 = __pyx_t_2;
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_v_py_str_value = ((PyObject*)__pyx_t_1);
      __pyx_t_1 = 0;

      /* "pyreadstat/_readstat_parser.pyx":721
 *             c_str_value = readstat_string_value(value)
 *             py_str_value = <str> c_str_value
 *             pyformat = VAR_FORMAT_STRING             # <<<<<<<<<<<<<<
 *         elif value_type == READSTAT_TYPE_INT8:
 *             c_int8_value = readstat_int8_value(value)
 */
      __pyx_v_pyformat = __pyx_e_10pyreadstat_16_readstat_parser_VAR_FORMAT_STRING;

      /* "pyreadstat/_readstat_parser.pyx":718
 *     else:
 * 
 *         if value_type == READSTAT_TYPE_STRING or value_type == READSTAT_TYPE_STRING_REF:             # <<<<<<<<<<<<<<
 *             c_str_value = readstat_string_value(value)
 *             py_str_value = <str> c_str_value
 */
      break;
      case READSTAT_TYPE_INT8:

      /* "pyreadstat/_readstat_parser.pyx":723
 *             pyformat = VAR_FORMAT_STRING
 *         elif value_type == READSTAT_TYPE_INT8:
 *             c_int8_value = readstat_int8_value(value)             # <<<<<<<<<<<<<<
 *             py_long_value = <long> c_int8_value
 *             pyformat = VAR_FORMAT_LONG
 */
      __pyx_v_c_int8_value = readstat_int8_value(__pyx_v_value);

      /* "pyreadstat/_readstat_parser.pyx":724
 *         elif value_type == READSTAT_TYPE_INT8:
 *             c_int8_value = readstat_int8_value(value)
 *             py_long_value = <long> c_int8_value             # <<<<<<<<<<<<<<
 *             pyformat = VAR_FORMAT_LONG
 *         elif value_type == READSTAT_TYPE_INT16:
 */
      __pyx_v_py_long_value = ((long)__pyx_v_c_int8_value);

      /* "pyreadstat/_readstat_parser.pyx":725
 *             c_int8_value = readstat_int8_value(value)
 *             py_long_value = <long> c_int8_value
 *             pyformat = VAR_FORMAT_LONG             # <<<<<<<<<<<<<<
 *         elif value_type == READSTAT_TYPE_INT16:
 *             c_int16_value = readstat_int16_value(value)
 */
      __pyx_v_pyformat = __pyx_e_10pyreadstat_16_readstat_parser_VAR_FORMAT_LONG;

      /* "pyreadstat/_readstat_parser.pyx":722
 *             py_str_value = <str> c_str_value
 *             pyformat = VAR_FORMAT_STRING
 *         elif value_type == READSTAT_TYPE_INT8:             # <<<<<<<<<<<<<<
 *             c_int8_value = readstat_int8_value(value)
 *             py_long_value = <long> c_int8_value
 */
      break;
      case READSTAT_TYPE_INT16:

      /* "pyreadstat/_readstat_parser.pyx":727
 *             pyformat = VAR_FORMAT_LONG
 *         elif value_type == READSTAT_TYPE_INT16:
 *             c_int16_value = readstat_int16_value(value)             # <<<<<<<<<<<<<<
 *             py_long_value = <long> c_int16_value
 *             pyformat = VAR_FORMAT_LONG
 */
      __pyx_v_c_int16_value = readstat_int16_value(__pyx_v_value);

      /* "pyreadstat/_readstat_parser.pyx":728
 *         elif value_type == READSTAT_TYPE_INT16:
 *             c_int16_value = readstat_int16_value(value)
 *             py_long_value = <long> c_int16_value             # <<<<<<<<<<<<<<
 *             pyformat = VAR_FORMAT_LONG
 *         elif value_type == READSTAT_TYPE_INT32:
 */
      __pyx_v_py_long_value = ((long)__pyx_v_c_int16_value);

      /* "pyreadstat/_readstat_parser.pyx":729
 *             c_int16_value = readstat_int16_value(value)
 *             py_long_value = <long> c_int16_value
 *             pyformat = VAR_FORMAT_LONG             # <<<<<<<<<<<<<<
 *         elif value_type == READSTAT_TYPE_INT32:
 *             c_int32_value = readstat_int32_value(value)
 */
      __pyx_v_pyformat = __pyx_e_10pyreadstat_16_readstat_parser_VAR_FORMAT_LONG;

      /* "pyreadstat/_readstat_parser.pyx":726
 *             py_long_value = <long> c_int8_value
 *             pyformat = VAR_FORMAT_LONG
 *         elif value_type == READSTAT_TYPE_INT16:             # <<<<<<<<<<<<<<
 *             c_int16_value = readstat_int16_value(value)
 *             py_long_value = <long> c_int16_value
 */
      break;
      case READSTAT_TYPE_INT32:

      /* "pyreadstat/_readstat_parser.pyx":731
 *             pyformat = VAR_FORMAT_LONG
 *         elif value_type == READSTAT_TYPE_INT32:
 *             c_int32_value = readstat_int32_value(value)             # <<<<<<<<<<<<<<
 *             py_long_value = <long> c_int32_value
 *             pyformat = VAR_FORMAT_LONG
 */
      __pyx_v_c_int32_value = readstat_int32_value(__pyx_v_value);

      /* "pyreadstat/_readstat_parser.pyx":732
 *         elif value_type == READSTAT_TYPE_INT32:
 *             c_int32_value = readstat_int32_value(value)
 *             py_long_value = <long> c_int32_value             # <<<<<<<<<<<<<<
 *             pyformat = VAR_FORMAT_LONG
 *         elif value_type == READSTAT_TYPE_FLOAT:
 */
      __pyx_v_py_long_value = ((long)__pyx_v_c_int32_value);

      /* "pyreadstat/_readstat_parser.pyx":733
 *             c_int32_value = readstat_int32_value(value)
 *             py_long_value = <long> c_int32_value
 *             pyformat = VAR_FORMAT_LONG             # <<<<<<<<<<<<<<
 *         elif value_type == READSTAT_TYPE_FLOAT:
 *             c_float_value = readstat_float_value(value)
 */
      __pyx_v_pyformat = __pyx_e_10pyreadstat_16_readstat_parser_VAR_FORMAT_LONG;

      /* "pyreadstat/_readstat_parser.pyx":730
 *             py_long_value = <long> c_int16_value
 *             pyformat = VAR_FORMAT_LONG
 *         elif value_type == READSTAT_TYPE_INT32:             # <<<<<<<<<<<<<<
 *             c_int32_value = readstat_int32_value(value)
 *             py_long_value = <long> c_int32_value
 */
      break;
      case READSTAT_TYPE_FLOAT:

      /* "pyreadstat/_readstat_parser.pyx":735
 *             pyformat = VAR_FORMAT_LONG
 *         elif value_type == READSTAT_TYPE_FLOAT:
 *             c_float_value = readstat_float_value(value)             # <<<<<<<<<<<<<<
 *             py_float_value = <double> c_float_value
 *             pyformat = VAR_FORMAT_FLOAT
 */
      __pyx_v_c_float_value = readstat_float_value(__pyx_v_value);

      /* "pyreadstat/_readstat_parser.pyx":736
 *         elif value_type == READSTAT_TYPE_FLOAT:
 *             c_float_value = readstat_float_value(value)
 *             py_float_value = <double> c_float_value             # <<<<<<<<<<<<<<
 *             pyformat = VAR_FORMAT_FLOAT
 *         elif value_type == READSTAT_TYPE_DOUBLE:
 */
      __pyx_v_py_float_value = ((double)__pyx_v_c_float_value);

      /* "pyreadstat/_readstat_parser.pyx":737
 *             c_float_value = readstat_float_value(value)
 *             py_float_value = <double> c_float_value
 *             pyformat = VAR_FORMAT_FLOAT             # <<<<<<<<<<<<<<
 *         elif value_type == READSTAT_TYPE_DOUBLE:
 *             c_double_value = readstat_double_value(value);
 */
      __pyx_v_pyformat = __pyx_e_10pyreadstat_16_readstat_parser_VAR_FORMAT_FLOAT;

      /* "pyreadstat/_readstat_parser.pyx":734
 *             py_long_value = <long> c_int32_value
 *             pyformat = VAR_FORMAT_LONG
 *         elif value_type == READSTAT_TYPE_FLOAT:             # <<<<<<<<<<<<<<
 *             c_float_value = readstat_float_value(value)
 *             py_float_value = <double> c_float_value
 */
      break;
      case READSTAT_TYPE_DOUBLE:

      /* "pyreadstat/_readstat_parser.pyx":739
 *             pyformat = VAR_FORMAT_FLOAT
 *         elif value_type == READSTAT_TYPE_DOUBLE:
 *             c_double_value = readstat_double_value(value);             # <<<<<<<<<<<<<<
 *             py_float_value = <double> c_double_value
 *             pyformat = VAR_FORMAT_FLOAT
 */
      __pyx_v_c_double_value = readstat_double_value(__pyx_v_value);

      /* "pyreadstat/_readstat_parser.pyx":740
 *         elif value_type == READSTAT_TYPE_DOUBLE:
 *             c_double_value = readstat_double_value(value);
 *             py_float_value = <double> c_double_value             # <<<<<<<<<<<<<<
 *             pyformat = VAR_FORMAT_FLOAT
 *         else:
 */
      __pyx_v_py_float_value = ((double)__pyx_v_c_double_value);

      /* "pyreadstat/_readstat_parser.pyx":741
 *             c_double_value = readstat_double_value(value);
 *             py_float_value = <double> c_double_value
 *             pyformat = VAR_FORMAT_FLOAT             # <<<<<<<<<<<<<<
 *         else:
 *             raise PyreadstatError("Unkown data type")
 */
      __pyx_v_pyformat = __pyx_e_10pyreadstat_16_readstat_parser_VAR_FORMAT_FLOAT;

      /* "pyreadstat/_readstat_parser.pyx":738
 *             py_float_value = <double> c_float_value
 *             pyformat = VAR_FORMAT_FLOAT
 *         elif value_type == READSTAT_TYPE_DOUBLE:             # <<<<<<<<<<<<<<
 *             c_double_value = readstat_double_value(value);
 *             py_float_value = <double> c_double_value
 */
      break;
      default:

      /* "pyreadstat/_readstat_parser.pyx":743
 *             pyformat = VAR_FORMAT_FLOAT
 *         else:
 *             raise PyreadstatError("Unkown data type")             # <<<<<<<<<<<<<<
 * 
 * 
 */
      __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_PyreadstatError); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 743, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_3 = NULL;
      __pyx_t_4 = 0;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
        __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
        if (likely(__pyx_t_3)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_2, function);
          __pyx_t_4 = 1;
        }
      }
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_kp_s_Unkown_data_type};
        __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 743, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      }
      __Pyx_Raise(__pyx_t_1, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __PYX_ERR(0, 743, __pyx_L1_error)
      break;
    }

    /* "pyreadstat/_readstat_parser.pyx":746
 * 
 * 
 *         if pyformat == VAR_FORMAT_STRING:             # <<<<<<<<<<<<<<
 *             cur_dict[py_str_value] = value_label_name
 *         elif pyformat == VAR_FORMAT_LONG:
 */
    switch (__pyx_v_pyformat) {
      case __pyx_e_10pyreadstat_16_readstat_parser_VAR_FORMAT_STRING:

      /* "pyreadstat/_readstat_parser.pyx":747
 * 
 *         if pyformat == VAR_FORMAT_STRING:
 *             cur_dict[py_str_value] = value_label_name             # <<<<<<<<<<<<<<
 *         elif pyformat == VAR_FORMAT_LONG:
 *             cur_dict[py_long_value] = value_label_name
 */
      if (unlikely(!__pyx_v_py_str_value)) { __Pyx_RaiseUnboundLocalError("py_str_value"); __PYX_ERR(0, 747, __pyx_L1_error) }
      if (unlikely((PyObject_SetItem(__pyx_v_cur_dict, __pyx_v_py_str_value, __pyx_v_value_label_name) < 0))) __PYX_ERR(0, 747, __pyx_L1_error)

      /* "pyreadstat/_readstat_parser.pyx":746
 * 
 * 
 *         if pyformat == VAR_FORMAT_STRING:             # <<<<<<<<<<<<<<
 *             cur_dict[py_str_value] = value_label_name
 *         elif pyformat == VAR_FORMAT_LONG:
 */
      break;
      case __pyx_e_10pyreadstat_16_readstat_parser_VAR_FORMAT_LONG:

      /* "pyreadstat/_readstat_parser.pyx":749
 *             cur_dict[py_str_value] = value_label_name
 *         elif pyformat == VAR_FORMAT_LONG:
 *             cur_dict[py_long_value] = value_label_name             # <<<<<<<<<<<<<<
 *         elif pyformat == VAR_FORMAT_FLOAT:
 *             cur_dict[py_float_value] = value_label_name
 */
      if (unlikely((__Pyx_SetItemInt(__pyx_v_cur_dict, __pyx_v_py_long_value, __pyx_v_value_label_name, long, 1, __Pyx_PyInt_From_long, 0, 1, 1) < 0))) __PYX_ERR(0, 749, __pyx_L1_error)

      /* "pyreadstat/_readstat_parser.pyx":748
 *         if pyformat == VAR_FORMAT_STRING:
 *             cur_dict[py_str_value] = value_label_name
 *         elif pyformat == VAR_FORMAT_LONG:             # <<<<<<<<<<<<<<
 *             cur_dict[py_long_value] = value_label_name
 *         elif pyformat == VAR_FORMAT_FLOAT:
 */
      break;
      case __pyx_e_10pyreadstat_16_readstat_parser_VAR_FORMAT_FLOAT:

      /* "pyreadstat/_readstat_parser.pyx":751
 *             cur_dict[py_long_value] = value_label_name
 *         elif pyformat == VAR_FORMAT_FLOAT:
 *             cur_dict[py_float_value] = value_label_name             # <<<<<<<<<<<<<<
 *         elif pyformat == VAR_FORMAT_MISSING:
 *             pass
 */
      __pyx_t_1 = PyFloat_FromDouble(__pyx_v_py_float_value); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 751, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (unlikely((PyObject_SetItem(__pyx_v_cur_dict, __pyx_t_1, __pyx_v_value_label_name) < 0))) __PYX_ERR(0, 751, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "pyreadstat/_readstat_parser.pyx":750
 *         elif pyformat == VAR_FORMAT_LONG:
 *             cur_dict[py_long_value] = value_label_name
 *         elif pyformat == VAR_FORMAT_FLOAT:             # <<<<<<<<<<<<<<
 *             cur_dict[py_float_value] = value_label_name
 *         elif pyformat == VAR_FORMAT_MISSING:
 */
      break;
      case __pyx_e_10pyreadstat_16_readstat_parser_VAR_FORMAT_MISSING:

      /* "pyreadstat/_readstat_parser.pyx":752
 *         elif pyformat == VAR_FORMAT_FLOAT:
 *             cur_dict[py_float_value] = value_label_name
 *         elif pyformat == VAR_FORMAT_MISSING:             # <<<<<<<<<<<<<<
 *             pass
 *         else:
 */
      break;
      default:

      /* "pyreadstat/_readstat_parser.pyx":755
 *             pass
 *         else:
 *             raise PyreadstatError("Failed convert C to python value")             # <<<<<<<<<<<<<<
 * 
 *     dc.labels_raw[var_label] = cur_dict
 */
      __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_PyreadstatError); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 755, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_3 = NULL;
      __pyx_t_4 = 0;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
        __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
        if (likely(__pyx_t_3)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_2, function);
          __pyx_t_4 = 1;
        }
      }
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_kp_s_Failed_convert_C_to_python_value};
        __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 755, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      }
      __Pyx_Raise(__pyx_t_1, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __PYX_ERR(0, 755, __pyx_L1_error)
      break;
    }
  }
  __pyx_L4:;

  /* "pyreadstat/_readstat_parser.pyx":757
 *             raise PyreadstatError("Failed convert C to python value")
 * 
 *     dc.labels_raw[var_label] = cur_dict             # <<<<<<<<<<<<<<
 * 
 *     return READSTAT_HANDLER_OK
 */
  if (unlikely((PyObject_SetItem(__pyx_v_dc->labels_raw, __pyx_v_var_label, __pyx_v_cur_dict) < 0))) __PYX_ERR(0, 757, __pyx_L1_error)

  /* "pyreadstat/_readstat_parser.pyx":759
 *     dc.labels_raw[var_label] = cur_dict
 * 
 *     return READSTAT_HANDLER_OK             # <<<<<<<<<<<<<<
 * 
 * cdef int handle_note (int note_index, char *note, void *ctx) except READSTAT_HANDLER_ABORT:
 */
  __pyx_r = READSTAT_HANDLER_OK;
  goto __pyx_L0;

  /* "pyreadstat/_readstat_parser.pyx":676
 * 
 * 
 * cdef int handle_value_label(char *val_labels, readstat_value_t value, char *label, void *ctx) except READSTAT_HANDLER_ABORT:             # <<<<<<<<<<<<<<
 *     """
 *     Reads the label for the value that belongs to the label set val_labels. In Handle variable we need to do a map
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pyreadstat._readstat_parser.handle_value_label", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = READSTAT_HANDLER_ABORT;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_dc);
  __Pyx_XDECREF(__pyx_v_py_str_value);
  __Pyx_XDECREF(__pyx_v_labels_raw);
  __Pyx_XDECREF(__pyx_v_var_label);
  __Pyx_XDECREF(__pyx_v_value_label_name);
  __Pyx_XDECREF(__pyx_v_cur_dict);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyreadstat/_readstat_parser.pyx":761
 *     return READSTAT_HANDLER_OK
 * 
 * cdef int handle_note (int note_index, char *note, void *ctx) except READSTAT_HANDLER_ABORT:             # <<<<<<<<<<<<<<
 *     """
 *     Collects notes (text annotations) attached to the documents. It happens for spss and stata
 */

static int __pyx_f_10pyreadstat_16_readstat_parser_handle_note(CYTHON_UNUSED int __pyx_v_note_index, char *__pyx_v_note, void *__pyx_v_ctx) {
  PyObject *__pyx_v_pynote = 0;
  struct __pyx_obj_10pyreadstat_16_readstat_parser_data_container *__pyx_v_dc = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("handle_note", 0);

  /* "pyreadstat/_readstat_parser.pyx":767
 * 
 *     cdef str pynote
 *     cdef  data_container dc = <data_container> ctx             # <<<<<<<<<<<<<<
 * 
 *     pynote = <str> note
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_ctx);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_dc = ((struct __pyx_obj_10pyreadstat_16_readstat_parser_data_container *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyreadstat/_readstat_parser.pyx":769
 *     cdef  data_container dc = <data_container> ctx
 * 
 *     pynote = <str> note             # <<<<<<<<<<<<<<
 *     dc.notes.append(pynote)
 * 
 */
  __pyx_t_1 = __Pyx_PyStr_FromString(__pyx_v_note); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 769, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __pyx_t_1;
  __Pyx_INCREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_pynote = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "pyreadstat/_readstat_parser.pyx":770
 * 
 *     pynote = <str> note
 *     dc.notes.append(pynote)             # <<<<<<<<<<<<<<
 * 
 *     return READSTAT_HANDLER_OK
 */
  __pyx_t_3 = __Pyx_PyObject_Append(__pyx_v_dc->notes, __pyx_v_pynote); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 770, __pyx_L1_error)

  /* "pyreadstat/_readstat_parser.pyx":772
 *     dc.notes.append(pynote)
 * 
 *     return READSTAT_HANDLER_OK             # <<<<<<<<<<<<<<
 * 
 * cdef int handle_open(const char *u8_path, void *io_ctx) except READSTAT_HANDLER_ABORT:
 */
  __pyx_r = READSTAT_HANDLER_OK;
  goto __pyx_L0;

  /* "pyreadstat/_readstat_parser.pyx":761
 *     return READSTAT_HANDLER_OK
 * 
 * cdef int handle_note (int note_index, char *note, void *ctx) except READSTAT_HANDLER_ABORT:             # <<<<<<<<<<<<<<
 *     """
 *     Collects notes (text annotations) attached to the documents. It happens for spss and stata
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pyreadstat._readstat_parser.handle_note", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = READSTAT_HANDLER_ABORT;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_pynote);
  __Pyx_XDECREF((PyObject *)__pyx_v_dc);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyreadstat/_readstat_parser.pyx":774
 *     return READSTAT_HANDLER_OK
 * 
 * cdef int handle_open(const char *u8_path, void *io_ctx) except READSTAT_HANDLER_ABORT:             # <<<<<<<<<<<<<<
 *     """
 *     Special open handler for windows in order to be able to handle paths with international characters
 */

static int __pyx_f_10pyreadstat_16_readstat_parser_handle_open(char const *__pyx_v_u8_path, void *__pyx_v_io_ctx) {
  int __pyx_v_fd;
  Py_ssize_t __pyx_v_length;
  wchar_t *__pyx_v_u16_path;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  wchar_t *__pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("handle_open", 0);

  /* "pyreadstat/_readstat_parser.pyx":782
 *     cdef Py_ssize_t length
 * 
 *     if not os.path.isfile(u8_path):             # <<<<<<<<<<<<<<
 *         return -1
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_os); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 782, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_path); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 782, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_isfile); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 782, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyUnicode_FromString(__pyx_v_u8_path); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 782, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_3};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 782, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(0, 782, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_7 = (!__pyx_t_6);
  if (__pyx_t_7) {

    /* "pyreadstat/_readstat_parser.pyx":783
 * 
 *     if not os.path.isfile(u8_path):
 *         return -1             # <<<<<<<<<<<<<<
 * 
 *     if os.name == "nt":
 */
    __pyx_r = -1;
    goto __pyx_L0;

    /* "pyreadstat/_readstat_parser.pyx":782
 *     cdef Py_ssize_t length
 * 
 *     if not os.path.isfile(u8_path):             # <<<<<<<<<<<<<<
 *         return -1
 * 
 */
  }

  /* "pyreadstat/_readstat_parser.pyx":785
 *         return -1
 * 
 *     if os.name == "nt":             # <<<<<<<<<<<<<<
 *         u16_path = PyUnicode_AsWideCharString(u8_path, &length)
 *         fd = _wsopen(u16_path, _O_RDONLY | _O_BINARY, _SH_DENYWR, 0)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_os); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 785, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_name); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 785, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_7 = (__Pyx_PyString_Equals(__pyx_t_2, __pyx_n_s_nt, Py_EQ)); if (unlikely((__pyx_t_7 < 0))) __PYX_ERR(0, 785, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_7) {

    /* "pyreadstat/_readstat_parser.pyx":786
 * 
 *     if os.name == "nt":
 *         u16_path = PyUnicode_AsWideCharString(u8_path, &length)             # <<<<<<<<<<<<<<
 *         fd = _wsopen(u16_path, _O_RDONLY | _O_BINARY, _SH_DENYWR, 0)
 *         assign_fd(io_ctx, fd)
 */
    __pyx_t_2 = __Pyx_PyUnicode_FromString(__pyx_v_u8_path); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 786, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_8 = PyUnicode_AsWideCharString(__pyx_t_2, (&__pyx_v_length)); if (unlikely(__pyx_t_8 == ((wchar_t *)NULL))) __PYX_ERR(0, 786, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_u16_path = __pyx_t_8;

    /* "pyreadstat/_readstat_parser.pyx":787
 *     if os.name == "nt":
 *         u16_path = PyUnicode_AsWideCharString(u8_path, &length)
 *         fd = _wsopen(u16_path, _O_RDONLY | _O_BINARY, _SH_DENYWR, 0)             # <<<<<<<<<<<<<<
 *         assign_fd(io_ctx, fd)
 *         return fd
 */
    __pyx_v_fd = _wsopen(__pyx_v_u16_path, (_O_RDONLY | _O_BINARY), _SH_DENYWR, 0);

    /* "pyreadstat/_readstat_parser.pyx":788
 *         u16_path = PyUnicode_AsWideCharString(u8_path, &length)
 *         fd = _wsopen(u16_path, _O_RDONLY | _O_BINARY, _SH_DENYWR, 0)
 *         assign_fd(io_ctx, fd)             # <<<<<<<<<<<<<<
 *         return fd
 *     else:
 */
    assign_fd(__pyx_v_io_ctx, __pyx_v_fd);

    /* "pyreadstat/_readstat_parser.pyx":789
 *         fd = _wsopen(u16_path, _O_RDONLY | _O_BINARY, _SH_DENYWR, 0)
 *         assign_fd(io_ctx, fd)
 *         return fd             # <<<<<<<<<<<<<<
 *     else:
 *         return -1
 */
    __pyx_r = __pyx_v_fd;
    goto __pyx_L0;

    /* "pyreadstat/_readstat_parser.pyx":785
 *         return -1
 * 
 *     if os.name == "nt":             # <<<<<<<<<<<<<<
 *         u16_path = PyUnicode_AsWideCharString(u8_path, &length)
 *         fd = _wsopen(u16_path, _O_RDONLY | _O_BINARY, _SH_DENYWR, 0)
 */
  }

  /* "pyreadstat/_readstat_parser.pyx":791
 *         return fd
 *     else:
 *         return -1             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __pyx_r = -1;
    goto __pyx_L0;
  }

  /* "pyreadstat/_readstat_parser.pyx":774
 *     return READSTAT_HANDLER_OK
 * 
 * cdef int handle_open(const char *u8_path, void *io_ctx) except READSTAT_HANDLER_ABORT:             # <<<<<<<<<<<<<<
 *     """
 *     Special open handler for windows in order to be able to handle paths with international characters
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pyreadstat._readstat_parser.handle_open", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = READSTAT_HANDLER_ABORT;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyreadstat/_readstat_parser.pyx":794
 * 
 * 
 * cdef void check_exit_status(readstat_error_t retcode) except *:             # <<<<<<<<<<<<<<
 *     """
 *     transforms a readstat exit status to a python error if status is not READSTAT OK
 */

static void __pyx_f_10pyreadstat_16_readstat_parser_check_exit_status(readstat_error_t __pyx_v_retcode) {
  char *__pyx_v_err_readstat;
  PyObject *__pyx_v_err_message = 0;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("check_exit_status", 0);

  /* "pyreadstat/_readstat_parser.pyx":801
 *     cdef char * err_readstat
 *     cdef str err_message
 *     if retcode != READSTAT_OK:             # <<<<<<<<<<<<<<
 *         err_readstat = readstat_error_message(retcode)
 *         err_message = <str> err_readstat
 */
  __pyx_t_1 = (__pyx_v_retcode != READSTAT_OK);
  if (unlikely(__pyx_t_1)) {

    /* "pyreadstat/_readstat_parser.pyx":802
 *     cdef str err_message
 *     if retcode != READSTAT_OK:
 *         err_readstat = readstat_error_message(retcode)             # <<<<<<<<<<<<<<
 *         err_message = <str> err_readstat
 *         raise ReadstatError(err_message)
 */
    __pyx_v_err_readstat = readstat_error_message(__pyx_v_retcode);

    /* "pyreadstat/_readstat_parser.pyx":803
 *     if retcode != READSTAT_OK:
 *         err_readstat = readstat_error_message(retcode)
 *         err_message = <str> err_readstat             # <<<<<<<<<<<<<<
 *         raise ReadstatError(err_message)
 * 
 */
    __pyx_t_2 = __Pyx_PyStr_FromString(__pyx_v_err_readstat); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 803, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __pyx_t_2;
    __Pyx_INCREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_err_message = ((PyObject*)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "pyreadstat/_readstat_parser.pyx":804
 *         err_readstat = readstat_error_message(retcode)
 *         err_message = <str> err_readstat
 *         raise ReadstatError(err_message)             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_ReadstatError); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 804, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = NULL;
    __pyx_t_5 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
        __pyx_t_5 = 1;
      }
    }
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v_err_message};
      __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 804, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 804, __pyx_L1_error)

    /* "pyreadstat/_readstat_parser.pyx":801
 *     cdef char * err_readstat
 *     cdef str err_message
 *     if retcode != READSTAT_OK:             # <<<<<<<<<<<<<<
 *         err_readstat = readstat_error_message(retcode)
 *         err_message = <str> err_readstat
 */
  }

  /* "pyreadstat/_readstat_parser.pyx":794
 * 
 * 
 * cdef void check_exit_status(readstat_error_t retcode) except *:             # <<<<<<<<<<<<<<
 *     """
 *     transforms a readstat exit status to a python error if status is not READSTAT OK
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pyreadstat._readstat_parser.check_exit_status", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_err_message);
  __Pyx_RefNannyFinishContext();
}

/* "pyreadstat/_readstat_parser.pyx":807
 * 
 * 
 * cdef void run_readstat_parser(char * filename, data_container data, py_file_extension file_extension, long row_limit, long row_offset) except *:             # <<<<<<<<<<<<<<
 *     """
 *     Runs the parsing of the file by readstat library
 */

static void __pyx_f_10pyreadstat_16_readstat_parser_run_readstat_parser(char *__pyx_v_filename, struct __pyx_obj_10pyreadstat_16_readstat_parser_data_container *__pyx_v_data, __pyx_t_10pyreadstat_16_readstat_parser_py_file_extension __pyx_v_file_extension, long __pyx_v_row_limit, long __pyx_v_row_offset) {
  readstat_parser_t *__pyx_v_parser;
  readstat_error_t __pyx_v_error;
  readstat_metadata_handler __pyx_v_metadata_handler;
  readstat_variable_handler __pyx_v_variable_handler;
  readstat_value_handler __pyx_v_value_handler;
  readstat_value_label_handler __pyx_v_value_label_handler;
  readstat_note_handler __pyx_v_note_handler;
  void *__pyx_v_ctx;
  PyObject *__pyx_v_pyerr;
  int __pyx_v_metaonly;
  readstat_open_handler __pyx_v_open_handler;
  PyObject *__pyx_v_encoding_bytes = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  char *__pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("run_readstat_parser", 0);

  /* "pyreadstat/_readstat_parser.pyx":827
 *     cdef bytes encoding_byte
 * 
 *     metaonly = data.metaonly             # <<<<<<<<<<<<<<
 *     ctx = <void *>data
 * 
 */
  __pyx_t_1 = __pyx_v_data->metaonly;
  __pyx_v_metaonly = __pyx_t_1;

  /* "pyreadstat/_readstat_parser.pyx":828
 * 
 *     metaonly = data.metaonly
 *     ctx = <void *>data             # <<<<<<<<<<<<<<
 * 
 *     #readstat_error_t error = READSTAT_OK;
 */
  __pyx_v_ctx = ((void *)__pyx_v_data);

  /* "pyreadstat/_readstat_parser.pyx":831
 * 
 *     #readstat_error_t error = READSTAT_OK;
 *     parser = readstat_parser_init()             # <<<<<<<<<<<<<<
 *     metadata_handler = <readstat_metadata_handler> handle_metadata
 *     variable_handler = <readstat_variable_handler> handle_variable
 */
  __pyx_v_parser = readstat_parser_init();

  /* "pyreadstat/_readstat_parser.pyx":832
 *     #readstat_error_t error = READSTAT_OK;
 *     parser = readstat_parser_init()
 *     metadata_handler = <readstat_metadata_handler> handle_metadata             # <<<<<<<<<<<<<<
 *     variable_handler = <readstat_variable_handler> handle_variable
 *     value_handler = <readstat_value_handler> handle_value
 */
  __pyx_v_metadata_handler = ((readstat_metadata_handler)__pyx_f_10pyreadstat_16_readstat_parser_handle_metadata);

  /* "pyreadstat/_readstat_parser.pyx":833
 *     parser = readstat_parser_init()
 *     metadata_handler = <readstat_metadata_handler> handle_metadata
 *     variable_handler = <readstat_variable_handler> handle_variable             # <<<<<<<<<<<<<<
 *     value_handler = <readstat_value_handler> handle_value
 *     value_label_handler = <readstat_value_label_handler> handle_value_label
 */
  __pyx_v_variable_handler = ((readstat_variable_handler)__pyx_f_10pyreadstat_16_readstat_parser_handle_variable);

  /* "pyreadstat/_readstat_parser.pyx":834
 *     metadata_handler = <readstat_metadata_handler> handle_metadata
 *     variable_handler = <readstat_variable_handler> handle_variable
 *     value_handler = <readstat_value_handler> handle_value             # <<<<<<<<<<<<<<
 *     value_label_handler = <readstat_value_label_handler> handle_value_label
 *     note_handler = <readstat_note_handler> handle_note
 */
  __pyx_v_value_handler = ((readstat_value_handler)__pyx_f_10pyreadstat_16_readstat_parser_handle_value);

  /* "pyreadstat/_readstat_parser.pyx":835
 *     variable_handler = <readstat_variable_handler> handle_variable
 *     value_handler = <readstat_value_handler> handle_value
 *     value_label_handler = <readstat_value_label_handler> handle_value_label             # <<<<<<<<<<<<<<
 *     note_handler = <readstat_note_handler> handle_note
 * 
 */
  __pyx_v_value_label_handler = ((readstat_value_label_handler)__pyx_f_10pyreadstat_16_readstat_parser_handle_value_label);

  /* "pyreadstat/_readstat_parser.pyx":836
 *     value_handler = <readstat_value_handler> handle_value
 *     value_label_handler = <readstat_value_label_handler> handle_value_label
 *     note_handler = <readstat_note_handler> handle_note             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_v_note_handler = ((readstat_note_handler)__pyx_f_10pyreadstat_16_readstat_parser_handle_note);

  /* "pyreadstat/_readstat_parser.pyx":839
 * 
 * 
 *     check_exit_status(readstat_set_metadata_handler(parser, metadata_handler))             # <<<<<<<<<<<<<<
 *     check_exit_status(readstat_set_variable_handler(parser, variable_handler))
 *     check_exit_status(readstat_set_value_label_handler(parser, value_label_handler))
 */
  __pyx_f_10pyreadstat_16_readstat_parser_check_exit_status(readstat_set_metadata_handler(__pyx_v_parser, __pyx_v_metadata_handler)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 839, __pyx_L1_error)

  /* "pyreadstat/_readstat_parser.pyx":840
 * 
 *     check_exit_status(readstat_set_metadata_handler(parser, metadata_handler))
 *     check_exit_status(readstat_set_variable_handler(parser, variable_handler))             # <<<<<<<<<<<<<<
 *     check_exit_status(readstat_set_value_label_handler(parser, value_label_handler))
 *     check_exit_status(readstat_set_note_handler(parser, note_handler))
 */
  __pyx_f_10pyreadstat_16_readstat_parser_check_exit_status(readstat_set_variable_handler(__pyx_v_parser, __pyx_v_variable_handler)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 840, __pyx_L1_error)

  /* "pyreadstat/_readstat_parser.pyx":841
 *     check_exit_status(readstat_set_metadata_handler(parser, metadata_handler))
 *     check_exit_status(readstat_set_variable_handler(parser, variable_handler))
 *     check_exit_status(readstat_set_value_label_handler(parser, value_label_handler))             # <<<<<<<<<<<<<<
 *     check_exit_status(readstat_set_note_handler(parser, note_handler))
 * 
 */
  __pyx_f_10pyreadstat_16_readstat_parser_check_exit_status(readstat_set_value_label_handler(__pyx_v_parser, __pyx_v_value_label_handler)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 841, __pyx_L1_error)

  /* "pyreadstat/_readstat_parser.pyx":842
 *     check_exit_status(readstat_set_variable_handler(parser, variable_handler))
 *     check_exit_status(readstat_set_value_label_handler(parser, value_label_handler))
 *     check_exit_status(readstat_set_note_handler(parser, note_handler))             # <<<<<<<<<<<<<<
 * 
 *     # on windows we need a custom open handler in order to deal with internation characters in the path.
 */
  __pyx_f_10pyreadstat_16_readstat_parser_check_exit_status(readstat_set_note_handler(__pyx_v_parser, __pyx_v_note_handler)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 842, __pyx_L1_error)

  /* "pyreadstat/_readstat_parser.pyx":845
 * 
 *     # on windows we need a custom open handler in order to deal with internation characters in the path.
 *     if os.name == "nt":             # <<<<<<<<<<<<<<
 *         open_handler = <readstat_open_handler> handle_open
 *         readstat_set_open_handler(parser, open_handler)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_os); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 845, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_name); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 845, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_1 = (__Pyx_PyString_Equals(__pyx_t_3, __pyx_n_s_nt, Py_EQ)); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 845, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_1) {

    /* "pyreadstat/_readstat_parser.pyx":846
 *     # on windows we need a custom open handler in order to deal with internation characters in the path.
 *     if os.name == "nt":
 *         open_handler = <readstat_open_handler> handle_open             # <<<<<<<<<<<<<<
 *         readstat_set_open_handler(parser, open_handler)
 * 
 */
    __pyx_v_open_handler = ((readstat_open_handler)__pyx_f_10pyreadstat_16_readstat_parser_handle_open);

    /* "pyreadstat/_readstat_parser.pyx":847
 *     if os.name == "nt":
 *         open_handler = <readstat_open_handler> handle_open
 *         readstat_set_open_handler(parser, open_handler)             # <<<<<<<<<<<<<<
 * 
 *     if not metaonly:
 */
    (void)(readstat_set_open_handler(__pyx_v_parser, __pyx_v_open_handler));

    /* "pyreadstat/_readstat_parser.pyx":845
 * 
 *     # on windows we need a custom open handler in order to deal with internation characters in the path.
 *     if os.name == "nt":             # <<<<<<<<<<<<<<
 *         open_handler = <readstat_open_handler> handle_open
 *         readstat_set_open_handler(parser, open_handler)
 */
  }

  /* "pyreadstat/_readstat_parser.pyx":849
 *         readstat_set_open_handler(parser, open_handler)
 * 
 *     if not metaonly:             # <<<<<<<<<<<<<<
 *         check_exit_status(readstat_set_value_handler(parser, value_handler))
 * 
 */
  __pyx_t_1 = (!__pyx_v_metaonly);
  if (__pyx_t_1) {

    /* "pyreadstat/_readstat_parser.pyx":850
 * 
 *     if not metaonly:
 *         check_exit_status(readstat_set_value_handler(parser, value_handler))             # <<<<<<<<<<<<<<
 * 
 *     # if the user set the encoding manually
 */
    __pyx_f_10pyreadstat_16_readstat_parser_check_exit_status(readstat_set_value_handler(__pyx_v_parser, __pyx_v_value_handler)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 850, __pyx_L1_error)

    /* "pyreadstat/_readstat_parser.pyx":849
 *         readstat_set_open_handler(parser, open_handler)
 * 
 *     if not metaonly:             # <<<<<<<<<<<<<<
 *         check_exit_status(readstat_set_value_handler(parser, value_handler))
 * 
 */
  }

  /* "pyreadstat/_readstat_parser.pyx":853
 * 
 *     # if the user set the encoding manually
 *     if data.user_encoding:             # <<<<<<<<<<<<<<
 *         encoding_bytes = data.user_encoding.encode("utf-8")
 *         readstat_set_file_character_encoding(parser, <char *> encoding_bytes)
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_data->user_encoding); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 853, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "pyreadstat/_readstat_parser.pyx":854
 *     # if the user set the encoding manually
 *     if data.user_encoding:
 *         encoding_bytes = data.user_encoding.encode("utf-8")             # <<<<<<<<<<<<<<
 *         readstat_set_file_character_encoding(parser, <char *> encoding_bytes)
 * 
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_data->user_encoding, __pyx_n_s_encode); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 854, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = NULL;
    __pyx_t_5 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
        __pyx_t_5 = 1;
      }
    }
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_kp_s_utf_8};
      __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 854, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
    __pyx_v_encoding_bytes = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "pyreadstat/_readstat_parser.pyx":855
 *     if data.user_encoding:
 *         encoding_bytes = data.user_encoding.encode("utf-8")
 *         readstat_set_file_character_encoding(parser, <char *> encoding_bytes)             # <<<<<<<<<<<<<<
 * 
 *     if row_limit:
 */
    __pyx_t_6 = __Pyx_PyObject_AsWritableString(__pyx_v_encoding_bytes); if (unlikely((!__pyx_t_6) && PyErr_Occurred())) __PYX_ERR(0, 855, __pyx_L1_error)
    (void)(readstat_set_file_character_encoding(__pyx_v_parser, ((char *)__pyx_t_6)));

    /* "pyreadstat/_readstat_parser.pyx":853
 * 
 *     # if the user set the encoding manually
 *     if data.user_encoding:             # <<<<<<<<<<<<<<
 *         encoding_bytes = data.user_encoding.encode("utf-8")
 *         readstat_set_file_character_encoding(parser, <char *> encoding_bytes)
 */
  }

  /* "pyreadstat/_readstat_parser.pyx":857
 *         readstat_set_file_character_encoding(parser, <char *> encoding_bytes)
 * 
 *     if row_limit:             # <<<<<<<<<<<<<<
 *         check_exit_status(readstat_set_row_limit(parser, row_limit))
 * 
 */
  __pyx_t_1 = (__pyx_v_row_limit != 0);
  if (__pyx_t_1) {

    /* "pyreadstat/_readstat_parser.pyx":858
 * 
 *     if row_limit:
 *         check_exit_status(readstat_set_row_limit(parser, row_limit))             # <<<<<<<<<<<<<<
 * 
 *     if row_offset:
 */
    __pyx_f_10pyreadstat_16_readstat_parser_check_exit_status(readstat_set_row_limit(__pyx_v_parser, __pyx_v_row_limit)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 858, __pyx_L1_error)

    /* "pyreadstat/_readstat_parser.pyx":857
 *         readstat_set_file_character_encoding(parser, <char *> encoding_bytes)
 * 
 *     if row_limit:             # <<<<<<<<<<<<<<
 *         check_exit_status(readstat_set_row_limit(parser, row_limit))
 * 
 */
  }

  /* "pyreadstat/_readstat_parser.pyx":860
 *         check_exit_status(readstat_set_row_limit(parser, row_limit))
 * 
 *     if row_offset:             # <<<<<<<<<<<<<<
 *         check_exit_status(readstat_set_row_offset(parser, row_offset))
 * 
 */
  __pyx_t_1 = (__pyx_v_row_offset != 0);
  if (__pyx_t_1) {

    /* "pyreadstat/_readstat_parser.pyx":861
 * 
 *     if row_offset:
 *         check_exit_status(readstat_set_row_offset(parser, row_offset))             # <<<<<<<<<<<<<<
 * 
 *     # parse!
 */
    __pyx_f_10pyreadstat_16_readstat_parser_check_exit_status(readstat_set_row_offset(__pyx_v_parser, __pyx_v_row_offset)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 861, __pyx_L1_error)

    /* "pyreadstat/_readstat_parser.pyx":860
 *         check_exit_status(readstat_set_row_limit(parser, row_limit))
 * 
 *     if row_offset:             # <<<<<<<<<<<<<<
 *         check_exit_status(readstat_set_row_offset(parser, row_offset))
 * 
 */
  }

  /* "pyreadstat/_readstat_parser.pyx":864
 * 
 *     # parse!
 *     if file_extension == FILE_EXT_SAV:             # <<<<<<<<<<<<<<
 *         error = readstat_parse_sav(parser, filename, ctx);
 *     elif file_extension == FILE_EXT_SAS7BDAT:
 */
  switch (__pyx_v_file_extension) {
    case __pyx_e_10pyreadstat_16_readstat_parser_FILE_EXT_SAV:

    /* "pyreadstat/_readstat_parser.pyx":865
 *     # parse!
 *     if file_extension == FILE_EXT_SAV:
 *         error = readstat_parse_sav(parser, filename, ctx);             # <<<<<<<<<<<<<<
 *     elif file_extension == FILE_EXT_SAS7BDAT:
 *         error = readstat_parse_sas7bdat(parser, filename, ctx);
 */
    __pyx_v_error = readstat_parse_sav(__pyx_v_parser, __pyx_v_filename, __pyx_v_ctx);

    /* "pyreadstat/_readstat_parser.pyx":864
 * 
 *     # parse!
 *     if file_extension == FILE_EXT_SAV:             # <<<<<<<<<<<<<<
 *         error = readstat_parse_sav(parser, filename, ctx);
 *     elif file_extension == FILE_EXT_SAS7BDAT:
 */
    break;
    case __pyx_e_10pyreadstat_16_readstat_parser_FILE_EXT_SAS7BDAT:

    /* "pyreadstat/_readstat_parser.pyx":867
 *         error = readstat_parse_sav(parser, filename, ctx);
 *     elif file_extension == FILE_EXT_SAS7BDAT:
 *         error = readstat_parse_sas7bdat(parser, filename, ctx);             # <<<<<<<<<<<<<<
 *     elif file_extension == FILE_EXT_DTA:
 *         error = readstat_parse_dta(parser, filename, ctx);
 */
    __pyx_v_error = readstat_parse_sas7bdat(__pyx_v_parser, __pyx_v_filename, __pyx_v_ctx);

    /* "pyreadstat/_readstat_parser.pyx":866
 *     if file_extension == FILE_EXT_SAV:
 *         error = readstat_parse_sav(parser, filename, ctx);
 *     elif file_extension == FILE_EXT_SAS7BDAT:             # <<<<<<<<<<<<<<
 *         error = readstat_parse_sas7bdat(parser, filename, ctx);
 *     elif file_extension == FILE_EXT_DTA:
 */
    break;
    case __pyx_e_10pyreadstat_16_readstat_parser_FILE_EXT_DTA:

    /* "pyreadstat/_readstat_parser.pyx":869
 *         error = readstat_parse_sas7bdat(parser, filename, ctx);
 *     elif file_extension == FILE_EXT_DTA:
 *         error = readstat_parse_dta(parser, filename, ctx);             # <<<<<<<<<<<<<<
 *     elif file_extension == FILE_EXT_XPORT:
 *         error = readstat_parse_xport(parser, filename, ctx);
 */
    __pyx_v_error = readstat_parse_dta(__pyx_v_parser, __pyx_v_filename, __pyx_v_ctx);

    /* "pyreadstat/_readstat_parser.pyx":868
 *     elif file_extension == FILE_EXT_SAS7BDAT:
 *         error = readstat_parse_sas7bdat(parser, filename, ctx);
 *     elif file_extension == FILE_EXT_DTA:             # <<<<<<<<<<<<<<
 *         error = readstat_parse_dta(parser, filename, ctx);
 *     elif file_extension == FILE_EXT_XPORT:
 */
    break;
    case __pyx_e_10pyreadstat_16_readstat_parser_FILE_EXT_XPORT:

    /* "pyreadstat/_readstat_parser.pyx":871
 *         error = readstat_parse_dta(parser, filename, ctx);
 *     elif file_extension == FILE_EXT_XPORT:
 *         error = readstat_parse_xport(parser, filename, ctx);             # <<<<<<<<<<<<<<
 *     elif file_extension == FILE_EXT_POR:
 *         error = readstat_parse_por(parser, filename, ctx);
 */
    __pyx_v_error = readstat_parse_xport(__pyx_v_parser, __pyx_v_filename, __pyx_v_ctx);

    /* "pyreadstat/_readstat_parser.pyx":870
 *     elif file_extension == FILE_EXT_DTA:
 *         error = readstat_parse_dta(parser, filename, ctx);
 *     elif file_extension == FILE_EXT_XPORT:             # <<<<<<<<<<<<<<
 *         error = readstat_parse_xport(parser, filename, ctx);
 *     elif file_extension == FILE_EXT_POR:
 */
    break;
    case __pyx_e_10pyreadstat_16_readstat_parser_FILE_EXT_POR:

    /* "pyreadstat/_readstat_parser.pyx":873
 *         error = readstat_parse_xport(parser, filename, ctx);
 *     elif file_extension == FILE_EXT_POR:
 *         error = readstat_parse_por(parser, filename, ctx);             # <<<<<<<<<<<<<<
 *     elif file_extension == FILE_EXT_SAS7BCAT:
 *         error = readstat_parse_sas7bcat(parser, filename, ctx);
 */
    __pyx_v_error = readstat_parse_por(__pyx_v_parser, __pyx_v_filename, __pyx_v_ctx);

    /* "pyreadstat/_readstat_parser.pyx":872
 *     elif file_extension == FILE_EXT_XPORT:
 *         error = readstat_parse_xport(parser, filename, ctx);
 *     elif file_extension == FILE_EXT_POR:             # <<<<<<<<<<<<<<
 *         error = readstat_parse_por(parser, filename, ctx);
 *     elif file_extension == FILE_EXT_SAS7BCAT:
 */
    break;
    case __pyx_e_10pyreadstat_16_readstat_parser_FILE_EXT_SAS7BCAT:

    /* "pyreadstat/_readstat_parser.pyx":875
 *         error = readstat_parse_por(parser, filename, ctx);
 *     elif file_extension == FILE_EXT_SAS7BCAT:
 *         error = readstat_parse_sas7bcat(parser, filename, ctx);             # <<<<<<<<<<<<<<
 *     #error = parse_func(parser, filename, ctx);
 *     readstat_parser_free(parser)
 */
    __pyx_v_error = readstat_parse_sas7bcat(__pyx_v_parser, __pyx_v_filename, __pyx_v_ctx);

    /* "pyreadstat/_readstat_parser.pyx":874
 *     elif file_extension == FILE_EXT_POR:
 *         error = readstat_parse_por(parser, filename, ctx);
 *     elif file_extension == FILE_EXT_SAS7BCAT:             # <<<<<<<<<<<<<<
 *         error = readstat_parse_sas7bcat(parser, filename, ctx);
 *     #error = parse_func(parser, filename, ctx);
 */
    break;
    default: break;
  }

  /* "pyreadstat/_readstat_parser.pyx":877
 *         error = readstat_parse_sas7bcat(parser, filename, ctx);
 *     #error = parse_func(parser, filename, ctx);
 *     readstat_parser_free(parser)             # <<<<<<<<<<<<<<
 *     # check if a python error ocurred, if yes, it will be printed by the interpreter,
 *     # if not, make sure that the return from parse_func is OK, if not print
 */
  readstat_parser_free(__pyx_v_parser);

  /* "pyreadstat/_readstat_parser.pyx":880
 *     # check if a python error ocurred, if yes, it will be printed by the interpreter,
 *     # if not, make sure that the return from parse_func is OK, if not print
 *     pyerr = PyErr_Occurred()             # <<<<<<<<<<<<<<
 *     if <void *>pyerr == NULL:
 *         check_exit_status(error)
 */
  __pyx_v_pyerr = PyErr_Occurred();

  /* "pyreadstat/_readstat_parser.pyx":881
 *     # if not, make sure that the return from parse_func is OK, if not print
 *     pyerr = PyErr_Occurred()
 *     if <void *>pyerr == NULL:             # <<<<<<<<<<<<<<
 *         check_exit_status(error)
 * 
 */
  __pyx_t_1 = (((void *)__pyx_v_pyerr) == NULL);
  if (__pyx_t_1) {

    /* "pyreadstat/_readstat_parser.pyx":882
 *     pyerr = PyErr_Occurred()
 *     if <void *>pyerr == NULL:
 *         check_exit_status(error)             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_f_10pyreadstat_16_readstat_parser_check_exit_status(__pyx_v_error); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 882, __pyx_L1_error)

    /* "pyreadstat/_readstat_parser.pyx":881
 *     # if not, make sure that the return from parse_func is OK, if not print
 *     pyerr = PyErr_Occurred()
 *     if <void *>pyerr == NULL:             # <<<<<<<<<<<<<<
 *         check_exit_status(error)
 * 
 */
  }

  /* "pyreadstat/_readstat_parser.pyx":807
 * 
 * 
 * cdef void run_readstat_parser(char * filename, data_container data, py_file_extension file_extension, long row_limit, long row_offset) except *:             # <<<<<<<<<<<<<<
 *     """
 *     Runs the parsing of the file by readstat library
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pyreadstat._readstat_parser.run_readstat_parser", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_encoding_bytes);
  __Pyx_RefNannyFinishContext();
}

/* "pyreadstat/_readstat_parser.pyx":885
 * 
 * 
 * cdef object data_container_to_dict(data_container data):             # <<<<<<<<<<<<<<
 *     """
 *     Transforms a data container object to a pandas data frame
 */

static PyObject *__pyx_f_10pyreadstat_16_readstat_parser_data_container_to_dict(struct __pyx_obj_10pyreadstat_16_readstat_parser_data_container *__pyx_v_data) {
  PyObject *__pyx_v_final_container = 0;
  PyObject *__pyx_v_col_data = 0;
  PyObject *__pyx_v_col_names = 0;
  PyObject *__pyx_v_cur_name_str = 0;
  int __pyx_v_fc_cnt;
  int __pyx_v_is_unkown_number_rows;
  int __pyx_v_max_n_obs;
  int __pyx_v_metaonly;
  PyObject *__pyx_v_cur_data = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("data_container_to_dict", 0);

  /* "pyreadstat/_readstat_parser.pyx":899
 *     cdef bint metaonly
 * 
 *     final_container = OrderedDict()             # <<<<<<<<<<<<<<
 *     col_data = data.col_data
 *     col_names = data.col_names
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_OrderedDict); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 899, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[1] = {__pyx_t_3, };
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 899, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_v_final_container = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pyreadstat/_readstat_parser.pyx":900
 * 
 *     final_container = OrderedDict()
 *     col_data = data.col_data             # <<<<<<<<<<<<<<
 *     col_names = data.col_names
 *     is_unkown_number_rows = data.is_unkown_number_rows
 */
  __pyx_t_1 = __pyx_v_data->col_data;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_col_data = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyreadstat/_readstat_parser.pyx":901
 *     final_container = OrderedDict()
 *     col_data = data.col_data
 *     col_names = data.col_names             # <<<<<<<<<<<<<<
 *     is_unkown_number_rows = data.is_unkown_number_rows
 *     max_n_obs = data.max_n_obs
 */
  __pyx_t_1 = __pyx_v_data->col_names;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_col_names = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyreadstat/_readstat_parser.pyx":902
 *     col_data = data.col_data
 *     col_names = data.col_names
 *     is_unkown_number_rows = data.is_unkown_number_rows             # <<<<<<<<<<<<<<
 *     max_n_obs = data.max_n_obs
 *     metaonly = data.metaonly
 */
  __pyx_t_5 = __pyx_v_data->is_unkown_number_rows;
  __pyx_v_is_unkown_number_rows = __pyx_t_5;

  /* "pyreadstat/_readstat_parser.pyx":903
 *     col_names = data.col_names
 *     is_unkown_number_rows = data.is_unkown_number_rows
 *     max_n_obs = data.max_n_obs             # <<<<<<<<<<<<<<
 *     metaonly = data.metaonly
 * 
 */
  __pyx_t_4 = __pyx_v_data->max_n_obs;
  __pyx_v_max_n_obs = __pyx_t_4;

  /* "pyreadstat/_readstat_parser.pyx":904
 *     is_unkown_number_rows = data.is_unkown_number_rows
 *     max_n_obs = data.max_n_obs
 *     metaonly = data.metaonly             # <<<<<<<<<<<<<<
 * 
 *     for fc_cnt in range(0, len(col_names)):
 */
  __pyx_t_5 = __pyx_v_data->metaonly;
  __pyx_v_metaonly = __pyx_t_5;

  /* "pyreadstat/_readstat_parser.pyx":906
 *     metaonly = data.metaonly
 * 
 *     for fc_cnt in range(0, len(col_names)):             # <<<<<<<<<<<<<<
 *         cur_name_str = col_names[fc_cnt]
 *         cur_data = col_data[fc_cnt]
 */
  if (unlikely(__pyx_v_col_names == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 906, __pyx_L1_error)
  }
  __pyx_t_6 = PyList_GET_SIZE(__pyx_v_col_names); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1))) __PYX_ERR(0, 906, __pyx_L1_error)
  __pyx_t_7 = __pyx_t_6;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_7; __pyx_t_4+=1) {
    __pyx_v_fc_cnt = __pyx_t_4;

    /* "pyreadstat/_readstat_parser.pyx":907
 * 
 *     for fc_cnt in range(0, len(col_names)):
 *         cur_name_str = col_names[fc_cnt]             # <<<<<<<<<<<<<<
 *         cur_data = col_data[fc_cnt]
 *         if is_unkown_number_rows and not metaonly:
 */
    if (unlikely(__pyx_v_col_names == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 907, __pyx_L1_error)
    }
    __pyx_t_1 = __Pyx_GetItemInt_List(__pyx_v_col_names, __pyx_v_fc_cnt, int, 1, __Pyx_PyInt_From_int, 1, 1, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 907, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!(likely(PyString_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("str", __pyx_t_1))) __PYX_ERR(0, 907, __pyx_L1_error)
    __Pyx_XDECREF_SET(__pyx_v_cur_name_str, ((PyObject*)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "pyreadstat/_readstat_parser.pyx":908
 *     for fc_cnt in range(0, len(col_names)):
 *         cur_name_str = col_names[fc_cnt]
 *         cur_data = col_data[fc_cnt]             # <<<<<<<<<<<<<<
 *         if is_unkown_number_rows and not metaonly:
 *             cur_data = cur_data[0:max_n_obs]
 */
    if (unlikely(__pyx_v_col_data == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 908, __pyx_L1_error)
    }
    __pyx_t_1 = __Pyx_GetItemInt_List(__pyx_v_col_data, __pyx_v_fc_cnt, int, 1, __Pyx_PyInt_From_int, 1, 1, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 908, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_cur_data, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "pyreadstat/_readstat_parser.pyx":909
 *         cur_name_str = col_names[fc_cnt]
 *         cur_data = col_data[fc_cnt]
 *         if is_unkown_number_rows and not metaonly:             # <<<<<<<<<<<<<<
 *             cur_data = cur_data[0:max_n_obs]
 *         if not metaonly:
 */
    if (__pyx_v_is_unkown_number_rows) {
    } else {
      __pyx_t_5 = __pyx_v_is_unkown_number_rows;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_8 = (!__pyx_v_metaonly);
    __pyx_t_5 = __pyx_t_8;
    __pyx_L6_bool_binop_done:;
    if (__pyx_t_5) {

      /* "pyreadstat/_readstat_parser.pyx":910
 *         cur_data = col_data[fc_cnt]
 *         if is_unkown_number_rows and not metaonly:
 *             cur_data = cur_data[0:max_n_obs]             # <<<<<<<<<<<<<<
 *         if not metaonly:
 *             final_container[cur_name_str] = cur_data
 */
      __pyx_t_1 = __Pyx_PyObject_GetSlice(__pyx_v_cur_data, 0, __pyx_v_max_n_obs, NULL, NULL, NULL, 1, 1, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 910, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF_SET(__pyx_v_cur_data, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "pyreadstat/_readstat_parser.pyx":909
 *         cur_name_str = col_names[fc_cnt]
 *         cur_data = col_data[fc_cnt]
 *         if is_unkown_number_rows and not metaonly:             # <<<<<<<<<<<<<<
 *             cur_data = cur_data[0:max_n_obs]
 *         if not metaonly:
 */
    }

    /* "pyreadstat/_readstat_parser.pyx":911
 *         if is_unkown_number_rows and not metaonly:
 *             cur_data = cur_data[0:max_n_obs]
 *         if not metaonly:             # <<<<<<<<<<<<<<
 *             final_container[cur_name_str] = cur_data
 *         else:
 */
    __pyx_t_5 = (!__pyx_v_metaonly);
    if (__pyx_t_5) {

      /* "pyreadstat/_readstat_parser.pyx":912
 *             cur_data = cur_data[0:max_n_obs]
 *         if not metaonly:
 *             final_container[cur_name_str] = cur_data             # <<<<<<<<<<<<<<
 *         else:
 *             final_container[cur_name_str] = list()
 */
      if (unlikely((PyObject_SetItem(__pyx_v_final_container, __pyx_v_cur_name_str, __pyx_v_cur_data) < 0))) __PYX_ERR(0, 912, __pyx_L1_error)

      /* "pyreadstat/_readstat_parser.pyx":911
 *         if is_unkown_number_rows and not metaonly:
 *             cur_data = cur_data[0:max_n_obs]
 *         if not metaonly:             # <<<<<<<<<<<<<<
 *             final_container[cur_name_str] = cur_data
 *         else:
 */
      goto __pyx_L8;
    }

    /* "pyreadstat/_readstat_parser.pyx":914
 *             final_container[cur_name_str] = cur_data
 *         else:
 *             final_container[cur_name_str] = list()             # <<<<<<<<<<<<<<
 * 
 *     return final_container
 */
    /*else*/ {
      __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 914, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (unlikely((PyObject_SetItem(__pyx_v_final_container, __pyx_v_cur_name_str, __pyx_t_1) < 0))) __PYX_ERR(0, 914, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_L8:;
  }

  /* "pyreadstat/_readstat_parser.pyx":916
 *             final_container[cur_name_str] = list()
 * 
 *     return final_container             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_final_container);
  __pyx_r = __pyx_v_final_container;
  goto __pyx_L0;

  /* "pyreadstat/_readstat_parser.pyx":885
 * 
 * 
 * cdef object data_container_to_dict(data_container data):             # <<<<<<<<<<<<<<
 *     """
 *     Transforms a data container object to a pandas data frame
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pyreadstat._readstat_parser.data_container_to_dict", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_final_container);
  __Pyx_XDECREF(__pyx_v_col_data);
  __Pyx_XDECREF(__pyx_v_col_names);
  __Pyx_XDECREF(__pyx_v_cur_name_str);
  __Pyx_XDECREF(__pyx_v_cur_data);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyreadstat/_readstat_parser.pyx":919
 * 
 * 
 * cdef object dict_to_pandas_dataframe(object dict_data, data_container dc):             # <<<<<<<<<<<<<<
 *     """
 *     Transforms a dict of numpy arrays to a pandas data frame
 */

static PyObject *__pyx_f_10pyreadstat_16_readstat_parser_dict_to_pandas_dataframe(PyObject *__pyx_v_dict_data, struct __pyx_obj_10pyreadstat_16_readstat_parser_data_container *__pyx_v_dc) {
  int __pyx_v_dates_as_pandas;
  int __pyx_v_index;
  PyObject *__pyx_v_column = 0;
  __pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format __pyx_v_var_format;
  PyObject *__pyx_v_dtypes = 0;
  PyObject *__pyx_v_data_frame = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  PyObject *(*__pyx_t_7)(PyObject *);
  __pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format __pyx_t_8;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  int __pyx_t_12;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("dict_to_pandas_dataframe", 0);

  /* "pyreadstat/_readstat_parser.pyx":930
 *     cdef list dtypes
 * 
 *     dates_as_pandas = dc.dates_as_pandas             # <<<<<<<<<<<<<<
 * 
 *     if dict_data:
 */
  __pyx_t_1 = __pyx_v_dc->dates_as_pandas;
  __pyx_v_dates_as_pandas = __pyx_t_1;

  /* "pyreadstat/_readstat_parser.pyx":932
 *     dates_as_pandas = dc.dates_as_pandas
 * 
 *     if dict_data:             # <<<<<<<<<<<<<<
 *         data_frame = pd.DataFrame.from_dict(dict_data)
 *         if dates_as_pandas:
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_dict_data); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 932, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "pyreadstat/_readstat_parser.pyx":933
 * 
 *     if dict_data:
 *         data_frame = pd.DataFrame.from_dict(dict_data)             # <<<<<<<<<<<<<<
 *         if dates_as_pandas:
 *             dtypes = data_frame.dtypes.tolist()
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pd); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 933, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_DataFrame); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 933, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_from_dict); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 933, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = NULL;
    __pyx_t_5 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_5 = 1;
      }
    }
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v_dict_data};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 933, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __pyx_v_data_frame = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "pyreadstat/_readstat_parser.pyx":934
 *     if dict_data:
 *         data_frame = pd.DataFrame.from_dict(dict_data)
 *         if dates_as_pandas:             # <<<<<<<<<<<<<<
 *             dtypes = data_frame.dtypes.tolist()
 *             # check that datetime columns are datetime type
 */
    if (__pyx_v_dates_as_pandas) {

      /* "pyreadstat/_readstat_parser.pyx":935
 *         data_frame = pd.DataFrame.from_dict(dict_data)
 *         if dates_as_pandas:
 *             dtypes = data_frame.dtypes.tolist()             # <<<<<<<<<<<<<<
 *             # check that datetime columns are datetime type
 *             # this is needed in case all date values are nan
 */
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_data_frame, __pyx_n_s_dtypes); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 935, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_tolist); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 935, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = NULL;
      __pyx_t_5 = 0;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_3)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
          __pyx_t_5 = 1;
        }
      }
      {
        PyObject *__pyx_callargs[1] = {__pyx_t_3, };
        __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 935, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      }
      if (!(likely(PyList_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None) || __Pyx_RaiseUnexpectedTypeError("list", __pyx_t_2))) __PYX_ERR(0, 935, __pyx_L1_error)
      __pyx_v_dtypes = ((PyObject*)__pyx_t_2);
      __pyx_t_2 = 0;

      /* "pyreadstat/_readstat_parser.pyx":938
 *             # check that datetime columns are datetime type
 *             # this is needed in case all date values are nan
 *             for index, column in enumerate(data_frame.columns):             # <<<<<<<<<<<<<<
 *                 var_format = dc.col_formats[index]
 *                 if dtypes[index] != '<M8[ns]' and (var_format == DATE_FORMAT_DATE or var_format == DATE_FORMAT_DATETIME):
 */
      __pyx_t_5 = 0;
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_data_frame, __pyx_n_s_columns); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 938, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      if (likely(PyList_CheckExact(__pyx_t_2)) || PyTuple_CheckExact(__pyx_t_2)) {
        __pyx_t_4 = __pyx_t_2; __Pyx_INCREF(__pyx_t_4); __pyx_t_6 = 0;
        __pyx_t_7 = NULL;
      } else {
        __pyx_t_6 = -1; __pyx_t_4 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 938, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_7 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_4); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 938, __pyx_L1_error)
      }
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      for (;;) {
        if (likely(!__pyx_t_7)) {
          if (likely(PyList_CheckExact(__pyx_t_4))) {
            if (__pyx_t_6 >= PyList_GET_SIZE(__pyx_t_4)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_2 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_6); __Pyx_INCREF(__pyx_t_2); __pyx_t_6++; if (unlikely((0 < 0))) __PYX_ERR(0, 938, __pyx_L1_error)
            #else
            __pyx_t_2 = PySequence_ITEM(__pyx_t_4, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 938, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_2);
            #endif
          } else {
            if (__pyx_t_6 >= PyTuple_GET_SIZE(__pyx_t_4)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_4, __pyx_t_6); __Pyx_INCREF(__pyx_t_2); __pyx_t_6++; if (unlikely((0 < 0))) __PYX_ERR(0, 938, __pyx_L1_error)
            #else
            __pyx_t_2 = PySequence_ITEM(__pyx_t_4, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 938, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_2);
            #endif
          }
        } else {
          __pyx_t_2 = __pyx_t_7(__pyx_t_4);
          if (unlikely(!__pyx_t_2)) {
            PyObject* exc_type = PyErr_Occurred();
            if (exc_type) {
              if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
              else __PYX_ERR(0, 938, __pyx_L1_error)
            }
            break;
          }
          __Pyx_GOTREF(__pyx_t_2);
        }
        if (!(likely(PyString_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None) || __Pyx_RaiseUnexpectedTypeError("str", __pyx_t_2))) __PYX_ERR(0, 938, __pyx_L1_error)
        __Pyx_XDECREF_SET(__pyx_v_column, ((PyObject*)__pyx_t_2));
        __pyx_t_2 = 0;
        __pyx_v_index = __pyx_t_5;
        __pyx_t_5 = (__pyx_t_5 + 1);

        /* "pyreadstat/_readstat_parser.pyx":939
 *             # this is needed in case all date values are nan
 *             for index, column in enumerate(data_frame.columns):
 *                 var_format = dc.col_formats[index]             # <<<<<<<<<<<<<<
 *                 if dtypes[index] != '<M8[ns]' and (var_format == DATE_FORMAT_DATE or var_format == DATE_FORMAT_DATETIME):
 *                     data_frame[column] = pd.to_datetime(data_frame[column])
 */
        if (unlikely(__pyx_v_dc->col_formats == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
          __PYX_ERR(0, 939, __pyx_L1_error)
        }
        __pyx_t_2 = __Pyx_GetItemInt_List(__pyx_v_dc->col_formats, __pyx_v_index, int, 1, __Pyx_PyInt_From_int, 1, 1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 939, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_8 = ((__pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format)__Pyx_PyInt_As___pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format(__pyx_t_2)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 939, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_v_var_format = __pyx_t_8;

        /* "pyreadstat/_readstat_parser.pyx":940
 *             for index, column in enumerate(data_frame.columns):
 *                 var_format = dc.col_formats[index]
 *                 if dtypes[index] != '<M8[ns]' and (var_format == DATE_FORMAT_DATE or var_format == DATE_FORMAT_DATETIME):             # <<<<<<<<<<<<<<
 *                     data_frame[column] = pd.to_datetime(data_frame[column])
 *     else:
 */
        if (unlikely(__pyx_v_dtypes == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
          __PYX_ERR(0, 940, __pyx_L1_error)
        }
        __pyx_t_2 = __Pyx_GetItemInt_List(__pyx_v_dtypes, __pyx_v_index, int, 1, __Pyx_PyInt_From_int, 1, 1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 940, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_9 = (__Pyx_PyString_Equals(__pyx_t_2, __pyx_kp_s_M8_ns, Py_NE)); if (unlikely((__pyx_t_9 < 0))) __PYX_ERR(0, 940, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (__pyx_t_9) {
        } else {
          __pyx_t_1 = __pyx_t_9;
          goto __pyx_L8_bool_binop_done;
        }
        switch (__pyx_v_var_format) {
          case __pyx_e_10pyreadstat_16_readstat_parser_DATE_FORMAT_DATE:
          case __pyx_e_10pyreadstat_16_readstat_parser_DATE_FORMAT_DATETIME:
          __pyx_t_9 = 1;
          break;
          default:
          __pyx_t_9 = 0;
          break;
        }
        __pyx_t_1 = __pyx_t_9;
        __pyx_L8_bool_binop_done:;
        if (__pyx_t_1) {

          /* "pyreadstat/_readstat_parser.pyx":941
 *                 var_format = dc.col_formats[index]
 *                 if dtypes[index] != '<M8[ns]' and (var_format == DATE_FORMAT_DATE or var_format == DATE_FORMAT_DATETIME):
 *                     data_frame[column] = pd.to_datetime(data_frame[column])             # <<<<<<<<<<<<<<
 *     else:
 *         data_frame = pd.DataFrame()
 */
          __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pd); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 941, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_to_datetime); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 941, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_10);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = __Pyx_PyObject_Dict_GetItem(__pyx_v_data_frame, __pyx_v_column); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 941, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_11 = NULL;
          __pyx_t_12 = 0;
          if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_10))) {
            __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_10);
            if (likely(__pyx_t_11)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
              __Pyx_INCREF(__pyx_t_11);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_10, function);
              __pyx_t_12 = 1;
            }
          }
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_11, __pyx_t_3};
            __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_10, __pyx_callargs+1-__pyx_t_12, 1+__pyx_t_12);
            __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 941, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          }
          if (unlikely((PyObject_SetItem(__pyx_v_data_frame, __pyx_v_column, __pyx_t_2) < 0))) __PYX_ERR(0, 941, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

          /* "pyreadstat/_readstat_parser.pyx":940
 *             for index, column in enumerate(data_frame.columns):
 *                 var_format = dc.col_formats[index]
 *                 if dtypes[index] != '<M8[ns]' and (var_format == DATE_FORMAT_DATE or var_format == DATE_FORMAT_DATETIME):             # <<<<<<<<<<<<<<
 *                     data_frame[column] = pd.to_datetime(data_frame[column])
 *     else:
 */
        }

        /* "pyreadstat/_readstat_parser.pyx":938
 *             # check that datetime columns are datetime type
 *             # this is needed in case all date values are nan
 *             for index, column in enumerate(data_frame.columns):             # <<<<<<<<<<<<<<
 *                 var_format = dc.col_formats[index]
 *                 if dtypes[index] != '<M8[ns]' and (var_format == DATE_FORMAT_DATE or var_format == DATE_FORMAT_DATETIME):
 */
      }
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "pyreadstat/_readstat_parser.pyx":934
 *     if dict_data:
 *         data_frame = pd.DataFrame.from_dict(dict_data)
 *         if dates_as_pandas:             # <<<<<<<<<<<<<<
 *             dtypes = data_frame.dtypes.tolist()
 *             # check that datetime columns are datetime type
 */
    }

    /* "pyreadstat/_readstat_parser.pyx":932
 *     dates_as_pandas = dc.dates_as_pandas
 * 
 *     if dict_data:             # <<<<<<<<<<<<<<
 *         data_frame = pd.DataFrame.from_dict(dict_data)
 *         if dates_as_pandas:
 */
    goto __pyx_L3;
  }

  /* "pyreadstat/_readstat_parser.pyx":943
 *                     data_frame[column] = pd.to_datetime(data_frame[column])
 *     else:
 *         data_frame = pd.DataFrame()             # <<<<<<<<<<<<<<
 * 
 *     return data_frame
 */
  /*else*/ {
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_pd); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 943, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_DataFrame); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 943, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = NULL;
    __pyx_t_5 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_10))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_10);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_10, function);
        __pyx_t_5 = 1;
      }
    }
    {
      PyObject *__pyx_callargs[1] = {__pyx_t_2, };
      __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_10, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 943, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    }
    __pyx_v_data_frame = __pyx_t_4;
    __pyx_t_4 = 0;
  }
  __pyx_L3:;

  /* "pyreadstat/_readstat_parser.pyx":945
 *         data_frame = pd.DataFrame()
 * 
 *     return data_frame             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_data_frame);
  __pyx_r = __pyx_v_data_frame;
  goto __pyx_L0;

  /* "pyreadstat/_readstat_parser.pyx":919
 * 
 * 
 * cdef object dict_to_pandas_dataframe(object dict_data, data_container dc):             # <<<<<<<<<<<<<<
 *     """
 *     Transforms a dict of numpy arrays to a pandas data frame
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("pyreadstat._readstat_parser.dict_to_pandas_dataframe", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_column);
  __Pyx_XDECREF(__pyx_v_dtypes);
  __Pyx_XDECREF(__pyx_v_data_frame);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyreadstat/_readstat_parser.pyx":948
 * 
 * 
 * cdef object data_container_extract_metadata(data_container data):             # <<<<<<<<<<<<<<
 *     """
 *     Extracts metadata from a data container and puts it into a metadata
 */

static PyObject *__pyx_f_10pyreadstat_16_readstat_parser_data_container_extract_metadata(struct __pyx_obj_10pyreadstat_16_readstat_parser_data_container *__pyx_v_data) {
  int __pyx_v_metaonly;
  int __pyx_v_is_unkown_number_rows;
  PyObject *__pyx_v_label_to_var_name = 0;
  PyObject *__pyx_v_labels_raw = 0;
  PyObject *__pyx_v_var_name = 0;
  PyObject *__pyx_v_var_label = 0;
  PyObject *__pyx_v_current_labels = 0;
  PyObject *__pyx_v_original_types = 0;
  readstat_type_t __pyx_v_var_type;
  PyObject *__pyx_v_metadata = 0;
  PyObject *__pyx_v_variable_value_labels = NULL;
  PyObject *__pyx_v_readstat_types = NULL;
  PyObject *__pyx_v_indx = NULL;
  PyObject *__pyx_v_cur_col = NULL;
  PyObject *__pyx_v_cur_type = NULL;
  PyObject *__pyx_v_curset = NULL;
  PyObject *__pyx_7genexpr__pyx_v_k = NULL;
  PyObject *__pyx_7genexpr__pyx_v_v = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  PyObject *(*__pyx_t_7)(PyObject *);
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *(*__pyx_t_10)(PyObject *);
  readstat_type_t __pyx_t_11;
  int __pyx_t_12;
  PyObject *__pyx_t_13 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("data_container_extract_metadata", 0);

  /* "pyreadstat/_readstat_parser.pyx":965
 *     cdef readstat_type_t var_type
 * 
 *     metaonly = data.metaonly             # <<<<<<<<<<<<<<
 *     is_unkown_number_rows = data.is_unkown_number_rows
 * 
 */
  __pyx_t_1 = __pyx_v_data->metaonly;
  __pyx_v_metaonly = __pyx_t_1;

  /* "pyreadstat/_readstat_parser.pyx":966
 * 
 *     metaonly = data.metaonly
 *     is_unkown_number_rows = data.is_unkown_number_rows             # <<<<<<<<<<<<<<
 * 
 *     cdef object metadata = metadata_container()
 */
  __pyx_t_1 = __pyx_v_data->is_unkown_number_rows;
  __pyx_v_is_unkown_number_rows = __pyx_t_1;

  /* "pyreadstat/_readstat_parser.pyx":968
 *     is_unkown_number_rows = data.is_unkown_number_rows
 * 
 *     cdef object metadata = metadata_container()             # <<<<<<<<<<<<<<
 * 
 *     # number of rows
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_metadata_container); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 968, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[1] = {__pyx_t_4, };
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 968, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v_metadata = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "pyreadstat/_readstat_parser.pyx":971
 * 
 *     # number of rows
 *     metadata.number_columns = data.n_vars             # <<<<<<<<<<<<<<
 *     if is_unkown_number_rows:
 *         if not metaonly:
 */
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_data->n_vars); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 971, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_metadata, __pyx_n_s_number_columns, __pyx_t_2) < 0) __PYX_ERR(0, 971, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyreadstat/_readstat_parser.pyx":972
 *     # number of rows
 *     metadata.number_columns = data.n_vars
 *     if is_unkown_number_rows:             # <<<<<<<<<<<<<<
 *         if not metaonly:
 *             metadata.number_rows = data.max_n_obs
 */
  if (__pyx_v_is_unkown_number_rows) {

    /* "pyreadstat/_readstat_parser.pyx":973
 *     metadata.number_columns = data.n_vars
 *     if is_unkown_number_rows:
 *         if not metaonly:             # <<<<<<<<<<<<<<
 *             metadata.number_rows = data.max_n_obs
 *     else:
 */
    __pyx_t_1 = (!__pyx_v_metaonly);
    if (__pyx_t_1) {

      /* "pyreadstat/_readstat_parser.pyx":974
 *     if is_unkown_number_rows:
 *         if not metaonly:
 *             metadata.number_rows = data.max_n_obs             # <<<<<<<<<<<<<<
 *     else:
 *         metadata.number_rows = data.n_obs
 */
      __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_data->max_n_obs); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 974, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      if (__Pyx_PyObject_SetAttrStr(__pyx_v_metadata, __pyx_n_s_number_rows, __pyx_t_2) < 0) __PYX_ERR(0, 974, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "pyreadstat/_readstat_parser.pyx":973
 *     metadata.number_columns = data.n_vars
 *     if is_unkown_number_rows:
 *         if not metaonly:             # <<<<<<<<<<<<<<
 *             metadata.number_rows = data.max_n_obs
 *     else:
 */
    }

    /* "pyreadstat/_readstat_parser.pyx":972
 *     # number of rows
 *     metadata.number_columns = data.n_vars
 *     if is_unkown_number_rows:             # <<<<<<<<<<<<<<
 *         if not metaonly:
 *             metadata.number_rows = data.max_n_obs
 */
    goto __pyx_L3;
  }

  /* "pyreadstat/_readstat_parser.pyx":976
 *             metadata.number_rows = data.max_n_obs
 *     else:
 *         metadata.number_rows = data.n_obs             # <<<<<<<<<<<<<<
 * 
 *     # value labels
 */
  /*else*/ {
    __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_data->n_obs); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 976, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_metadata, __pyx_n_s_number_rows, __pyx_t_2) < 0) __PYX_ERR(0, 976, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_L3:;

  /* "pyreadstat/_readstat_parser.pyx":979
 * 
 *     # value labels
 *     labels_raw = data.labels_raw             # <<<<<<<<<<<<<<
 *     label_to_var_name = data.label_to_var_name
 *     variable_value_labels = dict()
 */
  __pyx_t_2 = __pyx_v_data->labels_raw;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_v_labels_raw = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "pyreadstat/_readstat_parser.pyx":980
 *     # value labels
 *     labels_raw = data.labels_raw
 *     label_to_var_name = data.label_to_var_name             # <<<<<<<<<<<<<<
 *     variable_value_labels = dict()
 * 
 */
  __pyx_t_2 = __pyx_v_data->label_to_var_name;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_v_label_to_var_name = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "pyreadstat/_readstat_parser.pyx":981
 *     labels_raw = data.labels_raw
 *     label_to_var_name = data.label_to_var_name
 *     variable_value_labels = dict()             # <<<<<<<<<<<<<<
 * 
 *     if labels_raw:
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 981, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_variable_value_labels = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "pyreadstat/_readstat_parser.pyx":983
 *     variable_value_labels = dict()
 * 
 *     if labels_raw:             # <<<<<<<<<<<<<<
 *         for var_name, var_label in label_to_var_name.items():
 *             current_labels = labels_raw.get(var_label)
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_labels_raw); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 983, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "pyreadstat/_readstat_parser.pyx":984
 * 
 *     if labels_raw:
 *         for var_name, var_label in label_to_var_name.items():             # <<<<<<<<<<<<<<
 *             current_labels = labels_raw.get(var_label)
 *             if current_labels:
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_label_to_var_name, __pyx_n_s_items); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 984, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = NULL;
    __pyx_t_5 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_5 = 1;
      }
    }
    {
      PyObject *__pyx_callargs[1] = {__pyx_t_4, };
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 984, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    if (likely(PyList_CheckExact(__pyx_t_2)) || PyTuple_CheckExact(__pyx_t_2)) {
      __pyx_t_3 = __pyx_t_2; __Pyx_INCREF(__pyx_t_3); __pyx_t_6 = 0;
      __pyx_t_7 = NULL;
    } else {
      __pyx_t_6 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 984, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_7 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 984, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    for (;;) {
      if (likely(!__pyx_t_7)) {
        if (likely(PyList_CheckExact(__pyx_t_3))) {
          if (__pyx_t_6 >= PyList_GET_SIZE(__pyx_t_3)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_2 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_6); __Pyx_INCREF(__pyx_t_2); __pyx_t_6++; if (unlikely((0 < 0))) __PYX_ERR(0, 984, __pyx_L1_error)
          #else
          __pyx_t_2 = PySequence_ITEM(__pyx_t_3, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 984, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          #endif
        } else {
          if (__pyx_t_6 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_6); __Pyx_INCREF(__pyx_t_2); __pyx_t_6++; if (unlikely((0 < 0))) __PYX_ERR(0, 984, __pyx_L1_error)
          #else
          __pyx_t_2 = PySequence_ITEM(__pyx_t_3, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 984, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          #endif
        }
      } else {
        __pyx_t_2 = __pyx_t_7(__pyx_t_3);
        if (unlikely(!__pyx_t_2)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 984, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_2);
      }
      if ((likely(PyTuple_CheckExact(__pyx_t_2))) || (PyList_CheckExact(__pyx_t_2))) {
        PyObject* sequence = __pyx_t_2;
        Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
        if (unlikely(size != 2)) {
          if (size > 2) __Pyx_RaiseTooManyValuesError(2);
          else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
          __PYX_ERR(0, 984, __pyx_L1_error)
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        if (likely(PyTuple_CheckExact(sequence))) {
          __pyx_t_4 = PyTuple_GET_ITEM(sequence, 0); 
          __pyx_t_8 = PyTuple_GET_ITEM(sequence, 1); 
        } else {
          __pyx_t_4 = PyList_GET_ITEM(sequence, 0); 
          __pyx_t_8 = PyList_GET_ITEM(sequence, 1); 
        }
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_8);
        #else
        __pyx_t_4 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 984, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_8 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 984, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        #endif
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      } else {
        Py_ssize_t index = -1;
        __pyx_t_9 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 984, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_10 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_9);
        index = 0; __pyx_t_4 = __pyx_t_10(__pyx_t_9); if (unlikely(!__pyx_t_4)) goto __pyx_L8_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_4);
        index = 1; __pyx_t_8 = __pyx_t_10(__pyx_t_9); if (unlikely(!__pyx_t_8)) goto __pyx_L8_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_8);
        if (__Pyx_IternextUnpackEndCheck(__pyx_t_10(__pyx_t_9), 2) < 0) __PYX_ERR(0, 984, __pyx_L1_error)
        __pyx_t_10 = NULL;
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        goto __pyx_L9_unpacking_done;
        __pyx_L8_unpacking_failed:;
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __pyx_t_10 = NULL;
        if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
        __PYX_ERR(0, 984, __pyx_L1_error)
        __pyx_L9_unpacking_done:;
      }
      if (!(likely(PyString_CheckExact(__pyx_t_4))||((__pyx_t_4) == Py_None) || __Pyx_RaiseUnexpectedTypeError("str", __pyx_t_4))) __PYX_ERR(0, 984, __pyx_L1_error)
      if (!(likely(PyString_CheckExact(__pyx_t_8))||((__pyx_t_8) == Py_None) || __Pyx_RaiseUnexpectedTypeError("str", __pyx_t_8))) __PYX_ERR(0, 984, __pyx_L1_error)
      __Pyx_XDECREF_SET(__pyx_v_var_name, ((PyObject*)__pyx_t_4));
      __pyx_t_4 = 0;
      __Pyx_XDECREF_SET(__pyx_v_var_label, ((PyObject*)__pyx_t_8));
      __pyx_t_8 = 0;

      /* "pyreadstat/_readstat_parser.pyx":985
 *     if labels_raw:
 *         for var_name, var_label in label_to_var_name.items():
 *             current_labels = labels_raw.get(var_label)             # <<<<<<<<<<<<<<
 *             if current_labels:
 *                 variable_value_labels[var_name] = current_labels
 */
      __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_labels_raw, __pyx_n_s_get); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 985, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_4 = NULL;
      __pyx_t_5 = 0;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_8))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_8);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_8, function);
          __pyx_t_5 = 1;
        }
      }
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v_var_label};
        __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_8, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 985, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      }
      __Pyx_XDECREF_SET(__pyx_v_current_labels, __pyx_t_2);
      __pyx_t_2 = 0;

      /* "pyreadstat/_readstat_parser.pyx":986
 *         for var_name, var_label in label_to_var_name.items():
 *             current_labels = labels_raw.get(var_label)
 *             if current_labels:             # <<<<<<<<<<<<<<
 *                 variable_value_labels[var_name] = current_labels
 * 
 */
      __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_current_labels); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 986, __pyx_L1_error)
      if (__pyx_t_1) {

        /* "pyreadstat/_readstat_parser.pyx":987
 *             current_labels = labels_raw.get(var_label)
 *             if current_labels:
 *                 variable_value_labels[var_name] = current_labels             # <<<<<<<<<<<<<<
 * 
 *     original_types = dict()
 */
        if (unlikely((PyDict_SetItem(__pyx_v_variable_value_labels, __pyx_v_var_name, __pyx_v_current_labels) < 0))) __PYX_ERR(0, 987, __pyx_L1_error)

        /* "pyreadstat/_readstat_parser.pyx":986
 *         for var_name, var_label in label_to_var_name.items():
 *             current_labels = labels_raw.get(var_label)
 *             if current_labels:             # <<<<<<<<<<<<<<
 *                 variable_value_labels[var_name] = current_labels
 * 
 */
      }

      /* "pyreadstat/_readstat_parser.pyx":984
 * 
 *     if labels_raw:
 *         for var_name, var_label in label_to_var_name.items():             # <<<<<<<<<<<<<<
 *             current_labels = labels_raw.get(var_label)
 *             if current_labels:
 */
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pyreadstat/_readstat_parser.pyx":983
 *     variable_value_labels = dict()
 * 
 *     if labels_raw:             # <<<<<<<<<<<<<<
 *         for var_name, var_label in label_to_var_name.items():
 *             current_labels = labels_raw.get(var_label)
 */
  }

  /* "pyreadstat/_readstat_parser.pyx":989
 *                 variable_value_labels[var_name] = current_labels
 * 
 *     original_types = dict()             # <<<<<<<<<<<<<<
 *     readstat_types = dict()
 *     for indx in range(metadata.number_columns):
 */
  __pyx_t_3 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 989, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v_original_types = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "pyreadstat/_readstat_parser.pyx":990
 * 
 *     original_types = dict()
 *     readstat_types = dict()             # <<<<<<<<<<<<<<
 *     for indx in range(metadata.number_columns):
 *         cur_col = data.col_names[indx]
 */
  __pyx_t_3 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 990, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v_readstat_types = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "pyreadstat/_readstat_parser.pyx":991
 *     original_types = dict()
 *     readstat_types = dict()
 *     for indx in range(metadata.number_columns):             # <<<<<<<<<<<<<<
 *         cur_col = data.col_names[indx]
 *         cur_type = data.col_formats_original[indx]
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_metadata, __pyx_n_s_number_columns); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 991, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_range, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 991, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (likely(PyList_CheckExact(__pyx_t_2)) || PyTuple_CheckExact(__pyx_t_2)) {
    __pyx_t_3 = __pyx_t_2; __Pyx_INCREF(__pyx_t_3); __pyx_t_6 = 0;
    __pyx_t_7 = NULL;
  } else {
    __pyx_t_6 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 991, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_7 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 991, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  for (;;) {
    if (likely(!__pyx_t_7)) {
      if (likely(PyList_CheckExact(__pyx_t_3))) {
        if (__pyx_t_6 >= PyList_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_6); __Pyx_INCREF(__pyx_t_2); __pyx_t_6++; if (unlikely((0 < 0))) __PYX_ERR(0, 991, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_3, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 991, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      } else {
        if (__pyx_t_6 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_6); __Pyx_INCREF(__pyx_t_2); __pyx_t_6++; if (unlikely((0 < 0))) __PYX_ERR(0, 991, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_3, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 991, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      }
    } else {
      __pyx_t_2 = __pyx_t_7(__pyx_t_3);
      if (unlikely(!__pyx_t_2)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 991, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_XDECREF_SET(__pyx_v_indx, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "pyreadstat/_readstat_parser.pyx":992
 *     readstat_types = dict()
 *     for indx in range(metadata.number_columns):
 *         cur_col = data.col_names[indx]             # <<<<<<<<<<<<<<
 *         cur_type = data.col_formats_original[indx]
 *         original_types[cur_col] = cur_type
 */
    if (unlikely(__pyx_v_data->col_names == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 992, __pyx_L1_error)
    }
    __pyx_t_2 = __Pyx_PyObject_GetItem(__pyx_v_data->col_names, __pyx_v_indx); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 992, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_XDECREF_SET(__pyx_v_cur_col, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "pyreadstat/_readstat_parser.pyx":993
 *     for indx in range(metadata.number_columns):
 *         cur_col = data.col_names[indx]
 *         cur_type = data.col_formats_original[indx]             # <<<<<<<<<<<<<<
 *         original_types[cur_col] = cur_type
 *         var_type = data.col_dtypes[indx]
 */
    if (unlikely(__pyx_v_data->col_formats_original == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 993, __pyx_L1_error)
    }
    __pyx_t_2 = __Pyx_PyObject_GetItem(__pyx_v_data->col_formats_original, __pyx_v_indx); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 993, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_XDECREF_SET(__pyx_v_cur_type, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "pyreadstat/_readstat_parser.pyx":994
 *         cur_col = data.col_names[indx]
 *         cur_type = data.col_formats_original[indx]
 *         original_types[cur_col] = cur_type             # <<<<<<<<<<<<<<
 *         var_type = data.col_dtypes[indx]
 *         if var_type == READSTAT_TYPE_STRING or var_type == READSTAT_TYPE_STRING_REF:
 */
    if (unlikely((PyObject_SetItem(__pyx_v_original_types, __pyx_v_cur_col, __pyx_v_cur_type) < 0))) __PYX_ERR(0, 994, __pyx_L1_error)

    /* "pyreadstat/_readstat_parser.pyx":995
 *         cur_type = data.col_formats_original[indx]
 *         original_types[cur_col] = cur_type
 *         var_type = data.col_dtypes[indx]             # <<<<<<<<<<<<<<
 *         if var_type == READSTAT_TYPE_STRING or var_type == READSTAT_TYPE_STRING_REF:
 *             readstat_types[cur_col] = "string"
 */
    if (unlikely(__pyx_v_data->col_dtypes == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 995, __pyx_L1_error)
    }
    __pyx_t_2 = __Pyx_PyObject_GetItem(__pyx_v_data->col_dtypes, __pyx_v_indx); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 995, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_11 = ((readstat_type_t)__Pyx_PyInt_As_readstat_type_t(__pyx_t_2)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 995, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_var_type = __pyx_t_11;

    /* "pyreadstat/_readstat_parser.pyx":996
 *         original_types[cur_col] = cur_type
 *         var_type = data.col_dtypes[indx]
 *         if var_type == READSTAT_TYPE_STRING or var_type == READSTAT_TYPE_STRING_REF:             # <<<<<<<<<<<<<<
 *             readstat_types[cur_col] = "string"
 *         elif var_type == READSTAT_TYPE_INT8:
 */
    switch (__pyx_v_var_type) {
      case READSTAT_TYPE_STRING:
      case READSTAT_TYPE_STRING_REF:

      /* "pyreadstat/_readstat_parser.pyx":997
 *         var_type = data.col_dtypes[indx]
 *         if var_type == READSTAT_TYPE_STRING or var_type == READSTAT_TYPE_STRING_REF:
 *             readstat_types[cur_col] = "string"             # <<<<<<<<<<<<<<
 *         elif var_type == READSTAT_TYPE_INT8:
 *             readstat_types[cur_col] = "int8"
 */
      if (unlikely((PyDict_SetItem(__pyx_v_readstat_types, __pyx_v_cur_col, __pyx_n_s_string) < 0))) __PYX_ERR(0, 997, __pyx_L1_error)

      /* "pyreadstat/_readstat_parser.pyx":996
 *         original_types[cur_col] = cur_type
 *         var_type = data.col_dtypes[indx]
 *         if var_type == READSTAT_TYPE_STRING or var_type == READSTAT_TYPE_STRING_REF:             # <<<<<<<<<<<<<<
 *             readstat_types[cur_col] = "string"
 *         elif var_type == READSTAT_TYPE_INT8:
 */
      break;
      case READSTAT_TYPE_INT8:

      /* "pyreadstat/_readstat_parser.pyx":999
 *             readstat_types[cur_col] = "string"
 *         elif var_type == READSTAT_TYPE_INT8:
 *             readstat_types[cur_col] = "int8"             # <<<<<<<<<<<<<<
 *         elif var_type == READSTAT_TYPE_INT16:
 *             readstat_types[cur_col] = "int16"
 */
      if (unlikely((PyDict_SetItem(__pyx_v_readstat_types, __pyx_v_cur_col, __pyx_n_s_int8) < 0))) __PYX_ERR(0, 999, __pyx_L1_error)

      /* "pyreadstat/_readstat_parser.pyx":998
 *         if var_type == READSTAT_TYPE_STRING or var_type == READSTAT_TYPE_STRING_REF:
 *             readstat_types[cur_col] = "string"
 *         elif var_type == READSTAT_TYPE_INT8:             # <<<<<<<<<<<<<<
 *             readstat_types[cur_col] = "int8"
 *         elif var_type == READSTAT_TYPE_INT16:
 */
      break;
      case READSTAT_TYPE_INT16:

      /* "pyreadstat/_readstat_parser.pyx":1001
 *             readstat_types[cur_col] = "int8"
 *         elif var_type == READSTAT_TYPE_INT16:
 *             readstat_types[cur_col] = "int16"             # <<<<<<<<<<<<<<
 *         elif var_type == READSTAT_TYPE_INT32:
 *             readstat_types[cur_col] = "int32"
 */
      if (unlikely((PyDict_SetItem(__pyx_v_readstat_types, __pyx_v_cur_col, __pyx_n_s_int16) < 0))) __PYX_ERR(0, 1001, __pyx_L1_error)

      /* "pyreadstat/_readstat_parser.pyx":1000
 *         elif var_type == READSTAT_TYPE_INT8:
 *             readstat_types[cur_col] = "int8"
 *         elif var_type == READSTAT_TYPE_INT16:             # <<<<<<<<<<<<<<
 *             readstat_types[cur_col] = "int16"
 *         elif var_type == READSTAT_TYPE_INT32:
 */
      break;
      case READSTAT_TYPE_INT32:

      /* "pyreadstat/_readstat_parser.pyx":1003
 *             readstat_types[cur_col] = "int16"
 *         elif var_type == READSTAT_TYPE_INT32:
 *             readstat_types[cur_col] = "int32"             # <<<<<<<<<<<<<<
 *         elif var_type == READSTAT_TYPE_FLOAT:
 *             readstat_types[cur_col] = "float"
 */
      if (unlikely((PyDict_SetItem(__pyx_v_readstat_types, __pyx_v_cur_col, __pyx_n_s_int32) < 0))) __PYX_ERR(0, 1003, __pyx_L1_error)

      /* "pyreadstat/_readstat_parser.pyx":1002
 *         elif var_type == READSTAT_TYPE_INT16:
 *             readstat_types[cur_col] = "int16"
 *         elif var_type == READSTAT_TYPE_INT32:             # <<<<<<<<<<<<<<
 *             readstat_types[cur_col] = "int32"
 *         elif var_type == READSTAT_TYPE_FLOAT:
 */
      break;
      case READSTAT_TYPE_FLOAT:

      /* "pyreadstat/_readstat_parser.pyx":1005
 *             readstat_types[cur_col] = "int32"
 *         elif var_type == READSTAT_TYPE_FLOAT:
 *             readstat_types[cur_col] = "float"             # <<<<<<<<<<<<<<
 *         elif var_type == READSTAT_TYPE_DOUBLE:
 *             readstat_types[cur_col] = "double"
 */
      if (unlikely((PyDict_SetItem(__pyx_v_readstat_types, __pyx_v_cur_col, __pyx_n_s_float) < 0))) __PYX_ERR(0, 1005, __pyx_L1_error)

      /* "pyreadstat/_readstat_parser.pyx":1004
 *         elif var_type == READSTAT_TYPE_INT32:
 *             readstat_types[cur_col] = "int32"
 *         elif var_type == READSTAT_TYPE_FLOAT:             # <<<<<<<<<<<<<<
 *             readstat_types[cur_col] = "float"
 *         elif var_type == READSTAT_TYPE_DOUBLE:
 */
      break;
      case READSTAT_TYPE_DOUBLE:

      /* "pyreadstat/_readstat_parser.pyx":1007
 *             readstat_types[cur_col] = "float"
 *         elif var_type == READSTAT_TYPE_DOUBLE:
 *             readstat_types[cur_col] = "double"             # <<<<<<<<<<<<<<
 *         else:
 *             raise PyreadstatError("Unkown data type")
 */
      if (unlikely((PyDict_SetItem(__pyx_v_readstat_types, __pyx_v_cur_col, __pyx_n_s_double) < 0))) __PYX_ERR(0, 1007, __pyx_L1_error)

      /* "pyreadstat/_readstat_parser.pyx":1006
 *         elif var_type == READSTAT_TYPE_FLOAT:
 *             readstat_types[cur_col] = "float"
 *         elif var_type == READSTAT_TYPE_DOUBLE:             # <<<<<<<<<<<<<<
 *             readstat_types[cur_col] = "double"
 *         else:
 */
      break;
      default:

      /* "pyreadstat/_readstat_parser.pyx":1009
 *             readstat_types[cur_col] = "double"
 *         else:
 *             raise PyreadstatError("Unkown data type")             # <<<<<<<<<<<<<<
 * 
 *     for indx, curset in data.missing_user_values.items():
 */
      __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_PyreadstatError); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1009, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_4 = NULL;
      __pyx_t_5 = 0;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_8))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_8);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_8, function);
          __pyx_t_5 = 1;
        }
      }
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_kp_s_Unkown_data_type};
        __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_8, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1009, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      }
      __Pyx_Raise(__pyx_t_2, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __PYX_ERR(0, 1009, __pyx_L1_error)
      break;
    }

    /* "pyreadstat/_readstat_parser.pyx":991
 *     original_types = dict()
 *     readstat_types = dict()
 *     for indx in range(metadata.number_columns):             # <<<<<<<<<<<<<<
 *         cur_col = data.col_names[indx]
 *         cur_type = data.col_formats_original[indx]
 */
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "pyreadstat/_readstat_parser.pyx":1011
 *             raise PyreadstatError("Unkown data type")
 * 
 *     for indx, curset in data.missing_user_values.items():             # <<<<<<<<<<<<<<
 *         cur_col = data.col_names[indx]
 *         metadata.missing_user_values[cur_col] = sorted(list(curset))
 */
  if (unlikely(__pyx_v_data->missing_user_values == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "items");
    __PYX_ERR(0, 1011, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyDict_Items(__pyx_v_data->missing_user_values); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1011, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (likely(PyList_CheckExact(__pyx_t_3)) || PyTuple_CheckExact(__pyx_t_3)) {
    __pyx_t_2 = __pyx_t_3; __Pyx_INCREF(__pyx_t_2); __pyx_t_6 = 0;
    __pyx_t_7 = NULL;
  } else {
    __pyx_t_6 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1011, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_7 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1011, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  for (;;) {
    if (likely(!__pyx_t_7)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_6 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_3 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_6); __Pyx_INCREF(__pyx_t_3); __pyx_t_6++; if (unlikely((0 < 0))) __PYX_ERR(0, 1011, __pyx_L1_error)
        #else
        __pyx_t_3 = PySequence_ITEM(__pyx_t_2, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1011, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        #endif
      } else {
        if (__pyx_t_6 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_6); __Pyx_INCREF(__pyx_t_3); __pyx_t_6++; if (unlikely((0 < 0))) __PYX_ERR(0, 1011, __pyx_L1_error)
        #else
        __pyx_t_3 = PySequence_ITEM(__pyx_t_2, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1011, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        #endif
      }
    } else {
      __pyx_t_3 = __pyx_t_7(__pyx_t_2);
      if (unlikely(!__pyx_t_3)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 1011, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_3);
    }
    if ((likely(PyTuple_CheckExact(__pyx_t_3))) || (PyList_CheckExact(__pyx_t_3))) {
      PyObject* sequence = __pyx_t_3;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 1011, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_8 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_4 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_8 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_4 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_8);
      __Pyx_INCREF(__pyx_t_4);
      #else
      __pyx_t_8 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1011, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_4 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1011, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      #endif
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_9 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1011, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_10 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_9);
      index = 0; __pyx_t_8 = __pyx_t_10(__pyx_t_9); if (unlikely(!__pyx_t_8)) goto __pyx_L17_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_8);
      index = 1; __pyx_t_4 = __pyx_t_10(__pyx_t_9); if (unlikely(!__pyx_t_4)) goto __pyx_L17_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_4);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_10(__pyx_t_9), 2) < 0) __PYX_ERR(0, 1011, __pyx_L1_error)
      __pyx_t_10 = NULL;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      goto __pyx_L18_unpacking_done;
      __pyx_L17_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_t_10 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 1011, __pyx_L1_error)
      __pyx_L18_unpacking_done:;
    }
    __Pyx_XDECREF_SET(__pyx_v_indx, __pyx_t_8);
    __pyx_t_8 = 0;
    __Pyx_XDECREF_SET(__pyx_v_curset, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "pyreadstat/_readstat_parser.pyx":1012
 * 
 *     for indx, curset in data.missing_user_values.items():
 *         cur_col = data.col_names[indx]             # <<<<<<<<<<<<<<
 *         metadata.missing_user_values[cur_col] = sorted(list(curset))
 * 
 */
    if (unlikely(__pyx_v_data->col_names == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 1012, __pyx_L1_error)
    }
    __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_v_data->col_names, __pyx_v_indx); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1012, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_XDECREF_SET(__pyx_v_cur_col, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "pyreadstat/_readstat_parser.pyx":1013
 *     for indx, curset in data.missing_user_values.items():
 *         cur_col = data.col_names[indx]
 *         metadata.missing_user_values[cur_col] = sorted(list(curset))             # <<<<<<<<<<<<<<
 * 
 *     metadata.notes = data.notes
 */
    __pyx_t_4 = PySequence_List(__pyx_v_curset); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1013, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_8 = PySequence_List(__pyx_t_4); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1013, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_3 = ((PyObject*)__pyx_t_8);
    __pyx_t_8 = 0;
    __pyx_t_12 = PyList_Sort(__pyx_t_3); if (unlikely(__pyx_t_12 == ((int)-1))) __PYX_ERR(0, 1013, __pyx_L1_error)
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_metadata, __pyx_n_s_missing_user_values); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1013, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    if (unlikely((PyObject_SetItem(__pyx_t_8, __pyx_v_cur_col, __pyx_t_3) < 0))) __PYX_ERR(0, 1013, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pyreadstat/_readstat_parser.pyx":1011
 *             raise PyreadstatError("Unkown data type")
 * 
 *     for indx, curset in data.missing_user_values.items():             # <<<<<<<<<<<<<<
 *         cur_col = data.col_names[indx]
 *         metadata.missing_user_values[cur_col] = sorted(list(curset))
 */
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyreadstat/_readstat_parser.pyx":1015
 *         metadata.missing_user_values[cur_col] = sorted(list(curset))
 * 
 *     metadata.notes = data.notes             # <<<<<<<<<<<<<<
 *     metadata.column_names = data.col_names
 *     metadata.column_labels = data.col_labels
 */
  __pyx_t_2 = __pyx_v_data->notes;
  __Pyx_INCREF(__pyx_t_2);
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_metadata, __pyx_n_s_notes, __pyx_t_2) < 0) __PYX_ERR(0, 1015, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyreadstat/_readstat_parser.pyx":1016
 * 
 *     metadata.notes = data.notes
 *     metadata.column_names = data.col_names             # <<<<<<<<<<<<<<
 *     metadata.column_labels = data.col_labels
 *     metadata.column_names_to_labels = {k:v for k,v in zip(data.col_names, data.col_labels)}
 */
  __pyx_t_2 = __pyx_v_data->col_names;
  __Pyx_INCREF(__pyx_t_2);
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_metadata, __pyx_n_s_column_names, __pyx_t_2) < 0) __PYX_ERR(0, 1016, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyreadstat/_readstat_parser.pyx":1017
 *     metadata.notes = data.notes
 *     metadata.column_names = data.col_names
 *     metadata.column_labels = data.col_labels             # <<<<<<<<<<<<<<
 *     metadata.column_names_to_labels = {k:v for k,v in zip(data.col_names, data.col_labels)}
 *     metadata.file_encoding = data.file_encoding
 */
  __pyx_t_2 = __pyx_v_data->col_labels;
  __Pyx_INCREF(__pyx_t_2);
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_metadata, __pyx_n_s_column_labels, __pyx_t_2) < 0) __PYX_ERR(0, 1017, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyreadstat/_readstat_parser.pyx":1018
 *     metadata.column_names = data.col_names
 *     metadata.column_labels = data.col_labels
 *     metadata.column_names_to_labels = {k:v for k,v in zip(data.col_names, data.col_labels)}             # <<<<<<<<<<<<<<
 *     metadata.file_encoding = data.file_encoding
 *     metadata.file_label = data.file_label
 */
  { /* enter inner scope */
    __pyx_t_2 = PyDict_New(); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1018, __pyx_L22_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1018, __pyx_L22_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_v_data->col_names);
    __Pyx_GIVEREF(__pyx_v_data->col_names);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_data->col_names);
    __Pyx_INCREF(__pyx_v_data->col_labels);
    __Pyx_GIVEREF(__pyx_v_data->col_labels);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_data->col_labels);
    __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_zip, __pyx_t_3, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1018, __pyx_L22_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (likely(PyList_CheckExact(__pyx_t_8)) || PyTuple_CheckExact(__pyx_t_8)) {
      __pyx_t_3 = __pyx_t_8; __Pyx_INCREF(__pyx_t_3); __pyx_t_6 = 0;
      __pyx_t_7 = NULL;
    } else {
      __pyx_t_6 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1018, __pyx_L22_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_7 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1018, __pyx_L22_error)
    }
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    for (;;) {
      if (likely(!__pyx_t_7)) {
        if (likely(PyList_CheckExact(__pyx_t_3))) {
          if (__pyx_t_6 >= PyList_GET_SIZE(__pyx_t_3)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_8 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_6); __Pyx_INCREF(__pyx_t_8); __pyx_t_6++; if (unlikely((0 < 0))) __PYX_ERR(0, 1018, __pyx_L22_error)
          #else
          __pyx_t_8 = PySequence_ITEM(__pyx_t_3, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1018, __pyx_L22_error)
          __Pyx_GOTREF(__pyx_t_8);
          #endif
        } else {
          if (__pyx_t_6 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_8 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_6); __Pyx_INCREF(__pyx_t_8); __pyx_t_6++; if (unlikely((0 < 0))) __PYX_ERR(0, 1018, __pyx_L22_error)
          #else
          __pyx_t_8 = PySequence_ITEM(__pyx_t_3, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1018, __pyx_L22_error)
          __Pyx_GOTREF(__pyx_t_8);
          #endif
        }
      } else {
        __pyx_t_8 = __pyx_t_7(__pyx_t_3);
        if (unlikely(!__pyx_t_8)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 1018, __pyx_L22_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_8);
      }
      if ((likely(PyTuple_CheckExact(__pyx_t_8))) || (PyList_CheckExact(__pyx_t_8))) {
        PyObject* sequence = __pyx_t_8;
        Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
        if (unlikely(size != 2)) {
          if (size > 2) __Pyx_RaiseTooManyValuesError(2);
          else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
          __PYX_ERR(0, 1018, __pyx_L22_error)
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        if (likely(PyTuple_CheckExact(sequence))) {
          __pyx_t_4 = PyTuple_GET_ITEM(sequence, 0); 
          __pyx_t_9 = PyTuple_GET_ITEM(sequence, 1); 
        } else {
          __pyx_t_4 = PyList_GET_ITEM(sequence, 0); 
          __pyx_t_9 = PyList_GET_ITEM(sequence, 1); 
        }
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_9);
        #else
        __pyx_t_4 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1018, __pyx_L22_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_9 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1018, __pyx_L22_error)
        __Pyx_GOTREF(__pyx_t_9);
        #endif
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      } else {
        Py_ssize_t index = -1;
        __pyx_t_13 = PyObject_GetIter(__pyx_t_8); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1018, __pyx_L22_error)
        __Pyx_GOTREF(__pyx_t_13);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __pyx_t_10 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_13);
        index = 0; __pyx_t_4 = __pyx_t_10(__pyx_t_13); if (unlikely(!__pyx_t_4)) goto __pyx_L25_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_4);
        index = 1; __pyx_t_9 = __pyx_t_10(__pyx_t_13); if (unlikely(!__pyx_t_9)) goto __pyx_L25_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_9);
        if (__Pyx_IternextUnpackEndCheck(__pyx_t_10(__pyx_t_13), 2) < 0) __PYX_ERR(0, 1018, __pyx_L22_error)
        __pyx_t_10 = NULL;
        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
        goto __pyx_L26_unpacking_done;
        __pyx_L25_unpacking_failed:;
        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
        __pyx_t_10 = NULL;
        if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
        __PYX_ERR(0, 1018, __pyx_L22_error)
        __pyx_L26_unpacking_done:;
      }
      __Pyx_XDECREF_SET(__pyx_7genexpr__pyx_v_k, __pyx_t_4);
      __pyx_t_4 = 0;
      __Pyx_XDECREF_SET(__pyx_7genexpr__pyx_v_v, __pyx_t_9);
      __pyx_t_9 = 0;
      if (unlikely(PyDict_SetItem(__pyx_t_2, (PyObject*)__pyx_7genexpr__pyx_v_k, (PyObject*)__pyx_7genexpr__pyx_v_v))) __PYX_ERR(0, 1018, __pyx_L22_error)
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_7genexpr__pyx_v_k); __pyx_7genexpr__pyx_v_k = 0;
    __Pyx_XDECREF(__pyx_7genexpr__pyx_v_v); __pyx_7genexpr__pyx_v_v = 0;
    goto __pyx_L28_exit_scope;
    __pyx_L22_error:;
    __Pyx_XDECREF(__pyx_7genexpr__pyx_v_k); __pyx_7genexpr__pyx_v_k = 0;
    __Pyx_XDECREF(__pyx_7genexpr__pyx_v_v); __pyx_7genexpr__pyx_v_v = 0;
    goto __pyx_L1_error;
    __pyx_L28_exit_scope:;
  } /* exit inner scope */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_metadata, __pyx_n_s_column_names_to_labels, __pyx_t_2) < 0) __PYX_ERR(0, 1018, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyreadstat/_readstat_parser.pyx":1019
 *     metadata.column_labels = data.col_labels
 *     metadata.column_names_to_labels = {k:v for k,v in zip(data.col_names, data.col_labels)}
 *     metadata.file_encoding = data.file_encoding             # <<<<<<<<<<<<<<
 *     metadata.file_label = data.file_label
 *     metadata.variable_value_labels = variable_value_labels
 */
  __pyx_t_2 = __pyx_v_data->file_encoding;
  __Pyx_INCREF(__pyx_t_2);
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_metadata, __pyx_n_s_file_encoding, __pyx_t_2) < 0) __PYX_ERR(0, 1019, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyreadstat/_readstat_parser.pyx":1020
 *     metadata.column_names_to_labels = {k:v for k,v in zip(data.col_names, data.col_labels)}
 *     metadata.file_encoding = data.file_encoding
 *     metadata.file_label = data.file_label             # <<<<<<<<<<<<<<
 *     metadata.variable_value_labels = variable_value_labels
 *     metadata.value_labels = labels_raw
 */
  __pyx_t_2 = __pyx_v_data->file_label;
  __Pyx_INCREF(__pyx_t_2);
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_metadata, __pyx_n_s_file_label, __pyx_t_2) < 0) __PYX_ERR(0, 1020, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyreadstat/_readstat_parser.pyx":1021
 *     metadata.file_encoding = data.file_encoding
 *     metadata.file_label = data.file_label
 *     metadata.variable_value_labels = variable_value_labels             # <<<<<<<<<<<<<<
 *     metadata.value_labels = labels_raw
 *     metadata.variable_to_label = label_to_var_name
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_metadata, __pyx_n_s_variable_value_labels, __pyx_v_variable_value_labels) < 0) __PYX_ERR(0, 1021, __pyx_L1_error)

  /* "pyreadstat/_readstat_parser.pyx":1022
 *     metadata.file_label = data.file_label
 *     metadata.variable_value_labels = variable_value_labels
 *     metadata.value_labels = labels_raw             # <<<<<<<<<<<<<<
 *     metadata.variable_to_label = label_to_var_name
 *     metadata.original_variable_types = original_types
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_metadata, __pyx_n_s_value_labels, __pyx_v_labels_raw) < 0) __PYX_ERR(0, 1022, __pyx_L1_error)

  /* "pyreadstat/_readstat_parser.pyx":1023
 *     metadata.variable_value_labels = variable_value_labels
 *     metadata.value_labels = labels_raw
 *     metadata.variable_to_label = label_to_var_name             # <<<<<<<<<<<<<<
 *     metadata.original_variable_types = original_types
 *     metadata.readstat_variable_types = readstat_types
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_metadata, __pyx_n_s_variable_to_label, __pyx_v_label_to_var_name) < 0) __PYX_ERR(0, 1023, __pyx_L1_error)

  /* "pyreadstat/_readstat_parser.pyx":1024
 *     metadata.value_labels = labels_raw
 *     metadata.variable_to_label = label_to_var_name
 *     metadata.original_variable_types = original_types             # <<<<<<<<<<<<<<
 *     metadata.readstat_variable_types = readstat_types
 *     metadata.table_name = data.table_name
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_metadata, __pyx_n_s_original_variable_types, __pyx_v_original_types) < 0) __PYX_ERR(0, 1024, __pyx_L1_error)

  /* "pyreadstat/_readstat_parser.pyx":1025
 *     metadata.variable_to_label = label_to_var_name
 *     metadata.original_variable_types = original_types
 *     metadata.readstat_variable_types = readstat_types             # <<<<<<<<<<<<<<
 *     metadata.table_name = data.table_name
 *     metadata.missing_ranges = data.missing_ranges
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_metadata, __pyx_n_s_readstat_variable_types, __pyx_v_readstat_types) < 0) __PYX_ERR(0, 1025, __pyx_L1_error)

  /* "pyreadstat/_readstat_parser.pyx":1026
 *     metadata.original_variable_types = original_types
 *     metadata.readstat_variable_types = readstat_types
 *     metadata.table_name = data.table_name             # <<<<<<<<<<<<<<
 *     metadata.missing_ranges = data.missing_ranges
 *     metadata.variable_storage_width = data.variable_storage_width
 */
  __pyx_t_2 = __pyx_v_data->table_name;
  __Pyx_INCREF(__pyx_t_2);
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_metadata, __pyx_n_s_table_name, __pyx_t_2) < 0) __PYX_ERR(0, 1026, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyreadstat/_readstat_parser.pyx":1027
 *     metadata.readstat_variable_types = readstat_types
 *     metadata.table_name = data.table_name
 *     metadata.missing_ranges = data.missing_ranges             # <<<<<<<<<<<<<<
 *     metadata.variable_storage_width = data.variable_storage_width
 *     metadata.variable_display_width = data.variable_display_width
 */
  __pyx_t_2 = __pyx_v_data->missing_ranges;
  __Pyx_INCREF(__pyx_t_2);
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_metadata, __pyx_n_s_missing_ranges, __pyx_t_2) < 0) __PYX_ERR(0, 1027, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyreadstat/_readstat_parser.pyx":1028
 *     metadata.table_name = data.table_name
 *     metadata.missing_ranges = data.missing_ranges
 *     metadata.variable_storage_width = data.variable_storage_width             # <<<<<<<<<<<<<<
 *     metadata.variable_display_width = data.variable_display_width
 *     metadata.variable_alignment = data.variable_alignment
 */
  __pyx_t_2 = __pyx_v_data->variable_storage_width;
  __Pyx_INCREF(__pyx_t_2);
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_metadata, __pyx_n_s_variable_storage_width, __pyx_t_2) < 0) __PYX_ERR(0, 1028, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyreadstat/_readstat_parser.pyx":1029
 *     metadata.missing_ranges = data.missing_ranges
 *     metadata.variable_storage_width = data.variable_storage_width
 *     metadata.variable_display_width = data.variable_display_width             # <<<<<<<<<<<<<<
 *     metadata.variable_alignment = data.variable_alignment
 *     metadata.variable_measure = data.variable_measure
 */
  __pyx_t_2 = __pyx_v_data->variable_display_width;
  __Pyx_INCREF(__pyx_t_2);
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_metadata, __pyx_n_s_variable_display_width, __pyx_t_2) < 0) __PYX_ERR(0, 1029, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyreadstat/_readstat_parser.pyx":1030
 *     metadata.variable_storage_width = data.variable_storage_width
 *     metadata.variable_display_width = data.variable_display_width
 *     metadata.variable_alignment = data.variable_alignment             # <<<<<<<<<<<<<<
 *     metadata.variable_measure = data.variable_measure
 *     metadata.creation_time = datetime.datetime.fromtimestamp(data.ctime)
 */
  __pyx_t_2 = __pyx_v_data->variable_alignment;
  __Pyx_INCREF(__pyx_t_2);
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_metadata, __pyx_n_s_variable_alignment, __pyx_t_2) < 0) __PYX_ERR(0, 1030, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyreadstat/_readstat_parser.pyx":1031
 *     metadata.variable_display_width = data.variable_display_width
 *     metadata.variable_alignment = data.variable_alignment
 *     metadata.variable_measure = data.variable_measure             # <<<<<<<<<<<<<<
 *     metadata.creation_time = datetime.datetime.fromtimestamp(data.ctime)
 *     metadata.modification_time = datetime.datetime.fromtimestamp(data.mtime)
 */
  __pyx_t_2 = __pyx_v_data->variable_measure;
  __Pyx_INCREF(__pyx_t_2);
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_metadata, __pyx_n_s_variable_measure, __pyx_t_2) < 0) __PYX_ERR(0, 1031, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyreadstat/_readstat_parser.pyx":1032
 *     metadata.variable_alignment = data.variable_alignment
 *     metadata.variable_measure = data.variable_measure
 *     metadata.creation_time = datetime.datetime.fromtimestamp(data.ctime)             # <<<<<<<<<<<<<<
 *     metadata.modification_time = datetime.datetime.fromtimestamp(data.mtime)
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_datetime); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1032, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_datetime); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1032, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_fromtimestamp); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1032, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_PyInt_From_int(__pyx_v_data->ctime); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1032, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_9)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_9);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_9, __pyx_t_8};
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1032, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_metadata, __pyx_n_s_creation_time, __pyx_t_2) < 0) __PYX_ERR(0, 1032, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyreadstat/_readstat_parser.pyx":1033
 *     metadata.variable_measure = data.variable_measure
 *     metadata.creation_time = datetime.datetime.fromtimestamp(data.ctime)
 *     metadata.modification_time = datetime.datetime.fromtimestamp(data.mtime)             # <<<<<<<<<<<<<<
 * 
 *     return metadata
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_datetime); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1033, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_datetime); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1033, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_fromtimestamp); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1033, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_PyInt_From_int(__pyx_v_data->mtime); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1033, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_9)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_9);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_9, __pyx_t_8};
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1033, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_metadata, __pyx_n_s_modification_time, __pyx_t_2) < 0) __PYX_ERR(0, 1033, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyreadstat/_readstat_parser.pyx":1035
 *     metadata.modification_time = datetime.datetime.fromtimestamp(data.mtime)
 * 
 *     return metadata             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_metadata);
  __pyx_r = __pyx_v_metadata;
  goto __pyx_L0;

  /* "pyreadstat/_readstat_parser.pyx":948
 * 
 * 
 * cdef object data_container_extract_metadata(data_container data):             # <<<<<<<<<<<<<<
 *     """
 *     Extracts metadata from a data container and puts it into a metadata
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_AddTraceback("pyreadstat._readstat_parser.data_container_extract_metadata", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_label_to_var_name);
  __Pyx_XDECREF(__pyx_v_labels_raw);
  __Pyx_XDECREF(__pyx_v_var_name);
  __Pyx_XDECREF(__pyx_v_var_label);
  __Pyx_XDECREF(__pyx_v_current_labels);
  __Pyx_XDECREF(__pyx_v_original_types);
  __Pyx_XDECREF(__pyx_v_metadata);
  __Pyx_XDECREF(__pyx_v_variable_value_labels);
  __Pyx_XDECREF(__pyx_v_readstat_types);
  __Pyx_XDECREF(__pyx_v_indx);
  __Pyx_XDECREF(__pyx_v_cur_col);
  __Pyx_XDECREF(__pyx_v_cur_type);
  __Pyx_XDECREF(__pyx_v_curset);
  __Pyx_XDECREF(__pyx_7genexpr__pyx_v_k);
  __Pyx_XDECREF(__pyx_7genexpr__pyx_v_v);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyreadstat/_readstat_parser.pyx":1038
 * 
 * 
 * cdef object run_conversion(object filename_path, py_file_format file_format, py_file_extension file_extension,             # <<<<<<<<<<<<<<
 *                            str encoding, bint metaonly, bint dates_as_pandas, list usecols, bint usernan,
 *                            bint no_datetime_conversion, long row_limit, long row_offset, str output_format,
 */

static PyObject *__pyx_f_10pyreadstat_16_readstat_parser_run_conversion(PyObject *__pyx_v_filename_path, __pyx_t_10pyreadstat_16_readstat_parser_py_file_format __pyx_v_file_format, __pyx_t_10pyreadstat_16_readstat_parser_py_file_extension __pyx_v_file_extension, PyObject *__pyx_v_encoding, int __pyx_v_metaonly, int __pyx_v_dates_as_pandas, PyObject *__pyx_v_usecols, int __pyx_v_usernan, int __pyx_v_no_datetime_conversion, long __pyx_v_row_limit, long __pyx_v_row_offset, PyObject *__pyx_v_output_format, PyObject *__pyx_v_extra_datetime_formats, PyObject *__pyx_v_extra_date_formats) {
  PyObject *__pyx_v_filename_bytes = 0;
  char *__pyx_v_filename;
  struct __pyx_obj_10pyreadstat_16_readstat_parser_data_container *__pyx_v_data = 0;
  PyObject *__pyx_v_origin = 0;
  PyObject *__pyx_v_allowed_formats = 0;
  PyObject *__pyx_v_data_dict = 0;
  PyObject *__pyx_v_data_frame = 0;
  CYTHON_UNUSED void *__pyx_v_ctx;
  PyObject *__pyx_v_metadata = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  int __pyx_t_14;
  int __pyx_t_15;
  char *__pyx_t_16;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("run_conversion", 0);
  __Pyx_INCREF(__pyx_v_output_format);

  /* "pyreadstat/_readstat_parser.pyx":1055
 *     cdef object data_frame
 * 
 *     if hasattr(os, 'fsencode'):             # <<<<<<<<<<<<<<
 *         try:
 *             filename_bytes = os.fsencode(filename_path)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_os); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1055, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_HasAttr(__pyx_t_1, __pyx_n_s_fsencode); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(0, 1055, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "pyreadstat/_readstat_parser.pyx":1056
 * 
 *     if hasattr(os, 'fsencode'):
 *         try:             # <<<<<<<<<<<<<<
 *             filename_bytes = os.fsencode(filename_path)
 *         except UnicodeError:
 */
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_3, &__pyx_t_4, &__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_3);
      __Pyx_XGOTREF(__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_5);
      /*try:*/ {

        /* "pyreadstat/_readstat_parser.pyx":1057
 *     if hasattr(os, 'fsencode'):
 *         try:
 *             filename_bytes = os.fsencode(filename_path)             # <<<<<<<<<<<<<<
 *         except UnicodeError:
 *             warnings.warn("file path could not be encoded with %s which is set as your system encoding, trying to encode it as utf-8. Please set your system encoding correctly." % sys.getfilesystemencoding())
 */
        __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_os); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1057, __pyx_L4_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_fsencode); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1057, __pyx_L4_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_t_6 = NULL;
        __pyx_t_8 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_6)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_6);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
            __pyx_t_8 = 1;
          }
        }
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_6, __pyx_v_filename_path};
          __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_7, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
          __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
          if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1057, __pyx_L4_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        if (!(likely(PyBytes_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("bytes", __pyx_t_1))) __PYX_ERR(0, 1057, __pyx_L4_error)
        __pyx_v_filename_bytes = ((PyObject*)__pyx_t_1);
        __pyx_t_1 = 0;

        /* "pyreadstat/_readstat_parser.pyx":1056
 * 
 *     if hasattr(os, 'fsencode'):
 *         try:             # <<<<<<<<<<<<<<
 *             filename_bytes = os.fsencode(filename_path)
 *         except UnicodeError:
 */
      }
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      goto __pyx_L9_try_end;
      __pyx_L4_error:;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;

      /* "pyreadstat/_readstat_parser.pyx":1058
 *         try:
 *             filename_bytes = os.fsencode(filename_path)
 *         except UnicodeError:             # <<<<<<<<<<<<<<
 *             warnings.warn("file path could not be encoded with %s which is set as your system encoding, trying to encode it as utf-8. Please set your system encoding correctly." % sys.getfilesystemencoding())
 *             filename_bytes = os.fsdecode(filename_path).encode("utf-8", "surrogateescape")
 */
      __pyx_t_8 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_UnicodeError);
      if (__pyx_t_8) {
        __Pyx_AddTraceback("pyreadstat._readstat_parser.run_conversion", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_7, &__pyx_t_6) < 0) __PYX_ERR(0, 1058, __pyx_L6_except_error)
        __Pyx_XGOTREF(__pyx_t_1);
        __Pyx_XGOTREF(__pyx_t_7);
        __Pyx_XGOTREF(__pyx_t_6);

        /* "pyreadstat/_readstat_parser.pyx":1059
 *             filename_bytes = os.fsencode(filename_path)
 *         except UnicodeError:
 *             warnings.warn("file path could not be encoded with %s which is set as your system encoding, trying to encode it as utf-8. Please set your system encoding correctly." % sys.getfilesystemencoding())             # <<<<<<<<<<<<<<
 *             filename_bytes = os.fsdecode(filename_path).encode("utf-8", "surrogateescape")
 *     else:
 */
        __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_n_s_warnings); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1059, __pyx_L6_except_error)
        __Pyx_GOTREF(__pyx_t_10);
        __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_warn); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1059, __pyx_L6_except_error)
        __Pyx_GOTREF(__pyx_t_11);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_GetModuleGlobalName(__pyx_t_12, __pyx_n_s_sys); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1059, __pyx_L6_except_error)
        __Pyx_GOTREF(__pyx_t_12);
        __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_t_12, __pyx_n_s_getfilesystemencoding); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1059, __pyx_L6_except_error)
        __Pyx_GOTREF(__pyx_t_13);
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        __pyx_t_12 = NULL;
        __pyx_t_8 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_13))) {
          __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_13);
          if (likely(__pyx_t_12)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
            __Pyx_INCREF(__pyx_t_12);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_13, function);
            __pyx_t_8 = 1;
          }
        }
        {
          PyObject *__pyx_callargs[1] = {__pyx_t_12, };
          __pyx_t_10 = __Pyx_PyObject_FastCall(__pyx_t_13, __pyx_callargs+1-__pyx_t_8, 0+__pyx_t_8);
          __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
          if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1059, __pyx_L6_except_error)
          __Pyx_GOTREF(__pyx_t_10);
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
        }
        __pyx_t_13 = __Pyx_PyString_FormatSafe(__pyx_kp_s_file_path_could_not_be_encoded_w, __pyx_t_10); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1059, __pyx_L6_except_error)
        __Pyx_GOTREF(__pyx_t_13);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        __pyx_t_10 = NULL;
        __pyx_t_8 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_11))) {
          __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_11);
          if (likely(__pyx_t_10)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
            __Pyx_INCREF(__pyx_t_10);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_11, function);
            __pyx_t_8 = 1;
          }
        }
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_10, __pyx_t_13};
          __pyx_t_9 = __Pyx_PyObject_FastCall(__pyx_t_11, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
          __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1059, __pyx_L6_except_error)
          __Pyx_GOTREF(__pyx_t_9);
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        }
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

        /* "pyreadstat/_readstat_parser.pyx":1060
 *         except UnicodeError:
 *             warnings.warn("file path could not be encoded with %s which is set as your system encoding, trying to encode it as utf-8. Please set your system encoding correctly." % sys.getfilesystemencoding())
 *             filename_bytes = os.fsdecode(filename_path).encode("utf-8", "surrogateescape")             # <<<<<<<<<<<<<<
 *     else:
 *         if type(filename_path) == str:
 */
        __Pyx_GetModuleGlobalName(__pyx_t_11, __pyx_n_s_os); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1060, __pyx_L6_except_error)
        __Pyx_GOTREF(__pyx_t_11);
        __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_t_11, __pyx_n_s_fsdecode); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1060, __pyx_L6_except_error)
        __Pyx_GOTREF(__pyx_t_13);
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        __pyx_t_11 = NULL;
        __pyx_t_8 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_13))) {
          __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_13);
          if (likely(__pyx_t_11)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
            __Pyx_INCREF(__pyx_t_11);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_13, function);
            __pyx_t_8 = 1;
          }
        }
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_11, __pyx_v_filename_path};
          __pyx_t_9 = __Pyx_PyObject_FastCall(__pyx_t_13, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
          __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
          if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1060, __pyx_L6_except_error)
          __Pyx_GOTREF(__pyx_t_9);
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
        }
        __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_encode); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1060, __pyx_L6_except_error)
        __Pyx_GOTREF(__pyx_t_13);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __pyx_t_9 = __Pyx_PyObject_Call(__pyx_t_13, __pyx_tuple__7, NULL); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1060, __pyx_L6_except_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
        if (!(likely(PyBytes_CheckExact(__pyx_t_9))||((__pyx_t_9) == Py_None) || __Pyx_RaiseUnexpectedTypeError("bytes", __pyx_t_9))) __PYX_ERR(0, 1060, __pyx_L6_except_error)
        __Pyx_XDECREF_SET(__pyx_v_filename_bytes, ((PyObject*)__pyx_t_9));
        __pyx_t_9 = 0;
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        goto __pyx_L5_exception_handled;
      }
      goto __pyx_L6_except_error;

      /* "pyreadstat/_readstat_parser.pyx":1056
 * 
 *     if hasattr(os, 'fsencode'):
 *         try:             # <<<<<<<<<<<<<<
 *             filename_bytes = os.fsencode(filename_path)
 *         except UnicodeError:
 */
      __pyx_L6_except_error:;
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_4, __pyx_t_5);
      goto __pyx_L1_error;
      __pyx_L5_exception_handled:;
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_4, __pyx_t_5);
      __pyx_L9_try_end:;
    }

    /* "pyreadstat/_readstat_parser.pyx":1055
 *     cdef object data_frame
 * 
 *     if hasattr(os, 'fsencode'):             # <<<<<<<<<<<<<<
 *         try:
 *             filename_bytes = os.fsencode(filename_path)
 */
    goto __pyx_L3;
  }

  /* "pyreadstat/_readstat_parser.pyx":1062
 *             filename_bytes = os.fsdecode(filename_path).encode("utf-8", "surrogateescape")
 *     else:
 *         if type(filename_path) == str:             # <<<<<<<<<<<<<<
 *             filename_bytes = filename_path.encode('utf-8')
 *         elif type(filename_path) == bytes:
 */
  /*else*/ {
    __pyx_t_6 = PyObject_RichCompare(((PyObject *)Py_TYPE(__pyx_v_filename_path)), ((PyObject *)(&PyString_Type)), Py_EQ); __Pyx_XGOTREF(__pyx_t_6); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1062, __pyx_L1_error)
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_6); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 1062, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (__pyx_t_2) {

      /* "pyreadstat/_readstat_parser.pyx":1063
 *     else:
 *         if type(filename_path) == str:
 *             filename_bytes = filename_path.encode('utf-8')             # <<<<<<<<<<<<<<
 *         elif type(filename_path) == bytes:
 *             filename_bytes = filename_path
 */
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_filename_path, __pyx_n_s_encode); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1063, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_1 = NULL;
      __pyx_t_8 = 0;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_7);
        if (likely(__pyx_t_1)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_7, function);
          __pyx_t_8 = 1;
        }
      }
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_1, __pyx_kp_s_utf_8};
        __pyx_t_6 = __Pyx_PyObject_FastCall(__pyx_t_7, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1063, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      }
      if (!(likely(PyBytes_CheckExact(__pyx_t_6))||((__pyx_t_6) == Py_None) || __Pyx_RaiseUnexpectedTypeError("bytes", __pyx_t_6))) __PYX_ERR(0, 1063, __pyx_L1_error)
      __pyx_v_filename_bytes = ((PyObject*)__pyx_t_6);
      __pyx_t_6 = 0;

      /* "pyreadstat/_readstat_parser.pyx":1062
 *             filename_bytes = os.fsdecode(filename_path).encode("utf-8", "surrogateescape")
 *     else:
 *         if type(filename_path) == str:             # <<<<<<<<<<<<<<
 *             filename_bytes = filename_path.encode('utf-8')
 *         elif type(filename_path) == bytes:
 */
      goto __pyx_L12;
    }

    /* "pyreadstat/_readstat_parser.pyx":1064
 *         if type(filename_path) == str:
 *             filename_bytes = filename_path.encode('utf-8')
 *         elif type(filename_path) == bytes:             # <<<<<<<<<<<<<<
 *             filename_bytes = filename_path
 *         else:
 */
    __pyx_t_6 = PyObject_RichCompare(((PyObject *)Py_TYPE(__pyx_v_filename_path)), ((PyObject *)(&PyBytes_Type)), Py_EQ); __Pyx_XGOTREF(__pyx_t_6); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1064, __pyx_L1_error)
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_6); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 1064, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (likely(__pyx_t_2)) {

      /* "pyreadstat/_readstat_parser.pyx":1065
 *             filename_bytes = filename_path.encode('utf-8')
 *         elif type(filename_path) == bytes:
 *             filename_bytes = filename_path             # <<<<<<<<<<<<<<
 *         else:
 *             raise PyreadstatError("path must be either str or bytes")
 */
      if (!(likely(PyBytes_CheckExact(__pyx_v_filename_path))||((__pyx_v_filename_path) == Py_None) || __Pyx_RaiseUnexpectedTypeError("bytes", __pyx_v_filename_path))) __PYX_ERR(0, 1065, __pyx_L1_error)
      __pyx_t_6 = __pyx_v_filename_path;
      __Pyx_INCREF(__pyx_t_6);
      __pyx_v_filename_bytes = ((PyObject*)__pyx_t_6);
      __pyx_t_6 = 0;

      /* "pyreadstat/_readstat_parser.pyx":1064
 *         if type(filename_path) == str:
 *             filename_bytes = filename_path.encode('utf-8')
 *         elif type(filename_path) == bytes:             # <<<<<<<<<<<<<<
 *             filename_bytes = filename_path
 *         else:
 */
      goto __pyx_L12;
    }

    /* "pyreadstat/_readstat_parser.pyx":1067
 *             filename_bytes = filename_path
 *         else:
 *             raise PyreadstatError("path must be either str or bytes")             # <<<<<<<<<<<<<<
 *         if type(filename_path) not in (str, bytes, unicode):
 *             raise PyreadstatError("path must be str, bytes or unicode")
 */
    /*else*/ {
      __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_PyreadstatError); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1067, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_1 = NULL;
      __pyx_t_8 = 0;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_7);
        if (likely(__pyx_t_1)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_7, function);
          __pyx_t_8 = 1;
        }
      }
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_1, __pyx_kp_s_path_must_be_either_str_or_bytes};
        __pyx_t_6 = __Pyx_PyObject_FastCall(__pyx_t_7, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1067, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      }
      __Pyx_Raise(__pyx_t_6, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __PYX_ERR(0, 1067, __pyx_L1_error)
    }
    __pyx_L12:;

    /* "pyreadstat/_readstat_parser.pyx":1068
 *         else:
 *             raise PyreadstatError("path must be either str or bytes")
 *         if type(filename_path) not in (str, bytes, unicode):             # <<<<<<<<<<<<<<
 *             raise PyreadstatError("path must be str, bytes or unicode")
 *         filename_bytes = filename_path.encode('utf-8')
 */
    __Pyx_INCREF(((PyObject *)Py_TYPE(__pyx_v_filename_path)));
    __pyx_t_6 = ((PyObject *)Py_TYPE(__pyx_v_filename_path));
    __pyx_t_7 = PyObject_RichCompare(((PyObject *)__pyx_t_6), ((PyObject *)(&PyString_Type)), Py_NE); __Pyx_XGOTREF(__pyx_t_7); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1068, __pyx_L1_error)
    __pyx_t_14 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely((__pyx_t_14 < 0))) __PYX_ERR(0, 1068, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (__pyx_t_14) {
    } else {
      __pyx_t_2 = __pyx_t_14;
      goto __pyx_L14_bool_binop_done;
    }
    __pyx_t_7 = PyObject_RichCompare(((PyObject *)__pyx_t_6), ((PyObject *)(&PyBytes_Type)), Py_NE); __Pyx_XGOTREF(__pyx_t_7); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1068, __pyx_L1_error)
    __pyx_t_14 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely((__pyx_t_14 < 0))) __PYX_ERR(0, 1068, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (__pyx_t_14) {
    } else {
      __pyx_t_2 = __pyx_t_14;
      goto __pyx_L14_bool_binop_done;
    }
    __pyx_t_7 = PyObject_RichCompare(((PyObject *)__pyx_t_6), ((PyObject *)(&PyUnicode_Type)), Py_NE); __Pyx_XGOTREF(__pyx_t_7); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1068, __pyx_L1_error)
    __pyx_t_14 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely((__pyx_t_14 < 0))) __PYX_ERR(0, 1068, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_2 = __pyx_t_14;
    __pyx_L14_bool_binop_done:;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_14 = __pyx_t_2;
    if (unlikely(__pyx_t_14)) {

      /* "pyreadstat/_readstat_parser.pyx":1069
 *             raise PyreadstatError("path must be either str or bytes")
 *         if type(filename_path) not in (str, bytes, unicode):
 *             raise PyreadstatError("path must be str, bytes or unicode")             # <<<<<<<<<<<<<<
 *         filename_bytes = filename_path.encode('utf-8')
 * 
 */
      __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_PyreadstatError); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1069, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_1 = NULL;
      __pyx_t_8 = 0;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_7);
        if (likely(__pyx_t_1)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_7, function);
          __pyx_t_8 = 1;
        }
      }
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_1, __pyx_kp_s_path_must_be_str_bytes_or_unicod};
        __pyx_t_6 = __Pyx_PyObject_FastCall(__pyx_t_7, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1069, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      }
      __Pyx_Raise(__pyx_t_6, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __PYX_ERR(0, 1069, __pyx_L1_error)

      /* "pyreadstat/_readstat_parser.pyx":1068
 *         else:
 *             raise PyreadstatError("path must be either str or bytes")
 *         if type(filename_path) not in (str, bytes, unicode):             # <<<<<<<<<<<<<<
 *             raise PyreadstatError("path must be str, bytes or unicode")
 *         filename_bytes = filename_path.encode('utf-8')
 */
    }

    /* "pyreadstat/_readstat_parser.pyx":1070
 *         if type(filename_path) not in (str, bytes, unicode):
 *             raise PyreadstatError("path must be str, bytes or unicode")
 *         filename_bytes = filename_path.encode('utf-8')             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_filename_path, __pyx_n_s_encode); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1070, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_1 = NULL;
    __pyx_t_8 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_1)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
        __pyx_t_8 = 1;
      }
    }
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_1, __pyx_kp_s_utf_8};
      __pyx_t_6 = __Pyx_PyObject_FastCall(__pyx_t_7, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1070, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    }
    if (!(likely(PyBytes_CheckExact(__pyx_t_6))||((__pyx_t_6) == Py_None) || __Pyx_RaiseUnexpectedTypeError("bytes", __pyx_t_6))) __PYX_ERR(0, 1070, __pyx_L1_error)
    __Pyx_DECREF_SET(__pyx_v_filename_bytes, ((PyObject*)__pyx_t_6));
    __pyx_t_6 = 0;
  }
  __pyx_L3:;

  /* "pyreadstat/_readstat_parser.pyx":1073
 * 
 * 
 *     filename_bytes = os.path.expanduser(filename_bytes)             # <<<<<<<<<<<<<<
 *     if not os.path.isfile(filename_bytes):
 *         raise PyreadstatError("File {0} does not exist!".format(filename_path))
 */
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_os); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1073, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_path); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1073, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_expanduser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1073, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = NULL;
  __pyx_t_8 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_7);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_7, function);
      __pyx_t_8 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_1, __pyx_v_filename_bytes};
    __pyx_t_6 = __Pyx_PyObject_FastCall(__pyx_t_7, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1073, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  if (!(likely(PyBytes_CheckExact(__pyx_t_6))||((__pyx_t_6) == Py_None) || __Pyx_RaiseUnexpectedTypeError("bytes", __pyx_t_6))) __PYX_ERR(0, 1073, __pyx_L1_error)
  __Pyx_DECREF_SET(__pyx_v_filename_bytes, ((PyObject*)__pyx_t_6));
  __pyx_t_6 = 0;

  /* "pyreadstat/_readstat_parser.pyx":1074
 * 
 *     filename_bytes = os.path.expanduser(filename_bytes)
 *     if not os.path.isfile(filename_bytes):             # <<<<<<<<<<<<<<
 *         raise PyreadstatError("File {0} does not exist!".format(filename_path))
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_os); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1074, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_path); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1074, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_isfile); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1074, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = NULL;
  __pyx_t_8 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_7);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_7, function);
      __pyx_t_8 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_1, __pyx_v_filename_bytes};
    __pyx_t_6 = __Pyx_PyObject_FastCall(__pyx_t_7, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1074, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __pyx_t_14 = __Pyx_PyObject_IsTrue(__pyx_t_6); if (unlikely((__pyx_t_14 < 0))) __PYX_ERR(0, 1074, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_2 = (!__pyx_t_14);
  if (unlikely(__pyx_t_2)) {

    /* "pyreadstat/_readstat_parser.pyx":1075
 *     filename_bytes = os.path.expanduser(filename_bytes)
 *     if not os.path.isfile(filename_bytes):
 *         raise PyreadstatError("File {0} does not exist!".format(filename_path))             # <<<<<<<<<<<<<<
 * 
 *     if output_format is None:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_PyreadstatError); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1075, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_kp_s_File_0_does_not_exist, __pyx_n_s_format); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1075, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_13 = NULL;
    __pyx_t_8 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_9))) {
      __pyx_t_13 = PyMethod_GET_SELF(__pyx_t_9);
      if (likely(__pyx_t_13)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
        __Pyx_INCREF(__pyx_t_13);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_9, function);
        __pyx_t_8 = 1;
      }
    }
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_13, __pyx_v_filename_path};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_9, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1075, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    }
    __pyx_t_9 = NULL;
    __pyx_t_8 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
        __pyx_t_8 = 1;
      }
    }
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_9, __pyx_t_1};
      __pyx_t_6 = __Pyx_PyObject_FastCall(__pyx_t_7, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1075, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    }
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(0, 1075, __pyx_L1_error)

    /* "pyreadstat/_readstat_parser.pyx":1074
 * 
 *     filename_bytes = os.path.expanduser(filename_bytes)
 *     if not os.path.isfile(filename_bytes):             # <<<<<<<<<<<<<<
 *         raise PyreadstatError("File {0} does not exist!".format(filename_path))
 * 
 */
  }

  /* "pyreadstat/_readstat_parser.pyx":1077
 *         raise PyreadstatError("File {0} does not exist!".format(filename_path))
 * 
 *     if output_format is None:             # <<<<<<<<<<<<<<
 *         output_format = 'pandas'
 *     allowed_formats = {'pandas', 'dict'}
 */
  __pyx_t_2 = (__pyx_v_output_format == ((PyObject*)Py_None));
  if (__pyx_t_2) {

    /* "pyreadstat/_readstat_parser.pyx":1078
 * 
 *     if output_format is None:
 *         output_format = 'pandas'             # <<<<<<<<<<<<<<
 *     allowed_formats = {'pandas', 'dict'}
 *     if output_format not in allowed_formats:
 */
    __Pyx_INCREF(__pyx_n_s_pandas);
    __Pyx_DECREF_SET(__pyx_v_output_format, __pyx_n_s_pandas);

    /* "pyreadstat/_readstat_parser.pyx":1077
 *         raise PyreadstatError("File {0} does not exist!".format(filename_path))
 * 
 *     if output_format is None:             # <<<<<<<<<<<<<<
 *         output_format = 'pandas'
 *     allowed_formats = {'pandas', 'dict'}
 */
  }

  /* "pyreadstat/_readstat_parser.pyx":1079
 *     if output_format is None:
 *         output_format = 'pandas'
 *     allowed_formats = {'pandas', 'dict'}             # <<<<<<<<<<<<<<
 *     if output_format not in allowed_formats:
 *         raise PyreadstatError("output format must be one of {allowed_formats}, '{output_format}' was given".format(allowed_formats=allowed_formats, output_format=output_format))
 */
  __pyx_t_6 = PySet_New(0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1079, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (PySet_Add(__pyx_t_6, __pyx_n_s_pandas) < 0) __PYX_ERR(0, 1079, __pyx_L1_error)
  if (PySet_Add(__pyx_t_6, __pyx_n_s_dict) < 0) __PYX_ERR(0, 1079, __pyx_L1_error)
  __pyx_v_allowed_formats = ((PyObject*)__pyx_t_6);
  __pyx_t_6 = 0;

  /* "pyreadstat/_readstat_parser.pyx":1080
 *         output_format = 'pandas'
 *     allowed_formats = {'pandas', 'dict'}
 *     if output_format not in allowed_formats:             # <<<<<<<<<<<<<<
 *         raise PyreadstatError("output format must be one of {allowed_formats}, '{output_format}' was given".format(allowed_formats=allowed_formats, output_format=output_format))
 * 
 */
  __pyx_t_2 = (__Pyx_PySet_ContainsTF(__pyx_v_output_format, __pyx_v_allowed_formats, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 1080, __pyx_L1_error)
  if (unlikely(__pyx_t_2)) {

    /* "pyreadstat/_readstat_parser.pyx":1081
 *     allowed_formats = {'pandas', 'dict'}
 *     if output_format not in allowed_formats:
 *         raise PyreadstatError("output format must be one of {allowed_formats}, '{output_format}' was given".format(allowed_formats=allowed_formats, output_format=output_format))             # <<<<<<<<<<<<<<
 * 
 *     if extra_date_formats is not None:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_PyreadstatError); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1081, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_kp_s_output_format_must_be_one_of_all, __pyx_n_s_format); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1081, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_9 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1081, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_allowed_formats, __pyx_v_allowed_formats) < 0) __PYX_ERR(0, 1081, __pyx_L1_error)
    if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_output_format, __pyx_v_output_format) < 0) __PYX_ERR(0, 1081, __pyx_L1_error)
    __pyx_t_13 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_empty_tuple, __pyx_t_9); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1081, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_9 = NULL;
    __pyx_t_8 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
        __pyx_t_8 = 1;
      }
    }
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_9, __pyx_t_13};
      __pyx_t_6 = __Pyx_PyObject_FastCall(__pyx_t_7, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
      if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1081, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    }
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(0, 1081, __pyx_L1_error)

    /* "pyreadstat/_readstat_parser.pyx":1080
 *         output_format = 'pandas'
 *     allowed_formats = {'pandas', 'dict'}
 *     if output_format not in allowed_formats:             # <<<<<<<<<<<<<<
 *         raise PyreadstatError("output format must be one of {allowed_formats}, '{output_format}' was given".format(allowed_formats=allowed_formats, output_format=output_format))
 * 
 */
  }

  /* "pyreadstat/_readstat_parser.pyx":1083
 *         raise PyreadstatError("output format must be one of {allowed_formats}, '{output_format}' was given".format(allowed_formats=allowed_formats, output_format=output_format))
 * 
 *     if extra_date_formats is not None:             # <<<<<<<<<<<<<<
 *         if file_format == FILE_FORMAT_SAS:
 *             sas_date_formats.extend(extra_date_formats)
 */
  __pyx_t_2 = (__pyx_v_extra_date_formats != ((PyObject*)Py_None));
  if (__pyx_t_2) {

    /* "pyreadstat/_readstat_parser.pyx":1084
 * 
 *     if extra_date_formats is not None:
 *         if file_format == FILE_FORMAT_SAS:             # <<<<<<<<<<<<<<
 *             sas_date_formats.extend(extra_date_formats)
 *         elif file_format == FILE_FORMAT_SPSS:
 */
    switch (__pyx_v_file_format) {
      case __pyx_e_10pyreadstat_16_readstat_parser_FILE_FORMAT_SAS:

      /* "pyreadstat/_readstat_parser.pyx":1085
 *     if extra_date_formats is not None:
 *         if file_format == FILE_FORMAT_SAS:
 *             sas_date_formats.extend(extra_date_formats)             # <<<<<<<<<<<<<<
 *         elif file_format == FILE_FORMAT_SPSS:
 *             spss_date_formats.extend(extra_date_formats)
 */
      if (unlikely(__pyx_v_10pyreadstat_16_readstat_parser_sas_date_formats == Py_None)) {
        PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "extend");
        __PYX_ERR(0, 1085, __pyx_L1_error)
      }
      __pyx_t_15 = __Pyx_PyList_Extend(__pyx_v_10pyreadstat_16_readstat_parser_sas_date_formats, __pyx_v_extra_date_formats); if (unlikely(__pyx_t_15 == ((int)-1))) __PYX_ERR(0, 1085, __pyx_L1_error)

      /* "pyreadstat/_readstat_parser.pyx":1084
 * 
 *     if extra_date_formats is not None:
 *         if file_format == FILE_FORMAT_SAS:             # <<<<<<<<<<<<<<
 *             sas_date_formats.extend(extra_date_formats)
 *         elif file_format == FILE_FORMAT_SPSS:
 */
      break;
      case __pyx_e_10pyreadstat_16_readstat_parser_FILE_FORMAT_SPSS:

      /* "pyreadstat/_readstat_parser.pyx":1087
 *             sas_date_formats.extend(extra_date_formats)
 *         elif file_format == FILE_FORMAT_SPSS:
 *             spss_date_formats.extend(extra_date_formats)             # <<<<<<<<<<<<<<
 *         elif file_format == FILE_FORMAT_STATA:
 *             stata_date_formats.extend(extra_date_formats)
 */
      if (unlikely(__pyx_v_10pyreadstat_16_readstat_parser_spss_date_formats == Py_None)) {
        PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "extend");
        __PYX_ERR(0, 1087, __pyx_L1_error)
      }
      __pyx_t_15 = __Pyx_PyList_Extend(__pyx_v_10pyreadstat_16_readstat_parser_spss_date_formats, __pyx_v_extra_date_formats); if (unlikely(__pyx_t_15 == ((int)-1))) __PYX_ERR(0, 1087, __pyx_L1_error)

      /* "pyreadstat/_readstat_parser.pyx":1086
 *         if file_format == FILE_FORMAT_SAS:
 *             sas_date_formats.extend(extra_date_formats)
 *         elif file_format == FILE_FORMAT_SPSS:             # <<<<<<<<<<<<<<
 *             spss_date_formats.extend(extra_date_formats)
 *         elif file_format == FILE_FORMAT_STATA:
 */
      break;
      case __pyx_e_10pyreadstat_16_readstat_parser_FILE_FORMAT_STATA:

      /* "pyreadstat/_readstat_parser.pyx":1089
 *             spss_date_formats.extend(extra_date_formats)
 *         elif file_format == FILE_FORMAT_STATA:
 *             stata_date_formats.extend(extra_date_formats)             # <<<<<<<<<<<<<<
 *         else:
 *             raise PyreadstatError("Unknown file format")
 */
      if (unlikely(__pyx_v_10pyreadstat_16_readstat_parser_stata_date_formats == Py_None)) {
        PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "extend");
        __PYX_ERR(0, 1089, __pyx_L1_error)
      }
      __pyx_t_15 = __Pyx_PyList_Extend(__pyx_v_10pyreadstat_16_readstat_parser_stata_date_formats, __pyx_v_extra_date_formats); if (unlikely(__pyx_t_15 == ((int)-1))) __PYX_ERR(0, 1089, __pyx_L1_error)

      /* "pyreadstat/_readstat_parser.pyx":1088
 *         elif file_format == FILE_FORMAT_SPSS:
 *             spss_date_formats.extend(extra_date_formats)
 *         elif file_format == FILE_FORMAT_STATA:             # <<<<<<<<<<<<<<
 *             stata_date_formats.extend(extra_date_formats)
 *         else:
 */
      break;
      default:

      /* "pyreadstat/_readstat_parser.pyx":1091
 *             stata_date_formats.extend(extra_date_formats)
 *         else:
 *             raise PyreadstatError("Unknown file format")             # <<<<<<<<<<<<<<
 *     if extra_datetime_formats is not None:
 *         if file_format == FILE_FORMAT_SAS:
 */
      __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_PyreadstatError); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1091, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_13 = NULL;
      __pyx_t_8 = 0;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
        __pyx_t_13 = PyMethod_GET_SELF(__pyx_t_7);
        if (likely(__pyx_t_13)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
          __Pyx_INCREF(__pyx_t_13);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_7, function);
          __pyx_t_8 = 1;
        }
      }
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_13, __pyx_kp_s_Unknown_file_format};
        __pyx_t_6 = __Pyx_PyObject_FastCall(__pyx_t_7, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
        __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
        if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1091, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      }
      __Pyx_Raise(__pyx_t_6, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __PYX_ERR(0, 1091, __pyx_L1_error)
      break;
    }

    /* "pyreadstat/_readstat_parser.pyx":1083
 *         raise PyreadstatError("output format must be one of {allowed_formats}, '{output_format}' was given".format(allowed_formats=allowed_formats, output_format=output_format))
 * 
 *     if extra_date_formats is not None:             # <<<<<<<<<<<<<<
 *         if file_format == FILE_FORMAT_SAS:
 *             sas_date_formats.extend(extra_date_formats)
 */
  }

  /* "pyreadstat/_readstat_parser.pyx":1092
 *         else:
 *             raise PyreadstatError("Unknown file format")
 *     if extra_datetime_formats is not None:             # <<<<<<<<<<<<<<
 *         if file_format == FILE_FORMAT_SAS:
 *             sas_datetime_formats.extend(extra_datetime_formats)
 */
  __pyx_t_2 = (__pyx_v_extra_datetime_formats != ((PyObject*)Py_None));
  if (__pyx_t_2) {

    /* "pyreadstat/_readstat_parser.pyx":1093
 *             raise PyreadstatError("Unknown file format")
 *     if extra_datetime_formats is not None:
 *         if file_format == FILE_FORMAT_SAS:             # <<<<<<<<<<<<<<
 *             sas_datetime_formats.extend(extra_datetime_formats)
 *         elif file_format == FILE_FORMAT_SPSS:
 */
    switch (__pyx_v_file_format) {
      case __pyx_e_10pyreadstat_16_readstat_parser_FILE_FORMAT_SAS:

      /* "pyreadstat/_readstat_parser.pyx":1094
 *     if extra_datetime_formats is not None:
 *         if file_format == FILE_FORMAT_SAS:
 *             sas_datetime_formats.extend(extra_datetime_formats)             # <<<<<<<<<<<<<<
 *         elif file_format == FILE_FORMAT_SPSS:
 *             spss_datetime_formats.extend(extra_datetime_formats)
 */
      if (unlikely(__pyx_v_10pyreadstat_16_readstat_parser_sas_datetime_formats == Py_None)) {
        PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "extend");
        __PYX_ERR(0, 1094, __pyx_L1_error)
      }
      __pyx_t_15 = __Pyx_PyList_Extend(__pyx_v_10pyreadstat_16_readstat_parser_sas_datetime_formats, __pyx_v_extra_datetime_formats); if (unlikely(__pyx_t_15 == ((int)-1))) __PYX_ERR(0, 1094, __pyx_L1_error)

      /* "pyreadstat/_readstat_parser.pyx":1093
 *             raise PyreadstatError("Unknown file format")
 *     if extra_datetime_formats is not None:
 *         if file_format == FILE_FORMAT_SAS:             # <<<<<<<<<<<<<<
 *             sas_datetime_formats.extend(extra_datetime_formats)
 *         elif file_format == FILE_FORMAT_SPSS:
 */
      break;
      case __pyx_e_10pyreadstat_16_readstat_parser_FILE_FORMAT_SPSS:

      /* "pyreadstat/_readstat_parser.pyx":1096
 *             sas_datetime_formats.extend(extra_datetime_formats)
 *         elif file_format == FILE_FORMAT_SPSS:
 *             spss_datetime_formats.extend(extra_datetime_formats)             # <<<<<<<<<<<<<<
 *         elif file_format == FILE_FORMAT_STATA:
 *             stata_datetime_formats.extend(extra_datetime_formats)
 */
      if (unlikely(__pyx_v_10pyreadstat_16_readstat_parser_spss_datetime_formats == Py_None)) {
        PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "extend");
        __PYX_ERR(0, 1096, __pyx_L1_error)
      }
      __pyx_t_15 = __Pyx_PyList_Extend(__pyx_v_10pyreadstat_16_readstat_parser_spss_datetime_formats, __pyx_v_extra_datetime_formats); if (unlikely(__pyx_t_15 == ((int)-1))) __PYX_ERR(0, 1096, __pyx_L1_error)

      /* "pyreadstat/_readstat_parser.pyx":1095
 *         if file_format == FILE_FORMAT_SAS:
 *             sas_datetime_formats.extend(extra_datetime_formats)
 *         elif file_format == FILE_FORMAT_SPSS:             # <<<<<<<<<<<<<<
 *             spss_datetime_formats.extend(extra_datetime_formats)
 *         elif file_format == FILE_FORMAT_STATA:
 */
      break;
      case __pyx_e_10pyreadstat_16_readstat_parser_FILE_FORMAT_STATA:

      /* "pyreadstat/_readstat_parser.pyx":1098
 *             spss_datetime_formats.extend(extra_datetime_formats)
 *         elif file_format == FILE_FORMAT_STATA:
 *             stata_datetime_formats.extend(extra_datetime_formats)             # <<<<<<<<<<<<<<
 *         else:
 *             raise PyreadstatError("Unknown file format")
 */
      if (unlikely(__pyx_v_10pyreadstat_16_readstat_parser_stata_datetime_formats == Py_None)) {
        PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "extend");
        __PYX_ERR(0, 1098, __pyx_L1_error)
      }
      __pyx_t_15 = __Pyx_PyList_Extend(__pyx_v_10pyreadstat_16_readstat_parser_stata_datetime_formats, __pyx_v_extra_datetime_formats); if (unlikely(__pyx_t_15 == ((int)-1))) __PYX_ERR(0, 1098, __pyx_L1_error)

      /* "pyreadstat/_readstat_parser.pyx":1097
 *         elif file_format == FILE_FORMAT_SPSS:
 *             spss_datetime_formats.extend(extra_datetime_formats)
 *         elif file_format == FILE_FORMAT_STATA:             # <<<<<<<<<<<<<<
 *             stata_datetime_formats.extend(extra_datetime_formats)
 *         else:
 */
      break;
      default:

      /* "pyreadstat/_readstat_parser.pyx":1100
 *             stata_datetime_formats.extend(extra_datetime_formats)
 *         else:
 *             raise PyreadstatError("Unknown file format")             # <<<<<<<<<<<<<<
 *     global sas_all_formats, spss_all_formats, stata_all_formats
 *     sas_all_formats = sas_date_formats + sas_datetime_formats + sas_time_formats
 */
      __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_PyreadstatError); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1100, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_13 = NULL;
      __pyx_t_8 = 0;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
        __pyx_t_13 = PyMethod_GET_SELF(__pyx_t_7);
        if (likely(__pyx_t_13)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
          __Pyx_INCREF(__pyx_t_13);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_7, function);
          __pyx_t_8 = 1;
        }
      }
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_13, __pyx_kp_s_Unknown_file_format};
        __pyx_t_6 = __Pyx_PyObject_FastCall(__pyx_t_7, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
        __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
        if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1100, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      }
      __Pyx_Raise(__pyx_t_6, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __PYX_ERR(0, 1100, __pyx_L1_error)
      break;
    }

    /* "pyreadstat/_readstat_parser.pyx":1092
 *         else:
 *             raise PyreadstatError("Unknown file format")
 *     if extra_datetime_formats is not None:             # <<<<<<<<<<<<<<
 *         if file_format == FILE_FORMAT_SAS:
 *             sas_datetime_formats.extend(extra_datetime_formats)
 */
  }

  /* "pyreadstat/_readstat_parser.pyx":1102
 *             raise PyreadstatError("Unknown file format")
 *     global sas_all_formats, spss_all_formats, stata_all_formats
 *     sas_all_formats = sas_date_formats + sas_datetime_formats + sas_time_formats             # <<<<<<<<<<<<<<
 *     stata_all_formats = stata_date_formats + stata_datetime_formats + stata_time_formats
 *     spss_all_formats = spss_date_formats + spss_datetime_formats + spss_time_formats
 */
  __pyx_t_6 = PyNumber_Add(__pyx_v_10pyreadstat_16_readstat_parser_sas_date_formats, __pyx_v_10pyreadstat_16_readstat_parser_sas_datetime_formats); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1102, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = PyNumber_Add(__pyx_t_6, __pyx_v_10pyreadstat_16_readstat_parser_sas_time_formats); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1102, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_XGOTREF(__pyx_v_10pyreadstat_16_readstat_parser_sas_all_formats);
  __Pyx_DECREF_SET(__pyx_v_10pyreadstat_16_readstat_parser_sas_all_formats, ((PyObject*)__pyx_t_7));
  __Pyx_GIVEREF(__pyx_t_7);
  __pyx_t_7 = 0;

  /* "pyreadstat/_readstat_parser.pyx":1103
 *     global sas_all_formats, spss_all_formats, stata_all_formats
 *     sas_all_formats = sas_date_formats + sas_datetime_formats + sas_time_formats
 *     stata_all_formats = stata_date_formats + stata_datetime_formats + stata_time_formats             # <<<<<<<<<<<<<<
 *     spss_all_formats = spss_date_formats + spss_datetime_formats + spss_time_formats
 * 
 */
  __pyx_t_7 = PyNumber_Add(__pyx_v_10pyreadstat_16_readstat_parser_stata_date_formats, __pyx_v_10pyreadstat_16_readstat_parser_stata_datetime_formats); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1103, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_6 = PyNumber_Add(__pyx_t_7, __pyx_v_10pyreadstat_16_readstat_parser_stata_time_formats); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1103, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_XGOTREF(__pyx_v_10pyreadstat_16_readstat_parser_stata_all_formats);
  __Pyx_DECREF_SET(__pyx_v_10pyreadstat_16_readstat_parser_stata_all_formats, ((PyObject*)__pyx_t_6));
  __Pyx_GIVEREF(__pyx_t_6);
  __pyx_t_6 = 0;

  /* "pyreadstat/_readstat_parser.pyx":1104
 *     sas_all_formats = sas_date_formats + sas_datetime_formats + sas_time_formats
 *     stata_all_formats = stata_date_formats + stata_datetime_formats + stata_time_formats
 *     spss_all_formats = spss_date_formats + spss_datetime_formats + spss_time_formats             # <<<<<<<<<<<<<<
 * 
 *     filename = <char *> filename_bytes
 */
  __pyx_t_6 = PyNumber_Add(__pyx_v_10pyreadstat_16_readstat_parser_spss_date_formats, __pyx_v_10pyreadstat_16_readstat_parser_spss_datetime_formats); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1104, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = PyNumber_Add(__pyx_t_6, __pyx_v_10pyreadstat_16_readstat_parser_spss_time_formats); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1104, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_XGOTREF(__pyx_v_10pyreadstat_16_readstat_parser_spss_all_formats);
  __Pyx_DECREF_SET(__pyx_v_10pyreadstat_16_readstat_parser_spss_all_formats, ((PyObject*)__pyx_t_7));
  __Pyx_GIVEREF(__pyx_t_7);
  __pyx_t_7 = 0;

  /* "pyreadstat/_readstat_parser.pyx":1106
 *     spss_all_formats = spss_date_formats + spss_datetime_formats + spss_time_formats
 * 
 *     filename = <char *> filename_bytes             # <<<<<<<<<<<<<<
 * 
 *     data = data_container()
 */
  if (unlikely(__pyx_v_filename_bytes == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
    __PYX_ERR(0, 1106, __pyx_L1_error)
  }
  __pyx_t_16 = __Pyx_PyBytes_AsWritableString(__pyx_v_filename_bytes); if (unlikely((!__pyx_t_16) && PyErr_Occurred())) __PYX_ERR(0, 1106, __pyx_L1_error)
  __pyx_v_filename = ((char *)__pyx_t_16);

  /* "pyreadstat/_readstat_parser.pyx":1108
 *     filename = <char *> filename_bytes
 * 
 *     data = data_container()             # <<<<<<<<<<<<<<
 *     ctx = <void *>data
 * 
 */
  __pyx_t_7 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_10pyreadstat_16_readstat_parser_data_container)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1108, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_v_data = ((struct __pyx_obj_10pyreadstat_16_readstat_parser_data_container *)__pyx_t_7);
  __pyx_t_7 = 0;

  /* "pyreadstat/_readstat_parser.pyx":1109
 * 
 *     data = data_container()
 *     ctx = <void *>data             # <<<<<<<<<<<<<<
 * 
 *     data.file_format = file_format
 */
  __pyx_v_ctx = ((void *)__pyx_v_data);

  /* "pyreadstat/_readstat_parser.pyx":1111
 *     ctx = <void *>data
 * 
 *     data.file_format = file_format             # <<<<<<<<<<<<<<
 *     data.metaonly = metaonly
 *     data.dates_as_pandas = dates_as_pandas
 */
  __pyx_v_data->file_format = __pyx_v_file_format;

  /* "pyreadstat/_readstat_parser.pyx":1112
 * 
 *     data.file_format = file_format
 *     data.metaonly = metaonly             # <<<<<<<<<<<<<<
 *     data.dates_as_pandas = dates_as_pandas
 * 
 */
  __pyx_v_data->metaonly = __pyx_v_metaonly;

  /* "pyreadstat/_readstat_parser.pyx":1113
 *     data.file_format = file_format
 *     data.metaonly = metaonly
 *     data.dates_as_pandas = dates_as_pandas             # <<<<<<<<<<<<<<
 * 
 *     if encoding:
 */
  __pyx_v_data->dates_as_pandas = __pyx_v_dates_as_pandas;

  /* "pyreadstat/_readstat_parser.pyx":1115
 *     data.dates_as_pandas = dates_as_pandas
 * 
 *     if encoding:             # <<<<<<<<<<<<<<
 *         data.user_encoding = encoding
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_encoding); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 1115, __pyx_L1_error)
  if (__pyx_t_2) {

    /* "pyreadstat/_readstat_parser.pyx":1116
 * 
 *     if encoding:
 *         data.user_encoding = encoding             # <<<<<<<<<<<<<<
 * 
 *     if file_format == FILE_FORMAT_SAS:
 */
    __Pyx_INCREF(__pyx_v_encoding);
    __Pyx_GIVEREF(__pyx_v_encoding);
    __Pyx_GOTREF(__pyx_v_data->user_encoding);
    __Pyx_DECREF(__pyx_v_data->user_encoding);
    __pyx_v_data->user_encoding = __pyx_v_encoding;

    /* "pyreadstat/_readstat_parser.pyx":1115
 *     data.dates_as_pandas = dates_as_pandas
 * 
 *     if encoding:             # <<<<<<<<<<<<<<
 *         data.user_encoding = encoding
 * 
 */
  }

  /* "pyreadstat/_readstat_parser.pyx":1118
 *         data.user_encoding = encoding
 * 
 *     if file_format == FILE_FORMAT_SAS:             # <<<<<<<<<<<<<<
 *         origin = sas_origin
 *     elif file_format == FILE_FORMAT_SPSS:
 */
  switch (__pyx_v_file_format) {
    case __pyx_e_10pyreadstat_16_readstat_parser_FILE_FORMAT_SAS:

    /* "pyreadstat/_readstat_parser.pyx":1119
 * 
 *     if file_format == FILE_FORMAT_SAS:
 *         origin = sas_origin             # <<<<<<<<<<<<<<
 *     elif file_format == FILE_FORMAT_SPSS:
 *         origin = spss_origin
 */
    __Pyx_INCREF(__pyx_v_10pyreadstat_16_readstat_parser_sas_origin);
    __pyx_v_origin = __pyx_v_10pyreadstat_16_readstat_parser_sas_origin;

    /* "pyreadstat/_readstat_parser.pyx":1118
 *         data.user_encoding = encoding
 * 
 *     if file_format == FILE_FORMAT_SAS:             # <<<<<<<<<<<<<<
 *         origin = sas_origin
 *     elif file_format == FILE_FORMAT_SPSS:
 */
    break;
    case __pyx_e_10pyreadstat_16_readstat_parser_FILE_FORMAT_SPSS:

    /* "pyreadstat/_readstat_parser.pyx":1121
 *         origin = sas_origin
 *     elif file_format == FILE_FORMAT_SPSS:
 *         origin = spss_origin             # <<<<<<<<<<<<<<
 *     elif file_format == FILE_FORMAT_STATA:
 *         origin = stata_origin
 */
    __Pyx_INCREF(__pyx_v_10pyreadstat_16_readstat_parser_spss_origin);
    __pyx_v_origin = __pyx_v_10pyreadstat_16_readstat_parser_spss_origin;

    /* "pyreadstat/_readstat_parser.pyx":1120
 *     if file_format == FILE_FORMAT_SAS:
 *         origin = sas_origin
 *     elif file_format == FILE_FORMAT_SPSS:             # <<<<<<<<<<<<<<
 *         origin = spss_origin
 *     elif file_format == FILE_FORMAT_STATA:
 */
    break;
    case __pyx_e_10pyreadstat_16_readstat_parser_FILE_FORMAT_STATA:

    /* "pyreadstat/_readstat_parser.pyx":1123
 *         origin = spss_origin
 *     elif file_format == FILE_FORMAT_STATA:
 *         origin = stata_origin             # <<<<<<<<<<<<<<
 *     else:
 *         raise PyreadstatError("Unknown file format")
 */
    __Pyx_INCREF(__pyx_v_10pyreadstat_16_readstat_parser_stata_origin);
    __pyx_v_origin = __pyx_v_10pyreadstat_16_readstat_parser_stata_origin;

    /* "pyreadstat/_readstat_parser.pyx":1122
 *     elif file_format == FILE_FORMAT_SPSS:
 *         origin = spss_origin
 *     elif file_format == FILE_FORMAT_STATA:             # <<<<<<<<<<<<<<
 *         origin = stata_origin
 *     else:
 */
    break;
    default:

    /* "pyreadstat/_readstat_parser.pyx":1125
 *         origin = stata_origin
 *     else:
 *         raise PyreadstatError("Unknown file format")             # <<<<<<<<<<<<<<
 * 
 *     data.origin = origin
 */
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_PyreadstatError); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1125, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_13 = NULL;
    __pyx_t_8 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_13 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_13)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_13);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_13, __pyx_kp_s_Unknown_file_format};
      __pyx_t_7 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
      if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1125, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_Raise(__pyx_t_7, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __PYX_ERR(0, 1125, __pyx_L1_error)
    break;
  }

  /* "pyreadstat/_readstat_parser.pyx":1127
 *         raise PyreadstatError("Unknown file format")
 * 
 *     data.origin = origin             # <<<<<<<<<<<<<<
 * 
 *     if usecols is not None:
 */
  __Pyx_INCREF(__pyx_v_origin);
  __Pyx_GIVEREF(__pyx_v_origin);
  __Pyx_GOTREF(__pyx_v_data->origin);
  __Pyx_DECREF(__pyx_v_data->origin);
  __pyx_v_data->origin = __pyx_v_origin;

  /* "pyreadstat/_readstat_parser.pyx":1129
 *     data.origin = origin
 * 
 *     if usecols is not None:             # <<<<<<<<<<<<<<
 *         data.filter_cols = 1
 *         data.use_cols = usecols
 */
  __pyx_t_2 = (__pyx_v_usecols != ((PyObject*)Py_None));
  if (__pyx_t_2) {

    /* "pyreadstat/_readstat_parser.pyx":1130
 * 
 *     if usecols is not None:
 *         data.filter_cols = 1             # <<<<<<<<<<<<<<
 *         data.use_cols = usecols
 * 
 */
    __pyx_v_data->filter_cols = 1;

    /* "pyreadstat/_readstat_parser.pyx":1131
 *     if usecols is not None:
 *         data.filter_cols = 1
 *         data.use_cols = usecols             # <<<<<<<<<<<<<<
 * 
 *     data.usernan = usernan
 */
    __Pyx_INCREF(__pyx_v_usecols);
    __Pyx_GIVEREF(__pyx_v_usecols);
    __Pyx_GOTREF(__pyx_v_data->use_cols);
    __Pyx_DECREF(__pyx_v_data->use_cols);
    __pyx_v_data->use_cols = __pyx_v_usecols;

    /* "pyreadstat/_readstat_parser.pyx":1129
 *     data.origin = origin
 * 
 *     if usecols is not None:             # <<<<<<<<<<<<<<
 *         data.filter_cols = 1
 *         data.use_cols = usecols
 */
  }

  /* "pyreadstat/_readstat_parser.pyx":1133
 *         data.use_cols = usecols
 * 
 *     data.usernan = usernan             # <<<<<<<<<<<<<<
 *     data.no_datetime_conversion = no_datetime_conversion
 * 
 */
  __pyx_v_data->usernan = __pyx_v_usernan;

  /* "pyreadstat/_readstat_parser.pyx":1134
 * 
 *     data.usernan = usernan
 *     data.no_datetime_conversion = no_datetime_conversion             # <<<<<<<<<<<<<<
 * 
 *     # go!
 */
  __pyx_v_data->no_datetime_conversion = __pyx_v_no_datetime_conversion;

  /* "pyreadstat/_readstat_parser.pyx":1137
 * 
 *     # go!
 *     run_readstat_parser(filename, data, file_extension, row_limit, row_offset)             # <<<<<<<<<<<<<<
 *     data_dict = data_container_to_dict(data)
 *     if output_format == 'dict':
 */
  __pyx_f_10pyreadstat_16_readstat_parser_run_readstat_parser(__pyx_v_filename, __pyx_v_data, __pyx_v_file_extension, __pyx_v_row_limit, __pyx_v_row_offset); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1137, __pyx_L1_error)

  /* "pyreadstat/_readstat_parser.pyx":1138
 *     # go!
 *     run_readstat_parser(filename, data, file_extension, row_limit, row_offset)
 *     data_dict = data_container_to_dict(data)             # <<<<<<<<<<<<<<
 *     if output_format == 'dict':
 *         data_frame = data_dict
 */
  __pyx_t_7 = __pyx_f_10pyreadstat_16_readstat_parser_data_container_to_dict(__pyx_v_data); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1138, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_v_data_dict = __pyx_t_7;
  __pyx_t_7 = 0;

  /* "pyreadstat/_readstat_parser.pyx":1139
 *     run_readstat_parser(filename, data, file_extension, row_limit, row_offset)
 *     data_dict = data_container_to_dict(data)
 *     if output_format == 'dict':             # <<<<<<<<<<<<<<
 *         data_frame = data_dict
 *     elif output_format == 'pandas':
 */
  __pyx_t_2 = (__Pyx_PyString_Equals(__pyx_v_output_format, __pyx_n_s_dict, Py_EQ)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 1139, __pyx_L1_error)
  if (__pyx_t_2) {

    /* "pyreadstat/_readstat_parser.pyx":1140
 *     data_dict = data_container_to_dict(data)
 *     if output_format == 'dict':
 *         data_frame = data_dict             # <<<<<<<<<<<<<<
 *     elif output_format == 'pandas':
 *         data_frame = dict_to_pandas_dataframe(data_dict, data)
 */
    __Pyx_INCREF(__pyx_v_data_dict);
    __pyx_v_data_frame = __pyx_v_data_dict;

    /* "pyreadstat/_readstat_parser.pyx":1139
 *     run_readstat_parser(filename, data, file_extension, row_limit, row_offset)
 *     data_dict = data_container_to_dict(data)
 *     if output_format == 'dict':             # <<<<<<<<<<<<<<
 *         data_frame = data_dict
 *     elif output_format == 'pandas':
 */
    goto __pyx_L24;
  }

  /* "pyreadstat/_readstat_parser.pyx":1141
 *     if output_format == 'dict':
 *         data_frame = data_dict
 *     elif output_format == 'pandas':             # <<<<<<<<<<<<<<
 *         data_frame = dict_to_pandas_dataframe(data_dict, data)
 *     metadata = data_container_extract_metadata(data)
 */
  __pyx_t_2 = (__Pyx_PyString_Equals(__pyx_v_output_format, __pyx_n_s_pandas, Py_EQ)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 1141, __pyx_L1_error)
  if (__pyx_t_2) {

    /* "pyreadstat/_readstat_parser.pyx":1142
 *         data_frame = data_dict
 *     elif output_format == 'pandas':
 *         data_frame = dict_to_pandas_dataframe(data_dict, data)             # <<<<<<<<<<<<<<
 *     metadata = data_container_extract_metadata(data)
 * 
 */
    __pyx_t_7 = __pyx_f_10pyreadstat_16_readstat_parser_dict_to_pandas_dataframe(__pyx_v_data_dict, __pyx_v_data); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1142, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_v_data_frame = __pyx_t_7;
    __pyx_t_7 = 0;

    /* "pyreadstat/_readstat_parser.pyx":1141
 *     if output_format == 'dict':
 *         data_frame = data_dict
 *     elif output_format == 'pandas':             # <<<<<<<<<<<<<<
 *         data_frame = dict_to_pandas_dataframe(data_dict, data)
 *     metadata = data_container_extract_metadata(data)
 */
  }
  __pyx_L24:;

  /* "pyreadstat/_readstat_parser.pyx":1143
 *     elif output_format == 'pandas':
 *         data_frame = dict_to_pandas_dataframe(data_dict, data)
 *     metadata = data_container_extract_metadata(data)             # <<<<<<<<<<<<<<
 * 
 *     return data_frame, metadata
 */
  __pyx_t_7 = __pyx_f_10pyreadstat_16_readstat_parser_data_container_extract_metadata(__pyx_v_data); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1143, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_v_metadata = __pyx_t_7;
  __pyx_t_7 = 0;

  /* "pyreadstat/_readstat_parser.pyx":1145
 *     metadata = data_container_extract_metadata(data)
 * 
 *     return data_frame, metadata             # <<<<<<<<<<<<<<
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  if (unlikely(!__pyx_v_data_frame)) { __Pyx_RaiseUnboundLocalError("data_frame"); __PYX_ERR(0, 1145, __pyx_L1_error) }
  __pyx_t_7 = PyTuple_New(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1145, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_INCREF(__pyx_v_data_frame);
  __Pyx_GIVEREF(__pyx_v_data_frame);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_v_data_frame);
  __Pyx_INCREF(__pyx_v_metadata);
  __Pyx_GIVEREF(__pyx_v_metadata);
  PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_v_metadata);
  __pyx_r = __pyx_t_7;
  __pyx_t_7 = 0;
  goto __pyx_L0;

  /* "pyreadstat/_readstat_parser.pyx":1038
 * 
 * 
 * cdef object run_conversion(object filename_path, py_file_format file_format, py_file_extension file_extension,             # <<<<<<<<<<<<<<
 *                            str encoding, bint metaonly, bint dates_as_pandas, list usecols, bint usernan,
 *                            bint no_datetime_conversion, long row_limit, long row_offset, str output_format,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_AddTraceback("pyreadstat._readstat_parser.run_conversion", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_filename_bytes);
  __Pyx_XDECREF((PyObject *)__pyx_v_data);
  __Pyx_XDECREF(__pyx_v_origin);
  __Pyx_XDECREF(__pyx_v_allowed_formats);
  __Pyx_XDECREF(__pyx_v_data_dict);
  __Pyx_XDECREF(__pyx_v_data_frame);
  __Pyx_XDECREF(__pyx_v_metadata);
  __Pyx_XDECREF(__pyx_v_output_format);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_tp_new_10pyreadstat_16_readstat_parser_data_container(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_10pyreadstat_16_readstat_parser_data_container *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_10pyreadstat_16_readstat_parser_data_container *)o);
  p->col_data = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->col_data_len = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->col_names = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->col_labels = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->col_dtypes = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->col_numpy_dtypes = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->col_dtypes_isobject = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->col_dytpes_isfloat = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->col_formats = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->col_formats_original = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->origin = Py_None; Py_INCREF(Py_None);
  p->file_label = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->file_encoding = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->label_to_var_name = Py_None; Py_INCREF(Py_None);
  p->labels_raw = Py_None; Py_INCREF(Py_None);
  p->notes = Py_None; Py_INCREF(Py_None);
  p->user_encoding = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->table_name = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->use_cols = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->missing_ranges = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->missing_user_values = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->variable_storage_width = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->variable_display_width = Py_None; Py_INCREF(Py_None);
  p->variable_alignment = Py_None; Py_INCREF(Py_None);
  p->variable_measure = Py_None; Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_10pyreadstat_16_readstat_parser_14data_container_1__cinit__(o, __pyx_empty_tuple, NULL) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_10pyreadstat_16_readstat_parser_data_container(PyObject *o) {
  struct __pyx_obj_10pyreadstat_16_readstat_parser_data_container *p = (struct __pyx_obj_10pyreadstat_16_readstat_parser_data_container *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_10pyreadstat_16_readstat_parser_data_container) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->col_data);
  Py_CLEAR(p->col_data_len);
  Py_CLEAR(p->col_names);
  Py_CLEAR(p->col_labels);
  Py_CLEAR(p->col_dtypes);
  Py_CLEAR(p->col_numpy_dtypes);
  Py_CLEAR(p->col_dtypes_isobject);
  Py_CLEAR(p->col_dytpes_isfloat);
  Py_CLEAR(p->col_formats);
  Py_CLEAR(p->col_formats_original);
  Py_CLEAR(p->origin);
  Py_CLEAR(p->file_label);
  Py_CLEAR(p->file_encoding);
  Py_CLEAR(p->label_to_var_name);
  Py_CLEAR(p->labels_raw);
  Py_CLEAR(p->notes);
  Py_CLEAR(p->user_encoding);
  Py_CLEAR(p->table_name);
  Py_CLEAR(p->use_cols);
  Py_CLEAR(p->missing_ranges);
  Py_CLEAR(p->missing_user_values);
  Py_CLEAR(p->variable_storage_width);
  Py_CLEAR(p->variable_display_width);
  Py_CLEAR(p->variable_alignment);
  Py_CLEAR(p->variable_measure);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_10pyreadstat_16_readstat_parser_data_container(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_10pyreadstat_16_readstat_parser_data_container *p = (struct __pyx_obj_10pyreadstat_16_readstat_parser_data_container *)o;
  if (p->col_data) {
    e = (*v)(p->col_data, a); if (e) return e;
  }
  if (p->col_data_len) {
    e = (*v)(p->col_data_len, a); if (e) return e;
  }
  if (p->col_names) {
    e = (*v)(p->col_names, a); if (e) return e;
  }
  if (p->col_labels) {
    e = (*v)(p->col_labels, a); if (e) return e;
  }
  if (p->col_dtypes) {
    e = (*v)(p->col_dtypes, a); if (e) return e;
  }
  if (p->col_numpy_dtypes) {
    e = (*v)(p->col_numpy_dtypes, a); if (e) return e;
  }
  if (p->col_dtypes_isobject) {
    e = (*v)(p->col_dtypes_isobject, a); if (e) return e;
  }
  if (p->col_dytpes_isfloat) {
    e = (*v)(p->col_dytpes_isfloat, a); if (e) return e;
  }
  if (p->col_formats) {
    e = (*v)(p->col_formats, a); if (e) return e;
  }
  if (p->col_formats_original) {
    e = (*v)(p->col_formats_original, a); if (e) return e;
  }
  if (p->origin) {
    e = (*v)(p->origin, a); if (e) return e;
  }
  if (p->label_to_var_name) {
    e = (*v)(p->label_to_var_name, a); if (e) return e;
  }
  if (p->labels_raw) {
    e = (*v)(p->labels_raw, a); if (e) return e;
  }
  if (p->notes) {
    e = (*v)(p->notes, a); if (e) return e;
  }
  if (p->use_cols) {
    e = (*v)(p->use_cols, a); if (e) return e;
  }
  if (p->missing_ranges) {
    e = (*v)(p->missing_ranges, a); if (e) return e;
  }
  if (p->missing_user_values) {
    e = (*v)(p->missing_user_values, a); if (e) return e;
  }
  if (p->variable_storage_width) {
    e = (*v)(p->variable_storage_width, a); if (e) return e;
  }
  if (p->variable_display_width) {
    e = (*v)(p->variable_display_width, a); if (e) return e;
  }
  if (p->variable_alignment) {
    e = (*v)(p->variable_alignment, a); if (e) return e;
  }
  if (p->variable_measure) {
    e = (*v)(p->variable_measure, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_10pyreadstat_16_readstat_parser_data_container(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_10pyreadstat_16_readstat_parser_data_container *p = (struct __pyx_obj_10pyreadstat_16_readstat_parser_data_container *)o;
  tmp = ((PyObject*)p->col_data);
  p->col_data = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->col_data_len);
  p->col_data_len = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->col_names);
  p->col_names = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->col_labels);
  p->col_labels = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->col_dtypes);
  p->col_dtypes = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->col_numpy_dtypes);
  p->col_numpy_dtypes = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->col_dtypes_isobject);
  p->col_dtypes_isobject = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->col_dytpes_isfloat);
  p->col_dytpes_isfloat = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->col_formats);
  p->col_formats = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->col_formats_original);
  p->col_formats_original = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->origin);
  p->origin = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->label_to_var_name);
  p->label_to_var_name = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->labels_raw);
  p->labels_raw = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->notes);
  p->notes = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->use_cols);
  p->use_cols = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->missing_ranges);
  p->missing_ranges = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->missing_user_values);
  p->missing_user_values = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->variable_storage_width);
  p->variable_storage_width = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->variable_display_width);
  p->variable_display_width = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->variable_alignment);
  p->variable_alignment = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->variable_measure);
  p->variable_measure = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyMethodDef __pyx_methods_10pyreadstat_16_readstat_parser_data_container[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10pyreadstat_16_readstat_parser_14data_container_3__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10pyreadstat_16_readstat_parser_14data_container_5__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_10pyreadstat_16_readstat_parser_data_container_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_10pyreadstat_16_readstat_parser_data_container},
  {Py_tp_doc, (void *)PyDoc_STR("\n    This extension type holds all the data we need to get from our file from \n    the readstat parser and will be used later to compose our pandas data frame\n    ")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_10pyreadstat_16_readstat_parser_data_container},
  {Py_tp_clear, (void *)__pyx_tp_clear_10pyreadstat_16_readstat_parser_data_container},
  {Py_tp_methods, (void *)__pyx_methods_10pyreadstat_16_readstat_parser_data_container},
  {Py_tp_new, (void *)__pyx_tp_new_10pyreadstat_16_readstat_parser_data_container},
  {0, 0},
};
static PyType_Spec __pyx_type_10pyreadstat_16_readstat_parser_data_container_spec = {
  "pyreadstat._readstat_parser.data_container",
  sizeof(struct __pyx_obj_10pyreadstat_16_readstat_parser_data_container),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_10pyreadstat_16_readstat_parser_data_container_slots,
};
#else

static PyTypeObject __pyx_type_10pyreadstat_16_readstat_parser_data_container = {
  PyVarObject_HEAD_INIT(0, 0)
  "pyreadstat._readstat_parser.""data_container", /*tp_name*/
  sizeof(struct __pyx_obj_10pyreadstat_16_readstat_parser_data_container), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_10pyreadstat_16_readstat_parser_data_container, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("\n    This extension type holds all the data we need to get from our file from \n    the readstat parser and will be used later to compose our pandas data frame\n    "), /*tp_doc*/
  __pyx_tp_traverse_10pyreadstat_16_readstat_parser_data_container, /*tp_traverse*/
  __pyx_tp_clear_10pyreadstat_16_readstat_parser_data_container, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_10pyreadstat_16_readstat_parser_data_container, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_10pyreadstat_16_readstat_parser_data_container, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static PyMethodDef __pyx_methods[] = {
  {0, 0, 0, 0}
};
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 3))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif
/* #### Code section: pystring_table ### */

static int __Pyx_CreateStringTabAndInitStrings(void) {
  __Pyx_StringTabEntry __pyx_string_tab[] = {
    {&__pyx_n_s_ADATE, __pyx_k_ADATE, sizeof(__pyx_k_ADATE), 0, 0, 1, 1},
    {&__pyx_n_s_ADATE10, __pyx_k_ADATE10, sizeof(__pyx_k_ADATE10), 0, 0, 1, 1},
    {&__pyx_n_s_ADATE8, __pyx_k_ADATE8, sizeof(__pyx_k_ADATE8), 0, 0, 1, 1},
    {&__pyx_n_s_B8601DA, __pyx_k_B8601DA, sizeof(__pyx_k_B8601DA), 0, 0, 1, 1},
    {&__pyx_n_s_B8601DN, __pyx_k_B8601DN, sizeof(__pyx_k_B8601DN), 0, 0, 1, 1},
    {&__pyx_n_s_B8601DT, __pyx_k_B8601DT, sizeof(__pyx_k_B8601DT), 0, 0, 1, 1},
    {&__pyx_n_s_B8601TM, __pyx_k_B8601TM, sizeof(__pyx_k_B8601TM), 0, 0, 1, 1},
    {&__pyx_n_s_DATE, __pyx_k_DATE, sizeof(__pyx_k_DATE), 0, 0, 1, 1},
    {&__pyx_n_s_DATE11, __pyx_k_DATE11, sizeof(__pyx_k_DATE11), 0, 0, 1, 1},
    {&__pyx_n_s_DATE12, __pyx_k_DATE12, sizeof(__pyx_k_DATE12), 0, 0, 1, 1},
    {&__pyx_n_s_DATE8, __pyx_k_DATE8, sizeof(__pyx_k_DATE8), 0, 0, 1, 1},
    {&__pyx_n_s_DATE9, __pyx_k_DATE9, sizeof(__pyx_k_DATE9), 0, 0, 1, 1},
    {&__pyx_n_s_DATEAMPM, __pyx_k_DATEAMPM, sizeof(__pyx_k_DATEAMPM), 0, 0, 1, 1},
    {&__pyx_n_s_DATETIME, __pyx_k_DATETIME, sizeof(__pyx_k_DATETIME), 0, 0, 1, 1},
    {&__pyx_n_s_DATETIME17, __pyx_k_DATETIME17, sizeof(__pyx_k_DATETIME17), 0, 0, 1, 1},
    {&__pyx_n_s_DATETIME18, __pyx_k_DATETIME18, sizeof(__pyx_k_DATETIME18), 0, 0, 1, 1},
    {&__pyx_n_s_DATETIME19, __pyx_k_DATETIME19, sizeof(__pyx_k_DATETIME19), 0, 0, 1, 1},
    {&__pyx_n_s_DATETIME20, __pyx_k_DATETIME20, sizeof(__pyx_k_DATETIME20), 0, 0, 1, 1},
    {&__pyx_n_s_DATETIME21, __pyx_k_DATETIME21, sizeof(__pyx_k_DATETIME21), 0, 0, 1, 1},
    {&__pyx_n_s_DATETIME22, __pyx_k_DATETIME22, sizeof(__pyx_k_DATETIME22), 0, 0, 1, 1},
    {&__pyx_kp_s_DATETIME23_2, __pyx_k_DATETIME23_2, sizeof(__pyx_k_DATETIME23_2), 0, 0, 1, 0},
    {&__pyx_n_s_DATETIME8, __pyx_k_DATETIME8, sizeof(__pyx_k_DATETIME8), 0, 0, 1, 1},
    {&__pyx_n_s_DDMMYY, __pyx_k_DDMMYY, sizeof(__pyx_k_DDMMYY), 0, 0, 1, 1},
    {&__pyx_n_s_DDMMYYB, __pyx_k_DDMMYYB, sizeof(__pyx_k_DDMMYYB), 0, 0, 1, 1},
    {&__pyx_n_s_DDMMYYB10, __pyx_k_DDMMYYB10, sizeof(__pyx_k_DDMMYYB10), 0, 0, 1, 1},
    {&__pyx_n_s_DDMMYYC, __pyx_k_DDMMYYC, sizeof(__pyx_k_DDMMYYC), 0, 0, 1, 1},
    {&__pyx_n_s_DDMMYYC10, __pyx_k_DDMMYYC10, sizeof(__pyx_k_DDMMYYC10), 0, 0, 1, 1},
    {&__pyx_n_s_DDMMYYD, __pyx_k_DDMMYYD, sizeof(__pyx_k_DDMMYYD), 0, 0, 1, 1},
    {&__pyx_n_s_DDMMYYD10, __pyx_k_DDMMYYD10, sizeof(__pyx_k_DDMMYYD10), 0, 0, 1, 1},
    {&__pyx_n_s_DDMMYYN6, __pyx_k_DDMMYYN6, sizeof(__pyx_k_DDMMYYN6), 0, 0, 1, 1},
    {&__pyx_n_s_DDMMYYN8, __pyx_k_DDMMYYN8, sizeof(__pyx_k_DDMMYYN8), 0, 0, 1, 1},
    {&__pyx_n_s_DDMMYYP, __pyx_k_DDMMYYP, sizeof(__pyx_k_DDMMYYP), 0, 0, 1, 1},
    {&__pyx_n_s_DDMMYYP10, __pyx_k_DDMMYYP10, sizeof(__pyx_k_DDMMYYP10), 0, 0, 1, 1},
    {&__pyx_n_s_DDMMYYS, __pyx_k_DDMMYYS, sizeof(__pyx_k_DDMMYYS), 0, 0, 1, 1},
    {&__pyx_n_s_DDMMYYS10, __pyx_k_DDMMYYS10, sizeof(__pyx_k_DDMMYYS10), 0, 0, 1, 1},
    {&__pyx_n_s_DTDATE, __pyx_k_DTDATE, sizeof(__pyx_k_DTDATE), 0, 0, 1, 1},
    {&__pyx_n_s_DTIME, __pyx_k_DTIME, sizeof(__pyx_k_DTIME), 0, 0, 1, 1},
    {&__pyx_n_s_DataFrame, __pyx_k_DataFrame, sizeof(__pyx_k_DataFrame), 0, 0, 1, 1},
    {&__pyx_n_s_E8601DA, __pyx_k_E8601DA, sizeof(__pyx_k_E8601DA), 0, 0, 1, 1},
    {&__pyx_n_s_E8601DT, __pyx_k_E8601DT, sizeof(__pyx_k_E8601DT), 0, 0, 1, 1},
    {&__pyx_n_s_E8601TM, __pyx_k_E8601TM, sizeof(__pyx_k_E8601TM), 0, 0, 1, 1},
    {&__pyx_n_s_EDATE, __pyx_k_EDATE, sizeof(__pyx_k_EDATE), 0, 0, 1, 1},
    {&__pyx_n_s_EDATE10, __pyx_k_EDATE10, sizeof(__pyx_k_EDATE10), 0, 0, 1, 1},
    {&__pyx_n_s_EDATE8, __pyx_k_EDATE8, sizeof(__pyx_k_EDATE8), 0, 0, 1, 1},
    {&__pyx_kp_s_Failed_convert_C_to_python_value, __pyx_k_Failed_convert_C_to_python_value, sizeof(__pyx_k_Failed_convert_C_to_python_value), 0, 0, 1, 0},
    {&__pyx_kp_s_Failed_to_read_number_of_variabl, __pyx_k_Failed_to_read_number_of_variabl, sizeof(__pyx_k_Failed_to_read_number_of_variabl), 0, 0, 1, 0},
    {&__pyx_kp_s_File_0_does_not_exist, __pyx_k_File_0_does_not_exist, sizeof(__pyx_k_File_0_does_not_exist), 0, 0, 1, 0},
    {&__pyx_n_s_HHMM, __pyx_k_HHMM, sizeof(__pyx_k_HHMM), 0, 0, 1, 1},
    {&__pyx_n_s_IS8601DA, __pyx_k_IS8601DA, sizeof(__pyx_k_IS8601DA), 0, 0, 1, 1},
    {&__pyx_n_s_IS8601DT, __pyx_k_IS8601DT, sizeof(__pyx_k_IS8601DT), 0, 0, 1, 1},
    {&__pyx_n_s_IS8601TM, __pyx_k_IS8601TM, sizeof(__pyx_k_IS8601TM), 0, 0, 1, 1},
    {&__pyx_n_s_JDATE, __pyx_k_JDATE, sizeof(__pyx_k_JDATE), 0, 0, 1, 1},
    {&__pyx_n_s_JDATE5, __pyx_k_JDATE5, sizeof(__pyx_k_JDATE5), 0, 0, 1, 1},
    {&__pyx_n_s_JDATE7, __pyx_k_JDATE7, sizeof(__pyx_k_JDATE7), 0, 0, 1, 1},
    {&__pyx_kp_s_Just_defining_a_custom_exceptio, __pyx_k_Just_defining_a_custom_exceptio, sizeof(__pyx_k_Just_defining_a_custom_exceptio), 0, 0, 1, 0},
    {&__pyx_kp_s_Just_defining_a_custom_exceptio_2, __pyx_k_Just_defining_a_custom_exceptio_2, sizeof(__pyx_k_Just_defining_a_custom_exceptio_2), 0, 0, 1, 0},
    {&__pyx_kp_s_M8_ns, __pyx_k_M8_ns, sizeof(__pyx_k_M8_ns), 0, 0, 1, 0},
    {&__pyx_n_s_MDYAMPM, __pyx_k_MDYAMPM, sizeof(__pyx_k_MDYAMPM), 0, 0, 1, 1},
    {&__pyx_n_s_MMDDYY, __pyx_k_MMDDYY, sizeof(__pyx_k_MMDDYY), 0, 0, 1, 1},
    {&__pyx_n_s_MMDDYYB, __pyx_k_MMDDYYB, sizeof(__pyx_k_MMDDYYB), 0, 0, 1, 1},
    {&__pyx_n_s_MMDDYYB10, __pyx_k_MMDDYYB10, sizeof(__pyx_k_MMDDYYB10), 0, 0, 1, 1},
    {&__pyx_n_s_MMDDYYC, __pyx_k_MMDDYYC, sizeof(__pyx_k_MMDDYYC), 0, 0, 1, 1},
    {&__pyx_n_s_MMDDYYC10, __pyx_k_MMDDYYC10, sizeof(__pyx_k_MMDDYYC10), 0, 0, 1, 1},
    {&__pyx_n_s_MMDDYYD, __pyx_k_MMDDYYD, sizeof(__pyx_k_MMDDYYD), 0, 0, 1, 1},
    {&__pyx_n_s_MMDDYYD10, __pyx_k_MMDDYYD10, sizeof(__pyx_k_MMDDYYD10), 0, 0, 1, 1},
    {&__pyx_n_s_MMDDYYN6, __pyx_k_MMDDYYN6, sizeof(__pyx_k_MMDDYYN6), 0, 0, 1, 1},
    {&__pyx_n_s_MMDDYYN8, __pyx_k_MMDDYYN8, sizeof(__pyx_k_MMDDYYN8), 0, 0, 1, 1},
    {&__pyx_n_s_MMDDYYP, __pyx_k_MMDDYYP, sizeof(__pyx_k_MMDDYYP), 0, 0, 1, 1},
    {&__pyx_n_s_MMDDYYP10, __pyx_k_MMDDYYP10, sizeof(__pyx_k_MMDDYYP10), 0, 0, 1, 1},
    {&__pyx_n_s_MMDDYYS, __pyx_k_MMDDYYS, sizeof(__pyx_k_MMDDYYS), 0, 0, 1, 1},
    {&__pyx_n_s_MMDDYYS10, __pyx_k_MMDDYYS10, sizeof(__pyx_k_MMDDYYS10), 0, 0, 1, 1},
    {&__pyx_n_s_NULL, __pyx_k_NULL, sizeof(__pyx_k_NULL), 0, 0, 1, 1},
    {&__pyx_n_s_OrderedDict, __pyx_k_OrderedDict, sizeof(__pyx_k_OrderedDict), 0, 0, 1, 1},
    {&__pyx_n_s_PyreadstatError, __pyx_k_PyreadstatError, sizeof(__pyx_k_PyreadstatError), 0, 0, 1, 1},
    {&__pyx_n_s_ReadstatError, __pyx_k_ReadstatError, sizeof(__pyx_k_ReadstatError), 0, 0, 1, 1},
    {&__pyx_n_s_RuntimeError, __pyx_k_RuntimeError, sizeof(__pyx_k_RuntimeError), 0, 0, 1, 1},
    {&__pyx_n_s_SDATE, __pyx_k_SDATE, sizeof(__pyx_k_SDATE), 0, 0, 1, 1},
    {&__pyx_n_s_SDATE10, __pyx_k_SDATE10, sizeof(__pyx_k_SDATE10), 0, 0, 1, 1},
    {&__pyx_n_s_SDATE8, __pyx_k_SDATE8, sizeof(__pyx_k_SDATE8), 0, 0, 1, 1},
    {&__pyx_kp_s_STRING_type_with_value_s_with_da, __pyx_k_STRING_type_with_value_s_with_da, sizeof(__pyx_k_STRING_type_with_value_s_with_da), 0, 0, 1, 0},
    {&__pyx_n_s_TIME, __pyx_k_TIME, sizeof(__pyx_k_TIME), 0, 0, 1, 1},
    {&__pyx_kp_s_TIME11_2, __pyx_k_TIME11_2, sizeof(__pyx_k_TIME11_2), 0, 0, 1, 0},
    {&__pyx_n_s_TIME20, __pyx_k_TIME20, sizeof(__pyx_k_TIME20), 0, 0, 1, 1},
    {&__pyx_kp_s_TIME20_3, __pyx_k_TIME20_3, sizeof(__pyx_k_TIME20_3), 0, 0, 1, 0},
    {&__pyx_n_s_TIME5, __pyx_k_TIME5, sizeof(__pyx_k_TIME5), 0, 0, 1, 1},
    {&__pyx_n_s_TIME8, __pyx_k_TIME8, sizeof(__pyx_k_TIME8), 0, 0, 1, 1},
    {&__pyx_n_s_TIMEAMPM, __pyx_k_TIMEAMPM, sizeof(__pyx_k_TIMEAMPM), 0, 0, 1, 1},
    {&__pyx_n_s_TOD, __pyx_k_TOD, sizeof(__pyx_k_TOD), 0, 0, 1, 1},
    {&__pyx_kp_s_This_class_holds_metadata_we_wa, __pyx_k_This_class_holds_metadata_we_wa, sizeof(__pyx_k_This_class_holds_metadata_we_wa), 0, 0, 1, 0},
    {&__pyx_kp_s_Time_zones_are_not_available_fro, __pyx_k_Time_zones_are_not_available_fro, sizeof(__pyx_k_Time_zones_are_not_available_fro), 0, 0, 1, 0},
    {&__pyx_n_s_TypeError, __pyx_k_TypeError, sizeof(__pyx_k_TypeError), 0, 0, 1, 1},
    {&__pyx_n_s_UnicodeError, __pyx_k_UnicodeError, sizeof(__pyx_k_UnicodeError), 0, 0, 1, 1},
    {&__pyx_kp_s_Unknown_file_format, __pyx_k_Unknown_file_format, sizeof(__pyx_k_Unknown_file_format), 0, 0, 1, 0},
    {&__pyx_kp_s_Unkown_data_type, __pyx_k_Unkown_data_type, sizeof(__pyx_k_Unkown_data_type), 0, 0, 1, 0},
    {&__pyx_n_s_WEEKDATE, __pyx_k_WEEKDATE, sizeof(__pyx_k_WEEKDATE), 0, 0, 1, 1},
    {&__pyx_n_s_WEEKDATX, __pyx_k_WEEKDATX, sizeof(__pyx_k_WEEKDATX), 0, 0, 1, 1},
    {&__pyx_n_s_YMDHMS16, __pyx_k_YMDHMS16, sizeof(__pyx_k_YMDHMS16), 0, 0, 1, 1},
    {&__pyx_n_s_YMDHMS19, __pyx_k_YMDHMS19, sizeof(__pyx_k_YMDHMS19), 0, 0, 1, 1},
    {&__pyx_kp_s_YMDHMS19_2, __pyx_k_YMDHMS19_2, sizeof(__pyx_k_YMDHMS19_2), 0, 0, 1, 0},
    {&__pyx_n_s_YMDHMS20, __pyx_k_YMDHMS20, sizeof(__pyx_k_YMDHMS20), 0, 0, 1, 1},
    {&__pyx_n_s_YYMMDD, __pyx_k_YYMMDD, sizeof(__pyx_k_YYMMDD), 0, 0, 1, 1},
    {&__pyx_n_s_YYMMDD10, __pyx_k_YYMMDD10, sizeof(__pyx_k_YYMMDD10), 0, 0, 1, 1},
    {&__pyx_n_s_YYMMDDB, __pyx_k_YYMMDDB, sizeof(__pyx_k_YYMMDDB), 0, 0, 1, 1},
    {&__pyx_n_s_YYMMDDD, __pyx_k_YYMMDDD, sizeof(__pyx_k_YYMMDDD), 0, 0, 1, 1},
    {&__pyx_n_s_YYMMDDN, __pyx_k_YYMMDDN, sizeof(__pyx_k_YYMMDDN), 0, 0, 1, 1},
    {&__pyx_n_s_YYMMDDP, __pyx_k_YYMMDDP, sizeof(__pyx_k_YYMMDDP), 0, 0, 1, 1},
    {&__pyx_n_s_YYMMDDS, __pyx_k_YYMMDDS, sizeof(__pyx_k_YYMMDDS), 0, 0, 1, 1},
    {&__pyx_n_s__15, __pyx_k__15, sizeof(__pyx_k__15), 0, 0, 1, 1},
    {&__pyx_kp_s__2, __pyx_k__2, sizeof(__pyx_k__2), 0, 0, 1, 0},
    {&__pyx_kp_u__8, __pyx_k__8, sizeof(__pyx_k__8), 0, 1, 0, 0},
    {&__pyx_n_s__9, __pyx_k__9, sizeof(__pyx_k__9), 0, 0, 1, 1},
    {&__pyx_n_s_allowed_formats, __pyx_k_allowed_formats, sizeof(__pyx_k_allowed_formats), 0, 0, 1, 1},
    {&__pyx_n_s_append, __pyx_k_append, sizeof(__pyx_k_append), 0, 0, 1, 1},
    {&__pyx_n_s_astype, __pyx_k_astype, sizeof(__pyx_k_astype), 0, 0, 1, 1},
    {&__pyx_n_s_asyncio_coroutines, __pyx_k_asyncio_coroutines, sizeof(__pyx_k_asyncio_coroutines), 0, 0, 1, 1},
    {&__pyx_n_s_center, __pyx_k_center, sizeof(__pyx_k_center), 0, 0, 1, 1},
    {&__pyx_n_s_chr, __pyx_k_chr, sizeof(__pyx_k_chr), 0, 0, 1, 1},
    {&__pyx_n_s_class_getitem, __pyx_k_class_getitem, sizeof(__pyx_k_class_getitem), 0, 0, 1, 1},
    {&__pyx_n_s_cline_in_traceback, __pyx_k_cline_in_traceback, sizeof(__pyx_k_cline_in_traceback), 0, 0, 1, 1},
    {&__pyx_n_s_collections, __pyx_k_collections, sizeof(__pyx_k_collections), 0, 0, 1, 1},
    {&__pyx_kp_s_column_0_is_duplicated_renamed_t, __pyx_k_column_0_is_duplicated_renamed_t, sizeof(__pyx_k_column_0_is_duplicated_renamed_t), 0, 0, 1, 0},
    {&__pyx_n_s_column_labels, __pyx_k_column_labels, sizeof(__pyx_k_column_labels), 0, 0, 1, 1},
    {&__pyx_n_s_column_names, __pyx_k_column_names, sizeof(__pyx_k_column_names), 0, 0, 1, 1},
    {&__pyx_n_s_column_names_to_labels, __pyx_k_column_names_to_labels, sizeof(__pyx_k_column_names_to_labels), 0, 0, 1, 1},
    {&__pyx_n_s_columns, __pyx_k_columns, sizeof(__pyx_k_columns), 0, 0, 1, 1},
    {&__pyx_n_s_copy, __pyx_k_copy, sizeof(__pyx_k_copy), 0, 0, 1, 1},
    {&__pyx_n_s_creation_time, __pyx_k_creation_time, sizeof(__pyx_k_creation_time), 0, 0, 1, 1},
    {&__pyx_kp_s_d, __pyx_k_d, sizeof(__pyx_k_d), 0, 0, 1, 0},
    {&__pyx_n_s_data_container, __pyx_k_data_container, sizeof(__pyx_k_data_container), 0, 0, 1, 1},
    {&__pyx_n_s_data_container___reduce_cython, __pyx_k_data_container___reduce_cython, sizeof(__pyx_k_data_container___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_data_container___setstate_cython, __pyx_k_data_container___setstate_cython, sizeof(__pyx_k_data_container___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_date, __pyx_k_date, sizeof(__pyx_k_date), 0, 0, 1, 1},
    {&__pyx_n_s_datetime, __pyx_k_datetime, sizeof(__pyx_k_datetime), 0, 0, 1, 1},
    {&__pyx_n_s_dict, __pyx_k_dict, sizeof(__pyx_k_dict), 0, 0, 1, 1},
    {&__pyx_n_s_dict_2, __pyx_k_dict_2, sizeof(__pyx_k_dict_2), 0, 0, 1, 1},
    {&__pyx_kp_u_disable, __pyx_k_disable, sizeof(__pyx_k_disable), 0, 1, 0, 0},
    {&__pyx_n_s_doc, __pyx_k_doc, sizeof(__pyx_k_doc), 0, 0, 1, 1},
    {&__pyx_n_s_double, __pyx_k_double, sizeof(__pyx_k_double), 0, 0, 1, 1},
    {&__pyx_n_s_dtype, __pyx_k_dtype, sizeof(__pyx_k_dtype), 0, 0, 1, 1},
    {&__pyx_n_s_dtypes, __pyx_k_dtypes, sizeof(__pyx_k_dtypes), 0, 0, 1, 1},
    {&__pyx_n_s_duplicated, __pyx_k_duplicated, sizeof(__pyx_k_duplicated), 0, 0, 1, 1},
    {&__pyx_n_s_empty, __pyx_k_empty, sizeof(__pyx_k_empty), 0, 0, 1, 1},
    {&__pyx_kp_u_enable, __pyx_k_enable, sizeof(__pyx_k_enable), 0, 1, 0, 0},
    {&__pyx_n_s_encode, __pyx_k_encode, sizeof(__pyx_k_encode), 0, 0, 1, 1},
    {&__pyx_n_s_enumerate, __pyx_k_enumerate, sizeof(__pyx_k_enumerate), 0, 0, 1, 1},
    {&__pyx_n_s_expanduser, __pyx_k_expanduser, sizeof(__pyx_k_expanduser), 0, 0, 1, 1},
    {&__pyx_n_s_file_encoding, __pyx_k_file_encoding, sizeof(__pyx_k_file_encoding), 0, 0, 1, 1},
    {&__pyx_n_s_file_label, __pyx_k_file_label, sizeof(__pyx_k_file_label), 0, 0, 1, 1},
    {&__pyx_kp_s_file_path_could_not_be_encoded_w, __pyx_k_file_path_could_not_be_encoded_w, sizeof(__pyx_k_file_path_could_not_be_encoded_w), 0, 0, 1, 0},
    {&__pyx_n_s_fill, __pyx_k_fill, sizeof(__pyx_k_fill), 0, 0, 1, 1},
    {&__pyx_n_s_float, __pyx_k_float, sizeof(__pyx_k_float), 0, 0, 1, 1},
    {&__pyx_n_s_float64, __pyx_k_float64, sizeof(__pyx_k_float64), 0, 0, 1, 1},
    {&__pyx_n_s_format, __pyx_k_format, sizeof(__pyx_k_format), 0, 0, 1, 1},
    {&__pyx_n_s_from_dict, __pyx_k_from_dict, sizeof(__pyx_k_from_dict), 0, 0, 1, 1},
    {&__pyx_n_s_fromtimestamp, __pyx_k_fromtimestamp, sizeof(__pyx_k_fromtimestamp), 0, 0, 1, 1},
    {&__pyx_n_s_fsdecode, __pyx_k_fsdecode, sizeof(__pyx_k_fsdecode), 0, 0, 1, 1},
    {&__pyx_n_s_fsencode, __pyx_k_fsencode, sizeof(__pyx_k_fsencode), 0, 0, 1, 1},
    {&__pyx_kp_u_gc, __pyx_k_gc, sizeof(__pyx_k_gc), 0, 1, 0, 0},
    {&__pyx_n_s_get, __pyx_k_get, sizeof(__pyx_k_get), 0, 0, 1, 1},
    {&__pyx_n_s_getfilesystemencoding, __pyx_k_getfilesystemencoding, sizeof(__pyx_k_getfilesystemencoding), 0, 0, 1, 1},
    {&__pyx_n_s_getstate, __pyx_k_getstate, sizeof(__pyx_k_getstate), 0, 0, 1, 1},
    {&__pyx_n_s_hi, __pyx_k_hi, sizeof(__pyx_k_hi), 0, 0, 1, 1},
    {&__pyx_n_s_import, __pyx_k_import, sizeof(__pyx_k_import), 0, 0, 1, 1},
    {&__pyx_n_s_init, __pyx_k_init, sizeof(__pyx_k_init), 0, 0, 1, 1},
    {&__pyx_n_s_init_subclass, __pyx_k_init_subclass, sizeof(__pyx_k_init_subclass), 0, 0, 1, 1},
    {&__pyx_n_s_initializing, __pyx_k_initializing, sizeof(__pyx_k_initializing), 0, 0, 1, 1},
    {&__pyx_n_s_int16, __pyx_k_int16, sizeof(__pyx_k_int16), 0, 0, 1, 1},
    {&__pyx_n_s_int32, __pyx_k_int32, sizeof(__pyx_k_int32), 0, 0, 1, 1},
    {&__pyx_n_s_int64, __pyx_k_int64, sizeof(__pyx_k_int64), 0, 0, 1, 1},
    {&__pyx_n_s_int8, __pyx_k_int8, sizeof(__pyx_k_int8), 0, 0, 1, 1},
    {&__pyx_n_s_is_coroutine, __pyx_k_is_coroutine, sizeof(__pyx_k_is_coroutine), 0, 0, 1, 1},
    {&__pyx_kp_u_isenabled, __pyx_k_isenabled, sizeof(__pyx_k_isenabled), 0, 1, 0, 0},
    {&__pyx_n_s_isfile, __pyx_k_isfile, sizeof(__pyx_k_isfile), 0, 0, 1, 1},
    {&__pyx_n_s_items, __pyx_k_items, sizeof(__pyx_k_items), 0, 0, 1, 1},
    {&__pyx_n_s_left, __pyx_k_left, sizeof(__pyx_k_left), 0, 0, 1, 1},
    {&__pyx_n_s_lo, __pyx_k_lo, sizeof(__pyx_k_lo), 0, 0, 1, 1},
    {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
    {&__pyx_n_s_metaclass, __pyx_k_metaclass, sizeof(__pyx_k_metaclass), 0, 0, 1, 1},
    {&__pyx_n_s_metadata_container, __pyx_k_metadata_container, sizeof(__pyx_k_metadata_container), 0, 0, 1, 1},
    {&__pyx_n_s_metadata_container___init, __pyx_k_metadata_container___init, sizeof(__pyx_k_metadata_container___init), 0, 0, 1, 1},
    {&__pyx_n_s_missing_ranges, __pyx_k_missing_ranges, sizeof(__pyx_k_missing_ranges), 0, 0, 1, 1},
    {&__pyx_n_s_missing_user_values, __pyx_k_missing_user_values, sizeof(__pyx_k_missing_user_values), 0, 0, 1, 1},
    {&__pyx_n_s_modification_time, __pyx_k_modification_time, sizeof(__pyx_k_modification_time), 0, 0, 1, 1},
    {&__pyx_n_s_module, __pyx_k_module, sizeof(__pyx_k_module), 0, 0, 1, 1},
    {&__pyx_n_s_mro_entries, __pyx_k_mro_entries, sizeof(__pyx_k_mro_entries), 0, 0, 1, 1},
    {&__pyx_n_s_name, __pyx_k_name, sizeof(__pyx_k_name), 0, 0, 1, 1},
    {&__pyx_n_s_name_2, __pyx_k_name_2, sizeof(__pyx_k_name_2), 0, 0, 1, 1},
    {&__pyx_n_s_nan, __pyx_k_nan, sizeof(__pyx_k_nan), 0, 0, 1, 1},
    {&__pyx_kp_s_no_default___reduce___due_to_non, __pyx_k_no_default___reduce___due_to_non, sizeof(__pyx_k_no_default___reduce___due_to_non), 0, 0, 1, 0},
    {&__pyx_n_s_nominal, __pyx_k_nominal, sizeof(__pyx_k_nominal), 0, 0, 1, 1},
    {&__pyx_n_s_notes, __pyx_k_notes, sizeof(__pyx_k_notes), 0, 0, 1, 1},
    {&__pyx_n_s_np, __pyx_k_np, sizeof(__pyx_k_np), 0, 0, 1, 1},
    {&__pyx_n_s_nt, __pyx_k_nt, sizeof(__pyx_k_nt), 0, 0, 1, 1},
    {&__pyx_n_s_number_columns, __pyx_k_number_columns, sizeof(__pyx_k_number_columns), 0, 0, 1, 1},
    {&__pyx_n_s_number_rows, __pyx_k_number_rows, sizeof(__pyx_k_number_rows), 0, 0, 1, 1},
    {&__pyx_n_s_numpy, __pyx_k_numpy, sizeof(__pyx_k_numpy), 0, 0, 1, 1},
    {&__pyx_n_s_object, __pyx_k_object, sizeof(__pyx_k_object), 0, 0, 1, 1},
    {&__pyx_n_s_ordinal, __pyx_k_ordinal, sizeof(__pyx_k_ordinal), 0, 0, 1, 1},
    {&__pyx_n_s_original_variable_types, __pyx_k_original_variable_types, sizeof(__pyx_k_original_variable_types), 0, 0, 1, 1},
    {&__pyx_n_s_os, __pyx_k_os, sizeof(__pyx_k_os), 0, 0, 1, 1},
    {&__pyx_n_s_output_format, __pyx_k_output_format, sizeof(__pyx_k_output_format), 0, 0, 1, 1},
    {&__pyx_kp_s_output_format_must_be_one_of_all, __pyx_k_output_format_must_be_one_of_all, sizeof(__pyx_k_output_format_must_be_one_of_all), 0, 0, 1, 0},
    {&__pyx_n_s_pandas, __pyx_k_pandas, sizeof(__pyx_k_pandas), 0, 0, 1, 1},
    {&__pyx_n_s_path, __pyx_k_path, sizeof(__pyx_k_path), 0, 0, 1, 1},
    {&__pyx_kp_s_path_must_be_either_str_or_bytes, __pyx_k_path_must_be_either_str_or_bytes, sizeof(__pyx_k_path_must_be_either_str_or_bytes), 0, 0, 1, 0},
    {&__pyx_kp_s_path_must_be_str_bytes_or_unicod, __pyx_k_path_must_be_str_bytes_or_unicod, sizeof(__pyx_k_path_must_be_str_bytes_or_unicod), 0, 0, 1, 0},
    {&__pyx_n_s_pd, __pyx_k_pd, sizeof(__pyx_k_pd), 0, 0, 1, 1},
    {&__pyx_n_s_prepare, __pyx_k_prepare, sizeof(__pyx_k_prepare), 0, 0, 1, 1},
    {&__pyx_n_s_pyreadstat__readstat_parser, __pyx_k_pyreadstat__readstat_parser, sizeof(__pyx_k_pyreadstat__readstat_parser), 0, 0, 1, 1},
    {&__pyx_kp_s_pyreadstat__readstat_parser_pyx, __pyx_k_pyreadstat__readstat_parser_pyx, sizeof(__pyx_k_pyreadstat__readstat_parser_pyx), 0, 0, 1, 0},
    {&__pyx_n_s_pyx_capi, __pyx_k_pyx_capi, sizeof(__pyx_k_pyx_capi), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_state, __pyx_k_pyx_state, sizeof(__pyx_k_pyx_state), 0, 0, 1, 1},
    {&__pyx_n_s_qualname, __pyx_k_qualname, sizeof(__pyx_k_qualname), 0, 0, 1, 1},
    {&__pyx_n_s_range, __pyx_k_range, sizeof(__pyx_k_range), 0, 0, 1, 1},
    {&__pyx_n_s_readstat_to_numpy_types, __pyx_k_readstat_to_numpy_types, sizeof(__pyx_k_readstat_to_numpy_types), 0, 0, 1, 1},
    {&__pyx_n_s_readstat_variable_types, __pyx_k_readstat_variable_types, sizeof(__pyx_k_readstat_variable_types), 0, 0, 1, 1},
    {&__pyx_n_s_reduce, __pyx_k_reduce, sizeof(__pyx_k_reduce), 0, 0, 1, 1},
    {&__pyx_n_s_reduce_cython, __pyx_k_reduce_cython, sizeof(__pyx_k_reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_reduce_ex, __pyx_k_reduce_ex, sizeof(__pyx_k_reduce_ex), 0, 0, 1, 1},
    {&__pyx_n_s_right, __pyx_k_right, sizeof(__pyx_k_right), 0, 0, 1, 1},
    {&__pyx_n_s_sas_all_formats, __pyx_k_sas_all_formats, sizeof(__pyx_k_sas_all_formats), 0, 0, 1, 1},
    {&__pyx_n_s_sas_date_formats, __pyx_k_sas_date_formats, sizeof(__pyx_k_sas_date_formats), 0, 0, 1, 1},
    {&__pyx_n_s_sas_datetime_formats, __pyx_k_sas_datetime_formats, sizeof(__pyx_k_sas_datetime_formats), 0, 0, 1, 1},
    {&__pyx_n_s_sas_origin, __pyx_k_sas_origin, sizeof(__pyx_k_sas_origin), 0, 0, 1, 1},
    {&__pyx_n_s_sas_time_formats, __pyx_k_sas_time_formats, sizeof(__pyx_k_sas_time_formats), 0, 0, 1, 1},
    {&__pyx_n_s_scale, __pyx_k_scale, sizeof(__pyx_k_scale), 0, 0, 1, 1},
    {&__pyx_n_s_self, __pyx_k_self, sizeof(__pyx_k_self), 0, 0, 1, 1},
    {&__pyx_n_s_set_name, __pyx_k_set_name, sizeof(__pyx_k_set_name), 0, 0, 1, 1},
    {&__pyx_n_s_setstate, __pyx_k_setstate, sizeof(__pyx_k_setstate), 0, 0, 1, 1},
    {&__pyx_n_s_setstate_cython, __pyx_k_setstate_cython, sizeof(__pyx_k_setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_spec, __pyx_k_spec, sizeof(__pyx_k_spec), 0, 0, 1, 1},
    {&__pyx_n_s_spss_all_formats, __pyx_k_spss_all_formats, sizeof(__pyx_k_spss_all_formats), 0, 0, 1, 1},
    {&__pyx_n_s_spss_date_formats, __pyx_k_spss_date_formats, sizeof(__pyx_k_spss_date_formats), 0, 0, 1, 1},
    {&__pyx_n_s_spss_datetime_formats, __pyx_k_spss_datetime_formats, sizeof(__pyx_k_spss_datetime_formats), 0, 0, 1, 1},
    {&__pyx_n_s_spss_origin, __pyx_k_spss_origin, sizeof(__pyx_k_spss_origin), 0, 0, 1, 1},
    {&__pyx_n_s_spss_time_formats, __pyx_k_spss_time_formats, sizeof(__pyx_k_spss_time_formats), 0, 0, 1, 1},
    {&__pyx_n_s_stata_all_formats, __pyx_k_stata_all_formats, sizeof(__pyx_k_stata_all_formats), 0, 0, 1, 1},
    {&__pyx_n_s_stata_date_formats, __pyx_k_stata_date_formats, sizeof(__pyx_k_stata_date_formats), 0, 0, 1, 1},
    {&__pyx_n_s_stata_datetime_formats, __pyx_k_stata_datetime_formats, sizeof(__pyx_k_stata_datetime_formats), 0, 0, 1, 1},
    {&__pyx_n_s_stata_origin, __pyx_k_stata_origin, sizeof(__pyx_k_stata_origin), 0, 0, 1, 1},
    {&__pyx_n_s_stata_time_formats, __pyx_k_stata_time_formats, sizeof(__pyx_k_stata_time_formats), 0, 0, 1, 1},
    {&__pyx_n_s_string, __pyx_k_string, sizeof(__pyx_k_string), 0, 0, 1, 1},
    {&__pyx_kp_s_stringsource, __pyx_k_stringsource, sizeof(__pyx_k_stringsource), 0, 0, 1, 0},
    {&__pyx_n_s_super, __pyx_k_super, sizeof(__pyx_k_super), 0, 0, 1, 1},
    {&__pyx_n_s_surrogateescape, __pyx_k_surrogateescape, sizeof(__pyx_k_surrogateescape), 0, 0, 1, 1},
    {&__pyx_n_s_sys, __pyx_k_sys, sizeof(__pyx_k_sys), 0, 0, 1, 1},
    {&__pyx_kp_s_tC, __pyx_k_tC, sizeof(__pyx_k_tC), 0, 0, 1, 0},
    {&__pyx_n_s_table_name, __pyx_k_table_name, sizeof(__pyx_k_table_name), 0, 0, 1, 1},
    {&__pyx_kp_s_tc, __pyx_k_tc, sizeof(__pyx_k_tc), 0, 0, 1, 0},
    {&__pyx_kp_s_tcHH_MM, __pyx_k_tcHH_MM, sizeof(__pyx_k_tcHH_MM), 0, 0, 1, 0},
    {&__pyx_kp_s_tcHH_MM_SS, __pyx_k_tcHH_MM_SS, sizeof(__pyx_k_tcHH_MM_SS), 0, 0, 1, 0},
    {&__pyx_kp_s_td, __pyx_k_td, sizeof(__pyx_k_td), 0, 0, 1, 0},
    {&__pyx_kp_s_tdCCYY_NN_DD, __pyx_k_tdCCYY_NN_DD, sizeof(__pyx_k_tdCCYY_NN_DD), 0, 0, 1, 0},
    {&__pyx_kp_s_tdD_m_Y, __pyx_k_tdD_m_Y, sizeof(__pyx_k_tdD_m_Y), 0, 0, 1, 0},
    {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
    {&__pyx_n_s_time, __pyx_k_time, sizeof(__pyx_k_time), 0, 0, 1, 1},
    {&__pyx_n_s_to_datetime, __pyx_k_to_datetime, sizeof(__pyx_k_to_datetime), 0, 0, 1, 1},
    {&__pyx_n_s_tolist, __pyx_k_tolist, sizeof(__pyx_k_tolist), 0, 0, 1, 1},
    {&__pyx_n_s_undetermined, __pyx_k_undetermined, sizeof(__pyx_k_undetermined), 0, 0, 1, 1},
    {&__pyx_n_s_unknown, __pyx_k_unknown, sizeof(__pyx_k_unknown), 0, 0, 1, 1},
    {&__pyx_kp_s_utf_8, __pyx_k_utf_8, sizeof(__pyx_k_utf_8), 0, 0, 1, 0},
    {&__pyx_n_s_value_labels, __pyx_k_value_labels, sizeof(__pyx_k_value_labels), 0, 0, 1, 1},
    {&__pyx_n_s_variable_alignment, __pyx_k_variable_alignment, sizeof(__pyx_k_variable_alignment), 0, 0, 1, 1},
    {&__pyx_n_s_variable_display_width, __pyx_k_variable_display_width, sizeof(__pyx_k_variable_display_width), 0, 0, 1, 1},
    {&__pyx_n_s_variable_measure, __pyx_k_variable_measure, sizeof(__pyx_k_variable_measure), 0, 0, 1, 1},
    {&__pyx_n_s_variable_storage_width, __pyx_k_variable_storage_width, sizeof(__pyx_k_variable_storage_width), 0, 0, 1, 1},
    {&__pyx_n_s_variable_to_label, __pyx_k_variable_to_label, sizeof(__pyx_k_variable_to_label), 0, 0, 1, 1},
    {&__pyx_n_s_variable_value_labels, __pyx_k_variable_value_labels, sizeof(__pyx_k_variable_value_labels), 0, 0, 1, 1},
    {&__pyx_n_s_warn, __pyx_k_warn, sizeof(__pyx_k_warn), 0, 0, 1, 1},
    {&__pyx_n_s_warnings, __pyx_k_warnings, sizeof(__pyx_k_warnings), 0, 0, 1, 1},
    {&__pyx_n_s_zip, __pyx_k_zip, sizeof(__pyx_k_zip), 0, 0, 1, 1},
    {0, 0, 0, 0, 0, 0, 0}
  };
  return __Pyx_InitStrings(__pyx_string_tab);
}
/* #### Code section: cached_builtins ### */
static CYTHON_SMALL_CODE int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_object = __Pyx_GetBuiltinName(__pyx_n_s_object); if (!__pyx_builtin_object) __PYX_ERR(0, 71, __pyx_L1_error)
  __pyx_builtin_TypeError = __Pyx_GetBuiltinName(__pyx_n_s_TypeError); if (!__pyx_builtin_TypeError) __PYX_ERR(1, 2, __pyx_L1_error)
  __pyx_builtin_range = __Pyx_GetBuiltinName(__pyx_n_s_range); if (!__pyx_builtin_range) __PYX_ERR(0, 536, __pyx_L1_error)
  __pyx_builtin_chr = __Pyx_GetBuiltinName(__pyx_n_s_chr); if (!__pyx_builtin_chr) __PYX_ERR(0, 656, __pyx_L1_error)
  __pyx_builtin_enumerate = __Pyx_GetBuiltinName(__pyx_n_s_enumerate); if (!__pyx_builtin_enumerate) __PYX_ERR(0, 938, __pyx_L1_error)
  __pyx_builtin_zip = __Pyx_GetBuiltinName(__pyx_n_s_zip); if (!__pyx_builtin_zip) __PYX_ERR(0, 1018, __pyx_L1_error)
  __pyx_builtin_UnicodeError = __Pyx_GetBuiltinName(__pyx_n_s_UnicodeError); if (!__pyx_builtin_UnicodeError) __PYX_ERR(0, 1058, __pyx_L1_error)
  __pyx_builtin_RuntimeError = __Pyx_GetBuiltinName(__pyx_n_s_RuntimeError); if (!__pyx_builtin_RuntimeError) __PYX_ERR(2, 308, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}
/* #### Code section: cached_constants ### */

static CYTHON_SMALL_CODE int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "cpython/datetime.pxd":308
 * cdef inline object timezone_new(object offset, object name=None):
 *     if PY_VERSION_HEX < 0x030700b1:
 *         raise RuntimeError('Time zones are not available from the C-API.')             # <<<<<<<<<<<<<<
 *     return __Pyx_TimeZone_FromOffsetAndName(offset, <PyObject*>name if name is not None else NULL)
 * 
 */
  __pyx_tuple_ = PyTuple_Pack(1, __pyx_kp_s_Time_zones_are_not_available_fro); if (unlikely(!__pyx_tuple_)) __PYX_ERR(2, 308, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple_);
  __Pyx_GIVEREF(__pyx_tuple_);

  /* "pyreadstat/_readstat_parser.pyx":523
 *     # pre-allocate data
 *     if metaonly:
 *         row = np.empty(1, dtype=curnptype)             # <<<<<<<<<<<<<<
 *     else:
 *         obs_count = dc.n_obs
 */
  __pyx_tuple__3 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__3)) __PYX_ERR(0, 523, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__3);
  __Pyx_GIVEREF(__pyx_tuple__3);

  /* "pyreadstat/_readstat_parser.pyx":622
 *         if var_max_rows <= obs_index:
 *             curnptype = dc.col_numpy_dtypes[index]
 *             buf_list = np.empty(100000, dtype=curnptype)             # <<<<<<<<<<<<<<
 *             if iscurnptypeobject or iscurnptypefloat:
 *                 buf_list.fill(np.nan)
 */
  __pyx_tuple__4 = PyTuple_Pack(1, __pyx_int_100000); if (unlikely(!__pyx_tuple__4)) __PYX_ERR(0, 622, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__4);
  __Pyx_GIVEREF(__pyx_tuple__4);

  /* "pyreadstat/_readstat_parser.pyx":642
 *                 dc.col_dtypes_isobject[index] = 1
 *                 iscurnptypeobject = 1
 *                 dc.col_data[index] = dc.col_data[index].astype(object, copy=False)             # <<<<<<<<<<<<<<
 *                 dc.col_data[index][obs_index:] = np.nan
 *                 #dc.col_data[index][obs_index] = NAN
 */
  __pyx_tuple__5 = PyTuple_Pack(1, __pyx_builtin_object); if (unlikely(!__pyx_tuple__5)) __PYX_ERR(0, 642, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__5);
  __Pyx_GIVEREF(__pyx_tuple__5);

  /* "pyreadstat/_readstat_parser.pyx":662
 *                 dc.col_dytpes_isfloat[index] = 0
 *                 iscurnptypeobject = 1
 *                 dc.col_data[index] = dc.col_data[index].astype(object, copy=False)             # <<<<<<<<<<<<<<
 *                 dc.col_data[index][obs_index] =  chr(missing_tag)
 *             curset = dc.missing_user_values.get(index)
 */
  __pyx_tuple__6 = PyTuple_Pack(1, __pyx_builtin_object); if (unlikely(!__pyx_tuple__6)) __PYX_ERR(0, 662, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__6);
  __Pyx_GIVEREF(__pyx_tuple__6);

  /* "pyreadstat/_readstat_parser.pyx":1060
 *         except UnicodeError:
 *             warnings.warn("file path could not be encoded with %s which is set as your system encoding, trying to encode it as utf-8. Please set your system encoding correctly." % sys.getfilesystemencoding())
 *             filename_bytes = os.fsdecode(filename_path).encode("utf-8", "surrogateescape")             # <<<<<<<<<<<<<<
 *     else:
 *         if type(filename_path) == str:
 */
  __pyx_tuple__7 = PyTuple_Pack(2, __pyx_kp_s_utf_8, __pyx_n_s_surrogateescape); if (unlikely(!__pyx_tuple__7)) __PYX_ERR(0, 1060, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__7);
  __Pyx_GIVEREF(__pyx_tuple__7);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_tuple__10 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__10)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__10);
  __Pyx_GIVEREF(__pyx_tuple__10);
  __pyx_codeobj__11 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__10, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__11)) __PYX_ERR(1, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_tuple__12 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_pyx_state); if (unlikely(!__pyx_tuple__12)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__12);
  __Pyx_GIVEREF(__pyx_tuple__12);
  __pyx_codeobj__13 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__12, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__13)) __PYX_ERR(1, 3, __pyx_L1_error)

  /* "pyreadstat/_readstat_parser.pyx":123
 *     This class holds metadata we want to give back to python
 *     """
 *     def __init__(self):             # <<<<<<<<<<<<<<
 *         self.column_names = list()
 *         self.column_labels = list()
 */
  __pyx_codeobj__14 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__10, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_pyreadstat__readstat_parser_pyx, __pyx_n_s_init, 123, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__14)) __PYX_ERR(0, 123, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}
/* #### Code section: init_constants ### */

static CYTHON_SMALL_CODE int __Pyx_InitConstants(void) {
  __pyx_umethod_PyDict_Type_get.type = (PyObject*)&PyDict_Type;
  __pyx_umethod_PyDict_Type_get.method_name = &__pyx_n_s_get;
  __pyx_umethod_PyDict_Type_items.type = (PyObject*)&PyDict_Type;
  __pyx_umethod_PyDict_Type_items.method_name = &__pyx_n_s_items;
  if (__Pyx_CreateStringTabAndInitStrings() < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  __pyx_int_0 = PyInt_FromLong(0); if (unlikely(!__pyx_int_0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_1 = PyInt_FromLong(1); if (unlikely(!__pyx_int_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_100000 = PyInt_FromLong(100000L); if (unlikely(!__pyx_int_100000)) __PYX_ERR(0, 1, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}
/* #### Code section: init_globals ### */

static CYTHON_SMALL_CODE int __Pyx_InitGlobals(void) {
  return 0;
}
/* #### Code section: init_module ### */

static CYTHON_SMALL_CODE int __Pyx_modinit_global_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_import_code(void); /*proto*/

static int __Pyx_modinit_global_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  __pyx_v_10pyreadstat_16_readstat_parser_readstat_to_numpy_types = ((PyObject*)Py_None); Py_INCREF(Py_None);
  __pyx_v_10pyreadstat_16_readstat_parser_sas_date_formats = ((PyObject*)Py_None); Py_INCREF(Py_None);
  __pyx_v_10pyreadstat_16_readstat_parser_sas_datetime_formats = ((PyObject*)Py_None); Py_INCREF(Py_None);
  __pyx_v_10pyreadstat_16_readstat_parser_sas_time_formats = ((PyObject*)Py_None); Py_INCREF(Py_None);
  __pyx_v_10pyreadstat_16_readstat_parser_sas_all_formats = ((PyObject*)Py_None); Py_INCREF(Py_None);
  __pyx_v_10pyreadstat_16_readstat_parser_sas_origin = Py_None; Py_INCREF(Py_None);
  __pyx_v_10pyreadstat_16_readstat_parser_spss_datetime_formats = ((PyObject*)Py_None); Py_INCREF(Py_None);
  __pyx_v_10pyreadstat_16_readstat_parser_spss_date_formats = ((PyObject*)Py_None); Py_INCREF(Py_None);
  __pyx_v_10pyreadstat_16_readstat_parser_spss_time_formats = ((PyObject*)Py_None); Py_INCREF(Py_None);
  __pyx_v_10pyreadstat_16_readstat_parser_spss_all_formats = ((PyObject*)Py_None); Py_INCREF(Py_None);
  __pyx_v_10pyreadstat_16_readstat_parser_spss_origin = Py_None; Py_INCREF(Py_None);
  __pyx_v_10pyreadstat_16_readstat_parser_stata_datetime_formats = ((PyObject*)Py_None); Py_INCREF(Py_None);
  __pyx_v_10pyreadstat_16_readstat_parser_stata_date_formats = ((PyObject*)Py_None); Py_INCREF(Py_None);
  __pyx_v_10pyreadstat_16_readstat_parser_stata_time_formats = ((PyObject*)Py_None); Py_INCREF(Py_None);
  __pyx_v_10pyreadstat_16_readstat_parser_stata_all_formats = ((PyObject*)Py_None); Py_INCREF(Py_None);
  __pyx_v_10pyreadstat_16_readstat_parser_stata_origin = Py_None; Py_INCREF(Py_None);
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(void) {
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  if (__Pyx_ExportVoidPtr(__pyx_n_s_readstat_to_numpy_types, (void *)&__pyx_v_10pyreadstat_16_readstat_parser_readstat_to_numpy_types, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportVoidPtr(__pyx_n_s_sas_date_formats, (void *)&__pyx_v_10pyreadstat_16_readstat_parser_sas_date_formats, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportVoidPtr(__pyx_n_s_sas_datetime_formats, (void *)&__pyx_v_10pyreadstat_16_readstat_parser_sas_datetime_formats, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportVoidPtr(__pyx_n_s_sas_time_formats, (void *)&__pyx_v_10pyreadstat_16_readstat_parser_sas_time_formats, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportVoidPtr(__pyx_n_s_sas_all_formats, (void *)&__pyx_v_10pyreadstat_16_readstat_parser_sas_all_formats, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportVoidPtr(__pyx_n_s_sas_origin, (void *)&__pyx_v_10pyreadstat_16_readstat_parser_sas_origin, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportVoidPtr(__pyx_n_s_spss_datetime_formats, (void *)&__pyx_v_10pyreadstat_16_readstat_parser_spss_datetime_formats, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportVoidPtr(__pyx_n_s_spss_date_formats, (void *)&__pyx_v_10pyreadstat_16_readstat_parser_spss_date_formats, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportVoidPtr(__pyx_n_s_spss_time_formats, (void *)&__pyx_v_10pyreadstat_16_readstat_parser_spss_time_formats, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportVoidPtr(__pyx_n_s_spss_all_formats, (void *)&__pyx_v_10pyreadstat_16_readstat_parser_spss_all_formats, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportVoidPtr(__pyx_n_s_spss_origin, (void *)&__pyx_v_10pyreadstat_16_readstat_parser_spss_origin, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportVoidPtr(__pyx_n_s_stata_datetime_formats, (void *)&__pyx_v_10pyreadstat_16_readstat_parser_stata_datetime_formats, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportVoidPtr(__pyx_n_s_stata_date_formats, (void *)&__pyx_v_10pyreadstat_16_readstat_parser_stata_date_formats, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportVoidPtr(__pyx_n_s_stata_time_formats, (void *)&__pyx_v_10pyreadstat_16_readstat_parser_stata_time_formats, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportVoidPtr(__pyx_n_s_stata_all_formats, (void *)&__pyx_v_10pyreadstat_16_readstat_parser_stata_all_formats, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportVoidPtr(__pyx_n_s_stata_origin, (void *)&__pyx_v_10pyreadstat_16_readstat_parser_stata_origin, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_function_export_code(void) {
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  if (__Pyx_ExportFunction("transform_variable_format", (void (*)(void))__pyx_f_10pyreadstat_16_readstat_parser_transform_variable_format, "__pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format (PyObject *, __pyx_t_10pyreadstat_16_readstat_parser_py_file_format)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("transform_datetime", (void (*)(void))__pyx_f_10pyreadstat_16_readstat_parser_transform_datetime, "PyObject *(__pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format, double, __pyx_t_10pyreadstat_16_readstat_parser_py_file_format, PyObject *, int)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("handle_metadata", (void (*)(void))__pyx_f_10pyreadstat_16_readstat_parser_handle_metadata, "int (readstat_metadata_t *, void *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("handle_variable", (void (*)(void))__pyx_f_10pyreadstat_16_readstat_parser_handle_variable, "int (int, readstat_variable_t *, char *, void *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("handle_value", (void (*)(void))__pyx_f_10pyreadstat_16_readstat_parser_handle_value, "int (int, readstat_variable_t *, readstat_value_t, void *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("handle_value_label", (void (*)(void))__pyx_f_10pyreadstat_16_readstat_parser_handle_value_label, "int (char *, readstat_value_t, char *, void *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("handle_note", (void (*)(void))__pyx_f_10pyreadstat_16_readstat_parser_handle_note, "int (int, char *, void *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("check_exit_status", (void (*)(void))__pyx_f_10pyreadstat_16_readstat_parser_check_exit_status, "void (readstat_error_t)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("run_readstat_parser", (void (*)(void))__pyx_f_10pyreadstat_16_readstat_parser_run_readstat_parser, "void (char *, struct __pyx_obj_10pyreadstat_16_readstat_parser_data_container *, __pyx_t_10pyreadstat_16_readstat_parser_py_file_extension, long, long)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("data_container_to_dict", (void (*)(void))__pyx_f_10pyreadstat_16_readstat_parser_data_container_to_dict, "PyObject *(struct __pyx_obj_10pyreadstat_16_readstat_parser_data_container *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("dict_to_pandas_dataframe", (void (*)(void))__pyx_f_10pyreadstat_16_readstat_parser_dict_to_pandas_dataframe, "PyObject *(PyObject *, struct __pyx_obj_10pyreadstat_16_readstat_parser_data_container *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("data_container_extract_metadata", (void (*)(void))__pyx_f_10pyreadstat_16_readstat_parser_data_container_extract_metadata, "PyObject *(struct __pyx_obj_10pyreadstat_16_readstat_parser_data_container *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("run_conversion", (void (*)(void))__pyx_f_10pyreadstat_16_readstat_parser_run_conversion, "PyObject *(PyObject *, __pyx_t_10pyreadstat_16_readstat_parser_py_file_format, __pyx_t_10pyreadstat_16_readstat_parser_py_file_extension, PyObject *, int, int, PyObject *, int, int, long, long, PyObject *, PyObject *, PyObject *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_init_code(void) {
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_10pyreadstat_16_readstat_parser_data_container = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_10pyreadstat_16_readstat_parser_data_container_spec, NULL); if (unlikely(!__pyx_ptype_10pyreadstat_16_readstat_parser_data_container)) __PYX_ERR(0, 75, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_10pyreadstat_16_readstat_parser_data_container_spec, __pyx_ptype_10pyreadstat_16_readstat_parser_data_container) < 0) __PYX_ERR(0, 75, __pyx_L1_error)
  #else
  __pyx_ptype_10pyreadstat_16_readstat_parser_data_container = &__pyx_type_10pyreadstat_16_readstat_parser_data_container;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_10pyreadstat_16_readstat_parser_data_container) < 0) __PYX_ERR(0, 75, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_10pyreadstat_16_readstat_parser_data_container->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_10pyreadstat_16_readstat_parser_data_container->tp_dictoffset && __pyx_ptype_10pyreadstat_16_readstat_parser_data_container->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_10pyreadstat_16_readstat_parser_data_container->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_data_container, (PyObject *) __pyx_ptype_10pyreadstat_16_readstat_parser_data_container) < 0) __PYX_ERR(0, 75, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_10pyreadstat_16_readstat_parser_data_container) < 0) __PYX_ERR(0, 75, __pyx_L1_error)
  #endif
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_4type_type = __Pyx_ImportType_3_0_0(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "type", 
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyTypeObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_0(PyTypeObject),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(PyTypeObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_0(PyTypeObject),
  #else
  sizeof(PyHeapTypeObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_0(PyHeapTypeObject),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_0_0); if (!__pyx_ptype_7cpython_4type_type) __PYX_ERR(3, 9, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("datetime"); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 70, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_8datetime_date = __Pyx_ImportType_3_0_0(__pyx_t_1, "datetime", "date", sizeof(PyDateTime_Date), __PYX_GET_STRUCT_ALIGNMENT_3_0_0(PyDateTime_Date),__Pyx_ImportType_CheckSize_Warn_3_0_0); if (!__pyx_ptype_7cpython_8datetime_date) __PYX_ERR(2, 70, __pyx_L1_error)
  __pyx_ptype_7cpython_8datetime_time = __Pyx_ImportType_3_0_0(__pyx_t_1, "datetime", "time", sizeof(PyDateTime_Time), __PYX_GET_STRUCT_ALIGNMENT_3_0_0(PyDateTime_Time),__Pyx_ImportType_CheckSize_Warn_3_0_0); if (!__pyx_ptype_7cpython_8datetime_time) __PYX_ERR(2, 83, __pyx_L1_error)
  __pyx_ptype_7cpython_8datetime_datetime = __Pyx_ImportType_3_0_0(__pyx_t_1, "datetime", "datetime", sizeof(PyDateTime_DateTime), __PYX_GET_STRUCT_ALIGNMENT_3_0_0(PyDateTime_DateTime),__Pyx_ImportType_CheckSize_Warn_3_0_0); if (!__pyx_ptype_7cpython_8datetime_datetime) __PYX_ERR(2, 109, __pyx_L1_error)
  __pyx_ptype_7cpython_8datetime_timedelta = __Pyx_ImportType_3_0_0(__pyx_t_1, "datetime", "timedelta", sizeof(PyDateTime_Delta), __PYX_GET_STRUCT_ALIGNMENT_3_0_0(PyDateTime_Delta),__Pyx_ImportType_CheckSize_Warn_3_0_0); if (!__pyx_ptype_7cpython_8datetime_timedelta) __PYX_ERR(2, 147, __pyx_L1_error)
  __pyx_ptype_7cpython_8datetime_tzinfo = __Pyx_ImportType_3_0_0(__pyx_t_1, "datetime", "tzinfo", sizeof(PyDateTime_TZInfo), __PYX_GET_STRUCT_ALIGNMENT_3_0_0(PyDateTime_TZInfo),__Pyx_ImportType_CheckSize_Warn_3_0_0); if (!__pyx_ptype_7cpython_8datetime_tzinfo) __PYX_ERR(2, 160, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_variable_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}


#if PY_MAJOR_VERSION >= 3
#if CYTHON_PEP489_MULTI_PHASE_INIT
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/
static int __pyx_pymod_exec__readstat_parser(PyObject* module); /*proto*/
static PyModuleDef_Slot __pyx_moduledef_slots[] = {
  {Py_mod_create, (void*)__pyx_pymod_create},
  {Py_mod_exec, (void*)__pyx_pymod_exec__readstat_parser},
  {0, NULL}
};
#endif

#ifdef __cplusplus
namespace {
  struct PyModuleDef __pyx_moduledef =
  #else
  static struct PyModuleDef __pyx_moduledef =
  #endif
  {
      PyModuleDef_HEAD_INIT,
      "_readstat_parser",
      0, /* m_doc */
    #if CYTHON_PEP489_MULTI_PHASE_INIT
      0, /* m_size */
    #elif CYTHON_USE_MODULE_STATE
      sizeof(__pyx_mstate), /* m_size */
    #else
      -1, /* m_size */
    #endif
      __pyx_methods /* m_methods */,
    #if CYTHON_PEP489_MULTI_PHASE_INIT
      __pyx_moduledef_slots, /* m_slots */
    #else
      NULL, /* m_reload */
    #endif
    #if CYTHON_USE_MODULE_STATE
      __pyx_m_traverse, /* m_traverse */
      __pyx_m_clear, /* m_clear */
      NULL /* m_free */
    #else
      NULL, /* m_traverse */
      NULL, /* m_clear */
      NULL /* m_free */
    #endif
  };
  #ifdef __cplusplus
} /* anonymous namespace */
#endif
#endif

#ifndef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#elif PY_MAJOR_VERSION < 3
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" void
#else
#define __Pyx_PyMODINIT_FUNC void
#endif
#else
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" PyObject *
#else
#define __Pyx_PyMODINIT_FUNC PyObject *
#endif
#endif


#if PY_MAJOR_VERSION < 3
__Pyx_PyMODINIT_FUNC init_readstat_parser(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC init_readstat_parser(void)
#else
__Pyx_PyMODINIT_FUNC PyInit__readstat_parser(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit__readstat_parser(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
static CYTHON_SMALL_CODE int __Pyx_check_single_interpreter(void) {
    #if PY_VERSION_HEX >= 0x030700A1
    static PY_INT64_T main_interpreter_id = -1;
    PY_INT64_T current_id = PyInterpreterState_GetID(PyThreadState_Get()->interp);
    if (main_interpreter_id == -1) {
        main_interpreter_id = current_id;
        return (unlikely(current_id == -1)) ? -1 : 0;
    } else if (unlikely(main_interpreter_id != current_id))
    #else
    static PyInterpreterState *main_interpreter = NULL;
    PyInterpreterState *current_interpreter = PyThreadState_Get()->interp;
    if (!main_interpreter) {
        main_interpreter = current_interpreter;
    } else if (unlikely(main_interpreter != current_interpreter))
    #endif
    {
        PyErr_SetString(
            PyExc_ImportError,
            "Interpreter change detected - this module can only be loaded into one interpreter per process.");
        return -1;
    }
    return 0;
}
#if CYTHON_COMPILING_IN_LIMITED_API
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *module, const char* from_name, const char* to_name, int allow_none)
#else
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name, int allow_none)
#endif
{
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        if (allow_none || value != Py_None) {
#if CYTHON_COMPILING_IN_LIMITED_API
            result = PyModule_AddObject(module, to_name, value);
#else
            result = PyDict_SetItemString(moddict, to_name, value);
#endif
        }
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static CYTHON_SMALL_CODE PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    CYTHON_UNUSED_VAR(def);
    if (__Pyx_check_single_interpreter())
        return NULL;
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
#if CYTHON_COMPILING_IN_LIMITED_API
    moddict = module;
#else
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
#endif
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__", 0) < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static CYTHON_SMALL_CODE int __pyx_pymod_exec__readstat_parser(PyObject *__pyx_pyinit_module)
#endif
#endif
{
  int stringtab_initialized = 0;
  #if CYTHON_USE_MODULE_STATE
  int pystate_addmodule_run = 0;
  #endif
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m) {
    if (__pyx_m == __pyx_pyinit_module) return 0;
    PyErr_SetString(PyExc_RuntimeError, "Module '_readstat_parser' has already been imported. Re-initialisation is not supported.");
    return -1;
  }
  #elif PY_MAJOR_VERSION >= 3
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_m = __pyx_pyinit_module;
  Py_INCREF(__pyx_m);
  #else
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("_readstat_parser", __pyx_methods, 0, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 1, __pyx_L1_error)
  #elif CYTHON_USE_MODULE_STATE
  __pyx_t_1 = PyModule_Create(&__pyx_moduledef); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  {
    int add_module_result = PyState_AddModule(__pyx_t_1, &__pyx_moduledef);
    __pyx_t_1 = 0; /* transfer ownership from __pyx_t_1 to _readstat_parser pseudovariable */
    if (unlikely((add_module_result < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
    pystate_addmodule_run = 1;
  }
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #endif
  CYTHON_UNUSED_VAR(__pyx_t_1);
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = PyImport_AddModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_b);
  __pyx_cython_runtime = PyImport_AddModule((char *) "cython_runtime"); if (unlikely(!__pyx_cython_runtime)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_cython_runtime);
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #if CYTHON_REFNANNY
__Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
if (!__Pyx_RefNanny) {
  PyErr_Clear();
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
  if (!__Pyx_RefNanny)
      Py_FatalError("failed to import 'refnanny' module");
}
#endif
  __Pyx_RefNannySetupContext("__Pyx_PyMODINIT_FUNC PyInit__readstat_parser(void)", 0);
  if (__Pyx_check_binary_version() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pxy_PyFrame_Initialize_Offsets
  __Pxy_PyFrame_Initialize_Offsets();
  #endif
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init(__pyx_m) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init(__pyx_m) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init(__pyx_m) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init(__pyx_m) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_AsyncGen_USED
  if (__pyx_AsyncGen_init(__pyx_m) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init(__pyx_m) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(WITH_THREAD) && PY_VERSION_HEX < 0x030700F0 && defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  PyEval_InitThreads();
  #endif
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitConstants() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  stringtab_initialized = 1;
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main_pyreadstat___readstat_parser) {
    if (PyObject_SetAttr(__pyx_m, __pyx_n_s_name_2, __pyx_n_s_main) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "pyreadstat._readstat_parser")) {
      if (unlikely((PyDict_SetItemString(modules, "pyreadstat._readstat_parser", __pyx_m) < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code();
  if (unlikely((__Pyx_modinit_variable_export_code() < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
  if (unlikely((__Pyx_modinit_function_export_code() < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
  if (unlikely((__Pyx_modinit_type_init_code() < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
  if (unlikely((__Pyx_modinit_type_import_code() < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
  (void)__Pyx_modinit_variable_import_code();
  (void)__Pyx_modinit_function_import_code();
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif

  /* "pyreadstat/_readstat_parser.pyx":25
 * 
 * #from datetime import timedelta, datetime
 * from collections import OrderedDict             # <<<<<<<<<<<<<<
 * import datetime
 * import os
 */
  __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_OrderedDict);
  __Pyx_GIVEREF(__pyx_n_s_OrderedDict);
  PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_OrderedDict);
  __pyx_t_3 = __Pyx_Import(__pyx_n_s_collections, __pyx_t_2, -1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_OrderedDict); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_OrderedDict, __pyx_t_2) < 0) __PYX_ERR(0, 25, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "pyreadstat/_readstat_parser.pyx":26
 * #from datetime import timedelta, datetime
 * from collections import OrderedDict
 * import datetime             # <<<<<<<<<<<<<<
 * import os
 * import warnings
 */
  __pyx_t_3 = __Pyx_ImportDottedModuleRelFirst(__pyx_n_s_datetime, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_datetime, __pyx_t_3) < 0) __PYX_ERR(0, 26, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "pyreadstat/_readstat_parser.pyx":27
 * from collections import OrderedDict
 * import datetime
 * import os             # <<<<<<<<<<<<<<
 * import warnings
 * import sys
 */
  __pyx_t_3 = __Pyx_ImportDottedModuleRelFirst(__pyx_n_s_os, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_os, __pyx_t_3) < 0) __PYX_ERR(0, 27, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "pyreadstat/_readstat_parser.pyx":28
 * import datetime
 * import os
 * import warnings             # <<<<<<<<<<<<<<
 * import sys
 * 
 */
  __pyx_t_3 = __Pyx_ImportDottedModuleRelFirst(__pyx_n_s_warnings, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_warnings, __pyx_t_3) < 0) __PYX_ERR(0, 28, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "pyreadstat/_readstat_parser.pyx":29
 * import os
 * import warnings
 * import sys             # <<<<<<<<<<<<<<
 * 
 * import pandas as pd
 */
  __pyx_t_3 = __Pyx_ImportDottedModuleRelFirst(__pyx_n_s_sys, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_sys, __pyx_t_3) < 0) __PYX_ERR(0, 29, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "pyreadstat/_readstat_parser.pyx":31
 * import sys
 * 
 * import pandas as pd             # <<<<<<<<<<<<<<
 * import numpy as np
 * #from pandas._libs import Timestamp
 */
  __pyx_t_3 = __Pyx_ImportDottedModuleRelFirst(__pyx_n_s_pandas, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pd, __pyx_t_3) < 0) __PYX_ERR(0, 31, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "pyreadstat/_readstat_parser.pyx":32
 * 
 * import pandas as pd
 * import numpy as np             # <<<<<<<<<<<<<<
 * #from pandas._libs import Timestamp
 * 
 */
  __pyx_t_3 = __Pyx_ImportDottedModuleRelFirst(__pyx_n_s_numpy, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_np, __pyx_t_3) < 0) __PYX_ERR(0, 32, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "pyreadstat/_readstat_parser.pyx":38
 * 
 * # necessary to work with the datetime C API
 * import_datetime()             # <<<<<<<<<<<<<<
 * 
 * cdef list sas_date_formats = ["WEEKDATE", "MMDDYY", "DDMMYY", "YYMMDD", "DATE", "DATE9", "YYMMDD10",
 */
  __pyx_f_7cpython_8datetime_import_datetime(); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 38, __pyx_L1_error)

  /* "pyreadstat/_readstat_parser.pyx":40
 * import_datetime()
 * 
 * cdef list sas_date_formats = ["WEEKDATE", "MMDDYY", "DDMMYY", "YYMMDD", "DATE", "DATE9", "YYMMDD10",             # <<<<<<<<<<<<<<
 *                                 "DDMMYYB", "DDMMYYB10", "DDMMYYC", "DDMMYYC10", "DDMMYYD", "DDMMYYD10",
 *                                 "DDMMYYN6", "DDMMYYN8", "DDMMYYP", "DDMMYYP10", "DDMMYYS", "DDMMYYS10",
 */
  __pyx_t_3 = PyList_New(41); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 40, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_n_s_WEEKDATE);
  __Pyx_GIVEREF(__pyx_n_s_WEEKDATE);
  PyList_SET_ITEM(__pyx_t_3, 0, __pyx_n_s_WEEKDATE);
  __Pyx_INCREF(__pyx_n_s_MMDDYY);
  __Pyx_GIVEREF(__pyx_n_s_MMDDYY);
  PyList_SET_ITEM(__pyx_t_3, 1, __pyx_n_s_MMDDYY);
  __Pyx_INCREF(__pyx_n_s_DDMMYY);
  __Pyx_GIVEREF(__pyx_n_s_DDMMYY);
  PyList_SET_ITEM(__pyx_t_3, 2, __pyx_n_s_DDMMYY);
  __Pyx_INCREF(__pyx_n_s_YYMMDD);
  __Pyx_GIVEREF(__pyx_n_s_YYMMDD);
  PyList_SET_ITEM(__pyx_t_3, 3, __pyx_n_s_YYMMDD);
  __Pyx_INCREF(__pyx_n_s_DATE);
  __Pyx_GIVEREF(__pyx_n_s_DATE);
  PyList_SET_ITEM(__pyx_t_3, 4, __pyx_n_s_DATE);
  __Pyx_INCREF(__pyx_n_s_DATE9);
  __Pyx_GIVEREF(__pyx_n_s_DATE9);
  PyList_SET_ITEM(__pyx_t_3, 5, __pyx_n_s_DATE9);
  __Pyx_INCREF(__pyx_n_s_YYMMDD10);
  __Pyx_GIVEREF(__pyx_n_s_YYMMDD10);
  PyList_SET_ITEM(__pyx_t_3, 6, __pyx_n_s_YYMMDD10);
  __Pyx_INCREF(__pyx_n_s_DDMMYYB);
  __Pyx_GIVEREF(__pyx_n_s_DDMMYYB);
  PyList_SET_ITEM(__pyx_t_3, 7, __pyx_n_s_DDMMYYB);
  __Pyx_INCREF(__pyx_n_s_DDMMYYB10);
  __Pyx_GIVEREF(__pyx_n_s_DDMMYYB10);
  PyList_SET_ITEM(__pyx_t_3, 8, __pyx_n_s_DDMMYYB10);
  __Pyx_INCREF(__pyx_n_s_DDMMYYC);
  __Pyx_GIVEREF(__pyx_n_s_DDMMYYC);
  PyList_SET_ITEM(__pyx_t_3, 9, __pyx_n_s_DDMMYYC);
  __Pyx_INCREF(__pyx_n_s_DDMMYYC10);
  __Pyx_GIVEREF(__pyx_n_s_DDMMYYC10);
  PyList_SET_ITEM(__pyx_t_3, 10, __pyx_n_s_DDMMYYC10);
  __Pyx_INCREF(__pyx_n_s_DDMMYYD);
  __Pyx_GIVEREF(__pyx_n_s_DDMMYYD);
  PyList_SET_ITEM(__pyx_t_3, 11, __pyx_n_s_DDMMYYD);
  __Pyx_INCREF(__pyx_n_s_DDMMYYD10);
  __Pyx_GIVEREF(__pyx_n_s_DDMMYYD10);
  PyList_SET_ITEM(__pyx_t_3, 12, __pyx_n_s_DDMMYYD10);
  __Pyx_INCREF(__pyx_n_s_DDMMYYN6);
  __Pyx_GIVEREF(__pyx_n_s_DDMMYYN6);
  PyList_SET_ITEM(__pyx_t_3, 13, __pyx_n_s_DDMMYYN6);
  __Pyx_INCREF(__pyx_n_s_DDMMYYN8);
  __Pyx_GIVEREF(__pyx_n_s_DDMMYYN8);
  PyList_SET_ITEM(__pyx_t_3, 14, __pyx_n_s_DDMMYYN8);
  __Pyx_INCREF(__pyx_n_s_DDMMYYP);
  __Pyx_GIVEREF(__pyx_n_s_DDMMYYP);
  PyList_SET_ITEM(__pyx_t_3, 15, __pyx_n_s_DDMMYYP);
  __Pyx_INCREF(__pyx_n_s_DDMMYYP10);
  __Pyx_GIVEREF(__pyx_n_s_DDMMYYP10);
  PyList_SET_ITEM(__pyx_t_3, 16, __pyx_n_s_DDMMYYP10);
  __Pyx_INCREF(__pyx_n_s_DDMMYYS);
  __Pyx_GIVEREF(__pyx_n_s_DDMMYYS);
  PyList_SET_ITEM(__pyx_t_3, 17, __pyx_n_s_DDMMYYS);
  __Pyx_INCREF(__pyx_n_s_DDMMYYS10);
  __Pyx_GIVEREF(__pyx_n_s_DDMMYYS10);
  PyList_SET_ITEM(__pyx_t_3, 18, __pyx_n_s_DDMMYYS10);
  __Pyx_INCREF(__pyx_n_s_MMDDYYB);
  __Pyx_GIVEREF(__pyx_n_s_MMDDYYB);
  PyList_SET_ITEM(__pyx_t_3, 19, __pyx_n_s_MMDDYYB);
  __Pyx_INCREF(__pyx_n_s_MMDDYYB10);
  __Pyx_GIVEREF(__pyx_n_s_MMDDYYB10);
  PyList_SET_ITEM(__pyx_t_3, 20, __pyx_n_s_MMDDYYB10);
  __Pyx_INCREF(__pyx_n_s_MMDDYYC);
  __Pyx_GIVEREF(__pyx_n_s_MMDDYYC);
  PyList_SET_ITEM(__pyx_t_3, 21, __pyx_n_s_MMDDYYC);
  __Pyx_INCREF(__pyx_n_s_MMDDYYC10);
  __Pyx_GIVEREF(__pyx_n_s_MMDDYYC10);
  PyList_SET_ITEM(__pyx_t_3, 22, __pyx_n_s_MMDDYYC10);
  __Pyx_INCREF(__pyx_n_s_MMDDYYD);
  __Pyx_GIVEREF(__pyx_n_s_MMDDYYD);
  PyList_SET_ITEM(__pyx_t_3, 23, __pyx_n_s_MMDDYYD);
  __Pyx_INCREF(__pyx_n_s_MMDDYYD10);
  __Pyx_GIVEREF(__pyx_n_s_MMDDYYD10);
  PyList_SET_ITEM(__pyx_t_3, 24, __pyx_n_s_MMDDYYD10);
  __Pyx_INCREF(__pyx_n_s_MMDDYYN6);
  __Pyx_GIVEREF(__pyx_n_s_MMDDYYN6);
  PyList_SET_ITEM(__pyx_t_3, 25, __pyx_n_s_MMDDYYN6);
  __Pyx_INCREF(__pyx_n_s_MMDDYYN8);
  __Pyx_GIVEREF(__pyx_n_s_MMDDYYN8);
  PyList_SET_ITEM(__pyx_t_3, 26, __pyx_n_s_MMDDYYN8);
  __Pyx_INCREF(__pyx_n_s_MMDDYYP);
  __Pyx_GIVEREF(__pyx_n_s_MMDDYYP);
  PyList_SET_ITEM(__pyx_t_3, 27, __pyx_n_s_MMDDYYP);
  __Pyx_INCREF(__pyx_n_s_MMDDYYP10);
  __Pyx_GIVEREF(__pyx_n_s_MMDDYYP10);
  PyList_SET_ITEM(__pyx_t_3, 28, __pyx_n_s_MMDDYYP10);
  __Pyx_INCREF(__pyx_n_s_MMDDYYS);
  __Pyx_GIVEREF(__pyx_n_s_MMDDYYS);
  PyList_SET_ITEM(__pyx_t_3, 29, __pyx_n_s_MMDDYYS);
  __Pyx_INCREF(__pyx_n_s_MMDDYYS10);
  __Pyx_GIVEREF(__pyx_n_s_MMDDYYS10);
  PyList_SET_ITEM(__pyx_t_3, 30, __pyx_n_s_MMDDYYS10);
  __Pyx_INCREF(__pyx_n_s_WEEKDATX);
  __Pyx_GIVEREF(__pyx_n_s_WEEKDATX);
  PyList_SET_ITEM(__pyx_t_3, 31, __pyx_n_s_WEEKDATX);
  __Pyx_INCREF(__pyx_n_s_DTDATE);
  __Pyx_GIVEREF(__pyx_n_s_DTDATE);
  PyList_SET_ITEM(__pyx_t_3, 32, __pyx_n_s_DTDATE);
  __Pyx_INCREF(__pyx_n_s_IS8601DA);
  __Pyx_GIVEREF(__pyx_n_s_IS8601DA);
  PyList_SET_ITEM(__pyx_t_3, 33, __pyx_n_s_IS8601DA);
  __Pyx_INCREF(__pyx_n_s_E8601DA);
  __Pyx_GIVEREF(__pyx_n_s_E8601DA);
  PyList_SET_ITEM(__pyx_t_3, 34, __pyx_n_s_E8601DA);
  __Pyx_INCREF(__pyx_n_s_B8601DA);
  __Pyx_GIVEREF(__pyx_n_s_B8601DA);
  PyList_SET_ITEM(__pyx_t_3, 35, __pyx_n_s_B8601DA);
  __Pyx_INCREF(__pyx_n_s_YYMMDDB);
  __Pyx_GIVEREF(__pyx_n_s_YYMMDDB);
  PyList_SET_ITEM(__pyx_t_3, 36, __pyx_n_s_YYMMDDB);
  __Pyx_INCREF(__pyx_n_s_YYMMDDD);
  __Pyx_GIVEREF(__pyx_n_s_YYMMDDD);
  PyList_SET_ITEM(__pyx_t_3, 37, __pyx_n_s_YYMMDDD);
  __Pyx_INCREF(__pyx_n_s_YYMMDDN);
  __Pyx_GIVEREF(__pyx_n_s_YYMMDDN);
  PyList_SET_ITEM(__pyx_t_3, 38, __pyx_n_s_YYMMDDN);
  __Pyx_INCREF(__pyx_n_s_YYMMDDP);
  __Pyx_GIVEREF(__pyx_n_s_YYMMDDP);
  PyList_SET_ITEM(__pyx_t_3, 39, __pyx_n_s_YYMMDDP);
  __Pyx_INCREF(__pyx_n_s_YYMMDDS);
  __Pyx_GIVEREF(__pyx_n_s_YYMMDDS);
  PyList_SET_ITEM(__pyx_t_3, 40, __pyx_n_s_YYMMDDS);
  __Pyx_XGOTREF(__pyx_v_10pyreadstat_16_readstat_parser_sas_date_formats);
  __Pyx_DECREF_SET(__pyx_v_10pyreadstat_16_readstat_parser_sas_date_formats, ((PyObject*)__pyx_t_3));
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;

  /* "pyreadstat/_readstat_parser.pyx":49
 *                                 "IS8601DA", "E8601DA", "B8601DA",
 *                                 "YYMMDDB", "YYMMDDD", "YYMMDDN", "YYMMDDP", "YYMMDDS",]
 * cdef list sas_datetime_formats = ["DATETIME", "DATETIME18", "DATETIME19",  "DATETIME20", "DATETIME21", "DATETIME22",             # <<<<<<<<<<<<<<
 *                 "E8601DT", "DATEAMPM", "MDYAMPM", "IS8601DT", "B8601DT", "B8601DN"]
 * cdef list sas_time_formats = ["TIME", "HHMM", "TIME20.3", "TIME20", "TIME5", "TOD", "TIMEAMPM", "IS8601TM", "E8601TM", "B8601TM", ]
 */
  __pyx_t_3 = PyList_New(12); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 49, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_n_s_DATETIME);
  __Pyx_GIVEREF(__pyx_n_s_DATETIME);
  PyList_SET_ITEM(__pyx_t_3, 0, __pyx_n_s_DATETIME);
  __Pyx_INCREF(__pyx_n_s_DATETIME18);
  __Pyx_GIVEREF(__pyx_n_s_DATETIME18);
  PyList_SET_ITEM(__pyx_t_3, 1, __pyx_n_s_DATETIME18);
  __Pyx_INCREF(__pyx_n_s_DATETIME19);
  __Pyx_GIVEREF(__pyx_n_s_DATETIME19);
  PyList_SET_ITEM(__pyx_t_3, 2, __pyx_n_s_DATETIME19);
  __Pyx_INCREF(__pyx_n_s_DATETIME20);
  __Pyx_GIVEREF(__pyx_n_s_DATETIME20);
  PyList_SET_ITEM(__pyx_t_3, 3, __pyx_n_s_DATETIME20);
  __Pyx_INCREF(__pyx_n_s_DATETIME21);
  __Pyx_GIVEREF(__pyx_n_s_DATETIME21);
  PyList_SET_ITEM(__pyx_t_3, 4, __pyx_n_s_DATETIME21);
  __Pyx_INCREF(__pyx_n_s_DATETIME22);
  __Pyx_GIVEREF(__pyx_n_s_DATETIME22);
  PyList_SET_ITEM(__pyx_t_3, 5, __pyx_n_s_DATETIME22);
  __Pyx_INCREF(__pyx_n_s_E8601DT);
  __Pyx_GIVEREF(__pyx_n_s_E8601DT);
  PyList_SET_ITEM(__pyx_t_3, 6, __pyx_n_s_E8601DT);
  __Pyx_INCREF(__pyx_n_s_DATEAMPM);
  __Pyx_GIVEREF(__pyx_n_s_DATEAMPM);
  PyList_SET_ITEM(__pyx_t_3, 7, __pyx_n_s_DATEAMPM);
  __Pyx_INCREF(__pyx_n_s_MDYAMPM);
  __Pyx_GIVEREF(__pyx_n_s_MDYAMPM);
  PyList_SET_ITEM(__pyx_t_3, 8, __pyx_n_s_MDYAMPM);
  __Pyx_INCREF(__pyx_n_s_IS8601DT);
  __Pyx_GIVEREF(__pyx_n_s_IS8601DT);
  PyList_SET_ITEM(__pyx_t_3, 9, __pyx_n_s_IS8601DT);
  __Pyx_INCREF(__pyx_n_s_B8601DT);
  __Pyx_GIVEREF(__pyx_n_s_B8601DT);
  PyList_SET_ITEM(__pyx_t_3, 10, __pyx_n_s_B8601DT);
  __Pyx_INCREF(__pyx_n_s_B8601DN);
  __Pyx_GIVEREF(__pyx_n_s_B8601DN);
  PyList_SET_ITEM(__pyx_t_3, 11, __pyx_n_s_B8601DN);
  __Pyx_XGOTREF(__pyx_v_10pyreadstat_16_readstat_parser_sas_datetime_formats);
  __Pyx_DECREF_SET(__pyx_v_10pyreadstat_16_readstat_parser_sas_datetime_formats, ((PyObject*)__pyx_t_3));
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;

  /* "pyreadstat/_readstat_parser.pyx":51
 * cdef list sas_datetime_formats = ["DATETIME", "DATETIME18", "DATETIME19",  "DATETIME20", "DATETIME21", "DATETIME22",
 *                 "E8601DT", "DATEAMPM", "MDYAMPM", "IS8601DT", "B8601DT", "B8601DN"]
 * cdef list sas_time_formats = ["TIME", "HHMM", "TIME20.3", "TIME20", "TIME5", "TOD", "TIMEAMPM", "IS8601TM", "E8601TM", "B8601TM", ]             # <<<<<<<<<<<<<<
 * # "HOUR" # these do not print as full time formats in sas
 * #cdef list sas_all_formats = sas_date_formats + sas_datetime_formats + sas_time_formats
 */
  __pyx_t_3 = PyList_New(10); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 51, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_n_s_TIME);
  __Pyx_GIVEREF(__pyx_n_s_TIME);
  PyList_SET_ITEM(__pyx_t_3, 0, __pyx_n_s_TIME);
  __Pyx_INCREF(__pyx_n_s_HHMM);
  __Pyx_GIVEREF(__pyx_n_s_HHMM);
  PyList_SET_ITEM(__pyx_t_3, 1, __pyx_n_s_HHMM);
  __Pyx_INCREF(__pyx_kp_s_TIME20_3);
  __Pyx_GIVEREF(__pyx_kp_s_TIME20_3);
  PyList_SET_ITEM(__pyx_t_3, 2, __pyx_kp_s_TIME20_3);
  __Pyx_INCREF(__pyx_n_s_TIME20);
  __Pyx_GIVEREF(__pyx_n_s_TIME20);
  PyList_SET_ITEM(__pyx_t_3, 3, __pyx_n_s_TIME20);
  __Pyx_INCREF(__pyx_n_s_TIME5);
  __Pyx_GIVEREF(__pyx_n_s_TIME5);
  PyList_SET_ITEM(__pyx_t_3, 4, __pyx_n_s_TIME5);
  __Pyx_INCREF(__pyx_n_s_TOD);
  __Pyx_GIVEREF(__pyx_n_s_TOD);
  PyList_SET_ITEM(__pyx_t_3, 5, __pyx_n_s_TOD);
  __Pyx_INCREF(__pyx_n_s_TIMEAMPM);
  __Pyx_GIVEREF(__pyx_n_s_TIMEAMPM);
  PyList_SET_ITEM(__pyx_t_3, 6, __pyx_n_s_TIMEAMPM);
  __Pyx_INCREF(__pyx_n_s_IS8601TM);
  __Pyx_GIVEREF(__pyx_n_s_IS8601TM);
  PyList_SET_ITEM(__pyx_t_3, 7, __pyx_n_s_IS8601TM);
  __Pyx_INCREF(__pyx_n_s_E8601TM);
  __Pyx_GIVEREF(__pyx_n_s_E8601TM);
  PyList_SET_ITEM(__pyx_t_3, 8, __pyx_n_s_E8601TM);
  __Pyx_INCREF(__pyx_n_s_B8601TM);
  __Pyx_GIVEREF(__pyx_n_s_B8601TM);
  PyList_SET_ITEM(__pyx_t_3, 9, __pyx_n_s_B8601TM);
  __Pyx_XGOTREF(__pyx_v_10pyreadstat_16_readstat_parser_sas_time_formats);
  __Pyx_DECREF_SET(__pyx_v_10pyreadstat_16_readstat_parser_sas_time_formats, ((PyObject*)__pyx_t_3));
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;

  /* "pyreadstat/_readstat_parser.pyx":55
 * #cdef list sas_all_formats = sas_date_formats + sas_datetime_formats + sas_time_formats
 * cdef list sas_all_formats
 * cdef object sas_origin = datetime_new(1960, 1, 1, 0, 0, 0, 0, None)             # <<<<<<<<<<<<<<
 * 
 * cdef list spss_datetime_formats = ["DATETIME", "DATETIME8", 'DATETIME17', 'DATETIME20', 'DATETIME23.2',"YMDHMS16","YMDHMS19","YMDHMS19.2", "YMDHMS20"]
 */
  __pyx_t_3 = ((PyObject *)__pyx_f_7cpython_8datetime_datetime_new(0x7A8, 1, 1, 0, 0, 0, 0, Py_None, NULL)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 55, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_XGOTREF(__pyx_v_10pyreadstat_16_readstat_parser_sas_origin);
  __Pyx_DECREF_SET(__pyx_v_10pyreadstat_16_readstat_parser_sas_origin, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;

  /* "pyreadstat/_readstat_parser.pyx":57
 * cdef object sas_origin = datetime_new(1960, 1, 1, 0, 0, 0, 0, None)
 * 
 * cdef list spss_datetime_formats = ["DATETIME", "DATETIME8", 'DATETIME17', 'DATETIME20', 'DATETIME23.2',"YMDHMS16","YMDHMS19","YMDHMS19.2", "YMDHMS20"]             # <<<<<<<<<<<<<<
 * cdef list spss_date_formats = ["DATE",'DATE8','DATE11', 'DATE12', "ADATE","ADATE8", "ADATE10", "EDATE", 'EDATE8','EDATE10', "JDATE", "JDATE5", "JDATE7", "SDATE", "SDATE8", "SDATE10",]
 * cdef list spss_time_formats = ["TIME", "DTIME", 'TIME8', 'TIME5', 'TIME11.2']
 */
  __pyx_t_3 = PyList_New(9); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 57, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_n_s_DATETIME);
  __Pyx_GIVEREF(__pyx_n_s_DATETIME);
  PyList_SET_ITEM(__pyx_t_3, 0, __pyx_n_s_DATETIME);
  __Pyx_INCREF(__pyx_n_s_DATETIME8);
  __Pyx_GIVEREF(__pyx_n_s_DATETIME8);
  PyList_SET_ITEM(__pyx_t_3, 1, __pyx_n_s_DATETIME8);
  __Pyx_INCREF(__pyx_n_s_DATETIME17);
  __Pyx_GIVEREF(__pyx_n_s_DATETIME17);
  PyList_SET_ITEM(__pyx_t_3, 2, __pyx_n_s_DATETIME17);
  __Pyx_INCREF(__pyx_n_s_DATETIME20);
  __Pyx_GIVEREF(__pyx_n_s_DATETIME20);
  PyList_SET_ITEM(__pyx_t_3, 3, __pyx_n_s_DATETIME20);
  __Pyx_INCREF(__pyx_kp_s_DATETIME23_2);
  __Pyx_GIVEREF(__pyx_kp_s_DATETIME23_2);
  PyList_SET_ITEM(__pyx_t_3, 4, __pyx_kp_s_DATETIME23_2);
  __Pyx_INCREF(__pyx_n_s_YMDHMS16);
  __Pyx_GIVEREF(__pyx_n_s_YMDHMS16);
  PyList_SET_ITEM(__pyx_t_3, 5, __pyx_n_s_YMDHMS16);
  __Pyx_INCREF(__pyx_n_s_YMDHMS19);
  __Pyx_GIVEREF(__pyx_n_s_YMDHMS19);
  PyList_SET_ITEM(__pyx_t_3, 6, __pyx_n_s_YMDHMS19);
  __Pyx_INCREF(__pyx_kp_s_YMDHMS19_2);
  __Pyx_GIVEREF(__pyx_kp_s_YMDHMS19_2);
  PyList_SET_ITEM(__pyx_t_3, 7, __pyx_kp_s_YMDHMS19_2);
  __Pyx_INCREF(__pyx_n_s_YMDHMS20);
  __Pyx_GIVEREF(__pyx_n_s_YMDHMS20);
  PyList_SET_ITEM(__pyx_t_3, 8, __pyx_n_s_YMDHMS20);
  __Pyx_XGOTREF(__pyx_v_10pyreadstat_16_readstat_parser_spss_datetime_formats);
  __Pyx_DECREF_SET(__pyx_v_10pyreadstat_16_readstat_parser_spss_datetime_formats, ((PyObject*)__pyx_t_3));
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;

  /* "pyreadstat/_readstat_parser.pyx":58
 * 
 * cdef list spss_datetime_formats = ["DATETIME", "DATETIME8", 'DATETIME17', 'DATETIME20', 'DATETIME23.2',"YMDHMS16","YMDHMS19","YMDHMS19.2", "YMDHMS20"]
 * cdef list spss_date_formats = ["DATE",'DATE8','DATE11', 'DATE12', "ADATE","ADATE8", "ADATE10", "EDATE", 'EDATE8','EDATE10', "JDATE", "JDATE5", "JDATE7", "SDATE", "SDATE8", "SDATE10",]             # <<<<<<<<<<<<<<
 * cdef list spss_time_formats = ["TIME", "DTIME", 'TIME8', 'TIME5', 'TIME11.2']
 * #cdef list spss_all_formats = spss_date_formats + spss_datetime_formats + spss_time_formats
 */
  __pyx_t_3 = PyList_New(16); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 58, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_n_s_DATE);
  __Pyx_GIVEREF(__pyx_n_s_DATE);
  PyList_SET_ITEM(__pyx_t_3, 0, __pyx_n_s_DATE);
  __Pyx_INCREF(__pyx_n_s_DATE8);
  __Pyx_GIVEREF(__pyx_n_s_DATE8);
  PyList_SET_ITEM(__pyx_t_3, 1, __pyx_n_s_DATE8);
  __Pyx_INCREF(__pyx_n_s_DATE11);
  __Pyx_GIVEREF(__pyx_n_s_DATE11);
  PyList_SET_ITEM(__pyx_t_3, 2, __pyx_n_s_DATE11);
  __Pyx_INCREF(__pyx_n_s_DATE12);
  __Pyx_GIVEREF(__pyx_n_s_DATE12);
  PyList_SET_ITEM(__pyx_t_3, 3, __pyx_n_s_DATE12);
  __Pyx_INCREF(__pyx_n_s_ADATE);
  __Pyx_GIVEREF(__pyx_n_s_ADATE);
  PyList_SET_ITEM(__pyx_t_3, 4, __pyx_n_s_ADATE);
  __Pyx_INCREF(__pyx_n_s_ADATE8);
  __Pyx_GIVEREF(__pyx_n_s_ADATE8);
  PyList_SET_ITEM(__pyx_t_3, 5, __pyx_n_s_ADATE8);
  __Pyx_INCREF(__pyx_n_s_ADATE10);
  __Pyx_GIVEREF(__pyx_n_s_ADATE10);
  PyList_SET_ITEM(__pyx_t_3, 6, __pyx_n_s_ADATE10);
  __Pyx_INCREF(__pyx_n_s_EDATE);
  __Pyx_GIVEREF(__pyx_n_s_EDATE);
  PyList_SET_ITEM(__pyx_t_3, 7, __pyx_n_s_EDATE);
  __Pyx_INCREF(__pyx_n_s_EDATE8);
  __Pyx_GIVEREF(__pyx_n_s_EDATE8);
  PyList_SET_ITEM(__pyx_t_3, 8, __pyx_n_s_EDATE8);
  __Pyx_INCREF(__pyx_n_s_EDATE10);
  __Pyx_GIVEREF(__pyx_n_s_EDATE10);
  PyList_SET_ITEM(__pyx_t_3, 9, __pyx_n_s_EDATE10);
  __Pyx_INCREF(__pyx_n_s_JDATE);
  __Pyx_GIVEREF(__pyx_n_s_JDATE);
  PyList_SET_ITEM(__pyx_t_3, 10, __pyx_n_s_JDATE);
  __Pyx_INCREF(__pyx_n_s_JDATE5);
  __Pyx_GIVEREF(__pyx_n_s_JDATE5);
  PyList_SET_ITEM(__pyx_t_3, 11, __pyx_n_s_JDATE5);
  __Pyx_INCREF(__pyx_n_s_JDATE7);
  __Pyx_GIVEREF(__pyx_n_s_JDATE7);
  PyList_SET_ITEM(__pyx_t_3, 12, __pyx_n_s_JDATE7);
  __Pyx_INCREF(__pyx_n_s_SDATE);
  __Pyx_GIVEREF(__pyx_n_s_SDATE);
  PyList_SET_ITEM(__pyx_t_3, 13, __pyx_n_s_SDATE);
  __Pyx_INCREF(__pyx_n_s_SDATE8);
  __Pyx_GIVEREF(__pyx_n_s_SDATE8);
  PyList_SET_ITEM(__pyx_t_3, 14, __pyx_n_s_SDATE8);
  __Pyx_INCREF(__pyx_n_s_SDATE10);
  __Pyx_GIVEREF(__pyx_n_s_SDATE10);
  PyList_SET_ITEM(__pyx_t_3, 15, __pyx_n_s_SDATE10);
  __Pyx_XGOTREF(__pyx_v_10pyreadstat_16_readstat_parser_spss_date_formats);
  __Pyx_DECREF_SET(__pyx_v_10pyreadstat_16_readstat_parser_spss_date_formats, ((PyObject*)__pyx_t_3));
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;

  /* "pyreadstat/_readstat_parser.pyx":59
 * cdef list spss_datetime_formats = ["DATETIME", "DATETIME8", 'DATETIME17', 'DATETIME20', 'DATETIME23.2',"YMDHMS16","YMDHMS19","YMDHMS19.2", "YMDHMS20"]
 * cdef list spss_date_formats = ["DATE",'DATE8','DATE11', 'DATE12', "ADATE","ADATE8", "ADATE10", "EDATE", 'EDATE8','EDATE10', "JDATE", "JDATE5", "JDATE7", "SDATE", "SDATE8", "SDATE10",]
 * cdef list spss_time_formats = ["TIME", "DTIME", 'TIME8', 'TIME5', 'TIME11.2']             # <<<<<<<<<<<<<<
 * #cdef list spss_all_formats = spss_date_formats + spss_datetime_formats + spss_time_formats
 * cdef list spss_all_formats
 */
  __pyx_t_3 = PyList_New(5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 59, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_n_s_TIME);
  __Pyx_GIVEREF(__pyx_n_s_TIME);
  PyList_SET_ITEM(__pyx_t_3, 0, __pyx_n_s_TIME);
  __Pyx_INCREF(__pyx_n_s_DTIME);
  __Pyx_GIVEREF(__pyx_n_s_DTIME);
  PyList_SET_ITEM(__pyx_t_3, 1, __pyx_n_s_DTIME);
  __Pyx_INCREF(__pyx_n_s_TIME8);
  __Pyx_GIVEREF(__pyx_n_s_TIME8);
  PyList_SET_ITEM(__pyx_t_3, 2, __pyx_n_s_TIME8);
  __Pyx_INCREF(__pyx_n_s_TIME5);
  __Pyx_GIVEREF(__pyx_n_s_TIME5);
  PyList_SET_ITEM(__pyx_t_3, 3, __pyx_n_s_TIME5);
  __Pyx_INCREF(__pyx_kp_s_TIME11_2);
  __Pyx_GIVEREF(__pyx_kp_s_TIME11_2);
  PyList_SET_ITEM(__pyx_t_3, 4, __pyx_kp_s_TIME11_2);
  __Pyx_XGOTREF(__pyx_v_10pyreadstat_16_readstat_parser_spss_time_formats);
  __Pyx_DECREF_SET(__pyx_v_10pyreadstat_16_readstat_parser_spss_time_formats, ((PyObject*)__pyx_t_3));
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;

  /* "pyreadstat/_readstat_parser.pyx":62
 * #cdef list spss_all_formats = spss_date_formats + spss_datetime_formats + spss_time_formats
 * cdef list spss_all_formats
 * cdef object spss_origin = datetime_new(1582, 10, 14, 0, 0, 0, 0, None)             # <<<<<<<<<<<<<<
 * 
 * cdef list stata_datetime_formats = ["%tC", "%tc"]
 */
  __pyx_t_3 = ((PyObject *)__pyx_f_7cpython_8datetime_datetime_new(0x62E, 10, 14, 0, 0, 0, 0, Py_None, NULL)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 62, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_XGOTREF(__pyx_v_10pyreadstat_16_readstat_parser_spss_origin);
  __Pyx_DECREF_SET(__pyx_v_10pyreadstat_16_readstat_parser_spss_origin, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;

  /* "pyreadstat/_readstat_parser.pyx":64
 * cdef object spss_origin = datetime_new(1582, 10, 14, 0, 0, 0, 0, None)
 * 
 * cdef list stata_datetime_formats = ["%tC", "%tc"]             # <<<<<<<<<<<<<<
 * cdef list stata_date_formats = ["%td", "%d", "%tdD_m_Y", "%tdCCYY-NN-DD"]
 * cdef list stata_time_formats = ["%tcHH:MM:SS", "%tcHH:MM"]
 */
  __pyx_t_3 = PyList_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 64, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_kp_s_tC);
  __Pyx_GIVEREF(__pyx_kp_s_tC);
  PyList_SET_ITEM(__pyx_t_3, 0, __pyx_kp_s_tC);
  __Pyx_INCREF(__pyx_kp_s_tc);
  __Pyx_GIVEREF(__pyx_kp_s_tc);
  PyList_SET_ITEM(__pyx_t_3, 1, __pyx_kp_s_tc);
  __Pyx_XGOTREF(__pyx_v_10pyreadstat_16_readstat_parser_stata_datetime_formats);
  __Pyx_DECREF_SET(__pyx_v_10pyreadstat_16_readstat_parser_stata_datetime_formats, ((PyObject*)__pyx_t_3));
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;

  /* "pyreadstat/_readstat_parser.pyx":65
 * 
 * cdef list stata_datetime_formats = ["%tC", "%tc"]
 * cdef list stata_date_formats = ["%td", "%d", "%tdD_m_Y", "%tdCCYY-NN-DD"]             # <<<<<<<<<<<<<<
 * cdef list stata_time_formats = ["%tcHH:MM:SS", "%tcHH:MM"]
 * #cdef list stata_all_formats = stata_datetime_formats + stata_date_formats + stata_time_formats
 */
  __pyx_t_3 = PyList_New(4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 65, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_kp_s_td);
  __Pyx_GIVEREF(__pyx_kp_s_td);
  PyList_SET_ITEM(__pyx_t_3, 0, __pyx_kp_s_td);
  __Pyx_INCREF(__pyx_kp_s_d);
  __Pyx_GIVEREF(__pyx_kp_s_d);
  PyList_SET_ITEM(__pyx_t_3, 1, __pyx_kp_s_d);
  __Pyx_INCREF(__pyx_kp_s_tdD_m_Y);
  __Pyx_GIVEREF(__pyx_kp_s_tdD_m_Y);
  PyList_SET_ITEM(__pyx_t_3, 2, __pyx_kp_s_tdD_m_Y);
  __Pyx_INCREF(__pyx_kp_s_tdCCYY_NN_DD);
  __Pyx_GIVEREF(__pyx_kp_s_tdCCYY_NN_DD);
  PyList_SET_ITEM(__pyx_t_3, 3, __pyx_kp_s_tdCCYY_NN_DD);
  __Pyx_XGOTREF(__pyx_v_10pyreadstat_16_readstat_parser_stata_date_formats);
  __Pyx_DECREF_SET(__pyx_v_10pyreadstat_16_readstat_parser_stata_date_formats, ((PyObject*)__pyx_t_3));
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;

  /* "pyreadstat/_readstat_parser.pyx":66
 * cdef list stata_datetime_formats = ["%tC", "%tc"]
 * cdef list stata_date_formats = ["%td", "%d", "%tdD_m_Y", "%tdCCYY-NN-DD"]
 * cdef list stata_time_formats = ["%tcHH:MM:SS", "%tcHH:MM"]             # <<<<<<<<<<<<<<
 * #cdef list stata_all_formats = stata_datetime_formats + stata_date_formats + stata_time_formats
 * cdef list stata_all_formats
 */
  __pyx_t_3 = PyList_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 66, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_kp_s_tcHH_MM_SS);
  __Pyx_GIVEREF(__pyx_kp_s_tcHH_MM_SS);
  PyList_SET_ITEM(__pyx_t_3, 0, __pyx_kp_s_tcHH_MM_SS);
  __Pyx_INCREF(__pyx_kp_s_tcHH_MM);
  __Pyx_GIVEREF(__pyx_kp_s_tcHH_MM);
  PyList_SET_ITEM(__pyx_t_3, 1, __pyx_kp_s_tcHH_MM);
  __Pyx_XGOTREF(__pyx_v_10pyreadstat_16_readstat_parser_stata_time_formats);
  __Pyx_DECREF_SET(__pyx_v_10pyreadstat_16_readstat_parser_stata_time_formats, ((PyObject*)__pyx_t_3));
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;

  /* "pyreadstat/_readstat_parser.pyx":69
 * #cdef list stata_all_formats = stata_datetime_formats + stata_date_formats + stata_time_formats
 * cdef list stata_all_formats
 * cdef object stata_origin = datetime_new(1960, 1, 1, 0, 0, 0, 0, None)             # <<<<<<<<<<<<<<
 * 
 * cdef dict readstat_to_numpy_types = {READSTAT_TYPE_STRING: object, READSTAT_TYPE_STRING_REF: object,
 */
  __pyx_t_3 = ((PyObject *)__pyx_f_7cpython_8datetime_datetime_new(0x7A8, 1, 1, 0, 0, 0, 0, Py_None, NULL)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 69, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_XGOTREF(__pyx_v_10pyreadstat_16_readstat_parser_stata_origin);
  __Pyx_DECREF_SET(__pyx_v_10pyreadstat_16_readstat_parser_stata_origin, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;

  /* "pyreadstat/_readstat_parser.pyx":71
 * cdef object stata_origin = datetime_new(1960, 1, 1, 0, 0, 0, 0, None)
 * 
 * cdef dict readstat_to_numpy_types = {READSTAT_TYPE_STRING: object, READSTAT_TYPE_STRING_REF: object,             # <<<<<<<<<<<<<<
 *                                      READSTAT_TYPE_INT8: np.int64, READSTAT_TYPE_INT16: np.int64, READSTAT_TYPE_INT32:np.int64,
 *                                      READSTAT_TYPE_FLOAT: np.float64, READSTAT_TYPE_DOUBLE: np.float64}
 */
  __pyx_t_3 = __Pyx_PyDict_NewPresized(7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 71, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyInt_From_readstat_type_t(READSTAT_TYPE_STRING); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 71, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_3, __pyx_t_2, __pyx_builtin_object) < 0) __PYX_ERR(0, 71, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyInt_From_readstat_type_t(READSTAT_TYPE_STRING_REF); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 71, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_3, __pyx_t_2, __pyx_builtin_object) < 0) __PYX_ERR(0, 71, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyreadstat/_readstat_parser.pyx":72
 * 
 * cdef dict readstat_to_numpy_types = {READSTAT_TYPE_STRING: object, READSTAT_TYPE_STRING_REF: object,
 *                                      READSTAT_TYPE_INT8: np.int64, READSTAT_TYPE_INT16: np.int64, READSTAT_TYPE_INT32:np.int64,             # <<<<<<<<<<<<<<
 *                                      READSTAT_TYPE_FLOAT: np.float64, READSTAT_TYPE_DOUBLE: np.float64}
 * 
 */
  __pyx_t_2 = __Pyx_PyInt_From_readstat_type_t(READSTAT_TYPE_INT8); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 72, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 72, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_int64); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 72, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_t_3, __pyx_t_2, __pyx_t_5) < 0) __PYX_ERR(0, 71, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyInt_From_readstat_type_t(READSTAT_TYPE_INT16); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 72, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 72, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_int64); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 72, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_t_3, __pyx_t_5, __pyx_t_4) < 0) __PYX_ERR(0, 71, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyInt_From_readstat_type_t(READSTAT_TYPE_INT32); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 72, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_np); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 72, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_int64); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 72, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (PyDict_SetItem(__pyx_t_3, __pyx_t_4, __pyx_t_2) < 0) __PYX_ERR(0, 71, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyreadstat/_readstat_parser.pyx":73
 * cdef dict readstat_to_numpy_types = {READSTAT_TYPE_STRING: object, READSTAT_TYPE_STRING_REF: object,
 *                                      READSTAT_TYPE_INT8: np.int64, READSTAT_TYPE_INT16: np.int64, READSTAT_TYPE_INT32:np.int64,
 *                                      READSTAT_TYPE_FLOAT: np.float64, READSTAT_TYPE_DOUBLE: np.float64}             # <<<<<<<<<<<<<<
 * 
 * cdef class data_container:
 */
  __pyx_t_2 = __Pyx_PyInt_From_readstat_type_t(READSTAT_TYPE_FLOAT); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 73, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 73, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_float64); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 73, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_t_3, __pyx_t_2, __pyx_t_5) < 0) __PYX_ERR(0, 71, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyInt_From_readstat_type_t(READSTAT_TYPE_DOUBLE); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 73, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 73, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_float64); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 73, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_t_3, __pyx_t_5, __pyx_t_4) < 0) __PYX_ERR(0, 71, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_XGOTREF(__pyx_v_10pyreadstat_16_readstat_parser_readstat_to_numpy_types);
  __Pyx_DECREF_SET(__pyx_v_10pyreadstat_16_readstat_parser_readstat_to_numpy_types, ((PyObject*)__pyx_t_3));
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_10pyreadstat_16_readstat_parser_14data_container_3__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_data_container___reduce_cython, NULL, __pyx_n_s_pyreadstat__readstat_parser, __pyx_d, ((PyObject *)__pyx_codeobj__11)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_3) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_10pyreadstat_16_readstat_parser_14data_container_5__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_data_container___setstate_cython, NULL, __pyx_n_s_pyreadstat__readstat_parser, __pyx_d, ((PyObject *)__pyx_codeobj__13)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_3) < 0) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "pyreadstat/_readstat_parser.pyx":119
 *         self.mtime = 0
 * 
 * class metadata_container:             # <<<<<<<<<<<<<<
 *     """
 *     This class holds metadata we want to give back to python
 */
  __pyx_t_3 = __Pyx_Py3MetaclassPrepare((PyObject *) NULL, __pyx_empty_tuple, __pyx_n_s_metadata_container, __pyx_n_s_metadata_container, (PyObject *) NULL, __pyx_n_s_pyreadstat__readstat_parser, __pyx_kp_s_This_class_holds_metadata_we_wa); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 119, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);

  /* "pyreadstat/_readstat_parser.pyx":123
 *     This class holds metadata we want to give back to python
 *     """
 *     def __init__(self):             # <<<<<<<<<<<<<<
 *         self.column_names = list()
 *         self.column_labels = list()
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_10pyreadstat_16_readstat_parser_18metadata_container_1__init__, 0, __pyx_n_s_metadata_container___init, NULL, __pyx_n_s_pyreadstat__readstat_parser, __pyx_d, ((PyObject *)__pyx_codeobj__14)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 123, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetNameInClass(__pyx_t_3, __pyx_n_s_init, __pyx_t_4) < 0) __PYX_ERR(0, 123, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "pyreadstat/_readstat_parser.pyx":119
 *         self.mtime = 0
 * 
 * class metadata_container:             # <<<<<<<<<<<<<<
 *     """
 *     This class holds metadata we want to give back to python
 */
  __pyx_t_4 = __Pyx_Py3ClassCreate(((PyObject*)&__Pyx_DefaultClassType), __pyx_n_s_metadata_container, __pyx_empty_tuple, __pyx_t_3, NULL, 0, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 119, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_metadata_container, __pyx_t_4) < 0) __PYX_ERR(0, 119, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "pyreadstat/_readstat_parser.pyx":147
 * 
 * 
 * class ReadstatError(Exception):             # <<<<<<<<<<<<<<
 *     """
 *     Just defining a custom exception to raise when readstat gives an error return code.
 */
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 147, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0]));
  __Pyx_GIVEREF((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0]));
  PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
  __pyx_t_4 = __Pyx_PEP560_update_bases(__pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 147, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_CalculateMetaclass(NULL, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 147, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_2 = __Pyx_Py3MetaclassPrepare(__pyx_t_5, __pyx_t_4, __pyx_n_s_ReadstatError, __pyx_n_s_ReadstatError, (PyObject *) NULL, __pyx_n_s_pyreadstat__readstat_parser, __pyx_kp_s_Just_defining_a_custom_exceptio); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 147, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__pyx_t_4 != __pyx_t_3) {
    if (unlikely((PyDict_SetItemString(__pyx_t_2, "__orig_bases__", __pyx_t_3) < 0))) __PYX_ERR(0, 147, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_Py3ClassCreate(__pyx_t_5, __pyx_n_s_ReadstatError, __pyx_t_4, __pyx_t_2, NULL, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 147, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ReadstatError, __pyx_t_3) < 0) __PYX_ERR(0, 147, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "pyreadstat/_readstat_parser.pyx":153
 *     pass
 * 
 * class PyreadstatError(Exception):             # <<<<<<<<<<<<<<
 *     """
 *     Just defining a custom exception to raise when pyreadstat raises an exception.
 */
  __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 153, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0]));
  __Pyx_GIVEREF((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0]));
  PyTuple_SET_ITEM(__pyx_t_4, 0, ((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
  __pyx_t_5 = __Pyx_PEP560_update_bases(__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 153, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_2 = __Pyx_CalculateMetaclass(NULL, __pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 153, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_Py3MetaclassPrepare(__pyx_t_2, __pyx_t_5, __pyx_n_s_PyreadstatError, __pyx_n_s_PyreadstatError, (PyObject *) NULL, __pyx_n_s_pyreadstat__readstat_parser, __pyx_kp_s_Just_defining_a_custom_exceptio_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 153, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__pyx_t_5 != __pyx_t_4) {
    if (unlikely((PyDict_SetItemString(__pyx_t_3, "__orig_bases__", __pyx_t_4) < 0))) __PYX_ERR(0, 153, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_Py3ClassCreate(__pyx_t_2, __pyx_n_s_PyreadstatError, __pyx_t_5, __pyx_t_3, NULL, 0, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 153, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_PyreadstatError, __pyx_t_4) < 0) __PYX_ERR(0, 153, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "pyreadstat/_readstat_parser.pyx":1
 * # cython: c_string_type=unicode, c_string_encoding=utf8, language_level=2             # <<<<<<<<<<<<<<
 * 
 * # #############################################################################
 */
  __pyx_t_5 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_5) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  if (__pyx_m) {
    if (__pyx_d && stringtab_initialized) {
      __Pyx_AddTraceback("init pyreadstat._readstat_parser", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    #if !CYTHON_USE_MODULE_STATE
    Py_CLEAR(__pyx_m);
    #else
    Py_DECREF(__pyx_m);
    if (pystate_addmodule_run) {
      PyObject *tp, *value, *tb;
      PyErr_Fetch(&tp, &value, &tb);
      PyState_RemoveModule(&__pyx_moduledef);
      PyErr_Restore(tp, value, tb);
    }
    #endif
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init pyreadstat._readstat_parser");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  return (__pyx_m != NULL) ? 0 : -1;
  #elif PY_MAJOR_VERSION >= 3
  return __pyx_m;
  #else
  return;
  #endif
}
/* #### Code section: cleanup_globals ### */
/* #### Code section: cleanup_module ### */
/* #### Code section: main_method ### */
/* #### Code section: utility_code_pragmas ### */
#ifdef _MSC_VER
#pragma warning( push )
/* Warning 4127: conditional expression is constant
 * Cython uses constant conditional expressions to allow in inline functions to be optimized at
 * compile-time, so this warning is not useful
 */
#pragma warning( disable : 4127 )
#endif



/* #### Code section: utility_code_def ### */

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule(modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, "RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* PyErrExceptionMatches */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx_PyErr_ExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        if (__Pyx_PyErr_GivenExceptionMatches(exc_type, PyTuple_GET_ITEM(tuple, i))) return 1;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    int result;
    PyObject *exc_type;
#if PY_VERSION_HEX >= 0x030C00A6
    PyObject *current_exception = tstate->current_exception;
    if (unlikely(!current_exception)) return 0;
    exc_type = (PyObject*) Py_TYPE(current_exception);
    if (exc_type == err) return 1;
#else
    exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
#endif
    #if CYTHON_AVOID_BORROWED_REFS
    Py_INCREF(exc_type);
    #endif
    if (unlikely(PyTuple_Check(err))) {
        result = __Pyx_PyErr_ExceptionMatchesTuple(exc_type, err);
    } else {
        result = __Pyx_PyErr_GivenExceptionMatches(exc_type, err);
    }
    #if CYTHON_AVOID_BORROWED_REFS
    Py_DECREF(exc_type);
    #endif
    return result;
}
#endif

/* PyErrFetchRestore */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
#if PY_VERSION_HEX >= 0x030C00A6
    PyObject *tmp_value;
    assert(type == NULL || (value != NULL && type == (PyObject*) Py_TYPE(value)));
    if (value) {
        #if CYTHON_COMPILING_IN_CPYTHON
        if (unlikely(((PyBaseExceptionObject*) value)->traceback != tb))
        #endif
            PyException_SetTraceback(value, tb);
    }
    tmp_value = tstate->current_exception;
    tstate->current_exception = value;
    Py_XDECREF(tmp_value);
#else
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#endif
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
#if PY_VERSION_HEX >= 0x030C00A6
    PyObject* exc_value;
    exc_value = tstate->current_exception;
    tstate->current_exception = 0;
    *value = exc_value;
    *type = NULL;
    *tb = NULL;
    if (exc_value) {
        *type = (PyObject*) Py_TYPE(exc_value);
        Py_INCREF(*type);
        #if CYTHON_COMPILING_IN_CPYTHON
        *tb = ((PyBaseExceptionObject*) exc_value)->traceback;
        Py_XINCREF(*tb);
        #else
        *tb = PyException_GetTraceback(exc_value);
        #endif
    }
#else
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#endif
}
#endif

/* PyObjectGetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* PyObjectGetAttrStrNoError */
static void __Pyx_PyObject_GetAttrStr_ClearAttributeError(void) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (likely(__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        __Pyx_PyErr_Clear();
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name) {
    PyObject *result;
#if CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_TYPE_SLOTS && PY_VERSION_HEX >= 0x030700B1
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro == PyObject_GenericGetAttr)) {
        return _PyObject_GenericGetAttrWithDict(obj, attr_name, NULL, 1);
    }
#endif
    result = __Pyx_PyObject_GetAttrStr(obj, attr_name);
    if (unlikely(!result)) {
        __Pyx_PyObject_GetAttrStr_ClearAttributeError();
    }
    return result;
}

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStrNoError(__pyx_b, name);
    if (unlikely(!result) && !PyErr_Occurred()) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

/* ExtTypeTest */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type) {
    __Pyx_TypeName obj_type_name;
    __Pyx_TypeName type_name;
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    if (likely(__Pyx_TypeCheck(obj, type)))
        return 1;
    obj_type_name = __Pyx_PyType_GetName(Py_TYPE(obj));
    type_name = __Pyx_PyType_GetName(type);
    PyErr_Format(PyExc_TypeError,
                 "Cannot convert " __Pyx_FMT_TYPENAME " to " __Pyx_FMT_TYPENAME,
                 obj_type_name, type_name);
    __Pyx_DECREF_TypeName(obj_type_name);
    __Pyx_DECREF_TypeName(type_name);
    return 0;
}

/* PyObjectCall */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = Py_TYPE(func)->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* RaiseException */
#if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    __Pyx_PyThreadState_declare
    CYTHON_UNUSED_VAR(cause);
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    if (PyType_Check(type)) {
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
    }
    __Pyx_PyThreadState_assign
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
    if (cause) {
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
      #if PY_VERSION_HEX >= 0x030C00A6
        PyException_SetTraceback(value, tb);
      #elif CYTHON_FAST_THREAD_STATE
        PyThreadState *tstate = __Pyx_PyThreadState_Current;
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#else
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

/* TupleAndListFromArray */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE void __Pyx_copy_object_array(PyObject *const *CYTHON_RESTRICT src, PyObject** CYTHON_RESTRICT dest, Py_ssize_t length) {
    PyObject *v;
    Py_ssize_t i;
    for (i = 0; i < length; i++) {
        v = dest[i] = src[i];
        Py_INCREF(v);
    }
}
static CYTHON_INLINE PyObject *
__Pyx_PyTuple_FromArray(PyObject *const *src, Py_ssize_t n)
{
    PyObject *res;
    if (n <= 0) {
        Py_INCREF(__pyx_empty_tuple);
        return __pyx_empty_tuple;
    }
    res = PyTuple_New(n);
    if (unlikely(res == NULL)) return NULL;
    __Pyx_copy_object_array(src, ((PyTupleObject*)res)->ob_item, n);
    return res;
}
static CYTHON_INLINE PyObject *
__Pyx_PyList_FromArray(PyObject *const *src, Py_ssize_t n)
{
    PyObject *res;
    if (n <= 0) {
        return PyList_New(0);
    }
    res = PyList_New(n);
    if (unlikely(res == NULL)) return NULL;
    __Pyx_copy_object_array(src, ((PyListObject*)res)->ob_item, n);
    return res;
}
#endif

/* BytesEquals */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API
    return PyObject_RichCompareBool(s1, s2, equals);
#else
    if (s1 == s2) {
        return (equals == Py_EQ);
    } else if (PyBytes_CheckExact(s1) & PyBytes_CheckExact(s2)) {
        const char *ps1, *ps2;
        Py_ssize_t length = PyBytes_GET_SIZE(s1);
        if (length != PyBytes_GET_SIZE(s2))
            return (equals == Py_NE);
        ps1 = PyBytes_AS_STRING(s1);
        ps2 = PyBytes_AS_STRING(s2);
        if (ps1[0] != ps2[0]) {
            return (equals == Py_NE);
        } else if (length == 1) {
            return (equals == Py_EQ);
        } else {
            int result;
#if CYTHON_USE_UNICODE_INTERNALS && (PY_VERSION_HEX < 0x030B0000)
            Py_hash_t hash1, hash2;
            hash1 = ((PyBytesObject*)s1)->ob_shash;
            hash2 = ((PyBytesObject*)s2)->ob_shash;
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                return (equals == Py_NE);
            }
#endif
            result = memcmp(ps1, ps2, (size_t)length);
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & PyBytes_CheckExact(s2)) {
        return (equals == Py_NE);
    } else if ((s2 == Py_None) & PyBytes_CheckExact(s1)) {
        return (equals == Py_NE);
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
#endif
}

/* UnicodeEquals */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API
    return PyObject_RichCompareBool(s1, s2, equals);
#else
#if PY_MAJOR_VERSION < 3
    PyObject* owned_ref = NULL;
#endif
    int s1_is_unicode, s2_is_unicode;
    if (s1 == s2) {
        goto return_eq;
    }
    s1_is_unicode = PyUnicode_CheckExact(s1);
    s2_is_unicode = PyUnicode_CheckExact(s2);
#if PY_MAJOR_VERSION < 3
    if ((s1_is_unicode & (!s2_is_unicode)) && PyString_CheckExact(s2)) {
        owned_ref = PyUnicode_FromObject(s2);
        if (unlikely(!owned_ref))
            return -1;
        s2 = owned_ref;
        s2_is_unicode = 1;
    } else if ((s2_is_unicode & (!s1_is_unicode)) && PyString_CheckExact(s1)) {
        owned_ref = PyUnicode_FromObject(s1);
        if (unlikely(!owned_ref))
            return -1;
        s1 = owned_ref;
        s1_is_unicode = 1;
    } else if (((!s2_is_unicode) & (!s1_is_unicode))) {
        return __Pyx_PyBytes_Equals(s1, s2, equals);
    }
#endif
    if (s1_is_unicode & s2_is_unicode) {
        Py_ssize_t length;
        int kind;
        void *data1, *data2;
        if (unlikely(__Pyx_PyUnicode_READY(s1) < 0) || unlikely(__Pyx_PyUnicode_READY(s2) < 0))
            return -1;
        length = __Pyx_PyUnicode_GET_LENGTH(s1);
        if (length != __Pyx_PyUnicode_GET_LENGTH(s2)) {
            goto return_ne;
        }
#if CYTHON_USE_UNICODE_INTERNALS
        {
            Py_hash_t hash1, hash2;
        #if CYTHON_PEP393_ENABLED
            hash1 = ((PyASCIIObject*)s1)->hash;
            hash2 = ((PyASCIIObject*)s2)->hash;
        #else
            hash1 = ((PyUnicodeObject*)s1)->hash;
            hash2 = ((PyUnicodeObject*)s2)->hash;
        #endif
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                goto return_ne;
            }
        }
#endif
        kind = __Pyx_PyUnicode_KIND(s1);
        if (kind != __Pyx_PyUnicode_KIND(s2)) {
            goto return_ne;
        }
        data1 = __Pyx_PyUnicode_DATA(s1);
        data2 = __Pyx_PyUnicode_DATA(s2);
        if (__Pyx_PyUnicode_READ(kind, data1, 0) != __Pyx_PyUnicode_READ(kind, data2, 0)) {
            goto return_ne;
        } else if (length == 1) {
            goto return_eq;
        } else {
            int result = memcmp(data1, data2, (size_t)(length * kind));
            #if PY_MAJOR_VERSION < 3
            Py_XDECREF(owned_ref);
            #endif
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & s2_is_unicode) {
        goto return_ne;
    } else if ((s2 == Py_None) & s1_is_unicode) {
        goto return_ne;
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        #if PY_MAJOR_VERSION < 3
        Py_XDECREF(owned_ref);
        #endif
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
return_eq:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_EQ);
return_ne:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_NE);
#endif
}

/* fastcall */
#if CYTHON_METH_FASTCALL
static CYTHON_INLINE PyObject * __Pyx_GetKwValue_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues, PyObject *s)
{
    Py_ssize_t i, n = PyTuple_GET_SIZE(kwnames);
    for (i = 0; i < n; i++)
    {
        if (s == PyTuple_GET_ITEM(kwnames, i)) return kwvalues[i];
    }
    for (i = 0; i < n; i++)
    {
        int eq = __Pyx_PyUnicode_Equals(s, PyTuple_GET_ITEM(kwnames, i), Py_EQ);
        if (unlikely(eq != 0)) {
            if (unlikely(eq < 0)) return NULL;  // error
            return kwvalues[i];
        }
    }
    return NULL;  // not found (no exception set)
}
#endif

/* RaiseArgTupleInvalid */
static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* KeywordStringCheck */
static int __Pyx_CheckKeywordStrings(
    PyObject *kw,
    const char* function_name,
    int kw_allowed)
{
    PyObject* key = 0;
    Py_ssize_t pos = 0;
#if CYTHON_COMPILING_IN_PYPY
    if (!kw_allowed && PyDict_Next(kw, &pos, &key, 0))
        goto invalid_keyword;
    return 1;
#else
    if (CYTHON_METH_FASTCALL && likely(PyTuple_Check(kw))) {
        if (unlikely(PyTuple_GET_SIZE(kw) == 0))
            return 1;
        if (!kw_allowed) {
            key = PyTuple_GET_ITEM(kw, 0);
            goto invalid_keyword;
        }
#if PY_VERSION_HEX < 0x03090000
        for (pos = 0; pos < PyTuple_GET_SIZE(kw); pos++) {
            key = PyTuple_GET_ITEM(kw, pos);
            if (unlikely(!PyUnicode_Check(key)))
                goto invalid_keyword_type;
        }
#endif
        return 1;
    }
    while (PyDict_Next(kw, &pos, &key, 0)) {
        #if PY_MAJOR_VERSION < 3
        if (unlikely(!PyString_Check(key)))
        #endif
            if (unlikely(!PyUnicode_Check(key)))
                goto invalid_keyword_type;
    }
    if (!kw_allowed && unlikely(key))
        goto invalid_keyword;
    return 1;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    return 0;
#endif
invalid_keyword:
    #if PY_MAJOR_VERSION < 3
    PyErr_Format(PyExc_TypeError,
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
    PyErr_Format(PyExc_TypeError,
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
    return 0;
}

/* RaiseDoubleKeywords */
static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

/* ParseKeywords */
static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject *const *kwvalues,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    int kwds_is_tuple = CYTHON_METH_FASTCALL && likely(PyTuple_Check(kwds));
    while (1) {
        if (kwds_is_tuple) {
            if (pos >= PyTuple_GET_SIZE(kwds)) break;
            key = PyTuple_GET_ITEM(kwds, pos);
            value = kwvalues[pos];
            pos++;
        }
        else
        {
            if (!PyDict_Next(kwds, &pos, &key, &value)) break;
        }
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
            continue;
        }
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (__Pyx_PyUnicode_GET_LENGTH(**name) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key)
                );
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (__Pyx_PyUnicode_GET_LENGTH(**argname) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    #if PY_MAJOR_VERSION < 3
    PyErr_Format(PyExc_TypeError,
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
    PyErr_Format(PyExc_TypeError,
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    return -1;
}

/* PyObjectSetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_setattro))
        return tp->tp_setattro(obj, attr_name, value);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_setattr))
        return tp->tp_setattr(obj, PyString_AS_STRING(attr_name), value);
#endif
    return PyObject_SetAttr(obj, attr_name, value);
}
#endif

/* ModFloat[double] */
static CYTHON_INLINE double __Pyx_mod_double(double a, double b) {
    double r = fmod(a, b);
    r += ((r != 0) & ((r < 0) ^ (b < 0))) * b;
    return r;
}

/* PyFunctionFastCall */
#if CYTHON_FAST_PYCALL && !CYTHON_VECTORCALL
static PyObject* __Pyx_PyFunction_FastCallNoKw(PyCodeObject *co, PyObject **args, Py_ssize_t na,
                                               PyObject *globals) {
    PyFrameObject *f;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject **fastlocals;
    Py_ssize_t i;
    PyObject *result;
    assert(globals != NULL);
    /* XXX Perhaps we should create a specialized
       PyFrame_New() that doesn't take locals, but does
       take builtins without sanity checking them.
       */
    assert(tstate != NULL);
    f = PyFrame_New(tstate, co, globals, NULL);
    if (f == NULL) {
        return NULL;
    }
    fastlocals = __Pyx_PyFrame_GetLocalsplus(f);
    for (i = 0; i < na; i++) {
        Py_INCREF(*args);
        fastlocals[i] = *args++;
    }
    result = PyEval_EvalFrameEx(f,0);
    ++tstate->recursion_depth;
    Py_DECREF(f);
    --tstate->recursion_depth;
    return result;
}
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs) {
    PyCodeObject *co = (PyCodeObject *)PyFunction_GET_CODE(func);
    PyObject *globals = PyFunction_GET_GLOBALS(func);
    PyObject *argdefs = PyFunction_GET_DEFAULTS(func);
    PyObject *closure;
#if PY_MAJOR_VERSION >= 3
    PyObject *kwdefs;
#endif
    PyObject *kwtuple, **k;
    PyObject **d;
    Py_ssize_t nd;
    Py_ssize_t nk;
    PyObject *result;
    assert(kwargs == NULL || PyDict_Check(kwargs));
    nk = kwargs ? PyDict_Size(kwargs) : 0;
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object"))) {
        return NULL;
    }
    if (
#if PY_MAJOR_VERSION >= 3
            co->co_kwonlyargcount == 0 &&
#endif
            likely(kwargs == NULL || nk == 0) &&
            co->co_flags == (CO_OPTIMIZED | CO_NEWLOCALS | CO_NOFREE)) {
        if (argdefs == NULL && co->co_argcount == nargs) {
            result = __Pyx_PyFunction_FastCallNoKw(co, args, nargs, globals);
            goto done;
        }
        else if (nargs == 0 && argdefs != NULL
                 && co->co_argcount == Py_SIZE(argdefs)) {
            /* function called with no arguments, but all parameters have
               a default value: use default values as arguments .*/
            args = &PyTuple_GET_ITEM(argdefs, 0);
            result =__Pyx_PyFunction_FastCallNoKw(co, args, Py_SIZE(argdefs), globals);
            goto done;
        }
    }
    if (kwargs != NULL) {
        Py_ssize_t pos, i;
        kwtuple = PyTuple_New(2 * nk);
        if (kwtuple == NULL) {
            result = NULL;
            goto done;
        }
        k = &PyTuple_GET_ITEM(kwtuple, 0);
        pos = i = 0;
        while (PyDict_Next(kwargs, &pos, &k[i], &k[i+1])) {
            Py_INCREF(k[i]);
            Py_INCREF(k[i+1]);
            i += 2;
        }
        nk = i / 2;
    }
    else {
        kwtuple = NULL;
        k = NULL;
    }
    closure = PyFunction_GET_CLOSURE(func);
#if PY_MAJOR_VERSION >= 3
    kwdefs = PyFunction_GET_KW_DEFAULTS(func);
#endif
    if (argdefs != NULL) {
        d = &PyTuple_GET_ITEM(argdefs, 0);
        nd = Py_SIZE(argdefs);
    }
    else {
        d = NULL;
        nd = 0;
    }
#if PY_MAJOR_VERSION >= 3
    result = PyEval_EvalCodeEx((PyObject*)co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, kwdefs, closure);
#else
    result = PyEval_EvalCodeEx(co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, closure);
#endif
    Py_XDECREF(kwtuple);
done:
    Py_LeaveRecursiveCall();
    return result;
}
#endif

/* PyObjectCallMethO */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = PyCFunction_GET_FUNCTION(func);
    self = PyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectFastCall */
static PyObject* __Pyx_PyObject_FastCall_fallback(PyObject *func, PyObject **args, size_t nargs, PyObject *kwargs) {
    PyObject *argstuple;
    PyObject *result;
    size_t i;
    argstuple = PyTuple_New((Py_ssize_t)nargs);
    if (unlikely(!argstuple)) return NULL;
    for (i = 0; i < nargs; i++) {
        Py_INCREF(args[i]);
        PyTuple_SET_ITEM(argstuple, (Py_ssize_t)i, args[i]);
    }
    result = __Pyx_PyObject_Call(func, argstuple, kwargs);
    Py_DECREF(argstuple);
    return result;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_FastCallDict(PyObject *func, PyObject **args, size_t _nargs, PyObject *kwargs) {
    Py_ssize_t nargs = __Pyx_PyVectorcall_NARGS(_nargs);
#if CYTHON_COMPILING_IN_CPYTHON
    if (nargs == 0 && kwargs == NULL) {
#if defined(__Pyx_CyFunction_USED) && defined(NDEBUG)
        if (__Pyx_IsCyOrPyCFunction(func))
#else
        if (PyCFunction_Check(func))
#endif
        {
            if (likely(PyCFunction_GET_FLAGS(func) & METH_NOARGS)) {
                return __Pyx_PyObject_CallMethO(func, NULL);
            }
        }
    }
    else if (nargs == 1 && kwargs == NULL) {
        if (PyCFunction_Check(func))
        {
            if (likely(PyCFunction_GET_FLAGS(func) & METH_O)) {
                return __Pyx_PyObject_CallMethO(func, args[0]);
            }
        }
    }
#endif
    #if PY_VERSION_HEX < 0x030800B1
    #if CYTHON_FAST_PYCCALL
    if (PyCFunction_Check(func)) {
        if (kwargs) {
            return _PyCFunction_FastCallDict(func, args, nargs, kwargs);
        } else {
            return _PyCFunction_FastCallKeywords(func, args, nargs, NULL);
        }
    }
    #if PY_VERSION_HEX >= 0x030700A1
    if (!kwargs && __Pyx_IS_TYPE(func, &PyMethodDescr_Type)) {
        return _PyMethodDescr_FastCallKeywords(func, args, nargs, NULL);
    }
    #endif
    #endif
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCallDict(func, args, nargs, kwargs);
    }
    #endif
    #endif
    #if CYTHON_VECTORCALL
    vectorcallfunc f = _PyVectorcall_Function(func);
    if (f) {
        return f(func, args, (size_t)nargs, kwargs);
    }
    #elif defined(__Pyx_CyFunction_USED) && CYTHON_BACKPORT_VECTORCALL
    if (__Pyx_CyFunction_CheckExact(func)) {
        __pyx_vectorcallfunc f = __Pyx_CyFunction_func_vectorcall(func);
        if (f) return f(func, args, (size_t)nargs, kwargs);
    }
    #endif
    if (nargs == 0) {
        return __Pyx_PyObject_Call(func, __pyx_empty_tuple, kwargs);
    }
    return __Pyx_PyObject_FastCall_fallback(func, args, (size_t)nargs, kwargs);
}

/* GetItemInt */
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j) {
    PyObject *r;
    if (unlikely(!j)) return NULL;
    r = PyObject_GetItem(o, j);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyList_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyList_GET_SIZE(o)))) {
        PyObject *r = PyList_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyTuple_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyTuple_GET_SIZE(o)))) {
        PyObject *r = PyTuple_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i, int is_list,
                                                     CYTHON_NCP_UNUSED int wraparound,
                                                     CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyList_GET_SIZE(o);
        if ((!boundscheck) || (likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o))))) {
            PyObject *r = PyList_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    }
    else if (PyTuple_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyTuple_GET_SIZE(o);
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyTuple_GET_SIZE(o)))) {
            PyObject *r = PyTuple_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    } else {
        PyMappingMethods *mm = Py_TYPE(o)->tp_as_mapping;
        PySequenceMethods *sm = Py_TYPE(o)->tp_as_sequence;
        if (mm && mm->mp_subscript) {
            PyObject *r, *key = PyInt_FromSsize_t(i);
            if (unlikely(!key)) return NULL;
            r = mm->mp_subscript(o, key);
            Py_DECREF(key);
            return r;
        }
        if (likely(sm && sm->sq_item)) {
            if (wraparound && unlikely(i < 0) && likely(sm->sq_length)) {
                Py_ssize_t l = sm->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return NULL;
                    PyErr_Clear();
                }
            }
            return sm->sq_item(o, i);
        }
    }
#else
    if (is_list || PySequence_Check(o)) {
        return PySequence_GetItem(o, i);
    }
#endif
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
}

/* PyDictVersioning */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    return likely(dict) ? __PYX_GET_DICT_VERSION(dict) : 0;
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj) {
    PyObject **dictptr = NULL;
    Py_ssize_t offset = Py_TYPE(obj)->tp_dictoffset;
    if (offset) {
#if CYTHON_COMPILING_IN_CPYTHON
        dictptr = (likely(offset > 0)) ? (PyObject **) ((char *)obj + offset) : _PyObject_GetDictPtr(obj);
#else
        dictptr = _PyObject_GetDictPtr(obj);
#endif
    }
    return (dictptr && *dictptr) ? __PYX_GET_DICT_VERSION(*dictptr) : 0;
}
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    if (unlikely(!dict) || unlikely(tp_dict_version != __PYX_GET_DICT_VERSION(dict)))
        return 0;
    return obj_dict_version == __Pyx_get_object_dict_version(obj);
}
#endif

/* GetModuleGlobalName */
#if CYTHON_USE_DICT_VERSIONS
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value)
#else
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name)
#endif
{
    PyObject *result;
#if !CYTHON_AVOID_BORROWED_REFS
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1
    result = _PyDict_GetItem_KnownHash(__pyx_d, name, ((PyASCIIObject *) name)->hash);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    } else if (unlikely(PyErr_Occurred())) {
        return NULL;
    }
#elif CYTHON_COMPILING_IN_LIMITED_API
    if (unlikely(!__pyx_m)) {
        return NULL;
    }
    result = PyObject_GetAttr(__pyx_m, name);
    if (likely(result)) {
        return result;
    }
#else
    result = PyDict_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
#endif
#else
    result = PyObject_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
    PyErr_Clear();
#endif
    return __Pyx_GetBuiltinName(name);
}

/* RaiseUnboundLocalError */
static CYTHON_INLINE void __Pyx_RaiseUnboundLocalError(const char *varname) {
    PyErr_Format(PyExc_UnboundLocalError, "local variable '%s' referenced before assignment", varname);
}

/* UnicodeConcatInPlace */
# if CYTHON_COMPILING_IN_CPYTHON && PY_MAJOR_VERSION >= 3
static int
__Pyx_unicode_modifiable(PyObject *unicode)
{
    if (Py_REFCNT(unicode) != 1)
        return 0;
    if (!PyUnicode_CheckExact(unicode))
        return 0;
    if (PyUnicode_CHECK_INTERNED(unicode))
        return 0;
    return 1;
}
static CYTHON_INLINE PyObject *__Pyx_PyUnicode_ConcatInPlaceImpl(PyObject **p_left, PyObject *right
        #if CYTHON_REFNANNY
        , void* __pyx_refnanny
        #endif
    ) {
    PyObject *left = *p_left;
    Py_ssize_t left_len, right_len, new_len;
    if (unlikely(__Pyx_PyUnicode_READY(left) == -1))
        return NULL;
    if (unlikely(__Pyx_PyUnicode_READY(right) == -1))
        return NULL;
    left_len = PyUnicode_GET_LENGTH(left);
    if (left_len == 0) {
        Py_INCREF(right);
        return right;
    }
    right_len = PyUnicode_GET_LENGTH(right);
    if (right_len == 0) {
        Py_INCREF(left);
        return left;
    }
    if (unlikely(left_len > PY_SSIZE_T_MAX - right_len)) {
        PyErr_SetString(PyExc_OverflowError,
                        "strings are too large to concat");
        return NULL;
    }
    new_len = left_len + right_len;
    if (__Pyx_unicode_modifiable(left)
            && PyUnicode_CheckExact(right)
            && PyUnicode_KIND(right) <= PyUnicode_KIND(left)
            && !(PyUnicode_IS_ASCII(left) && !PyUnicode_IS_ASCII(right))) {
        __Pyx_GIVEREF(*p_left);
        if (unlikely(PyUnicode_Resize(p_left, new_len) != 0)) {
            __Pyx_GOTREF(*p_left);
            return NULL;
        }
        __Pyx_INCREF(*p_left);
        _PyUnicode_FastCopyCharacters(*p_left, left_len, right, 0, right_len);
        return *p_left;
    } else {
        return __Pyx_PyUnicode_Concat(left, right);
    }
  }
#endif

/* RaiseUnexpectedTypeError */
static int
__Pyx_RaiseUnexpectedTypeError(const char *expected, PyObject *obj)
{
    __Pyx_TypeName obj_type_name = __Pyx_PyType_GetName(Py_TYPE(obj));
    PyErr_Format(PyExc_TypeError, "Expected %s, got " __Pyx_FMT_TYPENAME,
                 expected, obj_type_name);
    __Pyx_DECREF_TypeName(obj_type_name);
    return 0;
}

/* DictGetItem */
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key) {
    PyObject *value;
    value = PyDict_GetItemWithError(d, key);
    if (unlikely(!value)) {
        if (!PyErr_Occurred()) {
            if (unlikely(PyTuple_Check(key))) {
                PyObject* args = PyTuple_Pack(1, key);
                if (likely(args)) {
                    PyErr_SetObject(PyExc_KeyError, args);
                    Py_DECREF(args);
                }
            } else {
                PyErr_SetObject(PyExc_KeyError, key);
            }
        }
        return NULL;
    }
    Py_INCREF(value);
    return value;
}
#endif

/* SetItemInt */
static int __Pyx_SetItemInt_Generic(PyObject *o, PyObject *j, PyObject *v) {
    int r;
    if (unlikely(!j)) return -1;
    r = PyObject_SetItem(o, j, v);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE int __Pyx_SetItemInt_Fast(PyObject *o, Py_ssize_t i, PyObject *v, int is_list,
                                               CYTHON_NCP_UNUSED int wraparound, CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = (!wraparound) ? i : ((likely(i >= 0)) ? i : i + PyList_GET_SIZE(o));
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o)))) {
            PyObject* old = PyList_GET_ITEM(o, n);
            Py_INCREF(v);
            PyList_SET_ITEM(o, n, v);
            Py_DECREF(old);
            return 1;
        }
    } else {
        PyMappingMethods *mm = Py_TYPE(o)->tp_as_mapping;
        PySequenceMethods *sm = Py_TYPE(o)->tp_as_sequence;
        if (mm && mm->mp_ass_subscript) {
            int r;
            PyObject *key = PyInt_FromSsize_t(i);
            if (unlikely(!key)) return -1;
            r = mm->mp_ass_subscript(o, key, v);
            Py_DECREF(key);
            return r;
        }
        if (likely(sm && sm->sq_ass_item)) {
            if (wraparound && unlikely(i < 0) && likely(sm->sq_length)) {
                Py_ssize_t l = sm->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return -1;
                    PyErr_Clear();
                }
            }
            return sm->sq_ass_item(o, i, v);
        }
    }
#else
#if CYTHON_COMPILING_IN_PYPY
    if (is_list || (PySequence_Check(o) && !PyDict_Check(o)))
#else
    if (is_list || PySequence_Check(o))
#endif
    {
        return PySequence_SetItem(o, i, v);
    }
#endif
    return __Pyx_SetItemInt_Generic(o, PyInt_FromSsize_t(i), v);
}

/* SliceObject */
static CYTHON_INLINE int __Pyx_PyObject_SetSlice(PyObject* obj, PyObject* value,
        Py_ssize_t cstart, Py_ssize_t cstop,
        PyObject** _py_start, PyObject** _py_stop, PyObject** _py_slice,
        int has_cstart, int has_cstop, int wraparound) {
    __Pyx_TypeName obj_type_name;
#if CYTHON_USE_TYPE_SLOTS
    PyMappingMethods* mp;
#if PY_MAJOR_VERSION < 3
    PySequenceMethods* ms = Py_TYPE(obj)->tp_as_sequence;
    if (likely(ms && ms->sq_ass_slice)) {
        if (!has_cstart) {
            if (_py_start && (*_py_start != Py_None)) {
                cstart = __Pyx_PyIndex_AsSsize_t(*_py_start);
                if ((cstart == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
            } else
                cstart = 0;
        }
        if (!has_cstop) {
            if (_py_stop && (*_py_stop != Py_None)) {
                cstop = __Pyx_PyIndex_AsSsize_t(*_py_stop);
                if ((cstop == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
            } else
                cstop = PY_SSIZE_T_MAX;
        }
        if (wraparound && unlikely((cstart < 0) | (cstop < 0)) && likely(ms->sq_length)) {
            Py_ssize_t l = ms->sq_length(obj);
            if (likely(l >= 0)) {
                if (cstop < 0) {
                    cstop += l;
                    if (cstop < 0) cstop = 0;
                }
                if (cstart < 0) {
                    cstart += l;
                    if (cstart < 0) cstart = 0;
                }
            } else {
                if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                    goto bad;
                PyErr_Clear();
            }
        }
        return ms->sq_ass_slice(obj, cstart, cstop, value);
    }
#else
    CYTHON_UNUSED_VAR(wraparound);
#endif
    mp = Py_TYPE(obj)->tp_as_mapping;
    if (likely(mp && mp->mp_ass_subscript))
#else
    CYTHON_UNUSED_VAR(wraparound);
#endif
    {
        int result;
        PyObject *py_slice, *py_start, *py_stop;
        if (_py_slice) {
            py_slice = *_py_slice;
        } else {
            PyObject* owned_start = NULL;
            PyObject* owned_stop = NULL;
            if (_py_start) {
                py_start = *_py_start;
            } else {
                if (has_cstart) {
                    owned_start = py_start = PyInt_FromSsize_t(cstart);
                    if (unlikely(!py_start)) goto bad;
                } else
                    py_start = Py_None;
            }
            if (_py_stop) {
                py_stop = *_py_stop;
            } else {
                if (has_cstop) {
                    owned_stop = py_stop = PyInt_FromSsize_t(cstop);
                    if (unlikely(!py_stop)) {
                        Py_XDECREF(owned_start);
                        goto bad;
                    }
                } else
                    py_stop = Py_None;
            }
            py_slice = PySlice_New(py_start, py_stop, Py_None);
            Py_XDECREF(owned_start);
            Py_XDECREF(owned_stop);
            if (unlikely(!py_slice)) goto bad;
        }
#if CYTHON_USE_TYPE_SLOTS
        result = mp->mp_ass_subscript(obj, py_slice, value);
#else
        result = value ? PyObject_SetItem(obj, py_slice, value) : PyObject_DelItem(obj, py_slice);
#endif
        if (!_py_slice) {
            Py_DECREF(py_slice);
        }
        return result;
    }
    obj_type_name = __Pyx_PyType_GetName(Py_TYPE(obj));
    PyErr_Format(PyExc_TypeError,
        "'" __Pyx_FMT_TYPENAME "' object does not support slice %.10s",
        obj_type_name, value ? "assignment" : "deletion");
    __Pyx_DECREF_TypeName(obj_type_name);
bad:
    return -1;
}

/* PyObjectCallOneArg */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *args[2] = {NULL, arg};
    return __Pyx_PyObject_FastCall(func, args+1, 1 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
}

/* UnpackUnboundCMethod */
static PyObject *__Pyx_SelflessCall(PyObject *method, PyObject *args, PyObject *kwargs) {
    PyObject *selfless_args = PyTuple_GetSlice(args, 1, PyTuple_Size(args));
    if (unlikely(!selfless_args)) return NULL;
    PyObject *result = PyObject_Call(method, selfless_args, kwargs);
    Py_DECREF(selfless_args);
    return result;
}
static PyMethodDef __Pyx_UnboundCMethod_Def = {
     "CythonUnboundCMethod",
     __PYX_REINTERPRET_FUNCION(PyCFunction, __Pyx_SelflessCall),
     METH_VARARGS | METH_KEYWORDS,
     NULL
};
static int __Pyx_TryUnpackUnboundCMethod(__Pyx_CachedCFunction* target) {
    PyObject *method;
    method = __Pyx_PyObject_GetAttrStr(target->type, *target->method_name);
    if (unlikely(!method))
        return -1;
    target->method = method;
#if CYTHON_COMPILING_IN_CPYTHON
    #if PY_MAJOR_VERSION >= 3
    if (likely(__Pyx_TypeCheck(method, &PyMethodDescr_Type)))
    #else
    if (likely(!PyCFunction_Check(method)))
    #endif
    {
        PyMethodDescrObject *descr = (PyMethodDescrObject*) method;
        target->func = descr->d_method->ml_meth;
        target->flag = descr->d_method->ml_flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_STACKLESS);
    } else
#endif
#if defined(CYTHON_COMPILING_IN_PYPY)
#elif PY_VERSION_HEX >= 0x03090000
    if (PyCFunction_CheckExact(method))
#else
    if (PyCFunction_Check(method))
#endif
    {
        PyObject *self;
        int self_found;
#if CYTHON_COMPILING_IN_LIMITED_API || CYTHON_COMPILING_IN_PYPY
        self = PyObject_GetAttrString(method, "__self__");
        if (!self) {
            PyErr_Clear();
        }
#else
        self = PyCFunction_GET_SELF(method);
#endif
        self_found = (self && self != Py_None);
#if CYTHON_COMPILING_IN_LIMITED_API || CYTHON_COMPILING_IN_PYPY
        Py_XDECREF(self);
#endif
        if (self_found) {
            PyObject *unbound_method = PyCFunction_New(&__Pyx_UnboundCMethod_Def, method);
            if (unlikely(!unbound_method)) return -1;
            Py_DECREF(method);
            target->method = unbound_method;
        }
    }
    return 0;
}

/* CallUnboundCMethod1 */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_CallUnboundCMethod1(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg) {
    if (likely(cfunc->func)) {
        int flag = cfunc->flag;
        if (flag == METH_O) {
            return (*(cfunc->func))(self, arg);
        } else if ((PY_VERSION_HEX >= 0x030600B1) && flag == METH_FASTCALL) {
            #if PY_VERSION_HEX >= 0x030700A0
                return (*(__Pyx_PyCFunctionFast)(void*)(PyCFunction)cfunc->func)(self, &arg, 1);
            #else
                return (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)cfunc->func)(self, &arg, 1, NULL);
            #endif
        } else if ((PY_VERSION_HEX >= 0x030700A0) && flag == (METH_FASTCALL | METH_KEYWORDS)) {
            return (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)cfunc->func)(self, &arg, 1, NULL);
        }
    }
    return __Pyx__CallUnboundCMethod1(cfunc, self, arg);
}
#endif
static PyObject* __Pyx__CallUnboundCMethod1(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg){
    PyObject *args, *result = NULL;
    if (unlikely(!cfunc->func && !cfunc->method) && unlikely(__Pyx_TryUnpackUnboundCMethod(cfunc) < 0)) return NULL;
#if CYTHON_COMPILING_IN_CPYTHON
    if (cfunc->func && (cfunc->flag & METH_VARARGS)) {
        args = PyTuple_New(1);
        if (unlikely(!args)) goto bad;
        Py_INCREF(arg);
        PyTuple_SET_ITEM(args, 0, arg);
        if (cfunc->flag & METH_KEYWORDS)
            result = (*(PyCFunctionWithKeywords)(void*)(PyCFunction)cfunc->func)(self, args, NULL);
        else
            result = (*cfunc->func)(self, args);
    } else {
        args = PyTuple_New(2);
        if (unlikely(!args)) goto bad;
        Py_INCREF(self);
        PyTuple_SET_ITEM(args, 0, self);
        Py_INCREF(arg);
        PyTuple_SET_ITEM(args, 1, arg);
        result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
    }
#else
    args = PyTuple_Pack(2, self, arg);
    if (unlikely(!args)) goto bad;
    result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
#endif
bad:
    Py_XDECREF(args);
    return result;
}

/* CallUnboundCMethod2 */
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030600B1
static CYTHON_INLINE PyObject *__Pyx_CallUnboundCMethod2(__Pyx_CachedCFunction *cfunc, PyObject *self, PyObject *arg1, PyObject *arg2) {
    if (likely(cfunc->func)) {
        PyObject *args[2] = {arg1, arg2};
        if (cfunc->flag == METH_FASTCALL) {
            #if PY_VERSION_HEX >= 0x030700A0
            return (*(__Pyx_PyCFunctionFast)(void*)(PyCFunction)cfunc->func)(self, args, 2);
            #else
            return (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)cfunc->func)(self, args, 2, NULL);
            #endif
        }
        #if PY_VERSION_HEX >= 0x030700A0
        if (cfunc->flag == (METH_FASTCALL | METH_KEYWORDS))
            return (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)cfunc->func)(self, args, 2, NULL);
        #endif
    }
    return __Pyx__CallUnboundCMethod2(cfunc, self, arg1, arg2);
}
#endif
static PyObject* __Pyx__CallUnboundCMethod2(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg1, PyObject* arg2){
    PyObject *args, *result = NULL;
    if (unlikely(!cfunc->func && !cfunc->method) && unlikely(__Pyx_TryUnpackUnboundCMethod(cfunc) < 0)) return NULL;
#if CYTHON_COMPILING_IN_CPYTHON
    if (cfunc->func && (cfunc->flag & METH_VARARGS)) {
        args = PyTuple_New(2);
        if (unlikely(!args)) goto bad;
        Py_INCREF(arg1);
        PyTuple_SET_ITEM(args, 0, arg1);
        Py_INCREF(arg2);
        PyTuple_SET_ITEM(args, 1, arg2);
        if (cfunc->flag & METH_KEYWORDS)
            result = (*(PyCFunctionWithKeywords)(void*)(PyCFunction)cfunc->func)(self, args, NULL);
        else
            result = (*cfunc->func)(self, args);
    } else {
        args = PyTuple_New(3);
        if (unlikely(!args)) goto bad;
        Py_INCREF(self);
        PyTuple_SET_ITEM(args, 0, self);
        Py_INCREF(arg1);
        PyTuple_SET_ITEM(args, 1, arg1);
        Py_INCREF(arg2);
        PyTuple_SET_ITEM(args, 2, arg2);
        result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
    }
#else
    args = PyTuple_Pack(3, self, arg1, arg2);
    if (unlikely(!args)) goto bad;
    result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
#endif
bad:
    Py_XDECREF(args);
    return result;
}

/* dict_getitem_default */
static PyObject* __Pyx_PyDict_GetItemDefault(PyObject* d, PyObject* key, PyObject* default_value) {
    PyObject* value;
#if PY_MAJOR_VERSION >= 3 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07020000)
    value = PyDict_GetItemWithError(d, key);
    if (unlikely(!value)) {
        if (unlikely(PyErr_Occurred()))
            return NULL;
        value = default_value;
    }
    Py_INCREF(value);
    if ((1));
#else
    if (PyString_CheckExact(key) || PyUnicode_CheckExact(key) || PyInt_CheckExact(key)) {
        value = PyDict_GetItem(d, key);
        if (unlikely(!value)) {
            value = default_value;
        }
        Py_INCREF(value);
    }
#endif
    else {
        if (default_value == Py_None)
            value = __Pyx_CallUnboundCMethod1(&__pyx_umethod_PyDict_Type_get, d, key);
        else
            value = __Pyx_CallUnboundCMethod2(&__pyx_umethod_PyDict_Type_get, d, key, default_value);
    }
    return value;
}

/* PyObjectCall2Args */
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2) {
    PyObject *args[3] = {NULL, arg1, arg2};
    return __Pyx_PyObject_FastCall(function, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
}

/* PyObjectGetMethod */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method) {
    PyObject *attr;
#if CYTHON_UNPACK_METHODS && CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_PYTYPE_LOOKUP
    __Pyx_TypeName type_name;
    PyTypeObject *tp = Py_TYPE(obj);
    PyObject *descr;
    descrgetfunc f = NULL;
    PyObject **dictptr, *dict;
    int meth_found = 0;
    assert (*method == NULL);
    if (unlikely(tp->tp_getattro != PyObject_GenericGetAttr)) {
        attr = __Pyx_PyObject_GetAttrStr(obj, name);
        goto try_unpack;
    }
    if (unlikely(tp->tp_dict == NULL) && unlikely(PyType_Ready(tp) < 0)) {
        return 0;
    }
    descr = _PyType_Lookup(tp, name);
    if (likely(descr != NULL)) {
        Py_INCREF(descr);
#if defined(Py_TPFLAGS_METHOD_DESCRIPTOR) && Py_TPFLAGS_METHOD_DESCRIPTOR
        if (__Pyx_PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_METHOD_DESCRIPTOR))
#elif PY_MAJOR_VERSION >= 3
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || __Pyx_IS_TYPE(descr, &PyMethodDescr_Type) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr) || __Pyx_IS_TYPE(descr, &PyMethodDescr_Type)))
        #endif
#else
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr)))
        #endif
#endif
        {
            meth_found = 1;
        } else {
            f = Py_TYPE(descr)->tp_descr_get;
            if (f != NULL && PyDescr_IsData(descr)) {
                attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
                Py_DECREF(descr);
                goto try_unpack;
            }
        }
    }
    dictptr = _PyObject_GetDictPtr(obj);
    if (dictptr != NULL && (dict = *dictptr) != NULL) {
        Py_INCREF(dict);
        attr = __Pyx_PyDict_GetItemStr(dict, name);
        if (attr != NULL) {
            Py_INCREF(attr);
            Py_DECREF(dict);
            Py_XDECREF(descr);
            goto try_unpack;
        }
        Py_DECREF(dict);
    }
    if (meth_found) {
        *method = descr;
        return 1;
    }
    if (f != NULL) {
        attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
        Py_DECREF(descr);
        goto try_unpack;
    }
    if (likely(descr != NULL)) {
        *method = descr;
        return 0;
    }
    type_name = __Pyx_PyType_GetName(tp);
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'" __Pyx_FMT_TYPENAME "' object has no attribute '%U'",
                 type_name, name);
#else
                 "'" __Pyx_FMT_TYPENAME "' object has no attribute '%.400s'",
                 type_name, PyString_AS_STRING(name));
#endif
    __Pyx_DECREF_TypeName(type_name);
    return 0;
#else
    attr = __Pyx_PyObject_GetAttrStr(obj, name);
    goto try_unpack;
#endif
try_unpack:
#if CYTHON_UNPACK_METHODS
    if (likely(attr) && PyMethod_Check(attr) && likely(PyMethod_GET_SELF(attr) == obj)) {
        PyObject *function = PyMethod_GET_FUNCTION(attr);
        Py_INCREF(function);
        Py_DECREF(attr);
        *method = function;
        return 1;
    }
#endif
    *method = attr;
    return 0;
}

/* PyObjectCallMethod1 */
static PyObject* __Pyx__PyObject_CallMethod1(PyObject* method, PyObject* arg) {
    PyObject *result = __Pyx_PyObject_CallOneArg(method, arg);
    Py_DECREF(method);
    return result;
}
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg) {
    PyObject *method = NULL, *result;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_Call2Args(method, obj, arg);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) return NULL;
    return __Pyx__PyObject_CallMethod1(method, arg);
}

/* append */
static CYTHON_INLINE int __Pyx_PyObject_Append(PyObject* L, PyObject* x) {
    if (likely(PyList_CheckExact(L))) {
        if (unlikely(__Pyx_PyList_Append(L, x) < 0)) return -1;
    } else {
        PyObject* retval = __Pyx_PyObject_CallMethod1(L, __pyx_n_s_append, x);
        if (unlikely(!retval))
            return -1;
        Py_DECREF(retval);
    }
    return 0;
}

/* SliceObject */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetSlice(PyObject* obj,
        Py_ssize_t cstart, Py_ssize_t cstop,
        PyObject** _py_start, PyObject** _py_stop, PyObject** _py_slice,
        int has_cstart, int has_cstop, int wraparound) {
    __Pyx_TypeName obj_type_name;
#if CYTHON_USE_TYPE_SLOTS
    PyMappingMethods* mp;
#if PY_MAJOR_VERSION < 3
    PySequenceMethods* ms = Py_TYPE(obj)->tp_as_sequence;
    if (likely(ms && ms->sq_slice)) {
        if (!has_cstart) {
            if (_py_start && (*_py_start != Py_None)) {
                cstart = __Pyx_PyIndex_AsSsize_t(*_py_start);
                if ((cstart == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
            } else
                cstart = 0;
        }
        if (!has_cstop) {
            if (_py_stop && (*_py_stop != Py_None)) {
                cstop = __Pyx_PyIndex_AsSsize_t(*_py_stop);
                if ((cstop == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
            } else
                cstop = PY_SSIZE_T_MAX;
        }
        if (wraparound && unlikely((cstart < 0) | (cstop < 0)) && likely(ms->sq_length)) {
            Py_ssize_t l = ms->sq_length(obj);
            if (likely(l >= 0)) {
                if (cstop < 0) {
                    cstop += l;
                    if (cstop < 0) cstop = 0;
                }
                if (cstart < 0) {
                    cstart += l;
                    if (cstart < 0) cstart = 0;
                }
            } else {
                if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                    goto bad;
                PyErr_Clear();
            }
        }
        return ms->sq_slice(obj, cstart, cstop);
    }
#else
    CYTHON_UNUSED_VAR(wraparound);
#endif
    mp = Py_TYPE(obj)->tp_as_mapping;
    if (likely(mp && mp->mp_subscript))
#else
    CYTHON_UNUSED_VAR(wraparound);
#endif
    {
        PyObject* result;
        PyObject *py_slice, *py_start, *py_stop;
        if (_py_slice) {
            py_slice = *_py_slice;
        } else {
            PyObject* owned_start = NULL;
            PyObject* owned_stop = NULL;
            if (_py_start) {
                py_start = *_py_start;
            } else {
                if (has_cstart) {
                    owned_start = py_start = PyInt_FromSsize_t(cstart);
                    if (unlikely(!py_start)) goto bad;
                } else
                    py_start = Py_None;
            }
            if (_py_stop) {
                py_stop = *_py_stop;
            } else {
                if (has_cstop) {
                    owned_stop = py_stop = PyInt_FromSsize_t(cstop);
                    if (unlikely(!py_stop)) {
                        Py_XDECREF(owned_start);
                        goto bad;
                    }
                } else
                    py_stop = Py_None;
            }
            py_slice = PySlice_New(py_start, py_stop, Py_None);
            Py_XDECREF(owned_start);
            Py_XDECREF(owned_stop);
            if (unlikely(!py_slice)) goto bad;
        }
#if CYTHON_USE_TYPE_SLOTS
        result = mp->mp_subscript(obj, py_slice);
#else
        result = PyObject_GetItem(obj, py_slice);
#endif
        if (!_py_slice) {
            Py_DECREF(py_slice);
        }
        return result;
    }
    obj_type_name = __Pyx_PyType_GetName(Py_TYPE(obj));
    PyErr_Format(PyExc_TypeError,
        "'" __Pyx_FMT_TYPENAME "' object is unsliceable", obj_type_name);
    __Pyx_DECREF_TypeName(obj_type_name);
bad:
    return NULL;
}

/* RaiseTooManyValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected) {
    PyErr_Format(PyExc_ValueError,
                 "too many values to unpack (expected %" CYTHON_FORMAT_SSIZE_T "d)", expected);
}

/* RaiseNeedMoreValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index) {
    PyErr_Format(PyExc_ValueError,
                 "need more than %" CYTHON_FORMAT_SSIZE_T "d value%.1s to unpack",
                 index, (index == 1) ? "" : "s");
}

/* IterFinish */
static CYTHON_INLINE int __Pyx_IterFinish(void) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    PyObject* exc_type = __Pyx_PyErr_CurrentExceptionType();
    if (unlikely(exc_type)) {
        if (unlikely(!__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration)))
            return -1;
        __Pyx_PyErr_Clear();
        return 0;
    }
    return 0;
}

/* UnpackItemEndCheck */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected) {
    if (unlikely(retval)) {
        Py_DECREF(retval);
        __Pyx_RaiseTooManyValuesError(expected);
        return -1;
    }
    return __Pyx_IterFinish();
}

/* ObjectGetItem */
#if CYTHON_USE_TYPE_SLOTS
static PyObject *__Pyx_PyObject_GetIndex(PyObject *obj, PyObject *index) {
    PyObject *runerr = NULL;
    Py_ssize_t key_value;
    key_value = __Pyx_PyIndex_AsSsize_t(index);
    if (likely(key_value != -1 || !(runerr = PyErr_Occurred()))) {
        return __Pyx_GetItemInt_Fast(obj, key_value, 0, 1, 1);
    }
    if (PyErr_GivenExceptionMatches(runerr, PyExc_OverflowError)) {
        __Pyx_TypeName index_type_name = __Pyx_PyType_GetName(Py_TYPE(index));
        PyErr_Clear();
        PyErr_Format(PyExc_IndexError,
            "cannot fit '" __Pyx_FMT_TYPENAME "' into an index-sized integer", index_type_name);
        __Pyx_DECREF_TypeName(index_type_name);
    }
    return NULL;
}
static PyObject *__Pyx_PyObject_GetItem_Slow(PyObject *obj, PyObject *key) {
    __Pyx_TypeName obj_type_name;
    if (likely(PyType_Check(obj))) {
        PyObject *meth = __Pyx_PyObject_GetAttrStrNoError(obj, __pyx_n_s_class_getitem);
        if (meth) {
            PyObject *result = __Pyx_PyObject_CallOneArg(meth, key);
            Py_DECREF(meth);
            return result;
        }
    }
    obj_type_name = __Pyx_PyType_GetName(Py_TYPE(obj));
    PyErr_Format(PyExc_TypeError,
        "'" __Pyx_FMT_TYPENAME "' object is not subscriptable", obj_type_name);
    __Pyx_DECREF_TypeName(obj_type_name);
    return NULL;
}
static PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject *key) {
    PyTypeObject *tp = Py_TYPE(obj);
    PyMappingMethods *mm = tp->tp_as_mapping;
    PySequenceMethods *sm = tp->tp_as_sequence;
    if (likely(mm && mm->mp_subscript)) {
        return mm->mp_subscript(obj, key);
    }
    if (likely(sm && sm->sq_item)) {
        return __Pyx_PyObject_GetIndex(obj, key);
    }
    return __Pyx_PyObject_GetItem_Slow(obj, key);
}
#endif

/* CallUnboundCMethod0 */
static PyObject* __Pyx__CallUnboundCMethod0(__Pyx_CachedCFunction* cfunc, PyObject* self) {
    PyObject *args, *result = NULL;
    if (unlikely(!cfunc->method) && unlikely(__Pyx_TryUnpackUnboundCMethod(cfunc) < 0)) return NULL;
#if CYTHON_ASSUME_SAFE_MACROS
    args = PyTuple_New(1);
    if (unlikely(!args)) goto bad;
    Py_INCREF(self);
    PyTuple_SET_ITEM(args, 0, self);
#else
    args = PyTuple_Pack(1, self);
    if (unlikely(!args)) goto bad;
#endif
    result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
    Py_DECREF(args);
bad:
    return result;
}

/* py_dict_items */
static CYTHON_INLINE PyObject* __Pyx_PyDict_Items(PyObject* d) {
    if (PY_MAJOR_VERSION >= 3)
        return __Pyx_CallUnboundCMethod0(&__pyx_umethod_PyDict_Type_items, d);
    else
        return PyDict_Items(d);
}

/* GetAttr */
static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *o, PyObject *n) {
#if CYTHON_USE_TYPE_SLOTS
#if PY_MAJOR_VERSION >= 3
    if (likely(PyUnicode_Check(n)))
#else
    if (likely(PyString_Check(n)))
#endif
        return __Pyx_PyObject_GetAttrStr(o, n);
#endif
    return PyObject_GetAttr(o, n);
}

/* HasAttr */
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *o, PyObject *n) {
    PyObject *r;
    if (unlikely(!__Pyx_PyBaseString_Check(n))) {
        PyErr_SetString(PyExc_TypeError,
                        "hasattr(): attribute name must be string");
        return -1;
    }
    r = __Pyx_GetAttr(o, n);
    if (!r) {
        PyErr_Clear();
        return 0;
    } else {
        Py_DECREF(r);
        return 1;
    }
}

/* GetTopmostException */
#if CYTHON_USE_EXC_INFO_STACK && CYTHON_FAST_THREAD_STATE
static _PyErr_StackItem *
__Pyx_PyErr_GetTopmostException(PyThreadState *tstate)
{
    _PyErr_StackItem *exc_info = tstate->exc_info;
    while ((exc_info->exc_value == NULL || exc_info->exc_value == Py_None) &&
           exc_info->previous_item != NULL)
    {
        exc_info = exc_info->previous_item;
    }
    return exc_info;
}
#endif

/* SaveResetException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
  #if CYTHON_USE_EXC_INFO_STACK && PY_VERSION_HEX >= 0x030B00a4
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    PyObject *exc_value = exc_info->exc_value;
    if (exc_value == NULL || exc_value == Py_None) {
        *value = NULL;
        *type = NULL;
        *tb = NULL;
    } else {
        *value = exc_value;
        Py_INCREF(*value);
        *type = (PyObject*) Py_TYPE(exc_value);
        Py_INCREF(*type);
        *tb = PyException_GetTraceback(exc_value);
    }
  #elif CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    *type = exc_info->exc_type;
    *value = exc_info->exc_value;
    *tb = exc_info->exc_traceback;
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
  #else
    *type = tstate->exc_type;
    *value = tstate->exc_value;
    *tb = tstate->exc_traceback;
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
  #endif
}
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
  #if CYTHON_USE_EXC_INFO_STACK && PY_VERSION_HEX >= 0x030B00a4
    _PyErr_StackItem *exc_info = tstate->exc_info;
    PyObject *tmp_value = exc_info->exc_value;
    exc_info->exc_value = value;
    Py_XDECREF(tmp_value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
  #else
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = type;
    exc_info->exc_value = value;
    exc_info->exc_traceback = tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = type;
    tstate->exc_value = value;
    tstate->exc_traceback = tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
  #endif
}
#endif

/* GetException */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb)
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb)
#endif
{
    PyObject *local_type = NULL, *local_value, *local_tb = NULL;
#if CYTHON_FAST_THREAD_STATE
    PyObject *tmp_type, *tmp_value, *tmp_tb;
  #if PY_VERSION_HEX >= 0x030C00A6
    local_value = tstate->current_exception;
    tstate->current_exception = 0;
    if (likely(local_value)) {
        local_type = (PyObject*) Py_TYPE(local_value);
        Py_INCREF(local_type);
        local_tb = PyException_GetTraceback(local_value);
    }
  #else
    local_type = tstate->curexc_type;
    local_value = tstate->curexc_value;
    local_tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
  #endif
#else
    PyErr_Fetch(&local_type, &local_value, &local_tb);
#endif
    PyErr_NormalizeException(&local_type, &local_value, &local_tb);
#if CYTHON_FAST_THREAD_STATE && PY_VERSION_HEX >= 0x030C00A6
    if (unlikely(tstate->current_exception))
#elif CYTHON_FAST_THREAD_STATE
    if (unlikely(tstate->curexc_type))
#else
    if (unlikely(PyErr_Occurred()))
#endif
        goto bad;
    #if PY_MAJOR_VERSION >= 3
    if (local_tb) {
        if (unlikely(PyException_SetTraceback(local_value, local_tb) < 0))
            goto bad;
    }
    #endif
    Py_XINCREF(local_tb);
    Py_XINCREF(local_type);
    Py_XINCREF(local_value);
    *type = local_type;
    *value = local_value;
    *tb = local_tb;
#if CYTHON_FAST_THREAD_STATE
    #if CYTHON_USE_EXC_INFO_STACK
    {
        _PyErr_StackItem *exc_info = tstate->exc_info;
      #if PY_VERSION_HEX >= 0x030B00a4
        tmp_value = exc_info->exc_value;
        exc_info->exc_value = local_value;
        tmp_type = NULL;
        tmp_tb = NULL;
        Py_XDECREF(local_type);
        Py_XDECREF(local_tb);
      #else
        tmp_type = exc_info->exc_type;
        tmp_value = exc_info->exc_value;
        tmp_tb = exc_info->exc_traceback;
        exc_info->exc_type = local_type;
        exc_info->exc_value = local_value;
        exc_info->exc_traceback = local_tb;
      #endif
    }
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = local_type;
    tstate->exc_value = local_value;
    tstate->exc_traceback = local_tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#else
    PyErr_SetExcInfo(local_type, local_value, local_tb);
#endif
    return 0;
bad:
    *type = 0;
    *value = 0;
    *tb = 0;
    Py_XDECREF(local_type);
    Py_XDECREF(local_value);
    Py_XDECREF(local_tb);
    return -1;
}

/* pyfrozenset_new */
static CYTHON_INLINE PyObject* __Pyx_PyFrozenSet_New(PyObject* it) {
    if (it) {
        PyObject* result;
#if CYTHON_COMPILING_IN_PYPY
        PyObject* args;
        args = PyTuple_Pack(1, it);
        if (unlikely(!args))
            return NULL;
        result = PyObject_Call((PyObject*)&PyFrozenSet_Type, args, NULL);
        Py_DECREF(args);
        return result;
#else
        if (PyFrozenSet_CheckExact(it)) {
            Py_INCREF(it);
            return it;
        }
        result = PyFrozenSet_New(it);
        if (unlikely(!result))
            return NULL;
        if ((PY_VERSION_HEX >= 0x031000A1) || likely(PySet_GET_SIZE(result)))
            return result;
        Py_DECREF(result);
#endif
    }
#if CYTHON_USE_TYPE_SLOTS
    return PyFrozenSet_Type.tp_new(&PyFrozenSet_Type, __pyx_empty_tuple, NULL);
#else
    return PyObject_Call((PyObject*)&PyFrozenSet_Type, __pyx_empty_tuple, NULL);
#endif
}

/* PySetContains */
static int __Pyx_PySet_ContainsUnhashable(PyObject *set, PyObject *key) {
    int result = -1;
    if (PySet_Check(key) && PyErr_ExceptionMatches(PyExc_TypeError)) {
        PyObject *tmpkey;
        PyErr_Clear();
        tmpkey = __Pyx_PyFrozenSet_New(key);
        if (tmpkey != NULL) {
            result = PySet_Contains(set, tmpkey);
            Py_DECREF(tmpkey);
        }
    }
    return result;
}
static CYTHON_INLINE int __Pyx_PySet_ContainsTF(PyObject* key, PyObject* set, int eq) {
    int result = PySet_Contains(set, key);
    if (unlikely(result < 0)) {
        result = __Pyx_PySet_ContainsUnhashable(set, key);
    }
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

/* PyObjectCallNoArg */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func) {
    PyObject *arg = NULL;
    return __Pyx_PyObject_FastCall(func, (&arg)+1, 0 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
}

/* FixUpExtensionType */
#if CYTHON_USE_TYPE_SPECS
static int __Pyx_fix_up_extension_type_from_spec(PyType_Spec *spec, PyTypeObject *type) {
#if PY_VERSION_HEX > 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
    CYTHON_UNUSED_VAR(spec);
    CYTHON_UNUSED_VAR(type);
#else
    const PyType_Slot *slot = spec->slots;
    while (slot && slot->slot && slot->slot != Py_tp_members)
        slot++;
    if (slot && slot->slot == Py_tp_members) {
        int changed = 0;
#if !(PY_VERSION_HEX <= 0x030900b1 && CYTHON_COMPILING_IN_CPYTHON)
        const
#endif
            PyMemberDef *memb = (PyMemberDef*) slot->pfunc;
        while (memb && memb->name) {
            if (memb->name[0] == '_' && memb->name[1] == '_') {
#if PY_VERSION_HEX < 0x030900b1
                if (strcmp(memb->name, "__weaklistoffset__") == 0) {
                    assert(memb->type == T_PYSSIZET);
                    assert(memb->flags == READONLY);
                    type->tp_weaklistoffset = memb->offset;
                    changed = 1;
                }
                else if (strcmp(memb->name, "__dictoffset__") == 0) {
                    assert(memb->type == T_PYSSIZET);
                    assert(memb->flags == READONLY);
                    type->tp_dictoffset = memb->offset;
                    changed = 1;
                }
#if CYTHON_METH_FASTCALL
                else if (strcmp(memb->name, "__vectorcalloffset__") == 0) {
                    assert(memb->type == T_PYSSIZET);
                    assert(memb->flags == READONLY);
#if PY_VERSION_HEX >= 0x030800b4
                    type->tp_vectorcall_offset = memb->offset;
#else
                    type->tp_print = (printfunc) memb->offset;
#endif
                    changed = 1;
                }
#endif
#else
                if ((0));
#endif
#if PY_VERSION_HEX <= 0x030900b1 && CYTHON_COMPILING_IN_CPYTHON
                else if (strcmp(memb->name, "__module__") == 0) {
                    PyObject *descr;
                    assert(memb->type == T_OBJECT);
                    assert(memb->flags == 0 || memb->flags == READONLY);
                    descr = PyDescr_NewMember(type, memb);
                    if (unlikely(!descr))
                        return -1;
                    if (unlikely(PyDict_SetItem(type->tp_dict, PyDescr_NAME(descr), descr) < 0)) {
                        Py_DECREF(descr);
                        return -1;
                    }
                    Py_DECREF(descr);
                    changed = 1;
                }
#endif
            }
            memb++;
        }
        if (changed)
            PyType_Modified(type);
    }
#endif
    return 0;
}
#endif

/* PyObjectCallMethod0 */
static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name) {
    PyObject *method = NULL, *result = NULL;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_CallOneArg(method, obj);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) goto bad;
    result = __Pyx_PyObject_CallNoArg(method);
    Py_DECREF(method);
bad:
    return result;
}

/* ValidateBasesTuple */
#if CYTHON_COMPILING_IN_CPYTHON || CYTHON_COMPILING_IN_LIMITED_API || CYTHON_USE_TYPE_SPECS
static int __Pyx_validate_bases_tuple(const char *type_name, Py_ssize_t dictoffset, PyObject *bases) {
    Py_ssize_t i, n = PyTuple_GET_SIZE(bases);
    for (i = 1; i < n; i++)
    {
        PyObject *b0 = PyTuple_GET_ITEM(bases, i);
        PyTypeObject *b;
#if PY_MAJOR_VERSION < 3
        if (PyClass_Check(b0))
        {
            PyErr_Format(PyExc_TypeError, "base class '%.200s' is an old-style class",
                         PyString_AS_STRING(((PyClassObject*)b0)->cl_name));
            return -1;
        }
#endif
        b = (PyTypeObject*) b0;
        if (!__Pyx_PyType_HasFeature(b, Py_TPFLAGS_HEAPTYPE))
        {
            __Pyx_TypeName b_name = __Pyx_PyType_GetName(b);
            PyErr_Format(PyExc_TypeError,
                "base class '" __Pyx_FMT_TYPENAME "' is not a heap type", b_name);
            __Pyx_DECREF_TypeName(b_name);
            return -1;
        }
        if (dictoffset == 0 && b->tp_dictoffset)
        {
            __Pyx_TypeName b_name = __Pyx_PyType_GetName(b);
            PyErr_Format(PyExc_TypeError,
                "extension type '%.200s' has no __dict__ slot, "
                "but base type '" __Pyx_FMT_TYPENAME "' has: "
                "either add 'cdef dict __dict__' to the extension type "
                "or add '__slots__ = [...]' to the base type",
                type_name, b_name);
            __Pyx_DECREF_TypeName(b_name);
            return -1;
        }
    }
    return 0;
}
#endif

/* PyType_Ready */
static int __Pyx_PyType_Ready(PyTypeObject *t) {
#if CYTHON_USE_TYPE_SPECS || !(CYTHON_COMPILING_IN_CPYTHON || CYTHON_COMPILING_IN_LIMITED_API) || defined(PYSTON_MAJOR_VERSION)
    (void)__Pyx_PyObject_CallMethod0;
#if CYTHON_USE_TYPE_SPECS
    (void)__Pyx_validate_bases_tuple;
#endif
    return PyType_Ready(t);
#else
    int r;
    PyObject *bases = __Pyx_PyType_GetSlot(t, tp_bases, PyObject*);
    if (bases && unlikely(__Pyx_validate_bases_tuple(t->tp_name, t->tp_dictoffset, bases) == -1))
        return -1;
#if PY_VERSION_HEX >= 0x03050000 && !defined(PYSTON_MAJOR_VERSION)
    {
        int gc_was_enabled;
    #if PY_VERSION_HEX >= 0x030A00b1
        gc_was_enabled = PyGC_Disable();
        (void)__Pyx_PyObject_CallMethod0;
    #else
        PyObject *ret, *py_status;
        PyObject *gc = NULL;
        #if PY_VERSION_HEX >= 0x030700a1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM+0 >= 0x07030400)
        gc = PyImport_GetModule(__pyx_kp_u_gc);
        #endif
        if (unlikely(!gc)) gc = PyImport_Import(__pyx_kp_u_gc);
        if (unlikely(!gc)) return -1;
        py_status = __Pyx_PyObject_CallMethod0(gc, __pyx_kp_u_isenabled);
        if (unlikely(!py_status)) {
            Py_DECREF(gc);
            return -1;
        }
        gc_was_enabled = __Pyx_PyObject_IsTrue(py_status);
        Py_DECREF(py_status);
        if (gc_was_enabled > 0) {
            ret = __Pyx_PyObject_CallMethod0(gc, __pyx_kp_u_disable);
            if (unlikely(!ret)) {
                Py_DECREF(gc);
                return -1;
            }
            Py_DECREF(ret);
        } else if (unlikely(gc_was_enabled == -1)) {
            Py_DECREF(gc);
            return -1;
        }
    #endif
        t->tp_flags |= Py_TPFLAGS_HEAPTYPE;
#if PY_VERSION_HEX >= 0x030A0000
        t->tp_flags |= Py_TPFLAGS_IMMUTABLETYPE;
#endif
#else
        (void)__Pyx_PyObject_CallMethod0;
#endif
    r = PyType_Ready(t);
#if PY_VERSION_HEX >= 0x03050000 && !defined(PYSTON_MAJOR_VERSION)
        t->tp_flags &= ~Py_TPFLAGS_HEAPTYPE;
    #if PY_VERSION_HEX >= 0x030A00b1
        if (gc_was_enabled)
            PyGC_Enable();
    #else
        if (gc_was_enabled) {
            PyObject *tp, *v, *tb;
            PyErr_Fetch(&tp, &v, &tb);
            ret = __Pyx_PyObject_CallMethod0(gc, __pyx_kp_u_enable);
            if (likely(ret || r == -1)) {
                Py_XDECREF(ret);
                PyErr_Restore(tp, v, tb);
            } else {
                Py_XDECREF(tp);
                Py_XDECREF(v);
                Py_XDECREF(tb);
                r = -1;
            }
        }
        Py_DECREF(gc);
    #endif
    }
#endif
    return r;
#endif
}

/* PyObject_GenericGetAttrNoDict */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject *__Pyx_RaiseGenericGetAttributeError(PyTypeObject *tp, PyObject *attr_name) {
    __Pyx_TypeName type_name = __Pyx_PyType_GetName(tp);
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'" __Pyx_FMT_TYPENAME "' object has no attribute '%U'",
                 type_name, attr_name);
#else
                 "'" __Pyx_FMT_TYPENAME "' object has no attribute '%.400s'",
                 type_name, PyString_AS_STRING(attr_name));
#endif
    __Pyx_DECREF_TypeName(type_name);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name) {
    PyObject *descr;
    PyTypeObject *tp = Py_TYPE(obj);
    if (unlikely(!PyString_Check(attr_name))) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    assert(!tp->tp_dictoffset);
    descr = _PyType_Lookup(tp, attr_name);
    if (unlikely(!descr)) {
        return __Pyx_RaiseGenericGetAttributeError(tp, attr_name);
    }
    Py_INCREF(descr);
    #if PY_MAJOR_VERSION < 3
    if (likely(PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_HAVE_CLASS)))
    #endif
    {
        descrgetfunc f = Py_TYPE(descr)->tp_descr_get;
        if (unlikely(f)) {
            PyObject *res = f(descr, obj, (PyObject *)tp);
            Py_DECREF(descr);
            return res;
        }
    }
    return descr;
}
#endif

/* PyObject_GenericGetAttr */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name) {
    if (unlikely(Py_TYPE(obj)->tp_dictoffset)) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    return __Pyx_PyObject_GenericGetAttrNoDict(obj, attr_name);
}
#endif

/* SetupReduce */
#if !CYTHON_COMPILING_IN_LIMITED_API
static int __Pyx_setup_reduce_is_named(PyObject* meth, PyObject* name) {
  int ret;
  PyObject *name_attr;
  name_attr = __Pyx_PyObject_GetAttrStrNoError(meth, __pyx_n_s_name_2);
  if (likely(name_attr)) {
      ret = PyObject_RichCompareBool(name_attr, name, Py_EQ);
  } else {
      ret = -1;
  }
  if (unlikely(ret < 0)) {
      PyErr_Clear();
      ret = 0;
  }
  Py_XDECREF(name_attr);
  return ret;
}
static int __Pyx_setup_reduce(PyObject* type_obj) {
    int ret = 0;
    PyObject *object_reduce = NULL;
    PyObject *object_getstate = NULL;
    PyObject *object_reduce_ex = NULL;
    PyObject *reduce = NULL;
    PyObject *reduce_ex = NULL;
    PyObject *reduce_cython = NULL;
    PyObject *setstate = NULL;
    PyObject *setstate_cython = NULL;
    PyObject *getstate = NULL;
#if CYTHON_USE_PYTYPE_LOOKUP
    getstate = _PyType_Lookup((PyTypeObject*)type_obj, __pyx_n_s_getstate);
#else
    getstate = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_getstate);
    if (!getstate && PyErr_Occurred()) {
        goto __PYX_BAD;
    }
#endif
    if (getstate) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_getstate = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_getstate);
#else
        object_getstate = __Pyx_PyObject_GetAttrStrNoError((PyObject*)&PyBaseObject_Type, __pyx_n_s_getstate);
        if (!object_getstate && PyErr_Occurred()) {
            goto __PYX_BAD;
        }
#endif
        if (object_getstate != getstate) {
            goto __PYX_GOOD;
        }
    }
#if CYTHON_USE_PYTYPE_LOOKUP
    object_reduce_ex = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#else
    object_reduce_ex = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#endif
    reduce_ex = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce_ex); if (unlikely(!reduce_ex)) goto __PYX_BAD;
    if (reduce_ex == object_reduce_ex) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_reduce = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto __PYX_BAD;
#else
        object_reduce = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto __PYX_BAD;
#endif
        reduce = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce); if (unlikely(!reduce)) goto __PYX_BAD;
        if (reduce == object_reduce || __Pyx_setup_reduce_is_named(reduce, __pyx_n_s_reduce_cython)) {
            reduce_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_reduce_cython);
            if (likely(reduce_cython)) {
                ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce, reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
            } else if (reduce == object_reduce || PyErr_Occurred()) {
                goto __PYX_BAD;
            }
            setstate = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_setstate);
            if (!setstate) PyErr_Clear();
            if (!setstate || __Pyx_setup_reduce_is_named(setstate, __pyx_n_s_setstate_cython)) {
                setstate_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_setstate_cython);
                if (likely(setstate_cython)) {
                    ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate, setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                    ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                } else if (!setstate || PyErr_Occurred()) {
                    goto __PYX_BAD;
                }
            }
            PyType_Modified((PyTypeObject*)type_obj);
        }
    }
    goto __PYX_GOOD;
__PYX_BAD:
    if (!PyErr_Occurred()) {
        __Pyx_TypeName type_obj_name =
            __Pyx_PyType_GetName((PyTypeObject*)type_obj);
        PyErr_Format(PyExc_RuntimeError,
            "Unable to initialize pickling for " __Pyx_FMT_TYPENAME, type_obj_name);
        __Pyx_DECREF_TypeName(type_obj_name);
    }
    ret = -1;
__PYX_GOOD:
#if !CYTHON_USE_PYTYPE_LOOKUP
    Py_XDECREF(object_reduce);
    Py_XDECREF(object_reduce_ex);
    Py_XDECREF(object_getstate);
    Py_XDECREF(getstate);
#endif
    Py_XDECREF(reduce);
    Py_XDECREF(reduce_ex);
    Py_XDECREF(reduce_cython);
    Py_XDECREF(setstate);
    Py_XDECREF(setstate_cython);
    return ret;
}
#endif

/* TypeImport */
#ifndef __PYX_HAVE_RT_ImportType_3_0_0
#define __PYX_HAVE_RT_ImportType_3_0_0
static PyTypeObject *__Pyx_ImportType_3_0_0(PyObject *module, const char *module_name, const char *class_name,
    size_t size, size_t alignment, enum __Pyx_ImportType_CheckSize_3_0_0 check_size)
{
    PyObject *result = 0;
    char warning[200];
    Py_ssize_t basicsize;
    Py_ssize_t itemsize;
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *py_basicsize;
    PyObject *py_itemsize;
#endif
    result = PyObject_GetAttrString(module, class_name);
    if (!result)
        goto bad;
    if (!PyType_Check(result)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s.%.200s is not a type object",
            module_name, class_name);
        goto bad;
    }
#if !CYTHON_COMPILING_IN_LIMITED_API
    basicsize = ((PyTypeObject *)result)->tp_basicsize;
    itemsize = ((PyTypeObject *)result)->tp_itemsize;
#else
    py_basicsize = PyObject_GetAttrString(result, "__basicsize__");
    if (!py_basicsize)
        goto bad;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = 0;
    if (basicsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
    py_itemsize = PyObject_GetAttrString(result, "__itemsize__");
    if (!py_itemsize)
        goto bad;
    itemsize = PyLong_AsSsize_t(py_itemsize);
    Py_DECREF(py_itemsize);
    py_itemsize = 0;
    if (itemsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
#endif
    if (itemsize) {
        if (size % alignment) {
            alignment = size % alignment;
        }
        if (itemsize < (Py_ssize_t)alignment)
            itemsize = (Py_ssize_t)alignment;
    }
    if ((size_t)(basicsize + itemsize) < size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize+itemsize);
        goto bad;
    }
    if (check_size == __Pyx_ImportType_CheckSize_Error_3_0_0 &&
            ((size_t)basicsize > size || (size_t)(basicsize + itemsize) < size)) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd-%zd from PyObject",
            module_name, class_name, size, basicsize, basicsize+itemsize);
        goto bad;
    }
    else if (check_size == __Pyx_ImportType_CheckSize_Warn_3_0_0 && (size_t)basicsize > size) {
        PyOS_snprintf(warning, sizeof(warning),
            "%s.%s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        if (PyErr_WarnEx(NULL, warning, 0) < 0) goto bad;
    }
    return (PyTypeObject *)result;
bad:
    Py_XDECREF(result);
    return NULL;
}
#endif

/* Import */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level) {
    PyObject *module = 0;
    PyObject *empty_dict = 0;
    PyObject *empty_list = 0;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_import;
    py_import = __Pyx_PyObject_GetAttrStr(__pyx_b, __pyx_n_s_import);
    if (unlikely(!py_import))
        goto bad;
    if (!from_list) {
        empty_list = PyList_New(0);
        if (unlikely(!empty_list))
            goto bad;
        from_list = empty_list;
    }
    #endif
    empty_dict = PyDict_New();
    if (unlikely(!empty_dict))
        goto bad;
    {
        #if PY_MAJOR_VERSION >= 3
        if (level == -1) {
            if ((1) && (strchr(__Pyx_MODULE_NAME, '.'))) {
                #if CYTHON_COMPILING_IN_LIMITED_API
                module = PyImport_ImportModuleLevelObject(
                    name, empty_dict, empty_dict, from_list, 1);
                #else
                module = PyImport_ImportModuleLevelObject(
                    name, __pyx_d, empty_dict, from_list, 1);
                #endif
                if (unlikely(!module)) {
                    if (unlikely(!PyErr_ExceptionMatches(PyExc_ImportError)))
                        goto bad;
                    PyErr_Clear();
                }
            }
            level = 0;
        }
        #endif
        if (!module) {
            #if PY_MAJOR_VERSION < 3
            PyObject *py_level = PyInt_FromLong(level);
            if (unlikely(!py_level))
                goto bad;
            module = PyObject_CallFunctionObjArgs(py_import,
                name, __pyx_d, empty_dict, from_list, py_level, (PyObject *)NULL);
            Py_DECREF(py_level);
            #else
            #if CYTHON_COMPILING_IN_LIMITED_API
            module = PyImport_ImportModuleLevelObject(
                name, empty_dict, empty_dict, from_list, level);
            #else
            module = PyImport_ImportModuleLevelObject(
                name, __pyx_d, empty_dict, from_list, level);
            #endif
            #endif
        }
    }
bad:
    Py_XDECREF(empty_dict);
    Py_XDECREF(empty_list);
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_import);
    #endif
    return module;
}

/* ImportFrom */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name) {
    PyObject* value = __Pyx_PyObject_GetAttrStr(module, name);
    if (unlikely(!value) && PyErr_ExceptionMatches(PyExc_AttributeError)) {
        const char* module_name_str = 0;
        PyObject* module_name = 0;
        PyObject* module_dot = 0;
        PyObject* full_name = 0;
        PyErr_Clear();
        module_name_str = PyModule_GetName(module);
        if (unlikely(!module_name_str)) { goto modbad; }
        module_name = PyUnicode_FromString(module_name_str);
        if (unlikely(!module_name)) { goto modbad; }
        module_dot = PyUnicode_Concat(module_name, __pyx_kp_u__8);
        if (unlikely(!module_dot)) { goto modbad; }
        full_name = PyUnicode_Concat(module_dot, name);
        if (unlikely(!full_name)) { goto modbad; }
        #if PY_VERSION_HEX < 0x030700A1 || (CYTHON_COMPILING_IN_PYPY && PYPY_VERSION_NUM  < 0x07030400)
        {
            PyObject *modules = PyImport_GetModuleDict();
            if (unlikely(!modules))
                goto modbad;
            value = PyObject_GetItem(modules, full_name);
        }
        #else
        value = PyImport_GetModule(full_name);
        #endif
      modbad:
        Py_XDECREF(full_name);
        Py_XDECREF(module_dot);
        Py_XDECREF(module_name);
    }
    if (unlikely(!value)) {
        PyErr_Format(PyExc_ImportError,
        #if PY_MAJOR_VERSION < 3
            "cannot import name %.230s", PyString_AS_STRING(name));
        #else
            "cannot import name %S", name);
        #endif
    }
    return value;
}

/* ImportDottedModule */
#if PY_MAJOR_VERSION >= 3
static PyObject *__Pyx__ImportDottedModule_Error(PyObject *name, PyObject *parts_tuple, Py_ssize_t count) {
    PyObject *partial_name = NULL, *slice = NULL, *sep = NULL;
    if (unlikely(PyErr_Occurred())) {
        PyErr_Clear();
    }
    if (likely(PyTuple_GET_SIZE(parts_tuple) == count)) {
        partial_name = name;
    } else {
        slice = PySequence_GetSlice(parts_tuple, 0, count);
        if (unlikely(!slice))
            goto bad;
        sep = PyUnicode_FromStringAndSize(".", 1);
        if (unlikely(!sep))
            goto bad;
        partial_name = PyUnicode_Join(sep, slice);
    }
    PyErr_Format(
#if PY_MAJOR_VERSION < 3
        PyExc_ImportError,
        "No module named '%s'", PyString_AS_STRING(partial_name));
#else
#if PY_VERSION_HEX >= 0x030600B1
        PyExc_ModuleNotFoundError,
#else
        PyExc_ImportError,
#endif
        "No module named '%U'", partial_name);
#endif
bad:
    Py_XDECREF(sep);
    Py_XDECREF(slice);
    Py_XDECREF(partial_name);
    return NULL;
}
#endif
#if PY_MAJOR_VERSION >= 3
static PyObject *__Pyx__ImportDottedModule_Lookup(PyObject *name) {
    PyObject *imported_module;
#if PY_VERSION_HEX < 0x030700A1 || (CYTHON_COMPILING_IN_PYPY && PYPY_VERSION_NUM  < 0x07030400)
    PyObject *modules = PyImport_GetModuleDict();
    if (unlikely(!modules))
        return NULL;
    imported_module = __Pyx_PyDict_GetItemStr(modules, name);
    Py_XINCREF(imported_module);
#else
    imported_module = PyImport_GetModule(name);
#endif
    return imported_module;
}
#endif
#if PY_MAJOR_VERSION >= 3
static PyObject *__Pyx_ImportDottedModule_WalkParts(PyObject *module, PyObject *name, PyObject *parts_tuple) {
    Py_ssize_t i, nparts;
    nparts = PyTuple_GET_SIZE(parts_tuple);
    for (i=1; i < nparts && module; i++) {
        PyObject *part, *submodule;
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        part = PyTuple_GET_ITEM(parts_tuple, i);
#else
        part = PySequence_ITEM(parts_tuple, i);
#endif
        submodule = __Pyx_PyObject_GetAttrStrNoError(module, part);
#if !(CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
        Py_DECREF(part);
#endif
        Py_DECREF(module);
        module = submodule;
    }
    if (unlikely(!module)) {
        return __Pyx__ImportDottedModule_Error(name, parts_tuple, i);
    }
    return module;
}
#endif
static PyObject *__Pyx__ImportDottedModule(PyObject *name, PyObject *parts_tuple) {
#if PY_MAJOR_VERSION < 3
    PyObject *module, *from_list, *star = __pyx_n_s__9;
    CYTHON_UNUSED_VAR(parts_tuple);
    from_list = PyList_New(1);
    if (unlikely(!from_list))
        return NULL;
    Py_INCREF(star);
    PyList_SET_ITEM(from_list, 0, star);
    module = __Pyx_Import(name, from_list, 0);
    Py_DECREF(from_list);
    return module;
#else
    PyObject *imported_module;
    PyObject *module = __Pyx_Import(name, NULL, 0);
    if (!parts_tuple || unlikely(!module))
        return module;
    imported_module = __Pyx__ImportDottedModule_Lookup(name);
    if (likely(imported_module)) {
        Py_DECREF(module);
        return imported_module;
    }
    PyErr_Clear();
    return __Pyx_ImportDottedModule_WalkParts(module, name, parts_tuple);
#endif
}
static PyObject *__Pyx_ImportDottedModule(PyObject *name, PyObject *parts_tuple) {
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030400B1
    PyObject *module = __Pyx__ImportDottedModule_Lookup(name);
    if (likely(module)) {
        PyObject *spec = __Pyx_PyObject_GetAttrStrNoError(module, __pyx_n_s_spec);
        if (likely(spec)) {
            PyObject *unsafe = __Pyx_PyObject_GetAttrStrNoError(spec, __pyx_n_s_initializing);
            if (likely(!unsafe || !__Pyx_PyObject_IsTrue(unsafe))) {
                Py_DECREF(spec);
                spec = NULL;
            }
            Py_XDECREF(unsafe);
        }
        if (likely(!spec)) {
            PyErr_Clear();
            return module;
        }
        Py_DECREF(spec);
        Py_DECREF(module);
    } else if (PyErr_Occurred()) {
        PyErr_Clear();
    }
#endif
    return __Pyx__ImportDottedModule(name, parts_tuple);
}

/* ImportDottedModuleRelFirst */
static PyObject *__Pyx_ImportDottedModuleRelFirst(PyObject *name, PyObject *parts_tuple) {
    PyObject *module;
    PyObject *from_list = NULL;
#if PY_MAJOR_VERSION < 3
    PyObject *star = __pyx_n_s__9;
    from_list = PyList_New(1);
    if (unlikely(!from_list))
        return NULL;
    Py_INCREF(star);
    PyList_SET_ITEM(from_list, 0, star);
#endif
    module = __Pyx_Import(name, from_list, -1);
    Py_XDECREF(from_list);
    if (module) {
        #if PY_MAJOR_VERSION >= 3
        if (parts_tuple) {
            module = __Pyx_ImportDottedModule_WalkParts(module, name, parts_tuple);
        }
        #endif
        return module;
    }
    if (unlikely(!PyErr_ExceptionMatches(PyExc_ImportError)))
        return NULL;
    PyErr_Clear();
    return __Pyx_ImportDottedModule(name, parts_tuple);
}

/* FetchSharedCythonModule */
static PyObject *__Pyx_FetchSharedCythonABIModule(void) {
    PyObject *abi_module = PyImport_AddModule((char*) __PYX_ABI_MODULE_NAME);
    if (unlikely(!abi_module)) return NULL;
    Py_INCREF(abi_module);
    return abi_module;
}

/* FetchCommonType */
static int __Pyx_VerifyCachedType(PyObject *cached_type,
                               const char *name,
                               Py_ssize_t basicsize,
                               Py_ssize_t expected_basicsize) {
    if (!PyType_Check(cached_type)) {
        PyErr_Format(PyExc_TypeError,
            "Shared Cython type %.200s is not a type object", name);
        return -1;
    }
    if (basicsize != expected_basicsize) {
        PyErr_Format(PyExc_TypeError,
            "Shared Cython type %.200s has the wrong size, try recompiling",
            name);
        return -1;
    }
    return 0;
}
#if !CYTHON_USE_TYPE_SPECS
static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type) {
    PyObject* abi_module;
    const char* object_name;
    PyTypeObject *cached_type = NULL;
    abi_module = __Pyx_FetchSharedCythonABIModule();
    if (!abi_module) return NULL;
    object_name = strrchr(type->tp_name, '.');
    object_name = object_name ? object_name+1 : type->tp_name;
    cached_type = (PyTypeObject*) PyObject_GetAttrString(abi_module, object_name);
    if (cached_type) {
        if (__Pyx_VerifyCachedType(
              (PyObject *)cached_type,
              object_name,
              cached_type->tp_basicsize,
              type->tp_basicsize) < 0) {
            goto bad;
        }
        goto done;
    }
    if (!PyErr_ExceptionMatches(PyExc_AttributeError)) goto bad;
    PyErr_Clear();
    if (PyType_Ready(type) < 0) goto bad;
    if (PyObject_SetAttrString(abi_module, object_name, (PyObject *)type) < 0)
        goto bad;
    Py_INCREF(type);
    cached_type = type;
done:
    Py_DECREF(abi_module);
    return cached_type;
bad:
    Py_XDECREF(cached_type);
    cached_type = NULL;
    goto done;
}
#else
static PyTypeObject *__Pyx_FetchCommonTypeFromSpec(PyObject *module, PyType_Spec *spec, PyObject *bases) {
    PyObject *abi_module, *cached_type = NULL;
    const char* object_name = strrchr(spec->name, '.');
    object_name = object_name ? object_name+1 : spec->name;
    abi_module = __Pyx_FetchSharedCythonABIModule();
    if (!abi_module) return NULL;
    cached_type = PyObject_GetAttrString(abi_module, object_name);
    if (cached_type) {
        Py_ssize_t basicsize;
#if CYTHON_COMPILING_IN_LIMITED_API
        PyObject *py_basicsize;
        py_basicsize = PyObject_GetAttrString(cached_type, "__basicsize__");
        if (unlikely(!py_basicsize)) goto bad;
        basicsize = PyLong_AsSsize_t(py_basicsize);
        Py_DECREF(py_basicsize);
        py_basicsize = 0;
        if (unlikely(basicsize == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
#else
        basicsize = likely(PyType_Check(cached_type)) ? ((PyTypeObject*) cached_type)->tp_basicsize : -1;
#endif
        if (__Pyx_VerifyCachedType(
              cached_type,
              object_name,
              basicsize,
              spec->basicsize) < 0) {
            goto bad;
        }
        goto done;
    }
    if (!PyErr_ExceptionMatches(PyExc_AttributeError)) goto bad;
    PyErr_Clear();
    CYTHON_UNUSED_VAR(module);
    cached_type = __Pyx_PyType_FromModuleAndSpec(abi_module, spec, bases);
    if (unlikely(!cached_type)) goto bad;
    if (unlikely(__Pyx_fix_up_extension_type_from_spec(spec, (PyTypeObject *) cached_type) < 0)) goto bad;
    if (PyObject_SetAttrString(abi_module, object_name, cached_type) < 0) goto bad;
done:
    Py_DECREF(abi_module);
    assert(cached_type == NULL || PyType_Check(cached_type));
    return (PyTypeObject *) cached_type;
bad:
    Py_XDECREF(cached_type);
    cached_type = NULL;
    goto done;
}
#endif

/* PyVectorcallFastCallDict */
#if CYTHON_METH_FASTCALL
static PyObject *__Pyx_PyVectorcall_FastCallDict_kw(PyObject *func, __pyx_vectorcallfunc vc, PyObject *const *args, size_t nargs, PyObject *kw)
{
    PyObject *res = NULL;
    PyObject *kwnames;
    PyObject **newargs;
    PyObject **kwvalues;
    Py_ssize_t i, pos;
    size_t j;
    PyObject *key, *value;
    unsigned long keys_are_strings;
    Py_ssize_t nkw = PyDict_GET_SIZE(kw);
    newargs = (PyObject **)PyMem_Malloc((nargs + (size_t)nkw) * sizeof(args[0]));
    if (unlikely(newargs == NULL)) {
        PyErr_NoMemory();
        return NULL;
    }
    for (j = 0; j < nargs; j++) newargs[j] = args[j];
    kwnames = PyTuple_New(nkw);
    if (unlikely(kwnames == NULL)) {
        PyMem_Free(newargs);
        return NULL;
    }
    kwvalues = newargs + nargs;
    pos = i = 0;
    keys_are_strings = Py_TPFLAGS_UNICODE_SUBCLASS;
    while (PyDict_Next(kw, &pos, &key, &value)) {
        keys_are_strings &= Py_TYPE(key)->tp_flags;
        Py_INCREF(key);
        Py_INCREF(value);
        PyTuple_SET_ITEM(kwnames, i, key);
        kwvalues[i] = value;
        i++;
    }
    if (unlikely(!keys_are_strings)) {
        PyErr_SetString(PyExc_TypeError, "keywords must be strings");
        goto cleanup;
    }
    res = vc(func, newargs, nargs, kwnames);
cleanup:
    Py_DECREF(kwnames);
    for (i = 0; i < nkw; i++)
        Py_DECREF(kwvalues[i]);
    PyMem_Free(newargs);
    return res;
}
static CYTHON_INLINE PyObject *__Pyx_PyVectorcall_FastCallDict(PyObject *func, __pyx_vectorcallfunc vc, PyObject *const *args, size_t nargs, PyObject *kw)
{
    if (likely(kw == NULL) || PyDict_GET_SIZE(kw) == 0) {
        return vc(func, args, nargs, NULL);
    }
    return __Pyx_PyVectorcall_FastCallDict_kw(func, vc, args, nargs, kw);
}
#endif

/* CythonFunctionShared */
static CYTHON_INLINE void __Pyx__CyFunction_SetClassObj(__pyx_CyFunctionObject* f, PyObject* classobj) {
#if PY_VERSION_HEX < 0x030900B1
    __Pyx_Py_XDECREF_SET(
        __Pyx_CyFunction_GetClassObj(f),
            ((classobj) ? __Pyx_NewRef(classobj) : NULL));
#else
    __Pyx_Py_XDECREF_SET(
        ((PyCMethodObject *) (f))->mm_class,
        (PyTypeObject*)((classobj) ? __Pyx_NewRef(classobj) : NULL));
#endif
}
static PyObject *
__Pyx_CyFunction_get_doc(__pyx_CyFunctionObject *op, void *closure)
{
    CYTHON_UNUSED_VAR(closure);
    if (unlikely(op->func_doc == NULL)) {
        if (((PyCFunctionObject*)op)->m_ml->ml_doc) {
#if PY_MAJOR_VERSION >= 3
            op->func_doc = PyUnicode_FromString(((PyCFunctionObject*)op)->m_ml->ml_doc);
#else
            op->func_doc = PyString_FromString(((PyCFunctionObject*)op)->m_ml->ml_doc);
#endif
            if (unlikely(op->func_doc == NULL))
                return NULL;
        } else {
            Py_INCREF(Py_None);
            return Py_None;
        }
    }
    Py_INCREF(op->func_doc);
    return op->func_doc;
}
static int
__Pyx_CyFunction_set_doc(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (value == NULL) {
        value = Py_None;
    }
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->func_doc, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_name(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (unlikely(op->func_name == NULL)) {
#if PY_MAJOR_VERSION >= 3
        op->func_name = PyUnicode_InternFromString(((PyCFunctionObject*)op)->m_ml->ml_name);
#else
        op->func_name = PyString_InternFromString(((PyCFunctionObject*)op)->m_ml->ml_name);
#endif
        if (unlikely(op->func_name == NULL))
            return NULL;
    }
    Py_INCREF(op->func_name);
    return op->func_name;
}
static int
__Pyx_CyFunction_set_name(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->func_name, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_qualname(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(context);
    Py_INCREF(op->func_qualname);
    return op->func_qualname;
}
static int
__Pyx_CyFunction_set_qualname(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->func_qualname, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_dict(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (unlikely(op->func_dict == NULL)) {
        op->func_dict = PyDict_New();
        if (unlikely(op->func_dict == NULL))
            return NULL;
    }
    Py_INCREF(op->func_dict);
    return op->func_dict;
}
static int
__Pyx_CyFunction_set_dict(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (unlikely(value == NULL)) {
        PyErr_SetString(PyExc_TypeError,
               "function's dictionary may not be deleted");
        return -1;
    }
    if (unlikely(!PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
               "setting function's dictionary to a non-dict");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->func_dict, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_globals(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(context);
    Py_INCREF(op->func_globals);
    return op->func_globals;
}
static PyObject *
__Pyx_CyFunction_get_closure(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(op);
    CYTHON_UNUSED_VAR(context);
    Py_INCREF(Py_None);
    return Py_None;
}
static PyObject *
__Pyx_CyFunction_get_code(__pyx_CyFunctionObject *op, void *context)
{
    PyObject* result = (op->func_code) ? op->func_code : Py_None;
    CYTHON_UNUSED_VAR(context);
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_init_defaults(__pyx_CyFunctionObject *op) {
    int result = 0;
    PyObject *res = op->defaults_getter((PyObject *) op);
    if (unlikely(!res))
        return -1;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    op->defaults_tuple = PyTuple_GET_ITEM(res, 0);
    Py_INCREF(op->defaults_tuple);
    op->defaults_kwdict = PyTuple_GET_ITEM(res, 1);
    Py_INCREF(op->defaults_kwdict);
    #else
    op->defaults_tuple = PySequence_ITEM(res, 0);
    if (unlikely(!op->defaults_tuple)) result = -1;
    else {
        op->defaults_kwdict = PySequence_ITEM(res, 1);
        if (unlikely(!op->defaults_kwdict)) result = -1;
    }
    #endif
    Py_DECREF(res);
    return result;
}
static int
__Pyx_CyFunction_set_defaults(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    if (!value) {
        value = Py_None;
    } else if (unlikely(value != Py_None && !PyTuple_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__defaults__ must be set to a tuple object");
        return -1;
    }
    PyErr_WarnEx(PyExc_RuntimeWarning, "changes to cyfunction.__defaults__ will not "
                 "currently affect the values used in function calls", 1);
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->defaults_tuple, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_defaults(__pyx_CyFunctionObject *op, void *context) {
    PyObject* result = op->defaults_tuple;
    CYTHON_UNUSED_VAR(context);
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (unlikely(__Pyx_CyFunction_init_defaults(op) < 0)) return NULL;
            result = op->defaults_tuple;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_kwdefaults(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    if (!value) {
        value = Py_None;
    } else if (unlikely(value != Py_None && !PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__kwdefaults__ must be set to a dict object");
        return -1;
    }
    PyErr_WarnEx(PyExc_RuntimeWarning, "changes to cyfunction.__kwdefaults__ will not "
                 "currently affect the values used in function calls", 1);
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->defaults_kwdict, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_kwdefaults(__pyx_CyFunctionObject *op, void *context) {
    PyObject* result = op->defaults_kwdict;
    CYTHON_UNUSED_VAR(context);
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (unlikely(__Pyx_CyFunction_init_defaults(op) < 0)) return NULL;
            result = op->defaults_kwdict;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_annotations(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    if (!value || value == Py_None) {
        value = NULL;
    } else if (unlikely(!PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__annotations__ must be set to a dict object");
        return -1;
    }
    Py_XINCREF(value);
    __Pyx_Py_XDECREF_SET(op->func_annotations, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_annotations(__pyx_CyFunctionObject *op, void *context) {
    PyObject* result = op->func_annotations;
    CYTHON_UNUSED_VAR(context);
    if (unlikely(!result)) {
        result = PyDict_New();
        if (unlikely(!result)) return NULL;
        op->func_annotations = result;
    }
    Py_INCREF(result);
    return result;
}
static PyObject *
__Pyx_CyFunction_get_is_coroutine(__pyx_CyFunctionObject *op, void *context) {
    int is_coroutine;
    CYTHON_UNUSED_VAR(context);
    if (op->func_is_coroutine) {
        return __Pyx_NewRef(op->func_is_coroutine);
    }
    is_coroutine = op->flags & __Pyx_CYFUNCTION_COROUTINE;
#if PY_VERSION_HEX >= 0x03050000
    if (is_coroutine) {
        PyObject *module, *fromlist, *marker = __pyx_n_s_is_coroutine;
        fromlist = PyList_New(1);
        if (unlikely(!fromlist)) return NULL;
        Py_INCREF(marker);
        PyList_SET_ITEM(fromlist, 0, marker);
        module = PyImport_ImportModuleLevelObject(__pyx_n_s_asyncio_coroutines, NULL, NULL, fromlist, 0);
        Py_DECREF(fromlist);
        if (unlikely(!module)) goto ignore;
        op->func_is_coroutine = __Pyx_PyObject_GetAttrStr(module, marker);
        Py_DECREF(module);
        if (likely(op->func_is_coroutine)) {
            return __Pyx_NewRef(op->func_is_coroutine);
        }
ignore:
        PyErr_Clear();
    }
#endif
    op->func_is_coroutine = __Pyx_PyBool_FromLong(is_coroutine);
    return __Pyx_NewRef(op->func_is_coroutine);
}
static PyGetSetDef __pyx_CyFunction_getsets[] = {
    {(char *) "func_doc", (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "__doc__",  (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "func_name", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__name__", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__qualname__", (getter)__Pyx_CyFunction_get_qualname, (setter)__Pyx_CyFunction_set_qualname, 0, 0},
    {(char *) "func_dict", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "__dict__", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "func_globals", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "__globals__", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "func_closure", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "__closure__", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "func_code", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "__code__", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "func_defaults", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__defaults__", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__kwdefaults__", (getter)__Pyx_CyFunction_get_kwdefaults, (setter)__Pyx_CyFunction_set_kwdefaults, 0, 0},
    {(char *) "__annotations__", (getter)__Pyx_CyFunction_get_annotations, (setter)__Pyx_CyFunction_set_annotations, 0, 0},
    {(char *) "_is_coroutine", (getter)__Pyx_CyFunction_get_is_coroutine, 0, 0, 0},
    {0, 0, 0, 0, 0}
};
static PyMemberDef __pyx_CyFunction_members[] = {
    {(char *) "__module__", T_OBJECT, offsetof(PyCFunctionObject, m_module), 0, 0},
#if CYTHON_USE_TYPE_SPECS
    {(char *) "__dictoffset__", T_PYSSIZET, offsetof(__pyx_CyFunctionObject, func_dict), READONLY, 0},
#if CYTHON_METH_FASTCALL
#if CYTHON_BACKPORT_VECTORCALL
    {(char *) "__vectorcalloffset__", T_PYSSIZET, offsetof(__pyx_CyFunctionObject, func_vectorcall), READONLY, 0},
#else
    {(char *) "__vectorcalloffset__", T_PYSSIZET, offsetof(PyCFunctionObject, vectorcall), READONLY, 0},
#endif
#endif
#if PY_VERSION_HEX < 0x030500A0
    {(char *) "__weaklistoffset__", T_PYSSIZET, offsetof(__pyx_CyFunctionObject, func_weakreflist), READONLY, 0},
#else
    {(char *) "__weaklistoffset__", T_PYSSIZET, offsetof(PyCFunctionObject, m_weakreflist), READONLY, 0},
#endif
#endif
    {0, 0, 0,  0, 0}
};
static PyObject *
__Pyx_CyFunction_reduce(__pyx_CyFunctionObject *m, PyObject *args)
{
    CYTHON_UNUSED_VAR(args);
#if PY_MAJOR_VERSION >= 3
    Py_INCREF(m->func_qualname);
    return m->func_qualname;
#else
    return PyString_FromString(((PyCFunctionObject*)m)->m_ml->ml_name);
#endif
}
static PyMethodDef __pyx_CyFunction_methods[] = {
    {"__reduce__", (PyCFunction)__Pyx_CyFunction_reduce, METH_VARARGS, 0},
    {0, 0, 0, 0}
};
#if PY_VERSION_HEX < 0x030500A0
#define __Pyx_CyFunction_weakreflist(cyfunc) ((cyfunc)->func_weakreflist)
#else
#define __Pyx_CyFunction_weakreflist(cyfunc) (((PyCFunctionObject*)cyfunc)->m_weakreflist)
#endif
static PyObject *__Pyx_CyFunction_Init(__pyx_CyFunctionObject *op, PyMethodDef *ml, int flags, PyObject* qualname,
                                       PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
    PyCFunctionObject *cf = (PyCFunctionObject*) op;
    if (unlikely(op == NULL))
        return NULL;
    op->flags = flags;
    __Pyx_CyFunction_weakreflist(op) = NULL;
    cf->m_ml = ml;
    cf->m_self = (PyObject *) op;
    Py_XINCREF(closure);
    op->func_closure = closure;
    Py_XINCREF(module);
    cf->m_module = module;
    op->func_dict = NULL;
    op->func_name = NULL;
    Py_INCREF(qualname);
    op->func_qualname = qualname;
    op->func_doc = NULL;
#if PY_VERSION_HEX < 0x030900B1
    op->func_classobj = NULL;
#else
    ((PyCMethodObject*)op)->mm_class = NULL;
#endif
    op->func_globals = globals;
    Py_INCREF(op->func_globals);
    Py_XINCREF(code);
    op->func_code = code;
    op->defaults_pyobjects = 0;
    op->defaults_size = 0;
    op->defaults = NULL;
    op->defaults_tuple = NULL;
    op->defaults_kwdict = NULL;
    op->defaults_getter = NULL;
    op->func_annotations = NULL;
    op->func_is_coroutine = NULL;
#if CYTHON_METH_FASTCALL
    switch (ml->ml_flags & (METH_VARARGS | METH_FASTCALL | METH_NOARGS | METH_O | METH_KEYWORDS | METH_METHOD)) {
    case METH_NOARGS:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_NOARGS;
        break;
    case METH_O:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_O;
        break;
    case METH_METHOD | METH_FASTCALL | METH_KEYWORDS:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS_METHOD;
        break;
    case METH_FASTCALL | METH_KEYWORDS:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS;
        break;
    case METH_VARARGS | METH_KEYWORDS:
        __Pyx_CyFunction_func_vectorcall(op) = NULL;
        break;
    default:
        PyErr_SetString(PyExc_SystemError, "Bad call flags for CyFunction");
        Py_DECREF(op);
        return NULL;
    }
#endif
    return (PyObject *) op;
}
static int
__Pyx_CyFunction_clear(__pyx_CyFunctionObject *m)
{
    Py_CLEAR(m->func_closure);
    Py_CLEAR(((PyCFunctionObject*)m)->m_module);
    Py_CLEAR(m->func_dict);
    Py_CLEAR(m->func_name);
    Py_CLEAR(m->func_qualname);
    Py_CLEAR(m->func_doc);
    Py_CLEAR(m->func_globals);
    Py_CLEAR(m->func_code);
#if PY_VERSION_HEX < 0x030900B1
    Py_CLEAR(__Pyx_CyFunction_GetClassObj(m));
#else
    {
        PyObject *cls = (PyObject*) ((PyCMethodObject *) (m))->mm_class;
        ((PyCMethodObject *) (m))->mm_class = NULL;
        Py_XDECREF(cls);
    }
#endif
    Py_CLEAR(m->defaults_tuple);
    Py_CLEAR(m->defaults_kwdict);
    Py_CLEAR(m->func_annotations);
    Py_CLEAR(m->func_is_coroutine);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_XDECREF(pydefaults[i]);
        PyObject_Free(m->defaults);
        m->defaults = NULL;
    }
    return 0;
}
static void __Pyx__CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    if (__Pyx_CyFunction_weakreflist(m) != NULL)
        PyObject_ClearWeakRefs((PyObject *) m);
    __Pyx_CyFunction_clear(m);
    __Pyx_PyHeapTypeObject_GC_Del(m);
}
static void __Pyx_CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    PyObject_GC_UnTrack(m);
    __Pyx__CyFunction_dealloc(m);
}
static int __Pyx_CyFunction_traverse(__pyx_CyFunctionObject *m, visitproc visit, void *arg)
{
    Py_VISIT(m->func_closure);
    Py_VISIT(((PyCFunctionObject*)m)->m_module);
    Py_VISIT(m->func_dict);
    Py_VISIT(m->func_name);
    Py_VISIT(m->func_qualname);
    Py_VISIT(m->func_doc);
    Py_VISIT(m->func_globals);
    Py_VISIT(m->func_code);
    Py_VISIT(__Pyx_CyFunction_GetClassObj(m));
    Py_VISIT(m->defaults_tuple);
    Py_VISIT(m->defaults_kwdict);
    Py_VISIT(m->func_is_coroutine);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_VISIT(pydefaults[i]);
    }
    return 0;
}
static PyObject*
__Pyx_CyFunction_repr(__pyx_CyFunctionObject *op)
{
#if PY_MAJOR_VERSION >= 3
    return PyUnicode_FromFormat("<cyfunction %U at %p>",
                                op->func_qualname, (void *)op);
#else
    return PyString_FromFormat("<cyfunction %s at %p>",
                               PyString_AsString(op->func_qualname), (void *)op);
#endif
}
static PyObject * __Pyx_CyFunction_CallMethod(PyObject *func, PyObject *self, PyObject *arg, PyObject *kw) {
    PyCFunctionObject* f = (PyCFunctionObject*)func;
    PyCFunction meth = f->m_ml->ml_meth;
    Py_ssize_t size;
    switch (f->m_ml->ml_flags & (METH_VARARGS | METH_KEYWORDS | METH_NOARGS | METH_O)) {
    case METH_VARARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0))
            return (*meth)(self, arg);
        break;
    case METH_VARARGS | METH_KEYWORDS:
        return (*(PyCFunctionWithKeywords)(void*)meth)(self, arg, kw);
    case METH_NOARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
            size = PyTuple_GET_SIZE(arg);
            if (likely(size == 0))
                return (*meth)(self, NULL);
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes no arguments (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
            return NULL;
        }
        break;
    case METH_O:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
            size = PyTuple_GET_SIZE(arg);
            if (likely(size == 1)) {
                PyObject *result, *arg0;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                arg0 = PyTuple_GET_ITEM(arg, 0);
                #else
                arg0 = PySequence_ITEM(arg, 0); if (unlikely(!arg0)) return NULL;
                #endif
                result = (*meth)(self, arg0);
                #if !(CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
                Py_DECREF(arg0);
                #endif
                return result;
            }
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes exactly one argument (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
            return NULL;
        }
        break;
    default:
        PyErr_SetString(PyExc_SystemError, "Bad call flags for CyFunction");
        return NULL;
    }
    PyErr_Format(PyExc_TypeError, "%.200s() takes no keyword arguments",
                 f->m_ml->ml_name);
    return NULL;
}
static CYTHON_INLINE PyObject *__Pyx_CyFunction_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    return __Pyx_CyFunction_CallMethod(func, ((PyCFunctionObject*)func)->m_self, arg, kw);
}
static PyObject *__Pyx_CyFunction_CallAsMethod(PyObject *func, PyObject *args, PyObject *kw) {
    PyObject *result;
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *) func;
#if CYTHON_METH_FASTCALL
     __pyx_vectorcallfunc vc = __Pyx_CyFunction_func_vectorcall(cyfunc);
    if (vc) {
#if CYTHON_ASSUME_SAFE_MACROS
        return __Pyx_PyVectorcall_FastCallDict(func, vc, &PyTuple_GET_ITEM(args, 0), (size_t)PyTuple_GET_SIZE(args), kw);
#else
        (void) &__Pyx_PyVectorcall_FastCallDict;
        return PyVectorcall_Call(func, args, kw);
#endif
    }
#endif
    if ((cyfunc->flags & __Pyx_CYFUNCTION_CCLASS) && !(cyfunc->flags & __Pyx_CYFUNCTION_STATICMETHOD)) {
        Py_ssize_t argc;
        PyObject *new_args;
        PyObject *self;
        argc = PyTuple_GET_SIZE(args);
        new_args = PyTuple_GetSlice(args, 1, argc);
        if (unlikely(!new_args))
            return NULL;
        self = PyTuple_GetItem(args, 0);
        if (unlikely(!self)) {
            Py_DECREF(new_args);
#if PY_MAJOR_VERSION > 2
            PyErr_Format(PyExc_TypeError,
                         "unbound method %.200S() needs an argument",
                         cyfunc->func_qualname);
#else
            PyErr_SetString(PyExc_TypeError,
                            "unbound method needs an argument");
#endif
            return NULL;
        }
        result = __Pyx_CyFunction_CallMethod(func, self, new_args, kw);
        Py_DECREF(new_args);
    } else {
        result = __Pyx_CyFunction_Call(func, args, kw);
    }
    return result;
}
#if CYTHON_METH_FASTCALL
static CYTHON_INLINE int __Pyx_CyFunction_Vectorcall_CheckArgs(__pyx_CyFunctionObject *cyfunc, Py_ssize_t nargs, PyObject *kwnames)
{
    int ret = 0;
    if ((cyfunc->flags & __Pyx_CYFUNCTION_CCLASS) && !(cyfunc->flags & __Pyx_CYFUNCTION_STATICMETHOD)) {
        if (unlikely(nargs < 1)) {
            PyErr_Format(PyExc_TypeError, "%.200s() needs an argument",
                         ((PyCFunctionObject*)cyfunc)->m_ml->ml_name);
            return -1;
        }
        ret = 1;
    }
    if (unlikely(kwnames) && unlikely(PyTuple_GET_SIZE(kwnames))) {
        PyErr_Format(PyExc_TypeError,
                     "%.200s() takes no keyword arguments", ((PyCFunctionObject*)cyfunc)->m_ml->ml_name);
        return -1;
    }
    return ret;
}
static PyObject * __Pyx_CyFunction_Vectorcall_NOARGS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    PyMethodDef* def = ((PyCFunctionObject*)cyfunc)->m_ml;
#if CYTHON_BACKPORT_VECTORCALL
    Py_ssize_t nargs = (Py_ssize_t)nargsf;
#else
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
#endif
    PyObject *self;
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, kwnames)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
        self = ((PyCFunctionObject*)cyfunc)->m_self;
        break;
    default:
        return NULL;
    }
    if (unlikely(nargs != 0)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s() takes no arguments (%" CYTHON_FORMAT_SSIZE_T "d given)",
            def->ml_name, nargs);
        return NULL;
    }
    return def->ml_meth(self, NULL);
}
static PyObject * __Pyx_CyFunction_Vectorcall_O(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    PyMethodDef* def = ((PyCFunctionObject*)cyfunc)->m_ml;
#if CYTHON_BACKPORT_VECTORCALL
    Py_ssize_t nargs = (Py_ssize_t)nargsf;
#else
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
#endif
    PyObject *self;
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, kwnames)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
        self = ((PyCFunctionObject*)cyfunc)->m_self;
        break;
    default:
        return NULL;
    }
    if (unlikely(nargs != 1)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s() takes exactly one argument (%" CYTHON_FORMAT_SSIZE_T "d given)",
            def->ml_name, nargs);
        return NULL;
    }
    return def->ml_meth(self, args[0]);
}
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    PyMethodDef* def = ((PyCFunctionObject*)cyfunc)->m_ml;
#if CYTHON_BACKPORT_VECTORCALL
    Py_ssize_t nargs = (Py_ssize_t)nargsf;
#else
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
#endif
    PyObject *self;
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, NULL)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
        self = ((PyCFunctionObject*)cyfunc)->m_self;
        break;
    default:
        return NULL;
    }
    return ((_PyCFunctionFastWithKeywords)(void(*)(void))def->ml_meth)(self, args, nargs, kwnames);
}
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS_METHOD(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    PyMethodDef* def = ((PyCFunctionObject*)cyfunc)->m_ml;
    PyTypeObject *cls = (PyTypeObject *) __Pyx_CyFunction_GetClassObj(cyfunc);
#if CYTHON_BACKPORT_VECTORCALL
    Py_ssize_t nargs = (Py_ssize_t)nargsf;
#else
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
#endif
    PyObject *self;
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, NULL)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
        self = ((PyCFunctionObject*)cyfunc)->m_self;
        break;
    default:
        return NULL;
    }
    return ((__Pyx_PyCMethod)(void(*)(void))def->ml_meth)(self, cls, args, (size_t)nargs, kwnames);
}
#endif
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_CyFunctionType_slots[] = {
    {Py_tp_dealloc, (void *)__Pyx_CyFunction_dealloc},
    {Py_tp_repr, (void *)__Pyx_CyFunction_repr},
    {Py_tp_call, (void *)__Pyx_CyFunction_CallAsMethod},
    {Py_tp_traverse, (void *)__Pyx_CyFunction_traverse},
    {Py_tp_clear, (void *)__Pyx_CyFunction_clear},
    {Py_tp_methods, (void *)__pyx_CyFunction_methods},
    {Py_tp_members, (void *)__pyx_CyFunction_members},
    {Py_tp_getset, (void *)__pyx_CyFunction_getsets},
    {Py_tp_descr_get, (void *)__Pyx_PyMethod_New},
    {0, 0},
};
static PyType_Spec __pyx_CyFunctionType_spec = {
    __PYX_TYPE_MODULE_PREFIX "cython_function_or_method",
    sizeof(__pyx_CyFunctionObject),
    0,
#ifdef Py_TPFLAGS_METHOD_DESCRIPTOR
    Py_TPFLAGS_METHOD_DESCRIPTOR |
#endif
#if (defined(_Py_TPFLAGS_HAVE_VECTORCALL) && CYTHON_METH_FASTCALL)
    _Py_TPFLAGS_HAVE_VECTORCALL |
#endif
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_BASETYPE,
    __pyx_CyFunctionType_slots
};
#else
static PyTypeObject __pyx_CyFunctionType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    __PYX_TYPE_MODULE_PREFIX "cython_function_or_method",
    sizeof(__pyx_CyFunctionObject),
    0,
    (destructor) __Pyx_CyFunction_dealloc,
#if !CYTHON_METH_FASTCALL
    0,
#elif CYTHON_BACKPORT_VECTORCALL
    (printfunc)offsetof(__pyx_CyFunctionObject, func_vectorcall),
#else
    offsetof(PyCFunctionObject, vectorcall),
#endif
    0,
    0,
#if PY_MAJOR_VERSION < 3
    0,
#else
    0,
#endif
    (reprfunc) __Pyx_CyFunction_repr,
    0,
    0,
    0,
    0,
    __Pyx_CyFunction_CallAsMethod,
    0,
    0,
    0,
    0,
#ifdef Py_TPFLAGS_METHOD_DESCRIPTOR
    Py_TPFLAGS_METHOD_DESCRIPTOR |
#endif
#ifdef _Py_TPFLAGS_HAVE_VECTORCALL
    _Py_TPFLAGS_HAVE_VECTORCALL |
#endif
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_BASETYPE,
    0,
    (traverseproc) __Pyx_CyFunction_traverse,
    (inquiry) __Pyx_CyFunction_clear,
    0,
#if PY_VERSION_HEX < 0x030500A0
    offsetof(__pyx_CyFunctionObject, func_weakreflist),
#else
    offsetof(PyCFunctionObject, m_weakreflist),
#endif
    0,
    0,
    __pyx_CyFunction_methods,
    __pyx_CyFunction_members,
    __pyx_CyFunction_getsets,
    0,
    0,
    __Pyx_PyMethod_New,
    0,
    offsetof(__pyx_CyFunctionObject, func_dict),
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if PY_VERSION_HEX >= 0x030400a1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
    0,
#endif
#if __PYX_NEED_TP_PRINT_SLOT
    0,
#endif
#if PY_VERSION_HEX >= 0x030C0000
    0,
#endif
#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
    0,
#endif
};
#endif
static int __pyx_CyFunction_init(PyObject *module) {
#if CYTHON_USE_TYPE_SPECS
    __pyx_CyFunctionType = __Pyx_FetchCommonTypeFromSpec(module, &__pyx_CyFunctionType_spec, NULL);
#else
    CYTHON_UNUSED_VAR(module);
    __pyx_CyFunctionType = __Pyx_FetchCommonType(&__pyx_CyFunctionType_type);
#endif
    if (unlikely(__pyx_CyFunctionType == NULL)) {
        return -1;
    }
    return 0;
}
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *func, size_t size, int pyobjects) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults = PyObject_Malloc(size);
    if (unlikely(!m->defaults))
        return PyErr_NoMemory();
    memset(m->defaults, 0, size);
    m->defaults_pyobjects = pyobjects;
    m->defaults_size = size;
    return m->defaults;
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *func, PyObject *tuple) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_tuple = tuple;
    Py_INCREF(tuple);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_kwdict = dict;
    Py_INCREF(dict);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->func_annotations = dict;
    Py_INCREF(dict);
}

/* CythonFunction */
static PyObject *__Pyx_CyFunction_New(PyMethodDef *ml, int flags, PyObject* qualname,
                                      PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
    PyObject *op = __Pyx_CyFunction_Init(
        PyObject_GC_New(__pyx_CyFunctionObject, __pyx_CyFunctionType),
        ml, flags, qualname, closure, module, globals, code
    );
    if (likely(op)) {
        PyObject_GC_Track(op);
    }
    return op;
}

/* CalculateMetaclass */
static PyObject *__Pyx_CalculateMetaclass(PyTypeObject *metaclass, PyObject *bases) {
    Py_ssize_t i, nbases = PyTuple_GET_SIZE(bases);
    for (i=0; i < nbases; i++) {
        PyTypeObject *tmptype;
        PyObject *tmp = PyTuple_GET_ITEM(bases, i);
        tmptype = Py_TYPE(tmp);
#if PY_MAJOR_VERSION < 3
        if (tmptype == &PyClass_Type)
            continue;
#endif
        if (!metaclass) {
            metaclass = tmptype;
            continue;
        }
        if (PyType_IsSubtype(metaclass, tmptype))
            continue;
        if (PyType_IsSubtype(tmptype, metaclass)) {
            metaclass = tmptype;
            continue;
        }
        PyErr_SetString(PyExc_TypeError,
                        "metaclass conflict: "
                        "the metaclass of a derived class "
                        "must be a (non-strict) subclass "
                        "of the metaclasses of all its bases");
        return NULL;
    }
    if (!metaclass) {
#if PY_MAJOR_VERSION < 3
        metaclass = &PyClass_Type;
#else
        metaclass = &PyType_Type;
#endif
    }
    Py_INCREF((PyObject*) metaclass);
    return (PyObject*) metaclass;
}

/* PyObjectLookupSpecial */
#if CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx__PyObject_LookupSpecial(PyObject* obj, PyObject* attr_name, int with_error) {
    PyObject *res;
    PyTypeObject *tp = Py_TYPE(obj);
#if PY_MAJOR_VERSION < 3
    if (unlikely(PyInstance_Check(obj)))
        return with_error ? __Pyx_PyObject_GetAttrStr(obj, attr_name) : __Pyx_PyObject_GetAttrStrNoError(obj, attr_name);
#endif
    res = _PyType_Lookup(tp, attr_name);
    if (likely(res)) {
        descrgetfunc f = Py_TYPE(res)->tp_descr_get;
        if (!f) {
            Py_INCREF(res);
        } else {
            res = f(res, obj, (PyObject *)tp);
        }
    } else if (with_error) {
        PyErr_SetObject(PyExc_AttributeError, attr_name);
    }
    return res;
}
#endif

/* Py3ClassCreate */
static PyObject *__Pyx_Py3MetaclassPrepare(PyObject *metaclass, PyObject *bases, PyObject *name,
                                           PyObject *qualname, PyObject *mkw, PyObject *modname, PyObject *doc) {
    PyObject *ns;
    if (metaclass) {
        PyObject *prep = __Pyx_PyObject_GetAttrStrNoError(metaclass, __pyx_n_s_prepare);
        if (prep) {
            PyObject *pargs[3] = {NULL, name, bases};
            ns = __Pyx_PyObject_FastCallDict(prep, pargs+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, mkw);
            Py_DECREF(prep);
        } else {
            if (unlikely(PyErr_Occurred()))
                return NULL;
            ns = PyDict_New();
        }
    } else {
        ns = PyDict_New();
    }
    if (unlikely(!ns))
        return NULL;
    if (unlikely(PyObject_SetItem(ns, __pyx_n_s_module, modname) < 0)) goto bad;
#if PY_VERSION_HEX >= 0x03030000
    if (unlikely(PyObject_SetItem(ns, __pyx_n_s_qualname, qualname) < 0)) goto bad;
#else
    CYTHON_MAYBE_UNUSED_VAR(qualname);
#endif
    if (unlikely(doc && PyObject_SetItem(ns, __pyx_n_s_doc, doc) < 0)) goto bad;
    return ns;
bad:
    Py_DECREF(ns);
    return NULL;
}
#if PY_VERSION_HEX < 0x030600A4 && CYTHON_PEP487_INIT_SUBCLASS
static int __Pyx_SetNamesPEP487(PyObject *type_obj) {
    PyTypeObject *type = (PyTypeObject*) type_obj;
    PyObject *names_to_set, *key, *value, *set_name, *tmp;
    Py_ssize_t i = 0;
#if CYTHON_USE_TYPE_SLOTS
    names_to_set = PyDict_Copy(type->tp_dict);
#else
    {
        PyObject *d = PyObject_GetAttr(type_obj, __pyx_n_s_dict_2);
        names_to_set = NULL;
        if (likely(d)) {
            PyObject *names_to_set = PyDict_New();
            int ret = likely(names_to_set) ? PyDict_Update(names_to_set, d) : -1;
            Py_DECREF(d);
            if (unlikely(ret < 0))
                Py_CLEAR(names_to_set);
        }
    }
#endif
    if (unlikely(names_to_set == NULL))
        goto bad;
    while (PyDict_Next(names_to_set, &i, &key, &value)) {
        set_name = __Pyx_PyObject_LookupSpecialNoError(value, __pyx_n_s_set_name);
        if (unlikely(set_name != NULL)) {
            tmp = __Pyx_PyObject_Call2Args(set_name, type_obj, key);
            Py_DECREF(set_name);
            if (unlikely(tmp == NULL)) {
                __Pyx_TypeName value_type_name =
                    __Pyx_PyType_GetName(Py_TYPE(value));
                __Pyx_TypeName type_name = __Pyx_PyType_GetName(type);
                PyErr_Format(PyExc_RuntimeError,
#if PY_MAJOR_VERSION >= 3
                    "Error calling __set_name__ on '" __Pyx_FMT_TYPENAME "' instance %R " "in '" __Pyx_FMT_TYPENAME "'",
                    value_type_name, key, type_name);
#else
                    "Error calling __set_name__ on '" __Pyx_FMT_TYPENAME "' instance %.100s in '" __Pyx_FMT_TYPENAME "'",
                    value_type_name,
                    PyString_Check(key) ? PyString_AS_STRING(key) : "?",
                    type_name);
#endif
                goto bad;
            } else {
                Py_DECREF(tmp);
            }
        }
        else if (unlikely(PyErr_Occurred())) {
            goto bad;
        }
    }
    Py_DECREF(names_to_set);
    return 0;
bad:
    Py_XDECREF(names_to_set);
    return -1;
}
static PyObject *__Pyx_InitSubclassPEP487(PyObject *type_obj, PyObject *mkw) {
#if CYTHON_USE_TYPE_SLOTS && CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    PyTypeObject *type = (PyTypeObject*) type_obj;
    PyObject *mro = type->tp_mro;
    Py_ssize_t i, nbases;
    if (unlikely(!mro)) goto done;
    (void) &__Pyx_GetBuiltinName;
    Py_INCREF(mro);
    nbases = PyTuple_GET_SIZE(mro);
    assert(PyTuple_GET_ITEM(mro, 0) == type_obj);
    for (i = 1; i < nbases-1; i++) {
        PyObject *base, *dict, *meth;
        base = PyTuple_GET_ITEM(mro, i);
        dict = ((PyTypeObject *)base)->tp_dict;
        meth = __Pyx_PyDict_GetItemStrWithError(dict, __pyx_n_s_init_subclass);
        if (unlikely(meth)) {
            descrgetfunc f = Py_TYPE(meth)->tp_descr_get;
            PyObject *res;
            Py_INCREF(meth);
            if (likely(f)) {
                res = f(meth, NULL, type_obj);
                Py_DECREF(meth);
                if (unlikely(!res)) goto bad;
                meth = res;
            }
            res = __Pyx_PyObject_FastCallDict(meth, NULL, 0, mkw);
            Py_DECREF(meth);
            if (unlikely(!res)) goto bad;
            Py_DECREF(res);
            goto done;
        } else if (unlikely(PyErr_Occurred())) {
            goto bad;
        }
    }
done:
    Py_XDECREF(mro);
    return type_obj;
bad:
    Py_XDECREF(mro);
    Py_DECREF(type_obj);
    return NULL;
#else
    PyObject *super_type, *super, *func, *res;
#if CYTHON_COMPILING_IN_PYPY && !defined(PySuper_Type)
    super_type = __Pyx_GetBuiltinName(__pyx_n_s_super);
#else
    super_type = (PyObject*) &PySuper_Type;
    (void) &__Pyx_GetBuiltinName;
#endif
    super = likely(super_type) ? __Pyx_PyObject_Call2Args(super_type, type_obj, type_obj) : NULL;
#if CYTHON_COMPILING_IN_PYPY && !defined(PySuper_Type)
    Py_XDECREF(super_type);
#endif
    if (unlikely(!super)) {
        Py_CLEAR(type_obj);
        goto done;
    }
    func = __Pyx_PyObject_GetAttrStrNoError(super, __pyx_n_s_init_subclass);
    Py_DECREF(super);
    if (likely(!func)) {
        if (unlikely(PyErr_Occurred()))
            Py_CLEAR(type_obj);
        goto done;
    }
    res = __Pyx_PyObject_FastCallDict(func, NULL, 0, mkw);
    Py_DECREF(func);
    if (unlikely(!res))
        Py_CLEAR(type_obj);
    Py_XDECREF(res);
done:
    return type_obj;
#endif
}
#endif
static PyObject *__Pyx_Py3ClassCreate(PyObject *metaclass, PyObject *name, PyObject *bases,
                                      PyObject *dict, PyObject *mkw,
                                      int calculate_metaclass, int allow_py2_metaclass) {
    PyObject *result;
    PyObject *owned_metaclass = NULL;
    PyObject *margs[4] = {NULL, name, bases, dict};
    if (allow_py2_metaclass) {
        owned_metaclass = PyObject_GetItem(dict, __pyx_n_s_metaclass);
        if (owned_metaclass) {
            metaclass = owned_metaclass;
        } else if (likely(PyErr_ExceptionMatches(PyExc_KeyError))) {
            PyErr_Clear();
        } else {
            return NULL;
        }
    }
    if (calculate_metaclass && (!metaclass || PyType_Check(metaclass))) {
        metaclass = __Pyx_CalculateMetaclass((PyTypeObject*) metaclass, bases);
        Py_XDECREF(owned_metaclass);
        if (unlikely(!metaclass))
            return NULL;
        owned_metaclass = metaclass;
    }
    result = __Pyx_PyObject_FastCallDict(metaclass, margs+1, 3 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET,
#if PY_VERSION_HEX < 0x030600A4
        (metaclass == (PyObject*)&PyType_Type) ? NULL : mkw
#else
        mkw
#endif
    );
    Py_XDECREF(owned_metaclass);
#if PY_VERSION_HEX < 0x030600A4 && CYTHON_PEP487_INIT_SUBCLASS
    if (likely(result) && likely(PyType_Check(result))) {
        if (unlikely(__Pyx_SetNamesPEP487(result) < 0)) {
            Py_CLEAR(result);
        } else {
            result = __Pyx_InitSubclassPEP487(result, mkw);
        }
    }
#else
    (void) &__Pyx_GetBuiltinName;
#endif
    return result;
}

/* Py3UpdateBases */
static PyObject*
__Pyx_PEP560_update_bases(PyObject *bases)
{
    Py_ssize_t i, j, size_bases;
    PyObject *base, *meth, *new_base, *result, *new_bases = NULL;
    size_bases = PyTuple_GET_SIZE(bases);
    for (i = 0; i < size_bases; i++) {
        base  = PyTuple_GET_ITEM(bases, i);
        if (PyType_Check(base)) {
            if (new_bases) {
                if (PyList_Append(new_bases, base) < 0) {
                    goto error;
                }
            }
            continue;
        }
        meth = __Pyx_PyObject_GetAttrStrNoError(base, __pyx_n_s_mro_entries);
        if (!meth && PyErr_Occurred()) {
            goto error;
        }
        if (!meth) {
            if (new_bases) {
                if (PyList_Append(new_bases, base) < 0) {
                    goto error;
                }
            }
            continue;
        }
        new_base = __Pyx_PyObject_CallOneArg(meth, bases);
        Py_DECREF(meth);
        if (!new_base) {
            goto error;
        }
        if (!PyTuple_Check(new_base)) {
            PyErr_SetString(PyExc_TypeError,
                            "__mro_entries__ must return a tuple");
            Py_DECREF(new_base);
            goto error;
        }
        if (!new_bases) {
            if (!(new_bases = PyList_New(i))) {
                goto error;
            }
            for (j = 0; j < i; j++) {
                base = PyTuple_GET_ITEM(bases, j);
                PyList_SET_ITEM(new_bases, j, base);
                Py_INCREF(base);
            }
        }
        j = PyList_GET_SIZE(new_bases);
        if (PyList_SetSlice(new_bases, j, j, new_base) < 0) {
            goto error;
        }
        Py_DECREF(new_base);
    }
    if (!new_bases) {
        Py_INCREF(bases);
        return bases;
    }
    result = PyList_AsTuple(new_bases);
    Py_DECREF(new_bases);
    return result;
error:
    Py_XDECREF(new_bases);
    return NULL;
}

/* CLineInTraceback */
#ifndef CYTHON_CLINE_IN_TRACEBACK
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line) {
    PyObject *use_cline;
    PyObject *ptype, *pvalue, *ptraceback;
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject **cython_runtime_dict;
#endif
    CYTHON_MAYBE_UNUSED_VAR(tstate);
    if (unlikely(!__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
#if CYTHON_COMPILING_IN_CPYTHON
    cython_runtime_dict = _PyObject_GetDictPtr(__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
        __PYX_PY_DICT_LOOKUP_IF_MODIFIED(
            use_cline, *cython_runtime_dict,
            __Pyx_PyDict_GetItemStr(*cython_runtime_dict, __pyx_n_s_cline_in_traceback))
    } else
#endif
    {
      PyObject *use_cline_obj = __Pyx_PyObject_GetAttrStrNoError(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback);
      if (use_cline_obj) {
        use_cline = PyObject_Not(use_cline_obj) ? Py_False : Py_True;
        Py_DECREF(use_cline_obj);
      } else {
        PyErr_Clear();
        use_cline = NULL;
      }
    }
    if (!use_cline) {
        c_line = 0;
        (void) PyObject_SetAttr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback, Py_False);
    }
    else if (use_cline == Py_False || (use_cline != Py_True && PyObject_Not(use_cline) != 0)) {
        c_line = 0;
    }
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache */
#if !CYTHON_COMPILING_IN_LIMITED_API
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, ((size_t)new_max) * sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}
#endif

/* AddTraceback */
#include "compile.h"
#include "frameobject.h"
#include "traceback.h"
#if PY_VERSION_HEX >= 0x030b00a6
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    if (c_line) {
        (void) __pyx_cfilenm;
        (void) __Pyx_CLineForTraceback(__Pyx_PyThreadState_Current, c_line);
    }
    _PyTraceback_Add(funcname, filename, py_line);
}
#else
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = NULL;
    PyObject *py_funcname = NULL;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_srcfile = NULL;
    py_srcfile = PyString_FromString(filename);
    if (!py_srcfile) goto bad;
    #endif
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        if (!py_funcname) goto bad;
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        if (!py_funcname) goto bad;
        funcname = PyUnicode_AsUTF8(py_funcname);
        if (!funcname) goto bad;
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        if (!py_funcname) goto bad;
        #endif
    }
    #if PY_MAJOR_VERSION < 3
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    #else
    py_code = PyCode_NewEmpty(filename, funcname, py_line);
    #endif
    Py_XDECREF(py_funcname);  // XDECREF since it's only set on Py3 if cline
    return py_code;
bad:
    Py_XDECREF(py_funcname);
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_srcfile);
    #endif
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject *ptype, *pvalue, *ptraceback;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) {
            /* If the code object creation fails, then we should clear the
               fetched exception references and propagate the new exception */
            Py_XDECREF(ptype);
            Py_XDECREF(pvalue);
            Py_XDECREF(ptraceback);
            goto bad;
        }
        __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}
#endif

/* CIntFromPyVerify */
#define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_readstat_type_t(readstat_type_t value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const readstat_type_t neg_one = (readstat_type_t) -1, const_zero = (readstat_type_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(readstat_type_t) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(readstat_type_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(readstat_type_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(readstat_type_t) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(readstat_type_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(readstat_type_t),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
    }
}

/* CIntFromPy */
static CYTHON_INLINE readstat_type_t __Pyx_PyInt_As_readstat_type_t(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const readstat_type_t neg_one = (readstat_type_t) -1, const_zero = (readstat_type_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if ((sizeof(readstat_type_t) < sizeof(long))) {
            __PYX_VERIFY_RETURN_INT(readstat_type_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (readstat_type_t) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            if (unlikely(__Pyx_PyLong_IsNeg(x))) {
                goto raise_neg_overflow;
            } else if (__Pyx_PyLong_IsCompact(x)) {
                __PYX_VERIFY_RETURN_INT(readstat_type_t, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
            } else {
                const digit* digits = __Pyx_PyLong_Digits(x);
                assert(__Pyx_PyLong_DigitCount(x) > 1);
                switch (__Pyx_PyLong_DigitCount(x)) {
                    case 2:
                        if ((8 * sizeof(readstat_type_t) > 1 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(readstat_type_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(readstat_type_t) >= 2 * PyLong_SHIFT)) {
                                return (readstat_type_t) (((((readstat_type_t)digits[1]) << PyLong_SHIFT) | (readstat_type_t)digits[0]));
                            }
                        }
                        break;
                    case 3:
                        if ((8 * sizeof(readstat_type_t) > 2 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(readstat_type_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(readstat_type_t) >= 3 * PyLong_SHIFT)) {
                                return (readstat_type_t) (((((((readstat_type_t)digits[2]) << PyLong_SHIFT) | (readstat_type_t)digits[1]) << PyLong_SHIFT) | (readstat_type_t)digits[0]));
                            }
                        }
                        break;
                    case 4:
                        if ((8 * sizeof(readstat_type_t) > 3 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(readstat_type_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(readstat_type_t) >= 4 * PyLong_SHIFT)) {
                                return (readstat_type_t) (((((((((readstat_type_t)digits[3]) << PyLong_SHIFT) | (readstat_type_t)digits[2]) << PyLong_SHIFT) | (readstat_type_t)digits[1]) << PyLong_SHIFT) | (readstat_type_t)digits[0]));
                            }
                        }
                        break;
                }
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (readstat_type_t) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if ((sizeof(readstat_type_t) <= sizeof(unsigned long))) {
                __PYX_VERIFY_RETURN_INT_EXC(readstat_type_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if ((sizeof(readstat_type_t) <= sizeof(unsigned PY_LONG_LONG))) {
                __PYX_VERIFY_RETURN_INT_EXC(readstat_type_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            if (__Pyx_PyLong_IsCompact(x)) {
                __PYX_VERIFY_RETURN_INT(readstat_type_t, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
            } else {
                const digit* digits = __Pyx_PyLong_Digits(x);
                assert(__Pyx_PyLong_DigitCount(x) > 1);
                switch (__Pyx_PyLong_SignedDigitCount(x)) {
                    case -2:
                        if ((8 * sizeof(readstat_type_t) - 1 > 1 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(readstat_type_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(readstat_type_t) - 1 > 2 * PyLong_SHIFT)) {
                                return (readstat_type_t) (((readstat_type_t)-1)*(((((readstat_type_t)digits[1]) << PyLong_SHIFT) | (readstat_type_t)digits[0])));
                            }
                        }
                        break;
                    case 2:
                        if ((8 * sizeof(readstat_type_t) > 1 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(readstat_type_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(readstat_type_t) - 1 > 2 * PyLong_SHIFT)) {
                                return (readstat_type_t) ((((((readstat_type_t)digits[1]) << PyLong_SHIFT) | (readstat_type_t)digits[0])));
                            }
                        }
                        break;
                    case -3:
                        if ((8 * sizeof(readstat_type_t) - 1 > 2 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(readstat_type_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(readstat_type_t) - 1 > 3 * PyLong_SHIFT)) {
                                return (readstat_type_t) (((readstat_type_t)-1)*(((((((readstat_type_t)digits[2]) << PyLong_SHIFT) | (readstat_type_t)digits[1]) << PyLong_SHIFT) | (readstat_type_t)digits[0])));
                            }
                        }
                        break;
                    case 3:
                        if ((8 * sizeof(readstat_type_t) > 2 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(readstat_type_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(readstat_type_t) - 1 > 3 * PyLong_SHIFT)) {
                                return (readstat_type_t) ((((((((readstat_type_t)digits[2]) << PyLong_SHIFT) | (readstat_type_t)digits[1]) << PyLong_SHIFT) | (readstat_type_t)digits[0])));
                            }
                        }
                        break;
                    case -4:
                        if ((8 * sizeof(readstat_type_t) - 1 > 3 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(readstat_type_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(readstat_type_t) - 1 > 4 * PyLong_SHIFT)) {
                                return (readstat_type_t) (((readstat_type_t)-1)*(((((((((readstat_type_t)digits[3]) << PyLong_SHIFT) | (readstat_type_t)digits[2]) << PyLong_SHIFT) | (readstat_type_t)digits[1]) << PyLong_SHIFT) | (readstat_type_t)digits[0])));
                            }
                        }
                        break;
                    case 4:
                        if ((8 * sizeof(readstat_type_t) > 3 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(readstat_type_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(readstat_type_t) - 1 > 4 * PyLong_SHIFT)) {
                                return (readstat_type_t) ((((((((((readstat_type_t)digits[3]) << PyLong_SHIFT) | (readstat_type_t)digits[2]) << PyLong_SHIFT) | (readstat_type_t)digits[1]) << PyLong_SHIFT) | (readstat_type_t)digits[0])));
                            }
                        }
                        break;
                }
            }
#endif
            if ((sizeof(readstat_type_t) <= sizeof(long))) {
                __PYX_VERIFY_RETURN_INT_EXC(readstat_type_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if ((sizeof(readstat_type_t) <= sizeof(PY_LONG_LONG))) {
                __PYX_VERIFY_RETURN_INT_EXC(readstat_type_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        PyErr_SetString(PyExc_RuntimeError,
                        "_PyLong_AsByteArray() not available, cannot convert large enums");
        return (readstat_type_t) -1;
    } else {
        readstat_type_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (readstat_type_t) -1;
        val = __Pyx_PyInt_As_readstat_type_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to readstat_type_t");
    return (readstat_type_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to readstat_type_t");
    return (readstat_type_t) -1;
}

/* CIntFromPy */
static CYTHON_INLINE __pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format __Pyx_PyInt_As___pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const __pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format neg_one = (__pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format) -1, const_zero = (__pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if ((sizeof(__pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format) < sizeof(long))) {
            __PYX_VERIFY_RETURN_INT(__pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (__pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            if (unlikely(__Pyx_PyLong_IsNeg(x))) {
                goto raise_neg_overflow;
            } else if (__Pyx_PyLong_IsCompact(x)) {
                __PYX_VERIFY_RETURN_INT(__pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
            } else {
                const digit* digits = __Pyx_PyLong_Digits(x);
                assert(__Pyx_PyLong_DigitCount(x) > 1);
                switch (__Pyx_PyLong_DigitCount(x)) {
                    case 2:
                        if ((8 * sizeof(__pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format) > 1 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(__pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(__pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format) >= 2 * PyLong_SHIFT)) {
                                return (__pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format) (((((__pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format)digits[1]) << PyLong_SHIFT) | (__pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format)digits[0]));
                            }
                        }
                        break;
                    case 3:
                        if ((8 * sizeof(__pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format) > 2 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(__pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(__pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format) >= 3 * PyLong_SHIFT)) {
                                return (__pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format) (((((((__pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format)digits[2]) << PyLong_SHIFT) | (__pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format)digits[1]) << PyLong_SHIFT) | (__pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format)digits[0]));
                            }
                        }
                        break;
                    case 4:
                        if ((8 * sizeof(__pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format) > 3 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(__pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(__pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format) >= 4 * PyLong_SHIFT)) {
                                return (__pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format) (((((((((__pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format)digits[3]) << PyLong_SHIFT) | (__pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format)digits[2]) << PyLong_SHIFT) | (__pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format)digits[1]) << PyLong_SHIFT) | (__pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format)digits[0]));
                            }
                        }
                        break;
                }
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (__pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if ((sizeof(__pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format) <= sizeof(unsigned long))) {
                __PYX_VERIFY_RETURN_INT_EXC(__pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if ((sizeof(__pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format) <= sizeof(unsigned PY_LONG_LONG))) {
                __PYX_VERIFY_RETURN_INT_EXC(__pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            if (__Pyx_PyLong_IsCompact(x)) {
                __PYX_VERIFY_RETURN_INT(__pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
            } else {
                const digit* digits = __Pyx_PyLong_Digits(x);
                assert(__Pyx_PyLong_DigitCount(x) > 1);
                switch (__Pyx_PyLong_SignedDigitCount(x)) {
                    case -2:
                        if ((8 * sizeof(__pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format) - 1 > 1 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(__pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(__pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format) - 1 > 2 * PyLong_SHIFT)) {
                                return (__pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format) (((__pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format)-1)*(((((__pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format)digits[1]) << PyLong_SHIFT) | (__pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format)digits[0])));
                            }
                        }
                        break;
                    case 2:
                        if ((8 * sizeof(__pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format) > 1 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(__pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(__pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format) - 1 > 2 * PyLong_SHIFT)) {
                                return (__pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format) ((((((__pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format)digits[1]) << PyLong_SHIFT) | (__pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format)digits[0])));
                            }
                        }
                        break;
                    case -3:
                        if ((8 * sizeof(__pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format) - 1 > 2 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(__pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(__pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format) - 1 > 3 * PyLong_SHIFT)) {
                                return (__pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format) (((__pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format)-1)*(((((((__pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format)digits[2]) << PyLong_SHIFT) | (__pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format)digits[1]) << PyLong_SHIFT) | (__pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format)digits[0])));
                            }
                        }
                        break;
                    case 3:
                        if ((8 * sizeof(__pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format) > 2 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(__pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(__pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format) - 1 > 3 * PyLong_SHIFT)) {
                                return (__pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format) ((((((((__pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format)digits[2]) << PyLong_SHIFT) | (__pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format)digits[1]) << PyLong_SHIFT) | (__pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format)digits[0])));
                            }
                        }
                        break;
                    case -4:
                        if ((8 * sizeof(__pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format) - 1 > 3 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(__pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(__pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format) - 1 > 4 * PyLong_SHIFT)) {
                                return (__pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format) (((__pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format)-1)*(((((((((__pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format)digits[3]) << PyLong_SHIFT) | (__pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format)digits[2]) << PyLong_SHIFT) | (__pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format)digits[1]) << PyLong_SHIFT) | (__pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format)digits[0])));
                            }
                        }
                        break;
                    case 4:
                        if ((8 * sizeof(__pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format) > 3 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(__pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(__pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format) - 1 > 4 * PyLong_SHIFT)) {
                                return (__pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format) ((((((((((__pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format)digits[3]) << PyLong_SHIFT) | (__pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format)digits[2]) << PyLong_SHIFT) | (__pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format)digits[1]) << PyLong_SHIFT) | (__pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format)digits[0])));
                            }
                        }
                        break;
                }
            }
#endif
            if ((sizeof(__pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format) <= sizeof(long))) {
                __PYX_VERIFY_RETURN_INT_EXC(__pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if ((sizeof(__pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format) <= sizeof(PY_LONG_LONG))) {
                __PYX_VERIFY_RETURN_INT_EXC(__pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        PyErr_SetString(PyExc_RuntimeError,
                        "_PyLong_AsByteArray() not available, cannot convert large enums");
        return (__pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format) -1;
    } else {
        __pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (__pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format) -1;
        val = __Pyx_PyInt_As___pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to __pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format");
    return (__pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to __pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format");
    return (__pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format) -1;
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_char(char value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const char neg_one = (char) -1, const_zero = (char) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(char) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(char) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(char) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(char) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(char) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(char),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From___pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format(__pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const __pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format neg_one = (__pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format) -1, const_zero = (__pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(__pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(__pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(__pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(__pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(__pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(__pyx_t_10pyreadstat_16_readstat_parser_py_datetime_format),
                                     little, !is_unsigned);
    }
}

/* CIntFromPy */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if ((sizeof(long) < sizeof(long))) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            if (unlikely(__Pyx_PyLong_IsNeg(x))) {
                goto raise_neg_overflow;
            } else if (__Pyx_PyLong_IsCompact(x)) {
                __PYX_VERIFY_RETURN_INT(long, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
            } else {
                const digit* digits = __Pyx_PyLong_Digits(x);
                assert(__Pyx_PyLong_DigitCount(x) > 1);
                switch (__Pyx_PyLong_DigitCount(x)) {
                    case 2:
                        if ((8 * sizeof(long) > 1 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(long) >= 2 * PyLong_SHIFT)) {
                                return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                            }
                        }
                        break;
                    case 3:
                        if ((8 * sizeof(long) > 2 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(long) >= 3 * PyLong_SHIFT)) {
                                return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                            }
                        }
                        break;
                    case 4:
                        if ((8 * sizeof(long) > 3 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(long) >= 4 * PyLong_SHIFT)) {
                                return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                            }
                        }
                        break;
                }
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (long) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if ((sizeof(long) <= sizeof(unsigned long))) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if ((sizeof(long) <= sizeof(unsigned PY_LONG_LONG))) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            if (__Pyx_PyLong_IsCompact(x)) {
                __PYX_VERIFY_RETURN_INT(long, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
            } else {
                const digit* digits = __Pyx_PyLong_Digits(x);
                assert(__Pyx_PyLong_DigitCount(x) > 1);
                switch (__Pyx_PyLong_SignedDigitCount(x)) {
                    case -2:
                        if ((8 * sizeof(long) - 1 > 1 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(long) - 1 > 2 * PyLong_SHIFT)) {
                                return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                            }
                        }
                        break;
                    case 2:
                        if ((8 * sizeof(long) > 1 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(long) - 1 > 2 * PyLong_SHIFT)) {
                                return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                            }
                        }
                        break;
                    case -3:
                        if ((8 * sizeof(long) - 1 > 2 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(long) - 1 > 3 * PyLong_SHIFT)) {
                                return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                            }
                        }
                        break;
                    case 3:
                        if ((8 * sizeof(long) > 2 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(long) - 1 > 3 * PyLong_SHIFT)) {
                                return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                            }
                        }
                        break;
                    case -4:
                        if ((8 * sizeof(long) - 1 > 3 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(long) - 1 > 4 * PyLong_SHIFT)) {
                                return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                            }
                        }
                        break;
                    case 4:
                        if ((8 * sizeof(long) > 3 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(long) - 1 > 4 * PyLong_SHIFT)) {
                                return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                            }
                        }
                        break;
                }
            }
#endif
            if ((sizeof(long) <= sizeof(long))) {
                __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if ((sizeof(long) <= sizeof(PY_LONG_LONG))) {
                __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
            long val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
#if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
#endif
            if (likely(v)) {
                int ret = -1;
#if !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                ret = _PyLong_AsByteArray((PyLongObject *)v,
                                           bytes, sizeof(val),
                                           is_little, !is_unsigned);
#else
                PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
                int bits, remaining_bits, is_negative = 0;
                long idigit;
                int chunk_size = (sizeof(long) < 8) ? 30 : 62;
                if (unlikely(!PyLong_CheckExact(v))) {
                    PyObject *tmp = v;
                    v = PyNumber_Long(v);
                    assert(PyLong_CheckExact(v));
                    Py_DECREF(tmp);
                    if (unlikely(!v)) return (long) -1;
                }
#if CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030B0000
                if (Py_SIZE(x) == 0)
                    return (long) 0;
                is_negative = Py_SIZE(x) < 0;
#else
                {
                    int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                    if (unlikely(result < 0))
                        return (long) -1;
                    is_negative = result == 1;
                }
#endif
                if (is_unsigned && unlikely(is_negative)) {
                    goto raise_neg_overflow;
                } else if (is_negative) {
                    stepval = PyNumber_Invert(v);
                    if (unlikely(!stepval))
                        return (long) -1;
                } else {
                    stepval = __Pyx_NewRef(v);
                }
                val = (long) 0;
                mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
                shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
                for (bits = 0; bits < (int) sizeof(long) * 8 - chunk_size; bits += chunk_size) {
                    PyObject *tmp, *digit;
                    digit = PyNumber_And(stepval, mask);
                    if (unlikely(!digit)) goto done;
                    idigit = PyLong_AsLong(digit);
                    Py_DECREF(digit);
                    if (unlikely(idigit < 0)) goto done;
                    tmp = PyNumber_Rshift(stepval, shift);
                    if (unlikely(!tmp)) goto done;
                    Py_DECREF(stepval); stepval = tmp;
                    val |= ((long) idigit) << bits;
                    #if CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030B0000
                    if (Py_SIZE(stepval) == 0)
                        goto unpacking_done;
                    #endif
                }
                idigit = PyLong_AsLong(stepval);
                if (unlikely(idigit < 0)) goto done;
                remaining_bits = ((int) sizeof(long) * 8) - bits - (is_unsigned ? 0 : 1);
                if (unlikely(idigit >= (1L << remaining_bits)))
                    goto raise_overflow;
                val |= ((long) idigit) << bits;
            #if CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030B0000
            unpacking_done:
            #endif
                if (!is_unsigned) {
                    if (unlikely(val & (((long) 1) << (sizeof(long) * 8 - 1))))
                        goto raise_overflow;
                    if (is_negative)
                        val = ~val;
                }
                ret = 0;
            done:
                Py_XDECREF(shift);
                Py_XDECREF(mask);
                Py_XDECREF(stepval);
#endif
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
            return (long) -1;
        }
    } else {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* CIntFromPy */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if ((sizeof(int) < sizeof(long))) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            if (unlikely(__Pyx_PyLong_IsNeg(x))) {
                goto raise_neg_overflow;
            } else if (__Pyx_PyLong_IsCompact(x)) {
                __PYX_VERIFY_RETURN_INT(int, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
            } else {
                const digit* digits = __Pyx_PyLong_Digits(x);
                assert(__Pyx_PyLong_DigitCount(x) > 1);
                switch (__Pyx_PyLong_DigitCount(x)) {
                    case 2:
                        if ((8 * sizeof(int) > 1 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(int) >= 2 * PyLong_SHIFT)) {
                                return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                            }
                        }
                        break;
                    case 3:
                        if ((8 * sizeof(int) > 2 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(int) >= 3 * PyLong_SHIFT)) {
                                return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                            }
                        }
                        break;
                    case 4:
                        if ((8 * sizeof(int) > 3 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(int) >= 4 * PyLong_SHIFT)) {
                                return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                            }
                        }
                        break;
                }
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if ((sizeof(int) <= sizeof(unsigned long))) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if ((sizeof(int) <= sizeof(unsigned PY_LONG_LONG))) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            if (__Pyx_PyLong_IsCompact(x)) {
                __PYX_VERIFY_RETURN_INT(int, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
            } else {
                const digit* digits = __Pyx_PyLong_Digits(x);
                assert(__Pyx_PyLong_DigitCount(x) > 1);
                switch (__Pyx_PyLong_SignedDigitCount(x)) {
                    case -2:
                        if ((8 * sizeof(int) - 1 > 1 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(int) - 1 > 2 * PyLong_SHIFT)) {
                                return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                            }
                        }
                        break;
                    case 2:
                        if ((8 * sizeof(int) > 1 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(int) - 1 > 2 * PyLong_SHIFT)) {
                                return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                            }
                        }
                        break;
                    case -3:
                        if ((8 * sizeof(int) - 1 > 2 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(int) - 1 > 3 * PyLong_SHIFT)) {
                                return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                            }
                        }
                        break;
                    case 3:
                        if ((8 * sizeof(int) > 2 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(int) - 1 > 3 * PyLong_SHIFT)) {
                                return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                            }
                        }
                        break;
                    case -4:
                        if ((8 * sizeof(int) - 1 > 3 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(int) - 1 > 4 * PyLong_SHIFT)) {
                                return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                            }
                        }
                        break;
                    case 4:
                        if ((8 * sizeof(int) > 3 * PyLong_SHIFT)) {
                            if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                                __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                            } else if ((8 * sizeof(int) - 1 > 4 * PyLong_SHIFT)) {
                                return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                            }
                        }
                        break;
                }
            }
#endif
            if ((sizeof(int) <= sizeof(long))) {
                __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if ((sizeof(int) <= sizeof(PY_LONG_LONG))) {
                __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
            int val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
#if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
#endif
            if (likely(v)) {
                int ret = -1;
#if !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                ret = _PyLong_AsByteArray((PyLongObject *)v,
                                           bytes, sizeof(val),
                                           is_little, !is_unsigned);
#else
                PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
                int bits, remaining_bits, is_negative = 0;
                long idigit;
                int chunk_size = (sizeof(long) < 8) ? 30 : 62;
                if (unlikely(!PyLong_CheckExact(v))) {
                    PyObject *tmp = v;
                    v = PyNumber_Long(v);
                    assert(PyLong_CheckExact(v));
                    Py_DECREF(tmp);
                    if (unlikely(!v)) return (int) -1;
                }
#if CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030B0000
                if (Py_SIZE(x) == 0)
                    return (int) 0;
                is_negative = Py_SIZE(x) < 0;
#else
                {
                    int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                    if (unlikely(result < 0))
                        return (int) -1;
                    is_negative = result == 1;
                }
#endif
                if (is_unsigned && unlikely(is_negative)) {
                    goto raise_neg_overflow;
                } else if (is_negative) {
                    stepval = PyNumber_Invert(v);
                    if (unlikely(!stepval))
                        return (int) -1;
                } else {
                    stepval = __Pyx_NewRef(v);
                }
                val = (int) 0;
                mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
                shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
                for (bits = 0; bits < (int) sizeof(int) * 8 - chunk_size; bits += chunk_size) {
                    PyObject *tmp, *digit;
                    digit = PyNumber_And(stepval, mask);
                    if (unlikely(!digit)) goto done;
                    idigit = PyLong_AsLong(digit);
                    Py_DECREF(digit);
                    if (unlikely(idigit < 0)) goto done;
                    tmp = PyNumber_Rshift(stepval, shift);
                    if (unlikely(!tmp)) goto done;
                    Py_DECREF(stepval); stepval = tmp;
                    val |= ((int) idigit) << bits;
                    #if CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030B0000
                    if (Py_SIZE(stepval) == 0)
                        goto unpacking_done;
                    #endif
                }
                idigit = PyLong_AsLong(stepval);
                if (unlikely(idigit < 0)) goto done;
                remaining_bits = ((int) sizeof(int) * 8) - bits - (is_unsigned ? 0 : 1);
                if (unlikely(idigit >= (1L << remaining_bits)))
                    goto raise_overflow;
                val |= ((int) idigit) << bits;
            #if CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030B0000
            unpacking_done:
            #endif
                if (!is_unsigned) {
                    if (unlikely(val & (((int) 1) << (sizeof(int) * 8 - 1))))
                        goto raise_overflow;
                    if (is_negative)
                        val = ~val;
                }
                ret = 0;
            done:
                Py_XDECREF(shift);
                Py_XDECREF(mask);
                Py_XDECREF(stepval);
#endif
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
            return (int) -1;
        }
    } else {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* FormatTypeName */
#if CYTHON_COMPILING_IN_LIMITED_API
static __Pyx_TypeName
__Pyx_PyType_GetName(PyTypeObject* tp)
{
    PyObject *name = __Pyx_PyObject_GetAttrStr((PyObject *)tp,
                                               __pyx_n_s_name_2);
    if (unlikely(name == NULL) || unlikely(!PyUnicode_Check(name))) {
        PyErr_Clear();
        Py_XSETREF(name, __Pyx_NewRef(__pyx_n_s__15));
    }
    return name;
}
#endif

/* FastTypeChecks */
#if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = __Pyx_PyType_GetSlot(a, tp_base, PyTypeObject*);
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
static CYTHON_INLINE int __Pyx_IsAnySubtype2(PyTypeObject *cls, PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (cls == a || cls == b) return 1;
    mro = cls->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            PyObject *base = PyTuple_GET_ITEM(mro, i);
            if (base == (PyObject *)a || base == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(cls, a) || __Pyx_InBases(cls, b);
}
#if PY_MAJOR_VERSION == 2
static int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject* exc_type2) {
    PyObject *exception, *value, *tb;
    int res;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&exception, &value, &tb);
    res = exc_type1 ? PyObject_IsSubclass(err, exc_type1) : 0;
    if (unlikely(res == -1)) {
        PyErr_WriteUnraisable(err);
        res = 0;
    }
    if (!res) {
        res = PyObject_IsSubclass(err, exc_type2);
        if (unlikely(res == -1)) {
            PyErr_WriteUnraisable(err);
            res = 0;
        }
    }
    __Pyx_ErrRestore(exception, value, tb);
    return res;
}
#else
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    if (exc_type1) {
        return __Pyx_IsAnySubtype2((PyTypeObject*)err, (PyTypeObject*)exc_type1, (PyTypeObject*)exc_type2);
    } else {
        return __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
}
#endif
static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    assert(PyExceptionClass_Check(exc_type));
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        PyObject *t = PyTuple_GET_ITEM(tuple, i);
        #if PY_MAJOR_VERSION < 3
        if (likely(exc_type == t)) return 1;
        #endif
        if (likely(PyExceptionClass_Check(t))) {
            if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
        } else {
        }
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        if (likely(PyExceptionClass_Check(exc_type))) {
            return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
        } else if (likely(PyTuple_Check(exc_type))) {
            return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
        } else {
        }
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    assert(PyExceptionClass_Check(exc_type1));
    assert(PyExceptionClass_Check(exc_type2));
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* CheckBinaryVersion */
static int __Pyx_check_binary_version(void) {
    char ctversion[5];
    int same=1, i, found_dot;
    const char* rt_from_call = Py_GetVersion();
    PyOS_snprintf(ctversion, 5, "%d.%d", PY_MAJOR_VERSION, PY_MINOR_VERSION);
    found_dot = 0;
    for (i = 0; i < 4; i++) {
        if (!ctversion[i]) {
            same = (rt_from_call[i] < '0' || rt_from_call[i] > '9');
            break;
        }
        if (rt_from_call[i] != ctversion[i]) {
            same = 0;
            break;
        }
    }
    if (!same) {
        char rtversion[5] = {'\0'};
        char message[200];
        for (i=0; i<4; ++i) {
            if (rt_from_call[i] == '.') {
                if (found_dot) break;
                found_dot = 1;
            } else if (rt_from_call[i] < '0' || rt_from_call[i] > '9') {
                break;
            }
            rtversion[i] = rt_from_call[i];
        }
        PyOS_snprintf(message, sizeof(message),
                      "compile time version %s of module '%.100s' "
                      "does not match runtime version %s",
                      ctversion, __Pyx_MODULE_NAME, rtversion);
        return PyErr_WarnEx(NULL, message, 1);
    }
    return 0;
}

/* VoidPtrExport */
static int __Pyx_ExportVoidPtr(PyObject *name, void *p, const char *sig) {
    PyObject *d;
    PyObject *cobj = 0;
    d = PyDict_GetItem(__pyx_d, __pyx_n_s_pyx_capi);
    Py_XINCREF(d);
    if (!d) {
        d = PyDict_New();
        if (!d)
            goto bad;
        if (__Pyx_PyObject_SetAttrStr(__pyx_m, __pyx_n_s_pyx_capi, d) < 0)
            goto bad;
    }
    cobj = PyCapsule_New(p, sig, 0);
    if (!cobj)
        goto bad;
    if (PyDict_SetItem(d, name, cobj) < 0)
        goto bad;
    Py_DECREF(cobj);
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(cobj);
    Py_XDECREF(d);
    return -1;
}

/* FunctionExport */
static int __Pyx_ExportFunction(const char *name, void (*f)(void), const char *sig) {
    PyObject *d = 0;
    PyObject *cobj = 0;
    union {
        void (*fp)(void);
        void *p;
    } tmp;
    d = PyObject_GetAttrString(__pyx_m, (char *)"__pyx_capi__");
    if (!d) {
        PyErr_Clear();
        d = PyDict_New();
        if (!d)
            goto bad;
        Py_INCREF(d);
        if (PyModule_AddObject(__pyx_m, (char *)"__pyx_capi__", d) < 0)
            goto bad;
    }
    tmp.fp = f;
    cobj = PyCapsule_New(tmp.p, sig, 0);
    if (!cobj)
        goto bad;
    if (PyDict_SetItemString(d, name, cobj) < 0)
        goto bad;
    Py_DECREF(cobj);
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(cobj);
    Py_XDECREF(d);
    return -1;
}

/* InitStrings */
#if PY_MAJOR_VERSION >= 3
static int __Pyx_InitString(__Pyx_StringTabEntry t, PyObject **str) {
    if (t.is_unicode | t.is_str) {
        if (t.intern) {
            *str = PyUnicode_InternFromString(t.s);
        } else if (t.encoding) {
            *str = PyUnicode_Decode(t.s, t.n - 1, t.encoding, NULL);
        } else {
            *str = PyUnicode_FromStringAndSize(t.s, t.n - 1);
        }
    } else {
        *str = PyBytes_FromStringAndSize(t.s, t.n - 1);
    }
    if (!*str)
        return -1;
    if (PyObject_Hash(*str) == -1)
        return -1;
    return 0;
}
#endif
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION >= 3
        __Pyx_InitString(*t, t->p);
        #else
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        if (!*t->p)
            return -1;
        if (PyObject_Hash(*t->p) == -1)
            return -1;
        #endif
        ++t;
    }
    return 0;
}

static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    return __Pyx_PyUnicode_FromStringAndSize(c_str, (Py_ssize_t)strlen(c_str));
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
#if !CYTHON_PEP393_ENABLED
static const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    char* defenc_c;
    PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
    if (!defenc) return NULL;
    defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    {
        char* end = defenc_c + PyBytes_GET_SIZE(defenc);
        char* c;
        for (c = defenc_c; c < end; c++) {
            if ((unsigned char) (*c) >= 128) {
                PyUnicode_AsASCIIString(o);
                return NULL;
            }
        }
    }
#endif
    *length = PyBytes_GET_SIZE(defenc);
    return defenc_c;
}
#else
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
}
#endif
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY && !CYTHON_COMPILING_IN_LIMITED_API) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject* x) {
    int retval;
    if (unlikely(!x)) return -1;
    retval = __Pyx_PyObject_IsTrue(x);
    Py_DECREF(x);
    return retval;
}
static PyObject* __Pyx_PyNumber_IntOrLongWrongResultType(PyObject* result, const char* type_name) {
    __Pyx_TypeName result_type_name = __Pyx_PyType_GetName(Py_TYPE(result));
#if PY_MAJOR_VERSION >= 3
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type " __Pyx_FMT_TYPENAME ").  "
                "The ability to return an instance of a strict subclass of int is deprecated, "
                "and may be removed in a future version of Python.",
                result_type_name)) {
            __Pyx_DECREF_TypeName(result_type_name);
            Py_DECREF(result);
            return NULL;
        }
        __Pyx_DECREF_TypeName(result_type_name);
        return result;
    }
#endif
    PyErr_Format(PyExc_TypeError,
                 "__%.4s__ returned non-%.4s (type " __Pyx_FMT_TYPENAME ")",
                 type_name, type_name, result_type_name);
    __Pyx_DECREF_TypeName(result_type_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_Check(x) || PyLong_Check(x)))
#else
  if (likely(PyLong_Check(x)))
#endif
    return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  #if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = m->nb_int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = m->nb_long(x);
  }
  #else
  if (likely(m && m->nb_int)) {
    name = "int";
    res = m->nb_int(x);
  }
  #endif
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
    res = PyNumber_Int(x);
  }
#endif
  if (likely(res)) {
#if PY_MAJOR_VERSION < 3
    if (unlikely(!PyInt_Check(res) && !PyLong_Check(res))) {
#else
    if (unlikely(!PyLong_CheckExact(res))) {
#endif
        return __Pyx_PyNumber_IntOrLongWrongResultType(res, name);
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(b);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(__Pyx_PyLong_IsCompact(b))) {
        return __Pyx_PyLong_CompactValue(b);
    } else {
      const digit* digits = __Pyx_PyLong_Digits(b);
      const Py_ssize_t size = __Pyx_PyLong_SignedDigitCount(b);
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject* o) {
  if (sizeof(Py_hash_t) == sizeof(Py_ssize_t)) {
    return (Py_hash_t) __Pyx_PyIndex_AsSsize_t(o);
#if PY_MAJOR_VERSION < 3
  } else if (likely(PyInt_CheckExact(o))) {
    return PyInt_AS_LONG(o);
#endif
  } else {
    Py_ssize_t ival;
    PyObject *x;
    x = PyNumber_Index(o);
    if (!x) return -1;
    ival = PyInt_AsLong(x);
    Py_DECREF(x);
    return ival;
  }
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return b ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


/* #### Code section: utility_code_pragmas_end ### */
#ifdef _MSC_VER
#pragma warning( pop )
#endif



/* #### Code section: end ### */
#endif /* Py_PYTHON_H */
